// cpp-translate-file.cc

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

#include "stella/stella-system.hh"

namespace stella {

DEFINE_STELLA_SPECIAL(oCURRENT_STREAMo, Output_Stream* , NULL);

char* oCPP_TRANSLATED_FILE_SUFFIXo = "";

void cpptrans(Object* statement) {
  // Translate `statement' to C++ and print the result.
  { 
    BIND_STELLA_SPECIAL(oTRANSLATOROUTPUTLANGUAGEo, Keyword*, oTRANSLATOROUTPUTLANGUAGEo.get());
    BIND_STELLA_SPECIAL(oCURRENT_STREAMo, Output_Stream*, STANDARD_OUTPUT);
    set_translator_output_language(KWD_CPP_TRANSLATE_FILE_CPP);
    incrementally_translate(statement);
  }
}

char* cpp_make_code_output_file_name(char* filename) {
  return (make_file_name(filename, KWD_CPP_TRANSLATE_FILE_CPP_CODE, FALSE));
}

char* cpp_make_header_output_file_name(char* filename) {
  return (make_file_name(filename, KWD_CPP_TRANSLATE_FILE_CPP_HEADER, FALSE));
}

void cpp_output_include_directive(char* filename, boolean stripdirectoryP) {
  if (stripdirectoryP) {
    filename = file_name_without_directory(filename);
  }
  *(oCURRENT_STREAMo.get()->native_stream) << "#include \"" << filename << "\"" << endl;
}

void cpp_standalone_output_import_declarations(char* headerfilename) {
  cpp_output_include_directive(make_file_name("cpp-standalone", KWD_CPP_TRANSLATE_FILE_CPP_HEADER, TRUE), TRUE);
  cpp_output_include_directive(headerfilename, TRUE);
  *(oCURRENT_STREAMo.get()->native_stream) << endl;
}

char* cpp_relativize_header_file(char* headerfile, char* systemname) {
  if (systemname == NULL) {
    systemname = oCURRENTSYSTEMDEFINITIONo.get()->name;
  }
  { 
    BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONo, System_Definition*, get_system_definition(systemname));
    BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo, char*, (((boolean)(oCURRENTSYSTEMDEFINITIONo.get())) ? oCURRENTSYSTEMDEFINITIONo.get()->directory : NULL));
    if (((boolean)(oCURRENTSYSTEMDEFINITIONo.get()))) {
      {
        { char* directoryfile = make_file_name("dummy", KWD_CPP_TRANSLATE_FILE_CPP_HEADER, TRUE);

          headerfile = relativize_file_name(headerfile, directory_parent_directory(directoryfile, 2));
        }
      }
    }
    else {
      *(STANDARD_WARNING->native_stream) << "Warning: " << "Can't find a system named " << "`" << systemname << "'" << endl << endl;
    }
  }
  return (headerfile);
}

void cpp_output_import_declarations(char* systemname) {
  if (systemname == NULL) {
    if (!((boolean)(oCURRENTSYSTEMDEFINITIONo.get()))) {
      return;
    }
    systemname = oCURRENTSYSTEMDEFINITIONo.get()->name;
  }
  { 
    BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONo, System_Definition*, get_system_definition(systemname));
    BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo, char*, (((boolean)(oCURRENTSYSTEMDEFINITIONo.get())) ? oCURRENTSYSTEMDEFINITIONo.get()->directory : NULL));
    if (((boolean)(oCURRENTSYSTEMDEFINITIONo.get()))) {
      {
        cpp_output_include_directive(cpp_relativize_header_file(make_file_name(cpp_yield_system_header_file_name(systemname), KWD_CPP_TRANSLATE_FILE_CPP_HEADER, TRUE), systemname), FALSE);
        *(oCURRENT_STREAMo.get()->native_stream) << endl;
      }
    }
    else {
      *(STANDARD_WARNING->native_stream) << "Warning: " << "Can't find a system named " << "`" << systemname << "'" << endl << endl;
    }
  }
}

void cpp_output_class_head_declarations(Cons* classes) {
  if (!(classes == NIL)) {
    *(oCURRENT_STREAMo.get()->native_stream) << "// Class headers:" << endl;
  }
  { Object* renamed_Class = NULL;
    Cons* iter_000 = classes;

    while (!(iter_000 == NIL)) {
      renamed_Class = iter_000->value;
      iter_000 = iter_000->rest;
      *(oCURRENT_STREAMo.get()->native_stream) << "class " << ((String_Wrapper*)(((Cons*)(renamed_Class))->rest->value))->wrapper_value << ";" << endl;
    }
  }
  *(oCURRENT_STREAMo.get()->native_stream) << endl;
}

void cpp_output_namespace_header(Module* module) {
  *(oCURRENT_STREAMo.get()->native_stream) << "namespace " << cpp_yield_namespace_prefix_from_module(module) << " {" << endl;
  { Module* usee = NULL;
    Cons* iter_000 = visible_modules(module)->consify()->rest->reverse();

    while (!(iter_000 == NIL)) {
      usee = ((Module*)(iter_000->value));
      iter_000 = iter_000->rest;
      *(oCURRENT_STREAMo.get()->native_stream) << "  using namespace " << cpp_yield_namespace_prefix_from_module(usee) << ";" << endl;
    }
  }
  *(oCURRENT_STREAMo.get()->native_stream) << endl;
}

void cpp_output_namespace_footer(Module* module) {
  *(oCURRENT_STREAMo.get()->native_stream) << "} // end of namespace " << cpp_yield_namespace_prefix_from_module(module) << endl;
}

boolean cpp_unit_defines_mainP(Translation_Unit* unit) {
  return ((unit->category == SYM_CPP_TRANSLATE_FILE_STELLA_METHOD) &&
      (((Method_Slot*)(unit->the_object))->slot_name == SYM_CPP_TRANSLATE_FILE_STELLA_MAIN));
}

void cpp_output_main(Translation_Unit* unit) {
  { Module* usee = NULL;
    Cons* iter_000 = visible_modules(oMODULEo.get())->consify()->reverse();

    while (!(iter_000 == NIL)) {
      usee = ((Module*)(iter_000->value));
      iter_000 = iter_000->rest;
      *(oCURRENT_STREAMo.get()->native_stream) << "using namespace " << cpp_yield_namespace_prefix_from_module(usee) << ";" << endl;
    }
  }
  *(oCURRENT_STREAMo.get()->native_stream) << endl;
  cpp_output_statement(unit->translation);
}

List* cpp_sort_units_for_header_file(List* fileunits) {
  { Cons* globals = NIL;
    Cons* auxiliaryglobals = NIL;
    Cons* methods = NIL;
    Cons* includestatements = NIL;
    Cons* classes = NIL;
    Cons* main = NIL;

    { Translation_Unit* unit = NULL;
      Cons* iter_000 = fileunits->the_cons_list;

      while (!(iter_000 == NIL)) {
        unit = ((Translation_Unit*)(iter_000->value));
        iter_000 = iter_000->rest;
        { Symbol* test_value_000 = unit->category;

          if ((test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_METHOD) ||
              ((test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_PRINT_METHOD) ||
               (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_MACRO))) {
            if (cpp_unit_defines_mainP(unit)) {
              main = cons(unit, main);
            }
            else {
              methods = cons(unit, methods);
            }
          }
          else if (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_VERBATIM) {
            { Cons* translation = ((Cons*)(unit->translation));
              char* code = ((String_Wrapper*)(translation->rest->value))->wrapper_value;

              if ((code[0] == '#') &&
                  (!(oCHARACTER_TYPE_TABLEo[(int)(unsigned char) (code[1])] == KWD_CPP_TRANSLATE_FILE_WHITE_SPACE))) {
                { Translation_Unit* self_000 = new_translation_unit();

                  self_000->category = SYM_CPP_TRANSLATE_FILE_STELLA_VERBATIM;
                  self_000->translation = copy_cons_tree(unit->translation);
                  includestatements = cons(self_000, includestatements);
                }
                translation->second_setter(wrap_string(""));
              }
            }
          }
          else if (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_CLASS) {
            classes = cons(unit, classes);
          }
          else if (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_GLOBAL_VARIABLE) {
            if (auxiliary_variable_unitP(unit)) {
              auxiliaryglobals = cons(unit, auxiliaryglobals);
              unit->category = SYM_CPP_TRANSLATE_FILE_STELLA_AUXILIARY_VARIABLE;
            }
            else {
              globals = cons(unit, globals);
            }
          }
          else if (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_TYPE) {
          }
          else {
            { Output_String_Stream* stream_000 = new_output_string_stream();

              *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_000->the_string_reader());
            }
          }
        }
      }
    }
    return (list(6, includestatements->reverse(), classes->reverse(), globals->reverse(), methods->reverse(), auxiliaryglobals->reverse(), main));
  }
}

void cpp_output_namespace_change(Module* fromnamespace, Module* tonamespace) {
  if (!((tonamespace == fromnamespace) ||
      (((boolean)(fromnamespace)) &&
       string_eqlP(cpp_yield_namespace_prefix_from_module(fromnamespace), cpp_yield_namespace_prefix_from_module(tonamespace))))) {
    if (!(!((boolean)(fromnamespace)))) {
      *(oCURRENT_STREAMo.get()->native_stream) << endl;
      cpp_output_namespace_footer(fromnamespace);
      *(oCURRENT_STREAMo.get()->native_stream) << endl << endl;
    }
    cpp_output_namespace_header(tonamespace);
  }
}

void cpp_output_all_units_to_file(char* filename) {
  { char* codeoutputfile = make_file_name(filename, KWD_CPP_TRANSLATE_FILE_CPP_CODE, FALSE);
    char* headeroutputfile = make_file_name(filename, KWD_CPP_TRANSLATE_FILE_CPP_HEADER, FALSE);
    Output_File_Stream* codeoutputstream = new_output_file_stream(codeoutputfile);
    Output_File_Stream* headeroutputstream = new_output_file_stream(headeroutputfile);
    List* headerunits = cpp_sort_units_for_header_file(oTRANSLATIONUNITSo.get());
    Cons* auxiliaryglobals = ((Cons*)(headerunits->fifth()));
    Translation_Unit* main = ((Translation_Unit*)(((Cons*)(headerunits->nth(5)))->value));
    Module* currentnamespacemodule = NULL;
    Module* currentunitmodule = NULL;

    { 
      BIND_STELLA_SPECIAL(oCURRENT_STREAMo, Output_Stream*, headeroutputstream);
      cout << "Writing " << "`" << headeroutputfile << "'" << "..." << endl;
      output_file_header(headeroutputstream, headeroutputfile);
      { Cons* bucket = NULL;
        Cons* iter_000 = headerunits->the_cons_list;

        while (!(iter_000 == NIL)) {
          bucket = ((Cons*)(iter_000->value));
          iter_000 = iter_000->rest;
          { Translation_Unit* unit = NULL;
            int i = NULL_INTEGER;
            Cons* iter_001 = bucket;
            int iter_002 = 1;

            while (!(iter_001 == NIL)) {
              unit = ((Translation_Unit*)(iter_001->value));
              iter_001 = iter_001->rest;
              i = iter_002;
              iter_002 = iter_002 + 1;
              if (((boolean)(unit->the_object)) &&
                  (!(unit->category == SYM_CPP_TRANSLATE_FILE_STELLA_VERBATIM))) {
                currentunitmodule = unit->the_object->home_module();
                cpp_output_namespace_change(currentnamespacemodule, currentunitmodule);
                currentnamespacemodule = currentunitmodule;
              }
              { Symbol* test_value_000 = unit->category;

                if (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_VERBATIM) {
                  if (i == 1) {
                    *(oCURRENT_STREAMo.get()->native_stream) << "// Header files and macros:" << endl;
                  }
                  cpp_output_statement(unit->translation);
                  *(oCURRENT_STREAMo.get()->native_stream) << endl;
                }
                else if (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_CLASS) {
                  if (i == 1) {
                    *(oCURRENT_STREAMo.get()->native_stream) << "// Class definitions:" << endl;
                  }
                  cpp_output_class(((Cons*)(unit->translation))->rest);
                }
                else if (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_GLOBAL_VARIABLE) {
                  if (i == 1) {
                    *(oCURRENT_STREAMo.get()->native_stream) << "// Global declarations:" << endl;
                  }
                  cpp_output_global_declaration(((Cons*)(unit->translation)));
                }
                else if ((test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_METHOD) ||
                    ((test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_PRINT_METHOD) ||
                     (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_MACRO))) {
                  if (i == 1) {
                    *(oCURRENT_STREAMo.get()->native_stream) << "// Function signatures:" << endl;
                  }
                  { Cons* translation = ((Cons*)(unit->translation));

                    { Generalized_Symbol* test_value_001 = ((Generalized_Symbol*)(translation->value));

                      if (test_value_001 == SYM_CPP_TRANSLATE_FILE_STELLA_CPP_FUNCTION) {
                        cpp_output_function_signature(translation->rest);
                        *(oCURRENT_STREAMo.get()->native_stream) << ";" << endl;
                      }
                      else if (test_value_001 == SYM_CPP_TRANSLATE_FILE_STELLA_CPP_DEFPRINT) {
                        cpp_output_defprint_signature(translation->rest);
                        *(oCURRENT_STREAMo.get()->native_stream) << ";" << endl;
                      }
                      else {
                      }
                    }
                  }
                }
                else if (test_value_000 == SYM_CPP_TRANSLATE_FILE_STELLA_AUXILIARY_VARIABLE) {
                  if (i == 1) {
                    *(oCURRENT_STREAMo.get()->native_stream) << "// Auxiliary global declarations:" << endl;
                  }
                  cpp_output_global_declaration(((Cons*)(unit->translation)));
                }
                else {
                  { Output_String_Stream* stream_000 = new_output_string_stream();

                    *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
                    throw *new_stella_exception(stream_000->the_string_reader());
                  }
                }
              }
            }
          }
          *(oCURRENT_STREAMo.get()->native_stream) << endl;
        }
      }
      cpp_output_namespace_footer(currentnamespacemodule);
      cout << "Writing " << "`" << codeoutputfile << "'" << "..." << endl;
      oCURRENT_STREAMo.set(codeoutputstream);
      *(oCURRENT_STREAMo.get()->native_stream) << "// " << file_name_without_directory(codeoutputfile) << endl << endl;
      if (translate_with_copyright_headerP()) {
        output_copyright_header(oCURRENT_STREAMo.get());
      }
      if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_CPP_TRANSLATE_FILE_CPP) {
        cpp_output_import_declarations(NULL);
      }
      else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_CPP_TRANSLATE_FILE_CPP_STANDALONE) {
        cpp_standalone_output_import_declarations(headeroutputfile);
      }
      else {
        { Output_String_Stream* stream_001 = new_output_string_stream();

          *(stream_001->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_001->the_string_reader());
        }
      }
      currentnamespacemodule = NULL;
      { Translation_Unit* unit = NULL;
        Cons* iter_003 = oTRANSLATIONUNITSo.get()->the_cons_list;

        while (!(iter_003 == NIL)) {
          unit = ((Translation_Unit*)(iter_003->value));
          iter_003 = iter_003->rest;
          if (((boolean)(unit->the_object)) &&
              ((!(unit->category == SYM_CPP_TRANSLATE_FILE_STELLA_VERBATIM)) &&
               (!(unit->category == SYM_CPP_TRANSLATE_FILE_STELLA_AUXILIARY_VARIABLE)))) {
            currentunitmodule = unit->the_object->home_module();
            cpp_output_namespace_change(currentnamespacemodule, currentunitmodule);
            currentnamespacemodule = currentunitmodule;
          }
          { Symbol* test_value_002 = unit->category;

            if ((test_value_002 == SYM_CPP_TRANSLATE_FILE_STELLA_METHOD) ||
                ((test_value_002 == SYM_CPP_TRANSLATE_FILE_STELLA_PRINT_METHOD) ||
                 (test_value_002 == SYM_CPP_TRANSLATE_FILE_STELLA_MACRO))) {
              cpp_output_statement(unit->translation);
            }
            else if (test_value_002 == SYM_CPP_TRANSLATE_FILE_STELLA_VERBATIM) {
              cpp_output_verbatim(((String_Wrapper*)(((Cons*)(unit->translation))->rest->value)));
            }
            else if (test_value_002 == SYM_CPP_TRANSLATE_FILE_STELLA_GLOBAL_VARIABLE) {
              if (!(auxiliary_variable_unitP(unit))) {
                cpp_output_variable_definition(((Cons*)(unit->translation)));
              }
            }
            else {
            }
          }
        }
      }
      { Translation_Unit* unit = NULL;
        Cons* iter_004 = auxiliaryglobals;

        while (!(iter_004 == NIL)) {
          unit = ((Translation_Unit*)(iter_004->value));
          iter_004 = iter_004->rest;
          currentunitmodule = unit->the_object->home_module();
          cpp_output_namespace_change(currentnamespacemodule, currentunitmodule);
          currentnamespacemodule = currentunitmodule;
          cpp_output_variable_definition(((Cons*)(unit->translation)));
        }
      }
      cpp_output_namespace_footer(currentnamespacemodule);
      if (((boolean)(main))) {
        *(oCURRENT_STREAMo.get()->native_stream) << endl << endl;
        cpp_output_main(main);
      }
      codeoutputstream->free();
      headeroutputstream->free();
    }
  }
}

void cpp_translate_file(char* filename) {
  translate_file(filename, KWD_CPP_TRANSLATE_FILE_CPP, FALSE);
}

void cpp_standalone_translate_file(char* filename) {
  translate_file(filename, KWD_CPP_TRANSLATE_FILE_CPP_STANDALONE, FALSE);
}

void cpp_translate_system(char* systemname) {
  // Translate the system `systemName' to C++.
  translate_system(systemname, KWD_CPP_TRANSLATE_FILE_CPP, 6, KWD_CPP_TRANSLATE_FILE_TWO_PASSp, TRUE_WRAPPER, KWD_CPP_TRANSLATE_FILE_FORCE_TRANSLATIONp, TRUE_WRAPPER, KWD_CPP_TRANSLATE_FILE_PRODUCTION_SETTINGSp, TRUE_WRAPPER);
}

void cpp_translate_walked_system_units(Key_Value_List* systemunits) {
  { 
    BIND_STELLA_SPECIAL(oCURRENTFILEo, char*, NULL);
    BIND_STELLA_SPECIAL(oTRANSLATIONUNITSo, List*, NULL);
    { String_Wrapper* file = NULL;
      List* units = NULL;
      Kv_Cons* iter_000 = systemunits->the_kv_list;

      while (((boolean)(iter_000))) {
        file = ((String_Wrapper*)(iter_000->key));
        units = ((List*)(iter_000->value));
        iter_000 = iter_000->rest;
        oCURRENTFILEo.set(file_base_name(file->wrapper_value));
        oTRANSLATIONUNITSo.set(units);
        { 
          BIND_STELLA_SPECIAL(oMODULEo, Module*, compute_file_units_module(units));
          BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
          translate_and_output_units_to_file(file->wrapper_value, "C++");
        }
        clean_up_translation_units_special();
      }
    }
    cpp_output_system_specific_files();
  }
}

void cpp_output_system_specific_files() {
  cpp_output_system_header_file();
  cpp_output_system_makefile();
}

char* cpp_yield_system_header_file_name(char* systemname) {
  return (string_concatenate(string_downcase(systemname), "-system", 0));
}

void cpp_output_system_header_file() {
  { System_Definition* system = oCURRENTSYSTEMDEFINITIONo.get();
    char* systemname = system->name;
    char* headeroutputfile = make_file_name(cpp_yield_system_header_file_name(systemname), KWD_CPP_TRANSLATE_FILE_CPP_HEADER, TRUE);
    Output_File_Stream* headeroutputstream = new_output_file_stream(headeroutputfile);

    { 
      BIND_STELLA_SPECIAL(oCURRENT_STREAMo, Output_Stream*, headeroutputstream);
      BIND_STELLA_SPECIAL(oMODULEo, Module*, get_cardinal_module(oCURRENTSYSTEMDEFINITIONo.get()));
      BIND_STELLA_SPECIAL(oCPP_CURRENT_UNIT_MODULEo, Module*, NULL);
      *(oCURRENT_STREAMo.get()->native_stream) << "// " << file_name_without_directory(headeroutputfile) << endl << endl;
      if (((boolean)(system->required_systems))) {
        *(oCURRENT_STREAMo.get()->native_stream) << "// Required systems:" << endl;
        { String_Wrapper* reqsys = NULL;
          Cons* iter_000 = system->required_systems;

          while (!(iter_000 == NIL)) {
            reqsys = ((String_Wrapper*)(iter_000->value));
            iter_000 = iter_000->rest;
            cpp_output_import_declarations(reqsys->wrapper_value);
          }
        }
      }
      *(oCURRENT_STREAMo.get()->native_stream) << "// System-wide forward declarations:" << endl;
      { Module* currentnamespacemodule = NULL;

        { Class* renamed_Class = NULL;
          Iterator* iter_001 = all_classes(oMODULEo.get(), TRUE);

          while (iter_001->nextP()) {
            renamed_Class = ((Class*)(iter_001->value));
            if (create_native_classP(renamed_Class)) {
              oCPP_CURRENT_UNIT_MODULEo.set(renamed_Class->home_module());
              cpp_output_namespace_change(currentnamespacemodule, oCPP_CURRENT_UNIT_MODULEo.get());
              currentnamespacemodule = oCPP_CURRENT_UNIT_MODULEo.get();
              *(oCURRENT_STREAMo.get()->native_stream) << "class " << cpp_translate_class_name(renamed_Class->class_type)->wrapper_value << ";" << endl;
            }
          }
        }
        *(oCURRENT_STREAMo.get()->native_stream) << endl;
        if (((boolean)(currentnamespacemodule))) {
          cpp_output_namespace_footer(oCPP_CURRENT_UNIT_MODULEo.get());
        }
        *(oCURRENT_STREAMo.get()->native_stream) << endl << endl;
      }
      *(oCURRENT_STREAMo.get()->native_stream) << "// System components:" << endl;
      if (string_eqlP(systemname, "stella")) {
        cpp_output_include_directive(cpp_relativize_header_file(make_file_name_from_relative_path(listO(3, wrap_string("cpp-lib"), wrap_string("cpp-primal"), NIL), KWD_CPP_TRANSLATE_FILE_CPP_HEADER), systemname), FALSE);
      }
      { String_Wrapper* file = NULL;
        Cons* iter_002 = files_plus_system_startup(system_definition_source_files(system));

        while (!(iter_002 == NIL)) {
          file = ((String_Wrapper*)(iter_002->value));
          iter_002 = iter_002->rest;
          cpp_output_include_directive(cpp_relativize_header_file(make_file_name(file->wrapper_value, KWD_CPP_TRANSLATE_FILE_CPP_HEADER, TRUE), systemname), FALSE);
        }
      }
      headeroutputstream->free();
    }
  }
}

void cpp_output_system_makefile() {
  { char* makefile = make_file_name("Makefile", KWD_CPP_TRANSLATE_FILE_STELLA, TRUE);
    char* templatefile = string_concatenate(file_name_without_extension(makefile), ".in", 0);
    System_Definition* system = oCURRENTSYSTEMDEFINITIONo.get();

    if (!(probe_fileP(templatefile))) {
      { 
        BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONo, System_Definition*, get_system_definition("STELLA"));
        BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo, char*, (((boolean)(oCURRENTSYSTEMDEFINITIONo.get())) ? oCURRENTSYSTEMDEFINITIONo.get()->directory : NULL));
        if (((boolean)(oCURRENTSYSTEMDEFINITIONo.get()))) {
          {
            makefile = make_file_name("Makefile-default", KWD_CPP_TRANSLATE_FILE_STELLA, TRUE);
            templatefile = string_concatenate(file_name_without_extension(makefile), ".in", 0);
          }
        }
        else {
          *(STANDARD_WARNING->native_stream) << "Warning: " << "Can't find a system named " << "STELLA" << endl << endl;
        }
      }
    }
    { Key_Value_List* variables = new_key_value_list();
      Output_String_Stream* systemfiles = new_output_string_stream();
      Output_String_Stream* requiredsystems = new_output_string_stream();

      variables->insert_at(wrap_string("SYSTEM-NAME"), wrap_string(system->name));
      variables->insert_at(wrap_string("PROGRAM-NAME"), wrap_string(system->name));
      variables->insert_at(wrap_string("SYSTEMS-ROOT"), wrap_string(directory_file_name(directory_parent_directory(make_file_name("dummy", KWD_CPP_TRANSLATE_FILE_CPP_CODE, TRUE), 2))));
      { 
        BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONo, System_Definition*, get_system_definition("STELLA"));
        BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo, char*, (((boolean)(oCURRENTSYSTEMDEFINITIONo.get())) ? oCURRENTSYSTEMDEFINITIONo.get()->directory : NULL));
        if (((boolean)(oCURRENTSYSTEMDEFINITIONo.get()))) {
          {
            variables->insert_at(wrap_string("STELLA-ROOT"), wrap_string(directory_file_name(directory_parent_directory(make_file_name("dummy", KWD_CPP_TRANSLATE_FILE_CPP_CODE, TRUE), 1))));
          }
        }
        else {
          *(STANDARD_WARNING->native_stream) << "Warning: " << "Can't find a system named " << "STELLA" << endl << endl;
        }
      }
      { String_Wrapper* file = NULL;
        Cons* iter_000 = files_plus_system_startup(system_definition_source_files(system));

        while (!(iter_000 == NIL)) {
          file = ((String_Wrapper*)(iter_000->value));
          iter_000 = iter_000->rest;
          *(systemfiles->native_stream) << file_name_without_directory(make_file_name(file->wrapper_value, KWD_CPP_TRANSLATE_FILE_CPP_CODE, TRUE)) << " ";
        }
      }
      variables->insert_at(wrap_string("SYSTEM-FILES"), wrap_string(systemfiles->the_string_reader()));
      if (((boolean)(system->required_systems))) {
        { String_Wrapper* reqsys = NULL;
          Cons* iter_001 = system->required_systems;

          while (!(iter_001 == NIL)) {
            reqsys = ((String_Wrapper*)(iter_001->value));
            iter_001 = iter_001->rest;
            *(requiredsystems->native_stream) << reqsys->wrapper_value << " ";
          }
        }
      }
      variables->insert_at(wrap_string("REQUIRED-SYSTEMS"), wrap_string(requiredsystems->the_string_reader()));
      makefile = file_name_without_extension(make_file_name("Makefile", KWD_CPP_TRANSLATE_FILE_CPP_CODE, TRUE));
      substitute_template_variables_in_file(templatefile, makefile, variables);
    }
  }
}

char* oMAKEFILE_TEMPLATE_VARIABLE_PREFIXo = "#$";

char* substitute_template_variables_in_string(char* string, Key_Value_List* variables) {
  { int index = 0;
    int length = strlen(string);
    char* result = string;
    int variableend = 0;
    char* variablename = NULL;
    String_Wrapper* variablevalue = NULL;

    for (;;) {
      index = string_search(result, oMAKEFILE_TEMPLATE_VARIABLE_PREFIXo, index);
      if (index == NULL_INTEGER) {
        break;
      }
      index = index + 2;
      if (index >= length) {
        *(STANDARD_WARNING->native_stream) << "Warning: " << "Illegal template variable syntax in" << endl << "    " << "`" << string << "'" << endl;
        break;
      }
      if (!(result[index] == '(')) {
        *(STANDARD_WARNING->native_stream) << "Warning: " << "Illegal template variable syntax in" << endl << "    " << "`" << string << "'" << endl;
        continue;
      }
      variableend = string_position(result, ')', index);
      if (variableend == NULL_INTEGER) {
        *(STANDARD_WARNING->native_stream) << "Warning: " << "Illegal template variable syntax in" << endl << "    " << "`" << string << "'" << endl;
        continue;
      }
      index = index + 1;
      variablename = string_subsequence(result, index, variableend);
      variablevalue = ((String_Wrapper*)(variables->lookup(wrap_string(variablename))));
      if (!((boolean)(variablevalue))) {
        *(STANDARD_WARNING->native_stream) << "Warning: " << "Undefined template variable: " << "`" << variablename << "'" << endl;
        continue;
      }
      result = string_concatenate(string_subsequence(result, 0, index - 3), variablevalue->wrapper_value, 1, string_subsequence(result, variableend + 1, NULL_INTEGER));
      index = variableend + 1 + (strlen(result) - length);
      length = strlen(result);
    }
    return (result);
  }
}

void substitute_template_variables_in_file(char* templatefile, char* outputfile, Key_Value_List* variables) {
  { Output_File_Stream* out = new_output_file_stream(outputfile);

    *(out->native_stream) << "# This file has been generated automatically.  " << "Modify at your own risk!" << endl << endl;
    { char* line = NULL;
      Line_Iterator* iter_000 = lines(new_input_file_stream(templatefile));

      while (iter_000->nextP()) {
        line = iter_000->value;
        *(out->native_stream) << substitute_template_variables_in_string(line, variables) << endl;
      }
    }
    out->free();
  }
}

void startup_cpp_translate_file() {
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oSTELLA_MODULEo);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (current_startup_time_phaseP(2)) {
      KWD_CPP_TRANSLATE_FILE_CPP = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP", NULL, 2)));
      KWD_CPP_TRANSLATE_FILE_CPP_CODE = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP-CODE", NULL, 2)));
      KWD_CPP_TRANSLATE_FILE_CPP_HEADER = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP-HEADER", NULL, 2)));
      SYM_CPP_TRANSLATE_FILE_STELLA_METHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_MAIN = ((Symbol*)(intern_rigid_symbol_wrt_module("MAIN", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_PRINT_METHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("PRINT-METHOD", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_MACRO = ((Symbol*)(intern_rigid_symbol_wrt_module("MACRO", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_VERBATIM = ((Symbol*)(intern_rigid_symbol_wrt_module("VERBATIM", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_CLASS = ((Symbol*)(intern_rigid_symbol_wrt_module("CLASS", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_GLOBAL_VARIABLE = ((Symbol*)(intern_rigid_symbol_wrt_module("GLOBAL-VARIABLE", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_TYPE = ((Symbol*)(intern_rigid_symbol_wrt_module("TYPE", NULL, 0)));
      KWD_CPP_TRANSLATE_FILE_WHITE_SPACE = ((Keyword*)(intern_rigid_symbol_wrt_module("WHITE-SPACE", NULL, 2)));
      SYM_CPP_TRANSLATE_FILE_STELLA_AUXILIARY_VARIABLE = ((Symbol*)(intern_rigid_symbol_wrt_module("AUXILIARY-VARIABLE", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_CPP_FUNCTION = ((Symbol*)(intern_rigid_symbol_wrt_module("CPP_FUNCTION", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_CPP_DEFPRINT = ((Symbol*)(intern_rigid_symbol_wrt_module("CPP_DEFPRINT", NULL, 0)));
      KWD_CPP_TRANSLATE_FILE_CPP_STANDALONE = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP-STANDALONE", NULL, 2)));
      KWD_CPP_TRANSLATE_FILE_TWO_PASSp = ((Keyword*)(intern_rigid_symbol_wrt_module("TWO-PASS?", NULL, 2)));
      KWD_CPP_TRANSLATE_FILE_FORCE_TRANSLATIONp = ((Keyword*)(intern_rigid_symbol_wrt_module("FORCE-TRANSLATION?", NULL, 2)));
      KWD_CPP_TRANSLATE_FILE_PRODUCTION_SETTINGSp = ((Keyword*)(intern_rigid_symbol_wrt_module("PRODUCTION-SETTINGS?", NULL, 2)));
      KWD_CPP_TRANSLATE_FILE_STELLA = ((Keyword*)(intern_rigid_symbol_wrt_module("STELLA", NULL, 2)));
      SYM_CPP_TRANSLATE_FILE_STELLA_STARTUP_CPP_TRANSLATE_FILE = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-CPP-TRANSLATE-FILE", NULL, 0)));
      SYM_CPP_TRANSLATE_FILE_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-STARTUP-CLASSNAME", NULL, 0)));
    }
    if (current_startup_time_phaseP(4)) {
      oCURRENT_STREAMo.set(STANDARD_OUTPUT);
    }
    if (current_startup_time_phaseP(6)) {
      finalize_classes();
    }
    if (current_startup_time_phaseP(7)) {
      define_function_object("CPPTRANS", "(DEFUN CPPTRANS ((STATEMENT OBJECT)) :COMMAND? TRUE :PUBLIC? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Translate `statement' to C++ and print the result.\")", ((cpp_function_code)(&cpptrans)), NULL);
      define_function_object("CPP-MAKE-CODE-OUTPUT-FILE-NAME", "(DEFUN (CPP-MAKE-CODE-OUTPUT-FILE-NAME FILE-NAME) ((FILENAME FILE-NAME)))", ((cpp_function_code)(&cpp_make_code_output_file_name)), NULL);
      define_function_object("CPP-MAKE-HEADER-OUTPUT-FILE-NAME", "(DEFUN (CPP-MAKE-HEADER-OUTPUT-FILE-NAME FILE-NAME) ((FILENAME FILE-NAME)))", ((cpp_function_code)(&cpp_make_header_output_file_name)), NULL);
      define_function_object("CPP-OUTPUT-INCLUDE-DIRECTIVE", "(DEFUN CPP-OUTPUT-INCLUDE-DIRECTIVE ((FILENAME FILE-NAME) (STRIPDIRECTORY? BOOLEAN)))", ((cpp_function_code)(&cpp_output_include_directive)), NULL);
      define_function_object("CPP-STANDALONE-OUTPUT-IMPORT-DECLARATIONS", "(DEFUN CPP-STANDALONE-OUTPUT-IMPORT-DECLARATIONS ((HEADERFILENAME FILE-NAME)))", ((cpp_function_code)(&cpp_standalone_output_import_declarations)), NULL);
      define_function_object("CPP-RELATIVIZE-HEADER-FILE", "(DEFUN (CPP-RELATIVIZE-HEADER-FILE FILE-NAME) ((HEADERFILE FILE-NAME) (SYSTEMNAME STRING)))", ((cpp_function_code)(&cpp_relativize_header_file)), NULL);
      define_function_object("CPP-OUTPUT-IMPORT-DECLARATIONS", "(DEFUN CPP-OUTPUT-IMPORT-DECLARATIONS ((SYSTEMNAME STRING)))", ((cpp_function_code)(&cpp_output_import_declarations)), NULL);
      define_function_object("CPP-OUTPUT-CLASS-HEAD-DECLARATIONS", "(DEFUN CPP-OUTPUT-CLASS-HEAD-DECLARATIONS ((CLASSES CONS)))", ((cpp_function_code)(&cpp_output_class_head_declarations)), NULL);
      define_function_object("CPP-OUTPUT-NAMESPACE-HEADER", "(DEFUN CPP-OUTPUT-NAMESPACE-HEADER ((MODULE MODULE)))", ((cpp_function_code)(&cpp_output_namespace_header)), NULL);
      define_function_object("CPP-OUTPUT-NAMESPACE-FOOTER", "(DEFUN CPP-OUTPUT-NAMESPACE-FOOTER ((MODULE MODULE)))", ((cpp_function_code)(&cpp_output_namespace_footer)), NULL);
      define_function_object("CPP-UNIT-DEFINES-MAIN?", "(DEFUN (CPP-UNIT-DEFINES-MAIN? BOOLEAN) ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&cpp_unit_defines_mainP)), NULL);
      define_function_object("CPP-OUTPUT-MAIN", "(DEFUN CPP-OUTPUT-MAIN ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&cpp_output_main)), NULL);
      define_function_object("CPP-SORT-UNITS-FOR-HEADER-FILE", "(DEFUN (CPP-SORT-UNITS-FOR-HEADER-FILE (LIST OF (CONS OF TRANSLATION-UNIT))) ((FILEUNITS (LIST OF TRANSLATION-UNIT))))", ((cpp_function_code)(&cpp_sort_units_for_header_file)), NULL);
      define_function_object("CPP-OUTPUT-NAMESPACE-CHANGE", "(DEFUN CPP-OUTPUT-NAMESPACE-CHANGE ((FROMNAMESPACE MODULE) (TONAMESPACE MODULE)))", ((cpp_function_code)(&cpp_output_namespace_change)), NULL);
      define_function_object("CPP-OUTPUT-ALL-UNITS-TO-FILE", "(DEFUN CPP-OUTPUT-ALL-UNITS-TO-FILE ((FILENAME FILE-NAME)))", ((cpp_function_code)(&cpp_output_all_units_to_file)), NULL);
      define_function_object("CPP-TRANSLATE-FILE", "(DEFUN CPP-TRANSLATE-FILE ((FILENAME FILE-NAME)) :PUBLIC? TRUE)", ((cpp_function_code)(&cpp_translate_file)), NULL);
      define_function_object("CPP-STANDALONE-TRANSLATE-FILE", "(DEFUN CPP-STANDALONE-TRANSLATE-FILE ((FILENAME FILE-NAME)))", ((cpp_function_code)(&cpp_standalone_translate_file)), NULL);
      define_function_object("CPP-TRANSLATE-SYSTEM", "(DEFUN CPP-TRANSLATE-SYSTEM ((SYSTEMNAME STRING)) :DOCUMENTATION \"Translate the system `systemName' to C++.\" :PUBLIC? TRUE)", ((cpp_function_code)(&cpp_translate_system)), NULL);
      define_function_object("CPP-TRANSLATE-WALKED-SYSTEM-UNITS", "(DEFUN CPP-TRANSLATE-WALKED-SYSTEM-UNITS ((SYSTEMUNITS SYSTEM-UNITS-ALIST)))", ((cpp_function_code)(&cpp_translate_walked_system_units)), NULL);
      define_function_object("CPP-OUTPUT-SYSTEM-SPECIFIC-FILES", "(DEFUN CPP-OUTPUT-SYSTEM-SPECIFIC-FILES ())", ((cpp_function_code)(&cpp_output_system_specific_files)), NULL);
      define_function_object("CPP-YIELD-SYSTEM-HEADER-FILE-NAME", "(DEFUN (CPP-YIELD-SYSTEM-HEADER-FILE-NAME STRING) ((SYSTEMNAME STRING)))", ((cpp_function_code)(&cpp_yield_system_header_file_name)), NULL);
      define_function_object("CPP-OUTPUT-SYSTEM-HEADER-FILE", "(DEFUN CPP-OUTPUT-SYSTEM-HEADER-FILE ())", ((cpp_function_code)(&cpp_output_system_header_file)), NULL);
      define_function_object("CPP-OUTPUT-SYSTEM-MAKEFILE", "(DEFUN CPP-OUTPUT-SYSTEM-MAKEFILE ())", ((cpp_function_code)(&cpp_output_system_makefile)), NULL);
      define_function_object("SUBSTITUTE-TEMPLATE-VARIABLES-IN-STRING", "(DEFUN (SUBSTITUTE-TEMPLATE-VARIABLES-IN-STRING STRING) ((STRING STRING) (VARIABLES (KEY-VALUE-LIST OF STRING-WRAPPER STRING-WRAPPER))))", ((cpp_function_code)(&substitute_template_variables_in_string)), NULL);
      define_function_object("SUBSTITUTE-TEMPLATE-VARIABLES-IN-FILE", "(DEFUN SUBSTITUTE-TEMPLATE-VARIABLES-IN-FILE ((TEMPLATEFILE STRING) (OUTPUTFILE STRING) (VARIABLES (KEY-VALUE-LIST OF STRING-WRAPPER STRING-WRAPPER))))", ((cpp_function_code)(&substitute_template_variables_in_file)), NULL);
      define_function_object("STARTUP-CPP-TRANSLATE-FILE", "(DEFUN STARTUP-CPP-TRANSLATE-FILE () :PUBLIC? TRUE)", ((cpp_function_code)(&startup_cpp_translate_file)), NULL);
      { Method_Slot* function = lookup_function(SYM_CPP_TRANSLATE_FILE_STELLA_STARTUP_CPP_TRANSLATE_FILE);

        set_dynamic_slot_value(function->dynamic_slots, SYM_CPP_TRANSLATE_FILE_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string("Startup-Cpp-Translate-File"), NULL_STRING_WRAPPER);
      }
    }
    if (current_startup_time_phaseP(8)) {
      finalize_slots();
      cleanup_unfinalized_classes();
    }
    if (current_startup_time_phaseP(9)) {
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *CURRENT-STREAM* OUTPUT-STREAM STANDARD-OUTPUT)");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *CPP-TRANSLATED-FILE-SUFFIX* STRING \"\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *MAKEFILE-TEMPLATE-VARIABLE-PREFIX* STRING \"#$\")");
    }
  }
}

Keyword* KWD_CPP_TRANSLATE_FILE_CPP = NULL;

Keyword* KWD_CPP_TRANSLATE_FILE_CPP_CODE = NULL;

Keyword* KWD_CPP_TRANSLATE_FILE_CPP_HEADER = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_METHOD = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_MAIN = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_PRINT_METHOD = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_MACRO = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_VERBATIM = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_CLASS = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_GLOBAL_VARIABLE = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_TYPE = NULL;

Keyword* KWD_CPP_TRANSLATE_FILE_WHITE_SPACE = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_AUXILIARY_VARIABLE = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_CPP_FUNCTION = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_CPP_DEFPRINT = NULL;

Keyword* KWD_CPP_TRANSLATE_FILE_CPP_STANDALONE = NULL;

Keyword* KWD_CPP_TRANSLATE_FILE_TWO_PASSp = NULL;

Keyword* KWD_CPP_TRANSLATE_FILE_FORCE_TRANSLATIONp = NULL;

Keyword* KWD_CPP_TRANSLATE_FILE_PRODUCTION_SETTINGSp = NULL;

Keyword* KWD_CPP_TRANSLATE_FILE_STELLA = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_STARTUP_CPP_TRANSLATE_FILE = NULL;

Symbol* SYM_CPP_TRANSLATE_FILE_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace stella
