// foreach.cc

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

#include "stella/stella-system.hh"

namespace stella {

Cons* extract_one_in_clause(Cons* tail, Cons*& _return1) {
  { Object* variable = tail->value;
    Object* renamed_Operator = tail->rest->value;
    Cons* clause = tail;

    if (!((symbolP(variable) ||
        consP(variable)) &&
        (get_quoted_tree("((IN ON ISA) \"/STELLA\")", "/STELLA")->memberP(renamed_Operator) &&
         ((boolean)(tail->rest->rest->value))))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal syntax where 'IN', 'ON', or 'ISA' clause expected:" << endl << "   " << "`" << de_uglify_parse_tree(tail) << "'" << "." << endl;
        }
      }
      _return1 = NIL;
      return (NULL);
    }
    clause->first_setter(renamed_Operator);
    clause->second_setter(variable);
    tail = tail->rest->rest->rest;
    clause->rest->rest->rest = NIL;
    _return1 = tail;
    return (clause);
  }
}

Cons* extract_one_generator_term(Cons* tail, Cons*& _return1) {
  { Cons* inclause = NULL;
    Cons* residue = NULL;
    Cons* parallelterms = NULL;
    Cons* temp = NULL;

    inclause = extract_one_in_clause(tail, residue);
    if (!(residue->value == SYM_FOREACH_STELLA_AS)) {
      _return1 = residue;
      return (inclause);
    }
    parallelterms = cons(inclause, cons(SYM_FOREACH_STELLA_AS, NIL));
    while (residue->value == SYM_FOREACH_STELLA_AS) {
      temp = residue;
      residue = residue->rest;
      inclause = extract_one_in_clause(residue, residue);
      parallelterms = cons(inclause, parallelterms);
    }
    _return1 = residue;
    return (parallelterms->reverse());
  }
}

Cons* extract_generators_clause(Cons* tail, Cons*& _return1) {
  { Cons* generatorterm = NULL;
    Cons* residue = NULL;
    Cons* generatorsclause = NULL;

    generatorterm = extract_one_generator_term(tail, residue);
    if (!(residue->value == SYM_FOREACH_STELLA_COMMA)) {
      _return1 = residue;
      return (generatorterm);
    }
    generatorsclause = cons(generatorterm, cons(SYM_FOREACH_STELLA_COMMA, NIL));
    while (residue->value == SYM_FOREACH_STELLA_COMMA) {
      generatorterm = extract_one_generator_term(residue->rest, residue);
      generatorsclause = cons(generatorterm, generatorsclause);
    }
    _return1 = residue;
    return (generatorsclause->reverse());
  }
}

Object* extract_option_and_residue(Symbol* option, Cons* tail, Cons*& _return1) {
  if (!(tail->value == option)) {
    _return1 = tail;
    return (NULL);
  }
  { Object* argument = NULL;
    Cons* residue = NULL;

    argument = tail->rest->value;
    residue = tail->rest->rest;
    _return1 = residue;
    return (argument);
  }
}

Cons* extract_do_clause(Cons* tail, Cons*& _return1) {
  { Standard_Object* dummy1;

    if (!(tail->value == SYM_FOREACH_STELLA_DO)) {
      _return1 = tail;
      return (NULL);
    }
    { Cons* doclause = tail;
      Cons* lastdocons = NULL;

      tail = tail->rest;
      while ((!(tail == NIL)) &&
          (!(tail->value == SYM_FOREACH_STELLA_COLLECT))) {
        lastdocons = tail;
        tail = tail->rest;
      }
      if (!((boolean)(lastdocons))) {
        {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Missing arguments to DO separator" << "." << endl;
            }
          }
          doclause->rest = cons_list(1, walk_dont_call_me_tree(NIL, SGT_FOREACH_STELLA_VOID, dummy1));
        }
      }
      else {
        lastdocons->rest = NIL;
      }
      _return1 = tail;
      return (doclause);
    }
  }
}

Cons* foreach_to_prefix_foreach_tree(Cons* foreachtree) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;

    { Cons* generatorsclause = NULL;
      Object* filterclause = NULL;
      Cons* doclause = NULL;
      Object* collectclause = NULL;
      Object* intoclause = NULL;
      Cons* residue = ((Cons*)(foreachtree->rest));
      Cons* otree = NULL;

      generatorsclause = extract_generators_clause(residue, residue);
      if (!((boolean)(generatorsclause))) {
        return (listO(4, SYM_FOREACH_STELLA_PREFIX_FOREACH, listO(4, SYM_FOREACH_STELLA_IN, SYM_FOREACH_STELLA_X, SYM_FOREACH_STELLA_NIL, NIL), listO(3, SYM_FOREACH_STELLA_DO, walk_dont_call_me_tree(otree, SGT_FOREACH_STELLA_VOID, dummy1), NIL), NIL));
      }
      filterclause = extract_option_and_residue(SYM_FOREACH_STELLA_WHERE, residue, residue);
      if (((boolean)(filterclause))) {
        filterclause = listO(3, SYM_FOREACH_STELLA_WHERE, filterclause, NIL);
      }
      doclause = extract_do_clause(residue, residue);
      collectclause = extract_option_and_residue(SYM_FOREACH_STELLA_COLLECT, residue, residue);
      intoclause = extract_option_and_residue(SYM_FOREACH_STELLA_INTO, residue, residue);
      if (((boolean)(collectclause)) ||
          ((boolean)(intoclause))) {
        collectclause = listO(3, SYM_FOREACH_STELLA_COLLECT_INTO, collectclause, cons(intoclause, NIL));
      }
      otree = listO(3, SYM_FOREACH_STELLA_PREFIX_FOREACH, generatorsclause, ((((boolean)(filterclause)) ? cons(filterclause, NIL) : NIL))->concatenate(((((boolean)(doclause)) ? cons(doclause, NIL) : NIL))->concatenate(((((boolean)(collectclause)) ? cons(collectclause, NIL) : NIL))->concatenate(NIL, 0), 0), 0));
      if ((((boolean)(collectclause)) &&
          (!((boolean)(intoclause)))) ||
          (((boolean)(intoclause)) &&
           (!((boolean)(collectclause))))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Missing COLLECT or INTO clause in" << "." << endl;
          }
        }
        return (listO(4, SYM_FOREACH_STELLA_PREFIX_FOREACH, listO(4, SYM_FOREACH_STELLA_IN, SYM_FOREACH_STELLA_X, SYM_FOREACH_STELLA_NIL, NIL), listO(3, SYM_FOREACH_STELLA_DO, walk_dont_call_me_tree(otree, SGT_FOREACH_STELLA_VOID, dummy2), NIL), NIL));
      }
      if (!(residue == NIL)) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Extra expressions at the end of foreach statement:" << "." << endl;
          }
        }
        return (listO(4, SYM_FOREACH_STELLA_PREFIX_FOREACH, listO(4, SYM_FOREACH_STELLA_IN, SYM_FOREACH_STELLA_X, SYM_FOREACH_STELLA_NIL, NIL), listO(3, SYM_FOREACH_STELLA_DO, walk_dont_call_me_tree(otree, SGT_FOREACH_STELLA_VOID, dummy3), NIL), NIL));
      }
      return (otree);
    }
  }
}

Cons* walk_foreach_tree(Cons* tree, Standard_Object*& _return1) {
  { 
    BIND_STELLA_SPECIAL(oNOFSPECIALSATLOOPENTRYo, int, oSPECIALVARIABLESTACKo.get()->length());
    { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

      if (test_value_000 == SYM_FOREACH_STELLA_FOREACH) {
        return (walk_prefix_foreach_tree(foreach_to_prefix_foreach_tree(tree), _return1));
      }
      else if (test_value_000 == SYM_FOREACH_STELLA_EXISTS) {
        return (walk_exists_tree(tree, _return1));
      }
      else if (test_value_000 == SYM_FOREACH_STELLA_FORALL) {
        return (walk_forall_tree(tree, _return1));
      }
      else if (test_value_000 == SYM_FOREACH_STELLA_SOME) {
        return (walk_some_tree(tree, _return1));
      }
      else if ((test_value_000 == SYM_FOREACH_STELLA_SETOF) ||
          (test_value_000 == SYM_FOREACH_STELLA_SET_OF)) {
        return (walk_setof_tree(tree, _return1));
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
    }
  }
}

Object* walk_collection_tree(Object* tree, boolean dontoptimizeP, Standard_Object*& _return1) {
  if (safe_primary_type(tree) == SGT_FOREACH_STELLA_CONS) {
    { Object* tree_000 = tree;
      Cons* tree = ((Cons*)(tree_000));

      if ((!dontoptimizeP) &&
          ((tree->value == SYM_FOREACH_STELLA_INTERVAL) &&
           (tree->length() == 3))) {
        _return1 = SGT_FOREACH_STELLA_INTEGER_INTERVAL;
        return (tree);
      }
    }
  }
  else {
  }
  return (walk_expression_tree(tree, SGT_FOREACH_STELLA_UNKNOWN, SYM_FOREACH_STELLA_FOREACH, FALSE, _return1));
}

Cons* yield_in_cursor_clauses(Cons* intree, boolean dontoptimizeP, Cons*& _return1, Object*& _return2) {
  { Object* vartree = intree->rest->value;
    Symbol* keyvar = NULL;
    Symbol* valuevar = NULL;
    Surrogate* collectionbasetype = NULL;

    { Object* collectiontree = NULL;
      Standard_Object* collectiontype = NULL;

      collectiontree = walk_collection_tree(intree->rest->rest->value, dontoptimizeP, collectiontype);
      intree->third_setter(NULL);
      collectionbasetype = type_spec_to_base_type(collectiontype);
      if (dontoptimizeP &&
          (!safe_subtype_ofP(collectionbasetype, SGT_FOREACH_STELLA_ABSTRACT_ITERATOR))) {
        collectiontree = walk_collection_tree(listO(3, SYM_FOREACH_STELLA_ALLOCATE_ITERATOR, collectiontree, NIL), TRUE, collectiontype);
        collectionbasetype = type_spec_to_base_type(collectiontype);
      }
      if (safe_primary_type(vartree) == SGT_FOREACH_STELLA_CONS) {
        { Object* vartree_000 = vartree;
          Cons* vartree = ((Cons*)(vartree_000));

          if (!(vartree->length() == 2)) {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal number of variables in IN clause: " << "`" << de_uglify_parse_tree(intree) << "'" << "." << endl;
              }
            }
            _return1 = NIL;
            _return2 = SYM_FOREACH_STELLA_FALSE;
            return (NIL);
          }
          keyvar = ((Symbol*)(vartree->value));
          valuevar = ((Symbol*)(vartree->rest->value));
        }
      }
      else {
        valuevar = ((Symbol*)(vartree));
      }
      if (collectionbasetype == SGT_FOREACH_STELLA_INTEGER_INTERVAL) {
        return (yield_in_cursor_clauses_for_integer_interval(((Cons*)(collectiontree)), collectiontype, keyvar, valuevar, _return1, _return2));
      }
      else if (collectionbasetype == SGT_FOREACH_STELLA_ARGUMENT_LIST) {
        return (yield_in_cursor_clauses_for_argument_list(collectiontree, collectiontype, keyvar, valuevar, _return1, _return2));
      }
      else {
        if (use_vector_style_iterationP(collectionbasetype)) {
          return (yield_in_cursor_clauses_for_vector(collectiontree, collectiontype, keyvar, valuevar, _return1, _return2));
        }
        else {
          return (yield_in_cursor_clauses_for_general_collection(collectiontree, collectiontype, keyvar, valuevar, _return1, _return2));
        }
      }
    }
  }
}

boolean use_vector_style_iterationP(Surrogate* collectiontype) {
  return ((collectiontype == SGT_FOREACH_STELLA_STRING) ||
      ((collectiontype == SGT_FOREACH_STELLA_MUTABLE_STRING) ||
       (safe_subtype_ofP(collectiontype, SGT_FOREACH_STELLA_VECTOR) ||
        (safe_subtype_ofP(collectiontype, SGT_FOREACH_STELLA_SEQUENCE_MIXIN) &&
         ((boolean)(lookup_slot(((Class*)(collectiontype->surrogate_value)), SYM_FOREACH_STELLA_THE_ARRAY)))))));
}

Cons* yield_in_cursor_clauses_for_general_collection(Object* collectiontree, Standard_Object* collectiontype, Symbol* keyvar, Symbol* valuevar, Cons*& _return1, Object*& _return2) {
  { boolean dummy1;

    { Surrogate* collectionbasetype = type_spec_to_base_type(collectiontype);
      Symbol* iteratorvar = local_gensym("ITER");
      Cons* iteratorclauses = NIL;
      Cons* valueclauses = NIL;
      Object* continuationtest = NULL;
      Cons* keytree = NIL;
      Cons* valuetree = NIL;
      Cons* bumptrees = NIL;

      if (collectionbasetype == SGT_FOREACH_STELLA_LIST) {
        iteratorclauses = cons(cons(iteratorvar, listO(3, SGT_FOREACH_STELLA_CONS, listO(3, SYM_FOREACH_STELLA_THE_CONS_LIST, collectiontree, NIL), NIL)), NIL);
        collectionbasetype = SGT_FOREACH_STELLA_CONS;
      }
      else if (collectionbasetype == SGT_FOREACH_STELLA_KEY_VALUE_LIST) {
        iteratorclauses = cons(cons(iteratorvar, listO(3, SGT_FOREACH_STELLA_KV_CONS, listO(3, SYM_FOREACH_STELLA_THE_KV_LIST, collectiontree, NIL), NIL)), NIL);
        collectionbasetype = SGT_FOREACH_STELLA_KV_CONS;
      }
      else if (collectionbasetype == SGT_FOREACH_STELLA_PROPERTY_LIST) {
        iteratorclauses = cons(cons(iteratorvar, listO(3, SGT_FOREACH_STELLA_CONS, listO(3, SYM_FOREACH_STELLA_THE_PLIST, collectiontree, NIL), NIL)), NIL);
      }
      else if ((collectionbasetype == SGT_FOREACH_STELLA_CONS) ||
          (collectionbasetype == SGT_FOREACH_STELLA_KV_CONS)) {
        iteratorclauses = cons(cons(iteratorvar, cons(collectionbasetype, cons(collectiontree, NIL))), NIL);
      }
      else {
        if (!(safe_subtype_ofP(collectionbasetype, SGT_FOREACH_STELLA_ABSTRACT_ITERATOR))) {
          collectiontree = listO(3, SYM_FOREACH_STELLA_ALLOCATE_ITERATOR, collectiontree, NIL);
        }
        iteratorclauses = cons(cons(iteratorvar, listO(3, SGT_FOREACH_STELLA_ABSTRACT_ITERATOR, collectiontree, NIL)), NIL);
        collectionbasetype = SGT_FOREACH_STELLA_ABSTRACT_ITERATOR;
      }
      if (((boolean)(keyvar))) {
        keytree = listO(3, SYM_FOREACH_STELLA_KEY, iteratorvar, NIL);
      }
      valuetree = listO(3, SYM_FOREACH_STELLA_VALUE, iteratorvar, NIL);
      if (collectionbasetype == SGT_FOREACH_STELLA_CONS) {
        bumptrees = cons(listO(3, SYM_FOREACH_STELLA_SETQ, iteratorvar, cons(listO(3, SYM_FOREACH_STELLA_REST, iteratorvar, NIL), NIL)), NIL);
        continuationtest = listO(3, SYM_FOREACH_STELLA_NOT, listO(3, SYM_FOREACH_STELLA_NILp, iteratorvar, NIL), NIL);
      }
      else if (collectionbasetype == SGT_FOREACH_STELLA_KV_CONS) {
        bumptrees = cons(listO(3, SYM_FOREACH_STELLA_SETQ, iteratorvar, cons(listO(3, SYM_FOREACH_STELLA_REST, iteratorvar, NIL), NIL)), NIL);
        continuationtest = listO(3, SYM_FOREACH_STELLA_DEFINEDp, iteratorvar, NIL);
      }
      else if (collectionbasetype == SGT_FOREACH_STELLA_PROPERTY_LIST) {
        if (((boolean)(keyvar))) {
          keytree = listO(3, SYM_FOREACH_STELLA_VALUE, iteratorvar, NIL);
        }
        valuetree = listO(3, SYM_FOREACH_STELLA_VALUE, listO(3, SYM_FOREACH_STELLA_REST, iteratorvar, NIL), NIL);
        bumptrees = cons(listO(3, SYM_FOREACH_STELLA_SETQ, iteratorvar, cons(listO(3, SYM_FOREACH_STELLA_REST, listO(3, SYM_FOREACH_STELLA_REST, iteratorvar, NIL), NIL), NIL)), NIL);
        continuationtest = listO(3, SYM_FOREACH_STELLA_NOT, listO(3, SYM_FOREACH_STELLA_NILp, iteratorvar, NIL), NIL);
      }
      else if (collectionbasetype == SGT_FOREACH_STELLA_ABSTRACT_ITERATOR) {
        bumptrees = NIL;
        continuationtest = listO(3, SYM_FOREACH_STELLA_NEXTp, iteratorvar, NIL);
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << collectionbasetype << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
      valueclauses = cons(cons(valuevar, cons(extract_parameter_type(collectiontype, SYM_FOREACH_STELLA_ANY_VALUE, dummy1), cons(optimize_progn(listO(3, SYM_FOREACH_STELLA_PROGN, listO(3, SYM_FOREACH_STELLA_SETQ, valuevar, cons(valuetree, NIL)), bumptrees->concatenate(NIL, 0))), NIL))), valueclauses);
      if (((boolean)(keyvar))) {
        { Standard_Object* keytype = NULL;
          boolean parameterexistsP = FALSE;

          keytype = extract_parameter_type(collectiontype, SYM_FOREACH_STELLA_ANY_KEY, parameterexistsP);
          if (parameterexistsP) {
            valueclauses = cons(cons(keyvar, cons(keytype, cons(listO(3, SYM_FOREACH_STELLA_SETQ, keyvar, cons(keytree, NIL)), NIL))), valueclauses);
          }
          else {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal key variable specified for keyless collection " << "`" << de_uglify_parse_tree(collectiontype) << "'" << "." << endl;
              }
            }
            _return1 = NIL;
            _return2 = SYM_FOREACH_STELLA_FALSE;
            return (NIL);
          }
        }
      }
      _return1 = valueclauses;
      _return2 = continuationtest;
      return (iteratorclauses);
    }
  }
}

Cons* yield_in_cursor_clauses_for_integer_interval(Cons* collectiontree, Standard_Object* collectiontype, Symbol* keyvar, Symbol* valuevar, Cons*& _return1, Object*& _return2) {
  collectiontype = collectiontype;
  if (((boolean)(keyvar))) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_error();
      if (!(suppress_warningsP())) {
        print_error_context(">> ERROR: ", STANDARD_ERROR);
        *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal key variable specified for INTEGER-INTERVAL" << "." << endl;
      }
    }
    _return1 = NIL;
    _return2 = SYM_FOREACH_STELLA_FALSE;
    return (NIL);
  }
  { Object* upperbound = collectiontree->rest->rest->value;
    boolean knownunboundedP = (upperbound == SYM_FOREACH_STELLA_NULL) ||
        (upperbound == SYM_FOREACH_STELLA_NULL_INTEGER);
    boolean knownboundedP = (!knownunboundedP) &&
        (isaP(upperbound, SGT_FOREACH_STELLA_NUMBER_WRAPPER) ||
         (consP(upperbound) &&
          get_quoted_tree("((+ - * / 1+ 1- ++ -- LENGTH) \"/STELLA\")", "/STELLA")->memberP(((Cons*)(upperbound))->value)));
    Symbol* iteratorvar = local_gensym("ITER");
    Symbol* upperboundvar = (knownunboundedP ? ((Symbol*)(NULL)) : local_gensym("UPPER-BOUND"));
    Symbol* unboundedvar = ((knownunboundedP ||
        knownboundedP) ? ((Symbol*)(NULL)) : local_gensym("UNBOUNDED?"));

    { Cons* _return_temp = cons(cons(iteratorvar, listO(3, SGT_FOREACH_STELLA_INTEGER, collectiontree->rest->value, NIL)), ((knownunboundedP ? NIL : cons(cons(upperboundvar, listO(3, SGT_FOREACH_STELLA_INTEGER, upperbound, NIL)), NIL)))->concatenate((((knownunboundedP ||
          knownboundedP) ? NIL : cons(cons(unboundedvar, listO(3, SGT_FOREACH_STELLA_BOOLEAN, listO(3, SYM_FOREACH_STELLA_NULLp, upperboundvar, NIL), NIL)), NIL)))->concatenate(NIL, 0), 0));

      _return1 = cons(cons(valuevar, listO(3, SGT_FOREACH_STELLA_INTEGER, listO(4, SYM_FOREACH_STELLA_PROGN, listO(3, SYM_FOREACH_STELLA_SETQ, valuevar, cons(iteratorvar, NIL)), listO(3, SYM_FOREACH_STELLA_ii, iteratorvar, NIL), NIL), NIL)), NIL);
      _return2 = (knownunboundedP ? ((Standard_Object*)(SYM_FOREACH_STELLA_TRUE)) : ((Standard_Object*)((knownboundedP ? listO(3, SYM_FOREACH_STELLA_le, iteratorvar, cons(upperboundvar, NIL)) : listO(3, SYM_FOREACH_STELLA_OR, unboundedvar, cons(listO(3, SYM_FOREACH_STELLA_le, iteratorvar, cons(upperboundvar, NIL)), NIL))))));
      return (_return_temp);
    }
  }
}

Cons* yield_in_cursor_clauses_for_vector(Object* collectiontree, Standard_Object* collectiontype, Symbol* keyvar, Symbol* valuevar, Cons*& _return1, Object*& _return2) {
  { boolean dummy1;

    if (((boolean)(keyvar))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal key variable specified for " << "`" << de_uglify_parse_tree(collectiontype) << "'" << "." << endl;
        }
      }
      _return1 = NIL;
      _return2 = SYM_FOREACH_STELLA_FALSE;
      return (NIL);
    }
    { Symbol* vectorvar = local_gensym("VECTOR");
      Symbol* indexvar = local_gensym("INDEX");
      Symbol* lengthvar = local_gensym("LENGTH");

      { Cons* _return_temp = listO(4, cons(vectorvar, cons(collectiontype, cons(collectiontree, NIL))), cons(indexvar, listO(3, SGT_FOREACH_STELLA_INTEGER, wrap_integer(0), NIL)), cons(lengthvar, listO(3, SGT_FOREACH_STELLA_INTEGER, listO(3, SYM_FOREACH_STELLA_LENGTH, vectorvar, NIL), NIL)), NIL);

        _return1 = cons(cons(valuevar, cons((((collectiontype == SGT_FOREACH_STELLA_STRING) ||
            (collectiontype == SGT_FOREACH_STELLA_MUTABLE_STRING)) ? SGT_FOREACH_STELLA_CHARACTER : ((Standard_Object*)(extract_parameter_type(collectiontype, SYM_FOREACH_STELLA_ANY_VALUE, dummy1)))), cons(listO(4, SYM_FOREACH_STELLA_PROGN, listO(3, SYM_FOREACH_STELLA_SETQ, valuevar, cons(listO(3, SYM_FOREACH_STELLA_NTH, vectorvar, cons(indexvar, NIL)), NIL)), listO(3, SYM_FOREACH_STELLA_ii, indexvar, NIL), NIL), NIL))), NIL);
        _return2 = listO(3, SYM_FOREACH_STELLA_l, indexvar, cons(lengthvar, NIL));
        return (_return_temp);
      }
    }
  }
}

Cons* yield_in_cursor_clauses_for_argument_list(Object* collectiontree, Standard_Object* collectiontype, Symbol* keyvar, Symbol* valuevar, Cons*& _return1, Object*& _return2) {
  { boolean dummy1;
    Standard_Object* dummy2;
    boolean dummy3;
    Standard_Object* dummy4;

    if (((boolean)(keyvar))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal key variable specified for " << "`" << de_uglify_parse_tree(collectiontype) << "'" << "." << endl;
        }
      }
      _return1 = NIL;
      _return2 = SYM_FOREACH_STELLA_FALSE;
      return (NIL);
    }
    if (!pass_variable_arguments_as_listP(oMETHODBEINGWALKEDo.get())) {
      { Parametric_Type_Specifier* self_000 = new_parametric_type_specifier();

        self_000->specifier_base_type = SGT_FOREACH_STELLA_ARGUMENT_LIST_ITERATOR;
        self_000->specifier_parameter_types = list(1, extract_parameter_type(collectiontype, SYM_FOREACH_STELLA_ANY_VALUE, dummy1));
        { Parametric_Type_Specifier* iteratortype = self_000;

          return (yield_in_cursor_clauses_for_general_collection(sys_tree(listO(5, SYM_FOREACH_STELLA_SYS_CALL_METHOD, SGT_FOREACH_STELLA_ARGUMENT_LIST_ITERATOR, SYM_FOREACH_STELLA_ALLOCATE_ITERATOR, walk_without_type_tree(collectiontree), NIL), iteratortype, dummy2), iteratortype, keyvar, valuevar, _return1, _return2));
        }
      }
    }
    { Standard_Object* listifiedtype = yield_listified_variable_arguments_type(oMETHODBEINGWALKEDo.get());
      Standard_Object* elementtype = extract_parameter_type(collectiontype, SYM_FOREACH_STELLA_ANY_VALUE, dummy3);

      { Cons* iteratorclauses = NULL;
        Cons* valueclauses = NULL;
        Object* continuationtest = NULL;

        iteratorclauses = yield_in_cursor_clauses_for_general_collection(sys_tree(collectiontree, listifiedtype, dummy4), listifiedtype, keyvar, valuevar, valueclauses, continuationtest);
        if (sub_type_spec_ofP(elementtype, SGT_FOREACH_STELLA_LITERAL)) {
          { Cons* clause = NULL;
            Cons* iter_000 = ((Cons*)(valueclauses));

            while (!(iter_000 == NIL)) {
              clause = ((Cons*)(iter_000->value));
              iter_000 = iter_000->rest;
              if (clause->value == valuevar) {
                clause->second_setter(elementtype);
                break;
              }
            }
          }
        }
        _return1 = valueclauses;
        _return2 = continuationtest;
        return (iteratorclauses);
      }
    }
  }
}

Cons* yield_on_cursor_clauses(Cons* ontree, Cons*& _return1, Object*& _return2) {
  { Object* iteratorvar = ontree->rest->value;

    if (!symbolP(iteratorvar)) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal ON clause with non-atomic variable: " << "`" << de_uglify_parse_tree(ontree) << "'" << "." << endl;
        }
      }
      _return1 = NIL;
      _return2 = SYM_FOREACH_STELLA_FALSE;
      return (NIL);
    }
    { Object* collectiontree = NULL;
      Standard_Object* collectiontype = NULL;

      collectiontree = walk_collection_tree(ontree->rest->rest->value, TRUE, collectiontype);
      if (sub_type_spec_ofP(collectiontype, SGT_FOREACH_STELLA_ARGUMENT_LIST)) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "ON iteration not supported for ARGUMENT-LISTs: " << "`" << de_uglify_parse_tree(ontree) << "'" << "." << endl;
          }
        }
        _return1 = NIL;
        _return2 = SYM_FOREACH_STELLA_FALSE;
        return (NIL);
      }
      ontree->third_setter(NULL);
      if (!safe_subtype_ofP(type_spec_to_base_type(collectiontype), SGT_FOREACH_STELLA_ABSTRACT_ITERATOR)) {
        collectiontree = listO(3, SYM_FOREACH_STELLA_ALLOCATE_ITERATOR, collectiontree, NIL);
      }
      { Cons* _return_temp = cons(cons(iteratorvar, listO(3, SGT_FOREACH_STELLA_ABSTRACT_ITERATOR, collectiontree, NIL)), NIL);

        _return1 = NIL;
        _return2 = listO(3, SYM_FOREACH_STELLA_NEXTp, iteratorvar, NIL);
        return (_return_temp);
      }
    }
  }
}

Cons* collect_cursor_clauses(Cons* cursorstree, boolean dontoptimizeP, Cons*& _return1, Object*& _return2) {
  { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(cursorstree->value));

    if (test_value_000 == SYM_FOREACH_STELLA_IN) {
      return (yield_in_cursor_clauses(cursorstree, dontoptimizeP, _return1, _return2));
    }
    else if (test_value_000 == SYM_FOREACH_STELLA_ON) {
      return (yield_on_cursor_clauses(cursorstree, _return1, _return2));
    }
    else if (test_value_000 == SYM_FOREACH_STELLA_AS) {
      { Cons* alliteratorclauses = NIL;
        Cons* allvalueclauses = NIL;
        Cons* continuationtests = NIL;

        { Object* clause = NULL;
          Cons* iter_000 = cursorstree->rest;

          while (!(iter_000 == NIL)) {
            clause = iter_000->value;
            iter_000 = iter_000->rest;
            { Cons* iteratorclauses = NULL;
              Cons* valueclauses = NULL;
              Object* continuationtest = NULL;

              iteratorclauses = collect_cursor_clauses(((Cons*)(clause)), dontoptimizeP, valueclauses, continuationtest);
              alliteratorclauses = alliteratorclauses->concatenate(iteratorclauses, 0);
              allvalueclauses = allvalueclauses->concatenate(valueclauses, 0);
              continuationtests = cons(continuationtest, continuationtests);
            }
          }
        }
        { Cons* _return_temp = alliteratorclauses;

          _return1 = allvalueclauses;
          _return2 = ((continuationtests->length() == 1) ? continuationtests->value : cons(SYM_FOREACH_STELLA_AND, continuationtests->reverse()->remove(SYM_FOREACH_STELLA_TRUE)->concatenate(NIL, 0)));
          return (_return_temp);
        }
      }
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

Cons* walk_cursors_tree(Cons* cursorstree, boolean dontoptimizeP, Cons*& _return1, Object*& _return2) {
  { Standard_Object* dummy1;

    { Cons* iteratorclauses = NULL;
      Cons* valueclauses = NULL;
      Object* continuationtest = NULL;

      iteratorclauses = collect_cursor_clauses(cursorstree, dontoptimizeP, valueclauses, continuationtest);
      { Cons* clause = NULL;
        Cons* iter_000 = ((Cons*)(iteratorclauses));

        while (!(iter_000 == NIL)) {
          clause = ((Cons*)(iter_000->value));
          iter_000 = iter_000->rest;
          { Object* iteratortree = NULL;
            Standard_Object* iteratortype = NULL;

            iteratortree = walk_expression_tree(clause->rest->rest->value, ((Standard_Object*)(clause->rest->value)), SYM_FOREACH_STELLA_FOREACH, FALSE, iteratortype);
            clause->second_setter(iteratortype);
            clause->third_setter(coerce_a_tree(iteratortree, iteratortype, iteratortype, dummy1));
          }
          push_variable_binding(((Symbol*)(clause->value)), ((Standard_Object*)(clause->rest->value)));
        }
      }
      { Cons* clause = NULL;
        Cons* iter_001 = ((Cons*)(valueclauses));

        while (!(iter_001 == NIL)) {
          clause = ((Cons*)(iter_001->value));
          iter_001 = iter_001->rest;
          push_variable_binding(((Symbol*)(clause->value)), ((Standard_Object*)(clause->rest->value)));
          clause->third_setter(walk_without_type_tree(clause->rest->rest->value));
        }
      }
      continuationtest = walk_without_type_tree(continuationtest);
      { int i = NULL_INTEGER;
        int iter_002 = 1;
        int upper_bound_000 = iteratorclauses->length() + valueclauses->length();

        while (iter_002 <= upper_bound_000) {
          i = iter_002;
          iter_002 = iter_002 + 1;
          i = i;
          pop_variable_binding();
        }
      }
      _return1 = valueclauses;
      _return2 = continuationtest;
      return (iteratorclauses);
    }
  }
}

Cons* yield_collect_into_list_code(Symbol* collectvariable, Object* intovariable, Object* collecttree) {
  return (cons(listO(5, SYM_FOREACH_STELLA_IF, listO(3, SYM_FOREACH_STELLA_NULLp, collectvariable, NIL), listO(4, SYM_FOREACH_STELLA_PROGN, listO(3, SYM_FOREACH_STELLA_SETQ, collectvariable, cons(listO(3, SYM_FOREACH_STELLA_CONS, collecttree, cons(SYM_FOREACH_STELLA_NIL, NIL)), NIL)), listO(5, SYM_FOREACH_STELLA_IF, listO(3, SYM_FOREACH_STELLA_EMPTYp, listO(3, SYM_FOREACH_STELLA_THE_CONS_LIST, intovariable, NIL), NIL), listO(4, SYM_FOREACH_STELLA_SETF, listO(3, SYM_FOREACH_STELLA_THE_CONS_LIST, intovariable, NIL), collectvariable, NIL), listO(4, SYM_FOREACH_STELLA_ADD_CONS_TO_END_OF_CONS_LIST, listO(3, SYM_FOREACH_STELLA_THE_CONS_LIST, intovariable, NIL), collectvariable, NIL), NIL), NIL), listO(4, SYM_FOREACH_STELLA_PROGN, listO(4, SYM_FOREACH_STELLA_SETF, listO(3, SYM_FOREACH_STELLA_REST, collectvariable, NIL), listO(3, SYM_FOREACH_STELLA_CONS, copy_cons_tree(collecttree), cons(SYM_FOREACH_STELLA_NIL, NIL)), NIL), listO(3, SYM_FOREACH_STELLA_SETQ, collectvariable, cons(listO(3, SYM_FOREACH_STELLA_REST, collectvariable, NIL), NIL)), NIL), NIL), NIL));
}

Cons* yield_collect_into_cons_code(Symbol* collectvariable, Object* intovariable, Object* collecttree) {
  return (cons(listO(5, SYM_FOREACH_STELLA_IF, listO(3, SYM_FOREACH_STELLA_NULLp, collectvariable, NIL), listO(4, SYM_FOREACH_STELLA_PROGN, listO(3, SYM_FOREACH_STELLA_SETQ, collectvariable, cons(listO(3, SYM_FOREACH_STELLA_CONS, collecttree, cons(SYM_FOREACH_STELLA_NIL, NIL)), NIL)), listO(5, SYM_FOREACH_STELLA_IF, listO(3, SYM_FOREACH_STELLA_NILp, intovariable, NIL), listO(3, SYM_FOREACH_STELLA_SETQ, intovariable, cons(collectvariable, NIL)), listO(3, SYM_FOREACH_STELLA_ADD_CONS_TO_END_OF_CONS_LIST, intovariable, cons(collectvariable, NIL)), NIL), NIL), listO(4, SYM_FOREACH_STELLA_PROGN, listO(4, SYM_FOREACH_STELLA_SETF, listO(3, SYM_FOREACH_STELLA_REST, collectvariable, NIL), listO(3, SYM_FOREACH_STELLA_CONS, copy_cons_tree(collecttree), cons(SYM_FOREACH_STELLA_NIL, NIL)), NIL), listO(3, SYM_FOREACH_STELLA_SETQ, collectvariable, cons(listO(3, SYM_FOREACH_STELLA_REST, collectvariable, NIL), NIL)), NIL), NIL), NIL));
}

Cons* yield_collect_code(Symbol* collectvariable, Object* intovariable, Object* collecttree) {
  { Standard_Object* dummy1;

    { Surrogate* intovariabletype = type_spec_to_base_type(lookup_variable_type(((Symbol*)(intovariable))));

      if (subtype_ofP(intovariabletype, SGT_FOREACH_STELLA_LIST)) {
        return (yield_collect_into_list_code(collectvariable, intovariable, collecttree));
      }
      else if (subtype_ofP(intovariabletype, SGT_FOREACH_STELLA_CONS)) {
        return (yield_collect_into_cons_code(collectvariable, intovariable, collecttree));
      }
      else if (intovariabletype == SGT_FOREACH_STELLA_UNKNOWN) {
        return (cons(listO(5, SYM_FOREACH_STELLA_IF, listO(3, SYM_FOREACH_STELLA_NULLp, collectvariable, NIL), listO(4, SYM_FOREACH_STELLA_PROGN, listO(3, SYM_FOREACH_STELLA_SETQ, collectvariable, cons(listO(3, SYM_FOREACH_STELLA_CONS, collecttree, cons(SYM_FOREACH_STELLA_NIL, NIL)), NIL)), listO(5, SYM_FOREACH_STELLA_COND, listO(3, listO(3, SYM_FOREACH_STELLA_NILp, intovariable, NIL), listO(3, SYM_FOREACH_STELLA_SETQ, intovariable, cons(collectvariable, NIL)), NIL), listO(3, listO(3, SYM_FOREACH_STELLA_CONSp, intovariable, NIL), listO(3, SYM_FOREACH_STELLA_ADD_CONS_TO_END_OF_CONS_LIST, intovariable, cons(collectvariable, NIL)), NIL), listO(3, SYM_FOREACH_STELLA_OTHERWISE, listO(5, SYM_FOREACH_STELLA_IF, listO(3, SYM_FOREACH_STELLA_EMPTYp, listO(3, SYM_FOREACH_STELLA_THE_CONS_LIST, intovariable, NIL), NIL), listO(4, SYM_FOREACH_STELLA_SETF, listO(3, SYM_FOREACH_STELLA_THE_CONS_LIST, intovariable, NIL), collectvariable, NIL), listO(4, SYM_FOREACH_STELLA_ADD_CONS_TO_END_OF_CONS_LIST, listO(3, SYM_FOREACH_STELLA_THE_CONS_LIST, intovariable, NIL), collectvariable, NIL), NIL), NIL), NIL), NIL), listO(4, SYM_FOREACH_STELLA_PROGN, listO(4, SYM_FOREACH_STELLA_SETF, listO(3, SYM_FOREACH_STELLA_REST, collectvariable, NIL), listO(3, SYM_FOREACH_STELLA_CONS, copy_cons_tree(collecttree), cons(SYM_FOREACH_STELLA_NIL, NIL)), NIL), listO(3, SYM_FOREACH_STELLA_SETQ, collectvariable, cons(listO(3, SYM_FOREACH_STELLA_REST, collectvariable, NIL), NIL)), NIL), NIL), NIL));
      }
      else {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Can't collect into a datatype other than LIST or CONS." << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(collecttree, SGT_FOREACH_STELLA_VOID, dummy1));
      }
    }
  }
}

Cons* walk_prefix_collect_into_tree(Cons* collectintotree, Cons* iterators, Cons* valuebindings, Cons* docode, Cons*& _return1) {
  { Object* collecttree = collectintotree->rest->value;
    Object* intotree = collectintotree->rest->rest->value;
    Symbol* collectvariable = local_gensym("COLLECT");
    Cons* collectintobindings = cons(cons(collectvariable, listO(3, SGT_FOREACH_STELLA_CONS, SYM_FOREACH_STELLA_NULL, NIL)), NIL);
    Cons* collectcode = NULL;
    Symbol* intovariable = NULL;
    Standard_Object* intovariabletype = NULL;

    if (symbolP(intotree)) {
      {
        intovariable = ((Symbol*)(intotree));
        collectcode = yield_collect_code(collectvariable, intovariable, collecttree);
      }
    }
    else {
      {
        { Cons* d = NULL;
          Cons* iter_000 = ((Cons*)(valuebindings));

          while (!(iter_000 == NIL)) {
            d = ((Cons*)(iter_000->value));
            iter_000 = iter_000->rest;
            push_variable_binding(((Symbol*)(d->value)), ((Standard_Object*)(d->rest->value)));
          }
        }
        intotree = walk_expression_tree(intotree, NULL, SYM_FOREACH_STELLA_FOREACH, FALSE, intovariabletype);
        { Object* d = NULL;
          Cons* iter_001 = valuebindings;

          while (!(iter_001 == NIL)) {
            d = iter_001->value;
            iter_001 = iter_001->rest;
                        pop_variable_binding();
          }
        }
        intovariable = local_gensym("INTO");
        collectintobindings = cons(cons_list(3, intovariable, intovariabletype, intotree), collectintobindings);
        push_variable_binding(intovariable, intovariabletype);
        collectcode = yield_collect_code(collectvariable, intovariable, collecttree);
        pop_variable_binding();
      }
    }
    { Cons* _return_temp = iterators->concatenate(collectintobindings, 0);

      _return1 = docode->concatenate(collectcode, 0);
      return (_return_temp);
    }
  }
}

Cons* wrap_where_test(Object* wheretest, Cons* actions) {
  if (((boolean)(wheretest))) {
    return (cons(listO(3, SYM_FOREACH_STELLA_WHEN, wheretest, actions->concatenate(NIL, 0)), NIL));
  }
  else {
    return (actions);
  }
}

Cons* walk_prefix_foreach_body(Cons* iteratorclauses, Cons* valueclauses, Object* wheretest, Cons* doactions) {
  { Cons* otree = NULL;

    { Cons* c = NULL;
      Cons* iter_000 = ((Cons*)(iteratorclauses));

      while (!(iter_000 == NIL)) {
        c = ((Cons*)(iter_000->value));
        iter_000 = iter_000->rest;
        push_variable_binding(((Symbol*)(c->value)), ((Standard_Object*)(c->rest->value)));
        c->first_setter(true_variable_name(((Symbol*)(c->value))));
      }
    }
    { Cons* c = NULL;
      Cons* iter_001 = ((Cons*)(valueclauses));

      while (!(iter_001 == NIL)) {
        c = ((Cons*)(iter_001->value));
        iter_001 = iter_001->rest;
        push_variable_binding(((Symbol*)(c->value)), ((Standard_Object*)(c->rest->value)));
        c->first_setter(true_variable_name(((Symbol*)(c->value))));
      }
    }
    otree = walk_list_of_trees(wrap_where_test(wheretest, doactions));
    { Object* c = NULL;
      Cons* iter_002 = valueclauses;

      while (!(iter_002 == NIL)) {
        c = iter_002->value;
        iter_002 = iter_002->rest;
                pop_variable_binding();
      }
    }
    { Object* c = NULL;
      Cons* iter_003 = iteratorclauses;

      while (!(iter_003 == NIL)) {
        c = iter_003->value;
        iter_003 = iter_003->rest;
                pop_variable_binding();
      }
    }
    return (otree);
  }
}

Cons* walk_prefix_foreach_tree(Cons* tree, Standard_Object*& _return1) {
  { Cons* tail = NULL;
    Cons* iteratorclauses = NULL;
    Cons* valueclauses = NULL;
    Object* continuationtest = NULL;
    Object* wheretest = NULL;
    Cons* doactions = NIL;

    iteratorclauses = walk_cursors_tree(((Cons*)(tree->rest->value)), FALSE, valueclauses, continuationtest);
    tree->second_setter(NULL);
    tail = ((Cons*)(tree->rest->rest));
    if (((Cons*)(tail->value))->value == SYM_FOREACH_STELLA_WHERE) {
      wheretest = ((Cons*)(tail->value))->rest->value;
      ((Cons*)(tail->value))->second_setter(NULL);
      tail = tail->rest;
    }
    if (((Cons*)(tail->value))->value == SYM_FOREACH_STELLA_DO) {
      doactions = ((Cons*)(tail->value))->rest;
      ((Cons*)(tail->value))->rest = NIL;
      tail = tail->rest;
    }
    if (!(tail == NIL)) {
      iteratorclauses = walk_prefix_collect_into_tree(((Cons*)(tail->value)), iteratorclauses, valueclauses, doactions, doactions);
      tail->first_setter(NULL);
    }
    _return1 = SGT_FOREACH_STELLA_VOID;
    return (listO(3, SYM_FOREACH_STELLA_SYS_FOREACH, iteratorclauses, cons(valueclauses, cons(continuationtest, walk_prefix_foreach_body(iteratorclauses, valueclauses, wheretest, doactions)->concatenate(NIL, 0)))));
  }
}

Cons* walk_exists_tree(Cons* tree, Standard_Object*& _return1) {
  { Symbol* foundP = local_gensym("FOUND?");
    Cons* otree = listO(5, SYM_FOREACH_STELLA_VRLET, cons(cons(foundP, cons(SYM_FOREACH_STELLA_FALSE, NIL)), NIL), cons(SYM_FOREACH_STELLA_FOREACH, tree->rest->concatenate(listO(4, SYM_FOREACH_STELLA_DO, listO(3, SYM_FOREACH_STELLA_SETQ, foundP, cons(SYM_FOREACH_STELLA_TRUE, NIL)), cons(SYM_FOREACH_STELLA_BREAK, NIL), NIL), 0)), foundP, NIL);

    return (walk_a_cons_tree(otree, _return1));
  }
}

Cons* walk_forall_tree(Cons* tree, Standard_Object*& _return1) {
  { Symbol* alwaysP = local_gensym("ALWAYS?");
    Cons* iteratorsandwhere = NIL;
    Cons* alwaystest = NIL;
    Cons* precursor = NIL;
    Cons* cursor = tree;

    while (!(cursor == NIL)) {
      if (cursor->value == SYM_FOREACH_STELLA_ALWAYS) {
        alwaystest = ((Cons*)(cursor->rest->value));
        precursor->rest = NIL;
        break;
      }
      precursor = cursor;
      cursor = cursor->rest;
    }
    iteratorsandwhere = tree->rest;
    _return1 = SGT_FOREACH_STELLA_BOOLEAN;
    return (((Cons*)(listO(5, SYM_FOREACH_STELLA_VRLET, cons(cons(alwaysP, cons(SYM_FOREACH_STELLA_TRUE, NIL)), NIL), cons(SYM_FOREACH_STELLA_FOREACH, iteratorsandwhere->concatenate(listO(3, SYM_FOREACH_STELLA_DO, listO(5, SYM_FOREACH_STELLA_WHEN, listO(3, SYM_FOREACH_STELLA_NOT, alwaystest, NIL), listO(3, SYM_FOREACH_STELLA_SETQ, alwaysP, cons(SYM_FOREACH_STELLA_FALSE, NIL)), cons(SYM_FOREACH_STELLA_BREAK, NIL), NIL), NIL), 0)), alwaysP, NIL))));
  }
}

Cons* walk_some_tree(Cons* tree, Standard_Object*& _return1) {
  { boolean dummy1;

    { Symbol* valuevar = local_gensym("VALUE");
      Object* iterationvar = tree->rest->value;
      Cons* otree = NULL;

      { Object* sequenceref = NULL;
        Standard_Object* sequencetype = NULL;

        sequenceref = walk_a_tree(tree->fourth(), sequencetype);
        tree->fourth_setter(sequenceref);
        otree = listO(5, SYM_FOREACH_STELLA_VRLET, cons(cons(valuevar, cons(extract_parameter_type(sequencetype, SYM_FOREACH_STELLA_ANY_VALUE, dummy1), cons(SYM_FOREACH_STELLA_NULL, NIL))), NIL), cons(SYM_FOREACH_STELLA_FOREACH, tree->rest->concatenate(listO(4, SYM_FOREACH_STELLA_DO, listO(3, SYM_FOREACH_STELLA_SETQ, valuevar, cons(iterationvar, NIL)), cons(SYM_FOREACH_STELLA_BREAK, NIL), NIL), 0)), valuevar, NIL);
        return (walk_a_cons_tree(otree, _return1));
      }
    }
  }
}

Cons* extract_setof_nested_iterator_tree(Cons* iteratorclauses) {
  { Cons* iterclauses = ((Cons*)(iteratorclauses));
    Object* subtree = ((Cons*)(iterclauses->value))->rest->rest->value;

    ((Cons*)(iterclauses->value))->third_setter(NULL);
    if (!(iterclauses->rest == NIL)) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Too many iterators in 'setof' expression" << "." << endl;
        }
      }
    }
    return (((Cons*)(subtree)));
  }
}

Symbol* yield_setof_next_function_name(Cons* valueclauses, Cons* wheretest) {
  { Cons* onlyvalueclause = ((Cons*)(valueclauses->value));
    Object* valuevariable = onlyvalueclause->value;
    Object* valuetype = onlyvalueclause->rest->value;
    Symbol* nextfunctionname = NULL;

    nextfunctionname = gensym("SETOF-NEXT-P");
    walk_auxiliary_tree(listO(5, SYM_FOREACH_STELLA_DEFUN, cons(nextfunctionname, cons(SYM_FOREACH_STELLA_BOOLEAN, NIL)), cons(listO(3, SYM_FOREACH_STELLA_SELF, SYM_FOREACH_STELLA_ALL_PURPOSE_ITERATOR, NIL), NIL), listO(5, SYM_FOREACH_STELLA_LET, listO(3, listO(3, SYM_FOREACH_STELLA_NESTEDITERATOR, listO(3, SYM_FOREACH_STELLA_ITERATOR_NESTED_ITERATOR, SYM_FOREACH_STELLA_SELF, NIL), NIL), cons(valuevariable, cons(valuetype, cons(SYM_FOREACH_STELLA_NULL, NIL))), NIL), listO(5, SYM_FOREACH_STELLA_WHILE, listO(3, SYM_FOREACH_STELLA_NEXTp, SYM_FOREACH_STELLA_NESTEDITERATOR, NIL), listO(3, SYM_FOREACH_STELLA_SETQ, valuevariable, cons(listO(3, SYM_FOREACH_STELLA_VALUE, SYM_FOREACH_STELLA_NESTEDITERATOR, NIL), NIL)), listO(3, SYM_FOREACH_STELLA_WHEN, wheretest, listO(3, listO(4, SYM_FOREACH_STELLA_SETF, listO(3, SYM_FOREACH_STELLA_VALUE, SYM_FOREACH_STELLA_SELF, NIL), valuevariable, NIL), listO(3, SYM_FOREACH_STELLA_RETURN, SYM_FOREACH_STELLA_TRUE, NIL), NIL)), NIL), listO(3, SYM_FOREACH_STELLA_RETURN, SYM_FOREACH_STELLA_FALSE, NIL), NIL), NIL));
    return (nextfunctionname);
  }
}

Cons* walk_setof_tree(Cons* tree, Standard_Object*& _return1) {
  tree = foreach_to_prefix_foreach_tree(tree);
  { Cons* iteratorssubtree = ((Cons*)(tree->rest->value));
    Symbol* iteratorvariable = local_gensym("ITER");
    Cons* tail = NULL;
    Cons* iteratorclauses = NULL;
    Cons* valueclauses = NULL;
    Object* continuationtest = NULL;
    Object* wheretest = NULL;

    tree->second_setter(NULL);
    if (get_quoted_tree("((ON AS) \"/STELLA\")", "/STELLA")->memberP(iteratorssubtree->value)) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal reserved word " << "`" << iteratorssubtree->value << "'" << " in 'setof' expression." << "." << endl;
        }
      }
    }
    iteratorclauses = walk_cursors_tree(iteratorssubtree, TRUE, valueclauses, continuationtest);
    tail = ((Cons*)(tree->rest->rest));
    if ((!(tail == NIL)) &&
        (((Cons*)(tail->value))->value == SYM_FOREACH_STELLA_WHERE)) {
      wheretest = ((Cons*)(tail->value))->rest->value;
      ((Cons*)(tail->value))->second_setter(NULL);
      tail = tail->rest;
    }
    if (!(tail == NIL)) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal 'setof' expression.  Syntax is " << endl << "   (SETOF <iterators> WHERE <boolean expression>)" << "." << endl;
        }
      }
    }
    _return1 = SGT_FOREACH_STELLA_ALL_PURPOSE_ITERATOR;
    return (listO(6, SYM_FOREACH_STELLA_VRLET, cons(cons(iteratorvariable, cons(listO(3, SYM_FOREACH_STELLA_NEW, SYM_FOREACH_STELLA_ALL_PURPOSE_ITERATOR, NIL), NIL)), NIL), listO(4, SYM_FOREACH_STELLA_SETF, listO(3, SYM_FOREACH_STELLA_ITERATOR_NESTED_ITERATOR, iteratorvariable, NIL), extract_setof_nested_iterator_tree(iteratorclauses), NIL), listO(4, SYM_FOREACH_STELLA_SETF, listO(3, SYM_FOREACH_STELLA_ITERATOR_NEXT_CODE, iteratorvariable, NIL), listO(4, SYM_FOREACH_STELLA_THE_CODE, KWD_FOREACH_FUNCTION, yield_setof_next_function_name(valueclauses, ((Cons*)(wheretest))), NIL), NIL), iteratorvariable, NIL));
  }
}

void help_startup_foreach1() {
  {
    SYM_FOREACH_STELLA_AS = ((Symbol*)(intern_rigid_symbol_wrt_module("AS", NULL, 0)));
    SYM_FOREACH_STELLA_COMMA = ((Symbol*)(intern_rigid_symbol_wrt_module("COMMA", NULL, 0)));
    SYM_FOREACH_STELLA_DO = ((Symbol*)(intern_rigid_symbol_wrt_module("DO", NULL, 0)));
    SYM_FOREACH_STELLA_COLLECT = ((Symbol*)(intern_rigid_symbol_wrt_module("COLLECT", NULL, 0)));
    SGT_FOREACH_STELLA_VOID = ((Surrogate*)(intern_rigid_symbol_wrt_module("VOID", NULL, 1)));
    SYM_FOREACH_STELLA_PREFIX_FOREACH = ((Symbol*)(intern_rigid_symbol_wrt_module("PREFIX-FOREACH", NULL, 0)));
    SYM_FOREACH_STELLA_IN = ((Symbol*)(intern_rigid_symbol_wrt_module("IN", NULL, 0)));
    SYM_FOREACH_STELLA_X = ((Symbol*)(intern_rigid_symbol_wrt_module("X", NULL, 0)));
    SYM_FOREACH_STELLA_NIL = ((Symbol*)(intern_rigid_symbol_wrt_module("NIL", NULL, 0)));
    SYM_FOREACH_STELLA_WHERE = ((Symbol*)(intern_rigid_symbol_wrt_module("WHERE", NULL, 0)));
    SYM_FOREACH_STELLA_INTO = ((Symbol*)(intern_rigid_symbol_wrt_module("INTO", NULL, 0)));
    SYM_FOREACH_STELLA_COLLECT_INTO = ((Symbol*)(intern_rigid_symbol_wrt_module("COLLECT-INTO", NULL, 0)));
    SYM_FOREACH_STELLA_FOREACH = ((Symbol*)(intern_rigid_symbol_wrt_module("FOREACH", NULL, 0)));
    SYM_FOREACH_STELLA_EXISTS = ((Symbol*)(intern_rigid_symbol_wrt_module("EXISTS", NULL, 0)));
    SYM_FOREACH_STELLA_FORALL = ((Symbol*)(intern_rigid_symbol_wrt_module("FORALL", NULL, 0)));
    SYM_FOREACH_STELLA_SOME = ((Symbol*)(intern_rigid_symbol_wrt_module("SOME", NULL, 0)));
    SYM_FOREACH_STELLA_SETOF = ((Symbol*)(intern_rigid_symbol_wrt_module("SETOF", NULL, 0)));
    SYM_FOREACH_STELLA_SET_OF = ((Symbol*)(intern_rigid_symbol_wrt_module("SET-OF", NULL, 0)));
    SGT_FOREACH_STELLA_CONS = ((Surrogate*)(intern_rigid_symbol_wrt_module("CONS", NULL, 1)));
    SYM_FOREACH_STELLA_INTERVAL = ((Symbol*)(intern_rigid_symbol_wrt_module("INTERVAL", NULL, 0)));
    SGT_FOREACH_STELLA_INTEGER_INTERVAL = ((Surrogate*)(intern_rigid_symbol_wrt_module("INTEGER-INTERVAL", NULL, 1)));
    SGT_FOREACH_STELLA_UNKNOWN = ((Surrogate*)(intern_rigid_symbol_wrt_module("UNKNOWN", NULL, 1)));
    SGT_FOREACH_STELLA_ABSTRACT_ITERATOR = ((Surrogate*)(intern_rigid_symbol_wrt_module("ABSTRACT-ITERATOR", NULL, 1)));
    SYM_FOREACH_STELLA_ALLOCATE_ITERATOR = ((Symbol*)(intern_rigid_symbol_wrt_module("ALLOCATE-ITERATOR", NULL, 0)));
    SYM_FOREACH_STELLA_FALSE = ((Symbol*)(intern_rigid_symbol_wrt_module("FALSE", NULL, 0)));
    SGT_FOREACH_STELLA_ARGUMENT_LIST = ((Surrogate*)(intern_rigid_symbol_wrt_module("ARGUMENT-LIST", NULL, 1)));
    SGT_FOREACH_STELLA_STRING = ((Surrogate*)(intern_rigid_symbol_wrt_module("STRING", NULL, 1)));
    SGT_FOREACH_STELLA_MUTABLE_STRING = ((Surrogate*)(intern_rigid_symbol_wrt_module("MUTABLE-STRING", NULL, 1)));
    SGT_FOREACH_STELLA_VECTOR = ((Surrogate*)(intern_rigid_symbol_wrt_module("VECTOR", NULL, 1)));
    SGT_FOREACH_STELLA_SEQUENCE_MIXIN = ((Surrogate*)(intern_rigid_symbol_wrt_module("SEQUENCE-MIXIN", NULL, 1)));
    SYM_FOREACH_STELLA_THE_ARRAY = ((Symbol*)(intern_rigid_symbol_wrt_module("THE-ARRAY", NULL, 0)));
    SGT_FOREACH_STELLA_LIST = ((Surrogate*)(intern_rigid_symbol_wrt_module("LIST", NULL, 1)));
    SYM_FOREACH_STELLA_THE_CONS_LIST = ((Symbol*)(intern_rigid_symbol_wrt_module("THE-CONS-LIST", NULL, 0)));
    SGT_FOREACH_STELLA_KEY_VALUE_LIST = ((Surrogate*)(intern_rigid_symbol_wrt_module("KEY-VALUE-LIST", NULL, 1)));
    SGT_FOREACH_STELLA_KV_CONS = ((Surrogate*)(intern_rigid_symbol_wrt_module("KV-CONS", NULL, 1)));
    SYM_FOREACH_STELLA_THE_KV_LIST = ((Symbol*)(intern_rigid_symbol_wrt_module("THE-KV-LIST", NULL, 0)));
    SGT_FOREACH_STELLA_PROPERTY_LIST = ((Surrogate*)(intern_rigid_symbol_wrt_module("PROPERTY-LIST", NULL, 1)));
    SYM_FOREACH_STELLA_THE_PLIST = ((Symbol*)(intern_rigid_symbol_wrt_module("THE-PLIST", NULL, 0)));
    SYM_FOREACH_STELLA_KEY = ((Symbol*)(intern_rigid_symbol_wrt_module("KEY", NULL, 0)));
    SYM_FOREACH_STELLA_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("VALUE", NULL, 0)));
    SYM_FOREACH_STELLA_SETQ = ((Symbol*)(intern_rigid_symbol_wrt_module("SETQ", NULL, 0)));
    SYM_FOREACH_STELLA_REST = ((Symbol*)(intern_rigid_symbol_wrt_module("REST", NULL, 0)));
    SYM_FOREACH_STELLA_NOT = ((Symbol*)(intern_rigid_symbol_wrt_module("NOT", NULL, 0)));
    SYM_FOREACH_STELLA_NILp = ((Symbol*)(intern_rigid_symbol_wrt_module("NIL?", NULL, 0)));
    SYM_FOREACH_STELLA_DEFINEDp = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFINED?", NULL, 0)));
    SYM_FOREACH_STELLA_NEXTp = ((Symbol*)(intern_rigid_symbol_wrt_module("NEXT?", NULL, 0)));
    SYM_FOREACH_STELLA_ANY_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("ANY-VALUE", NULL, 0)));
    SYM_FOREACH_STELLA_PROGN = ((Symbol*)(intern_rigid_symbol_wrt_module("PROGN", NULL, 0)));
    SYM_FOREACH_STELLA_ANY_KEY = ((Symbol*)(intern_rigid_symbol_wrt_module("ANY-KEY", NULL, 0)));
    SYM_FOREACH_STELLA_NULL = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL", NULL, 0)));
    SYM_FOREACH_STELLA_NULL_INTEGER = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL-INTEGER", NULL, 0)));
    SGT_FOREACH_STELLA_NUMBER_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("NUMBER-WRAPPER", NULL, 1)));
    SGT_FOREACH_STELLA_INTEGER = ((Surrogate*)(intern_rigid_symbol_wrt_module("INTEGER", NULL, 1)));
    SGT_FOREACH_STELLA_BOOLEAN = ((Surrogate*)(intern_rigid_symbol_wrt_module("BOOLEAN", NULL, 1)));
    SYM_FOREACH_STELLA_NULLp = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL?", NULL, 0)));
    SYM_FOREACH_STELLA_ii = ((Symbol*)(intern_rigid_symbol_wrt_module("++", NULL, 0)));
    SYM_FOREACH_STELLA_TRUE = ((Symbol*)(intern_rigid_symbol_wrt_module("TRUE", NULL, 0)));
    SYM_FOREACH_STELLA_le = ((Symbol*)(intern_rigid_symbol_wrt_module("<=", NULL, 0)));
    SYM_FOREACH_STELLA_OR = ((Symbol*)(intern_rigid_symbol_wrt_module("OR", NULL, 0)));
    SYM_FOREACH_STELLA_LENGTH = ((Symbol*)(intern_rigid_symbol_wrt_module("LENGTH", NULL, 0)));
  }
}

void help_startup_foreach2() {
  {
    SYM_FOREACH_STELLA_NTH = ((Symbol*)(intern_rigid_symbol_wrt_module("NTH", NULL, 0)));
    SGT_FOREACH_STELLA_CHARACTER = ((Surrogate*)(intern_rigid_symbol_wrt_module("CHARACTER", NULL, 1)));
    SYM_FOREACH_STELLA_l = ((Symbol*)(intern_rigid_symbol_wrt_module("<", NULL, 0)));
    SGT_FOREACH_STELLA_ARGUMENT_LIST_ITERATOR = ((Surrogate*)(intern_rigid_symbol_wrt_module("ARGUMENT-LIST-ITERATOR", NULL, 1)));
    SYM_FOREACH_STELLA_SYS_CALL_METHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-CALL-METHOD", NULL, 0)));
    SGT_FOREACH_STELLA_LITERAL = ((Surrogate*)(intern_rigid_symbol_wrt_module("LITERAL", NULL, 1)));
    SYM_FOREACH_STELLA_ON = ((Symbol*)(intern_rigid_symbol_wrt_module("ON", NULL, 0)));
    SYM_FOREACH_STELLA_AND = ((Symbol*)(intern_rigid_symbol_wrt_module("AND", NULL, 0)));
    SYM_FOREACH_STELLA_IF = ((Symbol*)(intern_rigid_symbol_wrt_module("IF", NULL, 0)));
    SYM_FOREACH_STELLA_CONS = ((Symbol*)(intern_rigid_symbol_wrt_module("CONS", NULL, 0)));
    SYM_FOREACH_STELLA_EMPTYp = ((Symbol*)(intern_rigid_symbol_wrt_module("EMPTY?", NULL, 0)));
    SYM_FOREACH_STELLA_SETF = ((Symbol*)(intern_rigid_symbol_wrt_module("SETF", NULL, 0)));
    SYM_FOREACH_STELLA_ADD_CONS_TO_END_OF_CONS_LIST = ((Symbol*)(intern_rigid_symbol_wrt_module("ADD-CONS-TO-END-OF-CONS-LIST", NULL, 0)));
    SYM_FOREACH_STELLA_COND = ((Symbol*)(intern_rigid_symbol_wrt_module("COND", NULL, 0)));
    SYM_FOREACH_STELLA_CONSp = ((Symbol*)(intern_rigid_symbol_wrt_module("CONS?", NULL, 0)));
    SYM_FOREACH_STELLA_OTHERWISE = ((Symbol*)(intern_rigid_symbol_wrt_module("OTHERWISE", NULL, 0)));
    SYM_FOREACH_STELLA_WHEN = ((Symbol*)(intern_rigid_symbol_wrt_module("WHEN", NULL, 0)));
    SYM_FOREACH_STELLA_SYS_FOREACH = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-FOREACH", NULL, 0)));
    SYM_FOREACH_STELLA_VRLET = ((Symbol*)(intern_rigid_symbol_wrt_module("VRLET", NULL, 0)));
    SYM_FOREACH_STELLA_BREAK = ((Symbol*)(intern_rigid_symbol_wrt_module("BREAK", NULL, 0)));
    SYM_FOREACH_STELLA_ALWAYS = ((Symbol*)(intern_rigid_symbol_wrt_module("ALWAYS", NULL, 0)));
    SYM_FOREACH_STELLA_DEFUN = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFUN", NULL, 0)));
    SYM_FOREACH_STELLA_BOOLEAN = ((Symbol*)(intern_rigid_symbol_wrt_module("BOOLEAN", NULL, 0)));
    SYM_FOREACH_STELLA_SELF = ((Symbol*)(intern_rigid_symbol_wrt_module("SELF", NULL, 0)));
    SYM_FOREACH_STELLA_ALL_PURPOSE_ITERATOR = ((Symbol*)(intern_rigid_symbol_wrt_module("ALL-PURPOSE-ITERATOR", NULL, 0)));
    SYM_FOREACH_STELLA_LET = ((Symbol*)(intern_rigid_symbol_wrt_module("LET", NULL, 0)));
    SYM_FOREACH_STELLA_NESTEDITERATOR = ((Symbol*)(intern_rigid_symbol_wrt_module("NESTEDITERATOR", NULL, 0)));
    SYM_FOREACH_STELLA_ITERATOR_NESTED_ITERATOR = ((Symbol*)(intern_rigid_symbol_wrt_module("ITERATOR-NESTED-ITERATOR", NULL, 0)));
    SYM_FOREACH_STELLA_WHILE = ((Symbol*)(intern_rigid_symbol_wrt_module("WHILE", NULL, 0)));
    SYM_FOREACH_STELLA_RETURN = ((Symbol*)(intern_rigid_symbol_wrt_module("RETURN", NULL, 0)));
    SYM_FOREACH_STELLA_NEW = ((Symbol*)(intern_rigid_symbol_wrt_module("NEW", NULL, 0)));
    SYM_FOREACH_STELLA_ITERATOR_NEXT_CODE = ((Symbol*)(intern_rigid_symbol_wrt_module("ITERATOR-NEXT-CODE", NULL, 0)));
    SYM_FOREACH_STELLA_THE_CODE = ((Symbol*)(intern_rigid_symbol_wrt_module("THE-CODE", NULL, 0)));
    KWD_FOREACH_FUNCTION = ((Keyword*)(intern_rigid_symbol_wrt_module("FUNCTION", NULL, 2)));
    SGT_FOREACH_STELLA_ALL_PURPOSE_ITERATOR = ((Surrogate*)(intern_rigid_symbol_wrt_module("ALL-PURPOSE-ITERATOR", NULL, 1)));
    SYM_FOREACH_STELLA_STARTUP_FOREACH = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-FOREACH", NULL, 0)));
    SYM_FOREACH_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-STARTUP-CLASSNAME", NULL, 0)));
  }
}

void startup_foreach() {
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oSTELLA_MODULEo);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (current_startup_time_phaseP(2)) {
      help_startup_foreach1();
      help_startup_foreach2();
    }
    if (current_startup_time_phaseP(6)) {
      finalize_classes();
    }
    if (current_startup_time_phaseP(7)) {
      define_function_object("EXTRACT-ONE-IN-CLAUSE", "(DEFUN (EXTRACT-ONE-IN-CLAUSE CONS CONS) ((TAIL CONS)))", ((cpp_function_code)(&extract_one_in_clause)), NULL);
      define_function_object("EXTRACT-ONE-GENERATOR-TERM", "(DEFUN (EXTRACT-ONE-GENERATOR-TERM CONS CONS) ((TAIL CONS)))", ((cpp_function_code)(&extract_one_generator_term)), NULL);
      define_function_object("EXTRACT-GENERATORS-CLAUSE", "(DEFUN (EXTRACT-GENERATORS-CLAUSE CONS CONS) ((TAIL CONS)) :PUBLIC? TRUE)", ((cpp_function_code)(&extract_generators_clause)), NULL);
      define_function_object("EXTRACT-OPTION-AND-RESIDUE", "(DEFUN (EXTRACT-OPTION-AND-RESIDUE OBJECT CONS) ((OPTION SYMBOL) (TAIL CONS)) :PUBLIC? TRUE)", ((cpp_function_code)(&extract_option_and_residue)), NULL);
      define_function_object("EXTRACT-DO-CLAUSE", "(DEFUN (EXTRACT-DO-CLAUSE CONS CONS) ((TAIL CONS)))", ((cpp_function_code)(&extract_do_clause)), NULL);
      define_function_object("FOREACH-TO-PREFIX-FOREACH-TREE", "(DEFUN (FOREACH-TO-PREFIX-FOREACH-TREE CONS) ((FOREACHTREE CONS)))", ((cpp_function_code)(&foreach_to_prefix_foreach_tree)), NULL);
      define_function_object("WALK-FOREACH-TREE", "(DEFUN (WALK-FOREACH-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_foreach_tree)), NULL);
      define_function_object("WALK-COLLECTION-TREE", "(DEFUN (WALK-COLLECTION-TREE OBJECT TYPE-SPEC) ((TREE OBJECT) (DONTOPTIMIZE? BOOLEAN)))", ((cpp_function_code)(&walk_collection_tree)), NULL);
      define_function_object("YIELD-IN-CURSOR-CLAUSES", "(DEFUN (YIELD-IN-CURSOR-CLAUSES CONS CONS OBJECT) ((INTREE CONS) (DONTOPTIMIZE? BOOLEAN)))", ((cpp_function_code)(&yield_in_cursor_clauses)), NULL);
      define_function_object("USE-VECTOR-STYLE-ITERATION?", "(DEFUN (USE-VECTOR-STYLE-ITERATION? BOOLEAN) ((COLLECTIONTYPE TYPE)))", ((cpp_function_code)(&use_vector_style_iterationP)), NULL);
      define_function_object("YIELD-IN-CURSOR-CLAUSES-FOR-GENERAL-COLLECTION", "(DEFUN (YIELD-IN-CURSOR-CLAUSES-FOR-GENERAL-COLLECTION CONS CONS OBJECT) ((COLLECTIONTREE OBJECT) (COLLECTIONTYPE TYPE-SPEC) (KEYVAR SYMBOL) (VALUEVAR SYMBOL)))", ((cpp_function_code)(&yield_in_cursor_clauses_for_general_collection)), NULL);
      define_function_object("YIELD-IN-CURSOR-CLAUSES-FOR-INTEGER-INTERVAL", "(DEFUN (YIELD-IN-CURSOR-CLAUSES-FOR-INTEGER-INTERVAL CONS CONS OBJECT) ((COLLECTIONTREE CONS) (COLLECTIONTYPE TYPE-SPEC) (KEYVAR SYMBOL) (VALUEVAR SYMBOL)))", ((cpp_function_code)(&yield_in_cursor_clauses_for_integer_interval)), NULL);
      define_function_object("YIELD-IN-CURSOR-CLAUSES-FOR-VECTOR", "(DEFUN (YIELD-IN-CURSOR-CLAUSES-FOR-VECTOR CONS CONS OBJECT) ((COLLECTIONTREE OBJECT) (COLLECTIONTYPE TYPE-SPEC) (KEYVAR SYMBOL) (VALUEVAR SYMBOL)))", ((cpp_function_code)(&yield_in_cursor_clauses_for_vector)), NULL);
      define_function_object("YIELD-IN-CURSOR-CLAUSES-FOR-ARGUMENT-LIST", "(DEFUN (YIELD-IN-CURSOR-CLAUSES-FOR-ARGUMENT-LIST CONS CONS OBJECT) ((COLLECTIONTREE OBJECT) (COLLECTIONTYPE TYPE-SPEC) (KEYVAR SYMBOL) (VALUEVAR SYMBOL)))", ((cpp_function_code)(&yield_in_cursor_clauses_for_argument_list)), NULL);
      define_function_object("YIELD-ON-CURSOR-CLAUSES", "(DEFUN (YIELD-ON-CURSOR-CLAUSES CONS CONS OBJECT) ((ONTREE CONS)))", ((cpp_function_code)(&yield_on_cursor_clauses)), NULL);
      define_function_object("COLLECT-CURSOR-CLAUSES", "(DEFUN (COLLECT-CURSOR-CLAUSES CONS CONS OBJECT) ((CURSORSTREE CONS) (DONTOPTIMIZE? BOOLEAN)))", ((cpp_function_code)(&collect_cursor_clauses)), NULL);
      define_function_object("WALK-CURSORS-TREE", "(DEFUN (WALK-CURSORS-TREE CONS CONS OBJECT) ((CURSORSTREE CONS) (DONTOPTIMIZE? BOOLEAN)))", ((cpp_function_code)(&walk_cursors_tree)), NULL);
      define_function_object("YIELD-COLLECT-INTO-LIST-CODE", "(DEFUN (YIELD-COLLECT-INTO-LIST-CODE CONS) ((COLLECTVARIABLE SYMBOL) (INTOVARIABLE OBJECT) (COLLECTTREE OBJECT)))", ((cpp_function_code)(&yield_collect_into_list_code)), NULL);
      define_function_object("YIELD-COLLECT-INTO-CONS-CODE", "(DEFUN (YIELD-COLLECT-INTO-CONS-CODE CONS) ((COLLECTVARIABLE SYMBOL) (INTOVARIABLE OBJECT) (COLLECTTREE OBJECT)))", ((cpp_function_code)(&yield_collect_into_cons_code)), NULL);
      define_function_object("YIELD-COLLECT-CODE", "(DEFUN (YIELD-COLLECT-CODE CONS) ((COLLECTVARIABLE SYMBOL) (INTOVARIABLE OBJECT) (COLLECTTREE OBJECT)))", ((cpp_function_code)(&yield_collect_code)), NULL);
      define_function_object("WALK-PREFIX-COLLECT-INTO-TREE", "(DEFUN (WALK-PREFIX-COLLECT-INTO-TREE CONS CONS) ((COLLECTINTOTREE CONS) (ITERATORS CONS) (VALUEBINDINGS CONS) (DOCODE CONS)))", ((cpp_function_code)(&walk_prefix_collect_into_tree)), NULL);
      define_function_object("WRAP-WHERE-TEST", "(DEFUN (WRAP-WHERE-TEST CONS) ((WHERETEST OBJECT) (ACTIONS CONS)))", ((cpp_function_code)(&wrap_where_test)), NULL);
      define_function_object("WALK-PREFIX-FOREACH-BODY", "(DEFUN (WALK-PREFIX-FOREACH-BODY CONS) ((ITERATORCLAUSES CONS) (VALUECLAUSES CONS) (WHERETEST OBJECT) (DOACTIONS CONS)))", ((cpp_function_code)(&walk_prefix_foreach_body)), NULL);
      define_function_object("WALK-PREFIX-FOREACH-TREE", "(DEFUN (WALK-PREFIX-FOREACH-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_prefix_foreach_tree)), NULL);
      define_function_object("WALK-EXISTS-TREE", "(DEFUN (WALK-EXISTS-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_exists_tree)), NULL);
      define_function_object("WALK-FORALL-TREE", "(DEFUN (WALK-FORALL-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_forall_tree)), NULL);
      define_function_object("WALK-SOME-TREE", "(DEFUN (WALK-SOME-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_some_tree)), NULL);
      define_function_object("EXTRACT-SETOF-NESTED-ITERATOR-TREE", "(DEFUN (EXTRACT-SETOF-NESTED-ITERATOR-TREE CONS) ((ITERATORCLAUSES CONS)))", ((cpp_function_code)(&extract_setof_nested_iterator_tree)), NULL);
      define_function_object("YIELD-SETOF-NEXT-FUNCTION-NAME", "(DEFUN (YIELD-SETOF-NEXT-FUNCTION-NAME SYMBOL) ((VALUECLAUSES CONS) (WHERETEST CONS)))", ((cpp_function_code)(&yield_setof_next_function_name)), NULL);
      define_function_object("WALK-SETOF-TREE", "(DEFUN (WALK-SETOF-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_setof_tree)), NULL);
      define_function_object("STARTUP-FOREACH", "(DEFUN STARTUP-FOREACH () :PUBLIC? TRUE)", ((cpp_function_code)(&startup_foreach)), NULL);
      { Method_Slot* function = lookup_function(SYM_FOREACH_STELLA_STARTUP_FOREACH);

        set_dynamic_slot_value(function->dynamic_slots, SYM_FOREACH_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string("Startup-Foreach"), NULL_STRING_WRAPPER);
      }
    }
    if (current_startup_time_phaseP(8)) {
      finalize_slots();
      cleanup_unfinalized_classes();
    }
  }
}

Symbol* SYM_FOREACH_STELLA_AS = NULL;

Symbol* SYM_FOREACH_STELLA_COMMA = NULL;

Symbol* SYM_FOREACH_STELLA_DO = NULL;

Symbol* SYM_FOREACH_STELLA_COLLECT = NULL;

Surrogate* SGT_FOREACH_STELLA_VOID = NULL;

Symbol* SYM_FOREACH_STELLA_PREFIX_FOREACH = NULL;

Symbol* SYM_FOREACH_STELLA_IN = NULL;

Symbol* SYM_FOREACH_STELLA_X = NULL;

Symbol* SYM_FOREACH_STELLA_NIL = NULL;

Symbol* SYM_FOREACH_STELLA_WHERE = NULL;

Symbol* SYM_FOREACH_STELLA_INTO = NULL;

Symbol* SYM_FOREACH_STELLA_COLLECT_INTO = NULL;

Symbol* SYM_FOREACH_STELLA_FOREACH = NULL;

Symbol* SYM_FOREACH_STELLA_EXISTS = NULL;

Symbol* SYM_FOREACH_STELLA_FORALL = NULL;

Symbol* SYM_FOREACH_STELLA_SOME = NULL;

Symbol* SYM_FOREACH_STELLA_SETOF = NULL;

Symbol* SYM_FOREACH_STELLA_SET_OF = NULL;

Surrogate* SGT_FOREACH_STELLA_CONS = NULL;

Symbol* SYM_FOREACH_STELLA_INTERVAL = NULL;

Surrogate* SGT_FOREACH_STELLA_INTEGER_INTERVAL = NULL;

Surrogate* SGT_FOREACH_STELLA_UNKNOWN = NULL;

Surrogate* SGT_FOREACH_STELLA_ABSTRACT_ITERATOR = NULL;

Symbol* SYM_FOREACH_STELLA_ALLOCATE_ITERATOR = NULL;

Symbol* SYM_FOREACH_STELLA_FALSE = NULL;

Surrogate* SGT_FOREACH_STELLA_ARGUMENT_LIST = NULL;

Surrogate* SGT_FOREACH_STELLA_STRING = NULL;

Surrogate* SGT_FOREACH_STELLA_MUTABLE_STRING = NULL;

Surrogate* SGT_FOREACH_STELLA_VECTOR = NULL;

Surrogate* SGT_FOREACH_STELLA_SEQUENCE_MIXIN = NULL;

Symbol* SYM_FOREACH_STELLA_THE_ARRAY = NULL;

Surrogate* SGT_FOREACH_STELLA_LIST = NULL;

Symbol* SYM_FOREACH_STELLA_THE_CONS_LIST = NULL;

Surrogate* SGT_FOREACH_STELLA_KEY_VALUE_LIST = NULL;

Surrogate* SGT_FOREACH_STELLA_KV_CONS = NULL;

Symbol* SYM_FOREACH_STELLA_THE_KV_LIST = NULL;

Surrogate* SGT_FOREACH_STELLA_PROPERTY_LIST = NULL;

Symbol* SYM_FOREACH_STELLA_THE_PLIST = NULL;

Symbol* SYM_FOREACH_STELLA_KEY = NULL;

Symbol* SYM_FOREACH_STELLA_VALUE = NULL;

Symbol* SYM_FOREACH_STELLA_SETQ = NULL;

Symbol* SYM_FOREACH_STELLA_REST = NULL;

Symbol* SYM_FOREACH_STELLA_NOT = NULL;

Symbol* SYM_FOREACH_STELLA_NILp = NULL;

Symbol* SYM_FOREACH_STELLA_DEFINEDp = NULL;

Symbol* SYM_FOREACH_STELLA_NEXTp = NULL;

Symbol* SYM_FOREACH_STELLA_ANY_VALUE = NULL;

Symbol* SYM_FOREACH_STELLA_PROGN = NULL;

Symbol* SYM_FOREACH_STELLA_ANY_KEY = NULL;

Symbol* SYM_FOREACH_STELLA_NULL = NULL;

Symbol* SYM_FOREACH_STELLA_NULL_INTEGER = NULL;

Surrogate* SGT_FOREACH_STELLA_NUMBER_WRAPPER = NULL;

Surrogate* SGT_FOREACH_STELLA_INTEGER = NULL;

Surrogate* SGT_FOREACH_STELLA_BOOLEAN = NULL;

Symbol* SYM_FOREACH_STELLA_NULLp = NULL;

Symbol* SYM_FOREACH_STELLA_ii = NULL;

Symbol* SYM_FOREACH_STELLA_TRUE = NULL;

Symbol* SYM_FOREACH_STELLA_le = NULL;

Symbol* SYM_FOREACH_STELLA_OR = NULL;

Symbol* SYM_FOREACH_STELLA_LENGTH = NULL;

Symbol* SYM_FOREACH_STELLA_NTH = NULL;

Surrogate* SGT_FOREACH_STELLA_CHARACTER = NULL;

Symbol* SYM_FOREACH_STELLA_l = NULL;

Surrogate* SGT_FOREACH_STELLA_ARGUMENT_LIST_ITERATOR = NULL;

Symbol* SYM_FOREACH_STELLA_SYS_CALL_METHOD = NULL;

Surrogate* SGT_FOREACH_STELLA_LITERAL = NULL;

Symbol* SYM_FOREACH_STELLA_ON = NULL;

Symbol* SYM_FOREACH_STELLA_AND = NULL;

Symbol* SYM_FOREACH_STELLA_IF = NULL;

Symbol* SYM_FOREACH_STELLA_CONS = NULL;

Symbol* SYM_FOREACH_STELLA_EMPTYp = NULL;

Symbol* SYM_FOREACH_STELLA_SETF = NULL;

Symbol* SYM_FOREACH_STELLA_ADD_CONS_TO_END_OF_CONS_LIST = NULL;

Symbol* SYM_FOREACH_STELLA_COND = NULL;

Symbol* SYM_FOREACH_STELLA_CONSp = NULL;

Symbol* SYM_FOREACH_STELLA_OTHERWISE = NULL;

Symbol* SYM_FOREACH_STELLA_WHEN = NULL;

Symbol* SYM_FOREACH_STELLA_SYS_FOREACH = NULL;

Symbol* SYM_FOREACH_STELLA_VRLET = NULL;

Symbol* SYM_FOREACH_STELLA_BREAK = NULL;

Symbol* SYM_FOREACH_STELLA_ALWAYS = NULL;

Symbol* SYM_FOREACH_STELLA_DEFUN = NULL;

Symbol* SYM_FOREACH_STELLA_BOOLEAN = NULL;

Symbol* SYM_FOREACH_STELLA_SELF = NULL;

Symbol* SYM_FOREACH_STELLA_ALL_PURPOSE_ITERATOR = NULL;

Symbol* SYM_FOREACH_STELLA_LET = NULL;

Symbol* SYM_FOREACH_STELLA_NESTEDITERATOR = NULL;

Symbol* SYM_FOREACH_STELLA_ITERATOR_NESTED_ITERATOR = NULL;

Symbol* SYM_FOREACH_STELLA_WHILE = NULL;

Symbol* SYM_FOREACH_STELLA_RETURN = NULL;

Symbol* SYM_FOREACH_STELLA_NEW = NULL;

Symbol* SYM_FOREACH_STELLA_ITERATOR_NEXT_CODE = NULL;

Symbol* SYM_FOREACH_STELLA_THE_CODE = NULL;

Keyword* KWD_FOREACH_FUNCTION = NULL;

Surrogate* SGT_FOREACH_STELLA_ALL_PURPOSE_ITERATOR = NULL;

Symbol* SYM_FOREACH_STELLA_STARTUP_FOREACH = NULL;

Symbol* SYM_FOREACH_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace stella
