// idl-translate.cc

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

#include "stella/stella-system.hh"

namespace stella {

Object* idl_translate_a_tree(Object* tree) {
  if (safe_primary_type(tree) == SGT_IDL_TRANSLATE_STELLA_CONS) {
    { Object* tree_000 = tree;
      Cons* tree = ((Cons*)(tree_000));

      if (tree == NIL) {
        return (tree->idl_translate_atomic_tree());
      }
      else {
        return (idl_translate_null());
      }
    }
  }
  else {
    if (!((boolean)(tree))) {
      return (idl_translate_null());
    }
    return (tree->idl_translate_atomic_tree());
  }
}

Cons* idl_translate_unit(Translation_Unit* unit) {
  { Symbol* test_value_000 = unit->category;

    if (test_value_000 == SYM_IDL_TRANSLATE_STELLA_GLOBAL_VARIABLE) {
      return (NULL);
    }
    else if (test_value_000 == SYM_IDL_TRANSLATE_STELLA_TYPE) {
      return (idl_translate_deftype_unit(unit));
    }
    else if (test_value_000 == SYM_IDL_TRANSLATE_STELLA_CLASS) {
      return (idl_translate_define_native_class_unit(unit));
    }
    else if ((test_value_000 == SYM_IDL_TRANSLATE_STELLA_METHOD) ||
        (test_value_000 == SYM_IDL_TRANSLATE_STELLA_MACRO)) {
      return (NULL);
    }
    else if (test_value_000 == SYM_IDL_TRANSLATE_STELLA_PRINT_METHOD) {
      return (NULL);
    }
    else if ((test_value_000 == SYM_IDL_TRANSLATE_STELLA_STARTUP_TIME_PROGN) ||
        (test_value_000 == SYM_IDL_TRANSLATE_STELLA_VERBATIM)) {
      return (NULL);
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

Cons* idl_translate_deftype_unit(Translation_Unit* unit) {
  return (listO(3, SYM_IDL_TRANSLATE_STELLA_IDL_TYPEDEF, idl_translate_name(((Symbol*)(((Cons*)(unit->code_register))->rest->value))), cons(idl_translate_type_spec(((Standard_Object*)(((Surrogate*)(unit->the_object))->surrogate_value))), NIL)));
}

Cons* idl_translate_null() {
  return (listO(3, SYM_IDL_TRANSLATE_STELLA_IDL_IDENT, wrap_string("NULL"), NIL));
}

Cons* Object::idl_translate_atomic_tree() {
  { Object* tree = this;

    return (listO(3, SYM_IDL_TRANSLATE_STELLA_IDL_LITERAL, tree, NIL));
  }
}

Cons* Symbol::idl_translate_atomic_tree() {
  { Symbol* tree = this;

    return (listO(3, SYM_IDL_TRANSLATE_STELLA_IDL_IDENT, idl_translate_name(tree), NIL));
  }
}

String_Wrapper* idl_translate_type_spec(Standard_Object* typespec) {
  { Surrogate* test_value_000 = safe_primary_type(typespec);

    if (subtype_of_parametric_type_specifierP(test_value_000)) {
      { Standard_Object* typespec_000 = typespec;
        Parametric_Type_Specifier* typespec = ((Parametric_Type_Specifier*)(typespec_000));

        return (wrap_string(string_concatenate(idl_translate_type_spec(typespec->specifier_base_type)->wrapper_value, string_concatenate("<", string_concatenate(idl_translate_type_spec(((Standard_Object*)(typespec->specifier_parameter_types->first())))->wrapper_value, ">", 0), 0), 0)));
      }
    }
    else if (subtype_of_surrogateP(test_value_000)) {
      { Standard_Object* typespec_001 = typespec;
        Surrogate* typespec = ((Surrogate*)(typespec_001));

        if (((String_Wrapper*)(dynamic_slot_value(((Class*)(typespec->surrogate_value))->dynamic_slots, SYM_IDL_TRANSLATE_STELLA_CLASS_IDL_NATIVE_TYPE, NULL_STRING_WRAPPER)))->wrapper_value != NULL) {
          return (wrap_string(((String_Wrapper*)(dynamic_slot_value(((Class*)(typespec->surrogate_value))->dynamic_slots, SYM_IDL_TRANSLATE_STELLA_CLASS_IDL_NATIVE_TYPE, NULL_STRING_WRAPPER)))->wrapper_value));
        }
        else {
          return (idl_translate_class_name(type_to_symbol(((Class*)(typespec->surrogate_value))->class_type)));
        }
      }
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

String_Wrapper* idl_translate_class_name(Generalized_Symbol* namesymbol) {
  return (wrap_string(idl_substitute_forbidden_characters(idl_change_case(((Class*)(namesymbol->surrogatify()->surrogate_value))->class_type, KWD_IDL_TRANSLATE_CAPITALIZED), KWD_IDL_TRANSLATE_CAPITALIZED)));
}

char* idl_create_character_substitution_table() {
  { char* table = make_string(256, '_');

    { int code = NULL_INTEGER;
      int iter_000 = (int)(unsigned char) '0';
      int upper_bound_000 = (int)(unsigned char) '9';
      boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;

      while (unboundedP_000 ||
          (iter_000 <= upper_bound_000)) {
        code = iter_000;
        iter_000 = iter_000 + 1;
        table[code] = '=';
      }
    }
    { int code = NULL_INTEGER;
      int iter_001 = (int)(unsigned char) 'A';
      int upper_bound_001 = (int)(unsigned char) 'Z';
      boolean unboundedP_001 = upper_bound_001 == NULL_INTEGER;

      while (unboundedP_001 ||
          (iter_001 <= upper_bound_001)) {
        code = iter_001;
        iter_001 = iter_001 + 1;
        table[code] = '=';
      }
    }
    { int code = NULL_INTEGER;
      int iter_002 = (int)(unsigned char) 'a';
      int upper_bound_002 = (int)(unsigned char) 'z';
      boolean unboundedP_002 = upper_bound_002 == NULL_INTEGER;

      while (unboundedP_002 ||
          (iter_002 <= upper_bound_002)) {
        code = iter_002;
        iter_002 = iter_002 + 1;
        table[code] = '=';
      }
    }
    table[((int)(unsigned char) '_')] = '=';
    table[((int)(unsigned char) ' ')] = '_';
    table[((int)(unsigned char) '!')] = 'X';
    table[((int)(unsigned char) '"')] = '_';
    table[((int)(unsigned char) '#')] = 'H';
    table[((int)(unsigned char) '$')] = 'B';
    table[((int)(unsigned char) '%')] = 'R';
    table[((int)(unsigned char) '&')] = 'A';
    table[((int)(unsigned char) '\'')] = 'Q';
    table[((int)(unsigned char) '(')] = '_';
    table[((int)(unsigned char) ')')] = '_';
    table[((int)(unsigned char) '*')] = 'O';
    table[((int)(unsigned char) '+')] = 'I';
    table[((int)(unsigned char) ',')] = '_';
    table[((int)(unsigned char) '-')] = '_';
    table[((int)(unsigned char) '.')] = 'D';
    table[((int)(unsigned char) '/')] = 'S';
    table[((int)(unsigned char) ':')] = 'C';
    table[((int)(unsigned char) ';')] = '_';
    table[((int)(unsigned char) '<')] = 'L';
    table[((int)(unsigned char) '=')] = 'E';
    table[((int)(unsigned char) '>')] = 'G';
    table[((int)(unsigned char) '?')] = 'P';
    table[((int)(unsigned char) '@')] = 'M';
    table[((int)(unsigned char) '[')] = '_';
    table[((int)(unsigned char) '\\')] = '_';
    table[((int)(unsigned char) ']')] = '_';
    table[((int)(unsigned char) '^')] = 'U';
    table[((int)(unsigned char) '`')] = '_';
    table[((int)(unsigned char) '{')] = '_';
    table[((int)(unsigned char) '|')] = 'V';
    table[((int)(unsigned char) '}')] = '_';
    table[((int)(unsigned char) '~')] = 'T';
    return (table);
  }
}

char* oIDL_CHARACTER_SUBSTITUTION_TABLEo = NULL;

char* idl_substitute_forbidden_characters(char* name, Keyword* caseconvention) {
  { char* result = NULL;
    char substitution = NULL_CHARACTER;

    switch (name[0]) {
      case '0': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "zero"));
      break;
      case '1': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "one"));
      break;
      case '2': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "two"));
      break;
      case '3': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "three"));
      break;
      case '4': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "four"));
      break;
      case '5': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "five"));
      break;
      case '6': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "six"));
      break;
      case '7': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "seven"));
      break;
      case '8': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "eight"));
      break;
      case '9': 
        result = string_to_mutable_string(replace_first_character_with_string(name, "nine"));
      break;
      default:
        result = strcpy(new (GC) char[strlen(name)+1], name);
      break;
    }
    { int i = NULL_INTEGER;
      int iter_000 = 0;
      int upper_bound_000 = strlen(result) - 1;

      while (iter_000 <= upper_bound_000) {
        i = iter_000;
        iter_000 = iter_000 + 1;
        substitution = oIDL_CHARACTER_SUBSTITUTION_TABLEo[((int)(unsigned char) (result[i]))];
        if (!(substitution == '=')) {
          if (caseconvention == KWD_IDL_TRANSLATE_UPPERCASE) {
            result[i] = (oCHARACTER_DOWNCASE_TABLEo[((int)(unsigned char) substitution)]);
          }
          else if (caseconvention == KWD_IDL_TRANSLATE_LOWERCASE) {
            result[i] = (oCHARACTER_UPCASE_TABLEo[((int)(unsigned char) substitution)]);
          }
          else if (caseconvention == KWD_IDL_TRANSLATE_CAPITALIZED) {
            result[i] = (oCHARACTER_UPCASE_TABLEo[((int)(unsigned char) substitution)]);
          }
          else {
            { Output_String_Stream* stream_000 = new_output_string_stream();

              *(stream_000->native_stream) << "`" << caseconvention << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_000->the_string_reader());
            }
          }
        }
      }
    }
    return (result);
  }
}

char* idl_change_case(Generalized_Symbol* symbol, Keyword* caseconvention) {
  { Module* module = ((Module*)(symbol->home_context));
    char* name = symbol->symbol_name;

    if (!((boolean)(module))) {
      module = oMODULEo.get();
    }
    if (module->case_sensitiveP) {
      return (string_copy(name));
    }
    if (caseconvention == KWD_IDL_TRANSLATE_UPPERCASE) {
      return (string_upcase(name));
    }
    else if (caseconvention == KWD_IDL_TRANSLATE_LOWERCASE) {
      return (string_downcase(name));
    }
    else if (caseconvention == KWD_IDL_TRANSLATE_CAPITALIZED) {
      return (string_capitalize(name));
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << caseconvention << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

String_Wrapper* idl_yield_renamed_reserved_word(Symbol* word) {
  return (wrap_string(string_concatenate("renamed_", string_capitalize(word->symbol_name), 0)));
}

String_Hash_Table* idl_create_reserved_word_table() {
  { String_Hash_Table* table = new_string_hash_table();

    { Object* word = NULL;
      Cons* iter_000 = listO(49, SYM_IDL_TRANSLATE_STELLA_ASM, SYM_IDL_TRANSLATE_STELLA_AUTO, SYM_IDL_TRANSLATE_STELLA_BREAK, SYM_IDL_TRANSLATE_STELLA_CASE, SYM_IDL_TRANSLATE_STELLA_CATCH, SYM_IDL_TRANSLATE_STELLA_CHAR, SYM_IDL_TRANSLATE_STELLA_CLASS, SYM_IDL_TRANSLATE_STELLA_CONST, SYM_IDL_TRANSLATE_STELLA_CONTINUE, SYM_IDL_TRANSLATE_STELLA_DEFAULT, SYM_IDL_TRANSLATE_STELLA_DELETE, SYM_IDL_TRANSLATE_STELLA_DO, SYM_IDL_TRANSLATE_STELLA_DOUBLE, SYM_IDL_TRANSLATE_STELLA_ELSE, SYM_IDL_TRANSLATE_STELLA_ENUM, SYM_IDL_TRANSLATE_STELLA_EXTERN, SYM_IDL_TRANSLATE_STELLA_FLOAT, SYM_IDL_TRANSLATE_STELLA_FOR, SYM_IDL_TRANSLATE_STELLA_FRIEND, SYM_IDL_TRANSLATE_STELLA_GOTO, SYM_IDL_TRANSLATE_STELLA_IF, SYM_IDL_TRANSLATE_STELLA_INLINE, SYM_IDL_TRANSLATE_STELLA_INT, SYM_IDL_TRANSLATE_STELLA_LONG, SYM_IDL_TRANSLATE_STELLA_NEW, SYM_IDL_TRANSLATE_STELLA_OPERATOR, SYM_IDL_TRANSLATE_STELLA_PRIVATE, SYM_IDL_TRANSLATE_STELLA_PROTECTED, SYM_IDL_TRANSLATE_STELLA_PUBLIC, SYM_IDL_TRANSLATE_STELLA_REGISTER, SYM_IDL_TRANSLATE_STELLA_RETURN, SYM_IDL_TRANSLATE_STELLA_SHORT, SYM_IDL_TRANSLATE_STELLA_SIGNED, SYM_IDL_TRANSLATE_STELLA_SIZEOF, SYM_IDL_TRANSLATE_STELLA_STATIC, SYM_IDL_TRANSLATE_STELLA_STRUCT, SYM_IDL_TRANSLATE_STELLA_SWITCH, SYM_IDL_TRANSLATE_STELLA_TEMPLATE, SYM_IDL_TRANSLATE_STELLA_THROW, SYM_IDL_TRANSLATE_STELLA_TRY, SYM_IDL_TRANSLATE_STELLA_TYPEDEF, SYM_IDL_TRANSLATE_STELLA_UNION, SYM_IDL_TRANSLATE_STELLA_UNSIGNED, SYM_IDL_TRANSLATE_STELLA_VIRTUAL, SYM_IDL_TRANSLATE_STELLA_VOID, SYM_IDL_TRANSLATE_STELLA_VOLATILE, SYM_IDL_TRANSLATE_STELLA_WHILE, SYM_IDL_TRANSLATE_STELLA_TYPENAME, NIL);

      while (!(iter_000 == NIL)) {
        word = iter_000->value;
        iter_000 = iter_000->rest;
        table->insert_at(((Symbol*)(word))->symbol_name, idl_yield_renamed_reserved_word(((Symbol*)(word))));
      }
    }
    table->insert_at("++", wrap_string("stella_Increment"));
    table->insert_at("--", wrap_string("stella_Decrement"));
    return (table);
  }
}

String_Hash_Table* oIDL_RESERVED_WORD_TABLEo = NULL;

String_Wrapper* idl_lookup_reserved_word_table(Symbol* word) {
  { String_Wrapper* entry = ((String_Wrapper*)(oIDL_RESERVED_WORD_TABLEo->lookup(word->symbol_name)));

    if (((boolean)(entry))) {
      return (wrap_string(entry->wrapper_value));
    }
    else {
      return (NULL);
    }
  }
}

String_Wrapper* idl_translate_name(Symbol* namesymbol) {
  { String_Wrapper* reservedname = idl_lookup_reserved_word_table(namesymbol);

    if (!((boolean)(namesymbol))) {
      return (wrap_string("NULL"));
    }
    if (((boolean)(reservedname))) {
      return (reservedname);
    }
    return (wrap_string(idl_substitute_forbidden_characters(idl_change_case(namesymbol, KWD_IDL_TRANSLATE_LOWERCASE), KWD_IDL_TRANSLATE_LOWERCASE)));
  }
}

String_Wrapper* idl_translate_symbol_constant_name(Generalized_Symbol* symbol) {
  { char* prefix = NULL;

    { Surrogate* test_value_000 = safe_primary_type(symbol);

      if (subtype_of_symbolP(test_value_000)) {
        { Generalized_Symbol* symbol_000 = symbol;
          Symbol* symbol = ((Symbol*)(symbol_000));

          prefix = string_concatenate("SYM_", ((Module*)(symbol->home_context))->context_name(), 0);
        }
      }
      else if (subtype_of_surrogateP(test_value_000)) {
        { Generalized_Symbol* symbol_001 = symbol;
          Surrogate* symbol = ((Surrogate*)(symbol_001));

          prefix = string_concatenate("SGT_", ((Module*)(symbol->home_context))->context_name(), 0);
        }
      }
      else if (subtype_of_keywordP(test_value_000)) {
        { Generalized_Symbol* symbol_002 = symbol;
          Keyword* symbol = ((Keyword*)(symbol_002));

          prefix = "KWD";
        }
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
    }
    return (wrap_string(idl_substitute_forbidden_characters(string_concatenate(prefix, string_concatenate("_", symbol->symbol_name, 0), 0), KWD_IDL_TRANSLATE_UPPERCASE)));
  }
}

String_Wrapper* idl_translate_symbol_constant_id(Generalized_Symbol* symbol) {
  { String_Wrapper* translatedsymbol = idl_translate_symbol_constant_name(symbol);

    translatedsymbol->wrapper_value = string_concatenate(translatedsymbol->wrapper_value, "_id", 0);
    return (translatedsymbol);
  }
}

String_Wrapper* idl_translate_symbol_name(Symbol* symbol) {
  return (idl_translate_symbol_constant_name(symbol));
}

String_Wrapper* idl_translate_symbol_id(Symbol* symbol) {
  return (idl_translate_symbol_constant_id(symbol));
}

String_Wrapper* idl_translate_surrogate_name(Surrogate* surrogate) {
  return (idl_translate_symbol_constant_name(surrogate));
}

String_Wrapper* idl_translate_surrogate_id(Surrogate* surrogate) {
  return (idl_translate_symbol_constant_id(surrogate));
}

String_Wrapper* idl_translate_keyword_name(Keyword* keyword) {
  return (idl_translate_symbol_constant_name(keyword));
}

String_Wrapper* idl_translate_keyword_id(Keyword* keyword) {
  return (idl_translate_symbol_constant_id(keyword));
}

String_Wrapper* idl_translate_class_parameter_name(Symbol* namesymbol) {
  return (wrap_string(idl_substitute_forbidden_characters(idl_change_case(namesymbol, KWD_IDL_TRANSLATE_UPPERCASE), KWD_IDL_TRANSLATE_UPPERCASE)));
}

Symbol* idl_create_overloaded_function_name(Symbol* functionname, Surrogate* classtype) {
  return (intern_symbol(string_concatenate(classtype->symbol_name, string_concatenate("-", functionname->symbol_name, 0), 0)));
}

boolean idl_method_object_is_overloaded_functionP(Method_Slot* method) {
  return ((!method->method_functionP) &&
      (((String_Wrapper*)(dynamic_slot_value(((Class*)(method->slot_owner->surrogate_value))->dynamic_slots, SYM_IDL_TRANSLATE_STELLA_CLASS_IDL_NATIVE_TYPE, NULL_STRING_WRAPPER)))->wrapper_value != NULL));
}

String_Wrapper* idl_translate_function_name(Method_Slot* function) {
  { Symbol* functionname = function->slot_name;
    char* translatedname = NULL;

    if (idl_method_object_is_overloaded_functionP(function)) {
      functionname = idl_create_overloaded_function_name(functionname, function->slot_owner);
    }
    translatedname = idl_translate_name(functionname)->wrapper_value;
    if (function->method_setterP) {
      translatedname = string_concatenate(translatedname, "_setter", 0);
    }
    if (slot_readerP(function)) {
      translatedname = string_concatenate(translatedname, "_reader", 0);
    }
    return (wrap_string(translatedname));
  }
}

Object* idl_translate_method_parameter(Symbol* name, Standard_Object* type, Symbol* direction) {
  return (cons(idl_translate_name(direction), cons(idl_translate_type_spec(type), cons(idl_translate_name(name), NIL))));
}

Cons* idl_translate_method_parameters(Method_Slot* method) {
  { Cons* result = NIL;
    Cons* directions = NIL;

    if (method->method_parameter_directions_reader()->emptyP()) {
      { Symbol* name = NULL;
        Cons* iter_000 = method->method_parameter_names_reader()->rest();
        Cons* collect_000 = NULL;

        while (!(iter_000 == NIL)) {
          name = ((Symbol*)(iter_000->value));
          iter_000 = iter_000->rest;
          if (!((boolean)(collect_000))) {
            {
              collect_000 = cons(SYM_IDL_TRANSLATE_STELLA_IN, NIL);
              if (directions == NIL) {
                directions = collect_000;
              }
              else {
                add_cons_to_end_of_cons_list(directions, collect_000);
              }
            }
          }
          else {
            {
              collect_000->rest = cons(SYM_IDL_TRANSLATE_STELLA_IN, NIL);
              collect_000 = collect_000->rest;
            }
          }
        }
      }
    }
    else {
      directions = method->method_parameter_directions_reader()->rest();
    }
    { Symbol* name = NULL;
      Standard_Object* type = NULL;
      Object* direction = NULL;
      Cons* iter_001 = method->method_parameter_names_reader()->rest();
      Cons* iter_002 = method->method_parameter_type_specifiers_reader()->rest();
      Cons* iter_003 = directions;
      Cons* collect_001 = NULL;

      while ((!(iter_001 == NIL)) &&
          ((!(iter_002 == NIL)) &&
           (!(iter_003 == NIL)))) {
        name = ((Symbol*)(iter_001->value));
        iter_001 = iter_001->rest;
        type = ((Standard_Object*)(iter_002->value));
        iter_002 = iter_002->rest;
        direction = iter_003->value;
        iter_003 = iter_003->rest;
        if (!((boolean)(collect_001))) {
          {
            collect_001 = cons(idl_translate_method_parameter(name, type, ((Symbol*)(direction))), NIL);
            if (result == NIL) {
              result = collect_001;
            }
            else {
              add_cons_to_end_of_cons_list(result, collect_001);
            }
          }
        }
        else {
          {
            collect_001->rest = cons(idl_translate_method_parameter(name, type, ((Symbol*)(direction))), NIL);
            collect_001 = collect_001->rest;
          }
        }
      }
    }
    return (result);
  }
}

String_Wrapper* idl_translate_constructor_name(Method_Slot* constructor) {
  return (idl_translate_class_name(constructor->slot_name));
}

void help_startup_idl_translate1() {
  {
    SGT_IDL_TRANSLATE_STELLA_CONS = ((Surrogate*)(intern_rigid_symbol_wrt_module("CONS", NULL, 1)));
    SYM_IDL_TRANSLATE_STELLA_GLOBAL_VARIABLE = ((Symbol*)(intern_rigid_symbol_wrt_module("GLOBAL-VARIABLE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_TYPE = ((Symbol*)(intern_rigid_symbol_wrt_module("TYPE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_CLASS = ((Symbol*)(intern_rigid_symbol_wrt_module("CLASS", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_METHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_MACRO = ((Symbol*)(intern_rigid_symbol_wrt_module("MACRO", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_PRINT_METHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("PRINT-METHOD", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_STARTUP_TIME_PROGN = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-TIME-PROGN", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_VERBATIM = ((Symbol*)(intern_rigid_symbol_wrt_module("VERBATIM", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_IDL_TYPEDEF = ((Symbol*)(intern_rigid_symbol_wrt_module("IDL_TYPEDEF", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_IDL_IDENT = ((Symbol*)(intern_rigid_symbol_wrt_module("IDL_IDENT", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_IDL_LITERAL = ((Symbol*)(intern_rigid_symbol_wrt_module("IDL_LITERAL", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_CLASS_IDL_NATIVE_TYPE = ((Symbol*)(intern_rigid_symbol_wrt_module("CLASS-IDL-NATIVE-TYPE", NULL, 0)));
    KWD_IDL_TRANSLATE_CAPITALIZED = ((Keyword*)(intern_rigid_symbol_wrt_module("CAPITALIZED", NULL, 2)));
    KWD_IDL_TRANSLATE_UPPERCASE = ((Keyword*)(intern_rigid_symbol_wrt_module("UPPERCASE", NULL, 2)));
    KWD_IDL_TRANSLATE_LOWERCASE = ((Keyword*)(intern_rigid_symbol_wrt_module("LOWERCASE", NULL, 2)));
    SYM_IDL_TRANSLATE_STELLA_ASM = ((Symbol*)(intern_rigid_symbol_wrt_module("ASM", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_AUTO = ((Symbol*)(intern_rigid_symbol_wrt_module("AUTO", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_BREAK = ((Symbol*)(intern_rigid_symbol_wrt_module("BREAK", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_CASE = ((Symbol*)(intern_rigid_symbol_wrt_module("CASE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_CATCH = ((Symbol*)(intern_rigid_symbol_wrt_module("CATCH", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_CHAR = ((Symbol*)(intern_rigid_symbol_wrt_module("CHAR", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_CONST = ((Symbol*)(intern_rigid_symbol_wrt_module("CONST", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_CONTINUE = ((Symbol*)(intern_rigid_symbol_wrt_module("CONTINUE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_DEFAULT = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFAULT", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_DELETE = ((Symbol*)(intern_rigid_symbol_wrt_module("DELETE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_DO = ((Symbol*)(intern_rigid_symbol_wrt_module("DO", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_DOUBLE = ((Symbol*)(intern_rigid_symbol_wrt_module("DOUBLE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_ELSE = ((Symbol*)(intern_rigid_symbol_wrt_module("ELSE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_ENUM = ((Symbol*)(intern_rigid_symbol_wrt_module("ENUM", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_EXTERN = ((Symbol*)(intern_rigid_symbol_wrt_module("EXTERN", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_FLOAT = ((Symbol*)(intern_rigid_symbol_wrt_module("FLOAT", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_FOR = ((Symbol*)(intern_rigid_symbol_wrt_module("FOR", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_FRIEND = ((Symbol*)(intern_rigid_symbol_wrt_module("FRIEND", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_GOTO = ((Symbol*)(intern_rigid_symbol_wrt_module("GOTO", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_IF = ((Symbol*)(intern_rigid_symbol_wrt_module("IF", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_INLINE = ((Symbol*)(intern_rigid_symbol_wrt_module("INLINE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_INT = ((Symbol*)(intern_rigid_symbol_wrt_module("INT", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_LONG = ((Symbol*)(intern_rigid_symbol_wrt_module("LONG", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_NEW = ((Symbol*)(intern_rigid_symbol_wrt_module("NEW", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_OPERATOR = ((Symbol*)(intern_rigid_symbol_wrt_module("OPERATOR", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_PRIVATE = ((Symbol*)(intern_rigid_symbol_wrt_module("PRIVATE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_PROTECTED = ((Symbol*)(intern_rigid_symbol_wrt_module("PROTECTED", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_PUBLIC = ((Symbol*)(intern_rigid_symbol_wrt_module("PUBLIC", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_REGISTER = ((Symbol*)(intern_rigid_symbol_wrt_module("REGISTER", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_RETURN = ((Symbol*)(intern_rigid_symbol_wrt_module("RETURN", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_SHORT = ((Symbol*)(intern_rigid_symbol_wrt_module("SHORT", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_SIGNED = ((Symbol*)(intern_rigid_symbol_wrt_module("SIGNED", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_SIZEOF = ((Symbol*)(intern_rigid_symbol_wrt_module("SIZEOF", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_STATIC = ((Symbol*)(intern_rigid_symbol_wrt_module("STATIC", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_STRUCT = ((Symbol*)(intern_rigid_symbol_wrt_module("STRUCT", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_SWITCH = ((Symbol*)(intern_rigid_symbol_wrt_module("SWITCH", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_TEMPLATE = ((Symbol*)(intern_rigid_symbol_wrt_module("TEMPLATE", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_THROW = ((Symbol*)(intern_rigid_symbol_wrt_module("THROW", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_TRY = ((Symbol*)(intern_rigid_symbol_wrt_module("TRY", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_TYPEDEF = ((Symbol*)(intern_rigid_symbol_wrt_module("TYPEDEF", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_UNION = ((Symbol*)(intern_rigid_symbol_wrt_module("UNION", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_UNSIGNED = ((Symbol*)(intern_rigid_symbol_wrt_module("UNSIGNED", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_VIRTUAL = ((Symbol*)(intern_rigid_symbol_wrt_module("VIRTUAL", NULL, 0)));
    SYM_IDL_TRANSLATE_STELLA_VOID = ((Symbol*)(intern_rigid_symbol_wrt_module("VOID", NULL, 0)));
  }
}

void startup_idl_translate() {
  if (current_startup_time_phaseP(1)) {
    define_module_from_stringified_source("/IDL-THEORY", "(:CASE-SENSITIVE? TRUE :USES () :LISP-PACKAGE \"IDL\")");
    define_module_from_stringified_source("/IDL-THEORY/IDL-USER-THEORY", "(:CASE-SENSITIVE? TRUE :USES () :LISP-PACKAGE \"IDL-USER\")");
  }
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oSTELLA_MODULEo);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (current_startup_time_phaseP(2)) {
      help_startup_idl_translate1();
      SYM_IDL_TRANSLATE_STELLA_VOLATILE = ((Symbol*)(intern_rigid_symbol_wrt_module("VOLATILE", NULL, 0)));
      SYM_IDL_TRANSLATE_STELLA_WHILE = ((Symbol*)(intern_rigid_symbol_wrt_module("WHILE", NULL, 0)));
      SYM_IDL_TRANSLATE_STELLA_TYPENAME = ((Symbol*)(intern_rigid_symbol_wrt_module("TYPENAME", NULL, 0)));
      SYM_IDL_TRANSLATE_STELLA_IN = ((Symbol*)(intern_rigid_symbol_wrt_module("IN", NULL, 0)));
      SYM_IDL_TRANSLATE_STELLA_STARTUP_IDL_TRANSLATE = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-IDL-TRANSLATE", NULL, 0)));
      SYM_IDL_TRANSLATE_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-STARTUP-CLASSNAME", NULL, 0)));
    }
    if (current_startup_time_phaseP(4)) {
      oIDL_CHARACTER_SUBSTITUTION_TABLEo = idl_create_character_substitution_table();
      oIDL_RESERVED_WORD_TABLEo = idl_create_reserved_word_table();
    }
    if (current_startup_time_phaseP(6)) {
      finalize_classes();
    }
    if (current_startup_time_phaseP(7)) {
      define_function_object("IDL-TRANSLATE-A-TREE", "(DEFUN (IDL-TRANSLATE-A-TREE OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&idl_translate_a_tree)), NULL);
      define_function_object("IDL-TRANSLATE-UNIT", "(DEFUN (IDL-TRANSLATE-UNIT CONS) ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&idl_translate_unit)), NULL);
      define_function_object("IDL-TRANSLATE-DEFTYPE-UNIT", "(DEFUN (IDL-TRANSLATE-DEFTYPE-UNIT CONS) ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&idl_translate_deftype_unit)), NULL);
      define_function_object("IDL-TRANSLATE-NULL", "(DEFUN (IDL-TRANSLATE-NULL CONS) ())", ((cpp_function_code)(&idl_translate_null)), NULL);
      define_method_object("(DEFMETHOD (IDL-TRANSLATE-ATOMIC-TREE CONS) ((TREE OBJECT)))", ((cpp_method_code)(&Object::idl_translate_atomic_tree)), ((cpp_method_code)(NULL)));
      define_method_object("(DEFMETHOD (IDL-TRANSLATE-ATOMIC-TREE CONS) ((TREE SYMBOL)))", ((cpp_method_code)(&Symbol::idl_translate_atomic_tree)), ((cpp_method_code)(NULL)));
      define_function_object("IDL-TRANSLATE-TYPE-SPEC", "(DEFUN (IDL-TRANSLATE-TYPE-SPEC STRING-WRAPPER) ((TYPESPEC TYPE-SPEC)))", ((cpp_function_code)(&idl_translate_type_spec)), NULL);
      define_function_object("IDL-TRANSLATE-CLASS-NAME", "(DEFUN (IDL-TRANSLATE-CLASS-NAME STRING-WRAPPER) ((NAMESYMBOL GENERALIZED-SYMBOL)))", ((cpp_function_code)(&idl_translate_class_name)), NULL);
      define_function_object("IDL-CREATE-CHARACTER-SUBSTITUTION-TABLE", "(DEFUN (IDL-CREATE-CHARACTER-SUBSTITUTION-TABLE STRING) ())", ((cpp_function_code)(&idl_create_character_substitution_table)), NULL);
      define_function_object("IDL-SUBSTITUTE-FORBIDDEN-CHARACTERS", "(DEFUN (IDL-SUBSTITUTE-FORBIDDEN-CHARACTERS STRING) ((NAME STRING) (CASECONVENTION KEYWORD)))", ((cpp_function_code)(&idl_substitute_forbidden_characters)), NULL);
      define_function_object("IDL-CHANGE-CASE", "(DEFUN (IDL-CHANGE-CASE STRING) ((SYMBOL GENERALIZED-SYMBOL) (CASECONVENTION KEYWORD)))", ((cpp_function_code)(&idl_change_case)), NULL);
      define_function_object("IDL-YIELD-RENAMED-RESERVED-WORD", "(DEFUN (IDL-YIELD-RENAMED-RESERVED-WORD STRING-WRAPPER) ((WORD SYMBOL)))", ((cpp_function_code)(&idl_yield_renamed_reserved_word)), NULL);
      define_function_object("IDL-CREATE-RESERVED-WORD-TABLE", "(DEFUN (IDL-CREATE-RESERVED-WORD-TABLE STRING-HASH-TABLE) ())", ((cpp_function_code)(&idl_create_reserved_word_table)), NULL);
      define_function_object("IDL-LOOKUP-RESERVED-WORD-TABLE", "(DEFUN (IDL-LOOKUP-RESERVED-WORD-TABLE STRING-WRAPPER) ((WORD SYMBOL)))", ((cpp_function_code)(&idl_lookup_reserved_word_table)), NULL);
      define_function_object("IDL-TRANSLATE-NAME", "(DEFUN (IDL-TRANSLATE-NAME STRING-WRAPPER) ((NAMESYMBOL SYMBOL)))", ((cpp_function_code)(&idl_translate_name)), NULL);
      define_function_object("IDL-TRANSLATE-SYMBOL-CONSTANT-NAME", "(DEFUN (IDL-TRANSLATE-SYMBOL-CONSTANT-NAME STRING-WRAPPER) ((SYMBOL GENERALIZED-SYMBOL)))", ((cpp_function_code)(&idl_translate_symbol_constant_name)), NULL);
      define_function_object("IDL-TRANSLATE-SYMBOL-CONSTANT-ID", "(DEFUN (IDL-TRANSLATE-SYMBOL-CONSTANT-ID STRING-WRAPPER) ((SYMBOL GENERALIZED-SYMBOL)))", ((cpp_function_code)(&idl_translate_symbol_constant_id)), NULL);
      define_function_object("IDL-TRANSLATE-SYMBOL-NAME", "(DEFUN (IDL-TRANSLATE-SYMBOL-NAME STRING-WRAPPER) ((SYMBOL SYMBOL)))", ((cpp_function_code)(&idl_translate_symbol_name)), NULL);
      define_function_object("IDL-TRANSLATE-SYMBOL-ID", "(DEFUN (IDL-TRANSLATE-SYMBOL-ID STRING-WRAPPER) ((SYMBOL SYMBOL)))", ((cpp_function_code)(&idl_translate_symbol_id)), NULL);
      define_function_object("IDL-TRANSLATE-SURROGATE-NAME", "(DEFUN (IDL-TRANSLATE-SURROGATE-NAME STRING-WRAPPER) ((SURROGATE SURROGATE)))", ((cpp_function_code)(&idl_translate_surrogate_name)), NULL);
      define_function_object("IDL-TRANSLATE-SURROGATE-ID", "(DEFUN (IDL-TRANSLATE-SURROGATE-ID STRING-WRAPPER) ((SURROGATE SURROGATE)))", ((cpp_function_code)(&idl_translate_surrogate_id)), NULL);
      define_function_object("IDL-TRANSLATE-KEYWORD-NAME", "(DEFUN (IDL-TRANSLATE-KEYWORD-NAME STRING-WRAPPER) ((KEYWORD KEYWORD)))", ((cpp_function_code)(&idl_translate_keyword_name)), NULL);
      define_function_object("IDL-TRANSLATE-KEYWORD-ID", "(DEFUN (IDL-TRANSLATE-KEYWORD-ID STRING-WRAPPER) ((KEYWORD KEYWORD)))", ((cpp_function_code)(&idl_translate_keyword_id)), NULL);
      define_function_object("IDL-TRANSLATE-CLASS-PARAMETER-NAME", "(DEFUN (IDL-TRANSLATE-CLASS-PARAMETER-NAME STRING-WRAPPER) ((NAMESYMBOL SYMBOL)))", ((cpp_function_code)(&idl_translate_class_parameter_name)), NULL);
      define_function_object("IDL-CREATE-OVERLOADED-FUNCTION-NAME", "(DEFUN (IDL-CREATE-OVERLOADED-FUNCTION-NAME SYMBOL) ((FUNCTIONNAME SYMBOL) (CLASSTYPE TYPE)))", ((cpp_function_code)(&idl_create_overloaded_function_name)), NULL);
      define_function_object("IDL-METHOD-OBJECT-IS-OVERLOADED-FUNCTION?", "(DEFUN (IDL-METHOD-OBJECT-IS-OVERLOADED-FUNCTION? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&idl_method_object_is_overloaded_functionP)), NULL);
      define_function_object("IDL-TRANSLATE-FUNCTION-NAME", "(DEFUN (IDL-TRANSLATE-FUNCTION-NAME STRING-WRAPPER) ((FUNCTION METHOD-SLOT)))", ((cpp_function_code)(&idl_translate_function_name)), NULL);
      define_function_object("IDL-TRANSLATE-METHOD-PARAMETER", "(DEFUN (IDL-TRANSLATE-METHOD-PARAMETER OBJECT) ((NAME SYMBOL) (TYPE TYPE-SPEC) (DIRECTION SYMBOL)))", ((cpp_function_code)(&idl_translate_method_parameter)), NULL);
      define_function_object("IDL-TRANSLATE-METHOD-PARAMETERS", "(DEFUN (IDL-TRANSLATE-METHOD-PARAMETERS CONS) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&idl_translate_method_parameters)), NULL);
      define_function_object("IDL-TRANSLATE-CONSTRUCTOR-NAME", "(DEFUN (IDL-TRANSLATE-CONSTRUCTOR-NAME STRING-WRAPPER) ((CONSTRUCTOR METHOD-SLOT)))", ((cpp_function_code)(&idl_translate_constructor_name)), NULL);
      define_function_object("STARTUP-IDL-TRANSLATE", "(DEFUN STARTUP-IDL-TRANSLATE () :PUBLIC? TRUE)", ((cpp_function_code)(&startup_idl_translate)), NULL);
      { Method_Slot* function = lookup_function(SYM_IDL_TRANSLATE_STELLA_STARTUP_IDL_TRANSLATE);

        set_dynamic_slot_value(function->dynamic_slots, SYM_IDL_TRANSLATE_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string("Startup-Idl-Translate"), NULL_STRING_WRAPPER);
      }
    }
    if (current_startup_time_phaseP(8)) {
      finalize_slots();
      cleanup_unfinalized_classes();
    }
    if (current_startup_time_phaseP(9)) {
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *IDL-CHARACTER-SUBSTITUTION-TABLE* STRING (IDL-CREATE-CHARACTER-SUBSTITUTION-TABLE))");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *IDL-RESERVED-WORD-TABLE* (STRING-HASH-TABLE OF STRING STRING-WRAPPER) (IDL-CREATE-RESERVED-WORD-TABLE))");
    }
  }
}

Surrogate* SGT_IDL_TRANSLATE_STELLA_CONS = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_GLOBAL_VARIABLE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_TYPE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_CLASS = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_METHOD = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_MACRO = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_PRINT_METHOD = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_STARTUP_TIME_PROGN = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_VERBATIM = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_IDL_TYPEDEF = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_IDL_IDENT = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_IDL_LITERAL = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_CLASS_IDL_NATIVE_TYPE = NULL;

Keyword* KWD_IDL_TRANSLATE_CAPITALIZED = NULL;

Keyword* KWD_IDL_TRANSLATE_UPPERCASE = NULL;

Keyword* KWD_IDL_TRANSLATE_LOWERCASE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_ASM = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_AUTO = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_BREAK = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_CASE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_CATCH = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_CHAR = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_CONST = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_CONTINUE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_DEFAULT = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_DELETE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_DO = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_DOUBLE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_ELSE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_ENUM = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_EXTERN = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_FLOAT = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_FOR = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_FRIEND = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_GOTO = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_IF = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_INLINE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_INT = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_LONG = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_NEW = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_OPERATOR = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_PRIVATE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_PROTECTED = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_PUBLIC = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_REGISTER = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_RETURN = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_SHORT = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_SIGNED = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_SIZEOF = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_STATIC = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_STRUCT = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_SWITCH = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_TEMPLATE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_THROW = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_TRY = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_TYPEDEF = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_UNION = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_UNSIGNED = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_VIRTUAL = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_VOID = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_VOLATILE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_WHILE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_TYPENAME = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_IN = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_STARTUP_IDL_TRANSLATE = NULL;

Symbol* SYM_IDL_TRANSLATE_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace stella
