// memoize.cc

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

#include "stella/stella-system.hh"

namespace stella {

Memoization_Table* new_memoization_table() {
  { Memoization_Table* self = NULL;

    self = new Memoization_Table();
    self->timestamps = NIL;
    self->current_timestamp = NULL;
    self->table_name = NULL;
    self->hash_table = NULL;
    return (self);
  }
}

Surrogate* Memoization_Table::primary_type() {
  { Memoization_Table* self = this;

    return (SGT_MEMOIZE_STELLA_MEMOIZATION_TABLE);
  }
}

Object* access_memoization_table_slot_value(Memoization_Table* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_MEMOIZE_STELLA_HASH_TABLE) {
    if (setvalueP) {
      self->hash_table = ((Integer_Hash_Table*)(value));
    }
    else {
      value = self->hash_table;
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_TABLE_NAME) {
    if (setvalueP) {
      self->table_name = ((Surrogate*)(value));
    }
    else {
      value = self->table_name;
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_CURRENT_TIMESTAMP) {
    if (setvalueP) {
      self->current_timestamp = ((Cons*)(value));
    }
    else {
      value = self->current_timestamp;
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_TIMESTAMPS) {
    if (setvalueP) {
      self->timestamps = ((Cons*)(value));
    }
    else {
      value = self->timestamps;
    }
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << slotname << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
  return (value);
}

// Holds all currently active memoization tables for timestamp
// maintenance and clearance purposes.
List* oALL_MEMOIZATION_TABLESo = NULL;

boolean oMEMOIZATION_ENABLEDpo = TRUE;

void enable_memoization() {
  // Enable memoization and use of memoized expression results.
  oMEMOIZATION_ENABLEDpo = TRUE;
}

void disable_memoization() {
  // Enable memoization and use of memoized expression results.
  oMEMOIZATION_ENABLEDpo = FALSE;
}

// Used by memoization to indicate that a NULL value
// was cached.  Needed to distinguish between an undefined value and a
// cached NULL.
Symbol* MEMOIZED_NULL_VALUE = NULL;

int oTOTAL_MEMOIZATION_LOOKUPSo = 0;

int oFAILED_MEMOIZATION_LOOKUPSo = 0;

int hash_memoized_arguments(Object* arg1, Object* arg2, Object* arg3, Object* arg4, int eqvector) {
  { int hashcode = 0;

    if (eqvector < 0) {
      if (!((boolean)(arg1))) {
        arg1 = MEMOIZED_NULL_VALUE;
      }
      hashcode = (size_t)arg1;
      for (;;) {
        if (arg2 == MEMOIZED_NULL_VALUE) {
          break;
        }
        if (!((boolean)(arg2))) {
          arg2 = MEMOIZED_NULL_VALUE;
        }
        hashcode = (((hashcode << 1)) ^ ((size_t)arg2));
        if (arg3 == MEMOIZED_NULL_VALUE) {
          break;
        }
        if (!((boolean)(arg3))) {
          arg3 = MEMOIZED_NULL_VALUE;
        }
        hashcode = (((hashcode << 1)) ^ ((size_t)arg3));
        if (arg4 == MEMOIZED_NULL_VALUE) {
          break;
        }
        if (!((boolean)(arg4))) {
          arg4 = MEMOIZED_NULL_VALUE;
        }
        hashcode = (((hashcode << 1)) ^ ((size_t)arg4));
        break;
      }
    }
    else {
      if (!((boolean)(arg1))) {
        arg1 = MEMOIZED_NULL_VALUE;
      }
      hashcode = (((eqvector & 1) == 0) ? arg1->hash_code() : (size_t)arg1);
      for (;;) {
        if (arg2 == MEMOIZED_NULL_VALUE) {
          break;
        }
        if (!((boolean)(arg2))) {
          arg2 = MEMOIZED_NULL_VALUE;
        }
        hashcode = (((hashcode << 1)) ^ ((((eqvector & 2) == 0) ? arg2->hash_code() : (size_t)arg2)));
        if (arg3 == MEMOIZED_NULL_VALUE) {
          break;
        }
        if (!((boolean)(arg3))) {
          arg3 = MEMOIZED_NULL_VALUE;
        }
        hashcode = (((hashcode << 1)) ^ ((((eqvector & 4) == 0) ? arg3->hash_code() : (size_t)arg3)));
        if (arg4 == MEMOIZED_NULL_VALUE) {
          break;
        }
        if (!((boolean)(arg4))) {
          arg4 = MEMOIZED_NULL_VALUE;
        }
        hashcode = (((hashcode << 1)) ^ ((((eqvector & 8) == 0) ? arg4->hash_code() : (size_t)arg4)));
        break;
      }
    }
    if (hashcode < 0) {
      return (2 - hashcode);
    }
    else {
      return (hashcode);
    }
  }
}

int hash_memoized_argumentsn(Cons* tuple, int eqvector) {
  { int hashcode = 0;

    if (eqvector < 0) {
      { Object* arg = NULL;
        Cons* iter_000 = tuple;

        while (!(iter_000 == NIL)) {
          arg = iter_000->value;
          iter_000 = iter_000->rest;
          if (!((boolean)(arg))) {
            arg = MEMOIZED_NULL_VALUE;
          }
          hashcode = (((hashcode << 1)) ^ ((size_t)arg));
        }
      }
    }
    else {
      { Object* arg = NULL;
        Cons* iter_001 = tuple;

        while (!(iter_001 == NIL)) {
          arg = iter_001->value;
          iter_001 = iter_001->rest;
          if (!((boolean)(arg))) {
            arg = MEMOIZED_NULL_VALUE;
          }
          hashcode = (((hashcode << 1)) ^ ((((eqvector & 1) == 0) ? arg->hash_code() : (size_t)arg)));
          eqvector = (eqvector >> 1);
        }
      }
    }
    if (hashcode < 0) {
      return (2 - hashcode);
    }
    else {
      return (hashcode);
    }
  }
}

Cons* make_memoized_value_entry(Object* value, Object* arg1, Object* arg2, Object* arg3, Object* arg4) {
  { Cons* entry = NIL;
    Cons* cursor = NIL;
    Cons* cell = NIL;

    entry = new Cons();
    entry->value = value;
    cell = new Cons();
    entry->rest = cell;
    cell->value = arg1;
    if (arg2 == MEMOIZED_NULL_VALUE) {
      cell->rest = NIL;
      return (entry);
    }
    cursor = cell;
    cell = new Cons();
    cursor->rest = cell;
    cell->value = arg2;
    if (arg3 == MEMOIZED_NULL_VALUE) {
      cell->rest = NIL;
      return (entry);
    }
    cursor = cell;
    cell = new Cons();
    cursor->rest = cell;
    cell->value = arg3;
    if (arg4 == MEMOIZED_NULL_VALUE) {
      cell->rest = NIL;
      return (entry);
    }
    cursor = cell;
    cell = new Cons();
    cursor->rest = cell;
    cell->value = arg4;
    cell->rest = NIL;
    return (entry);
  }
}

Cons* make_memoized_value_entryn(Object* value, Cons* args) {
  return (cons(value, args));
}

Cons* find_memoized_value_entry(Cons* bucket, Object* arg1, Object* arg2, Object* arg3, Object* arg4, int eqvector, boolean deleteentryP) {
  { Cons* entries = bucket->rest;
    Cons* args = NIL;
    Cons* result = NULL;

    if (eqvector < 0) {
      while (!(entries == NIL)) {
        args = ((Cons*)(entries->value))->rest;
        if (args->value == arg1) {
          if (arg2 == MEMOIZED_NULL_VALUE) {
            result = ((Cons*)(entries->value));
            break;
          }
          args = args->rest;
          if (args->value == arg2) {
            if (arg3 == MEMOIZED_NULL_VALUE) {
              result = ((Cons*)(entries->value));
              break;
            }
            args = args->rest;
            if (args->value == arg3) {
              if (arg4 == MEMOIZED_NULL_VALUE) {
                result = ((Cons*)(entries->value));
                break;
              }
              args = args->rest;
              if (args->value == arg4) {
                result = ((Cons*)(entries->value));
                break;
              }
            }
          }
        }
        entries = entries->rest;
      }
    }
    else {
      { Object* arg = NULL;

        while (!(entries == NIL)) {
          args = ((Cons*)(entries->value))->rest;
          arg = args->value;
          if ((arg == arg1) ||
              (((eqvector & 1) == 0) &&
               (((boolean)(arg)) &&
                (((boolean)(arg1)) &&
                 arg->object_eqlP(arg1))))) {
            if (arg2 == MEMOIZED_NULL_VALUE) {
              result = ((Cons*)(entries->value));
              break;
            }
            args = args->rest;
            arg = args->value;
            if ((arg == arg2) ||
                (((eqvector & 2) == 0) &&
                 (((boolean)(arg)) &&
                  (((boolean)(arg2)) &&
                   arg->object_eqlP(arg2))))) {
              if (arg3 == MEMOIZED_NULL_VALUE) {
                result = ((Cons*)(entries->value));
                break;
              }
              args = args->rest;
              arg = args->value;
              if ((arg == arg3) ||
                  (((eqvector & 4) == 0) &&
                   (((boolean)(arg)) &&
                    (((boolean)(arg3)) &&
                     arg->object_eqlP(arg3))))) {
                if (arg4 == MEMOIZED_NULL_VALUE) {
                  result = ((Cons*)(entries->value));
                  break;
                }
                args = args->rest;
                arg = args->value;
                if ((arg == arg4) ||
                    (((eqvector & 8) == 0) &&
                     (((boolean)(arg)) &&
                      (((boolean)(arg4)) &&
                       arg->object_eqlP(arg4))))) {
                  result = ((Cons*)(entries->value));
                  break;
                }
              }
            }
          }
          entries = entries->rest;
        }
      }
    }
    if (deleteentryP &&
        ((boolean)(result))) {
      entries->value = ((Cons*)(bucket->rest->value));
      if ((bucket->rest = bucket->rest->rest) == NIL) {
        bucket->rest = NIL;
      }
    }
    return (result);
  }
}

Cons* find_memoized_value_entryn(Cons* bucket, Cons* tuple, int eqvector, boolean deleteentryP) {
  { Cons* entries = bucket->rest;
    Cons* result = NULL;

    if (eqvector < 0) {
      while (!(entries == NIL)) {
        { boolean alwaysP_000 = TRUE;

          { Object* entryarg = NULL;
            Object* inputarg = NULL;
            Cons* iter_000 = ((Cons*)(entries->value))->rest;
            Cons* iter_001 = tuple;

            while ((!(iter_000 == NIL)) &&
                (!(iter_001 == NIL))) {
              entryarg = iter_000->value;
              iter_000 = iter_000->rest;
              inputarg = iter_001->value;
              iter_001 = iter_001->rest;
              if (!(entryarg == inputarg)) {
                alwaysP_000 = FALSE;
                break;
              }
            }
          }
          if (alwaysP_000) {
            result = ((Cons*)(entries->value));
            break;
          }
        }
        entries = entries->rest;
      }
    }
    else {
      while (!(entries == NIL)) {
        result = ((Cons*)(entries->value));
        { Object* entryarg = NULL;
          Object* inputarg = NULL;
          Cons* iter_002 = result->rest;
          Cons* iter_003 = tuple;

          while ((!(iter_002 == NIL)) &&
              (!(iter_003 == NIL))) {
            entryarg = iter_002->value;
            iter_002 = iter_002->rest;
            inputarg = iter_003->value;
            iter_003 = iter_003->rest;
            if ((entryarg == inputarg) ||
                (((eqvector & 1) == 0) &&
                 eqlP(entryarg, inputarg))) {
              eqvector = (eqvector >> 1);
            }
            else {
              result = NULL;
              break;
            }
          }
        }
        if (((boolean)(result))) {
          break;
        }
        else {
          entries = entries->rest;
        }
      }
    }
    if (deleteentryP &&
        ((boolean)(result))) {
      entries->value = ((Cons*)(bucket->rest->value));
      if ((bucket->rest = bucket->rest->rest) == NIL) {
        bucket->rest = NIL;
      }
    }
    return (result);
  }
}

Cons* lookup_memoized_value(Memoization_Table* memotable, Object* arg1, Object* arg2, Object* arg3, Object* arg4, int eqvector) {
  { Integer_Hash_Table* hashtable = memotable->hash_table;
    Cons* timestamp = NULL;
    int hashcode = 0;
    Cons* entry = NIL;
    Cons* bucket = NIL;

    if (!((boolean)(hashtable))) {
      hashtable = new_integer_hash_table();
      memotable->hash_table = hashtable;
      memotable->current_timestamp = cons(NULL, NIL);
    }
    timestamp = memotable->current_timestamp;
    hashcode = hash_memoized_arguments(arg1, arg2, arg3, arg4, eqvector);
    bucket = ((Cons*)(hashtable->lookup(hashcode)));
    if (((boolean)(bucket))) {
      if (((Cons*)(bucket->value)) == timestamp) {
        entry = find_memoized_value_entry(bucket, arg1, arg2, arg3, arg4, eqvector, FALSE);
        if (((boolean)(entry))) {
          return (entry);
        }
      }
      else {
        bucket->value = timestamp;
        bucket->rest = NIL;
      }
    }
    entry = make_memoized_value_entry(NULL, arg1, arg2, arg3, arg4);
    if (((boolean)(bucket))) {
      bucket->rest = cons(entry, bucket->rest);
    }
    else {
      hashtable->insert_at(hashcode, cons(timestamp, cons(entry, NIL)));
    }
    return (entry);
  }
}

Cons* lookup_memoized_valuen(Memoization_Table* memotable, Cons* args, int eqvector) {
  { Integer_Hash_Table* hashtable = memotable->hash_table;
    Cons* timestamp = NULL;
    int hashcode = 0;
    Cons* entry = NIL;
    Cons* bucket = NIL;

    if (!((boolean)(hashtable))) {
      hashtable = new_integer_hash_table();
      memotable->hash_table = hashtable;
      memotable->current_timestamp = cons(NULL, NIL);
    }
    timestamp = memotable->current_timestamp;
    hashcode = hash_memoized_argumentsn(args, eqvector);
    bucket = ((Cons*)(hashtable->lookup(hashcode)));
    if (((boolean)(bucket))) {
      if (((Cons*)(bucket->value)) == timestamp) {
        entry = find_memoized_value_entryn(bucket, args, eqvector, FALSE);
        if (((boolean)(entry))) {
          return (entry);
        }
      }
      else {
        bucket->value = timestamp;
        bucket->rest = NIL;
      }
    }
    entry = cons(NULL, args);
    if (((boolean)(bucket))) {
      bucket->rest = cons(entry, bucket->rest);
    }
    else {
      hashtable->insert_at(hashcode, cons(timestamp, cons(entry, NIL)));
    }
    return (entry);
  }
}

Mru_Memoization_Table* new_mru_memoization_table() {
  { Mru_Memoization_Table* self = NULL;

    self = new Mru_Memoization_Table();
    self->timestamps = NIL;
    self->current_timestamp = NULL;
    self->table_name = NULL;
    self->hash_table = NULL;
    self->max_entries = NULL_INTEGER;
    self->free_entries = NULL_INTEGER;
    self->nof_buckets = NULL_INTEGER;
    self->lru_timestamp = NULL;
    self->mru_timestamp = NULL;
    self->lru_buckets = NULL;
    self->mru_buckets = NULL;
    self->lru_buckets_vector = NULL;
    self->mru_buckets_vector = NULL;
    return (self);
  }
}

Surrogate* Mru_Memoization_Table::primary_type() {
  { Mru_Memoization_Table* self = this;

    return (SGT_MEMOIZE_STELLA_MRU_MEMOIZATION_TABLE);
  }
}

Object* access_mru_memoization_table_slot_value(Mru_Memoization_Table* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_MEMOIZE_STELLA_MRU_BUCKETS_VECTOR) {
    if (setvalueP) {
      self->mru_buckets_vector = ((Vector*)(value));
    }
    else {
      value = self->mru_buckets_vector;
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_LRU_BUCKETS_VECTOR) {
    if (setvalueP) {
      self->lru_buckets_vector = ((Vector*)(value));
    }
    else {
      value = self->lru_buckets_vector;
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_MRU_TIMESTAMP) {
    if (setvalueP) {
      self->mru_timestamp = ((Cons*)(value));
    }
    else {
      value = self->mru_timestamp;
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_LRU_TIMESTAMP) {
    if (setvalueP) {
      self->lru_timestamp = ((Cons*)(value));
    }
    else {
      value = self->lru_timestamp;
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_NOF_BUCKETS) {
    if (setvalueP) {
      self->nof_buckets = ((Integer_Wrapper*)(value))->wrapper_value;
    }
    else {
      value = wrap_integer(self->nof_buckets);
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_FREE_ENTRIES) {
    if (setvalueP) {
      self->free_entries = ((Integer_Wrapper*)(value))->wrapper_value;
    }
    else {
      value = wrap_integer(self->free_entries);
    }
  }
  else if (slotname == SYM_MEMOIZE_STELLA_MAX_ENTRIES) {
    if (setvalueP) {
      self->max_entries = ((Integer_Wrapper*)(value))->wrapper_value;
    }
    else {
      value = wrap_integer(self->max_entries);
    }
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << slotname << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
  return (value);
}

void initialize_mru_bucket_tables(Mru_Memoization_Table* memotable) {
  memotable->nof_buckets = pick_hash_table_size_prime(memotable->max_entries);
  memotable->mru_buckets_vector = new_vector(memotable->nof_buckets);
  memotable->lru_buckets_vector = new_vector(memotable->nof_buckets);
  memotable->mru_buckets = memotable->mru_buckets_vector->the_array;
  memotable->lru_buckets = memotable->lru_buckets_vector->the_array;
  memotable->free_entries = memotable->max_entries;
  memotable->current_timestamp = cons(NULL, NIL);
  memotable->mru_timestamp = memotable->current_timestamp;
  memotable->lru_timestamp = memotable->current_timestamp;
}

Cons* lookup_mru_memoized_value(Mru_Memoization_Table* memotable, Object* arg1, Object* arg2, Object* arg3, Object* arg4, int eqvector) {
  { Object** mrubuckets = memotable->mru_buckets;
    Object** lrubuckets = NULL;
    Cons* mrubucket = NIL;
    Cons* lrubucket = NIL;
    Cons* mrutimestamp = NULL;
    int hashcode = 0;
    int bucketindex = 0;
    Cons* entry = NULL;

    if (mrubuckets == NULL) {
      initialize_mru_bucket_tables(memotable);
      mrubuckets = memotable->mru_buckets;
    }
    mrutimestamp = memotable->mru_timestamp;
    if (!(memotable->current_timestamp == mrutimestamp)) {
      mrutimestamp = memotable->current_timestamp;
      memotable->mru_timestamp = mrutimestamp;
      memotable->lru_timestamp = mrutimestamp;
      memotable->free_entries = memotable->max_entries;
    }
    hashcode = hash_memoized_arguments(arg1, arg2, arg3, arg4, eqvector);
    bucketindex = (hashcode % (memotable->nof_buckets));
    mrubucket = ((Cons*)(mrubuckets[bucketindex]));
    if (((boolean)(mrubucket))) {
      if (((Cons*)(mrubucket->value)) == mrutimestamp) {
        entry = find_memoized_value_entry(mrubucket, arg1, arg2, arg3, arg4, eqvector, FALSE);
        if (((boolean)(entry))) {
          return (entry);
        }
      }
      else {
        mrubucket->value = mrutimestamp;
        mrubucket->rest = NIL;
      }
    }
    lrubuckets = memotable->lru_buckets;
    lrubucket = ((Cons*)(lrubuckets[bucketindex]));
    if (((boolean)(lrubucket))) {
      if (((Cons*)(lrubucket->value)) == memotable->lru_timestamp) {
        entry = find_memoized_value_entry(lrubucket, arg1, arg2, arg3, arg4, eqvector, TRUE);
      }
      else {
        lrubucket->rest = NIL;
      }
    }
    if (!((boolean)(entry))) {
      entry = make_memoized_value_entry(NULL, arg1, arg2, arg3, arg4);
    }
    if (((boolean)(mrubucket))) {
      mrubucket->rest = cons(entry, mrubucket->rest);
    }
    else {
      mrubuckets[bucketindex] = (cons(mrutimestamp, cons(entry, NIL)));
    }
    if ((memotable->free_entries = memotable->free_entries - 1) == 0) {
      memotable->mru_buckets = lrubuckets;
      memotable->lru_buckets = mrubuckets;
      memotable->mru_timestamp = cons(NULL, NIL);
      memotable->current_timestamp = memotable->mru_timestamp;
      memotable->lru_timestamp = mrutimestamp;
      memotable->free_entries = memotable->max_entries;
    }
    return (entry);
  }
}

Cons* lookup_mru_memoized_valuen(Mru_Memoization_Table* memotable, Cons* args, int eqvector) {
  { Object** mrubuckets = memotable->mru_buckets;
    Object** lrubuckets = NULL;
    Cons* mrubucket = NIL;
    Cons* lrubucket = NIL;
    Cons* mrutimestamp = NULL;
    int hashcode = 0;
    int bucketindex = 0;
    Cons* entry = NULL;

    if (mrubuckets == NULL) {
      initialize_mru_bucket_tables(memotable);
      mrubuckets = memotable->mru_buckets;
    }
    mrutimestamp = memotable->mru_timestamp;
    if (!(memotable->current_timestamp == mrutimestamp)) {
      mrutimestamp = memotable->current_timestamp;
      memotable->mru_timestamp = mrutimestamp;
      memotable->lru_timestamp = mrutimestamp;
      memotable->free_entries = memotable->max_entries;
    }
    hashcode = hash_memoized_argumentsn(args, eqvector);
    bucketindex = (hashcode % (memotable->nof_buckets));
    mrubucket = ((Cons*)(mrubuckets[bucketindex]));
    if (((boolean)(mrubucket))) {
      if (((Cons*)(mrubucket->value)) == mrutimestamp) {
        entry = find_memoized_value_entryn(mrubucket, args, eqvector, FALSE);
        if (((boolean)(entry))) {
          return (entry);
        }
      }
      else {
        mrubucket->value = mrutimestamp;
        mrubucket->rest = NIL;
      }
    }
    lrubuckets = memotable->lru_buckets;
    lrubucket = ((Cons*)(lrubuckets[bucketindex]));
    if (((boolean)(lrubucket))) {
      if (((Cons*)(lrubucket->value)) == memotable->lru_timestamp) {
        entry = find_memoized_value_entryn(lrubucket, args, eqvector, TRUE);
      }
      else {
        lrubucket->rest = NIL;
      }
    }
    if (!((boolean)(entry))) {
      entry = cons(NULL, args);
    }
    if (((boolean)(mrubucket))) {
      mrubucket->rest = cons(entry, mrubucket->rest);
    }
    else {
      mrubuckets[bucketindex] = (cons(mrutimestamp, cons(entry, NIL)));
    }
    if ((memotable->free_entries = memotable->free_entries - 1) == 0) {
      memotable->mru_buckets = lrubuckets;
      memotable->lru_buckets = mrubuckets;
      memotable->mru_timestamp = cons(NULL, NIL);
      memotable->current_timestamp = memotable->mru_timestamp;
      memotable->lru_timestamp = mrutimestamp;
      memotable->free_entries = memotable->max_entries;
    }
    return (entry);
  }
}

void initialize_memoization_table(Surrogate* memotablesurrogate, char* options) {
  { Property_List* self_000 = new_property_list();

    self_000->the_plist = ((Cons*)(read_s_expression_from_string(options)));
    { Property_List* theoptions = self_000;
      Object* oldtable = memotablesurrogate->surrogate_value;
      Integer_Wrapper* size = ((Integer_Wrapper*)(theoptions->lookup(KWD_MEMOIZE_MAX_VALUES)));

      { Memoization_Table* choose_value_000 = NULL;

        if (((boolean)(size))) {
          { Mru_Memoization_Table* self_001 = new_mru_memoization_table();

            self_001->max_entries = size->wrapper_value;
            choose_value_000 = self_001;
          }
        }
        else {
          choose_value_000 = new_memoization_table();
        }
        { Memoization_Table* table = choose_value_000;

          table->table_name = memotablesurrogate;
          if (((boolean)(theoptions->lookup(KWD_MEMOIZE_TIMESTAMPS)))) {
            table->timestamps = ((Cons*)(theoptions->lookup(KWD_MEMOIZE_TIMESTAMPS)));
          }
          oALL_MEMOIZATION_TABLESo->remove(((Memoization_Table*)(oldtable)));
          oALL_MEMOIZATION_TABLESo->insert(table);
          memotablesurrogate->surrogate_value = table;
        }
      }
    }
  }
}

void clear_all_memoization_tables() {
  { Memoization_Table* table = NULL;
    Cons* iter_000 = oALL_MEMOIZATION_TABLESo->the_cons_list;

    while (!(iter_000 == NIL)) {
      table = ((Memoization_Table*)(iter_000->value));
      iter_000 = iter_000->rest;
      table->table_name->surrogate_value = NULL;
    }
  }
  oALL_MEMOIZATION_TABLESo->clear();
}

void clear_memoization_tables(Keyword* timestampname) {
  { Memoization_Table* table = NULL;
    Cons* iter_000 = oALL_MEMOIZATION_TABLESo->the_cons_list;

    while (!(iter_000 == NIL)) {
      table = ((Memoization_Table*)(iter_000->value));
      iter_000 = iter_000->rest;
      { boolean foundP_000 = FALSE;

        { Keyword* ts = NULL;
          Cons* iter_001 = table->timestamps;

          while (!(iter_001 == NIL)) {
            ts = ((Keyword*)(iter_001->value));
            iter_001 = iter_001->rest;
            if (ts == timestampname) {
              foundP_000 = TRUE;
              break;
            }
          }
        }
        if (foundP_000) {
          if (subtype_ofP(safe_primary_type(table), SGT_MEMOIZE_STELLA_MRU_MEMOIZATION_TABLE)) {
            { Memoization_Table* table_000 = table;
              Mru_Memoization_Table* table = ((Mru_Memoization_Table*)(table_000));

              if (((boolean)(table->mru_buckets_vector))) {
                table->mru_buckets_vector->clear();
              }
              if (((boolean)(table->lru_buckets_vector))) {
                table->lru_buckets_vector->clear();
              }
              table->free_entries = table->max_entries;
            }
          }
          else {
            table->hash_table = NULL;
          }
        }
      }
    }
  }
}

void bump_memoization_timestamp(Keyword* timestampname) {
  { Memoization_Table* table = NULL;
    Cons* iter_000 = oALL_MEMOIZATION_TABLESo->the_cons_list;

    while (!(iter_000 == NIL)) {
      table = ((Memoization_Table*)(iter_000->value));
      iter_000 = iter_000->rest;
      { boolean foundP_000 = FALSE;

        { Keyword* ts = NULL;
          Cons* iter_001 = table->timestamps;

          while (!(iter_001 == NIL)) {
            ts = ((Keyword*)(iter_001->value));
            iter_001 = iter_001->rest;
            if (ts == timestampname) {
              foundP_000 = TRUE;
              break;
            }
          }
        }
        if (foundP_000) {
          table->current_timestamp = cons(NULL, NIL);
        }
      }
    }
  }
}

Surrogate* make_memoization_table_surrogate(Symbol* memoname) {
  if (!((boolean)(memoname))) {
    memoname = (((boolean)(oMETHODBEINGWALKEDo.get())) ? local_gensym(string_concatenate(oMETHODBEINGWALKEDo.get()->slot_name->symbol_name, "-MEMO-TABLE", 0)) : gensym("MEMO-TABLE"));
  }
  return (memoname->surrogatify());
}

Property_List* parse_memoize_options(Cons* options) {
  { Property_List* self_000 = new_property_list();

    self_000->the_plist = options;
    { Property_List* givenoptions = self_000;
      Property_List* parsedoptions = new_property_list();

      { Object* key = NULL;
        Object* value = NULL;
        Cons* iter_000 = givenoptions->the_plist;

        while (!(iter_000 == NIL)) {
          key = iter_000->value;
          value = iter_000->rest->value;
          iter_000 = iter_000->rest->rest;
          { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(key));

            if (test_value_000 == KWD_MEMOIZE_TIMESTAMPS) {
              if (!consP(value)) {
                value = cons(value, NIL);
              }
              { boolean test_value_001 = FALSE;

                { boolean alwaysP_000 = TRUE;

                  { Object* ts = NULL;
                    Cons* iter_001 = ((Cons*)(value));

                    while (!(iter_001 == NIL)) {
                      ts = iter_001->value;
                      iter_001 = iter_001->rest;
                      if (!keywordP(ts)) {
                        alwaysP_000 = FALSE;
                        break;
                      }
                    }
                  }
                  test_value_001 = alwaysP_000;
                }
                test_value_001 = !test_value_001;
                if (test_value_001) {
                  { 
                    BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                    signal_translation_error();
                    if (!(suppress_warningsP())) {
                      print_error_context(">> ERROR: ", STANDARD_ERROR);
                      *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal :timestamp value in memoize: " << "`" << de_uglify_parse_tree(value) << "'" << "." << endl;
                    }
                  }
                }
                else {
                  parsedoptions->insert_at(key, value);
                }
              }
            }
            else if (test_value_000 == KWD_MEMOIZE_NAME) {
              if (!isaP(value, SGT_MEMOIZE_STELLA_GENERALIZED_SYMBOL)) {
                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  signal_translation_error();
                  if (!(suppress_warningsP())) {
                    print_error_context(">> ERROR: ", STANDARD_ERROR);
                    *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal :name value in memoize: " << "`" << de_uglify_parse_tree(value) << "'" << "." << endl;
                  }
                }
              }
              else {
                parsedoptions->insert_at(key, make_memoization_table_surrogate(intern_derived_symbol(((Generalized_Symbol*)(value)), ((Generalized_Symbol*)(value))->symbol_name)));
              }
            }
            else if (test_value_000 == KWD_MEMOIZE_MAX_VALUES) {
              if (!integerP(value)) {
                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  signal_translation_error();
                  if (!(suppress_warningsP())) {
                    print_error_context(">> ERROR: ", STANDARD_ERROR);
                    *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal :max-values value in memoize: " << "`" << de_uglify_parse_tree(value) << "'" << "." << endl;
                  }
                }
              }
              else {
                parsedoptions->insert_at(key, value);
              }
            }
            else {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal memoize option: " << "`" << de_uglify_parse_tree(key) << "'" << "." << endl;
                }
              }
            }
          }
        }
      }
      return (parsedoptions);
    }
  }
}

Object* memoize_expander(Cons* inputargs, Cons* body) {
  // Compute the value of an expression and memoize it relative to
  //    the values of `inputArgs'.
  // `inputArgs' should characterize the complete set of values upon which
  //    the computation of the result depended.
  // Calls to `memoize' should be of the form
  // 
  //       (memoize (<arg>+) {:<option> <value>}* <expression>)
  // 
  //    and have the status of an expression.
  //    The following options are supported:
  // 
  //       :timestamps   A single or list of keywords specifying the names of
  //                     timestamps which when bumped should invalidate all
  //                     entries currently memoized in this table.
  //       :name         Names the memoization table so it can be shared by other
  //                     memoization sites.  By default, a gensymed name is used.
  //                     CAUTION: IT IS ASSUMED THAT ALL ENTRIES IN A MEMOZATION
  //                     TABLE DEPEND ON THE SAME NUMBER OF ARGUMENTS!!
  //       :max-values   The maximum number of values to be memoized.  Only the
  //                     `:max-values' most recently used values will be kept
  //                     in the memoization table, older values will be discarded
  //                     and recomputed if needed.  Without a `:max-values'
  //                     specification, the memoization table will grow
  //                     indefinitely.
  // 
  // PERFORMANCE NOTES: For most efficient lookup, input arguments that vary the most
  // should be listed first.  Also, arguments of type STANDARD-OBJECT (and all its
  // subtypes) can be memoized more efficiently than arguments of type OBJECT or
  // (wrapped) literals (with the exception of BOOLEANs).
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    { Surrogate* memoname = NULL;
      Symbol* memotablevar = local_gensym("MEMO-TABLE");
      Symbol* memoentryvar = local_gensym("MEMOIZED-ENTRY");
      Symbol* memovaluevar = local_gensym("MEMOIZED-VALUE");
      int nofinputs = inputargs->length();
      Property_List* options = NULL;
      Object* lookupfunction = NULL;
      Cons* lookuptree = NIL;
      boolean alleqP = TRUE;
      int eqvector = 0;

      body = cons(NIL, body);
      options = parse_memoize_options(extract_options(body, NULL));
      body = body->rest;
      memoname = ((Surrogate*)(options->lookup(KWD_MEMOIZE_NAME)));
      if (!((boolean)(memoname))) {
        memoname = make_memoization_table_surrogate(NULL);
      }
      { int index = NULL_INTEGER;
        Cons_Iterator* it = inputargs->allocate_iterator();
        int iter_000 = 0;

        while (it->nextP()) {
          index = iter_000;
          iter_000 = iter_000 + 1;
          { Object* walkedarg = NULL;
            Standard_Object* type = NULL;

            walkedarg = walk_a_tree(it->value, type);
            if (sub_type_spec_ofP(type, SGT_MEMOIZE_STELLA_STANDARD_OBJECT) ||
                (sub_type_spec_ofP(type, SGT_MEMOIZE_STELLA_BOOLEAN_WRAPPER) ||
                 sub_type_spec_ofP(type, SGT_MEMOIZE_STELLA_BOOLEAN))) {
              if (index < 12) {
                eqvector = (eqvector | ((1 << index)));
              }
            }
            else {
              alleqP = FALSE;
            }
            it->value_setter(sys_tree(walkedarg, type, dummy1));
          }
        }
      }
      lookupfunction = ((!((boolean)(options->lookup(KWD_MEMOIZE_MAX_VALUES)))) ? (((nofinputs <= 4) ? SYM_MEMOIZE_STELLA_LOOKUP_MEMOIZED_VALUE : SYM_MEMOIZE_STELLA_LOOKUP_MEMOIZED_VALUEN)) : (((nofinputs <= 4) ? SYM_MEMOIZE_STELLA_LOOKUP_MRU_MEMOIZED_VALUE : SYM_MEMOIZE_STELLA_LOOKUP_MRU_MEMOIZED_VALUEN)));
      switch (nofinputs) {
        case 0: 
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Missing memoization input argument specification" << "." << endl;
            }
          }
          return (walk_dont_call_me_tree(body, SGT_MEMOIZE_STELLA_OBJECT, dummy2));
        case 1: 
          lookuptree = cons(lookupfunction, cons(memotablevar, cons(inputargs->value, listO(5, SYM_MEMOIZE_STELLA_MEMOIZED_NULL_VALUE, SYM_MEMOIZE_STELLA_NULL, SYM_MEMOIZE_STELLA_NULL, (alleqP ? wrap_integer(-1) : wrap_integer(eqvector)), NIL))));
        break;
        case 2: 
          lookuptree = cons(lookupfunction, cons(memotablevar, cons(inputargs->value, cons(inputargs->rest->value, listO(4, SYM_MEMOIZE_STELLA_MEMOIZED_NULL_VALUE, SYM_MEMOIZE_STELLA_NULL, (alleqP ? wrap_integer(-1) : wrap_integer(eqvector)), NIL)))));
        break;
        case 3: 
          lookuptree = cons(lookupfunction, cons(memotablevar, cons(inputargs->value, cons(inputargs->rest->value, cons(inputargs->rest->rest->value, listO(3, SYM_MEMOIZE_STELLA_MEMOIZED_NULL_VALUE, (alleqP ? wrap_integer(-1) : wrap_integer(eqvector)), NIL))))));
        break;
        case 4: 
          lookuptree = cons(lookupfunction, cons(memotablevar, cons(inputargs->value, cons(inputargs->rest->value, cons(inputargs->rest->rest->value, cons(inputargs->fourth(), cons((alleqP ? wrap_integer(-1) : wrap_integer(eqvector)), NIL)))))));
        break;
        default:
          lookuptree = cons(lookupfunction, cons(memotablevar, listO(3, cons(SYM_MEMOIZE_STELLA_CONS_LIST, inputargs->concatenate(NIL, 0)), (alleqP ? wrap_integer(-1) : wrap_integer(eqvector)), NIL)));
        break;
      }
      return (listO(7, SYM_MEMOIZE_STELLA_VRLET, listO(4, cons(memotablevar, listO(3, SYM_MEMOIZE_STELLA_MEMOIZATION_TABLE, listO(3, SYM_MEMOIZE_STELLA_SURROGATE_VALUE, memoname, NIL), NIL)), cons(memoentryvar, listO(3, SYM_MEMOIZE_STELLA_CONS, SYM_MEMOIZE_STELLA_NULL, NIL)), cons(memovaluevar, listO(3, SYM_MEMOIZE_STELLA_OBJECT, SYM_MEMOIZE_STELLA_NULL, NIL)), NIL), listO(5, SYM_MEMOIZE_STELLA_WHEN, listO(3, SYM_MEMOIZE_STELLA_NULLp, memotablevar, NIL), listO(3, SYM_MEMOIZE_STELLA_INITIALIZE_MEMOIZATION_TABLE, memoname, cons(wrap_string(stringify(options->the_plist)), NIL)), listO(3, SYM_MEMOIZE_STELLA_SETQ, memotablevar, cons(listO(3, SYM_MEMOIZE_STELLA_SURROGATE_VALUE, memoname, NIL), NIL)), NIL), listO(5, SYM_MEMOIZE_STELLA_WHEN, SYM_MEMOIZE_STELLA_oMEMOIZATION_ENABLEDpo, listO(3, SYM_MEMOIZE_STELLA_SETQ, memoentryvar, cons(lookuptree, NIL)), listO(3, SYM_MEMOIZE_STELLA_SETQ, memovaluevar, cons(listO(3, SYM_MEMOIZE_STELLA_FIRST, memoentryvar, NIL), NIL)), NIL), listO(4, SYM_MEMOIZE_STELLA_COND, listO(3, listO(3, SYM_MEMOIZE_STELLA_DEFINEDp, memovaluevar, NIL), listO(4, SYM_MEMOIZE_STELLA_WHEN, listO(3, SYM_MEMOIZE_STELLA_EQLp, memovaluevar, cons(SYM_MEMOIZE_STELLA_MEMOIZED_NULL_VALUE, NIL)), listO(3, SYM_MEMOIZE_STELLA_SETQ, memovaluevar, cons(SYM_MEMOIZE_STELLA_NULL, NIL)), NIL), NIL), listO(4, SYM_MEMOIZE_STELLA_OTHERWISE, listO(3, SYM_MEMOIZE_STELLA_SETQ, memovaluevar, body->concatenate(NIL, 0)), listO(4, SYM_MEMOIZE_STELLA_WHEN, SYM_MEMOIZE_STELLA_oMEMOIZATION_ENABLEDpo, listO(4, SYM_MEMOIZE_STELLA_SETF, listO(3, SYM_MEMOIZE_STELLA_VALUE, memoentryvar, NIL), listO(5, SYM_MEMOIZE_STELLA_CHOOSE, listO(3, SYM_MEMOIZE_STELLA_NULLp, memovaluevar, NIL), SYM_MEMOIZE_STELLA_MEMOIZED_NULL_VALUE, memovaluevar, NIL), NIL), NIL), NIL), NIL), memovaluevar, NIL));
    }
  }
}

void help_startup_memoize1() {
  {
    SGT_MEMOIZE_STELLA_MEMOIZATION_TABLE = ((Surrogate*)(intern_rigid_symbol_wrt_module("MEMOIZATION-TABLE", NULL, 1)));
    SYM_MEMOIZE_STELLA_HASH_TABLE = ((Symbol*)(intern_rigid_symbol_wrt_module("HASH-TABLE", NULL, 0)));
    SYM_MEMOIZE_STELLA_TABLE_NAME = ((Symbol*)(intern_rigid_symbol_wrt_module("TABLE-NAME", NULL, 0)));
    SYM_MEMOIZE_STELLA_CURRENT_TIMESTAMP = ((Symbol*)(intern_rigid_symbol_wrt_module("CURRENT-TIMESTAMP", NULL, 0)));
    SYM_MEMOIZE_STELLA_TIMESTAMPS = ((Symbol*)(intern_rigid_symbol_wrt_module("TIMESTAMPS", NULL, 0)));
    SGT_MEMOIZE_STELLA_MRU_MEMOIZATION_TABLE = ((Surrogate*)(intern_rigid_symbol_wrt_module("MRU-MEMOIZATION-TABLE", NULL, 1)));
    SYM_MEMOIZE_STELLA_MRU_BUCKETS_VECTOR = ((Symbol*)(intern_rigid_symbol_wrt_module("MRU-BUCKETS-VECTOR", NULL, 0)));
    SYM_MEMOIZE_STELLA_LRU_BUCKETS_VECTOR = ((Symbol*)(intern_rigid_symbol_wrt_module("LRU-BUCKETS-VECTOR", NULL, 0)));
    SYM_MEMOIZE_STELLA_MRU_TIMESTAMP = ((Symbol*)(intern_rigid_symbol_wrt_module("MRU-TIMESTAMP", NULL, 0)));
    SYM_MEMOIZE_STELLA_LRU_TIMESTAMP = ((Symbol*)(intern_rigid_symbol_wrt_module("LRU-TIMESTAMP", NULL, 0)));
    SYM_MEMOIZE_STELLA_NOF_BUCKETS = ((Symbol*)(intern_rigid_symbol_wrt_module("NOF-BUCKETS", NULL, 0)));
    SYM_MEMOIZE_STELLA_FREE_ENTRIES = ((Symbol*)(intern_rigid_symbol_wrt_module("FREE-ENTRIES", NULL, 0)));
    SYM_MEMOIZE_STELLA_MAX_ENTRIES = ((Symbol*)(intern_rigid_symbol_wrt_module("MAX-ENTRIES", NULL, 0)));
    KWD_MEMOIZE_MAX_VALUES = ((Keyword*)(intern_rigid_symbol_wrt_module("MAX-VALUES", NULL, 2)));
    KWD_MEMOIZE_TIMESTAMPS = ((Keyword*)(intern_rigid_symbol_wrt_module("TIMESTAMPS", NULL, 2)));
    KWD_MEMOIZE_NAME = ((Keyword*)(intern_rigid_symbol_wrt_module("NAME", NULL, 2)));
    SGT_MEMOIZE_STELLA_GENERALIZED_SYMBOL = ((Surrogate*)(intern_rigid_symbol_wrt_module("GENERALIZED-SYMBOL", NULL, 1)));
    SGT_MEMOIZE_STELLA_STANDARD_OBJECT = ((Surrogate*)(intern_rigid_symbol_wrt_module("STANDARD-OBJECT", NULL, 1)));
    SGT_MEMOIZE_STELLA_BOOLEAN_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("BOOLEAN-WRAPPER", NULL, 1)));
    SGT_MEMOIZE_STELLA_BOOLEAN = ((Surrogate*)(intern_rigid_symbol_wrt_module("BOOLEAN", NULL, 1)));
    SYM_MEMOIZE_STELLA_LOOKUP_MEMOIZED_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("LOOKUP-MEMOIZED-VALUE", NULL, 0)));
    SYM_MEMOIZE_STELLA_LOOKUP_MEMOIZED_VALUEN = ((Symbol*)(intern_rigid_symbol_wrt_module("LOOKUP-MEMOIZED-VALUEN", NULL, 0)));
    SYM_MEMOIZE_STELLA_LOOKUP_MRU_MEMOIZED_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("LOOKUP-MRU-MEMOIZED-VALUE", NULL, 0)));
    SYM_MEMOIZE_STELLA_LOOKUP_MRU_MEMOIZED_VALUEN = ((Symbol*)(intern_rigid_symbol_wrt_module("LOOKUP-MRU-MEMOIZED-VALUEN", NULL, 0)));
    SGT_MEMOIZE_STELLA_OBJECT = ((Surrogate*)(intern_rigid_symbol_wrt_module("OBJECT", NULL, 1)));
    SYM_MEMOIZE_STELLA_MEMOIZED_NULL_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("MEMOIZED-NULL-VALUE", NULL, 0)));
    SYM_MEMOIZE_STELLA_NULL = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL", NULL, 0)));
    SYM_MEMOIZE_STELLA_CONS_LIST = ((Symbol*)(intern_rigid_symbol_wrt_module("CONS-LIST", NULL, 0)));
    SYM_MEMOIZE_STELLA_VRLET = ((Symbol*)(intern_rigid_symbol_wrt_module("VRLET", NULL, 0)));
    SYM_MEMOIZE_STELLA_MEMOIZATION_TABLE = ((Symbol*)(intern_rigid_symbol_wrt_module("MEMOIZATION-TABLE", NULL, 0)));
    SYM_MEMOIZE_STELLA_SURROGATE_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("SURROGATE-VALUE", NULL, 0)));
    SYM_MEMOIZE_STELLA_CONS = ((Symbol*)(intern_rigid_symbol_wrt_module("CONS", NULL, 0)));
    SYM_MEMOIZE_STELLA_OBJECT = ((Symbol*)(intern_rigid_symbol_wrt_module("OBJECT", NULL, 0)));
    SYM_MEMOIZE_STELLA_WHEN = ((Symbol*)(intern_rigid_symbol_wrt_module("WHEN", NULL, 0)));
    SYM_MEMOIZE_STELLA_NULLp = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL?", NULL, 0)));
    SYM_MEMOIZE_STELLA_INITIALIZE_MEMOIZATION_TABLE = ((Symbol*)(intern_rigid_symbol_wrt_module("INITIALIZE-MEMOIZATION-TABLE", NULL, 0)));
    SYM_MEMOIZE_STELLA_SETQ = ((Symbol*)(intern_rigid_symbol_wrt_module("SETQ", NULL, 0)));
    SYM_MEMOIZE_STELLA_oMEMOIZATION_ENABLEDpo = ((Symbol*)(intern_rigid_symbol_wrt_module("*MEMOIZATION-ENABLED?*", NULL, 0)));
    SYM_MEMOIZE_STELLA_FIRST = ((Symbol*)(intern_rigid_symbol_wrt_module("FIRST", NULL, 0)));
    SYM_MEMOIZE_STELLA_COND = ((Symbol*)(intern_rigid_symbol_wrt_module("COND", NULL, 0)));
    SYM_MEMOIZE_STELLA_DEFINEDp = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFINED?", NULL, 0)));
    SYM_MEMOIZE_STELLA_EQLp = ((Symbol*)(intern_rigid_symbol_wrt_module("EQL?", NULL, 0)));
    SYM_MEMOIZE_STELLA_OTHERWISE = ((Symbol*)(intern_rigid_symbol_wrt_module("OTHERWISE", NULL, 0)));
    SYM_MEMOIZE_STELLA_SETF = ((Symbol*)(intern_rigid_symbol_wrt_module("SETF", NULL, 0)));
    SYM_MEMOIZE_STELLA_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("VALUE", NULL, 0)));
    SYM_MEMOIZE_STELLA_CHOOSE = ((Symbol*)(intern_rigid_symbol_wrt_module("CHOOSE", NULL, 0)));
    SYM_MEMOIZE_STELLA_MEMOIZE = ((Symbol*)(intern_rigid_symbol_wrt_module("MEMOIZE", NULL, 0)));
    SYM_MEMOIZE_STELLA_MEMOIZE_EXPANDER = ((Symbol*)(intern_rigid_symbol_wrt_module("MEMOIZE-EXPANDER", NULL, 0)));
    SYM_MEMOIZE_STELLA_STARTUP_MEMOIZE = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-MEMOIZE", NULL, 0)));
    SYM_MEMOIZE_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-STARTUP-CLASSNAME", NULL, 0)));
  }
}

void startup_memoize() {
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oSTELLA_MODULEo);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (current_startup_time_phaseP(2)) {
      help_startup_memoize1();
    }
    if (current_startup_time_phaseP(4)) {
      oALL_MEMOIZATION_TABLESo = new_list();
      { Symbol* self_056 = new_symbol("MEMOIZED-NULL-VALUE");

        self_056->home_context = NULL;
        MEMOIZED_NULL_VALUE = self_056;
      }
    }
    if (current_startup_time_phaseP(5)) {
      { Class* renamed_Class = define_class_from_stringified_source("MEMOIZATION-TABLE", "(DEFCLASS MEMOIZATION-TABLE (STANDARD-OBJECT) :DOCUMENTATION \"Table that maps argument values onto computed results for\none individual memoization site.\" :SLOTS ((HASH-TABLE :TYPE (INTEGER-HASH-TABLE OF INTEGER (CONS OF CONS)) :DOCUMENTATION \"The table holding the memoized values.\nWe use an integer table, since we explicitly compute a hash code by combining\nhash codes of argument values.\") (TABLE-NAME :TYPE SURROGATE :DOCUMENTATION \"The surrogate used to point to this table.\nUsed at the memoization site for quick memo table lookup.\") (CURRENT-TIMESTAMP :TYPE CONS :DOCUMENTATION \"Marker value used to indicate valid memoized entries.\nThis marker changes everytime one of the `timestamps' gets bumped.\") (TIMESTAMPS :TYPE (CONS OF KEYWORD) :INITIALLY NIL :DOCUMENTATION \"Names of timestamps that trigger invalidation of\nmemoized entries when they get bumped.\")))");

        renamed_Class->class_constructor_code = ((cpp_function_code)(&new_memoization_table));
        renamed_Class->class_slot_accessor_code = ((cpp_function_code)(&access_memoization_table_slot_value));
      }
      { Class* renamed_Class = define_class_from_stringified_source("MRU-MEMOIZATION-TABLE", "(DEFCLASS MRU-MEMOIZATION-TABLE (MEMOIZATION-TABLE) :SLOTS ((MRU-BUCKETS-VECTOR :TYPE (VECTOR OF (CONS OF CONS))) (LRU-BUCKETS-VECTOR :TYPE (VECTOR OF (CONS OF CONS))) (MRU-BUCKETS :TYPE (NATIVE-VECTOR OF (CONS OF CONS))) (LRU-BUCKETS :TYPE (NATIVE-VECTOR OF (CONS OF CONS))) (MRU-TIMESTAMP :TYPE CONS) (LRU-TIMESTAMP :TYPE CONS) (NOF-BUCKETS :TYPE INTEGER) (FREE-ENTRIES :TYPE INTEGER) (MAX-ENTRIES :TYPE INTEGER)))");

        renamed_Class->class_constructor_code = ((cpp_function_code)(&new_mru_memoization_table));
        renamed_Class->class_slot_accessor_code = ((cpp_function_code)(&access_mru_memoization_table_slot_value));
      }
    }
    if (current_startup_time_phaseP(6)) {
      finalize_classes();
    }
    if (current_startup_time_phaseP(7)) {
      define_function_object("ENABLE-MEMOIZATION", "(DEFUN ENABLE-MEMOIZATION () :COMMAND? TRUE :PUBLIC? TRUE :DOCUMENTATION \"Enable memoization and use of memoized expression results.\")", ((cpp_function_code)(&enable_memoization)), NULL);
      define_function_object("DISABLE-MEMOIZATION", "(DEFUN DISABLE-MEMOIZATION () :COMMAND? TRUE :PUBLIC? TRUE :DOCUMENTATION \"Enable memoization and use of memoized expression results.\")", ((cpp_function_code)(&disable_memoization)), NULL);
      define_function_object("HASH-MEMOIZED-ARGUMENTS", "(DEFUN (HASH-MEMOIZED-ARGUMENTS INTEGER) ((ARG1 OBJECT) (ARG2 OBJECT) (ARG3 OBJECT) (ARG4 OBJECT) (EQVECTOR INTEGER)))", ((cpp_function_code)(&hash_memoized_arguments)), NULL);
      define_function_object("HASH-MEMOIZED-ARGUMENTSN", "(DEFUN (HASH-MEMOIZED-ARGUMENTSN INTEGER) ((TUPLE CONS) (EQVECTOR INTEGER)))", ((cpp_function_code)(&hash_memoized_argumentsn)), NULL);
      define_function_object("MAKE-MEMOIZED-VALUE-ENTRY", "(DEFUN (MAKE-MEMOIZED-VALUE-ENTRY (CONS OF CONS)) ((VALUE OBJECT) (ARG1 OBJECT) (ARG2 OBJECT) (ARG3 OBJECT) (ARG4 OBJECT)))", ((cpp_function_code)(&make_memoized_value_entry)), NULL);
      define_function_object("MAKE-MEMOIZED-VALUE-ENTRYN", "(DEFUN (MAKE-MEMOIZED-VALUE-ENTRYN (CONS OF CONS)) ((VALUE OBJECT) (ARGS CONS)) :GLOBALLY-INLINE? TRUE (RETURN (CONS VALUE ARGS)))", ((cpp_function_code)(&make_memoized_value_entryn)), NULL);
      define_function_object("FIND-MEMOIZED-VALUE-ENTRY", "(DEFUN (FIND-MEMOIZED-VALUE-ENTRY CONS) ((BUCKET (CONS OF CONS)) (ARG1 OBJECT) (ARG2 OBJECT) (ARG3 OBJECT) (ARG4 OBJECT) (EQVECTOR INTEGER) (DELETEENTRY? BOOLEAN)))", ((cpp_function_code)(&find_memoized_value_entry)), NULL);
      define_function_object("FIND-MEMOIZED-VALUE-ENTRYN", "(DEFUN (FIND-MEMOIZED-VALUE-ENTRYN CONS) ((BUCKET (CONS OF CONS)) (TUPLE CONS) (EQVECTOR INTEGER) (DELETEENTRY? BOOLEAN)))", ((cpp_function_code)(&find_memoized_value_entryn)), NULL);
      define_function_object("LOOKUP-MEMOIZED-VALUE", "(DEFUN (LOOKUP-MEMOIZED-VALUE CONS) ((MEMOTABLE MEMOIZATION-TABLE) (ARG1 OBJECT) (ARG2 OBJECT) (ARG3 OBJECT) (ARG4 OBJECT) (EQVECTOR INTEGER)))", ((cpp_function_code)(&lookup_memoized_value)), NULL);
      define_function_object("LOOKUP-MEMOIZED-VALUEN", "(DEFUN (LOOKUP-MEMOIZED-VALUEN CONS) ((MEMOTABLE MEMOIZATION-TABLE) (ARGS CONS) (EQVECTOR INTEGER)))", ((cpp_function_code)(&lookup_memoized_valuen)), NULL);
      define_function_object("INITIALIZE-MRU-BUCKET-TABLES", "(DEFUN INITIALIZE-MRU-BUCKET-TABLES ((MEMOTABLE MRU-MEMOIZATION-TABLE)))", ((cpp_function_code)(&initialize_mru_bucket_tables)), NULL);
      define_function_object("LOOKUP-MRU-MEMOIZED-VALUE", "(DEFUN (LOOKUP-MRU-MEMOIZED-VALUE CONS) ((MEMOTABLE MRU-MEMOIZATION-TABLE) (ARG1 OBJECT) (ARG2 OBJECT) (ARG3 OBJECT) (ARG4 OBJECT) (EQVECTOR INTEGER)))", ((cpp_function_code)(&lookup_mru_memoized_value)), NULL);
      define_function_object("LOOKUP-MRU-MEMOIZED-VALUEN", "(DEFUN (LOOKUP-MRU-MEMOIZED-VALUEN CONS) ((MEMOTABLE MRU-MEMOIZATION-TABLE) (ARGS CONS) (EQVECTOR INTEGER)))", ((cpp_function_code)(&lookup_mru_memoized_valuen)), NULL);
      define_function_object("INITIALIZE-MEMOIZATION-TABLE", "(DEFUN INITIALIZE-MEMOIZATION-TABLE ((MEMOTABLESURROGATE SURROGATE) (OPTIONS STRING)))", ((cpp_function_code)(&initialize_memoization_table)), NULL);
      define_function_object("CLEAR-ALL-MEMOIZATION-TABLES", "(DEFUN CLEAR-ALL-MEMOIZATION-TABLES ())", ((cpp_function_code)(&clear_all_memoization_tables)), NULL);
      define_function_object("CLEAR-MEMOIZATION-TABLES", "(DEFUN CLEAR-MEMOIZATION-TABLES ((TIMESTAMPNAME KEYWORD)))", ((cpp_function_code)(&clear_memoization_tables)), NULL);
      define_function_object("BUMP-MEMOIZATION-TIMESTAMP", "(DEFUN BUMP-MEMOIZATION-TIMESTAMP ((TIMESTAMPNAME KEYWORD)))", ((cpp_function_code)(&bump_memoization_timestamp)), NULL);
      define_function_object("MAKE-MEMOIZATION-TABLE-SURROGATE", "(DEFUN (MAKE-MEMOIZATION-TABLE-SURROGATE SURROGATE) ((MEMONAME SYMBOL)))", ((cpp_function_code)(&make_memoization_table_surrogate)), NULL);
      define_function_object("PARSE-MEMOIZE-OPTIONS", "(DEFUN (PARSE-MEMOIZE-OPTIONS PROPERTY-LIST) ((OPTIONS CONS)))", ((cpp_function_code)(&parse_memoize_options)), NULL);
      define_function_object("MEMOIZE-EXPANDER", "(DEFUN (MEMOIZE-EXPANDER OBJECT) ((INPUTARGS CONS) |&BODY| (BODY CONS)) :DOCUMENTATION \"Compute the value of an expression and memoize it relative to\n   the values of `inputArgs'.\n`inputArgs' should characterize the complete set of values upon which\n   the computation of the result depended.\nCalls to `memoize' should be of the form\n\n      (memoize (<arg>+) {:<option> <value>}* <expression>)\n\n   and have the status of an expression.\n   The following options are supported:\n\n      :timestamps   A single or list of keywords specifying the names of\n                    timestamps which when bumped should invalidate all\n                    entries currently memoized in this table.\n      :name         Names the memoization table so it can be shared by other\n                    memoization sites.  By default, a gensymed name is used.\n                    CAUTION: IT IS ASSUMED THAT ALL ENTRIES IN A MEMOZATION\n                    TABLE DEPEND ON THE SAME NUMBER OF ARGUMENTS!!\n      :max-values   The maximum number of values to be memoized.  Only the\n                    `:max-values' most recently used values will be kept\n                    in the memoization table, older values will be discarded\n                    and recomputed if needed.  Without a `:max-values'\n                    specification, the memoization table will grow\n                    indefinitely.\n\nPERFORMANCE NOTES: For most efficient lookup, input arguments that vary the most\nshould be listed first.  Also, arguments of type STANDARD-OBJECT (and all its\nsubtypes) can be memoized more efficiently than arguments of type OBJECT or\n(wrapped) literals (with the exception of BOOLEANs).\" :PUBLIC? TRUE)", ((cpp_function_code)(&memoize_expander)), NULL);
      define_function_object("STARTUP-MEMOIZE", "(DEFUN STARTUP-MEMOIZE () :PUBLIC? TRUE)", ((cpp_function_code)(&startup_memoize)), NULL);
      { Method_Slot* function = lookup_function(SYM_MEMOIZE_STELLA_STARTUP_MEMOIZE);

        set_dynamic_slot_value(function->dynamic_slots, SYM_MEMOIZE_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string("Startup-Memoize"), NULL_STRING_WRAPPER);
      }
    }
    if (current_startup_time_phaseP(8)) {
      finalize_slots();
      cleanup_unfinalized_classes();
    }
    if (current_startup_time_phaseP(9)) {
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *ALL-MEMOIZATION-TABLES* (LIST OF MEMOIZATION-TABLE) (NEW LIST) :DOCUMENTATION \"Holds all currently active memoization tables for timestamp\nmaintenance and clearance purposes.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *MEMOIZATION-ENABLED?* BOOLEAN TRUE)");
      define_stella_global_variable_from_stringified_source("(DEFCONSTANT MEMOIZED-NULL-VALUE SYMBOL (NEW SYMBOL :SYMBOL-NAME \"MEMOIZED-NULL-VALUE\" :HOME-CONTEXT NULL) :DOCUMENTATION \"Used by memoization to indicate that a NULL value\nwas cached.  Needed to distinguish between an undefined value and a\ncached NULL.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *TOTAL-MEMOIZATION-LOOKUPS* INTEGER 0)");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *FAILED-MEMOIZATION-LOOKUPS* INTEGER 0)");
      store_macro(SYM_MEMOIZE_STELLA_MEMOIZE, SYM_MEMOIZE_STELLA_MEMOIZE_EXPANDER, ((cpp_function_code)(&memoize_expander)));
    }
  }
}

Surrogate* SGT_MEMOIZE_STELLA_MEMOIZATION_TABLE = NULL;

Symbol* SYM_MEMOIZE_STELLA_HASH_TABLE = NULL;

Symbol* SYM_MEMOIZE_STELLA_TABLE_NAME = NULL;

Symbol* SYM_MEMOIZE_STELLA_CURRENT_TIMESTAMP = NULL;

Symbol* SYM_MEMOIZE_STELLA_TIMESTAMPS = NULL;

Surrogate* SGT_MEMOIZE_STELLA_MRU_MEMOIZATION_TABLE = NULL;

Symbol* SYM_MEMOIZE_STELLA_MRU_BUCKETS_VECTOR = NULL;

Symbol* SYM_MEMOIZE_STELLA_LRU_BUCKETS_VECTOR = NULL;

Symbol* SYM_MEMOIZE_STELLA_MRU_TIMESTAMP = NULL;

Symbol* SYM_MEMOIZE_STELLA_LRU_TIMESTAMP = NULL;

Symbol* SYM_MEMOIZE_STELLA_NOF_BUCKETS = NULL;

Symbol* SYM_MEMOIZE_STELLA_FREE_ENTRIES = NULL;

Symbol* SYM_MEMOIZE_STELLA_MAX_ENTRIES = NULL;

Keyword* KWD_MEMOIZE_MAX_VALUES = NULL;

Keyword* KWD_MEMOIZE_TIMESTAMPS = NULL;

Keyword* KWD_MEMOIZE_NAME = NULL;

Surrogate* SGT_MEMOIZE_STELLA_GENERALIZED_SYMBOL = NULL;

Surrogate* SGT_MEMOIZE_STELLA_STANDARD_OBJECT = NULL;

Surrogate* SGT_MEMOIZE_STELLA_BOOLEAN_WRAPPER = NULL;

Surrogate* SGT_MEMOIZE_STELLA_BOOLEAN = NULL;

Symbol* SYM_MEMOIZE_STELLA_LOOKUP_MEMOIZED_VALUE = NULL;

Symbol* SYM_MEMOIZE_STELLA_LOOKUP_MEMOIZED_VALUEN = NULL;

Symbol* SYM_MEMOIZE_STELLA_LOOKUP_MRU_MEMOIZED_VALUE = NULL;

Symbol* SYM_MEMOIZE_STELLA_LOOKUP_MRU_MEMOIZED_VALUEN = NULL;

Surrogate* SGT_MEMOIZE_STELLA_OBJECT = NULL;

Symbol* SYM_MEMOIZE_STELLA_MEMOIZED_NULL_VALUE = NULL;

Symbol* SYM_MEMOIZE_STELLA_NULL = NULL;

Symbol* SYM_MEMOIZE_STELLA_CONS_LIST = NULL;

Symbol* SYM_MEMOIZE_STELLA_VRLET = NULL;

Symbol* SYM_MEMOIZE_STELLA_MEMOIZATION_TABLE = NULL;

Symbol* SYM_MEMOIZE_STELLA_SURROGATE_VALUE = NULL;

Symbol* SYM_MEMOIZE_STELLA_CONS = NULL;

Symbol* SYM_MEMOIZE_STELLA_OBJECT = NULL;

Symbol* SYM_MEMOIZE_STELLA_WHEN = NULL;

Symbol* SYM_MEMOIZE_STELLA_NULLp = NULL;

Symbol* SYM_MEMOIZE_STELLA_INITIALIZE_MEMOIZATION_TABLE = NULL;

Symbol* SYM_MEMOIZE_STELLA_SETQ = NULL;

Symbol* SYM_MEMOIZE_STELLA_oMEMOIZATION_ENABLEDpo = NULL;

Symbol* SYM_MEMOIZE_STELLA_FIRST = NULL;

Symbol* SYM_MEMOIZE_STELLA_COND = NULL;

Symbol* SYM_MEMOIZE_STELLA_DEFINEDp = NULL;

Symbol* SYM_MEMOIZE_STELLA_EQLp = NULL;

Symbol* SYM_MEMOIZE_STELLA_OTHERWISE = NULL;

Symbol* SYM_MEMOIZE_STELLA_SETF = NULL;

Symbol* SYM_MEMOIZE_STELLA_VALUE = NULL;

Symbol* SYM_MEMOIZE_STELLA_CHOOSE = NULL;

Symbol* SYM_MEMOIZE_STELLA_MEMOIZE = NULL;

Symbol* SYM_MEMOIZE_STELLA_MEMOIZE_EXPANDER = NULL;

Symbol* SYM_MEMOIZE_STELLA_STARTUP_MEMOIZE = NULL;

Symbol* SYM_MEMOIZE_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace stella
