// translate-file.cc

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

#include "stella/stella-system.hh"

namespace stella {

// If `true', record object signatures into a persistent table.
boolean oRECORD_SIGNATURESpo = FALSE;

// If `true', the walker will try to find previously recorded
// signatures for referenced but undefined functions, methods, etc.
boolean oWALK_WITH_RECORDED_SIGNATURESpo = FALSE;

// If `true', `safe-lookup-slot' and `lookup-function' will
// try to retrieve a previously recorded signature if normal lookup fails.
DEFINE_STELLA_SPECIAL(oUSERECORDEDSIGNATURESpo, boolean , FALSE);

DEFINE_STELLA_SPECIAL(oROOTSOURCEDIRECTORYo, char* , "PL:sources;");

DEFINE_STELLA_SPECIAL(oROOTNATIVEDIRECTORYo, char* , "PL:natives;");

DEFINE_STELLA_SPECIAL(oROOTBINARYDIRECTORYo, char* , "PL:binaries;");

// Points to the directory containing the SYSTEMS file.
DEFINE_STELLA_SPECIAL(oSYSTEMDEFINITIONSDIRECTORYo, char* , NULL);

char* root_source_directory() {
  return (translate_logical_pathname(oROOTSOURCEDIRECTORYo.get()));
}

char* root_native_directory() {
  return (translate_logical_pathname(oROOTNATIVEDIRECTORYo.get()));
}

char* root_binary_directory() {
  return (translate_logical_pathname(oROOTBINARYDIRECTORYo.get()));
}

char* system_definitions_directory() {
  if (oSYSTEMDEFINITIONSDIRECTORYo.get() != NULL) {
    return (translate_logical_pathname(oSYSTEMDEFINITIONSDIRECTORYo.get()));
  }
  else {
    return (string_concatenate(root_source_directory(), "systems", 0));
  }
}

// Points to the current system.
DEFINE_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONo, System_Definition* , NULL);

// Points to a path from the root directory down
// to a local directory containing sources, natives, or binaries.
DEFINE_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo, char* , "");

boolean handle_in_module_tree(Cons* tree, boolean seeninmoduleP, boolean& _return1) {
  { Object* renamed_Operator = tree->value;
    Object* firstarg = NULL;

    if (subtype_of_symbolP(safe_primary_type(renamed_Operator))) {
      { Object* operator_000 = renamed_Operator;
        Symbol* renamed_Operator = ((Symbol*)(operator_000));

        { char* test_value_000 = (oMODULEo.get()->case_sensitiveP ? string_upcase(renamed_Operator->symbol_name) : renamed_Operator->symbol_name);

          if (string_eqlP(test_value_000, "IN-PACKAGE")) {
            _return1 = seeninmoduleP;
            return (TRUE);
          }
          else if (string_eqlP(test_value_000, "IN-MODULE")) {
            if (seeninmoduleP) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_WARNING);
                  *(STANDARD_WARNING->native_stream) << endl << " " << "Ignoring duplicate IN-MODULE declaration: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
                }
              }
              _return1 = seeninmoduleP;
              return (TRUE);
            }
            firstarg = tree->rest->value;
            if (subtype_of_stringP(safe_primary_type(firstarg))) {
              { Object* firstarg_000 = firstarg;
                String_Wrapper* firstarg = ((String_Wrapper*)(firstarg_000));

                string_change_module(firstarg->wrapper_value);
                seeninmoduleP = TRUE;
              }
            }
            else {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_WARNING);
                  *(STANDARD_WARNING->native_stream) << endl << " " << "Ignoring illegal IN-MODULE declaration: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
                }
              }
            }
            _return1 = seeninmoduleP;
            return (TRUE);
          }
          else if (string_eqlP(test_value_000, "DEFMODULE") ||
              string_eqlP(test_value_000, "VERBATIM")) {
          }
          else {
            if (!(seeninmoduleP)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_WARNING);
                  *(STANDARD_WARNING->native_stream) << endl << " " << "Missing IN-MODULE declaration, translating " << "into module STELLA" << "." << endl;
                }
              }
              string_change_module("STELLA");
              seeninmoduleP = TRUE;
            }
          }
        }
      }
    }
    else {
    }
    _return1 = seeninmoduleP;
    return (FALSE);
  }
}

void translate_and_output_units_to_file(char* file, char* targetlanguage) {
  oTRANSLATIONPHASEo.set(KWD_TRANSLATE_FILE_TRANSLATE);
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_JAVA) {
  }
  else {
    if (oTRANSLATIONVERBOSITYLEVELo.get() >= 1) {
      cout << "Translating " << "`" << file << "'" << " to " << "`" << targetlanguage << "'" << "..." << endl;
    }
    translate_all_units();
  }
  output_all_units_to_file(file);
}

void operate_on_file(char* file, Keyword* operation) {
  { char* targetlanguage = oTRANSLATOROUTPUTLANGUAGEo.get()->symbol_name;
    Module* currentmodule = oMODULEo.get();
    Context* currentcontext = oCONTEXTo.get();
    boolean seeninmoduleP = FALSE;
    boolean skipP = FALSE;

    file = translate_logical_pathname(file);
    if (!((boolean)(operation))) {
      operation = KWD_TRANSLATE_FILE_TRANSLATE;
    }
    { 
      BIND_STELLA_SPECIAL(oTRANSLATIONPHASEo, Keyword*, NULL);
      BIND_STELLA_SPECIAL(oTRANSLATIONERRORSo, int, 0);
      BIND_STELLA_SPECIAL(oTRANSLATIONWARNINGSo, int, 0);
      BIND_STELLA_SPECIAL(oCURRENTFILEo, char*, file_base_name(file));
      BIND_STELLA_SPECIAL(oMODULEo, Module*, currentmodule);
      BIND_STELLA_SPECIAL(oCONTEXTo, Context*, currentcontext);
      if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_COMMON_LISP) {
        targetlanguage = "Common Lisp";
      }
      else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP) ||
          (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP_STANDALONE)) {
        targetlanguage = "C++";
      }
      else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_IDL) {
        targetlanguage = "IDL";
      }
      else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_JAVA) {
        targetlanguage = "Java";
      }
      else {
      }
      if (oTRANSLATIONVERBOSITYLEVELo.get() >= 1) {
        cout << "Processing " << "`" << file << "'" << ":" << endl;
        cout << "*** Pass 1, generating objects..." << endl;
      }
      { 
        BIND_STELLA_SPECIAL(oTRANSIENTOBJECTSpo, boolean, TRUE);
        { Object* tree = NULL;
          S_Expression_Iterator* iter_000 = s_expressions(new_input_file_stream(file));

          while (iter_000->nextP()) {
            tree = iter_000->value;
            { 
              BIND_STELLA_SPECIAL(oTRANSIENTOBJECTSpo, boolean, FALSE);
              if (safe_primary_type(tree) == SGT_TRANSLATE_FILE_STELLA_CONS) {
                { Object* tree_000 = tree;
                  Cons* tree = ((Cons*)(tree_000));

                  skipP = handle_in_module_tree(tree, seeninmoduleP, seeninmoduleP);
                  if (!(skipP)) {
                    { 
                      BIND_STELLA_SPECIAL(oTRANSLATIONPHASEo, Keyword*, KWD_TRANSLATE_FILE_DEFINE);
                      walk_top_level_tree(tree, TRUE);
                    }
                  }
                }
              }
              else {
              }
            }
          }
        }
      }
      oTRANSLATIONUNITSo.get()->reverse();
      if (!(operation == KWD_TRANSLATE_FILE_DEFINE)) {
        if (oTRANSLATIONVERBOSITYLEVELo.get() >= 1) {
          cout << "*** Pass 2, finalizing objects..." << endl;
        }
        oTRANSLATIONPHASEo.set(KWD_TRANSLATE_FILE_FINALIZE);
        finalize_classes_and_slots();
        if (!(operation == KWD_TRANSLATE_FILE_FINALIZE)) {
          if (oTRANSLATIONVERBOSITYLEVELo.get() >= 1) {
            cout << "*** Pass 3, walking objects..." << endl;
          }
          oTRANSLATIONPHASEo.set(KWD_TRANSLATE_FILE_WALK);
          if (!(use_hardcoded_symbolsP() ||
              (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_JAVA))) {
            clear_symbol_registry();
          }
          { 
            BIND_STELLA_SPECIAL(oUSERECORDEDSIGNATURESpo, boolean, oWALK_WITH_RECORDED_SIGNATURESpo);
            walk_all_phase_one_units();
          }
          if (!(operation == KWD_TRANSLATE_FILE_WALK)) {
            translate_and_output_units_to_file(file, targetlanguage);
          }
          if (oRECORD_SIGNATURESpo) {
            cout << "Recording object signatures..." << endl;
            record_signatures_of_all_units(file);
          }
        }
      }
      summarize_translation_errors();
    }
  }
}

void clean_up_translation_units_special() {
  { Translation_Unit* unit = NULL;
    Cons* iter_000 = oTRANSLATIONUNITSo.get()->the_cons_list;

    while (!(iter_000 == NIL)) {
      unit = ((Translation_Unit*)(iter_000->value));
      iter_000 = iter_000->rest;
      unit->free();
    }
  }
  oTRANSLATIONUNITSo.get()->clear();
  sweep_transients();
}

void translate_file(char* file, Keyword* outputlanguage, boolean aspartofsystemP) {
  { Keyword* mode = KWD_TRANSLATE_FILE_TRANSLATE;

    { 
      BIND_STELLA_SPECIAL(oJAVA_INDENT_CHARSo, int, 0);
      BIND_STELLA_SPECIAL(oIDL_INDENT_CHARSo, int, 0);
      BIND_STELLA_SPECIAL(oCPP_INDENT_CHARSo, int, 0);
      if (outputlanguage == KWD_TRANSLATE_FILE_JAVA) {
        if (aspartofsystemP) {
          mode = KWD_TRANSLATE_FILE_WALK;
        }
      }
      else {
      }
      if (aspartofsystemP) {
        operate_on_file(file, mode);
      }
      else {
        { 
          BIND_STELLA_SPECIAL(oTRANSLATOROUTPUTLANGUAGEo, Keyword*, oTRANSLATOROUTPUTLANGUAGEo.get());
          BIND_STELLA_SPECIAL(oTRANSLATIONUNITSo, List*, new_list());
          set_translator_output_language(outputlanguage);
          operate_on_file(file, mode);
          if (outputlanguage == KWD_TRANSLATE_FILE_CPP_STANDALONE) {
          }
          else {
            clean_up_translation_units_special();
          }
        }
      }
    }
  }
}

void output_all_units_to_file(char* sourcefile) {
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_COMMON_LISP) {
    cl_output_all_units_to_file(sourcefile);
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_JAVA) {
    { 
      BIND_STELLA_SPECIAL(oJAVA_INDENT_CHARSo, int, 0);
      java_output_all_units_to_file();
    }
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_IDL) {
    { 
      BIND_STELLA_SPECIAL(oIDL_INDENT_CHARSo, int, 0);
      idl_output_all_units_to_file(sourcefile);
    }
  }
  else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP) ||
      (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP_STANDALONE)) {
    { 
      BIND_STELLA_SPECIAL(oCPP_INDENT_CHARSo, int, 0);
      cpp_output_all_units_to_file(sourcefile);
    }
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
}

void record_signatures_of_all_units(char* sourcefile) {
}

boolean translate_with_copyright_headerP() {
  return (oCURRENT_ENVIRONMENT_FEATURESo->memberP(KWD_TRANSLATE_FILE_TRANSLATE_WITH_COPYRIGHT_HEADER));
}

void output_copyright_header(Output_Stream* stream) {
  if ((!translate_with_copyright_headerP()) ||
      ((!((boolean)(oCURRENTSYSTEMDEFINITIONo.get()))) ||
       (oCURRENTSYSTEMDEFINITIONo.get()->banner == NULL))) {
    return;
  }
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_COMMON_LISP) {
    *(stream->native_stream) << "#|";
  }
  else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_JAVA) ||
      ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP) ||
       (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP_STANDALONE))) {
    *(stream->native_stream) << "/*";
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
  *(stream->native_stream) << oCURRENTSYSTEMDEFINITIONo.get()->banner;
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_COMMON_LISP) {
    *(stream->native_stream) << "|#";
  }
  else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_JAVA) ||
      ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP) ||
       (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP_STANDALONE))) {
    *(stream->native_stream) << "*/";
  }
  else {
    { Output_String_Stream* stream_001 = new_output_string_stream();

      *(stream_001->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_001->the_string_reader());
    }
  }
  *(stream->native_stream) << endl << endl;
}

void output_file_header(Output_Stream* stream, char* file) {
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_COMMON_LISP) {
    cl_output_file_header(stream, file);
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_JAVA) {
    java_output_file_header(stream, file);
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_TRANSLATE_FILE_CPP) {
    *(stream->native_stream) << "// " << file << endl << endl;
    if (translate_with_copyright_headerP()) {
      output_copyright_header(stream);
    }
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
}

// The operating system we are running on.  Currently,
// either :UNIX, :WINDOWS, or :MAC.
Keyword* oOSo = NULL;

Keyword* operating_system() {
  if (probe_fileP("C:\\AUTOEXEC.BAT") ||
      (probe_fileP("C:\\IO.SYS") ||
       probe_fileP("C:\\BOOT.INI"))) {
    return (KWD_TRANSLATE_FILE_WINDOWS);
  }
  else if (probe_fileP("/usr") ||
      probe_fileP("/etc")) {
    return (KWD_TRANSLATE_FILE_UNIX);
  }
  else {
    return (KWD_TRANSLATE_FILE_MAC);
  }
}

char directory_separator() {
  { char separator = '?';

    if (((boolean)(oOSo))) {
      if (oOSo == KWD_TRANSLATE_FILE_UNIX) {
        separator = '/';
      }
      else if (oOSo == KWD_TRANSLATE_FILE_MAC) {
        separator = ':';
      }
      else if (oOSo == KWD_TRANSLATE_FILE_WINDOWS) {
        separator = '\\';
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << oOSo << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
    }
    return (separator);
  }
}

char* directory_separator_string() {
  return (make_string(1, directory_separator()));
}

char directory_separator_for_file(char* file) {
  if (logical_pathnameP(file)) {
    return (LOGICAL_DIRECTORY_SEPARATOR);
  }
  else {
    return (directory_separator());
  }
}

char TYPE_SEPARATOR = '.';

char* file_name_directory(char* file) {
  // Return the directory component of `file' including the final
  // directory separator or the empty string if `file' does not include a directory.
  { int start = 0;
    char separator = directory_separator_for_file(file);

    while (string_position(file, separator, start) != NULL_INTEGER) {
      start = string_position(file, separator, start) + 1;
    }
    return (string_subsequence(file, 0, start));
  }
}

char* directory_file_name(char* directory) {
  // Return `directory' as a file name, i.e., without a terminating
  // directory separator.
  { char separator = directory_separator_for_file(directory);
    int last = strlen(directory) - 1;

    if ((last >= 0) &&
        (directory[last] == separator)) {
      return (string_subsequence(directory, 0, last));
    }
    else {
      return (directory);
    }
  }
}

char* file_name_as_directory(char* file) {
  // Return `file' interpreted as a directory, i.e., with a
  // terminating directory separator.  If `file' is the empty string simply
  // return the empty string, i.e., interpret it as the current directory instead
  // of the root directory.
  { char separator = directory_separator_for_file(file);
    int last = strlen(file) - 1;

    if ((last < 0) ||
        (file[last] == separator)) {
      return (file);
    }
    else {
      return (string_concatenate(file, make_string(1, separator), 0));
    }
  }
}

char* directory_parent_directory(char* directory, int level) {
  // Return the `level'-th parent directory component of `directory'
  // including the final directory separator, or the empty string if `directory' does
  // not have that many parents.
  { int i = NULL_INTEGER;
    int iter_000 = 1;
    int upper_bound_000 = level;
    boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;

    while (unboundedP_000 ||
        (iter_000 <= upper_bound_000)) {
      i = iter_000;
      iter_000 = iter_000 + 1;
      directory = file_name_directory(directory_file_name(directory));
    }
  }
  return (directory);
}

char* file_name_without_directory(char* file) {
  { int start = 0;
    char separator = directory_separator_for_file(file);

    while (string_position(file, separator, start) != NULL_INTEGER) {
      start = string_position(file, separator, start) + 1;
    }
    return (string_subsequence(file, start, NULL_INTEGER));
  }
}

char* file_name_without_extension(char* file) {
  { int start = 0;
    int end = NULL_INTEGER;

    while (string_position(file, TYPE_SEPARATOR, start) != NULL_INTEGER) {
      end = string_position(file, TYPE_SEPARATOR, start);
      start = end + 1;
    }
    return (string_subsequence(file, 0, end));
  }
}

char* file_extension(char* file) {
  { int start = 0;
    int end = NULL_INTEGER;

    while (string_position(file, TYPE_SEPARATOR, start) != NULL_INTEGER) {
      end = string_position(file, TYPE_SEPARATOR, start);
      start = end + 1;
    }
    if (end == NULL_INTEGER) {
      return (NULL);
    }
    else {
      return (string_subsequence(file, end, NULL_INTEGER));
    }
  }
}

char* file_base_name(char* file) {
  return (file_name_without_extension(file_name_without_directory(file)));
}

char LOGICAL_DIRECTORY_SEPARATOR = ';';

char LOGICAL_HOST_SEPARATOR = ':';

String_Hash_Table* oLOGICAL_HOST_TABLEo = NULL;

void define_logical_host_property(char* host, Keyword* property, Object* value) {
  // Define `property' with `value' for the logical host `host'.
  // As a side-effect, this also defines `host' as a logical host (both
  // `property' and `value' can be supplied as NULL).  If :ROOT-DIRECTORY
  // is specified, all pathnames with `host' are assumed to be relative to
  // that directory (even if they are absolute) and will be rerooted upon
  // translation.  :ROOT-DIRECTORY can be a logical or physical pathname.
  // If :LISP-TRANSLATIONS is specified, those will be used verbatimely as
  // the value of `(CL:logical-pathname-translations host)' if we are
  // running in Lisp, which allows us to depend on the native
  // `CL:translate-logical-pathname' for more complex translation operations.
  host = string_upcase(host);
  { Key_Value_List* properties = ((Key_Value_List*)(oLOGICAL_HOST_TABLEo->lookup(host)));

    if (!((boolean)(properties))) {
      properties = new_key_value_list();
      oLOGICAL_HOST_TABLEo->insert_at(host, properties);
    }
    if (((boolean)(property))) {
      if (property == KWD_TRANSLATE_FILE_ROOT_DIRECTORY) {
        properties->insert_at(property, NULL);
        if (((boolean)(value))) {
          properties->insert_at(property, wrap_string(translate_logical_pathname(file_name_as_directory(((String_Wrapper*)(value))->wrapper_value))));
        }
      }
      else if (property == KWD_TRANSLATE_FILE_LISP_TRANSLATIONS) {
        properties->insert_at(property, value);
      }
      else {
        *(STANDARD_WARNING->native_stream) << "Warning: " << "Illegal host property: " << "`" << property << "'" << endl;
      }
    }
  }
}

Object* lookup_logical_host_property(char* host, Keyword* property) {
  host = string_upcase(host);
  { Key_Value_List* properties = ((Key_Value_List*)(oLOGICAL_HOST_TABLEo->lookup(host)));

    if (!((boolean)(properties))) {
      return (NULL);
    }
    else {
      return (properties->lookup(property));
    }
  }
}

boolean logical_hostP(char* host) {
  return (((boolean)(((Key_Value_List*)(oLOGICAL_HOST_TABLEo->lookup(string_upcase(host)))))));
}

boolean logical_pathnameP(char* pathname) {
  { char* host = get_logical_host_name(pathname);

    return ((host != NULL) &&
        logical_hostP(host));
  }
}

boolean absolute_pathnameP(char* pathname) {
  { char separator = directory_separator();

    if (logical_pathnameP(pathname)) {
      separator = LOGICAL_DIRECTORY_SEPARATOR;
    }
    pathname = logical_pathname_without_host(pathname);
    return ((strlen(pathname) > 0) &&
        (pathname[0] == separator));
  }
}

char* get_logical_host_name(char* pathname) {
  { int separatorpos = string_position(pathname, LOGICAL_HOST_SEPARATOR, 0);

    if ((separatorpos != NULL_INTEGER) &&
        (separatorpos >= 1)) {
      return (string_subsequence(pathname, 0, separatorpos));
    }
    else {
      return (NULL);
    }
  }
}

char* logical_pathname_without_host(char* pathname) {
  { char* hostname = get_logical_host_name(pathname);

    if (hostname == NULL) {
      return (pathname);
    }
    else {
      return (string_subsequence(pathname, strlen(hostname) + 1, NULL_INTEGER));
    }
  }
}

char* translate_logical_pathname(char* pathname) {
  if (!(logical_pathnameP(pathname))) {
    return (pathname);
  }
  { char* host = get_logical_host_name(pathname);
    char* translatedpathname = pathname;

    translatedpathname = string_substitute(logical_pathname_without_host(pathname), directory_separator(), LOGICAL_DIRECTORY_SEPARATOR);
    { String_Wrapper* rootdirectory = ((String_Wrapper*)(lookup_logical_host_property(host, KWD_TRANSLATE_FILE_ROOT_DIRECTORY)));

      if (((boolean)(rootdirectory))) {
        if (absolute_pathnameP(translatedpathname)) {
          translatedpathname = string_subsequence(translatedpathname, 1, NULL_INTEGER);
        }
        translatedpathname = string_concatenate(rootdirectory->wrapper_value, translatedpathname, 0);
      }
    }
    return (translatedpathname);
  }
}

char* unlogicalize_pathname(char* pathname) {
  return (translate_logical_pathname(pathname));
}

// Used by `select-file-extension'.
Cons* oTYPE_TO_FILE_EXTENSION_TABLEo = NULL;

char* compute_common_lisp_file_extension(Keyword* type) {
  if (!(running_as_lispP())) {
    throw *new_stella_exception("Shouldn't call COMPUTE-COMMON-LISP-FILE-EXTENSION unless running in Lisp");
  }
  { char* suffix = NULL;
    char* dummypathname = NULL;

    if (oCURRENT_ENVIRONMENT_FEATURESo->membP(KWD_TRANSLATE_FILE_USE_COMMON_LISP_STRUCTS)) {
      suffix = ".slisp";
    }
    else if (oCURRENT_ENVIRONMENT_FEATURESo->membP(KWD_TRANSLATE_FILE_USE_COMMON_LISP_VECTOR_STRUCTS)) {
      suffix = ".vslisp";
    }
    else {
      suffix = ".lisp";
    }
    if (type == KWD_TRANSLATE_FILE_LISP_BINARY) {
      if (oCURRENT_ENVIRONMENT_FEATURESo->membP(KWD_TRANSLATE_FILE_USE_COMMON_LISP_STRUCTS)) {
        dummypathname = "f.sfasl";
      }
      else if (oCURRENT_ENVIRONMENT_FEATURESo->membP(KWD_TRANSLATE_FILE_USE_COMMON_LISP_VECTOR_STRUCTS)) {
        dummypathname = "f.vsfasl";
      }
      else {
      }
    }
    return (string_subsequence(dummypathname, 1, NULL_INTEGER));
  }
}

char* default_common_lisp_file_extension(Keyword* type) {
  if (type == KWD_TRANSLATE_FILE_LISP) {
    if (oOSo == KWD_TRANSLATE_FILE_WINDOWS) {
      return (".lsp");
    }
    else {
      return (".lisp");
    }
  }
  else if (type == KWD_TRANSLATE_FILE_LISP_BINARY) {
    if (oOSo == KWD_TRANSLATE_FILE_MAC) {
      return (".pfsl");
    }
    else if (oOSo == KWD_TRANSLATE_FILE_WINDOWS) {
      return (".fas");
    }
    else {
      return (".fasl");
    }
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << type << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
}

char* select_file_extension(Keyword* type) {
  { char* extension = NULL;

    { Cons* entry = NULL;
      Cons* iter_000 = oTYPE_TO_FILE_EXTENSION_TABLEo;

      while (!(iter_000 == NIL)) {
        entry = ((Cons*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (entry->value == type) {
          extension = ((String_Wrapper*)(entry->rest->value))->wrapper_value;
        }
      }
    }
    if (type == KWD_TRANSLATE_FILE_DIRECTORY) {
      return (directory_separator_string());
    }
    else if ((type == KWD_TRANSLATE_FILE_LISP) ||
        (type == KWD_TRANSLATE_FILE_LISP_BINARY)) {
      if (running_as_lispP()) {
        return (compute_common_lisp_file_extension(type));
      }
      else {
        return (default_common_lisp_file_extension(type));
      }
    }
    else {
    }
    return (extension);
  }
}

Keyword* classify_file_extension(char* file) {
  { char* extension = file_extension(file);

    if (extension == NULL) {
      return (KWD_TRANSLATE_FILE_NONE);
    }
    { Cons* entry = NULL;
      Cons* iter_000 = oTYPE_TO_FILE_EXTENSION_TABLEo;

      while (!(iter_000 == NIL)) {
        entry = ((Cons*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (eql_to_stringP(entry->rest->value, extension)) {
          return (((Keyword*)(entry->value)));
        }
      }
    }
    if (running_as_lispP()) {
      if (string_eqlP(compute_common_lisp_file_extension(KWD_TRANSLATE_FILE_LISP), extension)) {
        return (KWD_TRANSLATE_FILE_LISP);
      }
      if (string_eqlP(compute_common_lisp_file_extension(KWD_TRANSLATE_FILE_LISP_BINARY), extension)) {
        return (KWD_TRANSLATE_FILE_LISP_BINARY);
      }
    }
    if (string_eqlP(default_common_lisp_file_extension(KWD_TRANSLATE_FILE_LISP), extension)) {
      return (KWD_TRANSLATE_FILE_LISP);
    }
    if (string_eqlP(default_common_lisp_file_extension(KWD_TRANSLATE_FILE_LISP_BINARY), extension)) {
      return (KWD_TRANSLATE_FILE_LISP_BINARY);
    }
    return (KWD_TRANSLATE_FILE_OTHER);
  }
}

char* select_root_directory(Keyword* type) {
  { char* root = NULL;

    if ((type == KWD_TRANSLATE_FILE_STELLA) ||
        (type == KWD_TRANSLATE_FILE_KB)) {
      root = root_source_directory();
    }
    else if ((type == KWD_TRANSLATE_FILE_LISP_BINARY) ||
        (type == KWD_TRANSLATE_FILE_DIRECTORY)) {
      root = root_binary_directory();
    }
    else {
      root = root_native_directory();
    }
    return (root);
  }
}

char* oLISP_SPLITTER_PATHo = "lisp";

char* oJAVA_SPLITTER_PATHo = "java";

char* oCPP_SPLITTER_PATHo = "cpp";

char* select_splitter_path(Keyword* type) {
  { char* splitterpath = "";

    if ((type == KWD_TRANSLATE_FILE_LISP) ||
        (type == KWD_TRANSLATE_FILE_COMMON_LISP)) {
      splitterpath = oLISP_SPLITTER_PATHo;
    }
    else if (type == KWD_TRANSLATE_FILE_JAVA) {
      splitterpath = oJAVA_SPLITTER_PATHo;
    }
    else if ((type == KWD_TRANSLATE_FILE_CPP) ||
        ((type == KWD_TRANSLATE_FILE_CPP_HEADER) ||
         (type == KWD_TRANSLATE_FILE_CPP_CODE))) {
      splitterpath = oCPP_SPLITTER_PATHo;
    }
    else if ((type == KWD_TRANSLATE_FILE_NONE) ||
        (type == KWD_TRANSLATE_FILE_OTHER)) {
      return (select_splitter_path(oTRANSLATOROUTPUTLANGUAGEo.get()));
    }
    else {
    }
    if (!string_eqlP(splitterpath, "")) {
      splitterpath = string_concatenate(splitterpath, directory_separator_string(), 0);
    }
    return (splitterpath);
  }
}

char* make_file_name_from_relative_path(Object* filepath, Keyword* type) {
  return (make_file_name(implode_pathname(filepath), type, TRUE));
}

char* root_directory_prefix(char* file) {
  { char* extension = file_extension(file);

    if (string_eqlP(extension, ".ste")) {
      return (root_source_directory());
    }
    else if (string_eqlP(extension, ".lisp") ||
        (string_eqlP(extension, ".java") ||
         string_eqlP(extension, ".cpp"))) {
      return (root_native_directory());
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << extension << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

char* make_file_name(char* filepath, Keyword* type, boolean relativeP) {
  { char* extension = select_file_extension(type);
    char* rootpath = "";
    char* splitterpath = "";
    char* subdirectory = "";

    if (relativeP) {
      rootpath = select_root_directory(type);
      splitterpath = select_splitter_path(type);
      filepath = file_name_without_extension(filepath);
      if (type == KWD_TRANSLATE_FILE_JAVA) {
        subdirectory = java_package_prefix(oMODULEo.get(), directory_separator_string());
      }
      else {
        if (!string_eqlP(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo.get(), "")) {
          subdirectory = string_concatenate(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo.get(), directory_separator_string(), 0);
        }
      }
    }
    else {
      { char* rootprefix = root_directory_prefix(filepath);

        filepath = file_name_without_extension(filepath);
        if (!get_quoted_tree("((:STELLA :KB) \"/STELLA\")", "/STELLA")->memberP(type)) {
          { char* relativepath = relativize_file_name(filepath, rootprefix);

            if (!string_eqlP(filepath, relativepath)) {
              rootpath = select_root_directory(type);
              splitterpath = select_splitter_path(type);
              filepath = relativepath;
            }
          }
        }
      }
    }
    return (string_concatenate(rootpath, splitterpath, 3, subdirectory, filepath, extension));
  }
}

char* relativize_file_name(char* file, char* rootpath) {
  { boolean absoluteP = string_search(file, rootpath, 0) == 0;

    if (absoluteP) {
      return (string_subsequence(file, strlen(rootpath), NULL_INTEGER));
    }
    else {
      return (file);
    }
  }
}

char* help_implode_pathname(char* filename) {
  { char* extension = file_extension(filename);

    if (extension == NULL) {
      return (maybe_truncate_filename(filename));
    }
    else {
      return (string_concatenate(maybe_truncate_filename(file_base_name(filename)), extension, 0));
    }
  }
}

char* implode_pathname(Object* filepath) {
  { char* file = NULL;

    if (safe_primary_type(filepath) == SGT_TRANSLATE_FILE_STELLA_CONS) {
      { Object* filepath_000 = filepath;
        Cons* filepath = ((Cons*)(filepath_000));

        { int path_length = filepath->length();

          file = "";
          { String_Wrapper* name = NULL;
            int n = NULL_INTEGER;
            Cons* iter_000 = ((Cons*)(filepath));
            int iter_001 = 1;
            int upper_bound_000 = path_length;
            boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;

            while ((!(iter_000 == NIL)) &&
                (unboundedP_000 ||
                 (iter_001 <= upper_bound_000))) {
              name = ((String_Wrapper*)(iter_000->value));
              iter_000 = iter_000->rest;
              n = iter_001;
              iter_001 = iter_001 + 1;
              if (n == path_length) {
                file = string_concatenate(file, help_implode_pathname(name->wrapper_value), 0);
              }
              else {
                file = string_concatenate(file, name->wrapper_value, 1, directory_separator_string());
              }
            }
          }
        }
      }
    }
    else {
      file = help_implode_pathname(((String_Wrapper*)(filepath))->wrapper_value);
    }
    return (file);
  }
}

// When TRUE, overrides normal truncation of file names.
DEFINE_STELLA_SPECIAL(oDONTTRUNCATEFILENAMESpo, boolean , FALSE);

char* maybe_truncate_filename(char* barefilename) {
  if (oDONTTRUNCATEFILENAMESpo.get()) {
    return (barefilename);
  }
  { int maximumlength = 999;

    if (oOSo == KWD_TRANSLATE_FILE_MAC) {
      maximumlength = 31 - 5;
    }
    else {
    }
    if (strlen(barefilename) > maximumlength) {
      { char* originalfilename = barefilename;

        barefilename = string_subsequence(barefilename, 0, maximumlength - 1);
        *(STANDARD_WARNING->native_stream) << "Warning: " << "Truncating name of file." << endl << "   Original: " << "`" << originalfilename << "'" << "    Truncated: " << "`" << barefilename << "'" << endl << endl;
      }
    }
    return (barefilename);
  }
}

Keyword* language_to_file_type(Keyword* language) {
  if (language == KWD_TRANSLATE_FILE_COMMON_LISP) {
    return (KWD_TRANSLATE_FILE_LISP);
  }
  else {
    return (language);
  }
}

char* reroot_file_name(char* file, char* fromroot, char* toroot) {
  { boolean absoluteP = string_search(file, fromroot, 0) == 0;

    if (absoluteP) {
      if (string_eqlP(fromroot, toroot)) {
        return (file);
      }
      else {
        return (string_concatenate(toroot, string_subsequence(file, strlen(fromroot), NULL_INTEGER), 0));
      }
    }
    else {
      return (string_concatenate(toroot, file, 0));
    }
  }
}

Boolean_Wrapper* file_younger_thanP(char* file1, char* file2) {
  if ((!probe_fileP(file1)) ||
      (!probe_fileP(file2))) {
    return (NULL);
  }
  else {
    { Calendar_Date* file1_date = file_write_date(file1);
      Calendar_Date* file2_date = file_write_date(file2);

      if ((!((boolean)(file1_date))) ||
          (!((boolean)(file2_date)))) {
        return (NULL);
      }
      else {
        return ((time_greaterP(file1_date, file2_date) ? TRUE_WRAPPER : FALSE_WRAPPER));
      }
    }
  }
}

boolean stella_need_to_translateP(Object* file, Keyword* language) {
  if (language == KWD_TRANSLATE_FILE_JAVA) {
    return (TRUE);
  }
  { char* sourcefile = make_file_name(((String_Wrapper*)(file))->wrapper_value, KWD_TRANSLATE_FILE_STELLA, TRUE);
    char* translatedfile = make_file_name(((String_Wrapper*)(file))->wrapper_value, language_to_file_type(language), TRUE);

    return (!(file_younger_thanP(translatedfile, sourcefile) == TRUE_WRAPPER));
  }
}

boolean stella_need_to_compileP(Object* file) {
  { char* sourcefile = make_file_name(((String_Wrapper*)(file))->wrapper_value, KWD_TRANSLATE_FILE_LISP, TRUE);
    char* compiledfile = make_file_name(((String_Wrapper*)(file))->wrapper_value, KWD_TRANSLATE_FILE_LISP_BINARY, TRUE);

    return (!(file_younger_thanP(compiledfile, sourcefile) == TRUE_WRAPPER));
  }
}

boolean system_needs_translationP(char* systemname, Keyword* language) {
  { 
    BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONo, System_Definition*, get_system_definition(systemname));
    BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo, char*, (((boolean)(oCURRENTSYSTEMDEFINITIONo.get())) ? oCURRENTSYSTEMDEFINITIONo.get()->directory : NULL));
    if (((boolean)(oCURRENTSYSTEMDEFINITIONo.get()))) {
      {
        { boolean foundP_000 = FALSE;

          { String_Wrapper* file = NULL;
            Cons* iter_000 = files_plus_system_startup(system_definition_source_files(oCURRENTSYSTEMDEFINITIONo.get()));

            while (!(iter_000 == NIL)) {
              file = ((String_Wrapper*)(iter_000->value));
              iter_000 = iter_000->rest;
              if (stella_need_to_translateP(file, language)) {
                foundP_000 = TRUE;
                break;
              }
            }
          }
          { boolean value_000 = foundP_000;

            return (value_000);
          }
        }
      }
    }
    else {
      *(STANDARD_WARNING->native_stream) << "Warning: " << "Can't find a system named " << "`" << systemname << "'" << endl << endl;
    }
  }
  return (FALSE);
}

Module* compute_file_units_module(List* fileunits) {
  { Translation_Unit* unit = NULL;
    Cons* iter_000 = fileunits->the_cons_list;

    while (!(iter_000 == NIL)) {
      unit = ((Translation_Unit*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (((boolean)(unit->the_object))) {
        { Symbol* test_value_000 = unit->category;

          if ((test_value_000 == SYM_TRANSLATE_FILE_STELLA_STARTUP_TIME_PROGN) ||
              (test_value_000 == SYM_TRANSLATE_FILE_STELLA_VERBATIM)) {
          }
          else {
            return (unit->the_object->home_module());
          }
        }
      }
    }
  }
  return (oMODULEo.get());
}

List* concatenate_system_units(Key_Value_List* systemunits) {
  { Cons* head = cons(NULL, NIL);
    Cons* cursor = head;

    { String_Wrapper* file = NULL;
      List* units = NULL;
      Kv_Cons* iter_000 = systemunits->the_kv_list;

      while (((boolean)(iter_000))) {
        file = ((String_Wrapper*)(iter_000->key));
        units = ((List*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (units->non_emptyP()) {
          cursor->concatenate(units->the_cons_list, 0);
          cursor = units->the_cons_list;
        }
      }
    }
    { List* self_000 = new_list();

      self_000->the_cons_list = head->rest;
      { List* value_000 = self_000;

        return (value_000);
      }
    }
  }
}

boolean translate_system(char* systemname, Keyword* outputlanguage, int options, ...) {
  // Translate all of the STELLA source files in `system' into a target language.
  // If :two-pass? is true, all files will be scanned twice, once to define
  //    all objects defined in them, and once to actually translate them.
  // If :force-translation?, files will be translated whether or not they are up-to-date.
  // If :production-settings?, optimization levels will favor efficient code.
  // Return TRUE if at least one file was translated.
  { Property_List* self_000 = new_property_list();

    { Cons* arglist_000 = NIL;

      { Object* arg_000 = NULL;
        va_list iter_000;
        int iter_000_count = options;
        Cons* collect_000 = NULL;

        va_start(iter_000, options);
        while ((iter_000_count--) > 0) {
          arg_000 = va_arg(iter_000, Object*);
          if (!((boolean)(collect_000))) {
            {
              collect_000 = cons(arg_000, NIL);
              if (arglist_000 == NIL) {
                arglist_000 = collect_000;
              }
              else {
                add_cons_to_end_of_cons_list(arglist_000, collect_000);
              }
            }
          }
          else {
            {
              collect_000->rest = cons(arg_000, NIL);
              collect_000 = collect_000->rest;
            }
          }
        }
        va_end(iter_000);
      }
      self_000->the_plist = arglist_000;
    }
    { Property_List* plist = self_000;
      boolean twopassP = FALSE;
      boolean forcetranslationP = FALSE;
      boolean productionmodeP = FALSE;

      outputlanguage = ((Keyword*)(stellify(outputlanguage)));
      vet_options(plist, get_quoted_tree("((:TWO-PASS? :FORCE-TRANSLATION? :PRODUCTION-SETTINGS? :DEVELOPMENT-SETTINGS?) \"/STELLA\")", "/STELLA"));
      { Object* key = NULL;
        Object* value = NULL;
        Cons* iter_001 = plist->the_plist;

        while (!(iter_001 == NIL)) {
          key = iter_001->value;
          value = iter_001->rest->value;
          iter_001 = iter_001->rest->rest;
          { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(key));

            if (test_value_000 == KWD_TRANSLATE_FILE_TWO_PASSp) {
              twopassP = eqlP(value, TRUE_WRAPPER);
            }
            else if (test_value_000 == KWD_TRANSLATE_FILE_FORCE_TRANSLATIONp) {
              forcetranslationP = eqlP(value, TRUE_WRAPPER);
            }
            else if (test_value_000 == KWD_TRANSLATE_FILE_PRODUCTION_SETTINGSp) {
              productionmodeP = eqlP(value, TRUE_WRAPPER);
            }
            else if (test_value_000 == KWD_TRANSLATE_FILE_DEVELOPMENT_SETTINGSp) {
              productionmodeP = !eqlP(value, TRUE_WRAPPER);
            }
            else {
              { Output_String_Stream* stream_000 = new_output_string_stream();

                *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
                throw *new_stella_exception(stream_000->the_string_reader());
              }
            }
          }
        }
      }
      if (!((boolean)(lookup_function(system_startup_function_symbol(get_system_definition(systemname)))))) {
        twopassP = TRUE;
      }
      { 
        BIND_STELLA_SPECIAL(oTRANSLATOROUTPUTLANGUAGEo, Keyword*, oTRANSLATOROUTPUTLANGUAGEo.get());
        BIND_STELLA_SPECIAL(oTRANSLATIONPHASEo, Keyword*, NULL);
        BIND_STELLA_SPECIAL(oSAFETYo, int, oSAFETYo.get());
        BIND_STELLA_SPECIAL(oRECYCLING_ENABLEDpo, boolean, !(outputlanguage == KWD_TRANSLATE_FILE_JAVA));
        oGENSYM_COUNTERo = -1;
        enable_environment_feature(3, KWD_TRANSLATE_FILE_TRANSLATE_WITH_COPYRIGHT_HEADER, KWD_TRANSLATE_FILE_WARN_ABOUT_MISSING_METHODS, KWD_TRANSLATE_FILE_WARN_ABOUT_UNDEFINED_METHODS);
        disable_environment_feature(1, KWD_TRANSLATE_FILE_USE_HARDCODED_SYMBOLS);
        set_translator_output_language(outputlanguage);
        { 
          BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONo, System_Definition*, get_system_definition(systemname));
          BIND_STELLA_SPECIAL(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo, char*, (((boolean)(oCURRENTSYSTEMDEFINITIONo.get())) ? oCURRENTSYSTEMDEFINITIONo.get()->directory : NULL));
          if (((boolean)(oCURRENTSYSTEMDEFINITIONo.get()))) {
            {
              compute_optimization_levels(oCURRENTSYSTEMDEFINITIONo.get(), productionmodeP);
              { Cons* files = files_plus_system_startup(system_definition_source_files(oCURRENTSYSTEMDEFINITIONo.get()));
                boolean translatedsomethingP = FALSE;

                if (forcetranslationP ||
                    system_needs_translationP(systemname, outputlanguage)) {
                  create_system_startup_file(oCURRENTSYSTEMDEFINITIONo.get());
                  translatedsomethingP = help_translate_system(files, twopassP, forcetranslationP);
                }
                transfer_native_system_files(outputlanguage);
                disable_environment_feature(1, KWD_TRANSLATE_FILE_TRANSLATE_WITH_COPYRIGHT_HEADER);
                return (translatedsomethingP);
              }
            }
          }
          else {
            *(STANDARD_WARNING->native_stream) << "Warning: " << "Can't find a system named " << "`" << systemname << "'" << endl << endl;
          }
        }
        return (FALSE);
      }
    }
  }
}

boolean help_translate_system(Cons* files, boolean twopassP, boolean forcetranslationP) {
  { 
    BIND_STELLA_SPECIAL(oTRANSLATIONUNITSo, List*, new_list());
    { Keyword* outputlanguage = oTRANSLATOROUTPUTLANGUAGEo.get();
      char* filename = NULL;
      boolean translatedfileP = FALSE;
      Key_Value_List* walkedunits = new_key_value_list();

      clear_symbol_registry();
      if (twopassP) {
        { String_Wrapper* file = NULL;
          Cons* iter_000 = files;

          while (!(iter_000 == NIL)) {
            file = ((String_Wrapper*)(iter_000->value));
            iter_000 = iter_000->rest;
            if (forcetranslationP ||
                (stella_need_to_translateP(file, outputlanguage) ||
                 (!oCURRENTSYSTEMDEFINITIONo.get()->loadedP))) {
              filename = make_file_name(file->wrapper_value, KWD_TRANSLATE_FILE_STELLA, TRUE);
              if (!(probe_fileP(filename))) {
                cout << "Warning: File " << "`" << filename << "'" << " does not exist." << endl;
                continue;
              }
              operate_on_file(filename, KWD_TRANSLATE_FILE_DEFINE);
              clean_up_translation_units_special();
              translatedfileP = TRUE;
            }
          }
        }
        if (translatedfileP) {
          if (oTRANSLATIONVERBOSITYLEVELo.get() >= 1) {
            cout << "*** Pass 2, finalizing objects..." << endl;
          }
          finalize_classes_and_slots();
        }
      }
      { String_Wrapper* file = NULL;
        Cons* iter_001 = files;

        while (!(iter_001 == NIL)) {
          file = ((String_Wrapper*)(iter_001->value));
          iter_001 = iter_001->rest;
          if (forcetranslationP ||
              stella_need_to_translateP(file, outputlanguage)) {
            filename = make_file_name(file->wrapper_value, KWD_TRANSLATE_FILE_STELLA, TRUE);
            if (!(probe_fileP(filename))) {
              cout << "Warning: File " << "`" << filename << "'" << " does not exist." << endl;
              continue;
            }
            operate_on_file(filename, KWD_TRANSLATE_FILE_WALK);
            walkedunits->insert_at(wrap_string(filename), oTRANSLATIONUNITSo.get());
            oTRANSLATIONUNITSo.set(new_list());
          }
        }
      }
      walkedunits->reverse();
      if (oTRANSLATIONVERBOSITYLEVELo.get() >= 1) {
        cout << "    ==============================" << endl;
      }
      if (outputlanguage == KWD_TRANSLATE_FILE_COMMON_LISP) {
        cl_translate_walked_system_units(walkedunits);
      }
      else if (outputlanguage == KWD_TRANSLATE_FILE_CPP) {
        cpp_translate_walked_system_units(walkedunits);
      }
      else if (outputlanguage == KWD_TRANSLATE_FILE_JAVA) {
        java_translate_walked_system_units(walkedunits);
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << outputlanguage << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
      clear_symbol_registry();
      return (translatedfileP);
    }
  }
}

void transfer_native_system_files(Keyword* outputlanguage) {
  { Cons* nativefiles = NULL;
    char* systemsubdirectory = (string_eqlP(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo.get(), "") ? (char*)"" : string_concatenate(oCURRENTSYSTEMDEFINITIONSUBDIRECTORYo.get(), directory_separator_string(), 0));
    char* flotsamfilename = "";

    if (outputlanguage == KWD_TRANSLATE_FILE_COMMON_LISP) {
      nativefiles = oCURRENTSYSTEMDEFINITIONo.get()->lisp_only_files;
    }
    else if (outputlanguage == KWD_TRANSLATE_FILE_JAVA) {
      nativefiles = oCURRENTSYSTEMDEFINITIONo.get()->java_only_files;
      flotsamfilename = java_yield_flotsam_class_name(get_cardinal_module(oCURRENTSYSTEMDEFINITIONo.get()));
    }
    else if (outputlanguage == KWD_TRANSLATE_FILE_CPP) {
      nativefiles = oCURRENTSYSTEMDEFINITIONo.get()->cpp_only_files;
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << outputlanguage << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
    if (!((boolean)(nativefiles))) {
      return;
    }
    if (string_eqlP(root_source_directory(), root_native_directory())) {
      return;
    }
    nativefiles = nativefiles->subtract(system_definition_source_files(oCURRENTSYSTEMDEFINITIONo.get()));
    { 
      BIND_STELLA_SPECIAL(oMODULEo, Module*, get_cardinal_module(oCURRENTSYSTEMDEFINITIONo.get()));
      BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
      { String_Wrapper* f = NULL;
        Cons* iter_000 = nativefiles;

        while (!(iter_000 == NIL)) {
          f = ((String_Wrapper*)(iter_000->value));
          iter_000 = iter_000->rest;
          { char* filename = f->wrapper_value;
            Keyword* type = classify_file_extension(filename);
            char* relativefilename = relativize_file_name(filename, root_source_directory());
            char* fromfilename = NULL;
            char* tofilename = NULL;

            if (type == KWD_TRANSLATE_FILE_JAVA) {
              if (string_eqlP(file_base_name(filename), flotsamfilename)) {
                *(STANDARD_WARNING->native_stream) << "Warning: " << "Native Java filename " << "`" << flotsamfilename << "'" << endl << " conflicts with the Java catchall class' filename" << endl;
              }
            }
            else {
            }
            fromfilename = string_concatenate(root_source_directory(), systemsubdirectory, 1, filename);
            tofilename = make_file_name(relativefilename, type, TRUE);
            if (!(file_younger_thanP(tofilename, fromfilename) == TRUE_WRAPPER)) {
              if (oTRANSLATIONVERBOSITYLEVELo.get() >= 1) {
                cout << "Copying " << "`" << fromfilename << "'" << endl << " to " << "`" << tofilename << "'" << " ..." << endl;
              }
              copy_file(fromfilename, tofilename);
            }
          }
        }
      }
    }
  }
}

void help_startup_translate_file1() {
  {
    KWD_TRANSLATE_FILE_TRANSLATE = ((Keyword*)(intern_rigid_symbol_wrt_module("TRANSLATE", NULL, 2)));
    KWD_TRANSLATE_FILE_JAVA = ((Keyword*)(intern_rigid_symbol_wrt_module("JAVA", NULL, 2)));
    KWD_TRANSLATE_FILE_COMMON_LISP = ((Keyword*)(intern_rigid_symbol_wrt_module("COMMON-LISP", NULL, 2)));
    KWD_TRANSLATE_FILE_CPP = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP", NULL, 2)));
    KWD_TRANSLATE_FILE_CPP_STANDALONE = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP-STANDALONE", NULL, 2)));
    KWD_TRANSLATE_FILE_IDL = ((Keyword*)(intern_rigid_symbol_wrt_module("IDL", NULL, 2)));
    SGT_TRANSLATE_FILE_STELLA_CONS = ((Surrogate*)(intern_rigid_symbol_wrt_module("CONS", NULL, 1)));
    KWD_TRANSLATE_FILE_DEFINE = ((Keyword*)(intern_rigid_symbol_wrt_module("DEFINE", NULL, 2)));
    KWD_TRANSLATE_FILE_FINALIZE = ((Keyword*)(intern_rigid_symbol_wrt_module("FINALIZE", NULL, 2)));
    KWD_TRANSLATE_FILE_WALK = ((Keyword*)(intern_rigid_symbol_wrt_module("WALK", NULL, 2)));
    KWD_TRANSLATE_FILE_TRANSLATE_WITH_COPYRIGHT_HEADER = ((Keyword*)(intern_rigid_symbol_wrt_module("TRANSLATE-WITH-COPYRIGHT-HEADER", NULL, 2)));
    KWD_TRANSLATE_FILE_WINDOWS = ((Keyword*)(intern_rigid_symbol_wrt_module("WINDOWS", NULL, 2)));
    KWD_TRANSLATE_FILE_UNIX = ((Keyword*)(intern_rigid_symbol_wrt_module("UNIX", NULL, 2)));
    KWD_TRANSLATE_FILE_MAC = ((Keyword*)(intern_rigid_symbol_wrt_module("MAC", NULL, 2)));
    KWD_TRANSLATE_FILE_ROOT_DIRECTORY = ((Keyword*)(intern_rigid_symbol_wrt_module("ROOT-DIRECTORY", NULL, 2)));
    KWD_TRANSLATE_FILE_LISP_TRANSLATIONS = ((Keyword*)(intern_rigid_symbol_wrt_module("LISP-TRANSLATIONS", NULL, 2)));
    KWD_TRANSLATE_FILE_STELLA = ((Keyword*)(intern_rigid_symbol_wrt_module("STELLA", NULL, 2)));
    KWD_TRANSLATE_FILE_KB = ((Keyword*)(intern_rigid_symbol_wrt_module("KB", NULL, 2)));
    KWD_TRANSLATE_FILE_CPP_CODE = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP-CODE", NULL, 2)));
    KWD_TRANSLATE_FILE_CPP_HEADER = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP-HEADER", NULL, 2)));
    KWD_TRANSLATE_FILE_NONE = ((Keyword*)(intern_rigid_symbol_wrt_module("NONE", NULL, 2)));
    KWD_TRANSLATE_FILE_USE_COMMON_LISP_STRUCTS = ((Keyword*)(intern_rigid_symbol_wrt_module("USE-COMMON-LISP-STRUCTS", NULL, 2)));
    KWD_TRANSLATE_FILE_USE_COMMON_LISP_VECTOR_STRUCTS = ((Keyword*)(intern_rigid_symbol_wrt_module("USE-COMMON-LISP-VECTOR-STRUCTS", NULL, 2)));
    KWD_TRANSLATE_FILE_LISP_BINARY = ((Keyword*)(intern_rigid_symbol_wrt_module("LISP-BINARY", NULL, 2)));
    KWD_TRANSLATE_FILE_LISP = ((Keyword*)(intern_rigid_symbol_wrt_module("LISP", NULL, 2)));
    KWD_TRANSLATE_FILE_DIRECTORY = ((Keyword*)(intern_rigid_symbol_wrt_module("DIRECTORY", NULL, 2)));
    KWD_TRANSLATE_FILE_OTHER = ((Keyword*)(intern_rigid_symbol_wrt_module("OTHER", NULL, 2)));
    SYM_TRANSLATE_FILE_STELLA_SYSTEM_UNITS_ALIST = ((Symbol*)(intern_rigid_symbol_wrt_module("SYSTEM-UNITS-ALIST", NULL, 0)));
    SYM_TRANSLATE_FILE_STELLA_STARTUP_TIME_PROGN = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-TIME-PROGN", NULL, 0)));
    SYM_TRANSLATE_FILE_STELLA_VERBATIM = ((Symbol*)(intern_rigid_symbol_wrt_module("VERBATIM", NULL, 0)));
    KWD_TRANSLATE_FILE_TWO_PASSp = ((Keyword*)(intern_rigid_symbol_wrt_module("TWO-PASS?", NULL, 2)));
    KWD_TRANSLATE_FILE_FORCE_TRANSLATIONp = ((Keyword*)(intern_rigid_symbol_wrt_module("FORCE-TRANSLATION?", NULL, 2)));
    KWD_TRANSLATE_FILE_PRODUCTION_SETTINGSp = ((Keyword*)(intern_rigid_symbol_wrt_module("PRODUCTION-SETTINGS?", NULL, 2)));
    KWD_TRANSLATE_FILE_DEVELOPMENT_SETTINGSp = ((Keyword*)(intern_rigid_symbol_wrt_module("DEVELOPMENT-SETTINGS?", NULL, 2)));
    KWD_TRANSLATE_FILE_WARN_ABOUT_MISSING_METHODS = ((Keyword*)(intern_rigid_symbol_wrt_module("WARN-ABOUT-MISSING-METHODS", NULL, 2)));
    KWD_TRANSLATE_FILE_WARN_ABOUT_UNDEFINED_METHODS = ((Keyword*)(intern_rigid_symbol_wrt_module("WARN-ABOUT-UNDEFINED-METHODS", NULL, 2)));
    KWD_TRANSLATE_FILE_USE_HARDCODED_SYMBOLS = ((Keyword*)(intern_rigid_symbol_wrt_module("USE-HARDCODED-SYMBOLS", NULL, 2)));
    SYM_TRANSLATE_FILE_STELLA_STARTUP_TRANSLATE_FILE = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-TRANSLATE-FILE", NULL, 0)));
    SYM_TRANSLATE_FILE_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-STARTUP-CLASSNAME", NULL, 0)));
  }
}

void help_startup_translate_file2() {
  {
    define_function_object("ROOT-SOURCE-DIRECTORY", "(DEFUN (ROOT-SOURCE-DIRECTORY FILE-NAME) ())", ((cpp_function_code)(&root_source_directory)), NULL);
    define_function_object("ROOT-NATIVE-DIRECTORY", "(DEFUN (ROOT-NATIVE-DIRECTORY FILE-NAME) ())", ((cpp_function_code)(&root_native_directory)), NULL);
    define_function_object("ROOT-BINARY-DIRECTORY", "(DEFUN (ROOT-BINARY-DIRECTORY FILE-NAME) ())", ((cpp_function_code)(&root_binary_directory)), NULL);
    define_function_object("SYSTEM-DEFINITIONS-DIRECTORY", "(DEFUN (SYSTEM-DEFINITIONS-DIRECTORY FILE-NAME) ())", ((cpp_function_code)(&system_definitions_directory)), NULL);
    define_function_object("HANDLE-IN-MODULE-TREE", "(DEFUN (HANDLE-IN-MODULE-TREE BOOLEAN BOOLEAN) ((TREE CONS) (SEENINMODULE? BOOLEAN)) :PUBLIC? TRUE)", ((cpp_function_code)(&handle_in_module_tree)), NULL);
    define_function_object("TRANSLATE-AND-OUTPUT-UNITS-TO-FILE", "(DEFUN TRANSLATE-AND-OUTPUT-UNITS-TO-FILE ((FILE FILE-NAME) (TARGETLANGUAGE STRING)))", ((cpp_function_code)(&translate_and_output_units_to_file)), NULL);
    define_function_object("OPERATE-ON-FILE", "(DEFUN OPERATE-ON-FILE ((FILE FILE-NAME) (OPERATION KEYWORD)))", ((cpp_function_code)(&operate_on_file)), NULL);
    define_function_object("CLEAN-UP-TRANSLATION-UNITS-SPECIAL", "(DEFUN CLEAN-UP-TRANSLATION-UNITS-SPECIAL ())", ((cpp_function_code)(&clean_up_translation_units_special)), NULL);
    define_function_object("TRANSLATE-FILE", "(DEFUN TRANSLATE-FILE ((FILE FILE-NAME) (OUTPUTLANGUAGE KEYWORD) (ASPARTOFSYSTEM? BOOLEAN)))", ((cpp_function_code)(&translate_file)), NULL);
    define_function_object("OUTPUT-ALL-UNITS-TO-FILE", "(DEFUN OUTPUT-ALL-UNITS-TO-FILE ((SOURCEFILE FILE-NAME)))", ((cpp_function_code)(&output_all_units_to_file)), NULL);
    define_function_object("RECORD-SIGNATURES-OF-ALL-UNITS", "(DEFUN RECORD-SIGNATURES-OF-ALL-UNITS ((SOURCEFILE FILE-NAME)))", ((cpp_function_code)(&record_signatures_of_all_units)), NULL);
    define_function_object("TRANSLATE-WITH-COPYRIGHT-HEADER?", "(DEFUN (TRANSLATE-WITH-COPYRIGHT-HEADER? BOOLEAN) ())", ((cpp_function_code)(&translate_with_copyright_headerP)), NULL);
    define_function_object("OUTPUT-COPYRIGHT-HEADER", "(DEFUN OUTPUT-COPYRIGHT-HEADER ((STREAM OUTPUT-STREAM)))", ((cpp_function_code)(&output_copyright_header)), NULL);
    define_function_object("OUTPUT-FILE-HEADER", "(DEFUN OUTPUT-FILE-HEADER ((STREAM OUTPUT-STREAM) (FILE FILE-NAME)))", ((cpp_function_code)(&output_file_header)), NULL);
    define_function_object("OPERATING-SYSTEM", "(DEFUN (OPERATING-SYSTEM KEYWORD) ())", ((cpp_function_code)(&operating_system)), NULL);
    define_function_object("DIRECTORY-SEPARATOR", "(DEFUN (DIRECTORY-SEPARATOR CHARACTER) ())", ((cpp_function_code)(&directory_separator)), NULL);
    define_function_object("DIRECTORY-SEPARATOR-STRING", "(DEFUN (DIRECTORY-SEPARATOR-STRING STRING) ())", ((cpp_function_code)(&directory_separator_string)), NULL);
    define_function_object("DIRECTORY-SEPARATOR-FOR-FILE", "(DEFUN (DIRECTORY-SEPARATOR-FOR-FILE CHARACTER) ((FILE FILE-NAME)))", ((cpp_function_code)(&directory_separator_for_file)), NULL);
    define_function_object("FILE-NAME-DIRECTORY", "(DEFUN (FILE-NAME-DIRECTORY FILE-NAME) ((FILE FILE-NAME)) :DOCUMENTATION \"Return the directory component of `file' including the final\ndirectory separator or the empty string if `file' does not include a directory.\" :PUBLIC? TRUE)", ((cpp_function_code)(&file_name_directory)), NULL);
    define_function_object("DIRECTORY-FILE-NAME", "(DEFUN (DIRECTORY-FILE-NAME FILE-NAME) ((DIRECTORY FILE-NAME)) :DOCUMENTATION \"Return `directory' as a file name, i.e., without a terminating\ndirectory separator.\" :PUBLIC? TRUE)", ((cpp_function_code)(&directory_file_name)), NULL);
    define_function_object("FILE-NAME-AS-DIRECTORY", "(DEFUN (FILE-NAME-AS-DIRECTORY FILE-NAME) ((FILE FILE-NAME)) :DOCUMENTATION \"Return `file' interpreted as a directory, i.e., with a\nterminating directory separator.  If `file' is the empty string simply\nreturn the empty string, i.e., interpret it as the current directory instead\nof the root directory.\" :PUBLIC? TRUE)", ((cpp_function_code)(&file_name_as_directory)), NULL);
    define_function_object("DIRECTORY-PARENT-DIRECTORY", "(DEFUN (DIRECTORY-PARENT-DIRECTORY FILE-NAME) ((DIRECTORY FILE-NAME) (LEVEL INTEGER)) :DOCUMENTATION \"Return the `level'-th parent directory component of `directory'\nincluding the final directory separator, or the empty string if `directory' does\nnot have that many parents.\" :PUBLIC? TRUE)", ((cpp_function_code)(&directory_parent_directory)), NULL);
    define_function_object("FILE-NAME-WITHOUT-DIRECTORY", "(DEFUN (FILE-NAME-WITHOUT-DIRECTORY FILE-NAME) ((FILE FILE-NAME)) :PUBLIC? TRUE)", ((cpp_function_code)(&file_name_without_directory)), NULL);
    define_function_object("FILE-NAME-WITHOUT-EXTENSION", "(DEFUN (FILE-NAME-WITHOUT-EXTENSION FILE-NAME) ((FILE FILE-NAME)) :PUBLIC? TRUE)", ((cpp_function_code)(&file_name_without_extension)), NULL);
    define_function_object("FILE-EXTENSION", "(DEFUN (FILE-EXTENSION STRING) ((FILE FILE-NAME)))", ((cpp_function_code)(&file_extension)), NULL);
    define_function_object("FILE-BASE-NAME", "(DEFUN (FILE-BASE-NAME FILE-NAME) ((FILE FILE-NAME)) :PUBLIC? TRUE)", ((cpp_function_code)(&file_base_name)), NULL);
    define_function_object("DEFINE-LOGICAL-HOST-PROPERTY", "(DEFUN DEFINE-LOGICAL-HOST-PROPERTY ((HOST STRING) (PROPERTY KEYWORD) (VALUE OBJECT)) :DOCUMENTATION \"Define `property' with `value' for the logical host `host'.\nAs a side-effect, this also defines `host' as a logical host (both\n`property' and `value' can be supplied as NULL).  If :ROOT-DIRECTORY\nis specified, all pathnames with `host' are assumed to be relative to\nthat directory (even if they are absolute) and will be rerooted upon\ntranslation.  :ROOT-DIRECTORY can be a logical or physical pathname.\nIf :LISP-TRANSLATIONS is specified, those will be used verbatimely as\nthe value of `(CL:logical-pathname-translations host)' if we are\nrunning in Lisp, which allows us to depend on the native\n`CL:translate-logical-pathname' for more complex translation operations.\" :PUBLIC? TRUE)", ((cpp_function_code)(&define_logical_host_property)), NULL);
    define_function_object("LOOKUP-LOGICAL-HOST-PROPERTY", "(DEFUN (LOOKUP-LOGICAL-HOST-PROPERTY OBJECT) ((HOST STRING) (PROPERTY KEYWORD)))", ((cpp_function_code)(&lookup_logical_host_property)), NULL);
    define_function_object("LOGICAL-HOST?", "(DEFUN (LOGICAL-HOST? BOOLEAN) ((HOST STRING)))", ((cpp_function_code)(&logical_hostP)), NULL);
    define_function_object("LOGICAL-PATHNAME?", "(DEFUN (LOGICAL-PATHNAME? BOOLEAN) ((PATHNAME STRING)))", ((cpp_function_code)(&logical_pathnameP)), NULL);
    define_function_object("ABSOLUTE-PATHNAME?", "(DEFUN (ABSOLUTE-PATHNAME? BOOLEAN) ((PATHNAME STRING)))", ((cpp_function_code)(&absolute_pathnameP)), NULL);
    define_function_object("GET-LOGICAL-HOST-NAME", "(DEFUN (GET-LOGICAL-HOST-NAME STRING) ((PATHNAME STRING)))", ((cpp_function_code)(&get_logical_host_name)), NULL);
    define_function_object("LOGICAL-PATHNAME-WITHOUT-HOST", "(DEFUN (LOGICAL-PATHNAME-WITHOUT-HOST STRING) ((PATHNAME STRING)))", ((cpp_function_code)(&logical_pathname_without_host)), NULL);
    define_function_object("TRANSLATE-LOGICAL-PATHNAME", "(DEFUN (TRANSLATE-LOGICAL-PATHNAME STRING) ((PATHNAME STRING)))", ((cpp_function_code)(&translate_logical_pathname)), NULL);
    define_function_object("UNLOGICALIZE-PATHNAME", "(DEFUN (UNLOGICALIZE-PATHNAME STRING) ((PATHNAME STRING)) :GLOBALLY-INLINE? TRUE (RETURN (TRANSLATE-LOGICAL-PATHNAME PATHNAME)))", ((cpp_function_code)(&unlogicalize_pathname)), NULL);
    define_function_object("COMPUTE-COMMON-LISP-FILE-EXTENSION", "(DEFUN (COMPUTE-COMMON-LISP-FILE-EXTENSION STRING) ((TYPE KEYWORD)))", ((cpp_function_code)(&compute_common_lisp_file_extension)), NULL);
    define_function_object("DEFAULT-COMMON-LISP-FILE-EXTENSION", "(DEFUN (DEFAULT-COMMON-LISP-FILE-EXTENSION STRING) ((TYPE KEYWORD)))", ((cpp_function_code)(&default_common_lisp_file_extension)), NULL);
    define_function_object("SELECT-FILE-EXTENSION", "(DEFUN (SELECT-FILE-EXTENSION STRING) ((TYPE KEYWORD)))", ((cpp_function_code)(&select_file_extension)), NULL);
    define_function_object("CLASSIFY-FILE-EXTENSION", "(DEFUN (CLASSIFY-FILE-EXTENSION KEYWORD) ((FILE FILE-NAME)))", ((cpp_function_code)(&classify_file_extension)), NULL);
    define_function_object("SELECT-ROOT-DIRECTORY", "(DEFUN (SELECT-ROOT-DIRECTORY STRING) ((TYPE KEYWORD)))", ((cpp_function_code)(&select_root_directory)), NULL);
    define_function_object("SELECT-SPLITTER-PATH", "(DEFUN (SELECT-SPLITTER-PATH STRING) ((TYPE KEYWORD)))", ((cpp_function_code)(&select_splitter_path)), NULL);
    define_function_object("MAKE-FILE-NAME-FROM-RELATIVE-PATH", "(DEFUN (MAKE-FILE-NAME-FROM-RELATIVE-PATH FILE-NAME) ((FILEPATH OBJECT) (TYPE KEYWORD)))", ((cpp_function_code)(&make_file_name_from_relative_path)), NULL);
    define_function_object("ROOT-DIRECTORY-PREFIX", "(DEFUN (ROOT-DIRECTORY-PREFIX FILE-NAME) ((FILE FILE-NAME)))", ((cpp_function_code)(&root_directory_prefix)), NULL);
    define_function_object("MAKE-FILE-NAME", "(DEFUN (MAKE-FILE-NAME FILE-NAME) ((FILEPATH STRING) (TYPE KEYWORD) (RELATIVE? BOOLEAN)))", ((cpp_function_code)(&make_file_name)), NULL);
    define_function_object("RELATIVIZE-FILE-NAME", "(DEFUN (RELATIVIZE-FILE-NAME FILE-NAME) ((FILE FILE-NAME) (ROOTPATH FILE-NAME)))", ((cpp_function_code)(&relativize_file_name)), NULL);
    define_function_object("HELP-IMPLODE-PATHNAME", "(DEFUN (HELP-IMPLODE-PATHNAME FILE-NAME) ((FILENAME FILE-NAME)))", ((cpp_function_code)(&help_implode_pathname)), NULL);
    define_function_object("IMPLODE-PATHNAME", "(DEFUN (IMPLODE-PATHNAME FILE-NAME) ((FILEPATH OBJECT)))", ((cpp_function_code)(&implode_pathname)), NULL);
    define_function_object("MAYBE-TRUNCATE-FILENAME", "(DEFUN (MAYBE-TRUNCATE-FILENAME FILE-NAME) ((BAREFILENAME FILE-NAME)))", ((cpp_function_code)(&maybe_truncate_filename)), NULL);
    define_function_object("LANGUAGE-TO-FILE-TYPE", "(DEFUN (LANGUAGE-TO-FILE-TYPE KEYWORD) ((LANGUAGE KEYWORD)))", ((cpp_function_code)(&language_to_file_type)), NULL);
    define_function_object("REROOT-FILE-NAME", "(DEFUN (REROOT-FILE-NAME FILE-NAME) ((FILE FILE-NAME) (FROMROOT FILE-NAME) (TOROOT FILE-NAME)))", ((cpp_function_code)(&reroot_file_name)), NULL);
    define_function_object("FILE-YOUNGER-THAN?", "(DEFUN (FILE-YOUNGER-THAN? THREE-VALUED-BOOLEAN) ((FILE1 STRING) (FILE2 STRING)))", ((cpp_function_code)(&file_younger_thanP)), NULL);
    define_function_object("STELLA-NEED-TO-TRANSLATE?", "(DEFUN (STELLA-NEED-TO-TRANSLATE? BOOLEAN) ((FILE OBJECT) (LANGUAGE KEYWORD)))", ((cpp_function_code)(&stella_need_to_translateP)), NULL);
    define_function_object("STELLA-NEED-TO-COMPILE?", "(DEFUN (STELLA-NEED-TO-COMPILE? BOOLEAN) ((FILE OBJECT)))", ((cpp_function_code)(&stella_need_to_compileP)), NULL);
    define_function_object("SYSTEM-NEEDS-TRANSLATION?", "(DEFUN (SYSTEM-NEEDS-TRANSLATION? BOOLEAN) ((SYSTEMNAME STRING) (LANGUAGE KEYWORD)))", ((cpp_function_code)(&system_needs_translationP)), NULL);
    define_function_object("COMPUTE-FILE-UNITS-MODULE", "(DEFUN (COMPUTE-FILE-UNITS-MODULE MODULE) ((FILEUNITS (LIST OF TRANSLATION-UNIT))))", ((cpp_function_code)(&compute_file_units_module)), NULL);
    define_function_object("CONCATENATE-SYSTEM-UNITS", "(DEFUN (CONCATENATE-SYSTEM-UNITS (LIST OF TRANSLATION-UNIT)) ((SYSTEMUNITS SYSTEM-UNITS-ALIST)))", ((cpp_function_code)(&concatenate_system_units)), NULL);
    define_function_object("TRANSLATE-SYSTEM", "(DEFUN (TRANSLATE-SYSTEM BOOLEAN) ((SYSTEMNAME STRING) (OUTPUTLANGUAGE KEYWORD) |&REST| (OPTIONS OBJECT)) :DOCUMENTATION \"Translate all of the STELLA source files in `system' into a target language.\nIf :two-pass? is true, all files will be scanned twice, once to define\n   all objects defined in them, and once to actually translate them.\nIf :force-translation?, files will be translated whether or not they are up-to-date.\nIf :production-settings?, optimization levels will favor efficient code.\nReturn TRUE if at least one file was translated.\")", ((cpp_function_code)(&translate_system)), NULL);
    define_function_object("HELP-TRANSLATE-SYSTEM", "(DEFUN (HELP-TRANSLATE-SYSTEM BOOLEAN) ((FILES (CONS OF STRING-WRAPPER)) (TWOPASS? BOOLEAN) (FORCETRANSLATION? BOOLEAN)))", ((cpp_function_code)(&help_translate_system)), NULL);
    define_function_object("TRANSFER-NATIVE-SYSTEM-FILES", "(DEFUN TRANSFER-NATIVE-SYSTEM-FILES ((OUTPUTLANGUAGE KEYWORD)))", ((cpp_function_code)(&transfer_native_system_files)), NULL);
    define_function_object("STARTUP-TRANSLATE-FILE", "(DEFUN STARTUP-TRANSLATE-FILE () :PUBLIC? TRUE)", ((cpp_function_code)(&startup_translate_file)), NULL);
    { Method_Slot* function = lookup_function(SYM_TRANSLATE_FILE_STELLA_STARTUP_TRANSLATE_FILE);

      set_dynamic_slot_value(function->dynamic_slots, SYM_TRANSLATE_FILE_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string("Startup-Translate-File"), NULL_STRING_WRAPPER);
    }
  }
}

void startup_translate_file() {
  if (current_startup_time_phaseP(0)) {
    oLOGICAL_HOST_TABLEo = new_string_hash_table();
    define_logical_host_property("PL", NULL, NULL);
  }
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oSTELLA_MODULEo);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (current_startup_time_phaseP(2)) {
      help_startup_translate_file1();
    }
    if (current_startup_time_phaseP(4)) {
      oOSo = operating_system();
      oTYPE_TO_FILE_EXTENSION_TABLEo = listO(8, listO(3, KWD_TRANSLATE_FILE_STELLA, wrap_string(".ste"), NIL), listO(3, KWD_TRANSLATE_FILE_KB, wrap_string(".kb"), NIL), listO(3, KWD_TRANSLATE_FILE_CPP_CODE, wrap_string(".cc"), NIL), listO(3, KWD_TRANSLATE_FILE_CPP_HEADER, wrap_string(".hh"), NIL), listO(3, KWD_TRANSLATE_FILE_JAVA, wrap_string(".java"), NIL), listO(3, KWD_TRANSLATE_FILE_IDL, wrap_string(".idl"), NIL), listO(3, KWD_TRANSLATE_FILE_NONE, wrap_string(""), NIL), NIL);
    }
    if (current_startup_time_phaseP(5)) {
      define_stella_type_from_stringified_source("(DEFTYPE SYSTEM-UNITS-ALIST (KEY-VALUE-LIST OF STRING-WRAPPER (LIST OF TRANSLATION-UNIT)))");
    }
    if (current_startup_time_phaseP(6)) {
      finalize_classes();
    }
    if (current_startup_time_phaseP(7)) {
      help_startup_translate_file2();
    }
    if (current_startup_time_phaseP(8)) {
      finalize_slots();
      cleanup_unfinalized_classes();
    }
    if (current_startup_time_phaseP(9)) {
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *RECORD-SIGNATURES?* BOOLEAN FALSE :DOCUMENTATION \"If `true', record object signatures into a persistent table.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *WALK-WITH-RECORDED-SIGNATURES?* BOOLEAN FALSE :DOCUMENTATION \"If `true', the walker will try to find previously recorded\nsignatures for referenced but undefined functions, methods, etc.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *USERECORDEDSIGNATURES?* BOOLEAN FALSE :DOCUMENTATION \"If `true', `safe-lookup-slot' and `lookup-function' will\ntry to retrieve a previously recorded signature if normal lookup fails.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *ROOTSOURCEDIRECTORY* STRING \"PL:sources;\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *ROOTNATIVEDIRECTORY* STRING \"PL:natives;\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *ROOTBINARYDIRECTORY* STRING \"PL:binaries;\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *SYSTEMDEFINITIONSDIRECTORY* STRING NULL :DOCUMENTATION \"Points to the directory containing the SYSTEMS file.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *CURRENTSYSTEMDEFINITION* SYSTEM-DEFINITION NULL :DOCUMENTATION \"Points to the current system.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *CURRENTSYSTEMDEFINITIONSUBDIRECTORY* STRING \"\" :DOCUMENTATION \"Points to a path from the root directory down\nto a local directory containing sources, natives, or binaries.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *OS* KEYWORD (OPERATING-SYSTEM) :DOCUMENTATION \"The operating system we are running on.  Currently,\neither :UNIX, :WINDOWS, or :MAC.\")");
      define_stella_global_variable_from_stringified_source("(DEFCONSTANT TYPE-SEPARATOR CHARACTER #\\.)");
      define_stella_global_variable_from_stringified_source("(DEFCONSTANT LOGICAL-DIRECTORY-SEPARATOR CHARACTER #\\;)");
      define_stella_global_variable_from_stringified_source("(DEFCONSTANT LOGICAL-HOST-SEPARATOR CHARACTER #\\:)");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *LOGICAL-HOST-TABLE* (STRING-HASH-TABLE OF STRING KEY-VALUE-LIST) NULL)");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *TYPE-TO-FILE-EXTENSION-TABLE* (CONS OF CONS) (BQUOTE ((:STELLA \".ste\") (:KB \".kb\") (:CPP-CODE \".cc\") (:CPP-HEADER \".hh\") (:JAVA \".java\") (:IDL \".idl\") (:NONE \"\"))) :DOCUMENTATION \"Used by `select-file-extension'.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *LISP-SPLITTER-PATH* STRING \"lisp\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *JAVA-SPLITTER-PATH* STRING \"java\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *CPP-SPLITTER-PATH* STRING \"cpp\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *DONTTRUNCATEFILENAMES?* BOOLEAN FALSE :DOCUMENTATION \"When TRUE, overrides normal truncation of file names.\")");
    }
  }
}

Keyword* KWD_TRANSLATE_FILE_TRANSLATE = NULL;

Keyword* KWD_TRANSLATE_FILE_JAVA = NULL;

Keyword* KWD_TRANSLATE_FILE_COMMON_LISP = NULL;

Keyword* KWD_TRANSLATE_FILE_CPP = NULL;

Keyword* KWD_TRANSLATE_FILE_CPP_STANDALONE = NULL;

Keyword* KWD_TRANSLATE_FILE_IDL = NULL;

Surrogate* SGT_TRANSLATE_FILE_STELLA_CONS = NULL;

Keyword* KWD_TRANSLATE_FILE_DEFINE = NULL;

Keyword* KWD_TRANSLATE_FILE_FINALIZE = NULL;

Keyword* KWD_TRANSLATE_FILE_WALK = NULL;

Keyword* KWD_TRANSLATE_FILE_TRANSLATE_WITH_COPYRIGHT_HEADER = NULL;

Keyword* KWD_TRANSLATE_FILE_WINDOWS = NULL;

Keyword* KWD_TRANSLATE_FILE_UNIX = NULL;

Keyword* KWD_TRANSLATE_FILE_MAC = NULL;

Keyword* KWD_TRANSLATE_FILE_ROOT_DIRECTORY = NULL;

Keyword* KWD_TRANSLATE_FILE_LISP_TRANSLATIONS = NULL;

Keyword* KWD_TRANSLATE_FILE_STELLA = NULL;

Keyword* KWD_TRANSLATE_FILE_KB = NULL;

Keyword* KWD_TRANSLATE_FILE_CPP_CODE = NULL;

Keyword* KWD_TRANSLATE_FILE_CPP_HEADER = NULL;

Keyword* KWD_TRANSLATE_FILE_NONE = NULL;

Keyword* KWD_TRANSLATE_FILE_USE_COMMON_LISP_STRUCTS = NULL;

Keyword* KWD_TRANSLATE_FILE_USE_COMMON_LISP_VECTOR_STRUCTS = NULL;

Keyword* KWD_TRANSLATE_FILE_LISP_BINARY = NULL;

Keyword* KWD_TRANSLATE_FILE_LISP = NULL;

Keyword* KWD_TRANSLATE_FILE_DIRECTORY = NULL;

Keyword* KWD_TRANSLATE_FILE_OTHER = NULL;

Symbol* SYM_TRANSLATE_FILE_STELLA_SYSTEM_UNITS_ALIST = NULL;

Symbol* SYM_TRANSLATE_FILE_STELLA_STARTUP_TIME_PROGN = NULL;

Symbol* SYM_TRANSLATE_FILE_STELLA_VERBATIM = NULL;

Keyword* KWD_TRANSLATE_FILE_TWO_PASSp = NULL;

Keyword* KWD_TRANSLATE_FILE_FORCE_TRANSLATIONp = NULL;

Keyword* KWD_TRANSLATE_FILE_PRODUCTION_SETTINGSp = NULL;

Keyword* KWD_TRANSLATE_FILE_DEVELOPMENT_SETTINGSp = NULL;

Keyword* KWD_TRANSLATE_FILE_WARN_ABOUT_MISSING_METHODS = NULL;

Keyword* KWD_TRANSLATE_FILE_WARN_ABOUT_UNDEFINED_METHODS = NULL;

Keyword* KWD_TRANSLATE_FILE_USE_HARDCODED_SYMBOLS = NULL;

Symbol* SYM_TRANSLATE_FILE_STELLA_STARTUP_TRANSLATE_FILE = NULL;

Symbol* SYM_TRANSLATE_FILE_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace stella
