// type-predicates.cc

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

#include "stella/stella-system.hh"

namespace stella {

boolean oCLASS_HIERARCHY_BOOTEDpo = FALSE;

Surrogate* safe_primary_type(Object* x) {
  return (((!((boolean)(x))) ? SGT_TYPE_PREDICATES_STELLA_UNKNOWN : x->primary_type()));
}

boolean bootstrap_isaP(Object* x, Surrogate* type) {
  return (((boolean)(x)) &&
      ((x->primary_type() == type) ||
       (((boolean)(type->surrogate_value)) &&
        (((boolean)(x->primary_type()->surrogate_value)) &&
         isaP(x, type)))));
}

boolean bootstrap_subtype_ofP(Surrogate* subtype, Surrogate* supertype) {
  return ((subtype == supertype) ||
      (((boolean)(((Class*)(subtype->surrogate_value)))) &&
       (((boolean)(((Class*)(supertype->surrogate_value)))) &&
        subtype_ofP(subtype, supertype))));
}

boolean booleanP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_BOOLEAN_WRAPPER));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_BOOLEAN_WRAPPER));
}

boolean subtype_of_booleanP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_BOOLEAN_WRAPPER->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_BOOLEAN_WRAPPER));
  }
}

boolean integerP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_INTEGER_WRAPPER));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_INTEGER_WRAPPER));
}

boolean subtype_of_integerP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_INTEGER_WRAPPER->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_INTEGER_WRAPPER));
  }
}

boolean floatP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_FLOAT_WRAPPER));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_FLOAT_WRAPPER));
}

boolean subtype_of_floatP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_FLOAT_WRAPPER->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_FLOAT_WRAPPER));
  }
}

boolean stringP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_STRING_WRAPPER));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_STRING_WRAPPER) ||
      bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_VERBATIM_STRING_WRAPPER));
}

boolean subtype_of_stringP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_STRING_WRAPPER->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_STRING_WRAPPER) ||
        bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_VERBATIM_STRING_WRAPPER));
  }
}

boolean verbatim_stringP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_VERBATIM_STRING_WRAPPER));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_VERBATIM_STRING_WRAPPER));
}

boolean subtype_of_verbatim_stringP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_VERBATIM_STRING_WRAPPER->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_VERBATIM_STRING_WRAPPER));
  }
}

boolean characterP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_CHARACTER_WRAPPER));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_CHARACTER_WRAPPER));
}

boolean subtype_of_characterP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_CHARACTER_WRAPPER->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_CHARACTER_WRAPPER));
  }
}

boolean wrapperP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_WRAPPER));
  }
  return (booleanP(x) ||
      (integerP(x) ||
       (floatP(x) ||
        (stringP(x) ||
         (characterP(x) ||
          bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_WRAPPER))))));
}

boolean subtype_of_wrapperP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_WRAPPER->surrogate_value))));
  }
  else {
    return (subtype_of_booleanP(type) ||
        (subtype_of_integerP(type) ||
         (subtype_of_floatP(type) ||
          (subtype_of_stringP(type) ||
           (subtype_of_characterP(type) ||
            bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_WRAPPER))))));
  }
}

boolean surrogateP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_SURROGATE));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_SURROGATE));
}

boolean subtype_of_surrogateP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_SURROGATE->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_SURROGATE));
  }
}

boolean typeP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_SURROGATE));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_SURROGATE));
}

boolean subtype_of_typeP(Surrogate* type) {
  return (subtype_of_surrogateP(type));
}

boolean symbolP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_SYMBOL));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_SYMBOL) ||
      bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_TRANSIENT_SYMBOL));
}

boolean subtype_of_symbolP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_SYMBOL->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_SYMBOL) ||
        bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_TRANSIENT_SYMBOL));
  }
}

boolean transient_symbolP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_TRANSIENT_SYMBOL));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_TRANSIENT_SYMBOL));
}

boolean subtype_of_transient_symbolP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_TRANSIENT_SYMBOL->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_TRANSIENT_SYMBOL));
  }
}

boolean keywordP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_KEYWORD));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_KEYWORD));
}

boolean subtype_of_keywordP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_KEYWORD->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_KEYWORD));
  }
}

boolean consP(Object* x) {
  if (((boolean)(x))) {
    return (x->primary_type() == SGT_TYPE_PREDICATES_STELLA_CONS);
  }
  else {
    return (FALSE);
  }
}

boolean subtype_of_consP(Surrogate* type) {
  return (type == SGT_TYPE_PREDICATES_STELLA_CONS);
}

boolean stella_classP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    if (((boolean)(x))) {
      { Class* typeclass = ((Class*)(x->primary_type()->surrogate_value));

        return (((boolean)(typeclass)) &&
            subclass_ofP(typeclass, ((Class*)(SGT_TYPE_PREDICATES_STELLA_CLASS->surrogate_value))));
      }
    }
    else {
      return (FALSE);
    }
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_CLASS));
}

boolean subtype_of_classP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    { Class* typeclass = ((Class*)(type->surrogate_value));

      return (((boolean)(typeclass)) &&
          subclass_ofP(typeclass, ((Class*)(SGT_TYPE_PREDICATES_STELLA_CLASS->surrogate_value))));
    }
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_CLASS));
  }
}

boolean storage_slotP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_STORAGE_SLOT));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_STORAGE_SLOT));
}

boolean subtype_of_storage_slotP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_STORAGE_SLOT->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_STORAGE_SLOT));
  }
}

boolean method_slotP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_METHOD_SLOT));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_METHOD_SLOT));
}

boolean subtype_of_method_slotP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_METHOD_SLOT->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_METHOD_SLOT));
  }
}

boolean anchored_type_specifierP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_ANCHORED_TYPE_SPECIFIER));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_ANCHORED_TYPE_SPECIFIER));
}

boolean subtype_of_anchored_type_specifierP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_ANCHORED_TYPE_SPECIFIER->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_ANCHORED_TYPE_SPECIFIER));
  }
}

boolean parametric_type_specifierP(Object* x) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (((boolean)(x)) &&
        isaP(x, SGT_TYPE_PREDICATES_STELLA_PARAMETRIC_TYPE_SPECIFIER));
  }
  return (bootstrap_isaP(x, SGT_TYPE_PREDICATES_STELLA_PARAMETRIC_TYPE_SPECIFIER));
}

boolean subtype_of_parametric_type_specifierP(Surrogate* type) {
  if (oCLASS_HIERARCHY_BOOTEDpo) {
    return (subclass_ofP(((Class*)(type->surrogate_value)), ((Class*)(SGT_TYPE_PREDICATES_STELLA_PARAMETRIC_TYPE_SPECIFIER->surrogate_value))));
  }
  else {
    return (bootstrap_subtype_ofP(type, SGT_TYPE_PREDICATES_STELLA_PARAMETRIC_TYPE_SPECIFIER));
  }
}

boolean Object::standard_objectP() {
  { Object* self = this;

    return (FALSE);
  }
}

boolean Standard_Object::standard_objectP() {
  { Standard_Object* self = this;

    return (TRUE);
  }
}

void help_startup_type_predicates1() {
  {
    SGT_TYPE_PREDICATES_STELLA_UNKNOWN = ((Surrogate*)(intern_rigid_symbol_wrt_module("UNKNOWN", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_BOOLEAN_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("BOOLEAN-WRAPPER", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_INTEGER_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("INTEGER-WRAPPER", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_FLOAT_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("FLOAT-WRAPPER", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_STRING_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("STRING-WRAPPER", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_VERBATIM_STRING_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("VERBATIM-STRING-WRAPPER", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_CHARACTER_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("CHARACTER-WRAPPER", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("WRAPPER", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_SURROGATE = ((Surrogate*)(intern_rigid_symbol_wrt_module("SURROGATE", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_SYMBOL = ((Surrogate*)(intern_rigid_symbol_wrt_module("SYMBOL", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_TRANSIENT_SYMBOL = ((Surrogate*)(intern_rigid_symbol_wrt_module("TRANSIENT-SYMBOL", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_KEYWORD = ((Surrogate*)(intern_rigid_symbol_wrt_module("KEYWORD", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_CONS = ((Surrogate*)(intern_rigid_symbol_wrt_module("CONS", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_CLASS = ((Surrogate*)(intern_rigid_symbol_wrt_module("CLASS", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_STORAGE_SLOT = ((Surrogate*)(intern_rigid_symbol_wrt_module("STORAGE-SLOT", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_METHOD_SLOT = ((Surrogate*)(intern_rigid_symbol_wrt_module("METHOD-SLOT", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_ANCHORED_TYPE_SPECIFIER = ((Surrogate*)(intern_rigid_symbol_wrt_module("ANCHORED-TYPE-SPECIFIER", NULL, 1)));
    SGT_TYPE_PREDICATES_STELLA_PARAMETRIC_TYPE_SPECIFIER = ((Surrogate*)(intern_rigid_symbol_wrt_module("PARAMETRIC-TYPE-SPECIFIER", NULL, 1)));
    SYM_TYPE_PREDICATES_STELLA_STARTUP_TYPE_PREDICATES = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-TYPE-PREDICATES", NULL, 0)));
    SYM_TYPE_PREDICATES_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-STARTUP-CLASSNAME", NULL, 0)));
  }
}

void startup_type_predicates() {
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oSTELLA_MODULEo);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (current_startup_time_phaseP(2)) {
      help_startup_type_predicates1();
    }
    if (current_startup_time_phaseP(6)) {
      finalize_classes();
    }
    if (current_startup_time_phaseP(7)) {
      define_function_object("SAFE-PRIMARY-TYPE", "(DEFUN (SAFE-PRIMARY-TYPE TYPE) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&safe_primary_type)), NULL);
      define_function_object("BOOTSTRAP-ISA?", "(DEFUN (BOOTSTRAP-ISA? BOOLEAN) ((X OBJECT) (TYPE TYPE)))", ((cpp_function_code)(&bootstrap_isaP)), NULL);
      define_function_object("BOOTSTRAP-SUBTYPE-OF?", "(DEFUN (BOOTSTRAP-SUBTYPE-OF? BOOLEAN) ((SUBTYPE TYPE) (SUPERTYPE TYPE)))", ((cpp_function_code)(&bootstrap_subtype_ofP)), NULL);
      define_function_object("BOOLEAN?", "(DEFUN (BOOLEAN? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&booleanP)), NULL);
      define_function_object("SUBTYPE-OF-BOOLEAN?", "(DEFUN (SUBTYPE-OF-BOOLEAN? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_booleanP)), NULL);
      define_function_object("INTEGER?", "(DEFUN (INTEGER? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&integerP)), NULL);
      define_function_object("SUBTYPE-OF-INTEGER?", "(DEFUN (SUBTYPE-OF-INTEGER? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_integerP)), NULL);
      define_function_object("FLOAT?", "(DEFUN (FLOAT? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&floatP)), NULL);
      define_function_object("SUBTYPE-OF-FLOAT?", "(DEFUN (SUBTYPE-OF-FLOAT? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_floatP)), NULL);
      define_function_object("STRING?", "(DEFUN (STRING? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&stringP)), NULL);
      define_function_object("SUBTYPE-OF-STRING?", "(DEFUN (SUBTYPE-OF-STRING? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_stringP)), NULL);
      define_function_object("VERBATIM-STRING?", "(DEFUN (VERBATIM-STRING? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&verbatim_stringP)), NULL);
      define_function_object("SUBTYPE-OF-VERBATIM-STRING?", "(DEFUN (SUBTYPE-OF-VERBATIM-STRING? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_verbatim_stringP)), NULL);
      define_function_object("CHARACTER?", "(DEFUN (CHARACTER? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&characterP)), NULL);
      define_function_object("SUBTYPE-OF-CHARACTER?", "(DEFUN (SUBTYPE-OF-CHARACTER? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_characterP)), NULL);
      define_function_object("WRAPPER?", "(DEFUN (WRAPPER? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&wrapperP)), NULL);
      define_function_object("SUBTYPE-OF-WRAPPER?", "(DEFUN (SUBTYPE-OF-WRAPPER? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_wrapperP)), NULL);
      define_function_object("SURROGATE?", "(DEFUN (SURROGATE? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&surrogateP)), NULL);
      define_function_object("SUBTYPE-OF-SURROGATE?", "(DEFUN (SUBTYPE-OF-SURROGATE? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_surrogateP)), NULL);
      define_function_object("TYPE?", "(DEFUN (TYPE? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&typeP)), NULL);
      define_function_object("SUBTYPE-OF-TYPE?", "(DEFUN (SUBTYPE-OF-TYPE? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE :GLOBALLY-INLINE? TRUE (RETURN (SUBTYPE-OF-SURROGATE? TYPE)))", ((cpp_function_code)(&subtype_of_typeP)), NULL);
      define_function_object("SYMBOL?", "(DEFUN (SYMBOL? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&symbolP)), NULL);
      define_function_object("SUBTYPE-OF-SYMBOL?", "(DEFUN (SUBTYPE-OF-SYMBOL? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_symbolP)), NULL);
      define_function_object("TRANSIENT-SYMBOL?", "(DEFUN (TRANSIENT-SYMBOL? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&transient_symbolP)), NULL);
      define_function_object("SUBTYPE-OF-TRANSIENT-SYMBOL?", "(DEFUN (SUBTYPE-OF-TRANSIENT-SYMBOL? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_transient_symbolP)), NULL);
      define_function_object("KEYWORD?", "(DEFUN (KEYWORD? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&keywordP)), NULL);
      define_function_object("SUBTYPE-OF-KEYWORD?", "(DEFUN (SUBTYPE-OF-KEYWORD? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_keywordP)), NULL);
      define_function_object("CONS?", "(DEFUN (CONS? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&consP)), NULL);
      define_function_object("SUBTYPE-OF-CONS?", "(DEFUN (SUBTYPE-OF-CONS? BOOLEAN) ((TYPE TYPE)) :GLOBALLY-INLINE? TRUE :PUBLIC? TRUE (RETURN (EQL? TYPE @CONS)))", ((cpp_function_code)(&subtype_of_consP)), NULL);
      define_function_object("STELLA-CLASS?", "(DEFUN (STELLA-CLASS? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&stella_classP)), NULL);
      define_function_object("SUBTYPE-OF-CLASS?", "(DEFUN (SUBTYPE-OF-CLASS? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_classP)), NULL);
      define_function_object("STORAGE-SLOT?", "(DEFUN (STORAGE-SLOT? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&storage_slotP)), NULL);
      define_function_object("SUBTYPE-OF-STORAGE-SLOT?", "(DEFUN (SUBTYPE-OF-STORAGE-SLOT? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_storage_slotP)), NULL);
      define_function_object("METHOD-SLOT?", "(DEFUN (METHOD-SLOT? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&method_slotP)), NULL);
      define_function_object("SUBTYPE-OF-METHOD-SLOT?", "(DEFUN (SUBTYPE-OF-METHOD-SLOT? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_method_slotP)), NULL);
      define_function_object("ANCHORED-TYPE-SPECIFIER?", "(DEFUN (ANCHORED-TYPE-SPECIFIER? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&anchored_type_specifierP)), NULL);
      define_function_object("SUBTYPE-OF-ANCHORED-TYPE-SPECIFIER?", "(DEFUN (SUBTYPE-OF-ANCHORED-TYPE-SPECIFIER? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_anchored_type_specifierP)), NULL);
      define_function_object("PARAMETRIC-TYPE-SPECIFIER?", "(DEFUN (PARAMETRIC-TYPE-SPECIFIER? BOOLEAN) ((X OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&parametric_type_specifierP)), NULL);
      define_function_object("SUBTYPE-OF-PARAMETRIC-TYPE-SPECIFIER?", "(DEFUN (SUBTYPE-OF-PARAMETRIC-TYPE-SPECIFIER? BOOLEAN) ((TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&subtype_of_parametric_type_specifierP)), NULL);
      define_method_object("(DEFMETHOD (STANDARD-OBJECT? BOOLEAN) ((SELF OBJECT)) :PUBLIC? TRUE)", ((cpp_method_code)(&Object::standard_objectP)), ((cpp_method_code)(NULL)));
      define_method_object("(DEFMETHOD (STANDARD-OBJECT? BOOLEAN) ((SELF STANDARD-OBJECT)) :PUBLIC? TRUE)", ((cpp_method_code)(&Standard_Object::standard_objectP)), ((cpp_method_code)(NULL)));
      define_function_object("STARTUP-TYPE-PREDICATES", "(DEFUN STARTUP-TYPE-PREDICATES () :PUBLIC? TRUE)", ((cpp_function_code)(&startup_type_predicates)), NULL);
      { Method_Slot* function = lookup_function(SYM_TYPE_PREDICATES_STELLA_STARTUP_TYPE_PREDICATES);

        set_dynamic_slot_value(function->dynamic_slots, SYM_TYPE_PREDICATES_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string("Startup-Type-Predicates"), NULL_STRING_WRAPPER);
      }
    }
    if (current_startup_time_phaseP(8)) {
      finalize_slots();
      cleanup_unfinalized_classes();
    }
    if (current_startup_time_phaseP(9)) {
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *CLASS-HIERARCHY-BOOTED?* BOOLEAN FALSE)");
    }
  }
}

Surrogate* SGT_TYPE_PREDICATES_STELLA_UNKNOWN = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_BOOLEAN_WRAPPER = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_INTEGER_WRAPPER = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_FLOAT_WRAPPER = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_STRING_WRAPPER = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_VERBATIM_STRING_WRAPPER = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_CHARACTER_WRAPPER = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_WRAPPER = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_SURROGATE = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_SYMBOL = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_TRANSIENT_SYMBOL = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_KEYWORD = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_CONS = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_CLASS = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_STORAGE_SLOT = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_METHOD_SLOT = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_ANCHORED_TYPE_SPECIFIER = NULL;

Surrogate* SGT_TYPE_PREDICATES_STELLA_PARAMETRIC_TYPE_SPECIFIER = NULL;

Symbol* SYM_TYPE_PREDICATES_STELLA_STARTUP_TYPE_PREDICATES = NULL;

Symbol* SYM_TYPE_PREDICATES_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace stella
