// walk.cc

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

#include "stella/stella-system.hh"

namespace stella {

// List of Stella environment features.
List* oAVAILABLE_ENVIRONMENT_FEATURESo = NULL;

// List of currently enabled environment features.
List* oCURRENT_ENVIRONMENT_FEATURESo = NULL;

// List of features enabled upon call to
// `reset-environment-features'.
List* oDEFAULT_ENVIRONMENT_FEATURESo = NULL;

void display_environment_features() {
  // Print a list of the currently enabled environment 
  // features and print a list of all Stella environment features.
  cout << "CURRENTLY ENABLED ENVIRONMENT FEATURES:" << endl << "   " << oCURRENT_ENVIRONMENT_FEATURESo << endl << endl << "AVAILABLE ENVIRONMENT FEATURES:" << endl << "   " << oAVAILABLE_ENVIRONMENT_FEATURESo << endl;
}

void enable_environment_feature(int features, ...) {
  // Enable the environment feature(s) indicated by the
  // keywords in `features'.
  { Keyword* f = NULL;
    va_list iter_000;
    int iter_000_count = features;

    va_start(iter_000, features);
    while ((iter_000_count--) > 0) {
      f = va_arg(iter_000, Keyword*);
      if ((f == KWD_WALK_SUPPRESS_WARNINGS) ||
          ((f == KWD_WALK_WARN_ABOUT_UNDEFINED_METHODS) ||
           ((f == KWD_WALK_WARN_ABOUT_MISSING_METHODS) ||
            ((f == KWD_WALK_USE_HARDCODED_SYMBOLS) ||
             ((f == KWD_WALK_USE_COMMON_LISP_STRUCTS) ||
              ((f == KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS) ||
               ((f == KWD_WALK_USE_COMMON_LISP_CONSES) ||
                ((f == KWD_WALK_USE_CPP_GARBAGE_COLLECTOR) ||
                 (f == KWD_WALK_TRANSLATE_WITH_COPYRIGHT_HEADER))))))))) {
        if (!(oCURRENT_ENVIRONMENT_FEATURESo->memberP(f))) {
          oCURRENT_ENVIRONMENT_FEATURESo->insert(f);
        }
      }
      else {
        if (oAVAILABLE_ENVIRONMENT_FEATURESo->memberP(f)) {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "Oops. Forgot to implement environment feature:  " << "`" << f << "'";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
        else {
          *(STANDARD_WARNING->native_stream) << "Warning: " << "Unrecognizable environment feature:  " << "`" << f << "'" << endl;
        }
      }
      if (f == KWD_WALK_USE_COMMON_LISP_STRUCTS) {
        oCURRENT_ENVIRONMENT_FEATURESo->remove(KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS);
      }
      else if (f == KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS) {
        if (!(oCURRENT_ENVIRONMENT_FEATURESo->memberP(KWD_WALK_USE_COMMON_LISP_STRUCTS))) {
          oCURRENT_ENVIRONMENT_FEATURESo->insert(KWD_WALK_USE_COMMON_LISP_STRUCTS);
        }
      }
      else {
      }
    }
    va_end(iter_000);
  }
}

void disable_environment_feature(int features, ...) {
  // Disable the environment feature(s) indicated by the
  // keywords in `features'.
  { Keyword* f = NULL;
    va_list iter_000;
    int iter_000_count = features;

    va_start(iter_000, features);
    while ((iter_000_count--) > 0) {
      f = va_arg(iter_000, Keyword*);
      if ((f == KWD_WALK_SUPPRESS_WARNINGS) ||
          ((f == KWD_WALK_WARN_ABOUT_UNDEFINED_METHODS) ||
           ((f == KWD_WALK_WARN_ABOUT_MISSING_METHODS) ||
            ((f == KWD_WALK_USE_HARDCODED_SYMBOLS) ||
             ((f == KWD_WALK_USE_COMMON_LISP_STRUCTS) ||
              ((f == KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS) ||
               ((f == KWD_WALK_USE_COMMON_LISP_CONSES) ||
                ((f == KWD_WALK_USE_CPP_GARBAGE_COLLECTOR) ||
                 (f == KWD_WALK_TRANSLATE_WITH_COPYRIGHT_HEADER))))))))) {
        oCURRENT_ENVIRONMENT_FEATURESo->remove(f);
      }
      else {
        if (oAVAILABLE_ENVIRONMENT_FEATURESo->memberP(f)) {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "Oops. Forgot to implement environment feature:  " << "`" << f << "'";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
        else {
          *(STANDARD_WARNING->native_stream) << "Warning: " << "Unrecognizable environment feature:  " << "`" << f << "'" << endl;
        }
      }
      if ((f == KWD_WALK_USE_COMMON_LISP_STRUCTS) ||
          (f == KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS)) {
        oCURRENT_ENVIRONMENT_FEATURESo->remove(KWD_WALK_USE_COMMON_LISP_STRUCTS);
        oCURRENT_ENVIRONMENT_FEATURESo->remove(KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS);
      }
      else {
      }
    }
    va_end(iter_000);
  }
}

void reset_environment_features() {
  // Reset the current environment features to their default
  // settings.
  { Keyword* f = NULL;
    Cons* iter_000 = oCURRENT_ENVIRONMENT_FEATURESo->the_cons_list;

    while (!(iter_000 == NIL)) {
      f = ((Keyword*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (!oDEFAULT_ENVIRONMENT_FEATURESo->memberP(f)) {
        disable_environment_feature(1, f);
      }
    }
  }
  { Keyword* f = NULL;
    Cons* iter_001 = oDEFAULT_ENVIRONMENT_FEATURESo->the_cons_list;

    while (!(iter_001 == NIL)) {
      f = ((Keyword*)(iter_001->value));
      iter_001 = iter_001->rest;
      if (!oCURRENT_ENVIRONMENT_FEATURESo->memberP(f)) {
        enable_environment_feature(1, f);
      }
    }
  }
}

List* oTRACED_KEYWORDSo = NULL;

Object* trace_if_expander(Object* keyword, Cons* body) {
  return (listO(4, SYM_WALK_STELLA_WHEN, listO(4, SYM_WALK_STELLA_AND, listO(3, SYM_WALK_STELLA_DEFINEDp, SYM_WALK_STELLA_oTRACED_KEYWORDSo, NIL), (consP(keyword) ? listO(7, SYM_WALK_STELLA_EXISTS, SYM_WALK_STELLA_KWD, SYM_WALK_STELLA_IN, listO(3, SYM_WALK_STELLA_QUOTE, keyword, NIL), SYM_WALK_STELLA_WHERE, listO(4, SYM_WALK_STELLA_MEMBp, SYM_WALK_STELLA_oTRACED_KEYWORDSo, SYM_WALK_STELLA_KWD, NIL), NIL) : listO(4, SYM_WALK_STELLA_MEMBp, SYM_WALK_STELLA_oTRACED_KEYWORDSo, keyword, NIL)), NIL), cons(SYM_WALK_STELLA_PRINT, copy_cons_list(body)->concatenate(NIL, 0)), NIL));
}

boolean trace_keywordP(Object* keyword) {
  { boolean test_value_000 = FALSE;

    test_value_000 = ((boolean)(oTRACED_KEYWORDSo));
    if (test_value_000) {
      { boolean choose_value_000 = FALSE;

        if (consP(keyword)) {
          { boolean foundP_000 = FALSE;

            { Object* kwd = NULL;
              Cons* iter_000 = ((Cons*)(keyword));

              while (!(iter_000 == NIL)) {
                kwd = iter_000->value;
                iter_000 = iter_000->rest;
                if (oTRACED_KEYWORDSo->membP(((Keyword*)(kwd)))) {
                  foundP_000 = TRUE;
                  break;
                }
              }
            }
            choose_value_000 = foundP_000;
          }
        }
        else {
          choose_value_000 = oTRACED_KEYWORDSo->membP(((Keyword*)(keyword)));
        }
        test_value_000 = choose_value_000;
      }
    }
    { boolean value_000 = test_value_000;

      return (value_000);
    }
  }
}

List* add_trace(Cons* keywords) {
  // Enable trace messages identified by any of the listed `keywords'.  After
  // calling (`add-trace' <keyword>) code guarded by (`trace-if' <keyword> ...)
  // will be executed when it is encountered.
  if (!((boolean)(oTRACED_KEYWORDSo))) {
    oTRACED_KEYWORDSo = list(0);
  }
  { Generalized_Symbol* k = NULL;
    Cons* iter_000 = keywords;

    while (!(iter_000 == NIL)) {
      k = ((Generalized_Symbol*)(iter_000->value));
      iter_000 = iter_000->rest;
      oTRACED_KEYWORDSo->insert_new(string_keywordify(k->symbol_name));
    }
  }
  return (oTRACED_KEYWORDSo);
}

List* add_trace_evaluator_wrapper(Cons* arguments) {
  return (add_trace(arguments));
}

List* drop_trace(Cons* keywords) {
  // Disable trace messages identified by any of the listed `keywords'.  After
  // calling (`drop-trace' <keyword>) code guarded by (`trace-if' <keyword> ...)
  // will not be executed when it is encountered.
  if (((boolean)(oTRACED_KEYWORDSo))) {
    { Generalized_Symbol* k = NULL;
      Cons* iter_000 = keywords;

      while (!(iter_000 == NIL)) {
        k = ((Generalized_Symbol*)(iter_000->value));
        iter_000 = iter_000->rest;
        oTRACED_KEYWORDSo->remove(string_keywordify(k->symbol_name));
      }
    }
    if (oTRACED_KEYWORDSo->emptyP()) {
      oTRACED_KEYWORDSo = NULL;
    }
  }
  return (oTRACED_KEYWORDSo);
}

List* drop_trace_evaluator_wrapper(Cons* arguments) {
  return (drop_trace(arguments));
}

void clear_trace() {
  // Disable all tracing previously enabled with `add-trace'.
  oTRACED_KEYWORDSo->clear();
}

// Integer between 0 and 3.  Higher levels call more
// safety checks.
DEFINE_STELLA_SPECIAL(oSAFETYo, int , 3);

// Integer between 0 and 3.  Higher levels generate more
// code to aid debugging.
DEFINE_STELLA_SPECIAL(oDEBUGLEVELo, int , 3);

// Integer between 0 and 3.  Higher levels optimize for
// greater execution speed.
DEFINE_STELLA_SPECIAL(oOPTIMIZESPEEDLEVELo, int , 3);

// Integer between 0 and 3.  Higher levels optimize for
// less code size and memory consumption.
DEFINE_STELLA_SPECIAL(oOPTIMIZESPACELEVELo, int , 3);

void set_optimization_levels(int safety, int debug, int speed, int space) {
  // Set optimization levels for the qualities `safety', `debug',
  // `speed', and `space'.
  if ((safety != NULL_INTEGER) &&
      ((safety >= 0) &&
       (safety <= 3))) {
    oSAFETYo.set(safety);
  }
  if ((debug != NULL_INTEGER) &&
      ((debug >= 0) &&
       (debug <= 3))) {
    oDEBUGLEVELo.set(debug);
  }
  if ((speed != NULL_INTEGER) &&
      ((speed >= 0) &&
       (speed <= 3))) {
    oOPTIMIZESPEEDLEVELo.set(speed);
  }
  if ((space != NULL_INTEGER) &&
      ((space >= 0) &&
       (space <= 3))) {
    oOPTIMIZESPACELEVELo.set(space);
  }
}

boolean preserve_tail_merge_optimizabilityP() {
  return ((oOPTIMIZESPEEDLEVELo.get() >= 3) &&
      (oDEBUGLEVELo.get() <= 0));
}

boolean method_call_inlining_enabledP() {
  return ((oOPTIMIZESPEEDLEVELo.get() >= 3) &&
      (oDEBUGLEVELo.get() <= 1));
}

boolean optimize_boolean_testsP() {
  return ((oOPTIMIZESPEEDLEVELo.get() >= 2) &&
      (oDEBUGLEVELo.get() <= 2));
}

boolean check_for_illegal_returnP() {
  return ((oSAFETYo.get() >= 3) &&
      ((!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA)) &&
       (!preserve_tail_merge_optimizabilityP())));
}

// Stores quoted cons trees.
String_Hash_Table* oQUOTATION_TABLEo = NULL;

char* name_quoted_tree(Cons* tree) {
  { Cons* nametree = cons(tree, cons(wrap_string(oMODULEo.get()->module_full_name), NIL));
    char* name = stringify(nametree);

    return (name);
  }
}

Cons* get_quoted_tree(char* tree_name, char* modulename) {
  // Return the quoted tree with name `tree-name'.
  { Cons* quotedtree = ((Cons*)(oQUOTATION_TABLEo->lookup(tree_name)));

    if (((boolean)(quotedtree))) {
      return (quotedtree);
    }
    { 
      BIND_STELLA_SPECIAL(oMODULEo, Module*, get_stella_module(modulename, TRUE));
      { Cons* tree = ((Cons*)(read_s_expression_from_string(tree_name)));

        quotedtree = ((Cons*)(permanentify_form(tree->value)));
        oQUOTATION_TABLEo->insert_at(tree_name, quotedtree);
        return (quotedtree);
      }
    }
  }
}

Translation_Unit* new_translation_unit() {
  { Translation_Unit* self = NULL;

    self = new Translation_Unit();
    self->referenced_globals = new_list();
    self->translation = NULL;
    self->code_register = NULL;
    self->auxiliaryP = FALSE;
    self->annotation = NULL;
    self->category = NULL;
    self->the_object = NULL;
    self->tu_home_module = NULL;
    return (self);
  }
}

Surrogate* Translation_Unit::primary_type() {
  { Translation_Unit* self = this;

    return (SGT_WALK_STELLA_TRANSLATION_UNIT);
  }
}

Module* Translation_Unit::home_module() {
  { Translation_Unit* self = this;

    return (self->tu_home_module);
  }
}

Object* access_translation_unit_slot_value(Translation_Unit* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_WALK_STELLA_TU_HOME_MODULE) {
    if (setvalueP) {
      self->tu_home_module = ((Module*)(value));
    }
    else {
      value = self->tu_home_module;
    }
  }
  else if (slotname == SYM_WALK_STELLA_THE_OBJECT) {
    if (setvalueP) {
      self->the_object = value;
    }
    else {
      value = self->the_object;
    }
  }
  else if (slotname == SYM_WALK_STELLA_CATEGORY) {
    if (setvalueP) {
      self->category = ((Symbol*)(value));
    }
    else {
      value = self->category;
    }
  }
  else if (slotname == SYM_WALK_STELLA_ANNOTATION) {
    if (setvalueP) {
      self->annotation = ((String_Wrapper*)(value))->wrapper_value;
    }
    else {
      value = wrap_string(self->annotation);
    }
  }
  else if (slotname == SYM_WALK_STELLA_AUXILIARYp) {
    if (setvalueP) {
      self->auxiliaryP = coerce_wrapped_boolean_to_boolean(((Boolean_Wrapper*)(value)));
    }
    else {
      value = (self->auxiliaryP ? TRUE_WRAPPER : FALSE_WRAPPER);
    }
  }
  else if (slotname == SYM_WALK_STELLA_CODE_REGISTER) {
    if (setvalueP) {
      self->code_register = value;
    }
    else {
      value = self->code_register;
    }
  }
  else if (slotname == SYM_WALK_STELLA_TRANSLATION) {
    if (setvalueP) {
      self->translation = value;
    }
    else {
      value = self->translation;
    }
  }
  else if (slotname == SYM_WALK_STELLA_REFERENCED_GLOBALS) {
    if (setvalueP) {
      self->referenced_globals = ((List*)(value));
    }
    else {
      value = self->referenced_globals;
    }
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << slotname << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
  return (value);
}

void Translation_Unit::print_object(ostream* stream) {
  { Translation_Unit* self = this;

    print_translation_unit(self, stream);
  }
}

void clear_translation_unit(Translation_Unit* self) {
  self->tu_home_module = NULL;
  self->the_object = NULL;
  self->category = NULL;
  self->code_register = NULL;
  self->translation = NULL;
}

void print_translation_unit(Translation_Unit* self, ostream* stream) {
  *(stream) << "|TU|[" << self->category;
  { Symbol* test_value_000 = self->category;

    if ((test_value_000 == SYM_WALK_STELLA_CLASS) ||
        ((test_value_000 == SYM_WALK_STELLA_METHOD) ||
         ((test_value_000 == SYM_WALK_STELLA_MACRO) ||
          (test_value_000 == SYM_WALK_STELLA_GLOBAL_VARIABLE)))) {
      *(stream) << " " << self->the_object;
    }
    else {
    }
  }
  *(stream) << "]";
}

// List of objects representing partially walked
// top-level definitions and auxiliary code.
DEFINE_STELLA_SPECIAL(oTRANSLATIONUNITSo, List* , NULL);

// The translation unit currently operated on.
DEFINE_STELLA_SPECIAL(oCURRENTTRANSLATIONUNITo, Translation_Unit* , NULL);

// Indicates the current translation phase which is one of
// :DEFINE, :FINALIZE, :WALK, or :TRANSLATE.
DEFINE_STELLA_SPECIAL(oTRANSLATIONPHASEo, Keyword* , NULL);

// The higher the level, the more progress annotations are
// generated during the translation of Stella declarations.
DEFINE_STELLA_SPECIAL(oTRANSLATIONVERBOSITYLEVELo, int , 1);

DEFINE_STELLA_SPECIAL(oUSEHARDCODEDSYMBOLSpo, boolean , FALSE);

boolean use_hardcoded_symbolsP() {
  return (oCURRENT_ENVIRONMENT_FEATURESo->memberP(KWD_WALK_USE_HARDCODED_SYMBOLS) ||
      oUSEHARDCODEDSYMBOLSpo.get());
}

// Specifies the current translator output language; either
// :common-lisp, :idl, :java, :cpp, or :cpp-standalone.
DEFINE_STELLA_SPECIAL(oTRANSLATOROUTPUTLANGUAGEo, Keyword* , NULL);

Keyword* translator_output_language() {
  return (oTRANSLATOROUTPUTLANGUAGEo.get());
}

char* translator_output_language_name() {
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
    return ("Common Lisp");
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) {
    return ("C++");
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) {
    return ("Java");
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE) {
    return ("standalone-C++");
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_IDL) {
    return ("IDL");
  }
  else {
    return (oTRANSLATOROUTPUTLANGUAGEo.get()->symbol_name);
  }
}

Keyword* set_translator_output_language(Keyword* new_language) {
  // Set output language to `new-language'.  Return previous language.
  { Keyword* oldlanguage = oTRANSLATOROUTPUTLANGUAGEo.get();

    if (!(listO(6, KWD_WALK_CPP, KWD_WALK_CPP_STANDALONE, KWD_WALK_COMMON_LISP, KWD_WALK_JAVA, KWD_WALK_IDL, NIL)->memberP(new_language))) {
      *(STANDARD_WARNING->native_stream) << "Warning: " << "`" << new_language << "'" << " is not a legal translation language" << endl;
    }
    oTRANSLATOROUTPUTLANGUAGEo.set(new_language);
    return (oldlanguage);
  }
}

boolean translate_to_common_lispP() {
  // Return `true' if current output language is Common-Lisp.
  return (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP);
}

boolean translate_to_cppP() {
  // Return `true' if current output language is C++
  return (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP);
}

boolean translate_to_javaP() {
  // Return `true' if current output language is Java
  return (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA);
}

boolean translate_to_single_inheritance_languageP() {
  if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) ||
      ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
       (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA))) {
    return (TRUE);
  }
  else {
    return (FALSE);
  }
}

Keyword* toggle_output_language() {
  // Switch between Common Lisp and C++ as output languages.
  if (translate_to_common_lispP()) {
    oTRANSLATOROUTPUTLANGUAGEo.set(KWD_WALK_CPP);
    return (KWD_WALK_CPP);
  }
  else if (translate_to_cppP()) {
    oTRANSLATOROUTPUTLANGUAGEo.set(KWD_WALK_JAVA);
    return (KWD_WALK_JAVA);
  }
  else {
    oTRANSLATOROUTPUTLANGUAGEo.set(KWD_WALK_COMMON_LISP);
    return (KWD_WALK_COMMON_LISP);
  }
}

boolean use_bootstrap_translatorP() {
  return (FALSE);
}

Keyword* toggle_translators() {
  return (NULL);
}

boolean declaration_operatorP(Symbol* renamed_Operator) {
  return (get_quoted_tree("((DEFCLASS DEFSLOT DEFMETHOD DEFUN DEFTYPE DEFGLOBAL DEFSPECIAL DEFCONSTANT DEFMACRO DEFMODULE STARTUP-TIME-PROGN) \"/STELLA\")", "/STELLA")->memberP(renamed_Operator));
}

boolean declaration_treeP(Cons* tree) {
  { Object* renamed_Operator = tree->value;

    if (declaration_operatorP(((Symbol*)(renamed_Operator)))) {
      return (TRUE);
    }
    if (renamed_Operator == SYM_WALK_STELLA_PROGN) {
      { boolean foundP_000 = FALSE;

        { Object* subtree = NULL;
          Cons* iter_000 = tree->rest;

          while (!(iter_000 == NIL)) {
            subtree = iter_000->value;
            iter_000 = iter_000->rest;
            if (consP(subtree) &&
                declaration_treeP(((Cons*)(subtree)))) {
              foundP_000 = TRUE;
              break;
            }
          }
        }
        { boolean value_000 = foundP_000;

          return (value_000);
        }
      }
    }
    if (((boolean)(lookup_macro(((Symbol*)(renamed_Operator)))))) {
      { Object* expansion = expand_macro(copy_cons_list(tree));

        if (((boolean)(expansion))) {
          if (safe_primary_type(expansion) == SGT_WALK_STELLA_CONS) {
            { Object* expansion_000 = expansion;
              Cons* expansion = ((Cons*)(expansion_000));

              tree->value = expansion->value;
              tree->rest = expansion->rest;
              return (declaration_treeP(tree));
            }
          }
          else {
            return (FALSE);
          }
        }
        else {
          tree->value = SYM_WALK_STELLA_PROGN;
          tree->rest = NIL;
          return (FALSE);
        }
      }
    }
    return (FALSE);
  }
}

char* create_annotation(Cons* top_level_tree) {
  return (string_concatenate("(", string_concatenate(((Symbol*)(top_level_tree->value))->symbol_name, string_concatenate(" ", string_concatenate(stringify(top_level_tree->rest->value), " ...)", 0), 0), 0), 0));
}

void walk_top_level_tree(Cons* tree, boolean createannotationP) {
  { 
    BIND_STELLA_SPECIAL(oCURRENTTRANSLATIONUNITo, Translation_Unit*, NULL);
    { char* annotation = NULL;

      if (createannotationP) {
        annotation = create_annotation(tree);
        if (oTRANSLATIONVERBOSITYLEVELo.get() >= 2) {
          cout << "Defining " << "`" << annotation << "'" << endl;
        }
      }
      { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

        if (test_value_000 == SYM_WALK_STELLA_DEFCLASS) {
          walk_defclass_tree(tree);
        }
        else if (test_value_000 == SYM_WALK_STELLA_DEFSLOT) {
          walk_defslot_tree(tree);
        }
        else if ((test_value_000 == SYM_WALK_STELLA_DEFMETHOD) ||
            ((test_value_000 == SYM_WALK_STELLA_DEFUN) ||
             (test_value_000 == SYM_WALK_STELLA_DEFOPERATOR))) {
          if (tree->value == SYM_WALK_STELLA_DEFOPERATOR) {
            tree->first_setter(SYM_WALK_STELLA_DEFMETHOD);
          }
          walk_defmethod_tree(tree);
        }
        else if ((test_value_000 == SYM_WALK_STELLA_DEFGLOBAL) ||
            ((test_value_000 == SYM_WALK_STELLA_DEFSPECIAL) ||
             (test_value_000 == SYM_WALK_STELLA_DEFCONSTANT))) {
          walk_defglobal_tree(tree);
        }
        else if (test_value_000 == SYM_WALK_STELLA_DEFTYPE) {
          walk_deftype_tree(tree);
        }
        else if (test_value_000 == SYM_WALK_STELLA_DEFMODULE) {
          walk_defmodule_tree(tree);
        }
        else if (test_value_000 == SYM_WALK_STELLA_DEFMACRO) {
          walk_defmacro_tree(tree);
        }
        else if (test_value_000 == SYM_WALK_STELLA_STARTUP_TIME_PROGN) {
          walk_startup_time_progn_tree(tree);
        }
        else if (test_value_000 == SYM_WALK_STELLA_VERBATIM) {
          walk_verbatim_definition_tree(tree);
        }
        else if (test_value_000 == SYM_WALK_STELLA_PROGN) {
          { Object* form = NULL;
            Cons* iter_000 = tree->rest;

            while (!(iter_000 == NIL)) {
              form = iter_000->value;
              iter_000 = iter_000->rest;
              walk_top_level_tree(((Cons*)(form)), createannotationP);
            }
          }
          return;
        }
        else {
          if (((boolean)(lookup_macro(((Symbol*)(tree->value)))))) {
            { Object* expansion = expand_macro(tree);

              if (((boolean)(expansion))) {
                walk_top_level_tree(((Cons*)(expansion)), createannotationP);
              }
            }
          }
          else {
            if (!(incremental_translationP() ||
                ((boolean)(lookup_command(((Symbol*)(tree->value))))))) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_WARNING);
                  *(STANDARD_WARNING->native_stream) << endl << " " << "Illegal top-level statement:" << endl << "   " << "`" << de_uglify_parse_tree(tree) << "'" << endl << " Wrapping it into a 'startup-time-progn'" << "." << endl;
                }
              }
            }
            tree = listO(3, SYM_WALK_STELLA_STARTUP_TIME_PROGN, tree, NIL);
            walk_top_level_tree(tree, FALSE);
          }
          return;
        }
      }
      if (((boolean)(oCURRENTTRANSLATIONUNITo.get()))) {
        oTRANSLATIONUNITSo.get()->push(oCURRENTTRANSLATIONUNITo.get());
        oCURRENTTRANSLATIONUNITo.get()->annotation = annotation;
      }
    }
  }
}

Object* walk_top_level_expression(Object* tree) {
  { 
    BIND_STELLA_SPECIAL(oLOCALVARIABLETYPETABLEo, Key_Value_List*, new_key_value_list());
    { Object* otree = walk_without_type_tree(tree);

      if (vrlet_expressionP(otree)) {
        return (walk_without_type_tree(listO(3, SYM_WALK_STELLA_PROGN, otree, NIL)));
      }
      return (otree);
    }
  }
}

void walk_phase_one_unit(Translation_Unit* unit) {
  { 
    BIND_STELLA_SPECIAL(oLOCALVARIABLETYPETABLEo, Key_Value_List*, new_key_value_list());
    BIND_STELLA_SPECIAL(oCURRENTTRANSLATIONUNITo, Translation_Unit*, unit);
    { Symbol* test_value_000 = unit->category;

      if (test_value_000 == SYM_WALK_STELLA_GLOBAL_VARIABLE) {
        walk_global_unit(unit);
      }
      else if (test_value_000 == SYM_WALK_STELLA_MODULE) {
        walk_module_unit(unit);
      }
      else if (test_value_000 == SYM_WALK_STELLA_TYPE) {
        walk_type_unit(unit);
      }
      else if ((test_value_000 == SYM_WALK_STELLA_METHOD) ||
          (test_value_000 == SYM_WALK_STELLA_PRINT_METHOD)) {
        walk_method_unit(unit);
      }
      else if (test_value_000 == SYM_WALK_STELLA_MACRO) {
        walk_macro_unit(unit);
      }
      else if (test_value_000 == SYM_WALK_STELLA_CLASS) {
        walk_class_unit(unit);
      }
      else if (test_value_000 == SYM_WALK_STELLA_SLOT) {
        walk_slot_unit(unit);
      }
      else if (test_value_000 == SYM_WALK_STELLA_STARTUP_TIME_PROGN) {
        walk_startup_time_progn_unit(unit);
      }
      else if (test_value_000 == SYM_WALK_STELLA_VERBATIM) {
        walk_verbatim_unit(unit);
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
    }
  }
}

void walk_auxiliary_tree(Cons* tree) {
  help_walk_auxiliary_tree(tree, TRUE);
}

Translation_Unit* help_walk_auxiliary_tree(Cons* tree, boolean finalizeP) {
  if (!(get_quoted_tree("((DEFCLASS DEFMETHOD DEFUN DEFMACRO DEFGLOBAL DEFSPECIAL DEFCONSTANT STARTUP-TIME-PROGN VERBATIM) \"/STELLA\")", "/STELLA")->memberP(tree->value))) {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "OOPS, illegal auxiliary statement:" << endl << "   " << "`" << tree << "'";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
  walk_top_level_tree(tree, FALSE);
  if (finalizeP) {
    finalize_classes_and_slots();
  }
  { Translation_Unit* unit = ((Translation_Unit*)(oTRANSLATIONUNITSo.get()->pop()));

    unit->auxiliaryP = TRUE;
    walk_phase_one_unit(unit);
    return (unit);
  }
}

void walk_all_phase_one_units() {
  { List* phaseoneunits = oTRANSLATIONUNITSo.get();

    oTRANSLATIONUNITSo.set(new_list());
    { Translation_Unit* unit = NULL;
      Cons* iter_000 = phaseoneunits->the_cons_list;

      while (!(iter_000 == NIL)) {
        unit = ((Translation_Unit*)(iter_000->value));
        iter_000 = iter_000->rest;
        if ((oTRANSLATIONVERBOSITYLEVELo.get() >= 2) &&
            (unit->annotation != NULL)) {
          cout << "Walking " << "`" << unit->annotation << "'" << endl;
        }
        walk_phase_one_unit(unit);
      }
    }
    create_finalization_units();
    create_startup_function_units();
    phaseoneunits->clear();
    oTRANSLATIONUNITSo.get()->reverse();
  }
}

void translate_all_units() {
  { Translation_Unit* unit = NULL;
    Cons* iter_000 = oTRANSLATIONUNITSo.get()->the_cons_list;

    while (!(iter_000 == NIL)) {
      unit = ((Translation_Unit*)(iter_000->value));
      iter_000 = iter_000->rest;
      { 
        BIND_STELLA_SPECIAL(oCURRENTTRANSLATIONUNITo, Translation_Unit*, unit);
        if ((oTRANSLATIONVERBOSITYLEVELo.get() >= 2) &&
            (unit->annotation != NULL)) {
          cout << "Translating " << "`" << unit->annotation << "'" << endl;
        }
        if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
          unit->translation = cl_translate_unit(unit);
        }
        else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_IDL) {
          unit->translation = idl_translate_unit(unit);
        }
        else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) ||
            (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA_STANDALONE)) {
          if (!((boolean)(unit->translation))) {
            unit->translation = java_translate_unit(unit);
          }
        }
        else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
            (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE)) {
          unit->translation = cpp_translate_unit(unit);
        }
        else {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
      }
    }
  }
}

boolean unit_included_in_other_unitsP(Translation_Unit* unit) {
  { Method_Slot* method = NULL;

    if (!(unit->category == SYM_WALK_STELLA_METHOD)) {
      return (FALSE);
    }
    else {
      {
        method = ((Method_Slot*)(unit->the_object));
        if (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_CONSTRUCTORp, FALSE_WRAPPER)))->wrapper_value) {
          { Translation_Unit* other_unit = NULL;
            Cons* iter_000 = oTRANSLATIONUNITSo.get()->the_cons_list;

            while (!(iter_000 == NIL)) {
              other_unit = ((Translation_Unit*)(iter_000->value));
              iter_000 = iter_000->rest;
              if ((other_unit->category == SYM_WALK_STELLA_CLASS) &&
                  (canonical_type(method->slot_base_type) == ((Class*)(other_unit->the_object))->class_type)) {
                return (TRUE);
              }
            }
          }
          return (FALSE);
        }
        else {
          { Translation_Unit* other_unit = NULL;
            Cons* iter_001 = oTRANSLATIONUNITSo.get()->the_cons_list;

            while (!(iter_001 == NIL)) {
              other_unit = ((Translation_Unit*)(iter_001->value));
              iter_001 = iter_001->rest;
              if ((other_unit->category == SYM_WALK_STELLA_CLASS) &&
                  (canonical_type(method->slot_owner) == ((Class*)(other_unit->the_object))->class_type)) {
                return (TRUE);
              }
            }
          }
          return (FALSE);
        }
      }
    }
  }
}

Cons* combine_translated_trees() {
  { Cons* otree = NIL;

    if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) ||
        ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
         ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE) ||
          (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_IDL)))) {
      { Translation_Unit* unit = NULL;
        Cons* iter_000 = oTRANSLATIONUNITSo.get()->the_cons_list;
        Cons* collect_000 = NULL;

        while (!(iter_000 == NIL)) {
          unit = ((Translation_Unit*)(iter_000->value));
          iter_000 = iter_000->rest;
          if (!((boolean)(collect_000))) {
            {
              collect_000 = cons(unit->translation, NIL);
              if (otree == NIL) {
                otree = collect_000;
              }
              else {
                add_cons_to_end_of_cons_list(otree, collect_000);
              }
            }
          }
          else {
            {
              collect_000->rest = cons(unit->translation, NIL);
              collect_000 = collect_000->rest;
            }
          }
        }
      }
    }
    else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) ||
        (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA_STANDALONE)) {
      { Translation_Unit* unit = NULL;
        Cons* iter_001 = oTRANSLATIONUNITSo.get()->the_cons_list;
        Cons* collect_001 = NULL;

        while (!(iter_001 == NIL)) {
          unit = ((Translation_Unit*)(iter_001->value));
          iter_001 = iter_001->rest;
          if (!unit_included_in_other_unitsP(unit)) {
            if (!((boolean)(collect_001))) {
              {
                collect_001 = cons(unit->translation, NIL);
                if (otree == NIL) {
                  otree = collect_001;
                }
                else {
                  add_cons_to_end_of_cons_list(otree, collect_001);
                }
              }
            }
            else {
              {
                collect_001->rest = cons(unit->translation, NIL);
                collect_001 = collect_001->rest;
              }
            }
          }
        }
      }
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
    if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
      otree = cons(intern_common_lisp_symbol("PROGN"), otree->concatenate(NIL, 0));
    }
    else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
        ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE) ||
         ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) ||
          (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_IDL)))) {
    }
    else {
      { Output_String_Stream* stream_001 = new_output_string_stream();

        *(stream_001->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_001->the_string_reader());
      }
    }
    return (otree);
  }
}

Object* translate_walked_tree(Object* codetree) {
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
    return (cl_translate_a_tree(codetree));
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_IDL) {
    return (idl_translate_a_tree(codetree));
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) {
    return (java_translate_a_tree(codetree));
  }
  else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
      (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE)) {
    return (cpp_translate_a_tree(codetree));
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
}

Object* incrementally_translate(Object* tree) {
  // Translate a single Stella expression `tree' and return
  // the result.  For C++ and Java print the translation to standard output and
  // return NIL instead.
  { boolean toplevelinvocationP = !((boolean)(oTRANSLATIONUNITSo.get()));
    Object* ocode = NULL;
    boolean declarationP = consP(tree) &&
        declaration_treeP(((Cons*)(tree)));

    { 
      BIND_STELLA_SPECIAL(oTRANSLATIONUNITSo, List*, new_list());
      BIND_STELLA_SPECIAL(oTRANSLATIONVERBOSITYLEVELo, int, 0);
      BIND_STELLA_SPECIAL(oTRANSLATIONPHASEo, Keyword*, NULL);
      BIND_STELLA_SPECIAL(oTRANSLATIONERRORSo, int, 0);
      BIND_STELLA_SPECIAL(oTRANSLATIONWARNINGSo, int, 0);
      if (declarationP) {
        oTRANSLATIONPHASEo.set(KWD_WALK_DEFINE);
        walk_top_level_tree(((Cons*)(tree)), FALSE);
        if (translation_errorsP()) {
          summarize_translation_errors();
          return (NIL);
        }
        oTRANSLATIONUNITSo.get()->reverse();
        oTRANSLATIONPHASEo.set(KWD_WALK_FINALIZE);
        finalize_classes_and_slots();
        oTRANSLATIONPHASEo.set(KWD_WALK_WALK);
        if (!(use_hardcoded_symbolsP())) {
          clear_symbol_registry();
        }
        walk_all_phase_one_units();
      }
      else {
        oTRANSLATIONPHASEo.set(KWD_WALK_WALK);
        { 
          BIND_STELLA_SPECIAL(oUSEHARDCODEDSYMBOLSpo, boolean, TRUE);
          ocode = walk_top_level_expression(tree);
        }
      }
      if (translation_errorsP()) {
        summarize_translation_errors();
        return (NIL);
      }
      oTRANSLATIONPHASEo.set(KWD_WALK_TRANSLATE);
      if (declarationP) {
        translate_all_units();
        ocode = combine_translated_trees();
      }
      else {
        { 
          BIND_STELLA_SPECIAL(oUSEHARDCODEDSYMBOLSpo, boolean, TRUE);
          ocode = translate_walked_tree(ocode);
        }
      }
      if (translation_errorsP()) {
        summarize_translation_errors();
        return (NIL);
      }
      if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
      }
      else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_IDL) {
        idl_output_parse_tree(((Cons*)(ocode)));
        ocode = NIL;
      }
      else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) {
        java_output_parse_tree(((Cons*)(ocode)));
        ocode = NIL;
      }
      else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
          (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE)) {
        cpp_output_parse_tree(((Cons*)(ocode)));
        ocode = NIL;
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
      { Translation_Unit* unit = NULL;
        Cons* iter_000 = oTRANSLATIONUNITSo.get()->the_cons_list;

        while (!(iter_000 == NIL)) {
          unit = ((Translation_Unit*)(iter_000->value));
          iter_000 = iter_000->rest;
          unit->free();
        }
      }
      oTRANSLATIONUNITSo.get()->clear();
      if (toplevelinvocationP) {
        sweep_transients();
      }
      summarize_translation_errors();
      return (ocode);
    }
  }
}

// Table mapping local variable names their declared types
// (declared explicitly or implicitly).
DEFINE_STELLA_SPECIAL(oLOCALVARIABLETYPETABLEo, Key_Value_List* , NULL);

// Contains the method or function being walked, or else `null'.
DEFINE_STELLA_SPECIAL(oMETHODBEINGWALKEDo, Method_Slot* , NULL);

// Indicates that one or more return statements have been found
// during the walk of the current method.
DEFINE_STELLA_SPECIAL(oFOUNDRETURNpo, boolean , FALSE);

// Bound to the target type for an expression currently walked.
// Used instead of an extra argument to `walk-a-tree', since only a few types
// of expressions need to know about their expected type (e.g., FUNCALL).
DEFINE_STELLA_SPECIAL(oTARGETTYPEo, Standard_Object* , NULL);

// Table that maps each prefix of a function-local gensym
// to its own gensym counter and/or to related gensyms.
DEFINE_STELLA_SPECIAL(oLOCALGENSYMTABLEo, Key_Value_List* , NULL);

DEFINE_STELLA_SPECIAL(oTRANSLATIONERRORSo, int , 0);

DEFINE_STELLA_SPECIAL(oTRANSLATIONWARNINGSo, int , 0);

void reset_translation_errors() {
  oTRANSLATIONERRORSo.set(0);
  oTRANSLATIONWARNINGSo.set(0);
}

void signal_translation_error() {
  oTRANSLATIONERRORSo.set(oTRANSLATIONERRORSo.get() + 1);
}

void signal_translation_warning() {
  oTRANSLATIONWARNINGSo.set(oTRANSLATIONWARNINGSo.get() + 1);
}

// If `true' all passes of a translation will always be performed
// regardless of whether any errors were encountered.  Otherwise, translation
// ends with the first pass that encountered an error.
DEFINE_STELLA_SPECIAL(oIGNORETRANSLATIONERRORSpo, boolean , TRUE);

boolean ignore_translation_errorsP() {
  return (oIGNORETRANSLATIONERRORSpo.get());
}

boolean translation_errorsP() {
  return ((oTRANSLATIONERRORSo.get() > 0) &&
      (!ignore_translation_errorsP()));
}

void summarize_translation_errors() {
  if (oTRANSLATIONERRORSo.get() > 0) {
    cout << oTRANSLATIONERRORSo.get() << " error";
    if (oTRANSLATIONERRORSo.get() > 1) {
      cout << "s";
    }
  }
  if (oTRANSLATIONWARNINGSo.get() > 0) {
    if (oTRANSLATIONERRORSo.get() > 0) {
      cout << ", ";
    }
    cout << oTRANSLATIONWARNINGSo.get() << " warning";
    if (oTRANSLATIONWARNINGSo.get() > 1) {
      cout << "s";
    }
  }
  if ((oTRANSLATIONERRORSo.get() > 0) ||
      (oTRANSLATIONWARNINGSo.get() > 0)) {
    cout << "." << endl;
  }
}

void print_error_context(char* prefix, Output_Stream* stream) {
  *(stream->native_stream) << prefix << "While ";
  if (((boolean)(oTRANSLATIONPHASEo.get()))) {
    if (oTRANSLATIONPHASEo.get() == KWD_WALK_DEFINE) {
      *(stream->native_stream) << "defining ";
    }
    else if (oTRANSLATIONPHASEo.get() == KWD_WALK_FINALIZE) {
      *(stream->native_stream) << "finalizing ";
    }
    else if (oTRANSLATIONPHASEo.get() == KWD_WALK_WALK) {
      *(stream->native_stream) << "walking ";
    }
    else if (oTRANSLATIONPHASEo.get() == KWD_WALK_TRANSLATE) {
      *(stream->native_stream) << "translating ";
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << oTRANSLATIONPHASEo.get() << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
  else {
    *(stream->native_stream) << "processing ";
  }
  if (((boolean)(oCURRENTTRANSLATIONUNITo.get()))) {
    { Object* object = oCURRENTTRANSLATIONUNITo.get()->the_object;
      Symbol* category = oCURRENTTRANSLATIONUNITo.get()->category;

      if (category == SYM_WALK_STELLA_CLASS) {
        *(stream->native_stream) << "class ";
        if (subtype_of_classP(safe_primary_type(object))) {
          { Object* object_000 = object;
            Class* object = ((Class*)(object_000));

            *(stream->native_stream) << object->name();
          }
        }
        else {
          *(stream->native_stream) << object;
        }
      }
      else if (category == SYM_WALK_STELLA_SLOT) {
        *(stream->native_stream) << "external slot ";
        { Surrogate* test_value_000 = safe_primary_type(object);

          if (test_value_000 == SGT_WALK_STELLA_CONS) {
            { Object* object_001 = object;
              Cons* object = ((Cons*)(object_001));

              *(stream->native_stream) << object->rest->value << "." << object->rest->rest->value;
            }
          }
          else if (subtype_of_storage_slotP(test_value_000)) {
            { Object* object_002 = object;
              Storage_Slot* object = ((Storage_Slot*)(object_002));

              *(stream->native_stream) << object->slot_owner << "." << object->slot_name->symbol_name;
            }
          }
          else {
            { Output_String_Stream* stream_001 = new_output_string_stream();

              *(stream_001->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_001->the_string_reader());
            }
          }
        }
      }
      else if ((category == SYM_WALK_STELLA_METHOD) ||
          ((category == SYM_WALK_STELLA_PRINT_METHOD) ||
           (category == SYM_WALK_STELLA_MACRO))) {
        { Surrogate* test_value_001 = safe_primary_type(object);

          if (subtype_of_method_slotP(test_value_001)) {
            { Object* object_003 = object;
              Method_Slot* object = ((Method_Slot*)(object_003));

              if (category == SYM_WALK_STELLA_MACRO) {
                *(stream->native_stream) << "macro expander ";
              }
              if (object->method_functionP) {
                *(stream->native_stream) << "function " << object->slot_name;
              }
              else {
                *(stream->native_stream) << "method " << object->slot_owner << "." << object->slot_name->symbol_name;
              }
            }
          }
          else if (test_value_001 == SGT_WALK_STELLA_CONS) {
            { Object* object_004 = object;
              Cons* object = ((Cons*)(object_004));

              if (object->value == SYM_WALK_STELLA_DEFUN) {
                *(stream->native_stream) << "function ";
              }
              else {
                *(stream->native_stream) << "method ";
              }
              *(stream->native_stream) << ((consP(object->rest->value) ? ((Cons*)(object->rest->value))->value : object->rest->value));
            }
          }
          else {
            { Output_String_Stream* stream_002 = new_output_string_stream();

              *(stream_002->native_stream) << "`" << test_value_001 << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_002->the_string_reader());
            }
          }
        }
      }
      else if (category == SYM_WALK_STELLA_GLOBAL_VARIABLE) {
        { Surrogate* test_value_002 = safe_primary_type(object);

          if (subtype_ofP(test_value_002, SGT_WALK_STELLA_GLOBAL_VARIABLE)) {
            { Object* object_005 = object;
              Global_Variable* object = ((Global_Variable*)(object_005));

              if (object->variable_constantP) {
                *(stream->native_stream) << "constant ";
              }
              else {
                if (object->variable_specialP) {
                  *(stream->native_stream) << "special variable ";
                }
                else {
                  *(stream->native_stream) << "global variable ";
                }
              }
              *(stream->native_stream) << object->variable_name;
            }
          }
          else if (test_value_002 == SGT_WALK_STELLA_CONS) {
            { Object* object_006 = object;
              Cons* object = ((Cons*)(object_006));

              *(stream->native_stream) << object->value << " " << object->rest->value;
            }
          }
          else {
            { Output_String_Stream* stream_003 = new_output_string_stream();

              *(stream_003->native_stream) << "`" << test_value_002 << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_003->the_string_reader());
            }
          }
        }
      }
      else if (category == SYM_WALK_STELLA_TYPE) {
        *(stream->native_stream) << "type ";
        { Surrogate* test_value_003 = safe_primary_type(object);

          if (test_value_003 == SGT_WALK_STELLA_CONS) {
            { Object* object_007 = object;
              Cons* object = ((Cons*)(object_007));

              *(stream->native_stream) << object->rest->value;
            }
          }
          else if (subtype_ofP(test_value_003, SGT_WALK_STELLA_TYPE_SPEC)) {
            { Object* object_008 = object;
              Standard_Object* object = ((Standard_Object*)(object_008));

              *(stream->native_stream) << yield_type_spec_tree(object);
            }
          }
          else {
            { Output_String_Stream* stream_004 = new_output_string_stream();

              *(stream_004->native_stream) << "`" << test_value_003 << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_004->the_string_reader());
            }
          }
        }
      }
      else if (category == SYM_WALK_STELLA_MODULE) {
        *(stream->native_stream) << "module ";
        { Surrogate* test_value_004 = safe_primary_type(object);

          if (subtype_ofP(test_value_004, SGT_WALK_STELLA_MODULE)) {
            { Object* object_009 = object;
              Module* object = ((Module*)(object_009));

              *(stream->native_stream) << object->module_name;
            }
          }
          else if (test_value_004 == SGT_WALK_STELLA_CONS) {
            { Object* object_010 = object;
              Cons* object = ((Cons*)(object_010));

              *(stream->native_stream) << object->rest->value;
            }
          }
          else {
            { Output_String_Stream* stream_005 = new_output_string_stream();

              *(stream_005->native_stream) << "`" << test_value_004 << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_005->the_string_reader());
            }
          }
        }
      }
      else {
        *(stream->native_stream) << category;
      }
    }
  }
  else {
    *(stream->native_stream) << "top-level statement";
  }
  *(stream->native_stream) << ":";
}

// The current log file to which function calls should be logged.
// A non-NULL value indicates that function call logging is enabled.
Output_Stream* oFUNCTION_CALL_LOG_STREAMo = NULL;

// Translation switch which indicates that methods should
// be instrumented to log their calls to a file.
boolean oLOG_FUNCTION_CALLSpo = FALSE;

boolean log_function_callP(Method_Slot* method) {
  if (oLOG_FUNCTION_CALLSpo &&
      ((!((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_SLOT_AUXILIARYp, FALSE_WRAPPER)))->wrapper_value) &&
       (!((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_GLOBALLY_INLINEp, FALSE_WRAPPER)))->wrapper_value))) {
    { Symbol* test_value_000 = method->slot_name;

      if ((test_value_000 == SYM_WALK_STELLA_LOG_FUNCTION_CALL) ||
          ((test_value_000 == SYM_WALK_STELLA_START_FUNCTION_CALL_LOGGING) ||
           ((test_value_000 == SYM_WALK_STELLA_STOP_FUNCTION_CALL_LOGGING) ||
            ((test_value_000 == SYM_WALK_STELLA_SET_CALL_LOG_BREAK_POINT) ||
             ((test_value_000 == SYM_WALK_STELLA_BREAK_PROGRAM) ||
              ((test_value_000 == SYM_WALK_STELLA_TERMINATE_PROGRAM) ||
               ((test_value_000 == SYM_WALK_STELLA_PO) ||
                ((test_value_000 == SYM_WALK_STELLA_DEFINEDp) ||
                 ((test_value_000 == SYM_WALK_STELLA_NULLp) ||
                  ((test_value_000 == SYM_WALK_STELLA_EQp) ||
                   ((test_value_000 == SYM_WALK_STELLA_EQLp) ||
                    ((test_value_000 == SYM_WALK_STELLA_l) ||
                     ((test_value_000 == SYM_WALK_STELLA_le) ||
                      ((test_value_000 == SYM_WALK_STELLA_g) ||
                       ((test_value_000 == SYM_WALK_STELLA_ge) ||
                        (test_value_000 == SYM_WALK_STELLA_CAST)))))))))))))))) {
        return (FALSE);
      }
      else {
        { boolean test_value_001 = FALSE;

          if (method->method_parameter_type_specifiers_reader()->memberP(SGT_WALK_STELLA_UNKNOWN)) {
            test_value_001 = TRUE;
          }
          else {
            if (method->method_parameter_type_specifiers_reader()->memberP(SGT_WALK_STELLA_LISP_CODE)) {
              test_value_001 = TRUE;
            }
            else {
              if (method->method_return_type_specifiers_reader()->memberP(SGT_WALK_STELLA_UNKNOWN)) {
                test_value_001 = TRUE;
              }
              else {
                if (method->method_return_type_specifiers_reader()->memberP(SGT_WALK_STELLA_LISP_CODE)) {
                  test_value_001 = TRUE;
                }
                else {
                  {
                    { boolean foundP_000 = FALSE;

                      { Standard_Object* tspec = NULL;
                        Cons* iter_000 = method->method_parameter_type_specifiers_reader()->the_cons_list;

                        while (!(iter_000 == NIL)) {
                          tspec = ((Standard_Object*)(iter_000->value));
                          iter_000 = iter_000->rest;
                          if (sub_type_spec_ofP(tspec, SGT_WALK_STELLA_OUTPUT_STREAM)) {
                            foundP_000 = TRUE;
                            break;
                          }
                        }
                      }
                      test_value_001 = foundP_000;
                    }
                    if (!test_value_001) {
                      { boolean foundP_001 = FALSE;

                        { Standard_Object* tspec = NULL;
                          Cons* iter_001 = method->method_parameter_type_specifiers_reader()->the_cons_list;

                          while (!(iter_001 == NIL)) {
                            tspec = ((Standard_Object*)(iter_001->value));
                            iter_001 = iter_001->rest;
                            if (sub_type_spec_ofP(tspec, SGT_WALK_STELLA_NATIVE_OUTPUT_STREAM)) {
                              foundP_001 = TRUE;
                              break;
                            }
                          }
                        }
                        test_value_001 = foundP_001;
                      }
                      if (!test_value_001) {
                        test_value_001 = (oCURRENTFILEo.get() != NULL) &&
                            ((string_search(oCURRENTFILEo.get(), "cl-primal", 0) != NULL_INTEGER) ||
                             (string_search(oCURRENTFILEo.get(), "stella-to-cl", 0) != NULL_INTEGER));
                      }
                    }
                  }
                }
              }
            }
          }
          if (test_value_001) {
            return (FALSE);
          }
        }
      }
    }
    return (TRUE);
  }
  return (FALSE);
}

Cons* wrap_body_with_log_function_call_tree(Method_Slot* method, Cons* body) {
  if (consP(body->value) &&
      (!(((Cons*)(body->value))->value == SYM_WALK_STELLA_INLINE))) {
    return (cons(listO(3, SYM_WALK_STELLA_LOG_FUNCTION_CALL, wrap_string(stringify(method)), NIL), body->concatenate(NIL, 0)));
  }
  else {
    return (body);
  }
}

int oLOG_BREAK_POINT_COUNTERo = NULL_INTEGER;

void log_function_call(char* name) {
  if (!(oFUNCTION_CALL_LOG_STREAMo == NULL)) {
    *(oFUNCTION_CALL_LOG_STREAMo->native_stream) << name << endl;
  }
  if (oLOG_BREAK_POINT_COUNTERo != NULL_INTEGER) {
    oLOG_BREAK_POINT_COUNTERo = oLOG_BREAK_POINT_COUNTERo - 1;
    if (oLOG_BREAK_POINT_COUNTERo == 0) {
      oLOG_BREAK_POINT_COUNTERo = NULL_INTEGER;
      break_program("Call log break point");
    }
  }
}

void start_function_call_logging(char* filename) {
  // Start function call logging to `fileName'.
  { Output_Stream* currentlogstream = oFUNCTION_CALL_LOG_STREAMo;

    oFUNCTION_CALL_LOG_STREAMo = NULL;
    if (((boolean)(currentlogstream))) {
      currentlogstream->free();
    }
    oFUNCTION_CALL_LOG_STREAMo = new_output_file_stream(filename);
  }
}

void start_function_call_logging_evaluator_wrapper(Cons* arguments) {
  start_function_call_logging(((String_Wrapper*)(arguments->value))->wrapper_value);
}

void stop_function_call_logging() {
  // Stop function call logging and close the current log file.
  { Output_Stream* currentlogstream = oFUNCTION_CALL_LOG_STREAMo;

    oFUNCTION_CALL_LOG_STREAMo = NULL;
    if (((boolean)(currentlogstream))) {
      currentlogstream->free();
    }
  }
}

void set_call_log_break_point(int count) {
  // Set a call log break point to `count'.  Execution will be
  // interrupted right at the entry of the `count'th logged function call.
  oLOG_BREAK_POINT_COUNTERo = count;
}

void set_call_log_break_point_evaluator_wrapper(Cons* arguments) {
  set_call_log_break_point(((Integer_Wrapper*)(arguments->value))->wrapper_value);
}

void break_program(char* message) {
  // Interrupt the program and print `message'.  Continue after
  // confirmation with the user.
  {
    cout << "Program break: " << message << endl << endl;
    if (!yes_or_noP("Continue?")) {
      terminate_program();
    }
  }
}

void terminate_program() {
  // Terminate and exit the program with normal exit code.
  exit(0);;
}

void po(Object* thing) {
  cout << thing << endl;
}

boolean variable_eqlP(Symbol* var1, Symbol* var2) {
  return ((var1 == var2) ||
      string_eqlP(var1->symbol_name, var2->symbol_name));
}

Object* lookup_variable_table(Key_Value_List* self, Symbol* variable) {
  { Symbol* key = NULL;
    Object* value = NULL;
    Kv_Cons* iter_000 = self->the_kv_list;

    while (((boolean)(iter_000))) {
      key = ((Symbol*)(iter_000->key));
      value = iter_000->value;
      iter_000 = iter_000->rest;
      if (variable_eqlP(key, variable)) {
        return (value);
      }
    }
  }
  return (NULL);
}

void insert_at_variable_table(Key_Value_List* self, Symbol* variable, Object* value) {
  { Kv_Cons* cursor = self->the_kv_list;

    while (((boolean)(cursor))) {
      if (variable_eqlP(((Symbol*)(cursor->key)), variable)) {
        cursor->value = value;
        return;
      }
      cursor = cursor->rest;
    }
    self->the_kv_list = kv_cons(variable, value, self->the_kv_list);
  }
}

void push_variable_binding(Symbol* variable, Standard_Object* type) {
  { Class* renamed_Class = type_spec_to_class(type);

    if (!((boolean)(variable))) {
      return;
    }
    if ((!((boolean)(renamed_Class))) &&
        (!(type == SGT_WALK_STELLA_UNINITIALIZED))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "No class defined for type: " << "`" << type_spec_to_base_type(type) << "'" << "." << endl;
        }
      }
      type = SGT_WALK_STELLA_UNKNOWN;
    }
    if ((!((!translate_to_single_inheritance_languageP()) ||
        (!(oMIXIN_IMPLEMENTATION_STYLEo == KWD_WALK_SECOND_CLASS)))) &&
        (((boolean)(renamed_Class)) &&
         (renamed_Class->mixinP &&
          ((!((boolean)(oMETHODBEINGWALKEDo.get()))) ||
           ((!mixin_methodP(oMETHODBEINGWALKEDo.get())) ||
            oLOCALVARIABLETYPETABLEo.get()->non_emptyP()))))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Variable " << "`" << de_uglify_parse_tree(variable) << "'" << " is typed with a mixin class which is" << endl << " illegal when translating to a single-inheritance language" << "." << endl;
        }
      }
    }
    oLOCALVARIABLETYPETABLEo.get()->the_kv_list = kv_cons(variable, type, oLOCALVARIABLETYPETABLEo.get()->the_kv_list);
    maybe_rename_local_variable(variable);
  }
}

void pop_variable_binding() {
  { Kv_Cons* kvlist = oLOCALVARIABLETYPETABLEo.get()->the_kv_list;

    oLOCALVARIABLETYPETABLEo.get()->the_kv_list = kvlist->rest;
    if (subtype_ofP(safe_primary_type(kvlist->value), SGT_WALK_STELLA_KEY_VALUE_LIST)) {
      kvlist->value->free();
    }
    else {
    }
    free_kv_cons(kvlist);
  }
}

Standard_Object* lookup_variable_type(Symbol* variablename) {
  { Standard_Object* entry = (((boolean)(oLOCALVARIABLETYPETABLEo.get())) ? ((Standard_Object*)(lookup_variable_table(oLOCALVARIABLETYPETABLEo.get(), variablename))) : ((Standard_Object*)(NULL)));
    Standard_Object* type = NULL;

    if (!((boolean)(entry))) {
      if (variablename == SYM_WALK_STELLA_NULL) {
        return (SGT_WALK_STELLA_UNKNOWN);
      }
      type = lookup_global_variable_type(variablename);
      register_reference_to_global_variable(variablename);
    }
    else {
      if (subtype_ofP(safe_primary_type(entry), SGT_WALK_STELLA_KEY_VALUE_LIST)) {
        { Standard_Object* entry_000 = entry;
          Key_Value_List* entry = ((Key_Value_List*)(entry_000));

          type = ((Standard_Object*)(entry->lookup(KWD_WALK_TYPE)));
        }
      }
      else {
        type = entry;
      }
    }
    if (!((boolean)(type))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Undeclared variable " << "`" << de_uglify_parse_tree(variablename) << "'" << "." << endl;
        }
      }
      return (SGT_WALK_STELLA_UNKNOWN);
    }
    if (type == SGT_WALK_STELLA_UNINITIALIZED) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Reference to defined but uninitialized variable " << "`" << de_uglify_parse_tree(variablename) << "'" << "." << endl;
        }
      }
      return (SGT_WALK_STELLA_UNKNOWN);
    }
    return (type);
  }
}

void set_local_variable_info(Symbol* variable, Keyword* key, Object* info) {
  { Standard_Object* entry = ((Standard_Object*)(lookup_variable_table(oLOCALVARIABLETYPETABLEo.get(), variable)));

    if (((boolean)(entry))) {
      if (subtype_ofP(safe_primary_type(entry), SGT_WALK_STELLA_KEY_VALUE_LIST)) {
        { Standard_Object* entry_000 = entry;
          Key_Value_List* entry = ((Key_Value_List*)(entry_000));

          entry->insert_at(key, info);
        }
      }
      else {
        { Key_Value_List* variableinfo = new_key_value_list();

          variableinfo->insert_at(key, info);
          variableinfo->insert_at(KWD_WALK_TYPE, entry);
          insert_at_variable_table(oLOCALVARIABLETYPETABLEo.get(), variable, variableinfo);
        }
      }
    }
  }
}

Object* get_local_variable_info(Symbol* variable, Keyword* key) {
  { Standard_Object* entry = (((boolean)(oLOCALVARIABLETYPETABLEo.get())) ? ((Standard_Object*)(lookup_variable_table(oLOCALVARIABLETYPETABLEo.get(), variable))) : ((Standard_Object*)(NULL)));

    if (((boolean)(entry))) {
      if (subtype_ofP(safe_primary_type(entry), SGT_WALK_STELLA_KEY_VALUE_LIST)) {
        { Standard_Object* entry_000 = entry;
          Key_Value_List* entry = ((Key_Value_List*)(entry_000));

          return (entry->lookup(key));
        }
      }
      else {
        if (key == KWD_WALK_TYPE) {
          return (entry);
        }
      }
    }
    return (NULL);
  }
}

void register_reference_to_global_variable(Symbol* variablename) {
  if (((boolean)(oCURRENTTRANSLATIONUNITo.get()))) {
    { Global_Variable* global = variablename->lookup_global_variable();
      List* referencedglobals = oCURRENTTRANSLATIONUNITo.get()->referenced_globals;

      if (((boolean)(global)) &&
          (!referencedglobals->memberP(global))) {
        referencedglobals->insert(global);
      }
    }
  }
}

Symbol* local_gensym(char* prefix) {
  if (!((boolean)(oLOCALGENSYMTABLEo.get()))) {
    return (gensym(prefix));
  }
  { Integer_Wrapper* prefixcounter = ((Integer_Wrapper*)(oLOCALGENSYMTABLEo.get()->lookup(wrap_string(prefix))));

    if (!((boolean)(prefixcounter))) {
      prefixcounter = wrap_integer(0);
    }
    else {
      prefixcounter = wrap_integer(prefixcounter->wrapper_value + 1);
    }
    oLOCALGENSYMTABLEo.get()->insert_at(wrap_string(prefix), prefixcounter);
    return (intern_transient_symbol(yield_gensym_name(prefix, prefixcounter->wrapper_value)));
  }
}

Symbol* yield_local_variable_alias(Symbol* variable, int level) {
  if (!(level >= 1)) {
    cerr << "Safety violation: " << "INTERNAL ERROR: Variable level must be >= 1";
  }
  if (!((boolean)(oLOCALGENSYMTABLEo.get()))) {
    return (local_gensym(variable->symbol_name));
  }
  { Cons* aliases = ((Cons*)(oLOCALGENSYMTABLEo.get()->lookup(variable)));
    Symbol* alias = NULL;

    if (!((boolean)(aliases))) {
      aliases = NIL;
    }
    if (aliases->length() < level) {
      alias = local_gensym(variable->symbol_name);
      oLOCALGENSYMTABLEo.get()->insert_at(variable, aliases->concatenate(cons(alias, NIL), 0));
      return (alias);
    }
    else {
      return (((Symbol*)(aliases->nth(level - 1))));
    }
  }
}

boolean rename_shadowing_local_variablesP() {
  return (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA);
}

void maybe_rename_local_variable(Symbol* variable) {
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) {
    { Kv_Cons* cursor = oLOCALVARIABLETYPETABLEo.get()->the_kv_list;
      int level = -1;

      while (((boolean)(cursor))) {
        if (variable_eqlP(((Symbol*)(cursor->key)), variable)) {
          level = level + 1;
        }
        cursor = cursor->rest;
      }
      if (level >= 1) {
        set_local_variable_info(variable, KWD_WALK_ALIAS, yield_local_variable_alias(variable, level));
      }
    }
  }
}

Symbol* true_variable_name(Symbol* variable) {
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) {
    { Object* alias = get_local_variable_info(variable, KWD_WALK_ALIAS);

      if (((boolean)(alias))) {
        return (((Symbol*)(alias)));
      }
    }
  }
  return (variable);
}

boolean suppress_warningsP() {
  return (oCURRENT_ENVIRONMENT_FEATURESo->memberP(KWD_WALK_SUPPRESS_WARNINGS));
}

boolean bad_argument_rangeP(Cons* tree, int minarity, int maxarity) {
  { int argnum = tree->length() - 1;

    if (argnum < minarity) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Missing arguments in the expression: " << endl << " " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
        }
      }
      return (TRUE);
    }
    else if ((maxarity != NULL_INTEGER) &&
        (argnum > maxarity)) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Too many arguments in the expression: " << endl << " " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
        }
      }
      return (TRUE);
    }
    else {
      return (FALSE);
    }
  }
}

boolean bad_argument_countP(Cons* tree, int arity) {
  return (bad_argument_rangeP(tree, arity, arity));
}

Cons* yield_illegal_statement_tree() {
  { 
    BIND_STELLA_SPECIAL(oLOCALVARIABLETYPETABLEo, Key_Value_List*, new_key_value_list());
    return (((Cons*)(walk_without_type_tree(listO(5, SYM_WALK_STELLA_ERROR, wrap_string("Attempt to execute illegal statement flagged by the translator"), SYM_WALK_STELLA_EOL, wrap_string("    but not yet fixed."), NIL)))));
  }
}

Object* yield_illegal_expression_tree() {
  return (SYM_WALK_STELLA_ILLEGAL_EXPRESSION_FLAGGED_BY_THE_TRANSLATOR);
}

Cons* walk_dont_call_me_tree(Object* tree, Standard_Object* returntype, Standard_Object*& _return1) {
  _return1 = returntype;
  return (listO(3, SYM_WALK_STELLA_BAD_SYS, tree, cons(type_spec_to_base_type(returntype), NIL)));
}

boolean illegal_treeP(Object* tree) {
  return (consP(tree) &&
      (((Cons*)(tree))->value == SYM_WALK_STELLA_BAD_SYS));
}

boolean native_class_method_inliningP() {
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
    return (FALSE);
  }
  else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) ||
      ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
       ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE) ||
        (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_IDL)))) {
    return (TRUE);
  }
  else {
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
}

Object* prognify(Cons* trees) {
  switch (trees->length()) {
    case 0: 
      return (NIL);
    case 1: 
      { Object* tree = trees->value;

        trees->first_setter(NULL);
        return (tree);
      }
    break;
    default:
      return (cons(SYM_WALK_STELLA_PROGN, trees->concatenate(NIL, 0)));
  }
}

Object* optimize_progn(Cons* progn_tree) {
  switch (progn_tree->rest->length()) {
    case 0: 
      return (NIL);
    case 1: 
      { Object* tree = progn_tree->rest->value;

        progn_tree->second_setter(NULL);
        return (tree);
      }
    break;
    default:
      return (progn_tree);
  }
}

Cons* tree_to_trees(Object* tree) {
  if (safe_primary_type(tree) == SGT_WALK_STELLA_CONS) {
    { Object* tree_000 = tree;
      Cons* tree = ((Cons*)(tree_000));

      if (tree->value == SYM_WALK_STELLA_PROGN) {
        return (tree->rest);
      }
    }
  }
  else {
  }
  return (cons(tree, NIL));
}

// Table of pairs used by `wrapper-value-type' and
// `type-to-wrapped-type'.
Cons* oWRAPPED_TYPE_TABLEo = NULL;

Surrogate* wrapped_type_to_type(Surrogate* self) {
  // Return the unwrapped type for the wrapped type `self',
  // or `self' if it is not a wrapped type.
  { Cons* entry = NULL;
    Cons* iter_000 = oWRAPPED_TYPE_TABLEo;

    while (!(iter_000 == NIL)) {
      entry = ((Cons*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (entry->value == self) {
        return (((Surrogate*)(entry->rest->value)));
      }
    }
  }
  return (self);
}

Surrogate* wrapper_value_type(Wrapper* self) {
  // Return the type of the value stored in the wrapper `self'.
  { Surrogate* wrappertype = self->primary_type();
    Surrogate* valuetype = wrapped_type_to_type(wrappertype);

    if (valuetype == wrappertype) {
      throw *new_stella_exception("Missing *wrapped-type-table* entry");
    }
    else {
      return (valuetype);
    }
  }
}

Surrogate* Surrogate::type_to_wrapped_type() {
  // Return the wrapped type for the type `self',
  // or 'self' if it is not a bare literal type.
  { Surrogate* self = this;

    if (!((boolean)(((Class*)(self->surrogate_value))))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Can't find a class named " << "`" << de_uglify_parse_tree(self) << "'" << "." << endl;
        }
      }
      return (SGT_WALK_STELLA_UNKNOWN);
    }
    if (subtype_ofP(self, SGT_WALK_STELLA_OBJECT)) {
      return (self);
    }
    { Cons* entry = NULL;
      Cons* iter_000 = oWRAPPED_TYPE_TABLEo;

      while (!(iter_000 == NIL)) {
        entry = ((Cons*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (entry->rest->value == self) {
          return (((Surrogate*)(entry->value)));
        }
      }
    }
    { Surrogate* realtype = ((Class*)(self->surrogate_value))->class_type;

      if (!(self == realtype)) {
        { Cons* entry = NULL;
          Cons* iter_001 = oWRAPPED_TYPE_TABLEo;

          while (!(iter_001 == NIL)) {
            entry = ((Cons*)(iter_001->value));
            iter_001 = iter_001->rest;
            if (entry->rest->value == realtype) {
              return (((Surrogate*)(entry->value)));
            }
          }
        }
      }
    }
    return (NULL);
  }
}

// Table of triples used by `lookup-coersion-method' to
// locate a coersion method.
Cons* oCOERSION_TABLEo = NULL;

Object* lookup_coersion_method(Surrogate* sourcetype, Surrogate* targettype, Object* expression) {
  expression = expression;
  sourcetype = canonical_type(sourcetype);
  { Cons* triple = NULL;
    Cons* iter_000 = oCOERSION_TABLEo;

    while (!(iter_000 == NIL)) {
      triple = ((Cons*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (subtype_ofP(sourcetype, ((Surrogate*)(triple->value))) &&
          subtype_ofP(((Surrogate*)(triple->rest->value)), targettype)) {
        return (triple->rest->rest->value);
      }
    }
  }
  return (NULL);
}

Object* apply_coercion_method(Object* expression, Surrogate* sourcetype, Surrogate* targettype, Object* coercionmethod, Standard_Object*& _return1) {
  if (!((boolean)(coercionmethod))) {
    coercionmethod = lookup_coersion_method(sourcetype, targettype, expression);
  }
  if (!((boolean)(coercionmethod))) {
    _return1 = sourcetype;
    return (expression);
  }
  if (coercionmethod == SYM_WALK_STELLA_IDENTITY) {
    _return1 = targettype;
    return (expression);
  }
  { Surrogate* test_value_000 = safe_primary_type(coercionmethod);

    if (subtype_of_symbolP(test_value_000)) {
      { Object* coercionmethod_000 = coercionmethod;
        Symbol* coercionmethod = ((Symbol*)(coercionmethod_000));

        return (walk_expression_tree(cons(coercionmethod, cons(expression, NIL)), targettype, coercionmethod, FALSE, _return1));
      }
    }
    else if (test_value_000 == SGT_WALK_STELLA_CONS) {
      { Object* coercionmethod_001 = coercionmethod;
        Cons* coercionmethod = ((Cons*)(coercionmethod_001));

        return (walk_expression_tree(copy_cons_tree(substitute_cons_tree(((Cons*)(copy_cons_tree(coercionmethod))), expression, SYM_WALK_STELLA_lXg)), targettype, ((Symbol*)(coercionmethod->value)), FALSE, _return1));
      }
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

boolean translating_codeP() {
  return ((oTRANSLATIONPHASEo.get() == KWD_WALK_WALK) ||
      (oTRANSLATIONPHASEo.get() == KWD_WALK_TRANSLATE));
}

Surrogate* verify_type(Surrogate* self) {
  if (((boolean)(self->surrogate_value)) &&
      (!stella_classP(self->surrogate_value))) {
    *(STANDARD_WARNING->native_stream) << "Warning: " << "Illegal object " << "`" << self->surrogate_value << "'" << " found" << endl << "   where STELLA class expected" << endl << endl;
    self->surrogate_value = ((Class*)(SGT_WALK_STELLA_UNKNOWN->surrogate_value));
  }
  if (translating_codeP() &&
      ((!((boolean)(((Class*)(self->surrogate_value))))) ||
       (((Class*)(self->surrogate_value)) == ((Class*)(SGT_WALK_STELLA_UNKNOWN->surrogate_value))))) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_warning();
      if (!(suppress_warningsP())) {
        print_error_context(">> WARNING: ", STANDARD_OUTPUT);
        cout << endl << " " << "Reference to undefined class " << "`" << self->symbol_name << "'" << "." << endl;
      }
    }
    self->surrogate_value = ((Class*)(SGT_WALK_STELLA_UNKNOWN->surrogate_value));
  }
  return (self);
}

boolean safe_subtype_ofP(Surrogate* subtype, Surrogate* supertype) {
  verify_type(subtype);
  verify_type(supertype);
  return (subtype_ofP(subtype, supertype));
}

void warn_about_unknown_source_type(Object* tree) {
  if (warn_about_undefined_methodsP()) {
    { Surrogate* test_value_000 = safe_primary_type(tree);

      if (subtype_of_symbolP(test_value_000)) {
        { Object* tree_000 = tree;
          Symbol* tree = ((Symbol*)(tree_000));

          return;
        }
      }
      else if (test_value_000 == SGT_WALK_STELLA_CONS) {
        { Object* tree_001 = tree;
          Cons* tree = ((Cons*)(tree_001));

          { Generalized_Symbol* test_value_001 = ((Generalized_Symbol*)(tree->value));

            if ((test_value_001 == SYM_WALK_STELLA_VERBATIM) ||
                (test_value_001 == SYM_WALK_STELLA_VRLET)) {
              return;
            }
            else if (test_value_001 == SYM_WALK_STELLA_SYS_CALL_METHOD) {
              if (symbol_common_lispP(((Symbol*)(tree->rest->rest->value))) ||
                  (!((boolean)(lookup_slot(((Class*)(((Surrogate*)(tree->rest->value))->surrogate_value)), ((Symbol*)(tree->rest->rest->value))))))) {
                return;
              }
            }
            else if (test_value_001 == SYM_WALK_STELLA_SYS_CALL_FUNCTION) {
              if (lookup_function(((Symbol*)(tree->rest->value)))->slot_base_type == SGT_WALK_STELLA_UNKNOWN) {
                return;
              }
            }
            else if ((test_value_001 == SYM_WALK_STELLA_TYPED_SYS) ||
                (test_value_001 == SYM_WALK_STELLA_SYS_INLINE_CALL)) {
              warn_about_unknown_source_type(tree->rest->value);
              return;
            }
            else {
            }
          }
        }
      }
      else {
        if (!((boolean)(tree))) {
          return;
        }
      }
    }
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_error();
      if (!(suppress_warningsP())) {
        print_error_context(">> ERROR: ", STANDARD_ERROR);
        *(STANDARD_ERROR->native_stream) << endl << " " << "The expression" << endl << "   " << "`" << de_uglify_parse_tree(tree) << "'" << endl << " has UNKNOWN type" << "." << endl;
      }
    }
  }
}

Object* coerce_a_tree(Object* tree, Standard_Object* sourcetype, Standard_Object* targettype, Standard_Object*& _return1) {
  { Object* otree = NULL;
    Standard_Object* otype = NULL;
    boolean coercibleP = FALSE;

    otree = help_coerce_a_tree(tree, sourcetype, targettype, FALSE, otype, coercibleP);
    if (!(coercibleP)) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Type conflict: the expression:" << endl << "   " << "`" << de_uglify_parse_tree(tree) << "'" << endl << " has type " << "`" << yield_type_spec_tree(sourcetype) << "'" << " but should have type " << "`" << yield_type_spec_tree(targettype) << "'" << "." << endl;
        }
      }
    }
    _return1 = otype;
    return (otree);
  }
}

Object* help_coerce_a_tree(Object* tree, Standard_Object* sourcetype, Standard_Object* targettype, boolean testonlyP, Standard_Object*& _return1, boolean& _return2) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    { Object* coersionfunction = NULL;
      Surrogate* realsourcetype = target_language_type(tree);
      Surrogate* sourcebasetype = type_spec_to_base_type(sourcetype);
      Surrogate* targetbasetype = type_spec_to_base_type(targettype);

      if (sourcetype == SGT_WALK_STELLA_UNKNOWN) {
        warn_about_unknown_source_type(tree);
        if ((tree == SYM_WALK_STELLA_NULL) &&
            ((boolean)(targetbasetype))) {
          _return1 = targettype;
          _return2 = TRUE;
          return (type_to_walked_null_value_tree(targetbasetype));
        }
        else {
          _return1 = targettype;
          _return2 = TRUE;
          return (tree);
        }
      }
      if (((boolean)(realsourcetype))) {
        set_target_language_type(((Cons*)(tree)), NULL, dummy1);
        tree = coerce_a_tree(tree, realsourcetype, sourcetype, dummy2);
      }
      if (targettype == SGT_WALK_STELLA_UNKNOWN) {
        _return1 = sourcetype;
        _return2 = TRUE;
        return (tree);
      }
      if (!((boolean)(((Class*)(sourcebasetype->surrogate_value))))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "No class defined for type: " << "`" << de_uglify_parse_tree(sourcebasetype) << "'" << "." << endl;
          }
        }
        _return1 = targettype;
        _return2 = TRUE;
        return (tree);
      }
      if (((boolean)(targetbasetype)) &&
          (!((boolean)(((Class*)(targetbasetype->surrogate_value)))))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "No class defined for type: " << "`" << de_uglify_parse_tree(targetbasetype) << "'" << "." << endl;
          }
        }
        _return1 = sourcetype;
        _return2 = TRUE;
        return (tree);
      }
      if (sub_type_spec_ofP(sourcetype, targettype)) {
        _return1 = sourcetype;
        _return2 = TRUE;
        return (tree);
      }
      if (((boolean)(targetbasetype)) &&
          (compatible_parameter_typesP(sourcetype, targettype) ||
           ((!array_type_specifierP(sourcetype)) &&
            (subtype_ofP(targetbasetype, sourcebasetype) &&
             compatible_parameter_typesP(targettype, sourcetype))))) {
        { boolean test_value_000 = FALSE;

          test_value_000 = typeP(sourcetype);
          if (test_value_000) {
            { 
              coersionfunction = lookup_coersion_method(((Surrogate*)(sourcetype)), targetbasetype, tree);
              test_value_000 = ((boolean)(coersionfunction));
            }
          }
          if (test_value_000) {
            if (!(testonlyP)) {
              tree = apply_coercion_method(tree, ((Surrogate*)(sourcetype)), targetbasetype, coersionfunction, targettype);
            }
            _return1 = targettype;
            _return2 = TRUE;
            return (tree);
          }
        }
        if (subtype_ofP(targetbasetype, sourcebasetype)) {
          if (testonlyP) {
            _return1 = targettype;
            _return2 = TRUE;
            return (tree);
          }
          _return1 = targettype;
          _return2 = TRUE;
          return (walk_without_type_tree(cons((((boolean)(realsourcetype)) ? SYM_WALK_STELLA_SAFE_CAST : SYM_WALK_STELLA_CAST), cons(tree, cons(targettype, NIL)))));
        }
        if (subtype_ofP(targetbasetype, SGT_WALK_STELLA_NON_OBJECT) &&
            (typeP(sourcetype) &&
             subtype_ofP(SGT_WALK_STELLA_WRAPPER, ((Surrogate*)(sourcetype))))) {
          { Surrogate* wrappertype = targetbasetype->type_to_wrapped_type();

            if (((boolean)(wrappertype))) {
              if (testonlyP) {
                _return1 = targettype;
                _return2 = TRUE;
                return (tree);
              }
              coersionfunction = lookup_coersion_method(wrappertype, targetbasetype, tree);
              if (((boolean)(coersionfunction))) {
                tree = apply_coercion_method(listO(3, SYM_WALK_STELLA_CAST, tree, cons(wrappertype, NIL)), ((Surrogate*)(sourcetype)), targetbasetype, coersionfunction, targettype);
                _return1 = targettype;
                _return2 = TRUE;
                return (tree);
              }
            }
          }
        }
      }
      _return1 = sourcetype;
      _return2 = FALSE;
      return (tree);
    }
  }
}

boolean coercibleP(Object* tree, Standard_Object* sourcetype, Standard_Object* targettype) {
  { Object* otree = NULL;
    Standard_Object* otype = NULL;
    boolean coercibleP = FALSE;

    otree = help_coerce_a_tree(tree, sourcetype, targettype, TRUE, otype, coercibleP);
    return (coercibleP);
  }
}

boolean walking_expressionP() {
  return (!(oTARGETTYPEo.get() == SGT_WALK_STELLA_VOID));
}

Object* walk_expression_tree(Object* exptree, Standard_Object* targettype, Symbol* operatorname, boolean vrletisokP, Standard_Object*& _return1) {
  { 
    BIND_STELLA_SPECIAL(oTARGETTYPEo, Standard_Object*, (((boolean)(targettype)) ? ((Standard_Object*)(targettype)) : SGT_WALK_STELLA_UNKNOWN));
    { Object* oexp = NULL;
      Standard_Object* sourcetype = NULL;

      oexp = walk_a_tree(exptree, sourcetype);
      if (vrlet_expressionP(oexp)) {
        if (!(vrletisokP)) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal procedural expression as argument to " << "`" << de_uglify_parse_tree(operatorname) << "'" << ":" << endl << "   " << "`" << de_uglify_parse_tree(exptree) << "'" << "." << endl;
            }
          }
        }
        _return1 = sourcetype;
        return (oexp);
      }
      if (voidP(sourcetype) &&
          (!voidP(targettype))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal argument to " << "`" << de_uglify_parse_tree(operatorname) << "'" << ":" << endl << "   " << "`" << de_uglify_parse_tree(oexp) << "'" << endl << "The argument must be an expression, not a statement" << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(oexp, sourcetype, _return1));
      }
      if (!((boolean)(sourcetype))) {
        sourcetype = SGT_WALK_STELLA_UNKNOWN;
      }
      if (!((boolean)(targettype))) {
        targettype = sourcetype;
      }
      return (coerce_a_tree(oexp, sourcetype, targettype, _return1));
    }
  }
}

Object* walk_without_type_tree(Object* tree) {
  { Object* otree = NULL;
    Standard_Object* unusedtypespec = NULL;

    otree = walk_a_tree(tree, unusedtypespec);
    return (otree);
  }
}

Cons* walk_list_of_trees(Cons* trees) {
  { Cons* cursor = trees;

    while (!(cursor == NIL)) {
      cursor->value = walk_statement(cursor->value, FALSE);
      cursor = cursor->rest;
    }
    return (trees);
  }
}

Cons* walk_list_of_statements(Cons* trees) {
  { Cons* otrees = NIL;

    { Object* otree = NULL;
      Cons* iter_000 = walk_list_of_trees(trees);
      Cons* collect_000 = NULL;

      while (!(iter_000 == NIL)) {
        otree = iter_000->value;
        iter_000 = iter_000->rest;
        if (consP(otree)) {
          if (!((boolean)(collect_000))) {
            {
              collect_000 = cons(otree, NIL);
              if (otrees == NIL) {
                otrees = collect_000;
              }
              else {
                add_cons_to_end_of_cons_list(otrees, collect_000);
              }
            }
          }
          else {
            {
              collect_000->rest = cons(otree, NIL);
              collect_000 = collect_000->rest;
            }
          }
        }
      }
    }
    return (otrees);
  }
}

Object* walk_statement(Object* tree, boolean warnonatomP) {
  { Object* otree = NULL;
    Standard_Object* unusedtypespec = NULL;

    otree = walk_a_tree(tree, unusedtypespec);
    if (safe_primary_type(otree) == SGT_WALK_STELLA_CONS) {
      { Object* otree_000 = otree;
        Cons* otree = ((Cons*)(otree_000));

        if (otree->value == SYM_WALK_STELLA_VRLET) {
          otree = eliminate_vrlet_statement(otree);
        }
      }
    }
    else {
      if (warnonatomP &&
          (((boolean)(otree)) &&
           (!(otree == SYM_WALK_STELLA_NULL)))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_warning();
          if (!(suppress_warningsP())) {
            print_error_context(">> WARNING: ", STANDARD_OUTPUT);
            cout << endl << " " << "Expression found where statement expected: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
          }
        }
      }
    }
    return (otree);
  }
}

Cons* eliminate_vrlet_statement(Cons* tree) {
  { Standard_Object* dummy1;

    tree->first_setter(SYM_WALK_STELLA_LET);
    return (((Cons*)(walk_a_tree(tree, dummy1))));
  }
}

Cons* sys_tree(Object* tree, Standard_Object* typespec, Standard_Object*& _return1) {
  if (safe_primary_type(tree) == SGT_WALK_STELLA_CONS) {
    { Object* tree_000 = tree;
      Cons* tree = ((Cons*)(tree_000));

      { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

        if (test_value_000 == SYM_WALK_STELLA_VOID_SYS) {
          if (voidP(typespec)) {
            _return1 = typespec;
            return (tree);
          }
        }
        else if (test_value_000 == SYM_WALK_STELLA_TYPED_SYS) {
          if (tree->rest->rest->value == typespec) {
            _return1 = typespec;
            return (tree);
          }
        }
        else if (test_value_000 == SYM_WALK_STELLA_VRLET) {
          _return1 = typespec;
          return (tree);
        }
        else {
        }
      }
    }
  }
  else {
  }
  if (voidP(typespec)) {
    _return1 = SGT_WALK_STELLA_VOID;
    return (listO(3, SYM_WALK_STELLA_VOID_SYS, tree, NIL));
  }
  else {
    _return1 = typespec;
    return (listO(3, SYM_WALK_STELLA_TYPED_SYS, tree, cons(typespec, NIL)));
  }
}

Standard_Object* walked_expression_type(Object* tree) {
  { Surrogate* test_value_000 = safe_primary_type(tree);

    if (test_value_000 == SGT_WALK_STELLA_CONS) {
      { Object* tree_000 = tree;
        Cons* tree = ((Cons*)(tree_000));

        if (tree->value == SYM_WALK_STELLA_TYPED_SYS) {
          return (((Standard_Object*)(tree->rest->rest->value)));
        }
      }
    }
    else if (subtype_of_wrapperP(test_value_000)) {
      { Object* tree_001 = tree;
        Wrapper* tree = ((Wrapper*)(tree_001));

        return (wrapped_type_to_type(tree->primary_type()));
      }
    }
    else {
    }
  }
  return (SGT_WALK_STELLA_UNKNOWN);
}

boolean need_identical_method_signaturesP() {
  if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
      ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE) ||
       (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA))) {
    return (TRUE);
  }
  else {
    return (FALSE);
  }
}

Standard_Object* compute_most_general_return_type(Method_Slot* method, Standard_Object* returntype) {
  { Standard_Object* mostgeneralreturntype = returntype;
    Method_Slot* mostgeneralmethod = method;

    while (((boolean)(mostgeneralmethod->slot_direct_equivalent))) {
      mostgeneralmethod = ((Method_Slot*)(mostgeneralmethod->slot_direct_equivalent));
      if (!(mostgeneralmethod->slot_base_type == SGT_WALK_STELLA_UNKNOWN)) {
        if (array_type_specifierP(mostgeneralreturntype)) {
          mostgeneralreturntype = mostgeneralmethod->compute_return_type_spec(mostgeneralmethod->slot_owner);
        }
        else {
          mostgeneralreturntype = mostgeneralmethod->slot_base_type;
        }
      }
    }
    return (mostgeneralreturntype);
  }
}

Standard_Object* compute_real_slot_type(Storage_Slot* slot, Standard_Object* firstargtype, Standard_Object* returntype) {
  { Standard_Object* slottype = returntype;
    Storage_Slot* canonicalslot = canonical_slot(slot);
    Class* nativeslothome = NULL;

    if (!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP)) {
      if (native_slotP(canonicalslot)) {
        nativeslothome = native_slot_home(slot, type_spec_to_class(firstargtype));
      }
      if (((boolean)(nativeslothome))) {
        slottype = type_spec_to_base_type(lookup_slot(nativeslothome, canonicalslot->slot_name)->compute_return_type_spec(nativeslothome->class_type));
      }
      else {
        slottype = type_spec_to_base_type(slot->compute_return_type_spec(type_spec_to_base_type(firstargtype)));
      }
      if (slottype == type_spec_to_base_type(returntype)) {
        slottype = returntype;
      }
    }
    if (!(slot == canonicalslot)) {
      slottype = canonicalslot->slot_base_type;
    }
    return (slottype);
  }
}

Object* sys_tree_if_needed(Slot* slot, Object* tree, Standard_Object* firstargtype, Standard_Object* returntype, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;

    if (((boolean)(slot)) &&
        (!procedural_expressionP(tree))) {
      { Surrogate* test_value_000 = safe_primary_type(slot);

        if (subtype_of_method_slotP(test_value_000)) {
          { Slot* slot_000 = slot;
            Method_Slot* slot = ((Method_Slot*)(slot_000));

            if ((!slot->method_functionP) &&
                need_identical_method_signaturesP()) {
              { Standard_Object* mostgeneralreturntype = compute_most_general_return_type(slot, returntype);

                if (!(returntype == mostgeneralreturntype)) {
                  return (set_target_language_type(sys_tree(tree, returntype, dummy1), ((Surrogate*)(mostgeneralreturntype)), _return1));
                }
              }
            }
          }
        }
        else if (subtype_of_storage_slotP(test_value_000)) {
          { Slot* slot_001 = slot;
            Storage_Slot* slot = ((Storage_Slot*)(slot_001));

            { Standard_Object* realslottype = compute_real_slot_type(slot, firstargtype, returntype);
              Slot* realslot = NULL;

              if (!(returntype == realslottype)) {
                if ((realslottype == SGT_WALK_STELLA_UNKNOWN) &&
                    ((!(slot == (realslot = slot_from_expression_tree(((Cons*)(tree)))))) &&
                     method_slotP(realslot))) {
                  return (sys_tree_if_needed(realslot, tree, firstargtype, returntype, _return1));
                }
                return (set_target_language_type(sys_tree(tree, returntype, dummy2), ((Surrogate*)(realslottype)), _return1));
              }
            }
          }
        }
        else {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
      }
      { Compound_Type_Specifier* slottype = ((Compound_Type_Specifier*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL)));

        { boolean test_value_001 = FALSE;

          test_value_001 = ((boolean)(slottype));
          if (test_value_001) {
            if (anchored_type_specifierP(slottype)) {
              test_value_001 = TRUE;
            }
            else {
              { boolean foundP_000 = FALSE;

                { Standard_Object* ts = NULL;
                  Cons* iter_000 = ((Parametric_Type_Specifier*)(slottype))->specifier_parameter_types->the_cons_list;

                  while (!(iter_000 == NIL)) {
                    ts = ((Standard_Object*)(iter_000->value));
                    iter_000 = iter_000->rest;
                    if (anchored_type_specifierP(ts)) {
                      foundP_000 = TRUE;
                      break;
                    }
                  }
                }
                test_value_001 = foundP_000;
              }
            }
          }
          if (test_value_001) {
            if (anchored_type_specifierP(slottype) &&
                (!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP))) {
              return (set_target_language_type(sys_tree(tree, returntype, dummy3), ((slot->slot_base_type == SGT_WALK_STELLA_UNKNOWN) ? type_spec_to_base_type(returntype) : slot->slot_base_type), _return1));
            }
            else {
              return (sys_tree(tree, returntype, _return1));
            }
          }
        }
      }
    }
    _return1 = returntype;
    return (tree);
  }
}

Surrogate* target_language_type(Object* tree) {
  if (safe_primary_type(tree) == SGT_WALK_STELLA_CONS) {
    { Object* tree_000 = tree;
      Cons* tree = ((Cons*)(tree_000));

      if ((tree->value == SYM_WALK_STELLA_TYPED_SYS) &&
          (tree->length() == 4)) {
        return (((Surrogate*)(tree->fourth())));
      }
    }
  }
  else {
  }
  return (NULL);
}

Cons* set_target_language_type(Cons* systree, Surrogate* type, Standard_Object*& _return1) {
  if (systree->length() == 4) {
    systree->fourth_setter(type);
  }
  else {
    systree->rest->rest->rest = cons_list(1, type);
  }
  _return1 = ((Standard_Object*)(systree->rest->rest->value));
  return (systree);
}

Object* walk_a_tree(Object* tree, Standard_Object*& _return1) {
  return (help_walk_a_tree(tree, _return1));
}

Object* help_walk_a_tree(Object* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    if (!((boolean)(tree))) {
      _return1 = SGT_WALK_STELLA_UNKNOWN;
      return (NULL);
    }
    else {
      if (safe_primary_type(tree) == SGT_WALK_STELLA_CONS) {
        { Object* tree_000 = tree;
          Cons* tree = ((Cons*)(tree_000));

          if (tree == NIL) {
            _return1 = SGT_WALK_STELLA_CONS;
            return (NIL);
          }
          if (consP(tree->value)) {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Unimplemented or illegal parse tree: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
              }
            }
            _return1 = SGT_WALK_STELLA_VOID;
            return (tree);
          }
          if (!(consP(tree->rest))) {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal dotted pair or list*" << "." << endl;
              }
            }
            _return1 = SGT_WALK_STELLA_VOID;
            return (tree);
          }
          { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

            if (test_value_000 == SYM_WALK_STELLA_VOID_SYS) {
              _return1 = SGT_WALK_STELLA_VOID;
              return (tree);
            }
            else if (test_value_000 == SYM_WALK_STELLA_TYPED_SYS) {
              _return1 = ((Standard_Object*)(tree->rest->rest->value));
              return (tree);
            }
            else if (test_value_000 == SYM_WALK_STELLA_SYS_FOREACH) {
              _return1 = SGT_WALK_STELLA_VOID;
              return (tree);
            }
            else if (test_value_000 == SYM_WALK_STELLA_BAD_SYS) {
              _return1 = ((Standard_Object*)(tree->rest->rest->value));
              return (tree);
            }
            else if (test_value_000 == SYM_WALK_STELLA_SETQ) {
              return (walk_setq_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_SETF) {
              return (walk_setf_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_SLOT_VALUE) {
              return (walk_slot_value_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_SLOT_VALUE_SETTER) {
              return (walk_slot_value_setter_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_AREF) ||
                (test_value_000 == SYM_WALK_STELLA_AREF_SETTER)) {
              return (walk_aref_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_PROGN) {
              tree->rest = walk_list_of_statements(tree->rest);
              return (sys_tree(tree, SGT_WALK_STELLA_VOID, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_WITH_PROCESS_LOCK) {
              walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_PROCESS_LOCK_OBJECT, ((Symbol*)(tree->value)), TRUE, dummy1);
              tree->rest->rest = walk_list_of_statements(tree->rest->rest);
              return (sys_tree(tree, SGT_WALK_STELLA_VOID, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_LOOP) {
              return (walk_loop_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_LET) {
              return (walk_let_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_VRLET) {
              _return1 = SGT_WALK_STELLA_UNKNOWN;
              return (tree);
            }
            else if (test_value_000 == SYM_WALK_STELLA_VALUES) {
              return (walk_values_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_SPECIAL) ||
                (test_value_000 == SYM_WALK_STELLA_SYS_SPECIAL)) {
              return (walk_special_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_CAST) ||
                (test_value_000 == SYM_WALK_STELLA_SAFE_CAST)) {
              return (walk_cast_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_RETURN) {
              return (walk_return_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_MV_SETQ) {
              return (walk_mv_setq_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_MV_BIND) {
              return (walk_mv_bind_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_BREAK) ||
                (test_value_000 == SYM_WALK_STELLA_CONTINUE)) {
              return (walk_loop_exit_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_WHILE) {
              return (walk_while_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_FOREACH) ||
                ((test_value_000 == SYM_WALK_STELLA_EXISTS) ||
                 ((test_value_000 == SYM_WALK_STELLA_FORALL) ||
                  ((test_value_000 == SYM_WALK_STELLA_SOME) ||
                   ((test_value_000 == SYM_WALK_STELLA_SETOF) ||
                    ((test_value_000 == SYM_WALK_STELLA_SET_OF) ||
                     (test_value_000 == SYM_WALK_STELLA_SELECT))))))) {
              return (walk_foreach_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_IF) ||
                ((test_value_000 == SYM_WALK_STELLA_WHEN) ||
                 (test_value_000 == SYM_WALK_STELLA_UNLESS))) {
              return (walk_conditional_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_CHOOSE) {
              return (walk_choose_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_AND) ||
                ((test_value_000 == SYM_WALK_STELLA_OR) ||
                 (test_value_000 == SYM_WALK_STELLA_NOT))) {
              return (walk_boolean_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_COND) {
              return (walk_cond_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_CASE) {
              return (walk_case_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_TYPECASE) {
              return (walk_typecase_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_EQp) ||
                ((test_value_000 == SYM_WALK_STELLA_EQLp) ||
                 (test_value_000 == SYM_WALK_STELLA_e))) {
              return (walk_equality_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_i) ||
                ((test_value_000 == SYM_WALK_STELLA__) ||
                 ((test_value_000 == SYM_WALK_STELLA_o) ||
                  (test_value_000 == SYM_WALK_STELLA_s)))) {
              { Object* value_000 = NULL;
                Surrogate* value_001 = NULL;

                value_000 = walk_arithmetic_tree(tree, value_001);
                _return1 = value_001;
                return (value_000);
              }
            }
            else if (test_value_000 == SYM_WALK_STELLA_QUOTE) {
              return (walk_quoted_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_BQUOTE) {
              return (walk_bquote_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_PRINT_STREAM) ||
                (test_value_000 == SYM_WALK_STELLA_PRINT_NATIVE_STREAM)) {
              return (walk_print_stream_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_NEW) ||
                ((test_value_000 == SYM_WALK_STELLA_ALLOCATE) ||
                 (test_value_000 == SYM_WALK_STELLA_LOCAL_NEW))) {
              return (walk_new_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_MAKE) {
              return (walk_make_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_FUNCALL) ||
                ((test_value_000 == SYM_WALK_STELLA_SYS_CALL_FUNCTION_CODE) ||
                 (test_value_000 == SYM_WALK_STELLA_SYS_CALL_METHOD_CODE))) {
              return (walk_funcall_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_THE_CODE) {
              return (walk_the_code_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_SIGNAL_EXCEPTION) {
              return (walk_signal_exception_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_EXCEPTION_CASE) {
              return (walk_exception_case_tree(tree, _return1));
            }
            else if ((test_value_000 == SYM_WALK_STELLA_UNWIND_PROTECT) ||
                (test_value_000 == SYM_WALK_STELLA_SYS_UNWIND_PROTECT)) {
              { Cons* value_002 = NULL;
                Surrogate* value_003 = NULL;

                value_002 = walk_unwind_protect_tree(tree, value_003);
                _return1 = value_003;
                return (value_002);
              }
            }
            else if (test_value_000 == SYM_WALK_STELLA_VERBATIM) {
              return (walk_verbatim_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_INLINE) {
              _return1 = SGT_WALK_STELLA_VOID;
              return (tree);
            }
            else if ((test_value_000 == SYM_WALK_STELLA_SYS_SLOT_VALUE) ||
                ((test_value_000 == SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER) ||
                 ((test_value_000 == SYM_WALK_STELLA_SYS_CALL_METHOD) ||
                  ((test_value_000 == SYM_WALK_STELLA_SYS_CALL_METHOD_SETTER) ||
                   (test_value_000 == SYM_WALK_STELLA_SYS_CALL_FUNCTION))))) {
              return (walk_sys_call_tree(tree, _return1));
            }
            else if (test_value_000 == SYM_WALK_STELLA_SYS_NEW) {
              _return1 = ((Standard_Object*)(tree->rest->value));
              return (tree);
            }
            else if ((test_value_000 == SYM_WALK_STELLA_DEFCLASS) ||
                ((test_value_000 == SYM_WALK_STELLA_DEFMETHOD) ||
                 ((test_value_000 == SYM_WALK_STELLA_DEFUN) ||
                  ((test_value_000 == SYM_WALK_STELLA_DEFMACRO) ||
                   ((test_value_000 == SYM_WALK_STELLA_STARTUP_TIME_PROGN) ||
                    ((test_value_000 == SYM_WALK_STELLA_DEFGLOBAL) ||
                     ((test_value_000 == SYM_WALK_STELLA_DEFSPECIAL) ||
                      (test_value_000 == SYM_WALK_STELLA_DEFCONSTANT)))))))) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal declaration -- must appear as a top-level statement;" << endl << "   not nested within other statements." << "." << endl;
                }
              }
              _return1 = SGT_WALK_STELLA_VOID;
              return (NULL);
            }
            else {
              if (symbolP(tree->value)) {
                return (walk_call_slot_tree(tree, _return1));
              }
              else {
                {
                  { 
                    BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                    signal_translation_error();
                    if (!(suppress_warningsP())) {
                      print_error_context(">> ERROR: ", STANDARD_ERROR);
                      *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal expression where method name expected:" << endl << "   " << "`" << tree->value << "'" << "." << endl;
                    }
                  }
                  return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
                }
              }
            }
          }
        }
      }
      else {
        return (tree->walk_atomic_tree(_return1));
      }
    }
  }
}

Cons* walk_a_cons_tree(Cons* tree, Standard_Object*& _return1) {
  { Object* otree = NULL;
    Standard_Object* otype = NULL;

    otree = walk_a_tree(tree, otype);
    _return1 = otype;
    return (((Cons*)(otree)));
  }
}

Hash_Table* oSYMBOL_REGISTRYo = NULL;

List* oSYMBOL_SETo = NULL;

// Name of file that is currently being translated.
// A NULL value indicates an incremental translation.
DEFINE_STELLA_SPECIAL(oCURRENTFILEo, char* , NULL);

boolean incremental_translationP() {
  return (oCURRENTFILEo.get() == NULL);
}

char* construct_symbol_constant_name(Generalized_Symbol* symbol) {
  { char* prefix = NULL;
    char* environmentname = "";

    if (((boolean)(((Module*)(symbol->home_context))))) {
      environmentname = string_concatenate(((Module*)(symbol->home_context))->module_name, "-", 0);
    }
    if (!(incremental_translationP() ||
        (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA))) {
      environmentname = string_concatenate(string_upcase(oCURRENTFILEo.get()), "-", 1, environmentname);
    }
    { Surrogate* test_value_000 = safe_primary_type(symbol);

      if (subtype_of_symbolP(test_value_000)) {
        { Generalized_Symbol* symbol_000 = symbol;
          Symbol* symbol = ((Symbol*)(symbol_000));

          prefix = "SYM-";
        }
      }
      else if (subtype_of_surrogateP(test_value_000)) {
        { Generalized_Symbol* symbol_001 = symbol;
          Surrogate* symbol = ((Surrogate*)(symbol_001));

          prefix = "SGT-";
        }
      }
      else if (subtype_of_keywordP(test_value_000)) {
        { Generalized_Symbol* symbol_002 = symbol;
          Keyword* symbol = ((Keyword*)(symbol_002));

          prefix = "KWD-";
        }
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
    }
    return (string_concatenate(prefix, environmentname, 1, symbol->symbol_name));
  }
}

Symbol* yield_symbol_constant_name(Generalized_Symbol* symbol) {
  symbol = ((Generalized_Symbol*)(symbol->permanentify()));
  if (subtype_of_keywordP(safe_primary_type(symbol))) {
    { Generalized_Symbol* symbol_000 = symbol;
      Keyword* symbol = ((Keyword*)(symbol_000));

      return (intern_symbol_in_module(string_upcase(construct_symbol_constant_name(symbol)), oMODULEo.get(), TRUE));
    }
  }
  else {
    return (intern_symbol_in_module(string_upcase(construct_symbol_constant_name(symbol)), oMODULEo.get(), TRUE));
  }
}

Symbol* create_startup_symbol(Generalized_Symbol* symbol) {
  { Module* symbolmodule = oMODULEo.get();
    Symbol* symbolconstant = yield_symbol_constant_name(symbol);
    char* symbolconstantname = symbolconstant->symbol_name;
    Object* symbolconstanttypetree = yield_type_spec_tree(symbol->primary_type());
    char* lisppackage = symbolmodule->lisp_package();
    Cons* tree = NULL;

    if (incremental_translationP()) {
    }
    tree = listO(3, SYM_WALK_STELLA_DEFGLOBAL, symbolconstant, cons(symbolconstanttypetree, listO(4, SYM_WALK_STELLA_NULL, KWD_WALK_PUBLICp, SYM_WALK_STELLA_TRUE, NIL)));
    walk_auxiliary_tree(tree);
    tree = listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_SYMBOLS, listO(3, SYM_WALK_STELLA_SETQ, symbolconstant, cons(listO(4, SYM_WALK_STELLA_SAFE_CAST, listO(3, SYM_WALK_STELLA_INTERN_RIGID_SYMBOL_WRT_MODULE, wrap_string(symbol->symbol_name), cons((((((Module*)(symbol->home_context)) == oMODULEo.get()) ||
        keywordP(symbol)) ? ((Standard_Object*)(SYM_WALK_STELLA_NULL)) : ((Standard_Object*)(listO(3, SYM_WALK_STELLA_GET_STELLA_MODULE, wrap_string(((Module*)(symbol->home_context))->module_full_name), cons(SYM_WALK_STELLA_TRUE, NIL))))), cons(wrap_integer((keywordP(symbol) ? KEYWORD_SYM : ((surrogateP(symbol) ? SURROGATE_SYM : SYMBOL_SYM)))), NIL))), symbolconstanttypetree, NIL), NIL)), NIL);
    walk_auxiliary_tree(tree);
    return (symbolconstant);
  }
}

Generalized_Symbol* register_symbol(Generalized_Symbol* symbol) {
  symbol = ((Generalized_Symbol*)(symbol->permanentify()));
  { Generalized_Symbol* registeredsymbol = ((Generalized_Symbol*)(oSYMBOL_REGISTRYo->lookup(symbol)));

    if (!(((boolean)(registeredsymbol)))) {
      oSYMBOL_SETo->insert(symbol);
      if (use_hardcoded_symbolsP()) {
        registeredsymbol = symbol;
      }
      else {
        registeredsymbol = create_startup_symbol(symbol);
      }
      oSYMBOL_REGISTRYo->insert_at(symbol, registeredsymbol);
    }
    return (registeredsymbol);
  }
}

void clear_symbol_registry() {
  { Generalized_Symbol* symbol = NULL;
    Cons* iter_000 = oSYMBOL_SETo->the_cons_list;

    while (!(iter_000 == NIL)) {
      symbol = ((Generalized_Symbol*)(iter_000->value));
      iter_000 = iter_000->rest;
      oSYMBOL_REGISTRYo->remove_at(symbol);
    }
  }
  oSYMBOL_SETo->clear();
}

Cons* yield_hardcoded_intern_registered_symbols_tree() {
  { Cons* interntrees = listO(3, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_SYMBOLS, NIL);

    { Generalized_Symbol* symbol = NULL;
      Cons* iter_000 = oSYMBOL_SETo->the_cons_list;
      Cons* collect_000 = NULL;

      while (!(iter_000 == NIL)) {
        symbol = ((Generalized_Symbol*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (!((boolean)(collect_000))) {
          {
            collect_000 = cons(cons((symbolP(symbol) ? SYM_WALK_STELLA_INTERN_SYMBOL_AT : ((surrogateP(symbol) ? SYM_WALK_STELLA_INTERN_SURROGATE_AT : SYM_WALK_STELLA_INTERN_KEYWORD_AT))), cons(wrap_string(symbol->symbol_name), cons(wrap_integer(symbol->symbol_id), NIL))), NIL);
            if (interntrees == NIL) {
              interntrees = collect_000;
            }
            else {
              add_cons_to_end_of_cons_list(interntrees, collect_000);
            }
          }
        }
        else {
          {
            collect_000->rest = cons(cons((symbolP(symbol) ? SYM_WALK_STELLA_INTERN_SYMBOL_AT : ((surrogateP(symbol) ? SYM_WALK_STELLA_INTERN_SURROGATE_AT : SYM_WALK_STELLA_INTERN_KEYWORD_AT))), cons(wrap_string(symbol->symbol_name), cons(wrap_integer(symbol->symbol_id), NIL))), NIL);
            collect_000 = collect_000->rest;
          }
        }
      }
    }
    return (interntrees);
  }
}

Object* Object::walk_atomic_tree(Standard_Object*& _return1) {
  { Object* self = this;

    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "walk-atomic-tree: Not defined on " << "`" << self << "'";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
}

Object* Wrapper::walk_atomic_tree(Standard_Object*& _return1) {
  { Wrapper* self = this;

    { Object* _return_temp = self;

      _return1 = wrapper_value_type(self);
      return (_return_temp);
    }
  }
}

Surrogate* lookup_constant_symbol(Symbol* self) {
  if ((self == SYM_WALK_STELLA_TRUE) ||
      (self == SYM_WALK_STELLA_FALSE)) {
    return (SGT_WALK_STELLA_BOOLEAN);
  }
  else if (self == SYM_WALK_STELLA_NIL) {
    return (SGT_WALK_STELLA_CONS);
  }
  else if (self == SYM_WALK_STELLA_NULL) {
    return (SGT_WALK_STELLA_UNKNOWN);
  }
  else if (self == SYM_WALK_STELLA_NULL_INTEGER) {
    return (SGT_WALK_STELLA_INTEGER);
  }
  else if (self == SYM_WALK_STELLA_NULL_FLOAT) {
    return (SGT_WALK_STELLA_FLOAT);
  }
  else {
    return (NULL);
  }
}

boolean constant_symbolP(Symbol* symbol) {
  return (((boolean)(lookup_constant_symbol(symbol))));
}

Symbol* walk_constant_symbol(Symbol* self, Surrogate*& _return1) {
  { Symbol* _return_temp = self;

    _return1 = lookup_constant_symbol(self);
    return (_return_temp);
  }
}

Object* Symbol::walk_atomic_tree(Standard_Object*& _return1) {
  { Symbol* self = this;
    Standard_Object* dummy1;

    if (constant_symbolP(self)) {
      { Symbol* value_000 = NULL;
        Surrogate* value_001 = NULL;

        value_000 = walk_constant_symbol(self, value_001);
        _return1 = value_001;
        return (value_000);
      }
    }
    if (oINLININGMETHODCALLpo.get() &&
        ((boolean)(get_local_variable_info(self, KWD_WALK_INLINE_ARGUMENT)))) {
      return (walk_inline_variable_reference(self, _return1));
    }
    { Symbol* truename = true_variable_name(self);
      Standard_Object* type = lookup_variable_type(self);

      if (truename == self) {
        _return1 = type;
        return (self);
      }
      else {
        _return1 = type;
        return (sys_tree(truename, type, dummy1));
      }
    }
  }
}

Object* Keyword::walk_atomic_tree(Standard_Object*& _return1) {
  { Keyword* self = this;

    register_symbol(self);
    _return1 = SGT_WALK_STELLA_KEYWORD;
    return (self);
  }
}

Object* Surrogate::walk_atomic_tree(Standard_Object*& _return1) {
  { Surrogate* self = this;

    register_symbol(self);
    _return1 = SGT_WALK_STELLA_SURROGATE;
    return (self);
  }
}

boolean atomic_expressionP(Object* tree) {
  if (safe_primary_type(tree) == SGT_WALK_STELLA_CONS) {
    { Object* tree_000 = tree;
      Cons* tree = ((Cons*)(tree_000));

      if (((Generalized_Symbol*)(tree->value)) == SYM_WALK_STELLA_TYPED_SYS) {
        return (atomic_expressionP(tree->rest->value));
      }
      else {
      }
      return (FALSE);
    }
  }
  else {
    return (TRUE);
  }
}

boolean variable_expressionP(Object* tree) {
  { Surrogate* test_value_000 = safe_primary_type(tree);

    if (test_value_000 == SGT_WALK_STELLA_CONS) {
      { Object* tree_000 = tree;
        Cons* tree = ((Cons*)(tree_000));

        if (((Generalized_Symbol*)(tree->value)) == SYM_WALK_STELLA_TYPED_SYS) {
          return (variable_expressionP(tree->rest->value));
        }
        else {
        }
        return (FALSE);
      }
    }
    else if (subtype_of_symbolP(test_value_000)) {
      { Object* tree_001 = tree;
        Symbol* tree = ((Symbol*)(tree_001));

        return (TRUE);
      }
    }
    else {
      return (FALSE);
    }
  }
}

boolean vrlet_expressionP(Object* tree) {
  return (consP(tree) &&
      (((Cons*)(tree))->value == SYM_WALK_STELLA_VRLET));
}

boolean procedural_expressionP(Object* tree) {
  { boolean test_value_000 = FALSE;

    test_value_000 = consP(tree);
    if (test_value_000) {
      if (((Cons*)(tree))->value == SYM_WALK_STELLA_VRLET) {
        test_value_000 = TRUE;
      }
      else {
        { boolean foundP_000 = FALSE;

          { Object* term = NULL;
            Cons* iter_000 = ((Cons*)(tree))->rest;

            while (!(iter_000 == NIL)) {
              term = iter_000->value;
              iter_000 = iter_000->rest;
              if (procedural_expressionP(term)) {
                foundP_000 = TRUE;
                break;
              }
            }
          }
          test_value_000 = foundP_000;
        }
      }
    }
    { boolean value_000 = test_value_000;

      return (value_000);
    }
  }
}

boolean substitute_once(Object* expression, Object* newtree, Object* oldtree) {
  if (safe_primary_type(expression) == SGT_WALK_STELLA_CONS) {
    { Object* expression_000 = expression;
      Cons* expression = ((Cons*)(expression_000));

      if (!(expression == NIL)) {
        if (eqlP(expression->value, oldtree)) {
          expression->value = newtree;
          return (TRUE);
        }
        return (substitute_once(expression->value, newtree, oldtree) ||
            substitute_once(expression->rest, newtree, oldtree));
      }
    }
  }
  else {
  }
  return (FALSE);
}

Cons* transform_vrlet_expression(Cons* statement, Cons* vrexpression, Standard_Object*& _return1) {
  { Cons* cursor = NULL;
    Object* laststatement = NULL;

    substitute_once(statement, KWD_WALK_PLACE_WHERE_VRLET_WAS, vrexpression);
    vrexpression->value = SYM_WALK_STELLA_LET;
    cursor = vrexpression->rest->rest;
    while (!(cursor == NIL)) {
      if (cursor->rest == NIL) {
        laststatement = cursor->value;
        cursor->value = statement;
        break;
      }
      cursor = cursor->rest;
    }
    substitute_once(statement, laststatement, KWD_WALK_PLACE_WHERE_VRLET_WAS);
    return (walk_a_cons_tree(vrexpression, _return1));
  }
}

Cons* help_transform_boolean_procedural_expression(Object* expression, Symbol* testvariable) {
  if ((!consP(expression)) ||
      (!procedural_expressionP(expression))) {
    return (cons(listO(3, SYM_WALK_STELLA_SETQ, testvariable, cons(expression, NIL)), NIL));
  }
  { Cons* exp = ((Cons*)(expression));
    Object* renamed_Operator = exp->value;
    Object* firstarg = exp->rest->value;
    Cons* otherargs = exp->rest->rest;

    { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(renamed_Operator));

      if (test_value_000 == SYM_WALK_STELLA_AND) {
        if (otherargs == NIL) {
          return (help_transform_boolean_procedural_expression(firstarg, testvariable));
        }
        else if (procedural_expressionP(firstarg)) {
          return (help_transform_boolean_procedural_expression(firstarg, testvariable)->concatenate(cons(listO(3, SYM_WALK_STELLA_WHEN, testvariable, help_transform_boolean_procedural_expression(cons(SYM_WALK_STELLA_AND, otherargs->concatenate(NIL, 0)), testvariable)->concatenate(NIL, 0)), NIL), 0));
        }
        else {
          return (listO(3, listO(3, SYM_WALK_STELLA_SETQ, testvariable, cons(firstarg, NIL)), listO(3, SYM_WALK_STELLA_WHEN, testvariable, help_transform_boolean_procedural_expression(cons(SYM_WALK_STELLA_AND, otherargs->concatenate(NIL, 0)), testvariable)->concatenate(NIL, 0)), NIL));
        }
      }
      else if (test_value_000 == SYM_WALK_STELLA_OR) {
        if (otherargs == NIL) {
          return (help_transform_boolean_procedural_expression(firstarg, testvariable));
        }
        else if (procedural_expressionP(firstarg)) {
          return (help_transform_boolean_procedural_expression(firstarg, testvariable)->concatenate(cons(listO(3, SYM_WALK_STELLA_WHEN, listO(3, SYM_WALK_STELLA_NOT, testvariable, NIL), help_transform_boolean_procedural_expression(cons(SYM_WALK_STELLA_OR, otherargs->concatenate(NIL, 0)), testvariable)->concatenate(NIL, 0)), NIL), 0));
        }
        else {
          { Cons* otherargstest = help_transform_boolean_procedural_expression(cons(SYM_WALK_STELLA_OR, otherargs->concatenate(NIL, 0)), testvariable);

            if (!(otherargstest->rest == NIL)) {
              otherargstest = cons(cons(SYM_WALK_STELLA_PROGN, otherargstest->concatenate(NIL, 0)), NIL);
            }
            return (cons(listO(3, SYM_WALK_STELLA_IF, firstarg, cons(listO(3, SYM_WALK_STELLA_SETQ, testvariable, cons(SYM_WALK_STELLA_TRUE, NIL)), otherargstest->concatenate(NIL, 0))), NIL));
          }
        }
      }
      else if (test_value_000 == SYM_WALK_STELLA_NOT) {
        return (help_transform_boolean_procedural_expression(firstarg, testvariable)->concatenate(cons(listO(3, SYM_WALK_STELLA_SETQ, testvariable, cons(listO(3, SYM_WALK_STELLA_NOT, testvariable, NIL), NIL)), NIL), 0));
      }
      else if (test_value_000 == SYM_WALK_STELLA_VRLET) {
        { Cons* cursor = otherargs;

          while (!(cursor->rest == NIL)) {
            cursor = cursor->rest;
          }
          cursor->value = listO(3, SYM_WALK_STELLA_SETQ, testvariable, cons(cursor->value, NIL));
        }
        return (cons(listO(3, SYM_WALK_STELLA_LET, firstarg, otherargs->concatenate(NIL, 0)), NIL));
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
    }
  }
}

Cons* transform_boolean_procedural_expression(Cons* statement, Cons* vrexpression, Standard_Object*& _return1) {
  if (vrexpression->value == SYM_WALK_STELLA_VRLET) {
    return (transform_vrlet_expression(statement, vrexpression, _return1));
  }
  { Symbol* testvariable = local_gensym("TEST-VALUE");
    Cons* otree = NULL;

    substitute_once(statement, testvariable, vrexpression);
    otree = ((Cons*)(walk_without_type_tree(listO(3, SYM_WALK_STELLA_LET, cons(cons(testvariable, cons(SYM_WALK_STELLA_FALSE, NIL)), NIL), help_transform_boolean_procedural_expression(vrexpression, testvariable)->concatenate(cons(statement, NIL), 0)))));
    _return1 = SGT_WALK_STELLA_BOOLEAN;
    return (otree);
  }
}

Cons* percolate_out_boolean_vrlet_expressions(Cons* booleanexpression) {
  { Symbol* testvariable = local_gensym("TEST-VALUE");
    Cons* otree = NULL;

    otree = listO(3, SYM_WALK_STELLA_VRLET, cons(cons(testvariable, cons(SYM_WALK_STELLA_FALSE, NIL)), NIL), help_transform_boolean_procedural_expression(booleanexpression, testvariable)->concatenate(cons(testvariable, NIL), 0));
    return (otree);
  }
}

Cons* transform_procedural_expression(Cons* statement, Cons* vrexpression, Standard_Object* type, Standard_Object*& _return1) {
  if (type_spec_to_base_type(type) == SGT_WALK_STELLA_BOOLEAN) {
    return (transform_boolean_procedural_expression(statement, vrexpression, _return1));
  }
  else {
    return (transform_vrlet_expression(statement, vrexpression, _return1));
  }
}

Cons* percolate_out_vrlet_expression(Cons* statement, Cons* vrexpression, Standard_Object* type, Standard_Object*& _return1) {
  { Object* laststatement = NULL;
    Cons* cursor = vrexpression->rest->rest;

    while (!(cursor == NIL)) {
      if (cursor->rest == NIL) {
        laststatement = cursor->value;
        cursor->value = KWD_WALK_PLACE_WHERE_LAST_EXPRESSION_WAS;
        break;
      }
      cursor = cursor->rest;
    }
    substitute_once(statement, laststatement, vrexpression);
    substitute_once(vrexpression, statement, KWD_WALK_PLACE_WHERE_LAST_EXPRESSION_WAS);
    _return1 = type;
    return (vrexpression);
  }
}

Cons* walk_setq_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    bad_argument_countP(tree, 2);
    { Object* variable = tree->rest->value;
      Object* value = tree->rest->rest->value;
      Standard_Object* variabletype = NULL;
      Object* ovalue = NULL;
      Standard_Object* otype = NULL;

      if (!(symbolP(variable))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal first argument to 'setq'; should be a symbol: " << "`" << de_uglify_parse_tree(variable) << "'" << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
      }
      variabletype = lookup_variable_type(((Symbol*)(variable)));
      ovalue = walk_expression_tree(value, variabletype, SYM_WALK_STELLA_SETQ, TRUE, otype);
      tree->third_setter(ovalue);
      if (!(procedural_expressionP(ovalue))) {
        tree->second_setter(true_variable_name(((Symbol*)(variable))));
        if (walking_expressionP() &&
            (!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP))) {
          return (set_target_language_type(sys_tree(tree, otype, dummy1), type_spec_to_base_type(variabletype), _return1));
        }
        if (!eqlP(variable, tree->rest->value)) {
          return (sys_tree(tree, otype, _return1));
        }
        else {
          _return1 = otype;
          return (tree);
        }
      }
      return (transform_procedural_expression(tree, ((Cons*)(ovalue)), ((otype == SGT_WALK_STELLA_UNKNOWN) ? variabletype : otype), _return1));
    }
  }
}

Cons* yield_fixed_slot_value_setter_tree(Slot* slot, Object* objectref, Object* valuetree) {
  { Standard_Object* dummy1;

    return (listO(3, SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER, slot->slot_owner, cons(slot->slot_name, cons(objectref, cons(walk_expression_tree(valuetree, slot->type(), SYM_WALK_STELLA_SLOT_VALUE_SETTER, FALSE, dummy1), NIL)))));
  }
}

Cons* walk_setf_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;

    { Cons* place = NULL;

      if (consP(tree->rest->value)) {
        place = ((Cons*)(tree->rest->value));
      }
      if (!(((boolean)(place)) &&
          (symbolP(place->value) &&
           (place->length() >= (((place->value == SYM_WALK_STELLA_SLOT_VALUE) ? 3 : 2)))))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal place argument in 'setf' expression: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
      }
      if (bad_argument_countP(tree, 2)) {
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
      }
      { Object* objectref = NULL;
        Standard_Object* objecttype = NULL;

        objectref = walk_a_tree(place->rest->value, objecttype);
        objectref = coerce_a_tree(objectref, objecttype, objecttype, dummy1);
        { Class* objectclass = type_spec_to_class(objecttype);
          Object* slotname = place->value;
          Slot* slot = lookup_slot(objectclass, ((Symbol*)(slotname)));
          Standard_Object* slottype = NULL;
          Cons* otherplacearguments = place->rest->rest;
          Cons* otree = NIL;

          if (((boolean)(slot))) {
            slottype = slot->compute_return_type_spec(objecttype);
          }
          { Object* valueref = NULL;
            Standard_Object* valuetypespec = NULL;

            valueref = walk_expression_tree(tree->rest->rest->value, slottype, SYM_WALK_STELLA_SETF, TRUE, valuetypespec);
            if (procedural_expressionP(valueref)) {
              tree->third_setter(valueref);
              { Cons* ootree = NULL;
                Standard_Object* otypespec = NULL;

                ootree = transform_procedural_expression(tree, ((Cons*)(valueref)), valuetypespec, otypespec);
                _return1 = otypespec;
                return (ootree);
              }
            }
            if (slotname == SYM_WALK_STELLA_SLOT_VALUE) {
              return (walk_a_cons_tree(listO(3, SYM_WALK_STELLA_SLOT_VALUE_SETTER, objectref, cons(valueref, otherplacearguments->concatenate(NIL, 0))), _return1));
            }
            if (subtype_of_storage_slotP(safe_primary_type(slot))) {
              { Slot* slot_000 = slot;
                Storage_Slot* slot = ((Storage_Slot*)(slot_000));

                if (((boolean)(slot->writer())) ||
                    ((boolean)(lookup_slot(objectclass, yield_setter_method_name(((Symbol*)(slotname))))))) {
                  return (walk_a_cons_tree(cons((((boolean)(slot->writer())) ? slot->writer() : yield_setter_method_name(((Symbol*)(slotname)))), cons(objectref, cons(valueref, otherplacearguments->concatenate(NIL, 0)))), _return1));
                }
                if (system_defined_slot_writerP(slot)) {
                  { Object* value_000 = NULL;
                    Standard_Object* value_001 = NULL;

                    value_000 = sys_tree_if_needed(slot, listO(3, SYM_WALK_STELLA_SYS_CALL_METHOD, slot->slot_owner, cons(yield_setter_method_name(((Symbol*)(slotname))), cons(objectref, cons(walk_expression_tree(valueref, slottype, SYM_WALK_STELLA_SETF, FALSE, dummy2), NIL)))), objecttype, slottype, value_001);
                    _return1 = value_001;
                    return (((Cons*)(value_000)));
                  }
                }
                otree = yield_slot_value_setter_tree(slot, objectref, objecttype, valueref, slottype, NULL, dummy3);
                if (procedural_expressionP(otree)) {
                  _return1 = slottype;
                  return (otree);
                }
                else {
                  { Object* value_002 = NULL;
                    Standard_Object* value_003 = NULL;

                    value_002 = sys_tree_if_needed(slot, otree, objecttype, slottype, value_003);
                    _return1 = value_003;
                    return (((Cons*)(value_002)));
                  }
                }
              }
            }
            else {
              return (walk_a_cons_tree(cons(yield_setter_method_name(((Symbol*)(slotname))), cons(objectref, cons(valueref, otherplacearguments->concatenate(NIL, 0)))), _return1));
            }
          }
        }
      }
    }
  }
}

Cons* yield_dynamic_slot_value_tree(Storage_Slot* slot, Object* objectref, Standard_Object* returntype) {
  if (standard_dynamic_slot_accessP(slot)) {
    return (yield_standard_dynamic_slot_value_tree(slot, objectref, returntype));
  }
  { Keyword* test_value_000 = slot->allocation();

    if (test_value_000 == KWD_WALK_DYNAMIC) {
      { Symbol* answervar = local_gensym("ANSWER");

        return (listO(5, SYM_WALK_STELLA_VRLET, cons(cons(answervar, cons(returntype, cons(type_to_null_value_tree(slot->type()), NIL))), NIL), listO(9, SYM_WALK_STELLA_FOREACH, SYM_WALK_STELLA_IT, SYM_WALK_STELLA_ON, listO(3, SYM_WALK_STELLA_DYNAMIC_SLOTS, objectref, NIL), SYM_WALK_STELLA_WHERE, listO(4, SYM_WALK_STELLA_EQp, listO(3, SYM_WALK_STELLA_KEY, SYM_WALK_STELLA_IT, NIL), listO(3, SYM_WALK_STELLA_BQUOTE, slot->slot_name, NIL), NIL), SYM_WALK_STELLA_DO, (slot_value_is_bare_literalP(slot) ? listO(4, SYM_WALK_STELLA_LET, cons(listO(3, SYM_WALK_STELLA_WRAPPEDVALUE, slot->type()->type_to_wrapped_type(), cons(listO(3, SYM_WALK_STELLA_VALUE, SYM_WALK_STELLA_IT, NIL), NIL)), NIL), listO(3, SYM_WALK_STELLA_SETQ, answervar, cons(listO(3, SYM_WALK_STELLA_WRAPPER_VALUE, SYM_WALK_STELLA_WRAPPEDVALUE, NIL), NIL)), NIL) : listO(3, SYM_WALK_STELLA_SETQ, answervar, cons(listO(3, SYM_WALK_STELLA_VALUE, SYM_WALK_STELLA_IT, NIL), NIL))), cons(cons(SYM_WALK_STELLA_BREAK, NIL), NIL)), answervar, NIL));
      }
    }
    else if (test_value_000 == KWD_WALK_BIT) {
      throw *new_stella_exception("Allocation :bit IS NOT YET IMPLEMENTED");
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

Cons* yield_dynamic_slot_value_setter_tree(Storage_Slot* slot, Object* objectref, Object* valueref, Standard_Object* returntype) {
  { Symbol* slotname = slot->slot_name;

    if (standard_dynamic_slot_accessP(slot)) {
      return (yield_standard_dynamic_slot_value_setter_tree(slot, objectref, valueref, returntype));
    }
    { Keyword* test_value_000 = slot->allocation();

      if (test_value_000 == KWD_WALK_DYNAMIC) {
        return (listO(6, SYM_WALK_STELLA_VRLET, listO(4, listO(3, SYM_WALK_STELLA_DYNAMICSLOTS, listO(3, SYM_WALK_STELLA_DYNAMIC_SLOTS, objectref, NIL), NIL), listO(3, SYM_WALK_STELLA_NEWVALUE, valueref, NIL), listO(3, SYM_WALK_STELLA_FOUNDMATCHINGENTRYp, SYM_WALK_STELLA_FALSE, NIL), NIL), listO(9, SYM_WALK_STELLA_FOREACH, SYM_WALK_STELLA_IT, SYM_WALK_STELLA_ON, SYM_WALK_STELLA_DYNAMICSLOTS, SYM_WALK_STELLA_WHERE, listO(4, SYM_WALK_STELLA_EQp, listO(3, SYM_WALK_STELLA_KEY, SYM_WALK_STELLA_IT, NIL), listO(3, SYM_WALK_STELLA_BQUOTE, slotname, NIL), NIL), SYM_WALK_STELLA_DO, (slot_value_is_bare_literalP(slot) ? listO(4, SYM_WALK_STELLA_LET, cons(listO(3, SYM_WALK_STELLA_OLDVALUE, type_to_symbol(slot->type()->type_to_wrapped_type()), cons(listO(3, SYM_WALK_STELLA_VALUE, SYM_WALK_STELLA_IT, NIL), NIL)), NIL), listO(4, SYM_WALK_STELLA_SETF, listO(3, SYM_WALK_STELLA_WRAPPER_VALUE, SYM_WALK_STELLA_OLDVALUE, NIL), SYM_WALK_STELLA_NEWVALUE, NIL), NIL) : listO(4, SYM_WALK_STELLA_SETF, listO(3, SYM_WALK_STELLA_VALUE, SYM_WALK_STELLA_IT, NIL), SYM_WALK_STELLA_NEWVALUE, NIL)), cons(listO(4, SYM_WALK_STELLA_SETQ, SYM_WALK_STELLA_FOUNDMATCHINGENTRYp, SYM_WALK_STELLA_TRUE, NIL), NIL)), listO(4, SYM_WALK_STELLA_WHEN, listO(4, SYM_WALK_STELLA_AND, listO(3, SYM_WALK_STELLA_NOT, SYM_WALK_STELLA_FOUNDMATCHINGENTRYp, NIL), listO(3, SYM_WALK_STELLA_DEFINEDp, SYM_WALK_STELLA_NEWVALUE, NIL), NIL), listO(4, SYM_WALK_STELLA_SETF, listO(3, SYM_WALK_STELLA_THE_KV_LIST, SYM_WALK_STELLA_DYNAMICSLOTS, NIL), listO(4, SYM_WALK_STELLA_KV_CONS, listO(3, SYM_WALK_STELLA_BQUOTE, slotname, NIL), (slot_value_is_bare_literalP(slot) ? listO(3, SYM_WALK_STELLA_WRAP_LITERAL, SYM_WALK_STELLA_NEWVALUE, NIL) : ((Object*)(SYM_WALK_STELLA_NEWVALUE))), cons(listO(3, SYM_WALK_STELLA_THE_KV_LIST, SYM_WALK_STELLA_DYNAMICSLOTS, NIL), NIL)), NIL), NIL), SYM_WALK_STELLA_NEWVALUE, NIL));
      }
      else if (test_value_000 == KWD_WALK_BIT) {
        throw *new_stella_exception("Allocation :bit IS NOT YET IMPLEMENTED");
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
    }
  }
}

Cons* yield_native_slot_value_tree(Storage_Slot* slot, Object* objectref, Standard_Object* objecttype) {
  { Standard_Object* dummy1;

    slot = canonical_slot(slot);
    { boolean mixinslotP = ((Class*)(slot->slot_owner->surrogate_value))->mixinP;
      Class* objectclass = type_spec_to_class(objecttype);

      if (mixinslotP &&
          translate_to_single_inheritance_languageP()) {
        if (objectclass->mixinP) {
          return (sys_tree(listO(4, SYM_WALK_STELLA_SYS_CALL_METHOD, SGT_WALK_STELLA_OBJECT, yield_native_slot_reader_name(slot), cons(objectref, NIL)), (((boolean)(((Compound_Type_Specifier*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL))))) ? ((Standard_Object*)(((Compound_Type_Specifier*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL))))) : ((Standard_Object*)(slot->slot_base_type))), dummy1));
        }
      }
      if (slot->slot_hardwiredP) {
        return (listO(3, SYM_WALK_STELLA_SYS_CALL_METHOD, objectclass->class_type, cons(slot->slot_name, cons(objectref, NIL))));
      }
      else {
        return (listO(3, SYM_WALK_STELLA_SYS_SLOT_VALUE, objectclass->class_type, cons(slot->slot_name, cons(objectref, NIL))));
      }
    }
  }
}

Cons* yield_context_sensitive_slot_value_tree(Storage_Slot* slot, Object* objectref, Standard_Object* returntype, Cons* csoptions, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    { boolean contextsensitiveP = !(((boolean)(csoptions)) &&
          (search_plist(csoptions, KWD_WALK_CONTEXT_SENSITIVEp) == SYM_WALK_STELLA_FALSE));
      Symbol* dontinheritP = ((((boolean)(csoptions)) &&
          (search_plist(csoptions, KWD_WALK_DONT_INHERITp) == SYM_WALK_STELLA_TRUE)) ? SYM_WALK_STELLA_TRUE : SYM_WALK_STELLA_FALSE);
      Object* objectrefcopy = objectref;
      boolean onceonlywrapperP = contextsensitiveP &&
          (!side_effect_free_expressionP(objectref));
      boolean literalslotP = slot_value_is_bare_literalP(slot);
      Surrogate* realbasetype = slot->slot_base_type;
      Compound_Type_Specifier* realtypespec = ((Compound_Type_Specifier*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL)));
      Object* oldvaluetree = NULL;
      Symbol* oldvalueref = local_gensym("OLD-VALUE");
      Cons* accesstree = NULL;

      if (onceonlywrapperP) {
        objectref = local_gensym("OBJECT");
        push_variable_binding(((Symbol*)(objectref)), slot->slot_owner);
      }
      else {
        objectref = sys_tree(walk_without_type_tree(objectref), slot->slot_owner, dummy1);
      }
      slot->slot_base_type = SGT_WALK_STELLA_OBJECT;
      if (((boolean)(realtypespec))) {
        set_dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL, NULL);
      }
      slot->slot_context_sensitiveP = FALSE;
      oldvaluetree = sys_tree(walk_without_type_tree(listO(3, SYM_WALK_STELLA_SLOT_VALUE, (contextsensitiveP ? copy_cons_tree(objectref) : objectref), cons(slot->slot_name, NIL))), SGT_WALK_STELLA_OBJECT, dummy2);
      slot->slot_base_type = realbasetype;
      if (((boolean)(realtypespec))) {
        set_dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, realtypespec, NULL);
      }
      slot->slot_context_sensitiveP = TRUE;
      if (!contextsensitiveP) {
        _return1 = SGT_WALK_STELLA_OBJECT;
        return (((Cons*)(oldvaluetree)));
      }
      if (onceonlywrapperP) {
        pop_variable_binding();
      }
      accesstree = listO(4, SYM_WALK_STELLA_SAFE_CAST, listO(3, SYM_WALK_STELLA_ACCESS_IN_CONTEXT, (onceonlywrapperP ? oldvalueref : oldvaluetree), listO(3, listO(3, SYM_WALK_STELLA_HOME_CONTEXT, objectref, NIL), dontinheritP, NIL)), yield_type_spec_tree((literalslotP ? slot->type()->type_to_wrapped_type() : ((Standard_Object*)(returntype)))), NIL);
      if (literalslotP) {
        accesstree = listO(3, SYM_WALK_STELLA_WRAPPER_VALUE, accesstree, NIL);
      }
      if (onceonlywrapperP) {
        _return1 = returntype;
        return (listO(4, SYM_WALK_STELLA_VRLET, listO(3, cons(objectref, cons(objectrefcopy, NIL)), cons(oldvalueref, cons(oldvaluetree, NIL)), NIL), accesstree, NIL));
      }
      else {
        _return1 = returntype;
        return (((Cons*)(walk_without_type_tree(accesstree))));
      }
    }
  }
}

Cons* yield_slot_value_tree(Storage_Slot* slot, Object* objectref, Standard_Object* objecttype, Standard_Object* returntype, Cons* csoptions, Standard_Object*& _return1) {
  if (slot->slot_context_sensitiveP) {
    return (yield_context_sensitive_slot_value_tree(slot, objectref, returntype, csoptions, _return1));
  }
  else {
    if (slot->dynamic_storageP()) {
      _return1 = returntype;
      return (((Cons*)(walk_without_type_tree(yield_dynamic_slot_value_tree(slot, objectref, returntype)))));
    }
    else {
      _return1 = returntype;
      return (yield_native_slot_value_tree(slot, objectref, objecttype));
    }
  }
}

Cons* walk_slot_value_tree(Cons* tree, Standard_Object*& _return1) {
  { Object* objectref = NULL;
    Standard_Object* objecttype = NULL;

    objectref = walk_a_tree(tree->rest->value, objecttype);
    { Object* slotname = tree->rest->rest->value;
      Cons* csoptions = tree->nth_rest(3);
      Class* objectclass = type_spec_to_class(objecttype);
      Slot* slot = lookup_slot(objectclass, ((Symbol*)(slotname)));
      Cons* otree = NULL;
      Standard_Object* otype = NULL;

      if (!((boolean)(slot))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Can't read a non-existent slot " << "`" << de_uglify_parse_tree(slotname) << "'" << endl << "   on the class " << "`" << class_name(objectclass) << "'" << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
      }
      { Surrogate* test_value_000 = safe_primary_type(slot);

        if (subtype_of_method_slotP(test_value_000)) {
          { Slot* slot_000 = slot;
            Method_Slot* slot = ((Method_Slot*)(slot_000));

            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Can't read the method slot " << "`" << de_uglify_parse_tree(slot) << "'" << " by calling 'slot-value'" << "." << endl;
              }
            }
            _return1 = NULL;
            return (NULL);
          }
        }
        else if (subtype_of_storage_slotP(test_value_000)) {
          { Slot* slot_001 = slot;
            Storage_Slot* slot = ((Storage_Slot*)(slot_001));

            otype = slot->compute_return_type_spec(objecttype);
            otree = yield_slot_value_tree(slot, objectref, objecttype, otype, csoptions, otype);
            if (procedural_expressionP(otree)) {
              _return1 = otype;
              return (otree);
            }
            else {
              { Object* value_000 = NULL;
                Standard_Object* value_001 = NULL;

                value_000 = sys_tree_if_needed(slot, otree, objecttype, otype, value_001);
                _return1 = value_001;
                return (((Cons*)(value_000)));
              }
            }
          }
        }
        else {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
      }
    }
  }
}

Cons* yield_native_slot_value_setter_tree(Storage_Slot* slot, Object* objectref, Standard_Object* objecttype, Object* valueref) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    slot = canonical_slot(slot);
    { boolean mixinslotP = ((Class*)(slot->slot_owner->surrogate_value))->mixinP;
      Class* objectclass = type_spec_to_class(objecttype);

      if (mixinslotP &&
          translate_to_single_inheritance_languageP()) {
        if (objectclass->mixinP) {
          return (sys_tree(listO(4, SYM_WALK_STELLA_SYS_CALL_METHOD, SGT_WALK_STELLA_OBJECT, yield_native_slot_reader_name(slot), cons(objectref, cons(valueref, NIL))), (((boolean)(((Compound_Type_Specifier*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL))))) ? ((Standard_Object*)(((Compound_Type_Specifier*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL))))) : ((Standard_Object*)(slot->slot_base_type))), dummy1));
        }
      }
      if (slot->slot_hardwiredP) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Can't set the value of the hardwired slot " << "`" << slot->slot_owner << "'" << "." << "`" << slot->slot_name << "'" << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(NULL, slot->slot_base_type, dummy2));
      }
      return (listO(3, SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER, objectclass->class_type, cons(slot->slot_name, cons(objectref, cons(valueref, NIL)))));
    }
  }
}

Cons* yield_context_sensitive_slot_value_setter_tree(Storage_Slot* slot, Object* objectref, Object* valueref, Standard_Object* returntype, Cons* csoptions, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;

    { boolean contextsensitiveP = !(((boolean)(csoptions)) &&
          (search_plist(csoptions, KWD_WALK_CONTEXT_SENSITIVEp) == SYM_WALK_STELLA_FALSE));
      Symbol* copytochildren = ((((boolean)(csoptions)) &&
          (search_plist(csoptions, KWD_WALK_COPY_TO_CHILDRENp) == SYM_WALK_STELLA_TRUE)) ? SYM_WALK_STELLA_TRUE : SYM_WALK_STELLA_FALSE);
      Surrogate* realbasetype = slot->slot_base_type;
      Compound_Type_Specifier* realtypespec = ((Compound_Type_Specifier*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL)));
      Symbol* objectvar = NULL;
      Symbol* valuevar = NULL;
      Symbol* oldvaluevar = NULL;
      Symbol* newvaluevar = NULL;
      Cons* oldvaluetree = NULL;
      Cons* setnewvaluetree = NULL;

      slot->slot_base_type = SGT_WALK_STELLA_OBJECT;
      if (((boolean)(realtypespec))) {
        set_dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL, NULL);
      }
      slot->slot_context_sensitiveP = FALSE;
      if (contextsensitiveP) {
        objectvar = local_gensym("OBJECT");
        valuevar = local_gensym("VALUE");
        oldvaluevar = local_gensym("OLD-VALUE");
        newvaluevar = local_gensym("NEW-VALUE");
        push_variable_binding(objectvar, slot->slot_owner);
        push_variable_binding(newvaluevar, SGT_WALK_STELLA_OBJECT);
        oldvaluetree = sys_tree(walk_without_type_tree(listO(3, SYM_WALK_STELLA_SLOT_VALUE, objectvar, cons(slot->slot_name, NIL))), SGT_WALK_STELLA_OBJECT, dummy1);
        setnewvaluetree = sys_tree(walk_without_type_tree(listO(4, SYM_WALK_STELLA_SETF, listO(3, SYM_WALK_STELLA_SLOT_VALUE, objectvar, cons(slot->slot_name, NIL)), newvaluevar, NIL)), SGT_WALK_STELLA_OBJECT, dummy2);
      }
      else {
        setnewvaluetree = sys_tree(walk_without_type_tree(listO(4, SYM_WALK_STELLA_SETF, listO(3, SYM_WALK_STELLA_SLOT_VALUE, objectref, cons(slot->slot_name, NIL)), valueref, NIL)), SGT_WALK_STELLA_OBJECT, dummy3);
      }
      slot->slot_base_type = realbasetype;
      if (((boolean)(realtypespec))) {
        set_dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, realtypespec, NULL);
      }
      slot->slot_context_sensitiveP = TRUE;
      if (!contextsensitiveP) {
        _return1 = SGT_WALK_STELLA_OBJECT;
        return (setnewvaluetree);
      }
      pop_variable_binding();
      pop_variable_binding();
      _return1 = returntype;
      return (listO(5, SYM_WALK_STELLA_VRLET, listO(5, cons(objectvar, cons(objectref, NIL)), cons(valuevar, cons(realbasetype, cons(valueref, NIL))), cons(oldvaluevar, cons(oldvaluetree, NIL)), cons(newvaluevar, cons(listO(3, SYM_WALK_STELLA_UPDATE_IN_CONTEXT, oldvaluevar, cons(valuevar, listO(3, listO(3, SYM_WALK_STELLA_HOME_CONTEXT, objectvar, NIL), copytochildren, NIL))), NIL)), NIL), listO(4, SYM_WALK_STELLA_WHEN, listO(3, SYM_WALK_STELLA_NOT, listO(3, SYM_WALK_STELLA_CS_VALUEp, oldvaluevar, NIL), NIL), setnewvaluetree, NIL), valuevar, NIL));
    }
  }
}

Cons* yield_slot_value_setter_tree(Storage_Slot* slot, Object* objectref, Standard_Object* objecttype, Object* valueref, Standard_Object* returntype, Cons* csoptions, Standard_Object*& _return1) {
  if (slot->slot_context_sensitiveP) {
    return (yield_context_sensitive_slot_value_setter_tree(slot, objectref, valueref, returntype, csoptions, _return1));
  }
  else {
    if (slot->dynamic_storageP()) {
      _return1 = returntype;
      return (((Cons*)(walk_without_type_tree(yield_dynamic_slot_value_setter_tree(slot, objectref, valueref, returntype)))));
    }
    else {
      _return1 = returntype;
      return (yield_native_slot_value_setter_tree(slot, objectref, objecttype, valueref));
    }
  }
}

Cons* walk_slot_value_setter_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    { Object* objectref = NULL;
      Standard_Object* objecttype = NULL;

      objectref = walk_a_tree(tree->rest->value, objecttype);
      { Object* slotname = tree->fourth();
        Cons* csoptions = tree->nth_rest(4);
        Class* objectclass = type_spec_to_class(objecttype);
        Slot* slot = lookup_slot(objectclass, ((Symbol*)(slotname)));
        Object* valueref = NULL;
        Cons* otree = NULL;
        Standard_Object* otype = NULL;

        if (!((boolean)(slot))) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Can't set a non-existent slot " << "`" << de_uglify_parse_tree(slotname) << "'" << endl << "   on the class " << "`" << class_name(objectclass) << "'" << "." << endl;
            }
          }
          return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
        }
        { Surrogate* test_value_000 = safe_primary_type(slot);

          if (subtype_of_method_slotP(test_value_000)) {
            { Slot* slot_000 = slot;
              Method_Slot* slot = ((Method_Slot*)(slot_000));

              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Can't set the method slot " << "`" << de_uglify_parse_tree(slot) << "'" << " by calling 'slot-value-setter'" << "." << endl;
                }
              }
              return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
            }
          }
          else if (subtype_of_storage_slotP(test_value_000)) {
            { Slot* slot_001 = slot;
              Storage_Slot* slot = ((Storage_Slot*)(slot_001));

              otype = slot->compute_return_type_spec(objecttype);
              valueref = walk_expression_tree(tree->rest->rest->value, otype, SYM_WALK_STELLA_SLOT_VALUE_SETTER, FALSE, dummy1);
              otree = yield_slot_value_setter_tree(slot, objectref, objecttype, valueref, otype, csoptions, otype);
              if (procedural_expressionP(otree)) {
                _return1 = otype;
                return (otree);
              }
              else {
                { Object* value_000 = NULL;
                  Standard_Object* value_001 = NULL;

                  value_000 = sys_tree_if_needed(slot, otree, objecttype, otype, value_001);
                  _return1 = value_001;
                  return (((Cons*)(value_000)));
                }
              }
            }
          }
          else {
            { Output_String_Stream* stream_000 = new_output_string_stream();

              *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_000->the_string_reader());
            }
          }
        }
      }
    }
  }
}

Cons* walk_aref_tree(Cons* tree, Standard_Object*& _return1) {
  { boolean dummy1;
    Standard_Object* dummy2;

    { Object* renamed_Operator = tree->value;
      Object* otree = NULL;
      Standard_Object* otype = NULL;
      Parametric_Type_Specifier* arraytype = NULL;
      Standard_Object* elementtype = NULL;
      Cons* arguments = tree->rest;

      otree = walk_expression_tree(arguments->value, SGT_WALK_STELLA_ARRAY, ((Symbol*)(renamed_Operator)), TRUE, otype);
      if (vrlet_expressionP(otree)) {
        arguments->first_setter(otree);
        return (percolate_out_vrlet_expression(tree, ((Cons*)(otree)), SGT_WALK_STELLA_UNKNOWN, _return1));
      }
      if (!(array_type_specifierP(otype))) {
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
      }
      arraytype = ((Parametric_Type_Specifier*)(otype));
      elementtype = extract_parameter_type(arraytype, SYM_WALK_STELLA_ANY_VALUE, dummy1);
      { boolean test_value_000 = FALSE;

        test_value_000 = indefinite_array_typeP(arraytype);
        if (test_value_000) {
          { boolean alwaysP_000 = TRUE;

            { Integer_Wrapper* dim = NULL;
              Cons* iter_000 = array_type_dimensions(arraytype)->rest();

              while (!(iter_000 == NIL)) {
                dim = ((Integer_Wrapper*)(iter_000->value));
                iter_000 = iter_000->rest;
                if (!((boolean)(dim))) {
                  alwaysP_000 = FALSE;
                  break;
                }
              }
            }
            test_value_000 = alwaysP_000;
          }
          test_value_000 = !test_value_000;
        }
        if (test_value_000) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Can't AREF underspecified array type" << endl << "   " << "`" << de_uglify_parse_tree(arraytype) << "'" << " in" << endl << "   " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
            }
          }
          return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
        }
      }
      arguments->first_setter(otree);
      arguments = arguments->rest;
      if (renamed_Operator == SYM_WALK_STELLA_AREF_SETTER) {
        otree = walk_expression_tree(arguments->value, elementtype, ((Symbol*)(renamed_Operator)), TRUE, otype);
        arguments->first_setter(otree);
        if (vrlet_expressionP(otree)) {
          return (percolate_out_vrlet_expression(tree, ((Cons*)(otree)), SGT_WALK_STELLA_UNKNOWN, _return1));
        }
        arguments = arguments->rest;
      }
      if (!(arguments->length() == array_type_rank(arraytype))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "AREF indices don't match the rank of the array in" << endl << "   " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
          }
        }
      }
      while (!(arguments == NIL)) {
        otree = walk_expression_tree(arguments->value, SGT_WALK_STELLA_INTEGER, ((Symbol*)(renamed_Operator)), TRUE, otype);
        arguments->first_setter(otree);
        if (vrlet_expressionP(otree)) {
          return (percolate_out_vrlet_expression(tree, ((Cons*)(otree)), SGT_WALK_STELLA_UNKNOWN, _return1));
        }
        arguments = arguments->rest;
      }
      return (sys_tree(listO(3, SYM_WALK_STELLA_SYS_CALL_METHOD, arraytype->specifier_base_type, cons(renamed_Operator, cons(sys_tree(tree->rest->value, arraytype, dummy2), tree->rest->rest->concatenate(NIL, 0)))), elementtype, _return1));
    }
  }
}

Standard_Object* safe_yield_type_specifier(Object* typetree) {
  if (((boolean)(typetree))) {
    { Standard_Object* typespec = typetree->yield_type_specifier();

      validate_type_specifier(typespec, NULL, TRUE);
      return (typespec);
    }
  }
  else {
    return (NULL);
  }
}

Cons* walk_a_declaration(Symbol* variable, Object* typetree, Object* value, boolean inputparameterP) {
  { Standard_Object* sourcetype = NULL;
    Standard_Object* targettype = safe_yield_type_specifier(typetree);
    Surrogate* methodownertype = (((boolean)(oMETHODBEINGWALKEDo.get())) ? oMETHODBEINGWALKEDo.get()->slot_owner : ((Surrogate*)(NULL)));
    Object* ovalue = NULL;

    if (((boolean)(targettype)) &&
        (((boolean)(methodownertype)) &&
         (!voidP(methodownertype)))) {
      targettype = compute_relative_type_spec(targettype, methodownertype);
    }
    if (!(inputparameterP)) {
      push_variable_binding(variable, SGT_WALK_STELLA_UNINITIALIZED);
      ovalue = walk_expression_tree(value, targettype, SYM_WALK_STELLA_VARIABLE_DECLARATION, TRUE, sourcetype);
      if (!((boolean)(targettype))) {
        targettype = sourcetype;
      }
      pop_variable_binding();
    }
    push_variable_binding(variable, targettype);
    if (inputparameterP) {
      return (cons(variable, cons(targettype, NIL)));
    }
    else {
      return (cons(true_variable_name(variable), cons(targettype, cons(ovalue, NIL))));
    }
  }
}

Cons* walk_variable_declarations(Cons* declarations) {
  { Cons* odeclarations = NIL;
    Symbol* variable = NULL;
    boolean vrletencounteredP = FALSE;

    { Object* decl = NULL;
      Cons* iter_000 = declarations;

      while (!(iter_000 == NIL)) {
        decl = iter_000->value;
        iter_000 = iter_000->rest;
        if (safe_primary_type(decl) == SGT_WALK_STELLA_CONS) {
          { Object* decl_000 = decl;
            Cons* decl = ((Cons*)(decl_000));

            if (vrletencounteredP) {
              odeclarations = cons(decl, odeclarations);
              continue;
            }
            else if (!symbolP(decl->value)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal variable in `let' declaration: " << "`" << de_uglify_parse_tree(decl) << "'" << "." << endl;
                }
              }
              variable = SYM_WALK_STELLA_ILLEGAL_VARIABLE;
            }
            else {
              variable = ((Symbol*)(decl->value));
              if (((boolean)(variable->lookup_global_variable()))) {
                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  signal_translation_error();
                  if (!(suppress_warningsP())) {
                    print_error_context(">> ERROR: ", STANDARD_ERROR);
                    *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal binding of global or special variable " << "in `let' declaration:" << endl << "    " << "`" << de_uglify_parse_tree(decl) << "'" << "." << endl;
                  }
                }
                variable = SYM_WALK_STELLA_ILLEGAL_VARIABLE;
              }
            }
            switch (decl->length()) {
              case 1: 
                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  signal_translation_error();
                  if (!(suppress_warningsP())) {
                    print_error_context(">> ERROR: ", STANDARD_ERROR);
                    *(STANDARD_ERROR->native_stream) << endl << " " << "Missing argument(s) in `let' declaration: " << "`" << de_uglify_parse_tree(decl) << "'" << "." << endl;
                  }
                }
              break;
              case 2: 
                if (decl->rest->value == SYM_WALK_STELLA_NULL) {
                  { 
                    BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                    signal_translation_error();
                    if (!(suppress_warningsP())) {
                      print_error_context(">> ERROR: ", STANDARD_ERROR);
                      *(STANDARD_ERROR->native_stream) << endl << " " << "Missing type specifier in `let' declaration: " << "`" << de_uglify_parse_tree(decl) << "'" << "." << endl;
                    }
                  }
                }
                odeclarations = cons(walk_a_declaration(variable, NULL, decl->rest->value, FALSE), odeclarations);
              break;
              case 3: 
                odeclarations = cons(walk_a_declaration(variable, decl->rest->value, decl->rest->rest->value, FALSE), odeclarations);
              break;
              default:
                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  signal_translation_error();
                  if (!(suppress_warningsP())) {
                    print_error_context(">> ERROR: ", STANDARD_ERROR);
                    *(STANDARD_ERROR->native_stream) << endl << " " << "Too many terms in `let' declaration:" << "`" << de_uglify_parse_tree(decl) << "'" << "." << endl;
                  }
                }
              break;
            }
            if (procedural_expressionP(((Cons*)(odeclarations->value))->rest->rest->value)) {
              vrletencounteredP = TRUE;
              pop_variable_binding();
            }
          }
        }
        else {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal `let' declaration: " << "`" << de_uglify_parse_tree(decl) << "'" << "." << endl;
            }
          }
        }
      }
    }
    return (odeclarations->reverse());
  }
}

Cons* transform_let_with_procedural_expression(Cons* tree) {
  { Cons* firstdeclaration = ((Cons*)(((Cons*)(tree->rest->value))->value));
    Object* typetree = firstdeclaration->rest->value;
    Cons* typetreecons = firstdeclaration->rest;
    Object* vrlettree = walk_without_type_tree(firstdeclaration->rest->rest->value);

    if (typetree == SGT_WALK_STELLA_UNKNOWN) {
      firstdeclaration->rest = firstdeclaration->rest->rest;
    }
    { Cons* otree = NULL;
      Standard_Object* unusedtypespec = NULL;

      otree = transform_procedural_expression(tree, ((Cons*)(vrlettree)), ((Standard_Object*)(typetree)), unusedtypespec);
      unusedtypespec = unusedtypespec;
      return (otree);
    }
  }
}

void pop_local_variable_bindings(Cons* declarations) {
  { Object* d = NULL;
    Cons* iter_000 = declarations;

    while (!(iter_000 == NIL)) {
      d = iter_000->value;
      iter_000 = iter_000->rest;
            pop_variable_binding();
    }
  }
}

Cons* walk_let_tree(Cons* tree, Standard_Object*& _return1) {
  tree->second_setter(walk_variable_declarations(((Cons*)(tree->rest->value))));
  { boolean test_value_000 = FALSE;

    { boolean foundP_000 = FALSE;

      { Cons* d = NULL;
        Cons* iter_000 = ((Cons*)(tree->rest->value));

        while (!(iter_000 == NIL)) {
          d = ((Cons*)(iter_000->value));
          iter_000 = iter_000->rest;
          if (procedural_expressionP(d->rest->rest->value)) {
            foundP_000 = TRUE;
            break;
          }
        }
      }
      test_value_000 = foundP_000;
    }
    test_value_000 = !test_value_000;
    if (test_value_000) {
      tree->rest->rest = walk_list_of_statements(tree->rest->rest);
      pop_local_variable_bindings(((Cons*)(tree->rest->value)));
      return (sys_tree(tree, SGT_WALK_STELLA_VOID, _return1));
    }
  }
  { Cons* originaldeclarations = ((Cons*)(tree->rest->value));
    Cons* leadingdeclarations = NIL;
    Cons* trailingdeclarations = NIL;
    boolean leadingP = TRUE;

    { Cons* d = NULL;
      Cons* iter_001 = ((Cons*)(originaldeclarations));

      while (!(iter_001 == NIL)) {
        d = ((Cons*)(iter_001->value));
        iter_001 = iter_001->rest;
        if (procedural_expressionP(d->rest->rest->value)) {
          leadingP = FALSE;
        }
        if (leadingP) {
          leadingdeclarations = cons(d, leadingdeclarations);
        }
        else {
          trailingdeclarations = cons(d, trailingdeclarations);
        }
      }
    }
    leadingdeclarations = leadingdeclarations->reverse();
    trailingdeclarations = trailingdeclarations->reverse();
    tree->second_setter(trailingdeclarations);
    tree = transform_let_with_procedural_expression(tree);
    if (!(leadingdeclarations == NIL)) {
      pop_local_variable_bindings(leadingdeclarations);
      tree = listO(3, SYM_WALK_STELLA_LET, leadingdeclarations, cons(tree, NIL));
    }
    return (walk_a_cons_tree(tree, _return1));
  }
}

// Stack mirroring the current state of bound specials
// with their associated old-value variables.
DEFINE_STELLA_SPECIAL(oSPECIALVARIABLESTACKo, Key_Value_List* , NULL);

// `true' if using specials is enabled and legal.
DEFINE_STELLA_SPECIAL(oSPECIALSENABLEDpo, boolean , TRUE);

// Number of specials bound at the most recent entry
// to a LOOP/WHILE/FOREACH construct.
DEFINE_STELLA_SPECIAL(oNOFSPECIALSATLOOPENTRYo, int , 0);

Keyword* special_implementation_style() {
  if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
    return (KWD_WALK_COMMON_LISP);
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) {
    return (KWD_WALK_UNBIND_WITH_DESTRUCTORS);
  }
  else if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) {
    return (KWD_WALK_COMMON_LISP);
  }
  else {
    return (KWD_WALK_UNWIND_PROTECT);
  }
}

void push_special(Symbol* variable, Symbol* oldvaluevariable) {
  oSPECIALVARIABLESTACKo.get()->the_kv_list = kv_cons(variable, oldvaluevariable, oSPECIALVARIABLESTACKo.get()->the_kv_list);
}

void pop_special() {
  { Kv_Cons* kvlist = oSPECIALVARIABLESTACKo.get()->the_kv_list;

    oSPECIALVARIABLESTACKo.get()->the_kv_list = kvlist->rest;
    free_kv_cons(kvlist);
  }
}

Symbol* lookup_old_value_variable(Symbol* variable) {
  return (((Symbol*)(lookup_variable_table(oSPECIALVARIABLESTACKo.get(), variable))));
}

boolean need_to_unbind_specialsP() {
  return (oSPECIALSENABLEDpo.get() &&
      ((oSPECIALVARIABLESTACKo.get()->length() > 0) &&
       (!get_quoted_tree("((:COMMON-LISP :UNBIND-WITH-DESTRUCTORS) \"/STELLA\")", "/STELLA")->memberP(special_implementation_style()))));
}

Cons* yield_special_unbind_tree(int nofbindings) {
  { Cons* otree = NIL;

    if (special_implementation_style() == KWD_WALK_UNWIND_PROTECT) {
      { Symbol* variable = NULL;
        Symbol* oldvaluevariable = NULL;
        Kv_Cons* iter_000 = oSPECIALVARIABLESTACKo.get()->the_kv_list;

        while (((boolean)(iter_000))) {
          variable = ((Symbol*)(iter_000->key));
          oldvaluevariable = ((Symbol*)(iter_000->value));
          iter_000 = iter_000->rest;
          if ((nofbindings = nofbindings - 1) >= 0) {
            { Cons* value_000 = NULL;

              { Cons* tree = NULL;
                Cons* iter_001 = otree;

                while (!(iter_001 == NIL)) {
                  tree = ((Cons*)(iter_001->value));
                  iter_001 = iter_001->rest;
                  if (tree->rest->value == variable) {
                    value_000 = tree;
                    break;
                  }
                }
              }
              otree = otree->remove(value_000);
            }
            otree = cons(listO(3, SYM_WALK_STELLA_SETQ, variable, cons(oldvaluevariable, NIL)), otree);
          }
        }
      }
      return (((Cons*)(prognify(otree->reverse()))));
    }
    else {
      return (NULL);
    }
  }
}

Cons* yield_return_special_unbind_tree() {
  return (yield_special_unbind_tree(oSPECIALVARIABLESTACKo.get()->length()));
}

Cons* yield_loop_exit_special_unbind_tree() {
  return (yield_special_unbind_tree(oSPECIALVARIABLESTACKo.get()->length() - oNOFSPECIALSATLOOPENTRYo.get()));
}

Cons* walk_special_tree(Cons* tree, Standard_Object*& _return1) {
  if (!(oSPECIALSENABLEDpo.get())) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_error();
      if (!(suppress_warningsP())) {
        print_error_context(">> ERROR: ", STANDARD_ERROR);
        *(STANDARD_ERROR->native_stream) << endl << " " << "'special' statement not legal in the current context: " << endl << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
      }
    }
    return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
  }
  { boolean test_value_000 = FALSE;

    if (tree->length() < 2) {
      test_value_000 = TRUE;
    }
    else {
      if (!consP(tree->rest->value)) {
        test_value_000 = TRUE;
      }
      else {
        {
          { boolean alwaysP_000 = TRUE;

            { Cons* binding = NULL;
              Cons* iter_000 = ((Cons*)(tree->rest->value));

              while (!(iter_000 == NIL)) {
                binding = ((Cons*)(iter_000->value));
                iter_000 = iter_000->rest;
                if (!(consP(binding) &&
                    (symbolP(binding->value) &&
                     (binding->length() == 2)))) {
                  alwaysP_000 = FALSE;
                  break;
                }
              }
            }
            test_value_000 = alwaysP_000;
          }
          test_value_000 = !test_value_000;
        }
      }
    }
    if (test_value_000) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal 'special' statement: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
    }
  }
  if (((Cons*)(tree->rest->value))->length() == 0) {
    tree->first_setter(SYM_WALK_STELLA_LET);
    return (walk_let_tree(tree, _return1));
  }
  { Global_Variable* variable = NULL;

    { Cons* binding = NULL;
      Cons* iter_001 = ((Cons*)(tree->rest->value));

      while (!(iter_001 == NIL)) {
        binding = ((Cons*)(iter_001->value));
        iter_001 = iter_001->rest;
        variable = ((Symbol*)(binding->value))->lookup_global_variable();
        if ((!((boolean)(variable))) ||
            (!variable->variable_specialP)) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Trying to bind non-special variable " << "`" << binding->value << "'" << endl << " in 'special' statement" << "." << endl;
            }
          }
        }
      }
    }
  }
  { Keyword* test_value_001 = special_implementation_style();

    if (test_value_001 == KWD_WALK_COMMON_LISP) {
      return (walk_cl_special_tree(tree, _return1));
    }
    else if (test_value_001 == KWD_WALK_UNBIND_WITH_DESTRUCTORS) {
      return (walk_cl_special_tree(tree, _return1));
    }
    else if (test_value_001 == KWD_WALK_UNWIND_PROTECT) {
      return (walk_unwind_protect_special_tree(tree, _return1));
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << test_value_001 << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

Cons* walk_cl_special_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;

    { Cons* bindings = ((Cons*)(tree->rest->value));
      Cons* body = tree->rest->rest;
      Symbol* variablename = NULL;
      Standard_Object* variabletype = NULL;
      boolean bindviasetqP = FALSE;
      Cons* bindviasetqtrees = NIL;
      Object* otree = NULL;

      { Cons* binding = NULL;
        Cons* iter_000 = bindings;

        while (!(iter_000 == NIL)) {
          binding = ((Cons*)(iter_000->value));
          iter_000 = iter_000->rest;
          variablename = ((Symbol*)(binding->value));
          variabletype = lookup_global_variable_type(variablename);
          if (bindviasetqP) {
            bindviasetqtrees = cons(walk_a_tree(listO(3, SYM_WALK_STELLA_SETQ, variablename, cons(binding->rest->value, NIL)), dummy1), bindviasetqtrees);
            binding->second_setter(variabletype);
            binding->rest->rest = cons(SYM_WALK_STELLA_NULL, NIL);
          }
          else {
            otree = walk_expression_tree(binding->rest->value, variabletype, SYM_WALK_STELLA_SPECIAL, TRUE, dummy2);
            if (procedural_expressionP(otree)) {
              bindviasetqP = TRUE;
              bindviasetqtrees = cons(walk_a_tree(listO(3, SYM_WALK_STELLA_SETQ, variablename, cons(otree, NIL)), dummy3), bindviasetqtrees);
              otree = SYM_WALK_STELLA_NULL;
            }
            binding->second_setter(variabletype);
            binding->rest->rest = cons(otree, NIL);
          }
        }
      }
      tree->first_setter(SYM_WALK_STELLA_SPECIAL);
      tree->rest->rest = bindviasetqtrees->reverse()->concatenate(walk_list_of_statements(body), 0);
      return (sys_tree(tree, SGT_WALK_STELLA_VOID, _return1));
    }
  }
}

Cons* walk_unwind_protect_special_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    { Cons* bindings = ((Cons*)(tree->rest->value));
      Cons* body = tree->rest->rest;
      Symbol* variable = NULL;
      Symbol* oldvaluevariable = NULL;
      Cons* oldvaluebindings = NIL;
      Cons* bindtrees = NIL;
      Cons* unbindtrees = NIL;

      { Cons* binding = NULL;
        Cons* iter_000 = bindings;

        while (!(iter_000 == NIL)) {
          binding = ((Cons*)(iter_000->value));
          iter_000 = iter_000->rest;
          variable = ((Symbol*)(binding->value));
          bindtrees = cons(listO(3, SYM_WALK_STELLA_SETQ, variable, cons(binding->rest->value, NIL)), bindtrees);
          oldvaluevariable = local_gensym(string_concatenate("OLD-", variable->symbol_name, 0));
          push_special(variable, oldvaluevariable);
          oldvaluebindings = cons(cons(oldvaluevariable, cons(variable, NIL)), oldvaluebindings);
          unbindtrees = cons(listO(3, SYM_WALK_STELLA_SETQ, variable, cons(oldvaluevariable, NIL)), unbindtrees);
        }
      }
      tree = ((Cons*)(walk_a_tree(listO(3, SYM_WALK_STELLA_LET, oldvaluebindings->reverse(), bindtrees->reverse()->concatenate(cons(listO(3, SYM_WALK_STELLA_UNWIND_PROTECT, prognify(body), unbindtrees->concatenate(NIL, 0)), NIL), 0)), dummy1)));
      { Cons* binding = NULL;
        Cons* iter_001 = bindings;

        while (!(iter_001 == NIL)) {
          binding = ((Cons*)(iter_001->value));
          iter_001 = iter_001->rest;
          binding = binding;
          pop_special();
        }
      }
      _return1 = SGT_WALK_STELLA_VOID;
      return (tree);
    }
  }
}

Cons* walk_cast_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    if (bad_argument_countP(tree, 2)) {
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
    }
    { Standard_Object* typespec = tree->rest->rest->value->yield_type_specifier();

      tree->second_setter(walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_UNKNOWN, SYM_WALK_STELLA_CAST, TRUE, dummy1));
      if (vrlet_expressionP(tree->rest->value)) {
        return (percolate_out_vrlet_expression(tree, ((Cons*)(tree->rest->value)), typespec, _return1));
      }
      tree->third_setter(typespec);
      if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) &&
          ((tree->value == SYM_WALK_STELLA_CAST) &&
           (oSAFETYo.get() >= 2))) {
        register_symbol(type_spec_to_base_type(typespec));
      }
      _return1 = typespec;
      return (tree);
    }
  }
}

boolean values_treeP(Object* tree) {
  return (consP(tree) &&
      (SYM_WALK_STELLA_VALUES == ((Cons*)(tree))->value));
}

Cons* walk_values_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    { Object* otree = NULL;
      Standard_Object* otype = NULL;
      Cons* bindtree = NIL;
      Symbol* valuevar = NULL;
      boolean hasproceduralvalueP = FALSE;

      { Cons_Iterator* it = tree->rest->allocate_iterator();

        while (it->nextP()) {
          otree = walk_a_tree(it->value, otype);
          if (vrlet_expressionP(otree)) {
            it->value_setter(otree);
            hasproceduralvalueP = TRUE;
            break;
          }
          otree = sys_tree(otree, otype, dummy1);
          if (voidP(otype)) {
            otree = ((Cons*)(otree))->concatenate(cons(SGT_WALK_STELLA_VOID, NIL), 0);
          }
          it->value_setter(otree);
        }
      }
      if (!hasproceduralvalueP) {
        tree->first_setter(SYM_WALK_STELLA_PROGN);
        _return1 = SGT_WALK_STELLA_VOID;
        return (tree);
      }
      { Cons_Iterator* it = tree->rest->allocate_iterator();

        while (it->nextP()) {
          { Cons* value = ((Cons*)(it->value));

            if (procedural_expressionP(value) ||
                (!atomic_expressionP(value->rest->value))) {
              valuevar = local_gensym("VALUE");
              bindtree = cons(cons(valuevar, cons((procedural_expressionP(value) ? value : value->rest->value), NIL)), bindtree);
              if (procedural_expressionP(value)) {
                it->value_setter(valuevar);
                break;
              }
              else {
                value->second_setter(valuevar);
              }
            }
          }
        }
      }
      return (walk_a_cons_tree(listO(3, SYM_WALK_STELLA_LET, bindtree->reverse(), cons(tree, NIL)), _return1));
    }
  }
}

Object* walk_mv_expression_tree(Object* tree, List* targettypes, Symbol* renamed_Operator, Cons*& _return1, List*& _return2) {
  { 
    BIND_STELLA_SPECIAL(oTARGETTYPEo, Standard_Object*, ((targettypes->length() == 1) ? ((Standard_Object*)(((Standard_Object*)(targettypes->first())))) : SGT_WALK_STELLA_UNKNOWN));
    { Object* otree = NULL;
      Cons* valuestree = NULL;
      List* sourcetypes = NULL;

      otree = walk_mv_tree(tree, valuestree, sourcetypes);
      if (!(sourcetypes->length() == targettypes->length())) {
        if ((sourcetypes->length() == 1) &&
            (SGT_WALK_STELLA_UNKNOWN == ((Standard_Object*)(sourcetypes->first())))) {
          warn_about_unknown_source_type(otree);
        }
        else {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "`" << de_uglify_parse_tree(renamed_Operator) << "'" << " operator expected " << "`" << targettypes->length() << "'" << " value(s) from " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
            }
          }
        }
        { Object* _return_temp = otree;

          _return1 = valuestree;
          _return2 = targettypes->copy();
          return (_return_temp);
        }
      }
      if (sourcetypes->emptyP()) {
        _return1 = valuestree;
        _return2 = sourcetypes;
        return (otree);
      }
      if (((boolean)(valuestree))) {
        tree = coerce_mv_tree(valuestree, renamed_Operator, sourcetypes, targettypes, valuestree, targettypes);
      }
      else {
        otree = coerce_mv_tree(otree, renamed_Operator, sourcetypes, targettypes, valuestree, targettypes);
      }
      sourcetypes->free();
      _return1 = valuestree;
      _return2 = targettypes;
      return (otree);
    }
  }
}

List* listify_type_spec(Standard_Object* typespec) {
  if (voidP(typespec)) {
    return (list(0));
  }
  else {
    return (list(1, typespec));
  }
}

Slot* slot_from_expression_tree(Cons* tree) {
  if (SYM_WALK_STELLA_TYPED_SYS == tree->value) {
    tree = ((Cons*)(tree->rest->value));
  }
  { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

    if ((test_value_000 == SYM_WALK_STELLA_SYS_SLOT_VALUE) ||
        ((test_value_000 == SYM_WALK_STELLA_SYS_CALL_METHOD) ||
         ((test_value_000 == SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER) ||
          (test_value_000 == SYM_WALK_STELLA_SYS_CALL_METHOD_SETTER)))) {
      return (lookup_slot(((Class*)(((Surrogate*)(tree->rest->value))->surrogate_value)), ((Symbol*)(tree->rest->rest->value))));
    }
    else if (test_value_000 == SYM_WALK_STELLA_SYS_CALL_FUNCTION) {
      { Method_Slot* function = lookup_function(((Symbol*)(tree->rest->value)));

        if (((boolean)(function)) &&
            isaP(function, SGT_WALK_STELLA_METHOD_SLOT)) {
          return (function);
        }
      }
    }
    else {
      return (NULL);
    }
  }
  return (NULL);
}

Object* walk_mv_tree(Object* tree, Cons*& _return1, List*& _return2) {
  { Object* otree = NULL;
    Standard_Object* otype = NULL;
    Slot* slot = NULL;

    if (values_treeP(tree)) {
      return (walk_mv_values_tree(((Cons*)(tree)), _return1, _return2));
    }
    otree = walk_a_tree(tree, otype);
    if (consP(otree)) {
      slot = slot_from_expression_tree(((Cons*)(otree)));
    }
    if (((boolean)(slot))) {
      if (subtype_of_method_slotP(safe_primary_type(slot))) {
        { Slot* slot_000 = slot;
          Method_Slot* slot = ((Method_Slot*)(slot_000));

          { Object* _return_temp = otree;

            _return1 = NULL;
            _return2 = slot->method_return_type_specifiers_reader()->copy();
            return (_return_temp);
          }
        }
      }
      else {
      }
    }
    { Object* _return_temp = otree;

      _return1 = NULL;
      _return2 = listify_type_spec(otype);
      return (_return_temp);
    }
  }
}

Cons* walk_mv_values_tree(Cons* tree, Cons*& _return1, List*& _return2) {
  if (tree->length() == 1) {
    { Cons* _return_temp = tree;

      _return1 = tree;
      _return2 = listify_type_spec(SGT_WALK_STELLA_VOID);
      return (_return_temp);
    }
  }
  { Object* otree = NULL;
    Standard_Object* otype = NULL;
    List* returntypes = new_list();
    Cons* typedsystree = NULL;

    otype = otype;
    otree = walk_values_tree(tree, otype);
    tree->first_setter(SYM_WALK_STELLA_VALUES);
    { Cons_Iterator* it = tree->rest->allocate_iterator();

      while (it->nextP()) {
        typedsystree = ((Cons*)(it->value));
        returntypes->push(((Standard_Object*)(typedsystree->rest->rest->value)));
        if (SGT_WALK_STELLA_UNKNOWN == typedsystree->rest->rest->value) {
          it->value_setter(typedsystree->rest->value);
        }
      }
    }
    { Cons* _return_temp = ((Cons*)(otree));

      _return1 = tree;
      _return2 = returntypes->reverse();
      return (_return_temp);
    }
  }
}

Object* coerce_mv_tree(Object* tree, Symbol* renamed_Operator, List* sourcetypes, List* targettypes, Cons*& _return1, List*& _return2) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;

    { boolean valuestreeP = values_treeP(tree);
      Cons* valuestree = NIL;
      boolean needtemporariesP = FALSE;
      boolean successP = FALSE;
      Object* otree = NULL;
      Standard_Object* otype = NULL;
      List* returntypes = new_list();

      if (valuestreeP) {
        valuestree = ((Cons*)(tree));
      }
      else {
        valuestree = cons(SYM_WALK_STELLA_VALUES, NIL);
        { Standard_Object* type = NULL;
          Cons* iter_000 = sourcetypes->the_cons_list;
          Cons* collect_000 = NULL;

          while (!(iter_000 == NIL)) {
            type = ((Standard_Object*)(iter_000->value));
            iter_000 = iter_000->rest;
            if (!((boolean)(collect_000))) {
              {
                collect_000 = cons(sys_tree(NULL, type, dummy1), NIL);
                if (valuestree == NIL) {
                  valuestree = collect_000;
                }
                else {
                  add_cons_to_end_of_cons_list(valuestree, collect_000);
                }
              }
            }
            else {
              {
                collect_000->rest = cons(sys_tree(NULL, type, dummy2), NIL);
                collect_000 = collect_000->rest;
              }
            }
          }
        }
      }
      { Standard_Object* tgtts = NULL;
        Standard_Object* srcts = NULL;
        int i = NULL_INTEGER;
        Cons* iter_001 = targettypes->the_cons_list;
        Cons* iter_002 = sourcetypes->the_cons_list;
        int iter_003 = 1;
        Cons_Iterator* it = valuestree->rest->allocate_iterator();
        Cons* collect_001 = NULL;

        while ((!(iter_001 == NIL)) &&
            ((!(iter_002 == NIL)) &&
             it->nextP())) {
          tgtts = ((Standard_Object*)(iter_001->value));
          iter_001 = iter_001->rest;
          srcts = ((Standard_Object*)(iter_002->value));
          iter_002 = iter_002->rest;
          i = iter_003;
          iter_003 = iter_003 + 1;
          otree = help_coerce_a_tree(it->value, srcts, tgtts, FALSE, otype, successP);
          if (!(successP)) {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Type clash on argument " << "`" << de_uglify_parse_tree(wrap_integer(i)) << "'" << " of " << "`" << de_uglify_parse_tree(renamed_Operator) << "'" << " operator." << endl << " " << "`" << de_uglify_parse_tree(srcts) << "'" << " found where " << "`" << de_uglify_parse_tree(tgtts) << "'" << " expected" << "." << endl;
              }
            }
            { Object* _return_temp = tree;

              _return1 = ((Cons*)((valuestreeP ? tree : ((Object*)(NULL)))));
              _return2 = targettypes->copy();
              return (_return_temp);
            }
          }
          if ((!eqlP(otree, it->value)) ||
              ((i > 1) &&
               ((!valuestreeP) &&
                ((!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP)) &&
                 ((!(type_spec_to_base_type(tgtts) == SGT_WALK_STELLA_UNKNOWN)) &&
                  (!(type_spec_to_base_type(srcts) == type_spec_to_base_type(tgtts)))))))) {
            needtemporariesP = TRUE;
          }
          it->value_setter(otree);
          if (!((boolean)(collect_001))) {
            {
              collect_001 = cons(otype, NIL);
              if (returntypes->the_cons_list == NIL) {
                returntypes->the_cons_list = collect_001;
              }
              else {
                add_cons_to_end_of_cons_list(returntypes->the_cons_list, collect_001);
              }
            }
          }
          else {
            {
              collect_001->rest = cons(otype, NIL);
              collect_001 = collect_001->rest;
            }
          }
        }
      }
      if (valuestreeP) {
        _return1 = ((Cons*)(tree));
        _return2 = returntypes;
        return (tree);
      }
      if (!needtemporariesP) {
        _return1 = NULL;
        _return2 = returntypes;
        return (tree);
      }
      valuestree = cons(SYM_WALK_STELLA_VALUES, NIL);
      { int i = NULL_INTEGER;
        int iter_004 = 1;
        int upper_bound_000 = targettypes->length();
        Cons* collect_002 = NULL;

        while (iter_004 <= upper_bound_000) {
          i = iter_004;
          iter_004 = iter_004 + 1;
          if (!((boolean)(collect_002))) {
            {
              collect_002 = cons(local_gensym("VALUE"), NIL);
              if (valuestree == NIL) {
                valuestree = collect_002;
              }
              else {
                add_cons_to_end_of_cons_list(valuestree, collect_002);
              }
            }
          }
          else {
            {
              collect_002->rest = cons(local_gensym("VALUE"), NIL);
              collect_002 = collect_002->rest;
            }
          }
        }
      }
      otree = walk_without_type_tree(listO(3, SYM_WALK_STELLA_MV_BIND, copy_cons_list(valuestree->rest), cons(tree, cons(valuestree, NIL))));
      valuestree->first_setter(get_quoted_tree("((VALUES) \"/STELLA\")", "/STELLA"));
      { Standard_Object* srcts = NULL;
        Standard_Object* tgtts = NULL;
        Cons_Iterator* it = valuestree->rest->allocate_iterator();
        Cons* iter_005 = sourcetypes->the_cons_list;
        Cons* iter_006 = targettypes->the_cons_list;

        while (it->nextP() &&
            ((!(iter_005 == NIL)) &&
             (!(iter_006 == NIL)))) {
          srcts = ((Standard_Object*)(iter_005->value));
          iter_005 = iter_005->rest;
          tgtts = ((Standard_Object*)(iter_006->value));
          iter_006 = iter_006->rest;
          it->value_setter(coerce_a_tree(it->value, srcts, tgtts, dummy3));
        }
      }
      _return1 = valuestree;
      _return2 = returntypes;
      return (otree);
    }
  }
}

Cons* walk_return_and_unbind_specials(Cons* tree, Standard_Object*& _return1) {
  { Cons* unbindtree = yield_return_special_unbind_tree();
    Cons* valuetrees = tree->rest;
    Cons* valuevariables = NIL;
    Cons* valuebindings = NIL;

    { 
      BIND_STELLA_SPECIAL(oSPECIALSENABLEDpo, boolean, FALSE);
      if (valuetrees->length() == 0) {
        return (walk_a_cons_tree(listO(3, SYM_WALK_STELLA_PROGN, unbindtree, cons(tree, NIL)), _return1));
      }
      if ((valuetrees->length() == 1) &&
          (!(oMETHODBEINGWALKEDo.get()->method_return_type_specifiers_reader()->rest() == NIL))) {
        { int i = NULL_INTEGER;
          int iter_000 = 0;
          int upper_bound_000 = oMETHODBEINGWALKEDo.get()->method_return_type_specifiers_reader()->length() - 1;
          Cons* collect_000 = NULL;

          while (iter_000 <= upper_bound_000) {
            i = iter_000;
            iter_000 = iter_000 + 1;
            i = i;
            if (!((boolean)(collect_000))) {
              {
                collect_000 = cons(local_gensym("RETURN-VALUE"), NIL);
                if (valuevariables == NIL) {
                  valuevariables = collect_000;
                }
                else {
                  add_cons_to_end_of_cons_list(valuevariables, collect_000);
                }
              }
            }
            else {
              {
                collect_000->rest = cons(local_gensym("RETURN-VALUE"), NIL);
                collect_000 = collect_000->rest;
              }
            }
          }
        }
        tree->rest = valuevariables;
        return (walk_a_cons_tree(listO(3, SYM_WALK_STELLA_MV_BIND, copy_cons_list(valuevariables), cons(valuetrees->value, tree_to_trees(unbindtree)->concatenate(cons(tree, NIL), 0))), _return1));
      }
      { Cons_Iterator* it = valuetrees->allocate_iterator();

        while (it->nextP()) {
          if (consP(it->value) ||
              (symbolP(it->value) &&
               ((boolean)(lookup_old_value_variable(((Symbol*)(it->value))))))) {
            valuebindings = cons(cons(local_gensym("RETURN-VALUE"), cons(it->value, NIL)), valuebindings);
            it->value_setter(((Cons*)(valuebindings->value))->value);
          }
        }
      }
      return (walk_a_cons_tree(listO(3, SYM_WALK_STELLA_LET, valuebindings->reverse(), tree_to_trees(unbindtree)->concatenate(cons(tree, NIL), 0)), _return1));
    }
  }
}

Cons* walk_return_tree(Cons* tree, Standard_Object*& _return1) {
  if (!((boolean)(oMETHODBEINGWALKEDo.get()))) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_error();
      if (!(suppress_warningsP())) {
        print_error_context(">> ERROR: ", STANDARD_ERROR);
        *(STANDARD_ERROR->native_stream) << endl << " " << "Return statement encountered outside of method or function definition" << "." << endl;
      }
    }
    return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
  }
  oFOUNDRETURNpo.set(TRUE);
  if (need_to_unbind_specialsP()) {
    return (walk_return_and_unbind_specials(tree, _return1));
  }
  { Method_Slot* method = oMETHODBEINGWALKEDo.get();
    List* targettypes = new_list();

    { Standard_Object* rtype = NULL;
      Cons* iter_000 = method->method_return_type_specifiers_reader()->the_cons_list;
      Cons* collect_000 = NULL;

      while (!(iter_000 == NIL)) {
        rtype = ((Standard_Object*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (!((boolean)(collect_000))) {
          {
            collect_000 = cons(compute_relative_type_spec(rtype, method->slot_owner), NIL);
            if (targettypes->the_cons_list == NIL) {
              targettypes->the_cons_list = collect_000;
            }
            else {
              add_cons_to_end_of_cons_list(targettypes->the_cons_list, collect_000);
            }
          }
        }
        else {
          {
            collect_000->rest = cons(compute_relative_type_spec(rtype, method->slot_owner), NIL);
            collect_000 = collect_000->rest;
          }
        }
      }
    }
    { Object* otree = NULL;
      Cons* valuestree = NULL;
      List* valuetypes = NULL;

      otree = walk_mv_expression_tree((((tree->rest->length() == 1) &&
          (!(targettypes->rest() == NIL))) ? tree->rest->value : cons(SYM_WALK_STELLA_VALUES, tree->rest->concatenate(NIL, 0))), targettypes, SYM_WALK_STELLA_RETURN, valuestree, valuetypes);
      targettypes->free();
      valuetypes->free();
      if (!((boolean)(valuestree))) {
        if (!(tree->rest == NIL)) {
          tree->second_setter(otree);
        }
        _return1 = SGT_WALK_STELLA_VOID;
        return (tree);
      }
      valuestree->first_setter(SYM_WALK_STELLA_RETURN);
      _return1 = SGT_WALK_STELLA_VOID;
      return (((Cons*)(otree)));
    }
  }
}

Cons* walk_mv_setq_tree(Cons* tree, Standard_Object*& _return1) {
  { boolean test_value_000 = FALSE;

    if (bad_argument_countP(tree, 2)) {
      test_value_000 = TRUE;
    }
    else {
      if (!consP(tree->rest->value)) {
        test_value_000 = TRUE;
      }
      else {
        {
          { boolean alwaysP_000 = TRUE;

            { Object* v = NULL;
              Cons* iter_000 = ((Cons*)(tree->rest->value));

              while (!(iter_000 == NIL)) {
                v = iter_000->value;
                iter_000 = iter_000->rest;
                if (!symbolP(v)) {
                  alwaysP_000 = FALSE;
                  break;
                }
              }
            }
            test_value_000 = alwaysP_000;
          }
          test_value_000 = !test_value_000;
        }
      }
    }
    if (test_value_000) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal 'mv-setq' expression: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
    }
  }
  { Cons* variables = ((Cons*)(tree->rest->value));
    List* targettypes = new_list();

    { Cons_Iterator* it = variables->allocate_iterator();
      Cons* collect_000 = NULL;

      while (it->nextP()) {
        it->value_setter(true_variable_name(((Symbol*)(it->value))));
        if (!((boolean)(collect_000))) {
          {
            collect_000 = cons(lookup_variable_type(((Symbol*)(it->value))), NIL);
            if (targettypes->the_cons_list == NIL) {
              targettypes->the_cons_list = collect_000;
            }
            else {
              add_cons_to_end_of_cons_list(targettypes->the_cons_list, collect_000);
            }
          }
        }
        else {
          {
            collect_000->rest = cons(lookup_variable_type(((Symbol*)(it->value))), NIL);
            collect_000 = collect_000->rest;
          }
        }
      }
    }
    { Object* otree = NULL;
      Cons* valuestree = NULL;
      List* valuetypes = NULL;

      otree = walk_mv_expression_tree(tree->rest->rest->value, targettypes, SYM_WALK_STELLA_MV_SETQ, valuestree, valuetypes);
      valuetypes->free();
      targettypes->free();
      if (((boolean)(valuestree))) {
        valuestree->first_setter(SYM_WALK_STELLA_PROGN);
        { Object* var = NULL;
          Cons_Iterator* it = valuestree->rest->allocate_iterator();
          Cons* iter_001 = variables;

          while (it->nextP() &&
              (!(iter_001 == NIL))) {
            var = iter_001->value;
            iter_001 = iter_001->rest;
            it->value_setter(listO(3, SYM_WALK_STELLA_SETQ, var, cons(it->value, NIL)));
          }
        }
        tree->rest->rest = NIL;
        return (sys_tree(otree, SGT_WALK_STELLA_VOID, _return1));
      }
      tree->third_setter(otree);
      return (sys_tree(tree, SGT_WALK_STELLA_VOID, _return1));
    }
  }
}

Cons* walk_mv_bind_tree(Cons* tree, Standard_Object*& _return1) {
  if ((tree->length() < 3) ||
      (!consP(tree->rest->value))) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_error();
      if (!(suppress_warningsP())) {
        print_error_context(">> ERROR: ", STANDARD_ERROR);
        *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal 'mv-bind' expression: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
      }
    }
    return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
  }
  { Cons* declarations = ((Cons*)(tree->rest->value));
    Object* mvtree = tree->rest->rest->value;
    Cons* body = tree->rest->rest->rest;
    Cons* variables = NIL;
    Object* decl = NULL;
    List* targettypes = new_list();

    { Cons_Iterator* it = declarations->allocate_iterator();

      while (it->nextP()) {
        decl = it->value;
        { Surrogate* test_value_000 = safe_primary_type(decl);

          if (test_value_000 == SGT_WALK_STELLA_CONS) {
            { Object* decl_000 = decl;
              Cons* decl = ((Cons*)(decl_000));

              if (!(symbolP(decl->value) &&
                  (decl->length() == 2))) {
                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  signal_translation_error();
                  if (!(suppress_warningsP())) {
                    print_error_context(">> ERROR: ", STANDARD_ERROR);
                    *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal variable declaration " << "`" << de_uglify_parse_tree(decl) << "'" << " in 'mv-bind' expression: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
                  }
                }
                return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
              }
              targettypes->push(decl->rest->value->yield_type_specifier());
              decl->rest->rest = cons(SYM_WALK_STELLA_NULL, NIL);
              variables = cons(decl->value, variables);
            }
          }
          else if (subtype_of_symbolP(test_value_000)) {
            { Object* decl_001 = decl;
              Symbol* decl = ((Symbol*)(decl_001));

              targettypes->push(SGT_WALK_STELLA_UNKNOWN);
              it->value_setter(cons(decl, listO(3, SGT_WALK_STELLA_UNKNOWN, SYM_WALK_STELLA_NULL, NIL)));
              variables = cons(decl, variables);
            }
          }
          else {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal variable " << "`" << de_uglify_parse_tree(decl) << "'" << " in 'mv-bind' expression: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
              }
            }
            return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
          }
        }
      }
    }
    variables = variables->reverse();
    { Object* variable = NULL;
      Cons* iter_000 = variables;

      while (!(iter_000 == NIL)) {
        variable = iter_000->value;
        iter_000 = iter_000->rest;
        push_variable_binding(((Symbol*)(variable)), SGT_WALK_STELLA_UNINITIALIZED);
      }
    }
    { Object* omvtree = NULL;
      Cons* valuestree = NULL;
      List* variabletypes = NULL;

      omvtree = walk_mv_expression_tree(mvtree, targettypes->reverse(), SYM_WALK_STELLA_MV_BIND, valuestree, variabletypes);
      { Object* variable = NULL;
        Cons* iter_001 = variables;

        while (!(iter_001 == NIL)) {
          variable = iter_001->value;
          iter_001 = iter_001->rest;
          pop_variable_binding();
        }
      }
      targettypes->free();
      tree->rest->rest->rest = NIL;
      { Cons* decl = NULL;
        Standard_Object* type = NULL;
        Cons* iter_002 = ((Cons*)(declarations));
        Cons* iter_003 = variabletypes->the_cons_list;

        while ((!(iter_002 == NIL)) &&
            (!(iter_003 == NIL))) {
          decl = ((Cons*)(iter_002->value));
          iter_002 = iter_002->rest;
          type = ((Standard_Object*)(iter_003->value));
          iter_003 = iter_003->rest;
          if (SGT_WALK_STELLA_UNKNOWN == decl->rest->value) {
            decl->second_setter(type);
          }
        }
      }
      variabletypes->free();
      if (((boolean)(valuestree))) {
        valuestree->first_setter(SYM_WALK_STELLA_PROGN);
        { Object* var = NULL;
          Cons_Iterator* it = valuestree->rest->allocate_iterator();
          Cons* iter_004 = variables;

          while (it->nextP() &&
              (!(iter_004 == NIL))) {
            var = iter_004->value;
            iter_004 = iter_004->rest;
            it->value_setter(listO(3, SYM_WALK_STELLA_SETQ, var, cons(it->value, NIL)));
          }
        }
        valuestree = ((Cons*)(omvtree));
      }
      else {
        valuestree = listO(3, SYM_WALK_STELLA_MV_SETQ, variables, cons(omvtree, NIL));
      }
      return (walk_a_cons_tree(listO(3, SYM_WALK_STELLA_LET, declarations, cons(valuestree, body->concatenate(NIL, 0))), _return1));
    }
  }
}

Cons* walk_loop_tree(Cons* tree, Standard_Object*& _return1) {
  { 
    BIND_STELLA_SPECIAL(oNOFSPECIALSATLOOPENTRYo, int, oSPECIALVARIABLESTACKo.get()->length());
    tree->rest = walk_list_of_statements(tree->rest);
    return (sys_tree(tree, SGT_WALK_STELLA_VOID, _return1));
  }
}

Cons* walk_loop_exit_tree(Cons* tree, Standard_Object*& _return1) {
  bad_argument_countP(tree, 0);
  if (need_to_unbind_specialsP()) {
    { Cons* unbindtree = yield_loop_exit_special_unbind_tree();

      { 
        BIND_STELLA_SPECIAL(oSPECIALSENABLEDpo, boolean, FALSE);
        return (walk_a_cons_tree(listO(3, SYM_WALK_STELLA_PROGN, unbindtree, cons(tree, NIL)), _return1));
      }
    }
  }
  else {
    _return1 = SGT_WALK_STELLA_VOID;
    return (tree);
  }
}

Cons* walk_while_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    { 
      BIND_STELLA_SPECIAL(oNOFSPECIALSATLOOPENTRYo, int, oSPECIALVARIABLESTACKo.get()->length());
      { Object* test = walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_BOOLEAN, SYM_WALK_STELLA_WHILE, TRUE, dummy1);
        Cons* body = tree->rest->rest;

        tree->second_setter(test);
        if (procedural_expressionP(test)) {
          tree->rest->rest = NIL;
          return (walk_a_cons_tree(listO(3, SYM_WALK_STELLA_LOOP, listO(3, SYM_WALK_STELLA_UNLESS, test, cons(cons(SYM_WALK_STELLA_BREAK, NIL), NIL)), body->concatenate(NIL, 0)), _return1));
        }
        tree->rest->rest = walk_list_of_statements(tree->rest->rest);
        _return1 = SGT_WALK_STELLA_VOID;
        return (tree);
      }
    }
  }
}

Cons* walk_conditional_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    { Object* renamed_Operator = tree->value;
      Object* test = walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_BOOLEAN, ((Symbol*)(renamed_Operator)), TRUE, dummy1);

      tree->second_setter(test);
      if (procedural_expressionP(test)) {
        _return1 = SGT_WALK_STELLA_VOID;
        return (transform_boolean_procedural_expression(tree, ((Cons*)(test)), dummy2));
      }
      { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(renamed_Operator));

        if (test_value_000 == SYM_WALK_STELLA_IF) {
          if (bad_argument_countP(tree, 3)) {
            return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
          }
          tree->third_setter(walk_statement(tree->rest->rest->value, TRUE));
          tree->fourth_setter(walk_statement(tree->fourth(), TRUE));
        }
        else if ((test_value_000 == SYM_WALK_STELLA_WHEN) ||
            (test_value_000 == SYM_WALK_STELLA_UNLESS)) {
          tree->rest->rest = walk_list_of_statements(tree->rest->rest);
        }
        else {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
      }
      _return1 = SGT_WALK_STELLA_VOID;
      return (tree);
    }
  }
}

Cons* walk_choose_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;

    { Object* renamed_Operator = tree->value;
      Object* test = walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_BOOLEAN, ((Symbol*)(renamed_Operator)), TRUE, dummy1);
      Object* temp = NULL;
      Standard_Object* type1 = NULL;
      Standard_Object* type2 = NULL;
      Standard_Object* type = NULL;

      tree->second_setter(test);
      if (vrlet_expressionP(test)) {
        return (percolate_out_vrlet_expression(tree, ((Cons*)(test)), SGT_WALK_STELLA_BOOLEAN, _return1));
      }
      if (bad_argument_countP(tree, 3)) {
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
      }
      temp = walk_expression_tree(tree->rest->rest->value, oTARGETTYPEo.get(), SYM_WALK_STELLA_CHOOSE, TRUE, type1);
      tree->third_setter(temp);
      temp = walk_expression_tree(tree->fourth(), oTARGETTYPEo.get(), SYM_WALK_STELLA_CHOOSE, TRUE, type2);
      tree->fourth_setter(temp);
      if (type1 == SGT_WALK_STELLA_UNKNOWN) {
        type = type2;
        tree->third_setter(coerce_a_tree(tree->rest->rest->value, type1, type, dummy2));
      }
      else if (type2 == SGT_WALK_STELLA_UNKNOWN) {
        type = type1;
        tree->fourth_setter(coerce_a_tree(tree->fourth(), type2, type, dummy3));
      }
      else {
        type = two_argument_least_common_supertype(type1, type2);
      }
      if ((type == SGT_WALK_STELLA_VOID) ||
          ((!sub_type_spec_ofP(type, SGT_WALK_STELLA_OBJECT)) &&
           type_spec_to_class(type)->abstractP)) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Incompatible subexpression types " << "`" << de_uglify_parse_tree(type1) << "'" << " and " << "`" << de_uglify_parse_tree(type2) << "'" << endl << " in 'choose' expression." << "." << endl;
          }
        }
      }
      if (vrlet_expressionP(tree->rest->rest->value) ||
          vrlet_expressionP(tree->fourth())) {
        { Symbol* resultvariable = local_gensym("CHOOSE-VALUE");

          { Object* value_000 = NULL;
            Standard_Object* value_001 = NULL;

            value_000 = walk_a_tree(listO(5, SYM_WALK_STELLA_VRLET, cons(cons(resultvariable, cons(type, cons(SYM_WALK_STELLA_NULL, NIL))), NIL), listO(3, SYM_WALK_STELLA_IF, tree->rest->value, listO(3, listO(3, SYM_WALK_STELLA_SETQ, resultvariable, cons(tree->rest->rest->value, NIL)), listO(3, SYM_WALK_STELLA_SETQ, resultvariable, cons(tree->fourth(), NIL)), NIL)), resultvariable, NIL), value_001);
            _return1 = value_001;
            return (((Cons*)(value_000)));
          }
        }
      }
      if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
      }
      else {
        if ((!(type1 == type)) &&
            (!sub_type_spec_ofP(type1, type2))) {
          tree->third_setter(listO(3, SYM_WALK_STELLA_CAST, tree->rest->rest->value, cons(type, NIL)));
        }
        if ((!(type2 == type)) &&
            (!sub_type_spec_ofP(type2, type1))) {
          tree->fourth_setter(listO(3, SYM_WALK_STELLA_CAST, tree->fourth(), cons(type, NIL)));
        }
      }
      _return1 = type;
      return (tree);
    }
  }
}

Object* walk_boolean_tree(Cons* tree, Standard_Object*& _return1) {
  { Object* otree = help_walk_boolean_tree(tree);

    if (procedural_expressionP(otree)) {
      otree = percolate_out_boolean_vrlet_expressions(((Cons*)(otree)));
    }
    _return1 = SGT_WALK_STELLA_BOOLEAN;
    return (otree);
  }
}

boolean and_or_not_treeP(Object* tree) {
  if (consP(tree)) {
    { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(((Cons*)(tree))->value));

      if ((test_value_000 == SYM_WALK_STELLA_AND) ||
          ((test_value_000 == SYM_WALK_STELLA_OR) ||
           (test_value_000 == SYM_WALK_STELLA_NOT))) {
        return (TRUE);
      }
      else {
        return (FALSE);
      }
    }
  }
  else {
    return (FALSE);
  }
}

Object* help_walk_boolean_tree(Cons* tree) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    { Cons* cursor = tree->rest;
      Object* temp = NULL;

      while (!(cursor == NIL)) {
        cursor->value = (and_or_not_treeP(cursor->value) ? help_walk_boolean_tree(((Cons*)(cursor->value))) : walk_expression_tree(cursor->value, SGT_WALK_STELLA_BOOLEAN, ((Symbol*)(tree->value)), TRUE, dummy1));
        cursor = cursor->rest;
      }
      { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

        if (test_value_000 == SYM_WALK_STELLA_AND) {
          if (tree->rest == NIL) {
            return (TRUE_WRAPPER);
          }
          if (tree->rest->rest == NIL) {
            temp = tree->rest->value;
            return (temp);
          }
        }
        else if (test_value_000 == SYM_WALK_STELLA_OR) {
          if (tree->rest == NIL) {
            return (FALSE_WRAPPER);
          }
          if (tree->rest->rest == NIL) {
            temp = tree->rest->value;
            return (temp);
          }
        }
        else if (test_value_000 == SYM_WALK_STELLA_NOT) {
          if (bad_argument_countP(tree, 1)) {
            return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_BOOLEAN, dummy2));
          }
        }
        else {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
      }
      return (tree);
    }
  }
}

Cons* walk_cond_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    { Object* condition = NULL;
      Cons* iter_000 = tree->rest;

      while (!(iter_000 == NIL)) {
        condition = iter_000->value;
        iter_000 = iter_000->rest;
        if (safe_primary_type(condition) == SGT_WALK_STELLA_CONS) {
          { Object* condition_000 = condition;
            Cons* condition = ((Cons*)(condition_000));

            if (!(condition->value == SYM_WALK_STELLA_OTHERWISE)) {
              condition->value = walk_expression_tree(condition->value, SGT_WALK_STELLA_BOOLEAN, SYM_WALK_STELLA_COND, TRUE, dummy1);
              if (procedural_expressionP(condition->value)) {
                return (walk_a_cons_tree(cond_tree_to_if_tree(tree), _return1));
              }
            }
            condition->rest = walk_list_of_statements(condition->rest);
          }
        }
        else {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal condition in COND statement: " << "`" << de_uglify_parse_tree(condition) << "'" << "." << endl;
            }
          }
          return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
        }
      }
    }
    _return1 = SGT_WALK_STELLA_VOID;
    return (tree);
  }
}

Cons* cond_tree_to_if_tree(Cons* tree) {
  { Standard_Object* dummy1;

    { Object* firstcondition = tree->rest->value;
      Cons* otherconditions = tree->rest->rest;
      Object* firsttest = NULL;

      if (safe_primary_type(firstcondition) == SGT_WALK_STELLA_CONS) {
        { Object* firstcondition_000 = firstcondition;
          Cons* firstcondition = ((Cons*)(firstcondition_000));

          firsttest = firstcondition->value;
          if (firsttest == SYM_WALK_STELLA_OTHERWISE) {
            if (otherconditions == NIL) {
              tree->first_setter(SYM_WALK_STELLA_PROGN);
              tree->rest = firstcondition->rest;
              return (tree);
            }
            else {
              firstcondition->value = SYM_WALK_STELLA_TRUE;
              return (cond_tree_to_if_tree(tree));
            }
          }
          else if (otherconditions == NIL) {
            tree->first_setter(SYM_WALK_STELLA_WHEN);
            tree->rest = firstcondition;
            return (tree);
          }
          else {
            firstcondition->value = SYM_WALK_STELLA_PROGN;
            tree->second_setter(SYM_WALK_STELLA_COND);
            return (listO(3, SYM_WALK_STELLA_IF, firsttest, cons(firstcondition, cons(cond_tree_to_if_tree(tree->rest), NIL))));
          }
        }
      }
      else {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal condition in COND statement: " << "`" << de_uglify_parse_tree(firstcondition) << "'" << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, dummy1));
      }
    }
  }
}

Object* yield_hardcoded_case_symbol_id_or_ids(Object* casetest) {
  if (casetest == SYM_WALK_STELLA_OTHERWISE) {
    return (casetest);
  }
  { Symbol* symbol = NULL;

    { Surrogate* test_value_000 = safe_primary_type(casetest);

      if (test_value_000 == SGT_WALK_STELLA_CONS) {
        { Object* casetest_000 = casetest;
          Cons* casetest = ((Cons*)(casetest_000));

          { Cons_Iterator* it = casetest->allocate_iterator();

            while (it->nextP()) {
              symbol = ((Symbol*)(it->value));
              if (!(symbol->symbol_id != NULL_INTEGER)) {
                symbol = intern_permanent_symbol(symbol->symbol_name);
              }
              it->value_setter(wrap_integer(symbol->symbol_id));
              register_symbol(symbol);
            }
          }
          return (casetest);
        }
      }
      else if (subtype_of_symbolP(test_value_000)) {
        { Object* casetest_001 = casetest;
          Symbol* casetest = ((Symbol*)(casetest_001));

          symbol = casetest;
          if (!(symbol->symbol_id != NULL_INTEGER)) {
            symbol = intern_permanent_symbol(symbol->symbol_name);
          }
          register_symbol(casetest);
          return (wrap_integer(symbol->symbol_id));
        }
      }
      else {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
          throw *new_stella_exception(stream_000->the_string_reader());
        }
      }
    }
  }
}

Cons* walk_hardcoded_symbol_case_tree(Cons* tree, Standard_Object*& _return1) {
  tree->second_setter(listO(3, SYM_WALK_STELLA_SYMBOL_ID, listO(3, SYM_WALK_STELLA_CAST, tree->rest->value, cons(SYM_WALK_STELLA_GENERALIZED_SYMBOL, NIL)), NIL));
  { Cons* cond = NULL;
    Cons* iter_000 = ((Cons*)(tree->rest->rest));

    while (!(iter_000 == NIL)) {
      cond = ((Cons*)(iter_000->value));
      iter_000 = iter_000->rest;
      cond->first_setter(yield_hardcoded_case_symbol_id_or_ids(cond->value));
    }
  }
  return (walk_case_tree(tree, _return1));
}

// Table of specialized type predicates for various types.
// These predicates have to be used instead of `isa?', since they also work
// during bootstrap when only some class objects are defined.
Cons* oTYPE_PREDICATE_TABLEo = NULL;

Cons* yield_isaP_case_test(Surrogate* casekey, Object* testexpression) {
  { Cons* entry = NULL;
    Cons* iter_000 = oTYPE_PREDICATE_TABLEo;

    while (!(iter_000 == NIL)) {
      entry = ((Cons*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (entry->value == casekey) {
        return (cons(entry->rest->value, cons(copy_cons_tree(testexpression), NIL)));
      }
    }
  }
  return (listO(3, SYM_WALK_STELLA_ISAp, copy_cons_tree(testexpression), cons(casekey, NIL)));
}

Cons* yield_subtype_ofP_case_test(Surrogate* casekey, Object* testexpression) {
  { Cons* entry = NULL;
    Cons* iter_000 = oTYPE_PREDICATE_TABLEo;

    while (!(iter_000 == NIL)) {
      entry = ((Cons*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (entry->value == casekey) {
        return (cons(entry->rest->rest->value, cons(copy_cons_tree(testexpression), NIL)));
      }
    }
  }
  return (listO(3, SYM_WALK_STELLA_SUBTYPE_OFp, copy_cons_tree(testexpression), cons(casekey, NIL)));
}

Cons* yield_cond_test(Object* casetest, Object* testvariable, Symbol* equalitytest) {
  if (symbolP(casetest) &&
      (!use_hardcoded_symbolsP())) {
    casetest = listO(3, SYM_WALK_STELLA_TYPED_SYS, register_symbol(((Generalized_Symbol*)(casetest))), cons(SGT_WALK_STELLA_SYMBOL, NIL));
  }
  if (equalitytest == SYM_WALK_STELLA_ISAp) {
    return (yield_isaP_case_test(typify(casetest), testvariable));
  }
  else if (equalitytest == SYM_WALK_STELLA_SUBTYPE_OFp) {
    return (yield_subtype_ofP_case_test(typify(casetest), testvariable));
  }
  else {
    return (cons(equalitytest, cons(copy_cons_tree(testvariable), cons(casetest, NIL))));
  }
}

Object* yield_cond_test_or_tests(Object* casetest, Object* testvariable, Symbol* equalitytest) {
  if (safe_primary_type(casetest) == SGT_WALK_STELLA_CONS) {
    { Object* casetest_000 = casetest;
      Cons* casetest = ((Cons*)(casetest_000));

      { Cons_Iterator* it = casetest->allocate_iterator();

        while (it->nextP()) {
          it->value_setter(yield_cond_test(it->value, testvariable, equalitytest));
        }
      }
      casetest = cons(SYM_WALK_STELLA_OR, casetest);
      return (casetest);
    }
  }
  else {
    if (casetest == SYM_WALK_STELLA_OTHERWISE) {
      return (casetest);
    }
    else {
      return (yield_cond_test(casetest, testvariable, equalitytest));
    }
  }
}

int count_case_test_clauses(Cons* tree, boolean& _return1) {
  { Cons* clauses = ((Cons*)(tree->rest->rest));
    int nofclauses = 0;
    boolean otherwiseP = FALSE;

    { Cons* clause = NULL;
      Cons* iter_000 = clauses;

      while (!(iter_000 == NIL)) {
        clause = ((Cons*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (safe_primary_type(clause->value) == SGT_WALK_STELLA_CONS) {
          nofclauses = nofclauses + ((Cons*)(clause->value))->length();
        }
        else {
          if (clause->value == SYM_WALK_STELLA_OTHERWISE) {
            otherwiseP = TRUE;
          }
          else {
            nofclauses = nofclauses + 1;
          }
        }
      }
    }
    _return1 = otherwiseP;
    return (nofclauses);
  }
}

Cons* attach_unhandled_case_error_clause(Cons* casetree, Object* testtree) {
  { Object* lastclause = casetree->last();

    if (consP(lastclause) &&
        (!(((Cons*)(lastclause))->value == SYM_WALK_STELLA_OTHERWISE))) {
      return (casetree->concatenate(cons(listO(3, SYM_WALK_STELLA_OTHERWISE, listO(3, SYM_WALK_STELLA_ERROR, copy_cons_tree(testtree), cons(wrap_string(" is not a valid case option"), NIL)), NIL), NIL), 0));
    }
    else {
      return (casetree);
    }
  }
}

Cons* walk_non_built_in_case_tree(Cons* tree, Symbol* equalitytest, Standard_Object*& _return1) {
  { boolean dummy1;

    { Object* testexpression = tree->rest->value;
      Object* testvariable = ((atomic_expressionP(testexpression) ||
          (count_case_test_clauses(tree, dummy1) <= 1)) ? tree->rest->value : local_gensym("TEST-VALUE"));
      Cons* conditions = tree->rest->rest;

      { Cons* cond = NULL;
        Cons* iter_000 = ((Cons*)(conditions));

        while (!(iter_000 == NIL)) {
          cond = ((Cons*)(iter_000->value));
          iter_000 = iter_000->rest;
          cond->first_setter(yield_cond_test_or_tests(cond->value, testvariable, equalitytest));
        }
      }
      conditions = attach_unhandled_case_error_clause(conditions, testvariable);
      return (walk_a_cons_tree((eqlP(testexpression, testvariable) ? cons(SYM_WALK_STELLA_COND, conditions->concatenate(NIL, 0)) : listO(4, SYM_WALK_STELLA_LET, cons(cons(testvariable, cons(testexpression, NIL)), NIL), cons(SYM_WALK_STELLA_COND, conditions->concatenate(NIL, 0)), NIL)), _return1));
    }
  }
}

Cons* walk_case_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    { boolean test_value_000 = FALSE;

      test_value_000 = tree->length() >= 3;
      if (test_value_000) {
        { boolean alwaysP_000 = TRUE;

          { Object* clause = NULL;
            Cons* iter_000 = tree->rest->rest;

            while (!(iter_000 == NIL)) {
              clause = iter_000->value;
              iter_000 = iter_000->rest;
              if (!consP(clause)) {
                alwaysP_000 = FALSE;
                break;
              }
            }
          }
          test_value_000 = alwaysP_000;
        }
      }
      test_value_000 = !test_value_000;
      if (test_value_000) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal `case' statement: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
      }
    }
    { Cons* clauses = ((Cons*)(tree->rest->rest));
      Object* firsttesttree = (consP(((Cons*)(clauses->value))->value) ? ((Cons*)(((Cons*)(clauses->value))->value))->value : ((Cons*)(clauses->value))->value);
      Surrogate* caseconstanttype = SGT_WALK_STELLA_UNKNOWN;

      { Surrogate* test_value_001 = safe_primary_type(firsttesttree);

        if (subtype_of_symbolP(test_value_001)) {
          { Object* firsttesttree_000 = firsttesttree;
            Symbol* firsttesttree = ((Symbol*)(firsttesttree_000));

            caseconstanttype = SGT_WALK_STELLA_GENERALIZED_SYMBOL;
          }
        }
        else if (subtype_of_surrogateP(test_value_001)) {
          { Object* firsttesttree_001 = firsttesttree;
            Surrogate* firsttesttree = ((Surrogate*)(firsttesttree_001));

            caseconstanttype = SGT_WALK_STELLA_GENERALIZED_SYMBOL;
          }
        }
        else if (subtype_of_keywordP(test_value_001)) {
          { Object* firsttesttree_002 = firsttesttree;
            Keyword* firsttesttree = ((Keyword*)(firsttesttree_002));

            caseconstanttype = SGT_WALK_STELLA_GENERALIZED_SYMBOL;
          }
        }
        else if (subtype_of_stringP(test_value_001)) {
          { Object* firsttesttree_003 = firsttesttree;
            String_Wrapper* firsttesttree = ((String_Wrapper*)(firsttesttree_003));

            caseconstanttype = SGT_WALK_STELLA_STRING;
          }
        }
        else if (subtype_of_characterP(test_value_001)) {
          { Object* firsttesttree_004 = firsttesttree;
            Character_Wrapper* firsttesttree = ((Character_Wrapper*)(firsttesttree_004));

            caseconstanttype = SGT_WALK_STELLA_CHARACTER;
          }
        }
        else if (subtype_of_integerP(test_value_001)) {
          { Object* firsttesttree_005 = firsttesttree;
            Integer_Wrapper* firsttesttree = ((Integer_Wrapper*)(firsttesttree_005));

            caseconstanttype = SGT_WALK_STELLA_INTEGER;
          }
        }
        else if (subtype_of_floatP(test_value_001)) {
          { Object* firsttesttree_006 = firsttesttree;
            Float_Wrapper* firsttesttree = ((Float_Wrapper*)(firsttesttree_006));

            caseconstanttype = SGT_WALK_STELLA_FLOAT;
          }
        }
        else {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "`case' statement contains illegal constant of type " << "`" << firsttesttree->primary_type() << "'" << endl << "    " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
            }
          }
          return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
        }
      }
      tree->second_setter(walk_expression_tree(tree->rest->value, caseconstanttype, SYM_WALK_STELLA_CASE, TRUE, dummy1));
      if (vrlet_expressionP(tree->rest->value)) {
        return (percolate_out_vrlet_expression(tree, ((Cons*)(tree->rest->value)), caseconstanttype, _return1));
      }
      if ((caseconstanttype == SGT_WALK_STELLA_GENERALIZED_SYMBOL) &&
          use_hardcoded_symbolsP()) {
        return (walk_hardcoded_symbol_case_tree(tree, _return1));
      }
      if (!((caseconstanttype == SGT_WALK_STELLA_INTEGER) ||
          (caseconstanttype == SGT_WALK_STELLA_CHARACTER))) {
        return (walk_non_built_in_case_tree(tree, ((caseconstanttype == SGT_WALK_STELLA_STRING) ? SYM_WALK_STELLA_STRING_EQLp : SYM_WALK_STELLA_EQLp), _return1));
      }
      if (!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP)) {
        tree = attach_unhandled_case_error_clause(tree, tree->rest->value);
      }
      { Cons* cursor = tree->rest->rest;
        Cons* conditiontree = NULL;

        while (!(cursor == NIL)) {
          conditiontree = ((Cons*)(cursor->value));
          if (!(conditiontree->value == SYM_WALK_STELLA_OTHERWISE)) {
            conditiontree->first_setter((consP(conditiontree->value) ? walk_list_of_trees(((Cons*)(conditiontree->value))) : walk_a_tree(conditiontree->value, dummy2)));
          }
          conditiontree->rest = walk_list_of_statements(conditiontree->rest);
          cursor = cursor->rest;
        }
        _return1 = SGT_WALK_STELLA_VOID;
        return (tree);
      }
    }
  }
}

Cons* walk_typecase_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    { Object* testtree = tree->rest->value;
      Standard_Object* testexpressiontype = NULL;
      Symbol* testvariable = NULL;
      Cons* expandedcases = NIL;
      boolean needcastP = symbolP(testtree);
      List* precedingclausetypes = new_list();

      if (needcastP) {
        testvariable = ((Symbol*)(testtree));
      }
      testtree = walk_a_tree(testtree, testexpressiontype);
      tree->second_setter(testtree);
      testexpressiontype = type_spec_to_base_type(testexpressiontype);
      if (unknown_typeP(((Surrogate*)(testexpressiontype)))) {
        testexpressiontype = SGT_WALK_STELLA_UNKNOWN;
      }
      tree->second_setter(sys_tree(testtree, testexpressiontype, dummy1));
      if (needcastP) {
        { Object* casetree = NULL;
          Cons* iter_000 = tree->rest->rest;

          while (!(iter_000 == NIL)) {
            casetree = iter_000->value;
            iter_000 = iter_000->rest;
            if (safe_primary_type(casetree) == SGT_WALK_STELLA_CONS) {
              { Object* casetree_000 = casetree;
                Cons* casetree = ((Cons*)(casetree_000));

                { Object* typetree = casetree->value;

                  if (safe_primary_type(typetree) == SGT_WALK_STELLA_CONS) {
                    { Object* typetree_000 = typetree;
                      Cons* typetree = ((Cons*)(typetree_000));

                      casetree->first_setter(typetree->value);
                      expandedcases = cons(casetree, expandedcases);
                      { Object* type = NULL;
                        Cons* iter_001 = typetree->rest;

                        while (!(iter_001 == NIL)) {
                          type = iter_001->value;
                          iter_001 = iter_001->rest;
                          expandedcases = cons(cons(type, ((Cons*)(((Cons*)(copy_cons_tree(casetree->rest)))))->concatenate(NIL, 0)), expandedcases);
                        }
                      }
                    }
                  }
                  else {
                    expandedcases = cons(casetree, expandedcases);
                  }
                }
              }
            }
            else {
              expandedcases = cons(casetree, expandedcases);
            }
          }
        }
        tree->rest->rest = expandedcases->reverse();
      }
      { Object* casetree = NULL;
        Cons* iter_002 = tree->rest->rest;

        while (!(iter_002 == NIL)) {
          casetree = iter_002->value;
          iter_002 = iter_002->rest;
          if (safe_primary_type(casetree) == SGT_WALK_STELLA_CONS) {
            { Object* casetree_001 = casetree;
              Cons* casetree = ((Cons*)(casetree_001));

              { Object* typetree = casetree->value;

                { Surrogate* test_value_000 = safe_primary_type(typetree);

                  if (subtype_of_symbolP(test_value_000)) {
                    { Object* typetree_001 = typetree;
                      Symbol* typetree = ((Symbol*)(typetree_001));

                      if (!(string_eqlP(typetree->symbol_name, "OTHERWISE"))) {
                        { Surrogate* type = type_spec_to_base_type(typetree->yield_type_specifier());

                          verify_typecase_clause_type(((Surrogate*)(testexpressiontype)), type, precedingclausetypes, casetree);
                          casetree->first_setter(type);
                          type = type->type_to_wrapped_type();
                          if (needcastP &&
                              (!(casetree->rest == NIL))) {
                            casetree->rest = yield_casted_typecase_clause_trees(testvariable, type, casetree->rest);
                          }
                        }
                      }
                    }
                  }
                  else if (subtype_of_surrogateP(test_value_000)) {
                    { Object* typetree_002 = typetree;
                      Surrogate* typetree = ((Surrogate*)(typetree_002));

                      if (!(string_eqlP(typetree->symbol_name, "OTHERWISE"))) {
                        { Surrogate* type = type_spec_to_base_type(typetree->yield_type_specifier());

                          verify_typecase_clause_type(((Surrogate*)(testexpressiontype)), type, precedingclausetypes, casetree);
                          casetree->first_setter(type);
                          type = type->type_to_wrapped_type();
                          if (needcastP &&
                              (!(casetree->rest == NIL))) {
                            casetree->rest = yield_casted_typecase_clause_trees(testvariable, type, casetree->rest);
                          }
                        }
                      }
                    }
                  }
                  else if (test_value_000 == SGT_WALK_STELLA_CONS) {
                    { Object* typetree_003 = typetree;
                      Cons* typetree = ((Cons*)(typetree_003));

                      { Cons_Iterator* it = typetree->allocate_iterator();

                        while (it->nextP()) {
                          if (consP(it->value)) {
                            { 
                              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                              signal_translation_error();
                              if (!(suppress_warningsP())) {
                                print_error_context(">> ERROR: ", STANDARD_ERROR);
                                *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal type expression in 'typecase': " << "`" << de_uglify_parse_tree(typetree) << "'" << "." << endl;
                              }
                            }
                            return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
                          }
                          it->value_setter(type_spec_to_base_type(it->value->yield_type_specifier()));
                          verify_typecase_clause_type(((Surrogate*)(testexpressiontype)), ((Surrogate*)(it->value)), precedingclausetypes, casetree);
                        }
                      }
                    }
                  }
                  else {
                    { 
                      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                      signal_translation_error();
                      if (!(suppress_warningsP())) {
                        print_error_context(">> ERROR: ", STANDARD_ERROR);
                        *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal type expression in 'typecase': " << "`" << de_uglify_parse_tree(typetree) << "'" << "." << endl;
                      }
                    }
                    return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
                  }
                }
              }
            }
          }
          else {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal case expression in 'typecase': " << "`" << de_uglify_parse_tree(casetree) << "'" << "." << endl;
              }
            }
            return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
          }
        }
      }
      return (walk_non_built_in_case_tree(listO(3, SYM_WALK_STELLA_CASE, listO(3, SYM_WALK_STELLA_SAFE_PRIMARY_TYPE, tree->rest->value, NIL), tree->rest->rest->concatenate(NIL, 0)), SYM_WALK_STELLA_SUBTYPE_OFp, _return1));
    }
  }
}

Cons* yield_casted_typecase_clause_trees(Symbol* testvariable, Standard_Object* clausetype, Cons* clausetrees) {
  { Standard_Object* dummy1;

    { Standard_Object* testvariabletype = lookup_variable_type(testvariable);
      boolean needrealcastP = !(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP);
      Symbol* auxvariable = (((!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA)) &&
          needrealcastP) ? local_gensym(testvariable->symbol_name) : ((Symbol*)(NULL)));
      Cons* castedclausetrees = NULL;

      if (!needrealcastP) {
        push_variable_binding(testvariable, clausetype);
        castedclausetrees = cons(listO(3, SYM_WALK_STELLA_VOID_SYS, cons(SYM_WALK_STELLA_PROGN, walk_list_of_statements(clausetrees)->concatenate(NIL, 0)), NIL), NIL);
        pop_variable_binding();
      }
      else {
        if (((boolean)(auxvariable))) {
          push_variable_binding(auxvariable, testvariabletype);
        }
        castedclausetrees = cons(listO(3, SYM_WALK_STELLA_LET, (((boolean)(auxvariable)) ? listO(3, cons(auxvariable, cons(testvariable, NIL)), cons(testvariable, cons(clausetype, cons(auxvariable, NIL))), NIL) : cons(cons(testvariable, cons(listO(3, SYM_WALK_STELLA_SAFE_CAST, sys_tree(walk_without_type_tree(testvariable), clausetype, dummy1), cons(clausetype, NIL)), NIL)), NIL)), clausetrees->concatenate(NIL, 0)), NIL);
        if (((boolean)(auxvariable))) {
          pop_variable_binding();
        }
      }
      return (castedclausetrees);
    }
  }
}

void verify_typecase_clause_type(Surrogate* testtype, Surrogate* clausetype, List* precedingclausetypes, Cons* clause) {
  { Standard_Object* dummy1;

    { Surrogate* shadowingtype = NULL;

      clausetype = clausetype->type_to_wrapped_type();
      if (unknown_typeP(clausetype)) {
        {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Reference to undefined type in `typecase': " << "`" << de_uglify_parse_tree(clausetype) << "'" << "." << endl;
            }
          }
          clause->rest = walk_dont_call_me_tree(clause->rest, SGT_WALK_STELLA_VOID, dummy1);
        }
      }
      else {
        if ((!(testtype == SGT_WALK_STELLA_UNKNOWN)) &&
            (!sub_type_spec_ofP(clausetype, testtype))) {
          {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_warning();
              if (!(suppress_warningsP())) {
                print_error_context(">> WARNING: ", STANDARD_WARNING);
                *(STANDARD_WARNING->native_stream) << endl << " " << "Type " << "`" << de_uglify_parse_tree(clausetype) << "'" << " of `typecase' clause " << endl << "    " << "`" << de_uglify_parse_tree(clause) << "'" << endl << " is not a subtype of the test expression type " << "`" << de_uglify_parse_tree(testtype) << "'" << "." << endl;
              }
            }
          }
        }
        else {
          { 
            { Surrogate* value_000 = NULL;

              { Surrogate* type = NULL;
                Cons* iter_000 = precedingclausetypes->the_cons_list;

                while (!(iter_000 == NIL)) {
                  type = ((Surrogate*)(iter_000->value));
                  iter_000 = iter_000->rest;
                  if (subtype_ofP(clausetype, type)) {
                    value_000 = type;
                    break;
                  }
                }
              }
              shadowingtype = value_000;
            }
            if (((boolean)(shadowingtype))) {
              {
                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  signal_translation_warning();
                  if (!(suppress_warningsP())) {
                    print_error_context(">> WARNING: ", STANDARD_WARNING);
                    *(STANDARD_WARNING->native_stream) << endl << " " << "Type " << "`" << de_uglify_parse_tree(clausetype) << "'" << " of `typecase' clause " << endl << "    " << "`" << de_uglify_parse_tree(clause) << "'" << endl << " is shadowed by type " << "`" << de_uglify_parse_tree(shadowingtype) << "'" << "." << endl;
                  }
                }
              }
            }
            else {
              {
                precedingclausetypes->push(clausetype);
              }
            }
          }
        }
      }
    }
  }
}

Cons* finish_walking_eql_tree(Cons* tree, Surrogate* type1, Surrogate* type2, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    if (subtype_ofP(type1, SGT_WALK_STELLA_STANDARD_OBJECT) ||
        (subtype_ofP(type2, SGT_WALK_STELLA_STANDARD_OBJECT) ||
         (((type1 == SGT_WALK_STELLA_BOOLEAN_WRAPPER) &&
        (type2 == SGT_WALK_STELLA_BOOLEAN_WRAPPER)) ||
          (((!subtype_ofP(type1, SGT_WALK_STELLA_OBJECT)) &&
        (!subtype_ofP(type1, SGT_WALK_STELLA_LITERAL))) ||
           ((!subtype_ofP(type2, SGT_WALK_STELLA_OBJECT)) &&
            (!subtype_ofP(type2, SGT_WALK_STELLA_LITERAL))))))) {
      tree->first_setter(SYM_WALK_STELLA_EQp);
      tree->second_setter(sys_tree(tree->rest->value, type1, dummy1));
      tree->third_setter(sys_tree(tree->rest->rest->value, type2, dummy2));
      _return1 = SGT_WALK_STELLA_BOOLEAN;
      return (tree);
    }
    if (subtype_ofP(type1, SGT_WALK_STELLA_LITERAL) &&
        subtype_ofP(type2, SGT_WALK_STELLA_LITERAL)) {
      if (type1 == SGT_WALK_STELLA_STRING) {
        tree->first_setter(SYM_WALK_STELLA_STRING_EQLp);
      }
      else {
        tree->first_setter(SYM_WALK_STELLA_EQp);
      }
      return (walk_a_cons_tree(tree, _return1));
    }
    if ((!subtype_ofP(type1, SGT_WALK_STELLA_LITERAL)) &&
        (!subtype_ofP(type2, SGT_WALK_STELLA_LITERAL))) {
      { Object* otree = NULL;
        Standard_Object* otype = NULL;

        otree = walk_call_slot_tree(tree, otype);
        _return1 = otype;
        return (((Cons*)(otree)));
      }
    }
    if (!subtype_ofP(type2, SGT_WALK_STELLA_LITERAL)) {
      { Object* temparg = tree->rest->rest->value;
        Surrogate* temptype = type2;

        tree->third_setter(tree->rest->value);
        tree->second_setter(temparg);
        type2 = type1;
        type1 = temptype;
      }
    }
    if (type2 == SGT_WALK_STELLA_BOOLEAN) {
      tree->first_setter(SYM_WALK_STELLA_EQL_TO_BOOLEANp);
    }
    else if (type2 == SGT_WALK_STELLA_INTEGER) {
      tree->first_setter(SYM_WALK_STELLA_EQL_TO_INTEGERp);
    }
    else if (type2 == SGT_WALK_STELLA_FLOAT) {
      tree->first_setter(SYM_WALK_STELLA_EQL_TO_FLOATp);
    }
    else if (type2 == SGT_WALK_STELLA_STRING) {
      tree->first_setter(SYM_WALK_STELLA_EQL_TO_STRINGp);
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << type2 << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
    _return1 = SGT_WALK_STELLA_BOOLEAN;
    return (((Cons*)(walk_without_type_tree(tree))));
  }
}

Cons* finish_walking_equal_tree(Cons* tree, Surrogate* type1, Surrogate* type2, Standard_Object*& _return1) {
  {
    tree = tree;
    type1 = type1;
    type2 = type2;
  }
  { 
    BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
    signal_translation_error();
    if (!(suppress_warningsP())) {
      print_error_context(">> ERROR: ", STANDARD_ERROR);
      *(STANDARD_ERROR->native_stream) << endl << " " << "finish-walking-equal-tree: NOT YET IMPLEMENTED." << "." << endl;
    }
  }
  _return1 = SGT_WALK_STELLA_BOOLEAN;
  return (NULL);
}

Object* walk_equality_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;
    Standard_Object* dummy4;

    { Object* arg1 = NULL;
      Object* arg2 = NULL;
      Standard_Object* type1 = NULL;
      Standard_Object* type2 = NULL;

      arg1 = walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_UNKNOWN, ((Symbol*)(tree->value)), TRUE, type1);
      arg2 = walk_expression_tree(tree->rest->rest->value, SGT_WALK_STELLA_UNKNOWN, ((Symbol*)(tree->value)), TRUE, type2);
      tree->second_setter(arg1);
      tree->third_setter(arg2);
      if (vrlet_expressionP(arg1)) {
        return (walk_a_tree(percolate_out_vrlet_expression(tree, ((Cons*)(arg1)), SGT_WALK_STELLA_BOOLEAN, dummy1), _return1));
      }
      if (vrlet_expressionP(arg2)) {
        return (walk_a_tree(percolate_out_vrlet_expression(tree, ((Cons*)(arg2)), SGT_WALK_STELLA_BOOLEAN, dummy2), _return1));
      }
      type1 = canonical_type(type_spec_to_base_type(type1));
      type2 = canonical_type(type_spec_to_base_type(type2));
      if (!((type1 == SGT_WALK_STELLA_UNKNOWN) ||
          ((type2 == SGT_WALK_STELLA_UNKNOWN) ||
           (subtype_ofP(((Surrogate*)(type1)), ((Surrogate*)(type2))) ||
            (subtype_ofP(((Surrogate*)(type2)), ((Surrogate*)(type1))) ||
             (subtype_ofP(((Surrogate*)(type1))->type_to_wrapped_type(), ((Surrogate*)(type2))) ||
              subtype_ofP(((Surrogate*)(type2))->type_to_wrapped_type(), ((Surrogate*)(type1))))))))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_warning();
          if (!(suppress_warningsP())) {
            print_error_context(">> WARNING: ", STANDARD_WARNING);
            *(STANDARD_WARNING->native_stream) << endl << " " << "This equality test will always fail due to disjoint " << "argument types:" << endl << "    " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
          }
        }
      }
      { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

        if ((test_value_000 == SYM_WALK_STELLA_EQp) ||
            (test_value_000 == SYM_WALK_STELLA_e)) {
          tree->first_setter(SYM_WALK_STELLA_EQp);
          tree->second_setter(sys_tree(tree->rest->value, type1, dummy3));
          tree->third_setter(sys_tree(tree->rest->rest->value, type2, dummy4));
          _return1 = SGT_WALK_STELLA_BOOLEAN;
          return (tree);
        }
        else if (test_value_000 == SYM_WALK_STELLA_EQLp) {
          return (finish_walking_eql_tree(tree, ((Surrogate*)(type1)), ((Surrogate*)(type2)), _return1));
        }
        else {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
      }
    }
  }
}

List* oNUMERIC_TYPE_HIERARCHYo = NULL;

Surrogate* more_general_numeric_type(Surrogate* type1, Surrogate* type2) {
  { int position1 = oNUMERIC_TYPE_HIERARCHYo->position(type1, 0);
    int position2 = oNUMERIC_TYPE_HIERARCHYo->position(type2, 0);

    if ((position1 == NULL_INTEGER) ||
        (position2 == NULL_INTEGER)) {
      return (SGT_WALK_STELLA_NUMBER);
    }
    if (position1 > position2) {
      return (type1);
    }
    else {
      return (type2);
    }
  }
}

Cons* yield_arithmetic_operator_call_tree(Symbol* renamed_Operator, Cons* operands) {
  return (listO(4, SYM_WALK_STELLA_SYS_CALL_METHOD, SGT_WALK_STELLA_NUMBER, renamed_Operator, operands->concatenate(NIL, 0)));
}

Cons* yield_nested_arithmetic_operator_call_tree(Symbol* renamed_Operator, Cons* operands) {
  { int nofoperands = operands->length();
    Cons* parenthesizedtree = yield_arithmetic_operator_call_tree(renamed_Operator, cons_list(2, operands->value, operands->rest->value));

    { int i = NULL_INTEGER;
      int iter_000 = 2;
      int upper_bound_000 = nofoperands - 1;

      while (iter_000 <= upper_bound_000) {
        i = iter_000;
        iter_000 = iter_000 + 1;
        parenthesizedtree = yield_arithmetic_operator_call_tree(renamed_Operator, cons_list(2, parenthesizedtree, operands->nth(i)));
      }
    }
    return (parenthesizedtree);
  }
}

Object* walk_arithmetic_tree(Cons* tree, Surrogate*& _return1) {
  { Object* renamed_Operator = tree->value;
    Cons* operands = tree->rest;
    int nofoperands = operands->length();
    Object* operand = NULL;
    Standard_Object* operandtype = NULL;
    Surrogate* returntype = SGT_WALK_STELLA_INTEGER;

    switch (nofoperands) {
      case 0: 
        { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(renamed_Operator));

          if (test_value_000 == SYM_WALK_STELLA_i) {
            _return1 = SGT_WALK_STELLA_INTEGER;
            return (wrap_integer(0));
          }
          else if (test_value_000 == SYM_WALK_STELLA_o) {
            _return1 = SGT_WALK_STELLA_INTEGER;
            return (wrap_integer(1));
          }
          else {
            bad_argument_countP(tree, 1);
            { Cons* value_000 = NULL;
              Standard_Object* value_001 = NULL;

              value_000 = walk_dont_call_me_tree(tree, SGT_WALK_STELLA_NUMBER, value_001);
              _return1 = ((Surrogate*)(value_001));
              return (value_000);
            }
          }
        }
      break;
      case 1: 
        { Generalized_Symbol* test_value_001 = ((Generalized_Symbol*)(renamed_Operator));

          if ((test_value_001 == SYM_WALK_STELLA_i) ||
              (test_value_001 == SYM_WALK_STELLA_o)) {
            { Object* value_002 = NULL;
              Standard_Object* value_003 = NULL;

              value_002 = walk_expression_tree(operands->value, SGT_WALK_STELLA_NUMBER, ((Symbol*)(renamed_Operator)), FALSE, value_003);
              _return1 = ((Surrogate*)(value_003));
              return (value_002);
            }
          }
          else if ((test_value_001 == SYM_WALK_STELLA__) ||
              (test_value_001 == SYM_WALK_STELLA_s)) {
            operands->rest = cons(operands->value, NIL);
            operands->first_setter(((renamed_Operator == SYM_WALK_STELLA__) ? wrap_integer(0) : wrap_integer(1)));
          }
          else {
            { Output_String_Stream* stream_000 = new_output_string_stream();

              *(stream_000->native_stream) << "`" << test_value_001 << "'" << " is not a valid case option";
              throw *new_stella_exception(stream_000->the_string_reader());
            }
          }
        }
      break;
      default:
      break;
    }
    { Cons_Iterator* it = operands->allocate_iterator();

      while (it->nextP()) {
        operand = walk_expression_tree(it->value, SGT_WALK_STELLA_NUMBER, ((Symbol*)(renamed_Operator)), FALSE, operandtype);
        returntype = more_general_numeric_type(returntype, type_spec_to_base_type(operandtype));
        it->value_setter(operand);
      }
    }
    if ((renamed_Operator == SYM_WALK_STELLA_s) &&
        (returntype == SGT_WALK_STELLA_INTEGER)) {
      if (wrapperP(operands->value)) {
        operands->first_setter(wrap_float(((Integer_Wrapper*)(operands->value))->wrapper_value * 1.0));
      }
      else if (wrapperP(operands->rest->value)) {
        operands->second_setter(wrap_float(((Integer_Wrapper*)(operands->rest->value))->wrapper_value * 1.0));
      }
      else {
        operands = cons(listO(3, SYM_WALK_STELLA_CAST, operands->value, cons(SGT_WALK_STELLA_FLOAT, NIL)), operands->rest->concatenate(NIL, 0));
        nofoperands = nofoperands + 1;
      }
      returntype = SGT_WALK_STELLA_FLOAT;
    }
    if ((nofoperands > 2) &&
        (!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP))) {
      { Cons* value_004 = NULL;
        Standard_Object* value_005 = NULL;

        value_004 = sys_tree(yield_nested_arithmetic_operator_call_tree(((Symbol*)(renamed_Operator)), operands), returntype, value_005);
        _return1 = ((Surrogate*)(value_005));
        return (value_004);
      }
    }
    else {
      { Cons* value_006 = NULL;
        Standard_Object* value_007 = NULL;

        value_006 = sys_tree(yield_arithmetic_operator_call_tree(((Symbol*)(renamed_Operator)), operands), returntype, value_007);
        _return1 = ((Surrogate*)(value_007));
        return (value_006);
      }
    }
  }
}

Object* walk_quoted_tree(Cons* tree, Standard_Object*& _return1) {
  if (bad_argument_countP(tree, 1)) {
    return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
  }
  { Object* argument = tree->rest->value;
    boolean quotedliteralP = (!consP(argument)) ||
        (argument == NIL);
    char* objectname = NULL;

    if (quotedliteralP) {
      { Object* otree = NULL;
        Standard_Object* otype = NULL;

        otree = walk_a_tree(bquotify(argument), otype);
        _return1 = otype;
        return (otree);
      }
    }
    objectname = name_quoted_tree(((Cons*)(argument)));
    return (walk_a_tree(listO(3, SYM_WALK_STELLA_GET_QUOTED_TREE, wrap_string(objectname), cons(wrap_string(oMODULEo.get()->module_full_name), NIL)), _return1));
  }
}

Object* walk_bquote_tree(Cons* tree, Standard_Object*& _return1) {
  if (bad_argument_countP(tree, 1)) {
    return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
  }
  { Object* otree = NULL;
    Standard_Object* otype = NULL;

    otree = walk_a_tree(simplify_bquote_tree(expand_bquote_tree(tree->rest->value)), otype);
    _return1 = otype;
    return (otree);
  }
}

Object* expand_macro(Cons* tree) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    { Symbol* macroname = ((Symbol*)(tree->value));
      Method_Slot* expanderobject = lookup_macro(macroname);
      cpp_function_code expandercode = expanderobject->function_code;
      Cons* args = tree->rest;
      Object* expansion = NULL;

      if (expandercode == NULL) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Macro expander code for " << "`" << de_uglify_parse_tree(macroname) << "'" << " is not available" << "." << endl;
          }
        }
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, dummy1));
      }
      if (((Boolean_Wrapper*)(dynamic_slot_value(expanderobject->dynamic_slots, SYM_WALK_STELLA_METHOD_BODY_ARGUMENTp, FALSE_WRAPPER)))->wrapper_value) {
        switch (expanderobject->method_argument_count()) {
          case 1: 
            expansion = ((Object*  (*) (Cons*))expandercode)(args);
            tree->rest = NIL;
          break;
          case 2: 
            expansion = ((Object*  (*) (Object*, Cons*))expandercode)(args->value, args->rest);
            args->rest = NIL;
          break;
          case 3: 
            expansion = ((Object*  (*) (Object*, Object*, Cons*))expandercode)(args->value, args->rest->value, args->rest->rest);
            args->rest->rest = NIL;
          break;
          case 4: 
            expansion = ((Object*  (*) (Object*, Object*, Object*, Cons*))expandercode)(args->value, args->rest->value, args->rest->rest->value, args->nth_rest(3));
            args->nth_rest_setter(NIL, 3);
          break;
          case 5: 
            expansion = ((Object*  (*) (Object*, Object*, Object*, Object*, Cons*))expandercode)(args->value, args->rest->value, args->rest->rest->value, args->fourth(), args->nth_rest(4));
            args->nth_rest_setter(NIL, 4);
          break;
          default:
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Too many arguments in definition of macro " << "`" << de_uglify_parse_tree(macroname) << "'" << "." << endl;
              }
            }
            return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, dummy2));
        }
      }
      else {
        expansion = apply(expandercode, tree->rest);
      }
      return (expansion);
    }
  }
}

Object* walk_macro_tree(Cons* tree, Standard_Object*& _return1) {
  return (walk_a_tree(expand_macro(tree), _return1));
}

Cons* walk_print_stream_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    if (tree->length() <= 2) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Missing arguments in print statement: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
    }
    { Object* renamed_Operator = tree->value;
      Object* streamtree = tree->rest->value;
      Object* otree = NULL;
      Standard_Object* otype = NULL;

      if (keywordP(streamtree)) {
        if (!((streamtree == KWD_WALK_WARN) ||
            ((streamtree == KWD_WALK_ERROR) ||
             (streamtree == KWD_WALK_CONTINUABLE_ERROR)))) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal stream argument in print statement: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
            }
          }
          return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
        }
      }
      else {
        otree = walk_expression_tree(streamtree, SGT_WALK_STELLA_NATIVE_OUTPUT_STREAM, ((Symbol*)(tree->value)), FALSE, otype);
        tree->first_setter(SYM_WALK_STELLA_PRINT_NATIVE_STREAM);
        tree->second_setter(otree);
      }
      { Cons_Iterator* it = tree->rest->rest->allocate_iterator();

        while (it->nextP()) {
          otree = walk_expression_tree(it->value, SGT_WALK_STELLA_UNKNOWN, ((Symbol*)(renamed_Operator)), FALSE, otype);
          it->value_setter(otree);
          if (!(otree == SYM_WALK_STELLA_EOL)) {
            { Surrogate* test_value_000 = safe_primary_type(otree);

              if (subtype_of_stringP(test_value_000)) {
                { Object* otree_000 = otree;
                  String_Wrapper* otree = ((String_Wrapper*)(otree_000));

                }
              }
              else if (subtype_of_integerP(test_value_000)) {
                { Object* otree_001 = otree;
                  Integer_Wrapper* otree = ((Integer_Wrapper*)(otree_001));

                  it->value_setter(wrap_string(stringify(otree)));
                }
              }
              else if (subtype_of_floatP(test_value_000)) {
                { Object* otree_002 = otree;
                  Float_Wrapper* otree = ((Float_Wrapper*)(otree_002));

                  it->value_setter(wrap_string(stringify(otree)));
                }
              }
              else if (subtype_of_characterP(test_value_000)) {
                { Object* otree_003 = otree;
                  Character_Wrapper* otree = ((Character_Wrapper*)(otree_003));

                  it->value_setter(wrap_string(make_string(1, otree->wrapper_value)));
                }
              }
              else {
                it->value_setter(sys_tree(otree, otype, dummy1));
              }
            }
          }
        }
      }
      _return1 = SGT_WALK_STELLA_VOID;
      return (tree);
    }
  }
}

boolean warn_about_undefined_methodsP() {
  return (oCURRENT_ENVIRONMENT_FEATURESo->memberP(KWD_WALK_WARN_ABOUT_UNDEFINED_METHODS));
}

boolean warn_about_missing_methodsP() {
  return ((!(oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP)) ||
      oCURRENT_ENVIRONMENT_FEATURESo->memberP(KWD_WALK_WARN_ABOUT_MISSING_METHODS));
}

Cons* walk_undefined_slot_tree(Cons* tree, Standard_Object* firstargtype, boolean warnP, Surrogate*& _return1) {
  { Standard_Object* dummy1;

    { Cons* cursor = tree->rest->rest;
      Surrogate* type = type_spec_to_base_type(firstargtype);
      Symbol* operatorname = ((Symbol*)(tree->value));

      while (!(cursor == NIL)) {
        cursor->value = walk_expression_tree(cursor->value, NULL, operatorname, TRUE, dummy1);
        cursor = cursor->rest;
      }
      if (warnP &&
          (!symbol_common_lispP(operatorname))) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Reference to undefined method or function named " << "`" << de_uglify_parse_tree(operatorname) << "'" << endl << "   in the expression:" << endl << "   " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
          }
        }
      }
      _return1 = SGT_WALK_STELLA_UNKNOWN;
      return (listO(3, SYM_WALK_STELLA_SYS_CALL_METHOD, type, tree->concatenate(NIL, 0)));
    }
  }
}

Cons* walk_slot_on_abstract_type_tree(Cons* tree, Surrogate* abstracttype, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    { Object* methodname = tree->value;
      Method_Slot* prototypemethod = find_prototype_method(((Symbol*)(methodname)), abstracttype);
      boolean undefinedmethodP = TRUE;
      Object* otree = NULL;
      Standard_Object* otype = NULL;

      if (((boolean)(prototypemethod))) {
        undefinedmethodP = FALSE;
        if (warn_about_missing_methodsP()) {
          { int nofsubs = NULL_INTEGER;
            int nofmethods = NULL_INTEGER;
            boolean compatibleP = FALSE;

            nofsubs = compatible_real_methods(prototypemethod, abstracttype, nofmethods, compatibleP);
            if ((nofsubs == nofmethods) &&
                compatibleP) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_OUTPUT);
                  cout << endl << " " << "Inferred the method " << "`" << de_uglify_parse_tree(methodname) << "'" << " on the abstract" << endl << "   type " << "`" << de_uglify_parse_tree(abstracttype) << "'" << ", since " << "`" << de_uglify_parse_tree(methodname) << "'" << " is implemented" << endl << "   on all non-abstract subtypes of " << "`" << de_uglify_parse_tree(abstracttype) << "'" << "." << endl;
                }
              }
              tree->second_setter(sys_tree(tree->rest->value, prototypemethod->slot_owner, dummy1));
              otree = walk_call_slot_tree(tree, otype);
            }
            else if (nofsubs == nofmethods) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_OUTPUT);
                  cout << endl << " " << "Could not infer " << "`" << de_uglify_parse_tree(methodname) << "'" << " on abstract type " << "`" << de_uglify_parse_tree(abstracttype) << "'" << "," << endl << "   since the implementations of " << "`" << de_uglify_parse_tree(methodname) << "'" << " on the non-abstract" << endl << "   subtypes of " << "`" << de_uglify_parse_tree(abstracttype) << "'" << " are not all compatible" << "." << endl;
                }
              }
            }
            else {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_OUTPUT);
                  cout << endl << " " << "Could not infer " << "`" << de_uglify_parse_tree(methodname) << "'" << " on abstract type " << "`" << de_uglify_parse_tree(abstracttype) << "'" << "," << endl << "   since " << "`" << de_uglify_parse_tree(methodname) << "'" << " is only implemented on " << "`" << de_uglify_parse_tree(wrap_integer(nofmethods)) << "'" << " of the " << "`" << de_uglify_parse_tree(wrap_integer(nofsubs)) << "'" << endl << "   non-abstract subtypes of " << "`" << de_uglify_parse_tree(abstracttype) << "'" << "." << endl;
                }
              }
              if (!(compatibleP)) {
                cout << " Additionally, the method signatures are " << "not compatible." << endl;
              }
            }
          }
        }
      }
      if (undefinedmethodP) {
        { Cons* value_000 = NULL;
          Surrogate* value_001 = NULL;

          value_000 = walk_undefined_slot_tree(tree, abstracttype, warn_about_undefined_methodsP(), value_001);
          {
            otree = value_000;
            otype = value_001;
          }
        }
      }
      else {
        { Cons* value_002 = NULL;
          Surrogate* value_003 = NULL;

          value_002 = walk_undefined_slot_tree(tree, abstracttype, FALSE, value_003);
          {
            otree = value_002;
            otype = value_003;
          }
        }
      }
      _return1 = otype;
      return (((Cons*)(otree)));
    }
  }
}

Method_Slot* find_prototype_method(Symbol* methodname, Surrogate* abstracttype) {
  { Slot* method = NULL;

    { Surrogate* sub = NULL;
      Cons* iter_000 = ((Class*)(abstracttype->surrogate_value))->class_direct_subs->the_cons_list;

      while (!(iter_000 == NIL)) {
        sub = ((Surrogate*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (((boolean)(((Class*)(sub->surrogate_value))))) {
          if (((Class*)(sub->surrogate_value))->abstractP) {
            method = find_prototype_method(methodname, sub);
          }
          else {
            method = lookup_slot(((Class*)(sub->surrogate_value)), methodname);
          }
          if (((boolean)(method)) &&
              (!storage_slotP(method))) {
            return (((Method_Slot*)(method)));
          }
        }
      }
    }
    return (NULL);
  }
}

int compatible_real_methods(Method_Slot* prototypemethod, Surrogate* abstracttype, int& _return1, boolean& _return2) {
  { Symbol* methodname = prototypemethod->slot_name;
    int nofrealsubtypes = 0;
    int nofrealmethods = 0;
    boolean compatiblesignaturesP = TRUE;

    { Surrogate* sub = NULL;
      Cons* iter_000 = ((Class*)(abstracttype->surrogate_value))->class_direct_subs->the_cons_list;

      while (!(iter_000 == NIL)) {
        sub = ((Surrogate*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (((Class*)(sub->surrogate_value))->abstractP) {
          { int nofsubs = NULL_INTEGER;
            int nofmethods = NULL_INTEGER;
            boolean compatibleP = FALSE;

            nofsubs = compatible_real_methods(prototypemethod, sub, nofmethods, compatibleP);
            nofrealsubtypes = nofrealsubtypes + nofsubs;
            nofrealmethods = nofrealmethods + nofmethods;
            if (!(compatibleP)) {
              compatiblesignaturesP = FALSE;
            }
          }
        }
        else {
          { Slot* thismethod = NULL;

            nofrealsubtypes = nofrealsubtypes + 1;
            thismethod = lookup_slot(((Class*)(sub->surrogate_value)), methodname);
            if (((boolean)(thismethod)) &&
                (!storage_slotP(thismethod))) {
              nofrealmethods = nofrealmethods + 1;
              if (!(identical_signaturesP(prototypemethod, ((Method_Slot*)(thismethod))))) {
                compatiblesignaturesP = FALSE;
              }
            }
          }
        }
      }
    }
    _return1 = nofrealmethods;
    _return2 = compatiblesignaturesP;
    return (nofrealsubtypes);
  }
}

Object* Slot::finish_walking_call_slot_tree(Cons* tree, Standard_Object* firstargtype, Standard_Object*& _return1) {
  { Slot* self = this;

    {
      tree = tree;
      firstargtype = firstargtype;
    }
    { Output_String_Stream* stream_000 = new_output_string_stream();

      *(stream_000->native_stream) << "finish-walking-call-slot-tree: Not defined on " << "`" << self << "'";
      throw *new_stella_exception(stream_000->the_string_reader());
    }
  }
}

Object* Storage_Slot::finish_walking_call_slot_tree(Cons* tree, Standard_Object* firstargtype, Standard_Object*& _return1) {
  { Storage_Slot* self = this;
    Standard_Object* dummy1;

    { Object* objectref = tree->rest->value;
      Standard_Object* otype = self->compute_return_type_spec(firstargtype);

      if (((boolean)(self->reader()))) {
        tree->first_setter(self->reader());
        return (walk_a_cons_tree(tree, _return1));
      }
      if (system_defined_slot_readerP(self)) {
        _return1 = otype;
        return (listO(3, SYM_WALK_STELLA_SYS_CALL_METHOD, self->slot_owner, cons(self->slot_name, cons(objectref, NIL))));
      }
      else {
        _return1 = otype;
        return (yield_slot_value_tree(self, objectref, firstargtype, otype, NULL, dummy1));
      }
    }
  }
}

boolean string_concatenate_methodP(Method_Slot* method) {
  return ((method->slot_name == SYM_WALK_STELLA_CONCATENATE) &&
      (method->slot_owner == SGT_WALK_STELLA_STRING));
}

boolean pass_variable_arguments_as_listP(Method_Slot* method) {
  return (method_must_be_evaluableP(method) ||
      ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) &&
       ((oMETHODBEINGWALKEDo.get() == method) ||
        (!string_concatenate_methodP(method)))));
}

Standard_Object* variable_arguments_type(Method_Slot* method) {
  { boolean dummy1;

    return (extract_parameter_type(((Standard_Object*)(method->method_parameter_type_specifiers_reader()->last())), SYM_WALK_STELLA_ANY_VALUE, dummy1));
  }
}

Symbol* variable_arguments_name(Method_Slot* method) {
  return (((Symbol*)(method->method_parameter_names_reader()->last())));
}

Standard_Object* yield_listified_variable_arguments_type(Method_Slot* method) {
  { Surrogate* listbasetype = SGT_WALK_STELLA_CONS;
    Standard_Object* elementtype = variable_arguments_type(method);

    if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) &&
        (!method_must_be_evaluableP(method))) {
    }
    if (typeP(elementtype)) {
      elementtype = ((Surrogate*)(elementtype))->type_to_wrapped_type();
    }
    { Parametric_Type_Specifier* self_000 = new_parametric_type_specifier();

      self_000->specifier_base_type = listbasetype;
      self_000->specifier_parameter_types = list(1, elementtype);
      { Parametric_Type_Specifier* value_000 = self_000;

        return (value_000);
      }
    }
  }
}

Cons* yield_listified_variable_arguments(Cons* walkedargs, Standard_Object* targettype, boolean wrapargsP) {
  { Standard_Object* dummy1;

    if (!(walkedargs == NIL)) {
      { Object* listifiedargs = SYM_WALK_STELLA_NIL;

        { Cons_Iterator* it = walkedargs->allocate_iterator();

          while (it->nextP()) {
            { Object* arg = it->value;

              arg = sys_tree(arg, targettype, dummy1);
              if (wrapargsP) {
                arg = cons(((targettype == SGT_WALK_STELLA_BOOLEAN) ? SYM_WALK_STELLA_WRAP_BOOLEAN : SYM_WALK_STELLA_WRAP_LITERAL), cons(arg, NIL));
              }
              it->value_setter(arg);
            }
          }
        }
        if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) {
          { Object* arg = NULL;
            Cons* iter_000 = walkedargs->reverse();

            while (!(iter_000 == NIL)) {
              arg = iter_000->value;
              iter_000 = iter_000->rest;
              listifiedargs = listO(3, SYM_WALK_STELLA_CONS, arg, cons(listifiedargs, NIL));
            }
          }
        }
        else {
          listifiedargs = listO(3, SYM_WALK_STELLA_CONS_LIST, walkedargs->value, walkedargs->rest->concatenate(NIL, 0));
        }
        return (((Cons*)(walk_without_type_tree(listifiedargs))));
      }
    }
    else {
      return (cons(SYM_WALK_STELLA_NIL, NIL));
    }
  }
}

Object* finish_walking_argument_list_tree(Slot* self, Cons* tree, Standard_Object* firstargtype, Standard_Object*& _return1) {
  if (self->slot_name == SYM_WALK_STELLA_ALLOCATE_ITERATOR) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_error();
      if (!(suppress_warningsP())) {
        print_error_context(">> ERROR: ", STANDARD_ERROR);
        *(STANDARD_ERROR->native_stream) << endl << " " << "Cannot invoke " << "`" << de_uglify_parse_tree(self) << "'" << " directly, use `foreach' instead" << "." << endl;
      }
    }
    return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_ARGUMENT_LIST_ITERATOR, _return1));
  }
  if (self->slot_owner == SGT_WALK_STELLA_CONS) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_warning();
      if (!(suppress_warningsP())) {
        print_error_context(">> WARNING: ", STANDARD_WARNING);
        *(STANDARD_WARNING->native_stream) << endl << " " << "Applying CONS-methods to &rest-arguments is deprecated." << endl << "   " << "`" << de_uglify_parse_tree(tree) << "'" << endl << " Use `foreach' or explicitly coerce with `coerce-&rest-to-cons'" << "." << endl;
      }
    }
    firstargtype = yield_listified_variable_arguments_type(oMETHODBEINGWALKEDo.get());
  }
  if (pass_variable_arguments_as_listP(oMETHODBEINGWALKEDo.get())) {
    if (self->slot_name == SYM_WALK_STELLA_LENGTH) {
      return (sys_tree(listO(3, SYM_WALK_STELLA_SYS_CALL_METHOD, type_spec_to_base_type(yield_listified_variable_arguments_type(oMETHODBEINGWALKEDo.get())), tree->concatenate(NIL, 0)), SGT_WALK_STELLA_INTEGER, _return1));
    }
    else {
      return (self->finish_walking_call_slot_tree(tree, firstargtype, _return1));
    }
  }
  else {
    return (self->finish_walking_call_slot_tree(tree, firstargtype, _return1));
  }
}

Cons* walk_variable_arguments(Cons* arguments, Method_Slot* method) {
  { Standard_Object* dummy1;

    { Standard_Object* targettype = variable_arguments_type(method);
      boolean listifyargsP = pass_variable_arguments_as_listP(method);
      boolean wrapargsP = listifyargsP &&
          sub_type_spec_ofP(targettype, SGT_WALK_STELLA_LITERAL);
      Cons* cursor = arguments;

      while (!(cursor == NIL)) {
        cursor->value = walk_expression_tree(cursor->value, targettype, method->slot_name, FALSE, dummy1);
        cursor = cursor->rest;
      }
      if (listifyargsP) {
        { Cons* listifiedargs = yield_listified_variable_arguments(arguments, targettype, wrapargsP);

          if (arguments == NIL) {
            return (listifiedargs);
          }
          arguments->value = listifiedargs;
          arguments->rest = NIL;
        }
      }
      return (arguments);
    }
  }
}

Object* Method_Slot::finish_walking_call_slot_tree(Cons* tree, Standard_Object* firstargtype, Standard_Object*& _return1) {
  { Method_Slot* self = this;
    Standard_Object* dummy1;

    { Symbol* methodname = self->slot_name;
      Standard_Object* otypespec = self->compute_return_type_spec(firstargtype);
      Object* otree = NULL;

      { List* ptypespecs = self->method_parameter_type_specifiers_reader();
        int nofparameters = ptypespecs->length();
        Cons* cursor = tree->rest->rest;

        { Standard_Object* targetts = NULL;
          int pindex = NULL_INTEGER;
          Cons* iter_000 = ptypespecs->rest();
          int iter_001 = 2;
          int upper_bound_000 = nofparameters;
          boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;

          while ((!(iter_000 == NIL)) &&
              (unboundedP_000 ||
               (iter_001 <= upper_bound_000))) {
            targetts = ((Standard_Object*)(iter_000->value));
            iter_000 = iter_000->rest;
            pindex = iter_001;
            iter_001 = iter_001 + 1;
            targetts = compute_relative_type_spec(targetts, firstargtype);
            if ((pindex == nofparameters) &&
                ((Boolean_Wrapper*)(dynamic_slot_value(self->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value) {
              tree->nth_rest_setter(walk_variable_arguments(cursor, self), pindex);
              break;
            }
            cursor->value = walk_expression_tree(cursor->value, targetts, methodname, TRUE, dummy1);
            cursor = cursor->rest;
          }
        }
      }
      { Object* arg = NULL;
        Cons* iter_002 = tree->rest;

        while (!(iter_002 == NIL)) {
          arg = iter_002->value;
          iter_002 = iter_002->rest;
          if (vrlet_expressionP(arg)) {
            return (percolate_out_vrlet_expression(tree, ((Cons*)(arg)), otypespec, _return1));
          }
        }
      }
      if (inline_method_callP(self)) {
        otree = walk_inline_method_call(self, tree->rest);
        if (((boolean)(otree))) {
          return (sys_tree(otree, otypespec, _return1));
        }
      }
      if (self->method_functionP) {
        otree = cons(SYM_WALK_STELLA_SYS_CALL_FUNCTION, tree->concatenate(NIL, 0));
      }
      else {
        otree = listO(3, SYM_WALK_STELLA_SYS_CALL_METHOD, self->slot_owner, tree->concatenate(NIL, 0));
      }
      _return1 = otypespec;
      return (otree);
    }
  }
}

Cons* quote_arguments(Method_Slot* method, Cons* arguments) {
  { List* ptypespecs = method->method_parameter_type_specifiers_reader();
    int lastparameterindex = ptypespecs->length() - 1;
    Standard_Object* varargstype = (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value ? variable_arguments_type(method) : ((Standard_Object*)(NULL)));
    Standard_Object* ptype = NULL;

    { int argindex = NULL_INTEGER;
      Cons_Iterator* it = arguments->allocate_iterator();
      int iter_000 = 0;

      while (it->nextP()) {
        argindex = iter_000;
        iter_000 = iter_000 + 1;
        if (!(surrogateP(it->value) ||
            keywordP(it->value))) {
          if ((argindex >= lastparameterindex) &&
              ((boolean)(varargstype))) {
            ptype = varargstype;
          }
          else {
            ptype = ((Standard_Object*)(ptypespecs->nth(argindex)));
          }
          if ((it->value == SYM_WALK_STELLA_NULL) ||
              (((ptype == SGT_WALK_STELLA_BOOLEAN) &&
              ((it->value == SYM_WALK_STELLA_TRUE) ||
               (it->value == SYM_WALK_STELLA_FALSE))) ||
               (wrapperP(it->value) &&
                (((boolean)(ptype)) &&
                 sub_type_spec_ofP(ptype, SGT_WALK_STELLA_LITERAL))))) {
            continue;
          }
          it->value_setter(listO(3, SYM_WALK_STELLA_BQUOTE, it->value, NIL));
        }
      }
    }
    return (arguments);
  }
}

Standard_Object* walk_first_argument_to_function(Method_Slot* fnslot, Cons* tree) {
  { List* ptypespecs = fnslot->method_parameter_type_specifiers_reader();
    Standard_Object* targetts = ((Standard_Object*)(ptypespecs->first()));

    if (!fnslot->method_evaluate_argumentsP_reader()) {
      tree->rest = quote_arguments(fnslot, tree->rest);
    }
    if (((Boolean_Wrapper*)(dynamic_slot_value(fnslot->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value &&
        (ptypespecs->length() == 1)) {
      tree->rest = walk_variable_arguments(tree->rest, fnslot);
      return (SGT_WALK_STELLA_UNKNOWN);
    }
    if ((tree->rest == NIL) ||
        fnslot->method_parameter_type_specifiers_reader()->emptyP()) {
      return (SGT_WALK_STELLA_UNKNOWN);
    }
    { Object* otree = NULL;
      Standard_Object* otype = NULL;

      otree = walk_expression_tree(tree->rest->value, targetts, fnslot->slot_name, TRUE, otype);
      tree->second_setter(otree);
      if (subtype_of_integerP(safe_primary_type(otree))) {
        { Object* otree_000 = otree;
          Integer_Wrapper* otree = ((Integer_Wrapper*)(otree_000));

          { Symbol* test_value_000 = fnslot->slot_name;

            if (test_value_000 == SYM_WALK_STELLA_GET_SYM) {
              register_symbol(get_sym_from_offset(otree->wrapper_value));
            }
            else if (test_value_000 == SYM_WALK_STELLA_GET_SGT) {
              register_symbol(get_sgt_from_offset(otree->wrapper_value));
            }
            else if (test_value_000 == SYM_WALK_STELLA_GET_KWD) {
              register_symbol(get_kwd_from_offset(otree->wrapper_value));
            }
            else {
            }
          }
        }
      }
      else {
      }
      return (otype);
    }
  }
}

Object* walk_call_slot_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    { Object* slotname = tree->value;
      Object* walkedfirstarg = NULL;
      Standard_Object* firstargtype = SGT_WALK_STELLA_VOID;
      Class* firstargclass = NULL;
      boolean illegalfirstargP = FALSE;
      int previouserrors = oTRANSLATIONERRORSo.get();
      Slot* slot = NULL;

      if (slotname == SYM_WALK_STELLA_ERROR) {
        oFOUNDRETURNpo.set(TRUE);
      }
      if (((boolean)(lookup_macro(((Symbol*)(slotname)))))) {
        return (walk_macro_tree(tree, _return1));
      }
      { 
        slot = lookup_function(((Symbol*)(slotname)));
        if (((boolean)(slot))) {
          firstargtype = walk_first_argument_to_function(((Method_Slot*)(slot)), tree);
        }
        else {
          if (!(tree->rest == NIL)) {
            walkedfirstarg = walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_UNKNOWN, ((Symbol*)(slotname)), TRUE, firstargtype);
            tree->second_setter(coerce_a_tree(walkedfirstarg, firstargtype, firstargtype, dummy1));
            if (voidP(firstargtype)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "First argument in call to " << "`" << de_uglify_parse_tree(slotname) << "'" << " returns @VOID" << "." << endl;
                }
              }
              return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
            }
            firstargclass = type_spec_to_class(firstargtype);
            slot = lookup_slot(firstargclass, ((Symbol*)(slotname)));
          }
        }
      }
      illegalfirstargP = oTRANSLATIONERRORSo.get() > previouserrors;
      if (vrlet_expressionP(tree->rest->value)) {
        return (percolate_out_vrlet_expression(tree, ((Cons*)(tree->rest->value)), SGT_WALK_STELLA_UNKNOWN, _return1));
      }
      if ((!((boolean)(slot))) &&
          sub_type_spec_ofP(firstargtype, SGT_WALK_STELLA_ARGUMENT_LIST)) {
        slot = lookup_slot(((Class*)(SGT_WALK_STELLA_CONS->surrogate_value)), ((Symbol*)(slotname)));
      }
      if (!((boolean)(slot))) {
        if (warn_about_undefined_methodsP() &&
            ((!(firstargtype == SGT_WALK_STELLA_UNKNOWN)) &&
             (((boolean)(firstargclass)) &&
              firstargclass->abstractP))) {
          return (walk_slot_on_abstract_type_tree(tree, type_spec_to_base_type(firstargtype), _return1));
        }
        else {
          { Cons* value_000 = NULL;
            Surrogate* value_001 = NULL;

            value_000 = walk_undefined_slot_tree(tree, (voidP(firstargtype) ? SGT_WALK_STELLA_UNKNOWN : ((Standard_Object*)(firstargtype))), (!illegalfirstargP) &&
                warn_about_undefined_methodsP(), value_001);
            _return1 = value_001;
            return (value_000);
          }
        }
      }
      { int minargs = slot->method_argument_count();
        int maxargs = minargs;

        if (subtype_of_method_slotP(safe_primary_type(slot))) {
          { Slot* slot_000 = slot;
            Method_Slot* slot = ((Method_Slot*)(slot_000));

            if (((Boolean_Wrapper*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value ||
                ((Boolean_Wrapper*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_METHOD_BODY_ARGUMENTp, FALSE_WRAPPER)))->wrapper_value) {
              minargs = minargs - 1;
              maxargs = NULL_INTEGER;
            }
          }
        }
        else {
        }
        if (bad_argument_rangeP(tree, minargs, maxargs)) {
          return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
        }
      }
      { Object* otree = NULL;
        Standard_Object* otype = NULL;

        if (sub_type_spec_ofP(firstargtype, SGT_WALK_STELLA_ARGUMENT_LIST)) {
          otree = finish_walking_argument_list_tree(slot, tree, firstargtype, otype);
        }
        else {
          otree = slot->finish_walking_call_slot_tree(tree, firstargtype, otype);
        }
        return (sys_tree_if_needed(slot, otree, firstargtype, otype, _return1));
      }
    }
  }
}

boolean inline_method_callP(Method_Slot* method) {
  return (method_call_inlining_enabledP() &&
      ((((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_GLOBALLY_INLINEp, FALSE_WRAPPER)))->wrapper_value ||
      (((boolean)(oMETHODBEINGWALKEDo.get())) &&
       oMETHODBEINGWALKEDo.get()->method_inlined_functions_reader()->memberP(method->slot_name))) &&
       method_inlinableP(method)));
}

boolean method_inlinableP(Method_Slot* method) {
  if (voidP(method->type()) ||
      ((method->method_return_type_specifiers_reader()->length() > 1) ||
       ((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value)) {
    return (FALSE);
  }
  if (method->method_functionP ||
      most_specific_methodP(method)) {
    return (((boolean)(inlinable_method_body(method))));
  }
  else {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_warning();
      if (!(suppress_warningsP())) {
        print_error_context(">> WARNING: ", STANDARD_WARNING);
        *(STANDARD_WARNING->native_stream) << endl << " " << "Cannot inline method " << "`" << de_uglify_parse_tree(method) << "'" << ", since there are" << endl << " one or more methods that specialize it" << "." << endl;
      }
    }
    return (FALSE);
  }
}

boolean most_specific_methodP(Method_Slot* method) {
  { boolean alwaysP_000 = TRUE;

    { Surrogate* sub = NULL;
      Cons* iter_000 = ((Class*)(method->slot_owner->surrogate_value))->class_direct_subs->the_cons_list;

      while (!(iter_000 == NIL)) {
        sub = ((Surrogate*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (!help_most_specific_methodP(((Class*)(sub->surrogate_value)), method)) {
          alwaysP_000 = FALSE;
          break;
        }
      }
    }
    { boolean value_000 = alwaysP_000;

      return (value_000);
    }
  }
}

boolean help_most_specific_methodP(Class* renamed_Class, Method_Slot* method) {
  if (!((boolean)(renamed_Class))) {
    return (TRUE);
  }
  { boolean test_value_000 = FALSE;

    { boolean foundP_000 = FALSE;

      { Slot* slot = NULL;
        Cons* iter_000 = renamed_Class->class_local_slots->the_cons_list;

        while (!(iter_000 == NIL)) {
          slot = ((Slot*)(iter_000->value));
          iter_000 = iter_000->rest;
          if (slot->slot_direct_equivalent == method) {
            foundP_000 = TRUE;
            break;
          }
        }
      }
      test_value_000 = foundP_000;
    }
    test_value_000 = !test_value_000;
    if (test_value_000) {
      { boolean alwaysP_000 = TRUE;

        { Surrogate* sub = NULL;
          Cons* iter_001 = renamed_Class->class_direct_subs->the_cons_list;

          while (!(iter_001 == NIL)) {
            sub = ((Surrogate*)(iter_001->value));
            iter_001 = iter_001->rest;
            if (!help_most_specific_methodP(((Class*)(sub->surrogate_value)), method)) {
              alwaysP_000 = FALSE;
              break;
            }
          }
        }
        test_value_000 = alwaysP_000;
      }
    }
    { boolean value_000 = test_value_000;

      return (value_000);
    }
  }
}

Object* inlinable_method_body(Method_Slot* method) {
  { Cons* definition = NULL;
    Object* body = dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_CACHED_INLINABLE_METHOD_BODY, NULL);

    if (body == method) {
      return (NULL);
    }
    if (((boolean)(body))) {
      return (body);
    }
    definition = ((Cons*)(unstringify_stella_source(method->method_stringified_source, method->home_module())));
    extract_options(definition, NULL);
    set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_CACHED_INLINABLE_METHOD_BODY, method, NULL);
    if (definition->length() == 4) {
      body = definition->last();
      if (safe_primary_type(body) == SGT_WALK_STELLA_CONS) {
        { Object* body_000 = body;
          Cons* body = ((Cons*)(body_000));

          if ((body->value == SYM_WALK_STELLA_RETURN) &&
              (verbatim_treeP(body->rest->value) ||
               (!search_cons_treeP(body, SYM_WALK_STELLA_VERBATIM)))) {
            return (set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_CACHED_INLINABLE_METHOD_BODY, permanentify_form(body->rest->value), NULL));
          }
        }
      }
      else {
      }
    }
    return (NULL);
  }
}

DEFINE_STELLA_SPECIAL(oINLININGMETHODCALLpo, boolean , FALSE);

Object* walk_inline_method_call(Method_Slot* method, Cons* walkedargs) {
  { Standard_Object* dummy1;

    { Object* body = inlinable_method_body(method);
      int nofreferences = NULL_INTEGER;
      Object* inlinearg = NULL;
      boolean successP = TRUE;

      if (verbatim_treeP(body) &&
          ((boolean)(lookup_verbatim_tree(((Cons*)(body)), NULL)))) {
        return (yield_verbatim_inline_call_tree(method, walkedargs));
      }
      body = copy_cons_tree(body);
      { 
        BIND_STELLA_SPECIAL(oINLININGMETHODCALLpo, boolean, TRUE);
        BIND_STELLA_SPECIAL(oMETHODBEINGWALKEDo, Method_Slot*, method);
        { Symbol* var = NULL;
          Standard_Object* tspec = NULL;
          Object* arg = NULL;
          Cons* iter_000 = method->method_parameter_names_reader()->the_cons_list;
          Cons* iter_001 = method->method_parameter_type_specifiers_reader()->the_cons_list;
          Cons* iter_002 = walkedargs;

          while ((!(iter_000 == NIL)) &&
              ((!(iter_001 == NIL)) &&
               (!(iter_002 == NIL)))) {
            var = ((Symbol*)(iter_000->value));
            iter_000 = iter_000->rest;
            tspec = ((Standard_Object*)(iter_001->value));
            iter_001 = iter_001->rest;
            arg = iter_002->value;
            iter_002 = iter_002->rest;
            walk_a_declaration(var, tspec, NULL, TRUE);
            set_local_variable_info(var, KWD_WALK_INLINE_ARGUMENT, arg);
            set_local_variable_info(var, KWD_WALK_INLINE_REFERENCES, wrap_integer(0));
          }
        }
        body = walk_expression_tree(body, method->type(), SYM_WALK_STELLA_INLINE_CALL, TRUE, dummy1);
      }
      if (procedural_expressionP(body)) {
        successP = FALSE;
      }
      if (successP) {
        { Symbol* var = NULL;
          Cons* iter_003 = method->method_parameter_names_reader()->the_cons_list;

          while (!(iter_003 == NIL)) {
            var = ((Symbol*)(iter_003->value));
            iter_003 = iter_003->rest;
            inlinearg = get_local_variable_info(var, KWD_WALK_INLINE_ARGUMENT);
            nofreferences = ((Integer_Wrapper*)(get_local_variable_info(var, KWD_WALK_INLINE_REFERENCES)))->wrapper_value;
            switch (nofreferences) {
              case 0: 
                if (!side_effect_free_expressionP(inlinearg)) {
                  successP = FALSE;
                  break;
                }
              break;
              case 1: 
              break;
              default:
                if (!(side_effect_free_expressionP(inlinearg) &&
                    ((estimated_evaluation_cost(inlinearg) * nofreferences) <= 2))) {
                  successP = FALSE;
                  break;
                }
              break;
            }
          }
        }
      }
      { Symbol* var = NULL;
        Cons* iter_004 = method->method_parameter_names_reader()->the_cons_list;

        while (!(iter_004 == NIL)) {
          var = ((Symbol*)(iter_004->value));
          iter_004 = iter_004->rest;
          var = var;
          pop_variable_binding();
        }
      }
      return ((successP ? body : NULL));
    }
  }
}

Cons* yield_verbatim_inline_call_tree(Method_Slot* method, Cons* walkedargs) {
  return (listO(3, SYM_WALK_STELLA_SYS_INLINE_CALL, cons((method->method_functionP ? SYM_WALK_STELLA_SYS_CALL_FUNCTION : SYM_WALK_STELLA_SYS_CALL_METHOD), ((method->method_functionP ? NIL : cons(method->slot_owner, NIL)))->concatenate(cons(method->slot_name, walkedargs->concatenate(NIL, 0)), 0)), NIL));
}

boolean inline_variable_referenceP(Symbol* self) {
  return (oINLININGMETHODCALLpo.get() &&
      ((boolean)(get_local_variable_info(self, KWD_WALK_INLINE_ARGUMENT))));
}

Cons* walk_inline_variable_reference(Symbol* self, Standard_Object*& _return1) {
  set_local_variable_info(self, KWD_WALK_INLINE_REFERENCES, wrap_integer(((Integer_Wrapper*)(get_local_variable_info(self, KWD_WALK_INLINE_REFERENCES)))->wrapper_value + 1));
  return (sys_tree(copy_cons_tree(get_local_variable_info(self, KWD_WALK_INLINE_ARGUMENT)), lookup_variable_type(self), _return1));
}

boolean side_effect_free_expressionP(Object* tree) {
  if (safe_primary_type(tree) == SGT_WALK_STELLA_CONS) {
    { Object* tree_000 = tree;
      Cons* tree = ((Cons*)(tree_000));

      { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

        if ((test_value_000 == SYM_WALK_STELLA_TYPED_SYS) ||
            ((test_value_000 == SYM_WALK_STELLA_CAST) ||
             (test_value_000 == SYM_WALK_STELLA_SAFE_CAST))) {
          return (side_effect_free_expressionP(tree->rest->value));
        }
        else if ((test_value_000 == SYM_WALK_STELLA_AND) ||
            ((test_value_000 == SYM_WALK_STELLA_OR) ||
             ((test_value_000 == SYM_WALK_STELLA_NOT) ||
              ((test_value_000 == SYM_WALK_STELLA_EQp) ||
               ((test_value_000 == SYM_WALK_STELLA_EQLp) ||
                ((test_value_000 == SYM_WALK_STELLA_e) ||
                 ((test_value_000 == SYM_WALK_STELLA_i) ||
                  ((test_value_000 == SYM_WALK_STELLA__) ||
                   ((test_value_000 == SYM_WALK_STELLA_o) ||
                    ((test_value_000 == SYM_WALK_STELLA_s) ||
                     ((test_value_000 == SYM_WALK_STELLA_g) ||
                      ((test_value_000 == SYM_WALK_STELLA_ge) ||
                       ((test_value_000 == SYM_WALK_STELLA_l) ||
                        ((test_value_000 == SYM_WALK_STELLA_le) ||
                         ((test_value_000 == SYM_WALK_STELLA_THE_CODE) ||
                          ((test_value_000 == SYM_WALK_STELLA_SYS_SLOT_VALUE) ||
                           (test_value_000 == SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER))))))))))))))))) {
          { boolean alwaysP_000 = TRUE;

            { Object* arg = NULL;
              Cons* iter_000 = tree->rest;

              while (!(iter_000 == NIL)) {
                arg = iter_000->value;
                iter_000 = iter_000->rest;
                if (!side_effect_free_expressionP(arg)) {
                  alwaysP_000 = FALSE;
                  break;
                }
              }
            }
            { boolean value_000 = alwaysP_000;

              return (value_000);
            }
          }
        }
        else {
        }
      }
    }
  }
  else {
    return (TRUE);
  }
  return (FALSE);
}

int estimated_evaluation_cost(Object* tree) {
  { int cost = 0;

    if (safe_primary_type(tree) == SGT_WALK_STELLA_CONS) {
      { Object* tree_000 = tree;
        Cons* tree = ((Cons*)(tree_000));

        { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->value));

          if ((test_value_000 == SYM_WALK_STELLA_TYPED_SYS) ||
              ((test_value_000 == SYM_WALK_STELLA_CAST) ||
               (test_value_000 == SYM_WALK_STELLA_SAFE_CAST))) {
            return (estimated_evaluation_cost(tree->rest->value));
          }
          else if ((test_value_000 == SYM_WALK_STELLA_AND) ||
              ((test_value_000 == SYM_WALK_STELLA_OR) ||
               ((test_value_000 == SYM_WALK_STELLA_NOT) ||
                ((test_value_000 == SYM_WALK_STELLA_EQp) ||
                 ((test_value_000 == SYM_WALK_STELLA_EQLp) ||
                  ((test_value_000 == SYM_WALK_STELLA_e) ||
                   ((test_value_000 == SYM_WALK_STELLA_i) ||
                    ((test_value_000 == SYM_WALK_STELLA__) ||
                     ((test_value_000 == SYM_WALK_STELLA_o) ||
                      ((test_value_000 == SYM_WALK_STELLA_s) ||
                       ((test_value_000 == SYM_WALK_STELLA_g) ||
                        ((test_value_000 == SYM_WALK_STELLA_ge) ||
                         ((test_value_000 == SYM_WALK_STELLA_l) ||
                          ((test_value_000 == SYM_WALK_STELLA_le) ||
                           ((test_value_000 == SYM_WALK_STELLA_THE_CODE) ||
                            ((test_value_000 == SYM_WALK_STELLA_SYS_SLOT_VALUE) ||
                             (test_value_000 == SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER))))))))))))))))) {
            cost = cost + 1;
            { Object* arg = NULL;
              Cons* iter_000 = tree->rest;

              while (!(iter_000 == NIL)) {
                arg = iter_000->value;
                iter_000 = iter_000->rest;
                cost = cost + estimated_evaluation_cost(arg);
              }
            }
          }
          else {
          }
        }
      }
    }
    else {
    }
    return (cost);
  }
}

Cons* yield_synthesized_method_body(Translation_Unit* unit) {
  { Method_Slot* method = ((Method_Slot*)(unit->the_object));
    Cons* body = ((Cons*)(unit->code_register));

    if (!((boolean)(method->method_parameter_names_reader()))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Missing parameters for :inherits-through method" << "." << endl;
        }
      }
    }
    if (!(body == NIL)) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Over-specified definition of :inherits-through method" << "." << endl;
        }
      }
      return (body);
    }
    else {
      return (synthesize_method_body(method));
    }
  }
}

Cons* walk_method_object(Translation_Unit* unit) {
  { Method_Slot* method = ((Method_Slot*)(unit->the_object));
    Cons* body = ((Cons*)(unit->code_register));

    { 
      BIND_STELLA_SPECIAL(oLOCALVARIABLETYPETABLEo, Key_Value_List*, new_key_value_list());
      BIND_STELLA_SPECIAL(oLOCALGENSYMTABLEo, Key_Value_List*, new_key_value_list());
      BIND_STELLA_SPECIAL(oSPECIALVARIABLESTACKo, Key_Value_List*, new_key_value_list());
      BIND_STELLA_SPECIAL(oMETHODBEINGWALKEDo, Method_Slot*, method);
      BIND_STELLA_SPECIAL(oFOUNDRETURNpo, boolean, FALSE);
      if (((boolean)(((Symbol*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_INHERITS_THROUGH, NULL)))))) {
        body = yield_synthesized_method_body(unit);
      }
      { Symbol* name = NULL;
        Standard_Object* typespec = NULL;
        Cons* iter_000 = method->method_parameter_names_reader()->the_cons_list;
        Cons* iter_001 = method->method_parameter_type_specifiers_reader()->the_cons_list;

        while ((!(iter_000 == NIL)) &&
            (!(iter_001 == NIL))) {
          name = ((Symbol*)(iter_000->value));
          iter_000 = iter_000->rest;
          typespec = ((Standard_Object*)(iter_001->value));
          iter_001 = iter_001->rest;
          walk_a_declaration(name, typespec, NULL, TRUE);
        }
      }
      if (!(body == NIL)) {
        { boolean test_value_000 = FALSE;

          test_value_000 = check_for_illegal_returnP();
          if (test_value_000) {
            test_value_000 = !voidP(method->type());
            if (test_value_000) {
              test_value_000 = !unit->auxiliaryP;
              if (test_value_000) {
                test_value_000 = !((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_SLOT_AUXILIARYp, FALSE_WRAPPER)))->wrapper_value;
                if (test_value_000) {
                  { boolean foundP_000 = FALSE;

                    { Object* form = NULL;
                      Cons* iter_002 = body;

                      while (!(iter_002 == NIL)) {
                        form = iter_002->value;
                        iter_002 = iter_002->rest;
                        if (consP(form) &&
                            (((Cons*)(form))->value == SYM_WALK_STELLA_RETURN)) {
                          foundP_000 = TRUE;
                          break;
                        }
                      }
                    }
                    test_value_000 = foundP_000;
                  }
                  test_value_000 = !test_value_000;
                }
              }
            }
          }
          if (test_value_000) {
            body = body->concatenate(cons(listO(4, SYM_WALK_STELLA_ERROR, wrap_string("Returned from `"), wrap_string(stringify(method)), cons(wrap_string("' without a `return'."), NIL)), NIL), 0);
          }
        }
        if (log_function_callP(method) &&
            (!unit->auxiliaryP)) {
          body = wrap_body_with_log_function_call_tree(method, body);
        }
        body = walk_list_of_statements(body);
        if ((!voidP(method->type())) &&
            ((!method->abstractP) &&
             (!oFOUNDRETURNpo.get()))) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Missing return statement" << "." << endl;
            }
          }
        }
      }
      oLOCALVARIABLETYPETABLEo.get()->free();
      oLOCALGENSYMTABLEo.get()->free();
      oSPECIALVARIABLESTACKo.get()->free();
      unit->code_register = body;
      return (body);
    }
  }
}

void walk_defmethod_tree(Cons* tree) {
  { Method_Slot* method = NULL;

    { Translation_Unit* self_000 = new_translation_unit();

      self_000->category = SYM_WALK_STELLA_METHOD;
      self_000->the_object = tree;
      self_000->tu_home_module = oMODULEo.get();
      oCURRENTTRANSLATIONUNITo.set(self_000);
    }
    method = define_method_from_parse_tree(tree);
    if ((!method->method_functionP) &&
        ((!((boolean)(method->slot_owner))) ||
         (!((boolean)(((Class*)(method->slot_owner->surrogate_value))))))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Can't define method " << "`" << method->slot_name << "'" << " on the non-existent class " << "`" << method->slot_owner << "'" << "." << endl;
        }
      }
      method->free();
      clear_translation_unit(oCURRENTTRANSLATIONUNITo.get());
      oCURRENTTRANSLATIONUNITo.set(NULL);
    }
    else {
      oCURRENTTRANSLATIONUNITo.get()->the_object = method;
      oCURRENTTRANSLATIONUNITo.get()->code_register = tree->nth_rest(3);
      if (method_needs_lisp_macroP(method)) {
        register_native_name(method->slot_name, KWD_WALK_COMMON_LISP, KWD_WALK_FUNCTION);
      }
    }
  }
}

void walk_method_unit(Translation_Unit* unit) {
  { 
    BIND_STELLA_SPECIAL(oCURRENTTRANSLATIONUNITo, Translation_Unit*, unit);
    { Method_Slot* method = ((Method_Slot*)(unit->the_object));
      Symbol* name = method->slot_name;
      Cons* body = ((Cons*)(unit->code_register));
      boolean createmethodobjectP = TRUE;
      boolean storecodepointerP = TRUE;
      Method_Slot* evaluatorwrappermethod = NULL;

      if (mixin_methodP(method) &&
          translate_to_single_inheritance_languageP()) {
        clear_translation_unit(unit);
        return;
      }
      if (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_FORWARD_DECLARATIONp, FALSE_WRAPPER)))->wrapper_value) {
        set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_FORWARD_DECLARATIONp, (FALSE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
      }
      if ((!method->method_functionP) &&
          (!(method->home_module() == ((Class*)(method->owner()->surrogate_value))->home_module()))) {
        if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_warning();
            if (!(suppress_warningsP())) {
              print_error_context(">> WARNING: ", STANDARD_WARNING);
              *(STANDARD_WARNING->native_stream) << endl << " " << "Defining method " << "`" << de_uglify_parse_tree(method) << "'" << " outside the home module of" << endl << "   its owner class will be illegal in C++ and Java" << "." << endl;
            }
          }
        }
        else {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Defining method " << "`" << de_uglify_parse_tree(method) << "'" << " outside the home module of" << endl << "   its owner class is illegal in " << "`" << translator_output_language_name() << "'" << "." << endl;
            }
          }
          clear_translation_unit(unit);
          return;
        }
      }
      if ((body == NIL) &&
          (!((boolean)(((Symbol*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_INHERITS_THROUGH, NULL))))))) {
        walk_method_object(unit);
        if (!(((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_NATIVEp, FALSE_WRAPPER)))->wrapper_value ||
            method->abstractP)) {
          createmethodobjectP = FALSE;
          set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_FORWARD_DECLARATIONp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
          if (oTRANSLATIONVERBOSITYLEVELo.get() >= 3) {
            cout << "Forward declaration of " << "`" << method << "'" << endl;
          }
        }
        clear_translation_unit(unit);
      }
      else if (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_SLOT_AUXILIARYp, FALSE_WRAPPER)))->wrapper_value ||
          unit->auxiliaryP) {
        createmethodobjectP = FALSE;
        oTRANSLATIONUNITSo.get()->push(unit);
        walk_method_object(unit);
      }
      else {
        oTRANSLATIONUNITSo.get()->push(unit);
        walk_method_object(unit);
      }
      if (createmethodobjectP) {
        if (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_COMMANDp, FALSE_WRAPPER)))->wrapper_value &&
            (!commandP(method))) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_warning();
            if (!(suppress_warningsP())) {
              print_error_context(">> WARNING: ", STANDARD_WARNING);
              *(STANDARD_WARNING->native_stream) << endl << " " << "Ignored :command? declaration.  " << "Currently, only functions can be commands." << "." << endl;
            }
          }
        }
        if (method_needs_evaluator_wrapperP(method)) {
          evaluatorwrappermethod = create_evaluator_wrapper_unit(method);
        }
        if (method_needs_lisp_macroP(method)) {
          create_lisp_macro_units(name, method);
        }
        if (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_NATIVEp, FALSE_WRAPPER)))->wrapper_value ||
            (method->abstractP ||
             (name == SYM_WALK_STELLA_MAIN))) {
          storecodepointerP = FALSE;
        }
        walk_auxiliary_tree(listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_METHODS, yield_define_stella_method_object(method, (storecodepointerP ? method : NULL), (storecodepointerP ? evaluatorwrappermethod : NULL)), NIL));
      }
    }
  }
}

void print_undefined_methods(Module* module, boolean localP) {
  // Print all declared but not yet defined functions and methods
  // in `module'.  If `local?' is true, do not consider any parent modules of
  // `module'.  If `module' is NULL, look at all modules in the system.  This
  // is handy to pinpoint forward declarations that haven't been followed up
  // by actual definitions.
  { Cons* undefinedmethods = NIL;

    { Method_Slot* method = NULL;
      Iterator* iter_000 = all_methods(module, localP);
      Cons* collect_000 = NULL;

      while (iter_000->nextP()) {
        method = ((Method_Slot*)(iter_000->value));
        if (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_FORWARD_DECLARATIONp, FALSE_WRAPPER)))->wrapper_value &&
            ((!method->abstractP) &&
             (!method_contains_unknown_typeP(method)))) {
          if (!((boolean)(collect_000))) {
            {
              collect_000 = cons(method, NIL);
              if (undefinedmethods == NIL) {
                undefinedmethods = collect_000;
              }
              else {
                add_cons_to_end_of_cons_list(undefinedmethods, collect_000);
              }
            }
          }
          else {
            {
              collect_000->rest = cons(method, NIL);
              collect_000 = collect_000->rest;
            }
          }
        }
      }
    }
    { Method_Slot* function = NULL;
      Iterator* iter_001 = all_functions(module, localP);
      Cons* collect_001 = NULL;

      while (iter_001->nextP()) {
        function = ((Method_Slot*)(iter_001->value));
        if (((Boolean_Wrapper*)(dynamic_slot_value(function->dynamic_slots, SYM_WALK_STELLA_FORWARD_DECLARATIONp, FALSE_WRAPPER)))->wrapper_value) {
          if (!((boolean)(collect_001))) {
            {
              collect_001 = cons(function, NIL);
              if (undefinedmethods == NIL) {
                undefinedmethods = collect_001;
              }
              else {
                add_cons_to_end_of_cons_list(undefinedmethods, collect_001);
              }
            }
          }
          else {
            {
              collect_001->rest = cons(function, NIL);
              collect_001 = collect_001->rest;
            }
          }
        }
      }
    }
    if (!(undefinedmethods == NIL)) {
      cout << "The following functions and methods are declared " << "but not yet defined:" << endl;
      { Object* method = NULL;
        Cons* iter_002 = undefinedmethods;

        while (!(iter_002 == NIL)) {
          method = iter_002->value;
          iter_002 = iter_002->rest;
          cout << "    " << method << endl;
        }
      }
    }
  }
}

Cons* walk_sys_call_tree(Cons* tree, Standard_Object*& _return1) {
  { Slot* slot = slot_from_expression_tree(tree);
    Standard_Object* slottype = SGT_WALK_STELLA_UNKNOWN;

    if (((boolean)(slot))) {
      slottype = ((Compound_Type_Specifier*)(dynamic_slot_value(slot->dynamic_slots, SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER, NULL)));
      { boolean test_value_000 = FALSE;

        test_value_000 = ((boolean)(slottype));
        if (test_value_000) {
          if (anchored_type_specifierP(slottype)) {
            test_value_000 = TRUE;
          }
          else {
            { boolean foundP_000 = FALSE;

              { Standard_Object* ts = NULL;
                Cons* iter_000 = ((Parametric_Type_Specifier*)(slottype))->specifier_parameter_types->the_cons_list;

                while (!(iter_000 == NIL)) {
                  ts = ((Standard_Object*)(iter_000->value));
                  iter_000 = iter_000->rest;
                  if (anchored_type_specifierP(ts)) {
                    foundP_000 = TRUE;
                    break;
                  }
                }
              }
              test_value_000 = foundP_000;
            }
          }
        }
        if (test_value_000) {
          { Output_String_Stream* stream_000 = new_output_string_stream();

            *(stream_000->native_stream) << "walk-sys-call-tree: OOPS, can't handle anchored slot types: " << "`" << tree << "'";
            throw *new_stella_exception(stream_000->the_string_reader());
          }
        }
      }
      if (!(((boolean)(slottype)))) {
        slottype = slot->type();
      }
    }
    _return1 = slottype;
    return (tree);
  }
}

void walk_defclass_tree(Cons* tree) {
  { Translation_Unit* self_000 = new_translation_unit();

    self_000->category = SYM_WALK_STELLA_CLASS;
    self_000->the_object = tree->rest->value;
    self_000->tu_home_module = oMODULEo.get();
    oCURRENTTRANSLATIONUNITo.set(self_000);
  }
  oCURRENTTRANSLATIONUNITo.get()->the_object = define_class_from_parse_tree(tree);
  if (!((boolean)(oCURRENTTRANSLATIONUNITo.get()->the_object))) {
    clear_translation_unit(oCURRENTTRANSLATIONUNITo.get());
    oCURRENTTRANSLATIONUNITo.set(NULL);
  }
}

Cons* yield_initialize_hardwired_slots(Class* renamed_Class) {
  { Cons* hardwiredslotinitializers = NIL;

    if (create_native_classP(renamed_Class)) {
      { Slot* slot = NULL;
        Iterator* iter_000 = renamed_Class->class_slots();

        while (iter_000->nextP()) {
          slot = ((Slot*)(iter_000->value));
          if (subtype_of_storage_slotP(safe_primary_type(slot))) {
            { Slot* slot_000 = slot;
              Storage_Slot* slot = ((Storage_Slot*)(slot_000));

              if (slot->slot_hardwiredP &&
                  ((boolean)(slot->initial_value()))) {
                hardwiredslotinitializers = cons(listO(3, SYM_WALK_STELLA_SETQ, yield_hardwired_slot_variable(slot), cons(slot->initial_value(), NIL)), hardwiredslotinitializers);
              }
            }
          }
          else {
          }
        }
      }
    }
    return (hardwiredslotinitializers);
  }
}

Object* yield_initial_value_expression(Storage_Slot* slot) {
  if (slot->abstractP ||
      (slot->dynamic_storageP() ||
       (slot->slot_hardwiredP ||
        (slot->allocation() == KWD_WALK_CLASS)))) {
    return (NULL);
  }
  { Object* initialvalueexpression = slot->initially();

    if (((boolean)(initialvalueexpression))) {
      return (copy_cons_tree(initialvalueexpression));
    }
    if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
      if (slot->allocation() == KWD_WALK_EMBEDDED) {
        return (listO(3, SYM_WALK_STELLA_ALLOCATE, type_to_symbol(slot->type()), NIL));
      }
    }
    else if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA) ||
        ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
         ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE) ||
          (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_IDL)))) {
      if (slot->allocation() == KWD_WALK_EMBEDDED) {
        return (listO(3, SYM_WALK_STELLA_ALLOCATE, type_to_symbol(slot->type()), NIL));
      }
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
    return (type_to_null_value_tree(slot->type()));
  }
}

Cons* yield_constructor_attachment(Class* renamed_Class, Symbol* classref) {
  if ((!renamed_Class->abstractP) &&
      create_native_classP(renamed_Class)) {
    { Cons* constructorfncode = listO(4, SYM_WALK_STELLA_THE_CODE, KWD_WALK_FUNCTION, yield_constructor_name(renamed_Class), NIL);

      return (cons(listO(4, SYM_WALK_STELLA_SETF, listO(3, SYM_WALK_STELLA_CLASS_CONSTRUCTOR_CODE, classref, NIL), constructorfncode, NIL), NIL));
    }
  }
  else {
    return (NIL);
  }
}

List* Class::class_auxiliary_methods_reader() {
  { Class* self = this;

    { List* answer = ((List*)(dynamic_slot_value(self->dynamic_slots, SYM_WALK_STELLA_CLASS_AUXILIARY_METHODS, NULL)));

      if (!((boolean)(answer))) {
        return (NIL_LIST);
      }
      else {
        return (answer);
      }
    }
  }
}

void cleanup_auxiliary_methods(Class* renamed_Class) {
  { List* auxiliarymethods = renamed_Class->class_auxiliary_methods_reader();
    Slot* reason = NULL;

    { Method_Slot* method = NULL;
      Cons* iter_000 = auxiliarymethods->the_cons_list;

      while (!(iter_000 == NIL)) {
        method = ((Method_Slot*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (!method->deletedP()) {
          if (coerce_wrapped_boolean_to_boolean(method->badP_reader())) {
            method->deletedP_setter(TRUE);
          }
          else {
            reason = ((Slot*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_AUXILIARY_METHOD_REASON, NULL)));
            if ((!((boolean)(reason))) ||
                (reason->deletedP() ||
                 (coerce_wrapped_boolean_to_boolean(reason->badP_reader()) ||
                  ((!((boolean)(reason->slot_owner))) ||
                   ((reason->slot_owner->symbol_id == -1) ||
                    ((!((boolean)(((Class*)(reason->slot_owner->surrogate_value))))) ||
                     (((Class*)(reason->slot_owner->surrogate_value))->deletedP() ||
                      (coerce_wrapped_boolean_to_boolean(((Class*)(reason->slot_owner->surrogate_value))->badP_reader()) ||
                       (!(lookup_slot(((Class*)(reason->slot_owner->surrogate_value)), reason->slot_name) == reason)))))))))) {
              set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_AUXILIARY_METHOD_REASON, NULL, NULL);
              method->deletedP_setter(TRUE);
            }
          }
        }
      }
    }
    auxiliarymethods->remove_deleted_members();
  }
}

void register_auxiliary_method(Method_Slot* method, Slot* reason) {
  if (((boolean)(method))) {
    { Class* renamed_Class = ((Class*)(method->slot_owner->surrogate_value));
      List* auxiliarymethods = NULL;

      if (((boolean)(renamed_Class))) {
        auxiliarymethods = renamed_Class->class_auxiliary_methods_reader();
        if (null_listP(auxiliarymethods)) {
          auxiliarymethods = ((List*)(set_dynamic_slot_value(renamed_Class->dynamic_slots, SYM_WALK_STELLA_CLASS_AUXILIARY_METHODS, new_list(), NULL)));
        }
        finalize_auxiliary_method(method);
        { Method_Slot* value_000 = NULL;

          { Method_Slot* aux = NULL;
            Cons* iter_000 = auxiliarymethods->the_cons_list;

            while (!(iter_000 == NIL)) {
              aux = ((Method_Slot*)(iter_000->value));
              iter_000 = iter_000->rest;
              if ((method->slot_name == aux->slot_name) &&
                  (method->slot_owner == aux->slot_owner)) {
                value_000 = aux;
                break;
              }
            }
          }
          auxiliarymethods->remove(value_000);
        }
        auxiliarymethods->insert(method);
        set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_AUXILIARY_METHOD_REASON, reason, NULL);
      }
    }
  }
}

// A keyword describing how mixin classes are handled in
// single-inheritance target languages.  The legal values are
// :FIRST-CLASS-WITH-METHOD, which means that variables of a mixin type
// are legal and that slot access on a mixin type is facilitated by
// inherited-down accessor methods and a catch-all method on OBJECT,
// :FIRST-CLASS-WITH-TYPECASE which is similar but replaces the catch-all
// method with a function using a TYPECASE, and :SECOND-CLASS, which
// means that variables of a mixin type are illegal and no additional
// accessors and catch-all methods are needed.
Keyword* oMIXIN_IMPLEMENTATION_STYLEo = NULL;

boolean mixin_type_variables_allowedP() {
  return ((!translate_to_single_inheritance_languageP()) ||
      (!(oMIXIN_IMPLEMENTATION_STYLEo == KWD_WALK_SECOND_CLASS)));
}

boolean create_native_mixin_slot_accessorsP() {
  return (translate_to_single_inheritance_languageP() &&
      ((!translate_to_single_inheritance_languageP()) ||
       (!(oMIXIN_IMPLEMENTATION_STYLEo == KWD_WALK_SECOND_CLASS))));
}

void create_accessor_units_for_slot(Storage_Slot* slot, Class* renamed_Class, boolean mixinaccessorsP, boolean signaturesonlyP) {
  { Cons* methodtree = NULL;
    Method_Slot* method = NULL;

    if (system_defined_slot_readerP(slot)) {
      methodtree = yield_slot_reader_tree(slot, renamed_Class);
      if (signaturesonlyP) {
        { 
          BIND_STELLA_SPECIAL(oMODULEo, Module*, slot->home_module());
          BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
          method = define_method_from_parse_tree(methodtree);
        }
      }
      else {
        method = ((Method_Slot*)(help_walk_auxiliary_tree(methodtree, TRUE)->the_object));
      }
      register_auxiliary_method(method, slot);
    }
    if (system_defined_slot_writerP(slot)) {
      methodtree = yield_slot_writer_tree(slot, renamed_Class);
      if (signaturesonlyP) {
        { 
          BIND_STELLA_SPECIAL(oMODULEo, Module*, slot->home_module());
          BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
          method = define_method_from_parse_tree(methodtree);
        }
      }
      else {
        method = ((Method_Slot*)(help_walk_auxiliary_tree(methodtree, TRUE)->the_object));
      }
      register_auxiliary_method(method, slot);
    }
    if (mixinaccessorsP &&
        (!slot->abstractP)) {
      methodtree = yield_native_slot_reader_tree(slot, renamed_Class);
      if (signaturesonlyP) {
        { 
          BIND_STELLA_SPECIAL(oMODULEo, Module*, slot->home_module());
          BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
          method = define_method_from_parse_tree(methodtree);
        }
      }
      else {
        method = ((Method_Slot*)(help_walk_auxiliary_tree(methodtree, TRUE)->the_object));
      }
      register_auxiliary_method(method, slot);
      if (!(slot->slot_hardwiredP)) {
        methodtree = yield_native_slot_writer_tree(slot, renamed_Class);
        if (signaturesonlyP) {
          { 
            BIND_STELLA_SPECIAL(oMODULEo, Module*, slot->home_module());
            BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
            method = define_method_from_parse_tree(methodtree);
          }
        }
        else {
          method = ((Method_Slot*)(help_walk_auxiliary_tree(methodtree, TRUE)->the_object));
        }
        register_auxiliary_method(method, slot);
      }
    }
  }
}

void create_accessor_units_for_mixin_slot(Storage_Slot* slot, Class* renamed_Class) {
  if (native_slotP(slot) &&
      (native_slot_home(slot, renamed_Class) == renamed_Class)) {
    if (oMIXIN_IMPLEMENTATION_STYLEo == KWD_WALK_FIRST_CLASS_WITH_TYPECASE) {
      walk_auxiliary_tree(yield_mixin_slot_reader_tree(slot));
      walk_auxiliary_tree(yield_mixin_slot_writer_tree(slot));
    }
    else {
      create_accessor_units_for_slot(slot, ((Class*)(SGT_WALK_STELLA_OBJECT->surrogate_value)), TRUE, FALSE);
    }
  }
  if (slot->slot_hardwiredP) {
    { Cons* readcode = NULL;
      Cons* auxiliarycode = NULL;

      readcode = yield_hardwired_slot_reader_body(slot, auxiliarycode);
      readcode = readcode;
      walk_auxiliary_tree(auxiliarycode);
    }
    create_accessor_units_for_slot(slot, ((Class*)(SGT_WALK_STELLA_OBJECT->surrogate_value)), FALSE, FALSE);
  }
}

void create_slot_accessor_units(Class* renamed_Class) {
  { Slot* slot = NULL;
    Cons* iter_000 = renamed_Class->local_slots()->the_cons_list;

    while (!(iter_000 == NIL)) {
      slot = ((Slot*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (subtype_of_storage_slotP(safe_primary_type(slot))) {
        { Slot* slot_000 = slot;
          Storage_Slot* slot = ((Storage_Slot*)(slot_000));

          if ((translate_to_single_inheritance_languageP() &&
              ((!translate_to_single_inheritance_languageP()) ||
               (!(oMIXIN_IMPLEMENTATION_STYLEo == KWD_WALK_SECOND_CLASS)))) &&
              renamed_Class->mixinP) {
            create_accessor_units_for_mixin_slot(slot, renamed_Class);
          }
          else {
            create_accessor_units_for_slot(slot, renamed_Class, FALSE, slot->slot_externalP);
          }
        }
      }
      else {
      }
    }
  }
  { boolean test_value_000 = FALSE;

    test_value_000 = translate_to_single_inheritance_languageP();
    if (test_value_000) {
      { boolean foundP_000 = FALSE;

        { Surrogate* super = NULL;
          Cons* iter_001 = renamed_Class->class_direct_supers->the_cons_list;

          while (!(iter_001 == NIL)) {
            super = ((Surrogate*)(iter_001->value));
            iter_001 = iter_001->rest;
            if (((Class*)(super->surrogate_value))->mixinP) {
              foundP_000 = TRUE;
              break;
            }
          }
        }
        test_value_000 = foundP_000;
      }
    }
    if (test_value_000) {
      { Slot* slot = NULL;
        Iterator* iter_002 = renamed_Class->class_slots();

        while (iter_002->nextP()) {
          slot = ((Slot*)(iter_002->value));
          if (mixin_slotP(slot) &&
              (native_slot_home(slot, renamed_Class) == renamed_Class)) {
            create_accessor_units_for_slot(((Storage_Slot*)(slot)), renamed_Class, translate_to_single_inheritance_languageP() &&
                ((!translate_to_single_inheritance_languageP()) ||
                 (!(oMIXIN_IMPLEMENTATION_STYLEo == KWD_WALK_SECOND_CLASS))), slot->slot_externalP);
          }
        }
      }
    }
  }
  create_generic_slot_accessor_unit(renamed_Class);
}

void create_inline_method_units(Class* renamed_Class) {
  { Slot* slot = NULL;
    Cons* iter_000 = renamed_Class->local_slots()->the_cons_list;

    while (!(iter_000 == NIL)) {
      slot = ((Slot*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (inline_methodP(slot)) {
        { Object* name = NULL;
          Cons* parameters = NULL;
          Cons* body = NULL;

          name = destructure_defmethod_tree(((Cons*)(unstringify_stella_source(((Method_Slot*)(slot))->method_stringified_source, slot->home_module()))), NULL, parameters, body);
          {
            name = name;
            parameters = parameters;
          }
          { Translation_Unit* self_000 = new_translation_unit();

            self_000->the_object = slot;
            self_000->category = SYM_WALK_STELLA_METHOD;
            self_000->code_register = body;
            self_000->auxiliaryP = TRUE;
            self_000->tu_home_module = oMODULEo.get();
            walk_method_unit(self_000);
          }
        }
      }
    }
  }
}

Cons* yield_mixin_method_tree(Method_Slot* method, Class* renamed_Class) {
  { Cons* tree = ((Cons*)(unstringify_stella_source(method->method_stringified_source, method->home_module())));
    Cons* parameters = ((Cons*)(tree->rest->rest->value));
    Cons* selftree = ((Cons*)(parameters->value));
    Object* selftypetree = yield_type_spec_tree(((Standard_Object*)(method->method_parameter_type_specifiers_reader()->first())));
    Cons* body = tree->nth_rest(3);

    if (safe_primary_type(selftypetree) == SGT_WALK_STELLA_CONS) {
      { Object* selftypetree_000 = selftypetree;
        Cons* selftypetree = ((Cons*)(selftypetree_000));

        selftypetree->first_setter(yield_type_spec_tree(renamed_Class->class_type));
      }
    }
    else {
      selftypetree = yield_type_spec_tree(renamed_Class->class_type);
    }
    last_cons(selftree)->first_setter(selftypetree);
    if (renamed_Class == ((Class*)(SGT_WALK_STELLA_OBJECT->surrogate_value))) {
      body = extract_options(tree, NULL)->concatenate(cons(SYM_WALK_STELLA_NULL, NIL), 0);
    }
    return (listO(3, SYM_WALK_STELLA_DEFMETHOD, tree->rest->value, cons(parameters, listO(5, KWD_WALK_AUXILIARYp, SYM_WALK_STELLA_TRUE, KWD_WALK_PUBLICp, (method->slot_publicP ? SYM_WALK_STELLA_TRUE : SYM_WALK_STELLA_FALSE), body->concatenate(NIL, 0)))));
  }
}

void create_mixin_method_units(Class* renamed_Class) {
  if (renamed_Class->mixinP &&
      ((!translate_to_single_inheritance_languageP()) ||
       (!(oMIXIN_IMPLEMENTATION_STYLEo == KWD_WALK_SECOND_CLASS)))) {
    { Slot* slot = NULL;
      Cons* iter_000 = renamed_Class->class_local_slots->the_cons_list;

      while (!(iter_000 == NIL)) {
        slot = ((Slot*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (method_slotP(slot)) {
          register_auxiliary_method(((Method_Slot*)(help_walk_auxiliary_tree(yield_mixin_method_tree(((Method_Slot*)(slot)), ((Class*)(SGT_WALK_STELLA_OBJECT->surrogate_value))), TRUE)->the_object)), slot);
        }
      }
    }
  }
  else {
    { List* mixinmethods = new_list();

      { Surrogate* super = NULL;
        Cons* iter_001 = renamed_Class->class_direct_supers->the_cons_list;

        while (!(iter_001 == NIL)) {
          super = ((Surrogate*)(iter_001->value));
          iter_001 = iter_001->rest;
          if (((Class*)(super->surrogate_value))->mixinP) {
            { Slot* slot = NULL;
              Iterator* iter_002 = ((Class*)(super->surrogate_value))->class_slots();

              while (iter_002->nextP()) {
                slot = ((Slot*)(iter_002->value));
                if (method_slotP(slot) &&
                    (native_slot_home(slot, renamed_Class) == renamed_Class)) {
                  mixinmethods->insert_new(((Method_Slot*)(slot)));
                }
              }
            }
          }
        }
      }
      { Method_Slot* method = NULL;
        Cons* iter_003 = mixinmethods->reverse()->the_cons_list;

        while (!(iter_003 == NIL)) {
          method = ((Method_Slot*)(iter_003->value));
          iter_003 = iter_003->rest;
          register_auxiliary_method(((Method_Slot*)(help_walk_auxiliary_tree(yield_mixin_method_tree(method, renamed_Class), TRUE)->the_object)), method);
        }
      }
    }
  }
}

void create_defprint_unit(Class* renamed_Class) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    if (!(!((boolean)(dynamic_slot_value(renamed_Class->dynamic_slots, SYM_WALK_STELLA_PRINT_FORM, NULL))))) {
      walk_auxiliary_tree(listO(7, SYM_WALK_STELLA_DEFMETHOD, SYM_WALK_STELLA_PRINT_OBJECT, listO(3, listO(3, SYM_WALK_STELLA_SELF, renamed_Class->class_type, NIL), listO(3, SYM_WALK_STELLA_STREAM, SYM_WALK_STELLA_NATIVE_OUTPUT_STREAM, NIL), NIL), KWD_WALK_PUBLICp, SYM_WALK_STELLA_TRUE, copy_cons_tree(dynamic_slot_value(renamed_Class->dynamic_slots, SYM_WALK_STELLA_PRINT_FORM, NULL)), NIL));
      { boolean test_value_000 = FALSE;

        test_value_000 = oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP;
        if (test_value_000) {
          { boolean foundP_000 = FALSE;

            { Class* super = NULL;
              Cons* iter_000 = renamed_Class->class_all_super_classes;

              while (!(iter_000 == NIL)) {
                super = ((Class*)(iter_000->value));
                iter_000 = iter_000->rest;
                if (((boolean)(dynamic_slot_value(super->dynamic_slots, SYM_WALK_STELLA_PRINT_FORM, NULL)))) {
                  foundP_000 = TRUE;
                  break;
                }
              }
            }
            test_value_000 = foundP_000;
          }
          test_value_000 = !test_value_000;
        }
        if (test_value_000) {
          push_variable_binding(SYM_WALK_STELLA_SELF, renamed_Class->class_type);
          push_variable_binding(SYM_WALK_STELLA_STREAM, SGT_WALK_STELLA_NATIVE_OUTPUT_STREAM);
          { Translation_Unit* self_000 = new_translation_unit();

            self_000->the_object = renamed_Class;
            self_000->category = SYM_WALK_STELLA_PRINT_METHOD;
            self_000->tu_home_module = oMODULEo.get();
            self_000->code_register = walk_a_tree(listO(5, SYM_WALK_STELLA_IF, listO(3, SYM_WALK_STELLA_NULLp, SYM_WALK_STELLA_SELF, NIL), listO(4, SYM_WALK_STELLA_PRINT_NATIVE_STREAM, SYM_WALK_STELLA_STREAM, wrap_string("!NULL!"), NIL), listO(4, SYM_WALK_STELLA_PRINT_OBJECT, SYM_WALK_STELLA_SELF, SYM_WALK_STELLA_STREAM, NIL), NIL), dummy1);
            self_000->auxiliaryP = TRUE;
            oTRANSLATIONUNITSo.get()->push(self_000);
          }
          pop_variable_binding();
          pop_variable_binding();
        }
      }
      { boolean test_value_001 = FALSE;

        test_value_001 = oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_JAVA;
        if (test_value_001) {
          { boolean foundP_001 = FALSE;

            { Class* super = NULL;
              Cons* iter_001 = renamed_Class->class_all_super_classes;

              while (!(iter_001 == NIL)) {
                super = ((Class*)(iter_001->value));
                iter_001 = iter_001->rest;
                if (((boolean)(dynamic_slot_value(super->dynamic_slots, SYM_WALK_STELLA_PRINT_FORM, NULL)))) {
                  foundP_001 = TRUE;
                  break;
                }
              }
            }
            test_value_001 = foundP_001;
          }
          test_value_001 = !test_value_001;
        }
        if (test_value_001) {
          { 
            BIND_STELLA_SPECIAL(oMETHODBEINGWALKEDo, Method_Slot*, new_method_slot());
            oMETHODBEINGWALKEDo.get()->method_return_type_specifiers = list(1, SGT_WALK_STELLA_STRING);
            push_variable_binding(SYM_WALK_STELLA_SELF, renamed_Class->class_type);
            { Translation_Unit* self_002 = new_translation_unit();

              self_002->the_object = renamed_Class;
              self_002->category = SYM_WALK_STELLA_PRINT_METHOD;
              self_002->tu_home_module = oMODULEo.get();
              self_002->code_register = walk_a_tree(listO(5, SYM_WALK_STELLA_IF, listO(3, SYM_WALK_STELLA_NULLp, SYM_WALK_STELLA_SELF, NIL), listO(3, SYM_WALK_STELLA_RETURN, wrap_string("!NULL!"), NIL), listO(3, SYM_WALK_STELLA_RETURN, listO(3, SYM_WALK_STELLA_STRINGIFY, SYM_WALK_STELLA_SELF, NIL), NIL), NIL), dummy2);
              self_002->auxiliaryP = TRUE;
              oTRANSLATIONUNITSo.get()->push(self_002);
            }
            pop_variable_binding();
          }
        }
      }
    }
  }
}

void walk_class_unit(Translation_Unit* unit) {
  { Class* renamed_Class = ((Class*)(unit->the_object));
    Cons* startuptimecode = NULL;

    if (!renamed_Class->class_finalizedP) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Cannot translate class " << "`" << class_symbol(renamed_Class) << "'" << ", since it could not be finalized" << "." << endl;
        }
      }
      clear_translation_unit(unit);
      return;
    }
    if (create_native_classP(renamed_Class)) {
      warn_about_multiple_parents(renamed_Class);
      warn_about_non_direct_supers(renamed_Class);
      cleanup_auxiliary_methods(renamed_Class);
      if (!(renamed_Class->mixinP &&
          translate_to_single_inheritance_languageP())) {
        oTRANSLATIONUNITSo.get()->push(unit);
        create_constructor_and_destructor_units(renamed_Class);
        create_inline_method_units(renamed_Class);
      }
      create_slot_accessor_units(renamed_Class);
      if (translate_to_single_inheritance_languageP()) {
        create_mixin_method_units(renamed_Class);
      }
      { Slot* slot = NULL;
        Cons* iter_000 = renamed_Class->class_local_slots->the_cons_list;

        while (!(iter_000 == NIL)) {
          slot = ((Slot*)(iter_000->value));
          iter_000 = iter_000->rest;
          if (storage_slotP(slot) &&
              (native_slotP(((Storage_Slot*)(slot))) &&
               (((boolean)(native_slot_home(slot, renamed_Class))) &&
                (!(native_slot_home(slot, renamed_Class)->home_module() == slot->home_module()))))) {
            if (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_COMMON_LISP) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_WARNING);
                  *(STANDARD_WARNING->native_stream) << endl << " " << "Defining slot " << "`" << de_uglify_parse_tree(slot) << "'" << " outside the home module of" << endl << "   its owner class will be illegal in C++ and Java" << "." << endl;
                }
              }
            }
            else {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Defining slot " << "`" << de_uglify_parse_tree(slot) << "'" << " outside the home module of" << endl << "   its owner class is illegal in " << "`" << translator_output_language_name() << "'" << "." << endl;
                }
              }
              while (!(((Translation_Unit*)(oTRANSLATIONUNITSo.get()->pop())) == unit)) {
              }
              clear_translation_unit(unit);
              return;
            }
          }
        }
      }
    }
    create_defprint_unit(renamed_Class);
    startuptimecode = yield_initialize_hardwired_slots(renamed_Class)->concatenate(yield_constructor_attachment(renamed_Class, SYM_WALK_STELLA_CLASS)->concatenate(yield_generic_slot_accessor_attachment(renamed_Class, SYM_WALK_STELLA_CLASS)->concatenate(NIL, 0), 0), 0);
    walk_auxiliary_tree(((startuptimecode == NIL) ? listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_CLASSES, yield_define_stella_class(renamed_Class), NIL) : listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_CLASSES, listO(3, SYM_WALK_STELLA_LET, cons(listO(3, SYM_WALK_STELLA_CLASS, yield_define_stella_class(renamed_Class), NIL), NIL), startuptimecode->concatenate(NIL, 0)), NIL)));
    { Surrogate* synonym = NULL;
      Cons* iter_001 = renamed_Class->class_synonyms_reader()->the_cons_list;

      while (!(iter_001 == NIL)) {
        synonym = ((Surrogate*)(iter_001->value));
        iter_001 = iter_001->rest;
        register_symbol(intern_symbol_in_module(synonym->symbol_name, ((Module*)(synonym->home_context)), TRUE));
      }
    }
  }
}

void create_finalization_units() {
  walk_auxiliary_tree(listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_FINALIZE_CLASSES, cons(SYM_WALK_STELLA_FINALIZE_CLASSES, NIL), NIL));
  walk_auxiliary_tree(listO(5, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_FINALIZE_METHODS, cons(SYM_WALK_STELLA_FINALIZE_SLOTS, NIL), cons(SYM_WALK_STELLA_CLEANUP_UNFINALIZED_CLASSES, NIL), NIL));
}

void walk_defslot_tree(Cons* tree) {
  { Translation_Unit* self_000 = new_translation_unit();

    self_000->category = SYM_WALK_STELLA_SLOT;
    self_000->the_object = tree;
    self_000->tu_home_module = oMODULEo.get();
    oCURRENTTRANSLATIONUNITo.set(self_000);
  }
  { Storage_Slot* slot = NULL;

    slot = define_external_slot_from_parse_tree(tree);
    if (!((boolean)(slot))) {
      clear_translation_unit(oCURRENTTRANSLATIONUNITo.get());
      oCURRENTTRANSLATIONUNITo.set(NULL);
      return;
    }
    oCURRENTTRANSLATIONUNITo.get()->the_object = slot;
    oCURRENTTRANSLATIONUNITo.get()->code_register = tree;
  }
}

void walk_slot_unit(Translation_Unit* unit) {
  { Storage_Slot* slot = ((Storage_Slot*)(unit->the_object));
    Class* ownerclass = ((Class*)(slot->slot_owner->surrogate_value));
    char* stringifieddefinition = stringify(unit->code_register);

    create_accessor_units_for_slot(slot, ownerclass, FALSE, FALSE);
    clear_translation_unit(unit);
    walk_auxiliary_tree(listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_METHODS, listO(3, SYM_WALK_STELLA_DEFINE_EXTERNAL_SLOT_FROM_STRINGIFIED_SOURCE, wrap_string(stringifieddefinition), NIL), NIL));
  }
}

// Associates macro names with the name and code of
// their expander function.
Hash_Table* oMACRO_TABLEo = NULL;

Method_Slot* lookup_macro(Symbol* name) {
  { Cons* macroinfo = ((Cons*)(oMACRO_TABLEo->lookup(name)));
    Method_Slot* expandermethod = NULL;

    if (((boolean)(macroinfo))) {
      expandermethod = lookup_function(((Symbol*)(macroinfo->value)));
      if (((boolean)(expandermethod))) {
        expandermethod->function_code = ((Function_Code_Wrapper*)(macroinfo->rest->value))->wrapper_value;
        return (expandermethod);
      }
    }
    return (NULL);
  }
}

void store_macro(Symbol* name, Symbol* expandername, cpp_function_code expandercode) {
  oMACRO_TABLEo->insert_at(name, cons_list(2, expandername, wrap_function_code(expandercode)));
}

void undefine_macro(Symbol* name) {
  oMACRO_TABLEo->remove_at(name);
}

Symbol* yield_macro_expander_function_name(Symbol* macroname) {
  return (intern_derived_symbol(macroname, string_concatenate(macroname->symbol_name, "-EXPANDER", 0)));
}

void walk_defmacro_tree(Cons* tree) {
  { Object* name = tree->rest->value;

    if (!(symbolP(name))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal macro name: " << "`" << de_uglify_parse_tree(name) << "'" << "." << endl;
        }
      }
    }
    tree->first_setter(SYM_WALK_STELLA_DEFUN);
    tree->second_setter(cons(yield_macro_expander_function_name(((Symbol*)(name))), cons(SYM_WALK_STELLA_OBJECT, NIL)));
    walk_defmethod_tree(tree);
    if (((boolean)(oCURRENTTRANSLATIONUNITo.get()))) {
      oCURRENTTRANSLATIONUNITo.get()->category = SYM_WALK_STELLA_MACRO;
      tree->second_setter(name);
      oCURRENTTRANSLATIONUNITo.get()->code_register = tree;
    }
  }
}

void walk_macro_unit(Translation_Unit* unit) {
  { Cons* definition = ((Cons*)(unit->code_register));
    Symbol* name = ((Symbol*)(definition->rest->value));
    Method_Slot* expandermethod = ((Method_Slot*)(unit->the_object));
    Symbol* expanderfnname = expandermethod->slot_name;
    boolean needslispmacroP = method_needs_lisp_macroP(expandermethod);

    if (needslispmacroP) {
      create_lisp_macro_units(name, expandermethod);
      set_dynamic_slot_value(expandermethod->dynamic_slots, SYM_WALK_STELLA_METHOD_LISP_MACROp, FALSE_WRAPPER, NULL);
    }
    unit->code_register = definition->nth_rest(3);
    walk_method_unit(unit);
    if (needslispmacroP) {
      set_dynamic_slot_value(expandermethod->dynamic_slots, SYM_WALK_STELLA_METHOD_LISP_MACROp, TRUE_WRAPPER, NULL);
    }
    if (expandermethod->method_argument_count() > 5) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Too many arguments in macro definition" << "." << endl;
        }
      }
    }
    walk_auxiliary_tree(listO(3, SYM_WALK_STELLA_STARTUP_TIME_PROGN, listO(5, SYM_WALK_STELLA_STORE_MACRO, listO(3, SYM_WALK_STELLA_BQUOTE, name, NIL), listO(3, SYM_WALK_STELLA_BQUOTE, expanderfnname, NIL), listO(4, SYM_WALK_STELLA_THE_CODE, KWD_WALK_FUNCTION, expanderfnname, NIL), NIL), NIL));
  }
}

void walk_defglobal_tree(Cons* tree) {
  { Translation_Unit* self_000 = new_translation_unit();

    self_000->category = SYM_WALK_STELLA_GLOBAL_VARIABLE;
    self_000->the_object = tree;
    self_000->tu_home_module = oMODULEo.get();
    oCURRENTTRANSLATIONUNITo.set(self_000);
  }
  if (tree->length() < 3) {
    bad_argument_countP(tree, 3);
    clear_translation_unit(oCURRENTTRANSLATIONUNITo.get());
    oCURRENTTRANSLATIONUNITo.set(NULL);
    return;
  }
  { Global_Variable* global = NULL;
    Object* initialvaluetree = NULL;

    global = define_stella_global_variable_from_parse_tree(tree, stringify(tree), initialvaluetree);
    oCURRENTTRANSLATIONUNITo.get()->the_object = global;
    oCURRENTTRANSLATIONUNITo.get()->code_register = initialvaluetree;
    oCURRENTTRANSLATIONUNITo.get()->auxiliaryP = global->variable_auxiliaryP;
  }
}

void walk_global_unit(Translation_Unit* unit) {
  { Standard_Object* dummy1;

    { Global_Variable* global = ((Global_Variable*)(unit->the_object));
      Object* initialvaluetree = unit->code_register;
      char* stringifiedsource = global->variable_stringified_source;

      oTRANSLATIONUNITSo.get()->push(unit);
      if (!(initialvaluetree == KWD_WALK_UNBOUND_SPECIAL_VARIABLE)) {
        if (global->variable_specialP &&
            translate_to_javaP()) {
          global->variable_constantP = FALSE;
          walk_auxiliary_tree(listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_GLOBALS, listO(3, SYM_WALK_STELLA_SETQ, global->variable_name, cons(initialvaluetree, NIL)), NIL));
          initialvaluetree = KWD_WALK_UNBOUND_SPECIAL_VARIABLE;
        }
        else if (wrapperP(initialvaluetree) ||
            ((symbolP(initialvaluetree) &&
            constant_symbolP(((Symbol*)(initialvaluetree)))) ||
             (consP(initialvaluetree) &&
              (((Cons*)(initialvaluetree))->value == SYM_WALK_STELLA_VERBATIM)))) {
          initialvaluetree = walk_expression_tree(initialvaluetree, global->variable_type, global->variable_name, TRUE, dummy1);
        }
        else {
          global->variable_constantP = FALSE;
          walk_auxiliary_tree(listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_GLOBALS, listO(3, SYM_WALK_STELLA_SETQ, global->variable_name, cons(initialvaluetree, NIL)), NIL));
          initialvaluetree = type_to_walked_null_value_tree(type_spec_to_base_type(global->variable_type));
        }
      }
      unit->code_register = initialvaluetree;
      if (!(unit->auxiliaryP)) {
        walk_auxiliary_tree(listO(3, SYM_WALK_STELLA_STARTUP_TIME_PROGN, listO(3, SYM_WALK_STELLA_DEFINE_STELLA_GLOBAL_VARIABLE_FROM_STRINGIFIED_SOURCE, wrap_string(stringifiedsource), NIL), NIL));
      }
    }
  }
}

void walk_deftype_tree(Cons* tree) {
  { Translation_Unit* self_000 = new_translation_unit();

    self_000->category = SYM_WALK_STELLA_TYPE;
    self_000->the_object = tree;
    self_000->tu_home_module = oMODULEo.get();
    oCURRENTTRANSLATIONUNITo.set(self_000);
  }
  if (tree->length() < 3) {
    bad_argument_countP(tree, 3);
    clear_translation_unit(oCURRENTTRANSLATIONUNITo.get());
    oCURRENTTRANSLATIONUNITo.set(NULL);
    return;
  }
  { Surrogate* type = NULL;

    type = define_stella_type_from_parse_tree(tree);
    oCURRENTTRANSLATIONUNITo.get()->the_object = type;
    oCURRENTTRANSLATIONUNITo.get()->code_register = tree;
  }
}

void walk_type_unit(Translation_Unit* unit) {
  { char* stringifieddefinition = stringify(unit->code_register);

    oTRANSLATIONUNITSo.get()->push(unit);
    walk_auxiliary_tree(listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_CLASSES, listO(3, SYM_WALK_STELLA_DEFINE_STELLA_TYPE_FROM_STRINGIFIED_SOURCE, wrap_string(stringifieddefinition), NIL), NIL));
    register_symbol(intern_symbol_in_module(((Surrogate*)(unit->the_object))->symbol_name, ((Module*)(((Surrogate*)(unit->the_object))->home_context)), TRUE));
  }
}

void walk_startup_time_progn_tree(Cons* tree) {
  { Translation_Unit* self_000 = new_translation_unit();

    self_000->category = SYM_WALK_STELLA_STARTUP_TIME_PROGN;
    self_000->tu_home_module = oMODULEo.get();
    self_000->code_register = tree;
    oCURRENTTRANSLATIONUNITo.set(self_000);
  }
}

Keyword* extract_startup_time_phase(Cons* tree) {
  { Keyword* startuptimephase = KWD_WALK_FINAL;
    Cons* startuptimephasetree = NULL;

    if ((tree->length() >= 2) &&
        (keywordP(tree->rest->value) &&
         oSTARTUP_TIME_PHASESo->memberP(((Keyword*)(tree->rest->value))))) {
      startuptimephase = ((Keyword*)(tree->rest->value));
      startuptimephasetree = tree->rest;
      tree->rest = tree->rest->rest;
      startuptimephasetree->rest = NIL;
    }
    return (startuptimephase);
  }
}

void walk_startup_time_progn_unit(Translation_Unit* unit) {
  { Cons* tree = ((Cons*)(unit->code_register));
    Keyword* phase = extract_startup_time_phase(tree);

    oTRANSLATIONUNITSo.get()->push(unit);
    unit->the_object = walk_list_of_statements(tree->rest);
    unit->code_register = phase;
  }
}

// The maximum number of startup units that can be combined
// into a single startup function (this avoids the construction of huge startup
// functions that would cause too much stress for some wimpy compilers).
int oMAX_NUMBER_OF_STARTUP_UNITSo = 60;

boolean earlier_startup_unitP(Translation_Unit* unit1, Translation_Unit* unit2) {
  return (encode_startup_time_phase(((Keyword*)(unit1->code_register))) < encode_startup_time_phase(((Keyword*)(unit2->code_register))));
}

Cons* combine_startup_function_units(Symbol* startupfnname) {
  { Cons* earlystartuptrees = NIL;
    Cons* startuptrees = NIL;
    Cons* startupphasetrees = NIL;
    List* startupunits = new_list();

    { List_Iterator* it = ((List_Iterator*)(oTRANSLATIONUNITSo.get()->allocate_iterator()));

      while (it->nextP()) {
        if (((Translation_Unit*)(it->value))->category == SYM_WALK_STELLA_STARTUP_TIME_PROGN) {
          startupunits->push(((Translation_Unit*)(it->value)));
          it->value_setter(NULL);
        }
      }
    }
    oTRANSLATIONUNITSo.get()->remove(NULL);
    startupunits->sort(((cpp_function_code)(&earlier_startup_unitP)));
    if (((boolean)(startupfnname)) &&
        (startupunits->length() > oMAX_NUMBER_OF_STARTUP_UNITSo)) {
      extract_startup_function_units(startupunits, startupfnname);
    }
    { Keyword* phase = NULL;
      Cons* iter_000 = oSTARTUP_TIME_PHASESo->the_cons_list;

      while (!(iter_000 == NIL)) {
        phase = ((Keyword*)(iter_000->value));
        iter_000 = iter_000->rest;
        while (startupunits->non_emptyP() &&
            (((Translation_Unit*)(startupunits->first()))->code_register == phase)) {
          startupphasetrees = startupphasetrees->concatenate(((Cons*)(((Translation_Unit*)(startupunits->first()))->the_object)), 0);
          clear_translation_unit(((Translation_Unit*)(startupunits->pop())));
        }
        if (!(startupphasetrees == NIL)) {
          startupphasetrees = listO(3, SYM_WALK_STELLA_VOID_SYS, listO(3, SYM_WALK_STELLA_WHEN, listO(4, SYM_WALK_STELLA_SYS_CALL_FUNCTION, SYM_WALK_STELLA_CURRENT_STARTUP_TIME_PHASEp, wrap_integer(encode_startup_time_phase(phase)), NIL), startupphasetrees->concatenate(NIL, 0)), NIL);
          if (encode_startup_time_phase(phase) <= encode_startup_time_phase(KWD_WALK_MODULES)) {
            earlystartuptrees = cons(startupphasetrees, earlystartuptrees);
          }
          else {
            startuptrees = cons(startupphasetrees, startuptrees);
          }
          startupphasetrees = NIL;
        }
      }
    }
    if (!(startuptrees == NIL)) {
      startuptrees = cons(listO(3, SYM_WALK_STELLA_WITHIN_MODULE, ((oMODULEo.get() == oSTELLA_MODULEo) ? ((Standard_Object*)(SYM_WALK_STELLA_oSTELLA_MODULEo)) : ((Standard_Object*)(listO(3, SYM_WALK_STELLA_GET_STELLA_MODULE, wrap_string(oMODULEo.get()->module_full_name), cons(listO(4, SYM_WALK_STELLA_g, SYM_WALK_STELLA_oSTARTUP_TIME_PHASEo, wrap_integer(encode_startup_time_phase(KWD_WALK_MODULES)), NIL), NIL))))), startuptrees->reverse()->concatenate(NIL, 0)), NIL);
    }
    if ((!(earlystartuptrees == NIL)) ||
        (!(startuptrees == NIL))) {
      startuptrees = earlystartuptrees->reverse()->concatenate(startuptrees, 0);
    }
    return (startuptrees);
  }
}

boolean Method_Slot::method_startup_functionP() {
  { Method_Slot* method = this;

    return (((String_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME, NULL_STRING_WRAPPER)))->wrapper_value != NULL);
  }
}

void extract_startup_function_units(List* startupunits, Symbol* startupfnname) {
  { int remainingunits = startupunits->length();
    int minunitsperfunction = floor_((((double)(oMAX_NUMBER_OF_STARTUP_UNITSo)) / oSTARTUP_TIME_PHASESo->length()) + 1);
    Cons* cursor = startupunits->the_cons_list;
    Cons* phasestart = cursor;
    int phaseunits = 0;
    Keyword* phase = ((Keyword*)(((Translation_Unit*)(cursor->value))->code_register));
    Keyword* unitphase = phase;
    int helpfncounter = 0;

    while (!(cursor == NIL)) {
      unitphase = ((Keyword*)(((Translation_Unit*)(cursor->value))->code_register));
      if (unitphase == phase) {
        phaseunits = phaseunits + 1;
        if (phaseunits < oMAX_NUMBER_OF_STARTUP_UNITSo) {
          cursor = cursor->rest;
          continue;
        }
      }
      if (phaseunits < minunitsperfunction) {
        phasestart = cursor;
        phaseunits = 0;
        phase = unitphase;
        continue;
      }
      { Symbol* functionname = intern_derived_symbol(startupfnname, string_concatenate("HELP-", string_concatenate(startupfnname->symbol_name, integer_to_string(helpfncounter = helpfncounter + 1), 0), 0));
        Cons* helpfntree = ((Cons*)(((Translation_Unit*)(phasestart->value))->the_object));

        ((Translation_Unit*)(phasestart->value))->the_object = cons(listO(3, SYM_WALK_STELLA_SYS_CALL_FUNCTION, functionname, NIL), NIL);
        phasestart = phasestart->rest;
        { int i = NULL_INTEGER;
          Cons_Iterator* it = phasestart->allocate_iterator();
          int iter_000 = 2;
          int upper_bound_000 = phaseunits;
          boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;

          while (it->nextP() &&
              (unboundedP_000 ||
               (iter_000 <= upper_bound_000))) {
            i = iter_000;
            iter_000 = iter_000 + 1;
            i = i;
            helpfntree = helpfntree->concatenate(((Cons*)(((Translation_Unit*)(it->value))->the_object)), 0);
            clear_translation_unit(((Translation_Unit*)(it->value)));
            it->value_setter(NULL);
          }
        }
        helpfntree = listO(3, SYM_WALK_STELLA_DEFUN, functionname, listO(5, NIL, KWD_WALK_PUBLICp, SYM_WALK_STELLA_FALSE, listO(3, SYM_WALK_STELLA_VOID_SYS, cons(SYM_WALK_STELLA_PROGN, helpfntree->concatenate(NIL, 0)), NIL), NIL));
        set_dynamic_slot_value(((Method_Slot*)(help_walk_auxiliary_tree(helpfntree, TRUE)->the_object))->dynamic_slots, SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string(string_capitalize(startupfnname->symbol_name)), NULL_STRING_WRAPPER);
        remainingunits = remainingunits - (phaseunits - 1);
        if (remainingunits <= oMAX_NUMBER_OF_STARTUP_UNITSo) {
          break;
        }
        if (unitphase == phase) {
          cursor = cursor->rest;
        }
        phasestart = cursor;
        phaseunits = 0;
        phase = ((Keyword*)(((Translation_Unit*)(cursor->value))->code_register));
      }
    }
    startupunits->remove(NULL);
  }
}

Symbol* yield_startup_function_name(char* file) {
  file = ((file == NULL) ? oCURRENTFILEo.get() : file);
  if (system_startup_fileP(file)) {
    return (system_startup_function_symbol(oCURRENTSYSTEMDEFINITIONo.get()));
  }
  else {
    return (intern_symbol_in_module(string_concatenate("STARTUP-", string_upcase(file_base_name(file)), 0), oMODULEo.get(), TRUE));
  }
}

void create_startup_function_units() {
  { Symbol* startupfnname = (incremental_translationP() ? ((Symbol*)(NULL)) : yield_startup_function_name(oCURRENTFILEo.get()));
    Cons* tree = NULL;
    Method_Slot* method = NULL;

    if (dont_generate_startup_codeP()) {
      remove_all_startup_time_progn_units();
    }
    else if (((boolean)(startupfnname))) {
      tree = listO(3, SYM_WALK_STELLA_DEFUN, startupfnname, listO(5, NIL, KWD_WALK_PUBLICp, SYM_WALK_STELLA_TRUE, SYM_WALK_STELLA_NULL, NIL));
      method = define_method_from_parse_tree(tree);
      set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_FORWARD_DECLARATIONp, (FALSE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
      set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string(string_capitalize(startupfnname->symbol_name)), NULL_STRING_WRAPPER);
      tree = listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_METHODS, yield_define_stella_method_object(method, method, NULL), cons(listO(4, SYM_WALK_STELLA_LET, cons(listO(3, SYM_WALK_STELLA_FUNCTION, listO(3, SYM_WALK_STELLA_LOOKUP_FUNCTION, listO(3, SYM_WALK_STELLA_QUOTE, startupfnname, NIL), NIL), NIL), NIL), listO(4, SYM_WALK_STELLA_SETF, listO(3, SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME, SYM_WALK_STELLA_FUNCTION, NIL), wrap_string(((String_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME, NULL_STRING_WRAPPER)))->wrapper_value), NIL), NIL), NIL));
      walk_auxiliary_tree(tree);
      tree = listO(3, SYM_WALK_STELLA_DEFUN, startupfnname, listO(6, NIL, KWD_WALK_PUBLICp, SYM_WALK_STELLA_TRUE, KWD_WALK_AUXILIARYp, SYM_WALK_STELLA_TRUE, combine_startup_function_units(startupfnname)->concatenate(NIL, 0)));
      set_dynamic_slot_value(((Method_Slot*)(help_walk_auxiliary_tree(tree, TRUE)->the_object))->dynamic_slots, SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string(string_capitalize(startupfnname->symbol_name)), NULL_STRING_WRAPPER);
    }
    else {
      tree = cons(SYM_WALK_STELLA_STARTUP_TIME_PROGN, combine_startup_function_units(startupfnname)->concatenate(NIL, 0));
      walk_auxiliary_tree(tree);
      tree = cons(SYM_WALK_STELLA_PROGN, ((Cons*)(((Translation_Unit*)(oTRANSLATIONUNITSo.get()->first()))->the_object)));
      ((Translation_Unit*)(oTRANSLATIONUNITSo.get()->first()))->the_object = tree;
    }
  }
}

boolean dont_generate_startup_codeP() {
  return (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE);
}

void remove_all_startup_time_progn_units() {
  { Translation_Unit* placeholderunit = NULL;

    { List_Iterator* it = ((List_Iterator*)(oTRANSLATIONUNITSo.get()->allocate_iterator()));

      while (it->nextP()) {
        if (((Translation_Unit*)(it->value))->category == SYM_WALK_STELLA_STARTUP_TIME_PROGN) {
          { Translation_Unit* unit = ((Translation_Unit*)(it->value));

            if (!((boolean)(placeholderunit))) {
              placeholderunit = unit;
            }
            else {
              clear_translation_unit(unit);
            }
            it->value_setter(placeholderunit);
          }
        }
      }
    }
    oTRANSLATIONUNITSo.get()->remove(placeholderunit);
    clear_translation_unit(placeholderunit);
  }
}

Cons* walk_signal_exception_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;

    _return1 = SGT_WALK_STELLA_VOID;
    return (listO(3, SYM_WALK_STELLA_SYS_SIGNAL, walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_STELLA_ROOT_EXCEPTION, SYM_WALK_STELLA_SIGNAL_EXCEPTION, FALSE, dummy1), NIL));
  }
}

Cons* walk_exception_case_tree(Cons* tree, Standard_Object*& _return1) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    if (tree->length() <= 2) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Missing handler(s) in 'exception-case': " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
    }
    { Object* protectedtree = walk_a_tree(tree->rest->value, dummy1);
      Cons* handlertrees = walk_exception_handler_trees(tree->rest->rest);

      if ((!((boolean)(protectedtree))) ||
          (!((boolean)(handlertrees)))) {
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_VOID, _return1));
      }
      _return1 = SGT_WALK_STELLA_VOID;
      return (listO(3, SYM_WALK_STELLA_SYS_HANDLER_CASE, walk_a_tree(protectedtree, dummy2), handlertrees->concatenate(NIL, 0)));
    }
  }
}

Cons* walk_exception_handler_trees(Cons* tree) {
  { Cons* clauses = NIL;
    Object* exceptiontype = NULL;
    Object* exceptionvariablespec = NULL;
    Symbol* exceptionvariable = NULL;

    { Object* clause = NULL;
      Cons* iter_000 = tree;

      while (!(iter_000 == NIL)) {
        clause = iter_000->value;
        iter_000 = iter_000->rest;
        if (safe_primary_type(clause) == SGT_WALK_STELLA_CONS) {
          { Object* clause_000 = clause;
            Cons* clause = ((Cons*)(clause_000));

            if (clause->length() < 2) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal exception handler clause: " << "`" << de_uglify_parse_tree(clause) << "'" << "." << endl;
                }
              }
              return (NULL);
            }
            exceptiontype = clause->value;
            if (!symbolP(exceptiontype)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal exception type: " << "`" << de_uglify_parse_tree(exceptiontype) << "'" << "." << endl;
                }
              }
              return (NULL);
            }
            exceptiontype = typify(exceptiontype);
            if (!((boolean)(type_to_class(((Surrogate*)(exceptiontype)))))) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Undefined exception type: " << "`" << de_uglify_parse_tree(exceptiontype) << "'" << "." << endl;
                }
              }
              return (NULL);
            }
            if (!subtype_ofP(((Surrogate*)(exceptiontype)), SGT_WALK_STELLA_NATIVE_EXCEPTION)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Exception type " << "`" << de_uglify_parse_tree(exceptiontype) << "'" << " is not a subtype of NATIVE-EXCEPTION" << "." << endl;
                }
              }
              return (NULL);
            }
            exceptionvariablespec = clause->rest->value;
            if ((!consP(exceptionvariablespec)) ||
                ((((Cons*)(exceptionvariablespec))->length() > 1) ||
                 ((((Cons*)(exceptionvariablespec))->length() == 1) &&
                  (!symbolP(((Cons*)(exceptionvariablespec))->value))))) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal exception variable specification: " << "`" << de_uglify_parse_tree(clause) << "'" << "." << endl;
                }
              }
              return (NULL);
            }
            exceptionvariable = ((Symbol*)(((Cons*)(exceptionvariablespec))->value));
            if (translate_to_javaP() &&
                (!((boolean)(exceptionvariable)))) {
              exceptionvariable = local_gensym("E");
            }
            clause = listO(3, SYM_WALK_STELLA_SYS_HANDLE_EXCEPTION, walk_a_declaration(exceptionvariable, exceptiontype, NULL, TRUE), walk_list_of_trees(clause->rest->rest)->concatenate(NIL, 0));
            if (((boolean)(exceptionvariable))) {
              pop_variable_binding();
            }
            clauses = cons(clause, clauses);
          }
        }
        else {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal exception handler clause: " << "`" << de_uglify_parse_tree(tree) << "'" << "." << endl;
            }
          }
          return (NULL);
        }
      }
    }
    return (clauses->reverse());
  }
}

Cons* walk_unwind_protect_tree(Cons* tree, Surrogate*& _return1) {
  if (tree->value == SYM_WALK_STELLA_SYS_UNWIND_PROTECT) {
    _return1 = SGT_WALK_STELLA_VOID;
    return (tree);
  }
  _return1 = SGT_WALK_STELLA_VOID;
  return (cons(SYM_WALK_STELLA_SYS_UNWIND_PROTECT, walk_list_of_statements(tree->rest)->concatenate(NIL, 0)));
}

void walk_verbatim_definition_tree(Cons* tree) {
  { Translation_Unit* self_000 = new_translation_unit();

    self_000->category = SYM_WALK_STELLA_VERBATIM;
    self_000->code_register = tree;
    oCURRENTTRANSLATIONUNITo.set(self_000);
  }
}

void walk_verbatim_unit(Translation_Unit* unit) {
  { Standard_Object* dummy1;

    unit->the_object = walk_a_tree(unit->code_register, dummy1);
    if (((boolean)(unit->the_object))) {
      oTRANSLATIONUNITSo.get()->push(unit);
      unit->code_register = NULL;
    }
    else {
      unit->code_register = NULL;
      clear_translation_unit(unit);
    }
  }
}

Object* walk_verbatim_tree(Cons* tree, Standard_Object*& _return1) {
  { Property_List* self_000 = new_property_list();

    self_000->the_plist = tree->rest;
    { Property_List* options = self_000;
      Object* verbatimtree = options->lookup(oTRANSLATOROUTPUTLANGUAGEo.get());

      if (!((boolean)(verbatimtree))) {
        verbatimtree = options->lookup(KWD_WALK_OTHERWISE);
        if (!((boolean)(verbatimtree))) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Verbatim statement has no " << "`" << oTRANSLATOROUTPUTLANGUAGEo.get() << "'" << " option." << "." << endl;
            }
          }
          _return1 = SGT_WALK_STELLA_UNKNOWN;
          return (NULL);
        }
        if (!(verbatimtree == SYM_WALK_STELLA_NULL)) {
          return (walk_a_tree(verbatimtree, _return1));
        }
      }
      if (verbatimtree == SYM_WALK_STELLA_NULL) {
        _return1 = SGT_WALK_STELLA_UNKNOWN;
        return (NULL);
      }
      if (subtype_of_stringP(safe_primary_type(verbatimtree))) {
        { Object* verbatimtree_000 = verbatimtree;
          String_Wrapper* verbatimtree = ((String_Wrapper*)(verbatimtree_000));

          options->insert_at(oTRANSLATOROUTPUTLANGUAGEo.get(), new_verbatim_string_wrapper(verbatimtree->wrapper_value));
        }
      }
      else {
      }
      _return1 = SGT_WALK_STELLA_UNKNOWN;
      return (tree);
    }
  }
}

Object* lookup_verbatim_tree(Cons* tree, Keyword* language) {
  if (!((boolean)(language))) {
    language = oTRANSLATOROUTPUTLANGUAGEo.get();
  }
  return (search_plist(tree->rest, language));
}

boolean verbatim_treeP(Object* tree) {
  return (consP(tree) &&
      (((Cons*)(tree))->value == SYM_WALK_STELLA_VERBATIM));
}

void walk_defmodule_tree(Cons* tree) {
  { Translation_Unit* self_000 = new_translation_unit();

    self_000->category = SYM_WALK_STELLA_MODULE;
    self_000->the_object = tree;
    oCURRENTTRANSLATIONUNITo.set(self_000);
  }
  { Module* module = NULL;
    char* modulename = coerce_to_module_name(tree->rest->value, FALSE);

    if (modulename == NULL) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal module name: " << "`" << tree->rest->value << "'" << "." << endl;
        }
      }
      clear_translation_unit(oCURRENTTRANSLATIONUNITo.get());
      oCURRENTTRANSLATIONUNITo.set(NULL);
      return;
    }
    module = define_module(modulename, tree->rest->rest);
    oCURRENTTRANSLATIONUNITo.get()->the_object = module;
  }
}

void walk_module_unit(Translation_Unit* unit) {
  { Module* module = ((Module*)(unit->the_object));

    walk_auxiliary_tree(listO(4, SYM_WALK_STELLA_STARTUP_TIME_PROGN, KWD_WALK_MODULES, yield_define_module(module), NIL));
  }
}

Cons* extract_required_argument_values(Standard_Object* classtype, Property_List* slotsandvalues) {
  { Standard_Object* dummy1;

    { boolean arraytypeP = array_type_specifierP(classtype);
      List* requiredvalues = new_list();
      boolean foundP = FALSE;

      if (arraytypeP) {
        preprocess_array_arguments(((Parametric_Type_Specifier*)(classtype)), slotsandvalues);
      }
      { Symbol* requiredslotname = NULL;
        Cons* iter_000 = type_spec_to_class(classtype)->class_required_slot_names_reader()->the_cons_list;

        while (!(iter_000 == NIL)) {
          requiredslotname = ((Symbol*)(iter_000->value));
          iter_000 = iter_000->rest;
          foundP = FALSE;
          { Slot* slot = NULL;
            Object* valueref = NULL;
            Cons* iter_001 = slotsandvalues->the_plist;

            while (!(iter_001 == NIL)) {
              slot = ((Slot*)(iter_001->value));
              valueref = iter_001->rest->value;
              iter_001 = iter_001->rest->rest;
              if (slot->slot_name == requiredslotname) {
                foundP = TRUE;
                if (arraytypeP) {
                  evaluate_array_argument_value(((Parametric_Type_Specifier*)(classtype)), requiredvalues, slot, valueref);
                }
                else {
                  requiredvalues->insert_last(walk_expression_tree(valueref, slot->compute_return_type_spec(classtype), SYM_WALK_STELLA_NEW, FALSE, dummy1));
                }
                slotsandvalues->remove_at(slot);
                break;
              }
            }
          }
          if (!(foundP)) {
            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              signal_translation_error();
              if (!(suppress_warningsP())) {
                print_error_context(">> ERROR: ", STANDARD_ERROR);
                *(STANDARD_ERROR->native_stream) << endl << " " << "Missing required NEW argument: " << "`" << de_uglify_parse_tree(requiredslotname) << "'" << "." << endl;
              }
            }
          }
        }
      }
      return (requiredvalues->the_cons_list);
    }
  }
}

void preprocess_array_arguments(Parametric_Type_Specifier* arraytype, Property_List* slotsandvalues) {
  { Class* arrayclass = ((Class*)(arraytype->specifier_base_type->surrogate_value));
    List* arraydimensions = array_type_dimensions(arraytype);
    int arrayrank = arraydimensions->length();
    Slot* initialelementslot = lookup_slot(arrayclass, SYM_WALK_STELLA_INITIAL_ELEMENT);
    Object* initialelementref = slotsandvalues->lookup(initialelementslot);
    Slot* arraysizeslot = lookup_slot(arrayclass, SYM_WALK_STELLA_SIZE);
    Object* sizeref = slotsandvalues->lookup(arraysizeslot);
    boolean nonlistsizeP = FALSE;

    if (!((boolean)(sizeref))) {
      if (!indefinite_array_typeP(arraytype)) {
        sizeref = array_type_dimensions(arraytype)->the_cons_list;
      }
    }
    if (!((boolean)(sizeref))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Missing array :size argument in NEW expression" << "." << endl;
        }
      }
    }
    else {
      if (safe_primary_type(sizeref) == SGT_WALK_STELLA_CONS) {
        { Object* sizeref_000 = sizeref;
          Cons* sizeref = ((Cons*)(sizeref_000));

          if ((arrayrank == 1) &&
              ((!(sizeref->length() == 1)) ||
               (symbolP(sizeref->value) &&
                (((boolean)(lookup_function(((Symbol*)(sizeref->value))))) ||
                 ((boolean)(lookup_macro(((Symbol*)(sizeref->value))))))))) {
            nonlistsizeP = TRUE;
          }
        }
      }
      else {
        nonlistsizeP = TRUE;
      }
      if (nonlistsizeP) {
        sizeref = cons(sizeref, NIL);
      }
      if (!(((Cons*)(sizeref))->length() == arrayrank)) {
        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          signal_translation_error();
          if (!(suppress_warningsP())) {
            print_error_context(">> ERROR: ", STANDARD_ERROR);
            *(STANDARD_ERROR->native_stream) << endl << " " << "Array :size initializer " << "`" << de_uglify_parse_tree(sizeref) << "'" << endl << "   doesn't match the rank of" << endl << "   " << "`" << de_uglify_parse_tree(arraytype) << "'" << "." << endl;
          }
        }
        sizeref = NULL;
      }
      else {
        { int dim = NULL_INTEGER;
          Cons_Iterator* sizeit = ((Cons*)(sizeref))->allocate_iterator();
          List_Iterator* dimit = ((List_Iterator*)(arraydimensions->allocate_iterator()));
          int iter_000 = 0;

          while (sizeit->nextP() &&
              dimit->nextP()) {
            dim = iter_000;
            iter_000 = iter_000 + 1;
            if (!((boolean)(((Integer_Wrapper*)(dimit->value))))) {
              if (integerP(sizeit->value)) {
                dimit->value_setter(((Integer_Wrapper*)(sizeit->value)));
              }
            }
            else if (!eqlP(((Integer_Wrapper*)(dimit->value)), sizeit->value)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Array :size initializer " << "`" << sizeit->value << "'" << " doesn't match array dimension " << "`" << de_uglify_parse_tree(wrap_integer(dim)) << "'" << " of" << endl << "   " << "`" << de_uglify_parse_tree(arraytype) << "'" << "." << endl;
                }
              }
              sizeit->value_setter(((Integer_Wrapper*)(dimit->value)));
            }
          }
        }
      }
    }
    slotsandvalues->insert_at(initialelementslot, initialelementref);
    if (((boolean)(sizeref))) {
      slotsandvalues->insert_at(arraysizeslot, sizeref);
    }
    else {
      slotsandvalues->remove_at(arraysizeslot);
    }
  }
}

void evaluate_array_argument_value(Parametric_Type_Specifier* arraytype, List* requiredvalues, Slot* slot, Object* valueref) {
  { Standard_Object* dummy1;
    Standard_Object* dummy2;
    Standard_Object* dummy3;

    if (slot->slot_name == SYM_WALK_STELLA_SIZE) {
      { Object* dim = NULL;
        Cons* iter_000 = ((Cons*)(valueref));
        Cons* collect_000 = NULL;

        while (!(iter_000 == NIL)) {
          dim = iter_000->value;
          iter_000 = iter_000->rest;
          if (!((boolean)(collect_000))) {
            {
              collect_000 = cons(walk_expression_tree(dim, SGT_WALK_STELLA_INTEGER, SYM_WALK_STELLA_NEW, FALSE, dummy1), NIL);
              if (requiredvalues->the_cons_list == NIL) {
                requiredvalues->the_cons_list = collect_000;
              }
              else {
                add_cons_to_end_of_cons_list(requiredvalues->the_cons_list, collect_000);
              }
            }
          }
          else {
            {
              collect_000->rest = cons(walk_expression_tree(dim, SGT_WALK_STELLA_INTEGER, SYM_WALK_STELLA_NEW, FALSE, dummy2), NIL);
              collect_000 = collect_000->rest;
            }
          }
        }
      }
    }
    else {
      requiredvalues->insert_last(walk_expression_tree(valueref, slot->compute_return_type_spec(arraytype), SYM_WALK_STELLA_NEW, FALSE, dummy3));
    }
  }
}

Cons* yield_new_arguments_tree(Cons* keywordsandvalues, Standard_Object* classtype, Symbol* selfvariable, Cons*& _return1) {
  if (((keywordsandvalues->length()) % 2)) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      signal_translation_error();
      if (!(suppress_warningsP())) {
        print_error_context(">> ERROR: ", STANDARD_ERROR);
        *(STANDARD_ERROR->native_stream) << endl << " " << "Odd number of arguments in 'new' expression: " << "`" << de_uglify_parse_tree(keywordsandvalues) << "'" << "." << endl;
      }
    }
    _return1 = NIL;
    return (NIL);
  }
  { Class* renamed_Class = type_spec_to_class(classtype);

    { Property_List* self_000 = new_property_list();

      self_000->the_plist = keywordsandvalues;
      { Property_List* plist = self_000;
        Cons* otherassignments = NIL;
        Cons* requiredvalues = NULL;
        Symbol* slotname = NULL;
        Slot* slot = NULL;

        { Property_List_Iterator* it = ((Property_List_Iterator*)(plist->allocate_iterator()));

          while (it->nextP()) {
            if (!keywordP(it->key)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal slot keyword " << "`" << it->key << "'" << " in NEW expression" << "." << endl;
                }
              }
              _return1 = NIL;
              return (NIL);
            }
            slotname = intern_symbol_in_module(((Keyword*)(it->key))->symbol_name, renamed_Class->home_module(), FALSE);
            slot = lookup_slot(renamed_Class, slotname);
            if (!((boolean)(slot))) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_error();
                if (!(suppress_warningsP())) {
                  print_error_context(">> ERROR: ", STANDARD_ERROR);
                  *(STANDARD_ERROR->native_stream) << endl << " " << "No such slot " << "`" << de_uglify_parse_tree(slotname) << "'" << " on the class " << "`" << class_name(renamed_Class) << "'" << "." << endl;
                }
              }
              _return1 = NIL;
              return (NIL);
            }
            it->key_setter(slot);
          }
        }
        requiredvalues = extract_required_argument_values(classtype, plist);
        { Slot* slot = NULL;
          Object* valueref = NULL;
          Cons* iter_000 = ((Property_List*)(plist))->the_plist;
          Cons* collect_000 = NULL;

          while (!(iter_000 == NIL)) {
            slot = ((Slot*)(iter_000->value));
            valueref = iter_000->rest->value;
            iter_000 = iter_000->rest->rest;
            if (!storage_slotP(slot)) {
              { Method_Slot* method = ((Method_Slot*)(slot));

                if (((boolean)(((Symbol*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_STORAGE_SLOT, NULL)))))) {
                  slot = lookup_slot(renamed_Class, ((Symbol*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_STORAGE_SLOT, NULL))));
                }
              }
            }
            if (!((boolean)(collect_000))) {
              {
                collect_000 = cons(listO(4, SYM_WALK_STELLA_SETF, cons(slot->slot_name, cons(selfvariable, NIL)), valueref, NIL), NIL);
                if (otherassignments == NIL) {
                  otherassignments = collect_000;
                }
                else {
                  add_cons_to_end_of_cons_list(otherassignments, collect_000);
                }
              }
            }
            else {
              {
                collect_000->rest = cons(listO(4, SYM_WALK_STELLA_SETF, cons(slot->slot_name, cons(selfvariable, NIL)), valueref, NIL), NIL);
                collect_000 = collect_000->rest;
              }
            }
          }
        }
        _return1 = otherassignments;
        return (requiredvalues);
      }
    }
  }
}

Cons* walk_new_tree(Cons* tree, Standard_Object*& _return1) {
  { Object* renamed_Operator = tree->value;
    Object* classtree = tree->rest->value;
    Object* baseclassname = (consP(classtree) ? ((Cons*)(classtree))->value : classtree);
    Cons* keywordarguments = tree->rest->rest;
    Symbol* selfvariable = local_gensym("SELF");
    Standard_Object* classtype = NULL;
    Cons* newtree = NULL;

    if ((!symbolP(baseclassname)) &&
        (!keywordP(baseclassname))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Illegal argument to " << "`" << de_uglify_parse_tree(renamed_Operator) << "'" << " where symbol expected:  " << "`" << de_uglify_parse_tree(baseclassname) << "'" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
    }
    classtype = safe_yield_type_specifier(classtree);
    if (!((boolean)(type_spec_to_class(classtype)))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Undefined class supplied to " << "`" << de_uglify_parse_tree(renamed_Operator) << "'" << " operator: " << "`" << de_uglify_parse_tree(baseclassname) << "'" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
    }
    if (type_spec_to_class(classtype)->abstractP) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Trying to create an instance of the ABSTRACT class " << "`" << de_uglify_parse_tree(baseclassname) << "'" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
    }
    { Cons* requiredargs = NULL;
      Cons* otherassignments = NULL;

      requiredargs = yield_new_arguments_tree(keywordarguments, classtype, selfvariable, otherassignments);
      newtree = listO(3, SYM_WALK_STELLA_SYS_NEW, classtype, requiredargs->concatenate(NIL, 0));
      if (otherassignments == NIL) {
        _return1 = classtype;
        return (newtree);
      }
      else {
        _return1 = classtype;
        return (listO(3, SYM_WALK_STELLA_VRLET, cons(cons(selfvariable, cons(newtree, NIL)), NIL), otherassignments->concatenate(cons(selfvariable, NIL), 0)));
      }
    }
  }
}

Cons* walk_make_tree(Cons* tree, Standard_Object*& _return1) {
  { Object* classname = tree->rest->value;
    Surrogate* type = NULL;

    if (symbolP(classname)) {
    }
    else if (typeP(classname)) {
      tree->second_setter(intern_symbol_in_module(((Surrogate*)(classname))->symbol_name, ((Module*)(((Surrogate*)(classname))->home_context)), TRUE));
    }
    else {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Bad argument to 'make':" << endl << "   " << "`" << de_uglify_parse_tree(classname) << "'" << " found where symbol expected" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
    }
    type = typify(classname);
    if (!((boolean)(((Class*)(type->surrogate_value))))) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_error();
        if (!(suppress_warningsP())) {
          print_error_context(">> ERROR: ", STANDARD_ERROR);
          *(STANDARD_ERROR->native_stream) << endl << " " << "Reference to undefined class " << "`" << de_uglify_parse_tree(classname) << "'" << "." << endl;
        }
      }
      return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_UNKNOWN, _return1));
    }
    _return1 = type;
    return (tree);
  }
}

Cons* walk_funcall_tree(Cons* tree, Standard_Object*& _return1) {
  if ((tree->value == SYM_WALK_STELLA_SYS_CALL_FUNCTION_CODE) ||
      (tree->value == SYM_WALK_STELLA_SYS_CALL_METHOD_CODE)) {
    _return1 = ((Standard_Object*)(((Cons*)(((Cons*)(tree->rest->value))->value))->value));
    return (tree);
  }
  { Standard_Object* returntype = ((oTARGETTYPEo.get() == SGT_WALK_STELLA_UNKNOWN) ? SGT_WALK_STELLA_VOID : ((Standard_Object*)(oTARGETTYPEo.get())));
    Cons* signature = cons(cons(returntype, NIL), NIL);

    { Object* otree = NULL;
      Standard_Object* otype = NULL;

      otree = walk_expression_tree(tree->rest->value, SGT_WALK_STELLA_CODE, SYM_WALK_STELLA_FUNCALL, TRUE, otype);
      tree->first_setter(((otype == SGT_WALK_STELLA_METHOD_CODE) ? SYM_WALK_STELLA_SYS_CALL_METHOD_CODE : SYM_WALK_STELLA_SYS_CALL_FUNCTION_CODE));
      tree->second_setter(otree);
    }
    { Cons_Iterator* it = tree->rest->rest->allocate_iterator();

      while (it->nextP()) {
        { Object* argtree = NULL;
          Standard_Object* argtype = NULL;

          argtree = walk_a_tree(it->value, argtype);
          it->value_setter(argtree);
          signature = cons(argtype, signature);
        }
      }
    }
    tree->rest = cons(signature->reverse(), tree->rest);
    _return1 = returntype;
    return (tree);
  }
}

Cons* walk_the_code_tree(Cons* tree, Standard_Object*& _return1) {
  { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(tree->rest->value));

    if (test_value_000 == KWD_WALK_FUNCTION) {
      if (bad_argument_rangeP(tree, 2, 3)) {
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_FUNCTION_CODE, _return1));
      }
      if (tree->rest->rest->value == SYM_WALK_STELLA_MAIN) {
        if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
            (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE)) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "Cannot generate a function pointer to `main' in C++" << "." << endl;
            }
          }
          return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_FUNCTION_CODE, _return1));
        }
        else {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_warning();
            if (!(suppress_warningsP())) {
              print_error_context(">> WARNING: ", STANDARD_WARNING);
              *(STANDARD_WARNING->native_stream) << endl << " " << "Won't be able to generate a function pointer to `main' in C++" << "." << endl;
            }
          }
        }
      }
      _return1 = SGT_WALK_STELLA_FUNCTION_CODE;
      return (tree);
    }
    else if (test_value_000 == KWD_WALK_METHOD) {
      if (bad_argument_countP(tree, 3)) {
        return (walk_dont_call_me_tree(tree, SGT_WALK_STELLA_METHOD_CODE, _return1));
      }
      { Surrogate* type = typify(tree->rest->rest->value);

        tree->third_setter(type);
        if (!((boolean)(((Class*)(type->surrogate_value))))) {
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            signal_translation_error();
            if (!(suppress_warningsP())) {
              print_error_context(">> ERROR: ", STANDARD_ERROR);
              *(STANDARD_ERROR->native_stream) << endl << " " << "No class defined for the type: " << "`" << de_uglify_parse_tree(type) << "'" << "." << endl;
            }
          }
        }
        else {
          if ((oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP) ||
              (oTRANSLATOROUTPUTLANGUAGEo.get() == KWD_WALK_CPP_STANDALONE)) {
            if (subtype_ofP(type, SGT_WALK_STELLA_LITERAL)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_WARNING);
                  *(STANDARD_WARNING->native_stream) << endl << " " << "Cannot have methods on literals in " << "`" << translator_output_language_name() << "'" << ", hence, cannot generate" << endl << " a method-code pointer for " << "`" << tree->rest->rest->value << "'" << "." << "`" << tree->fourth() << "'" << "." << endl;
                }
              }
              { Object* value_000 = NULL;
                Standard_Object* value_001 = NULL;

                value_000 = walk_a_tree(listO(4, SYM_WALK_STELLA_CAST, SYM_WALK_STELLA_NULL, SYM_WALK_STELLA_METHOD_CODE, NIL), value_001);
                _return1 = value_001;
                return (((Cons*)(value_000)));
              }
            }
            if (!subtype_ofP(type, SGT_WALK_STELLA_OBJECT)) {
              { 
                BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                signal_translation_warning();
                if (!(suppress_warningsP())) {
                  print_error_context(">> WARNING: ", STANDARD_WARNING);
                  *(STANDARD_WARNING->native_stream) << endl << " " << "Cannot store a method-code pointer for " << endl << " " << "`" << tree->rest->rest->value << "'" << "." << "`" << tree->fourth() << "'" << endl << " since it is not defined on a subtype of @OBJECT" << "." << endl;
                }
              }
              { Object* value_002 = NULL;
                Standard_Object* value_003 = NULL;

                value_002 = walk_a_tree(listO(4, SYM_WALK_STELLA_CAST, SYM_WALK_STELLA_NULL, SYM_WALK_STELLA_METHOD_CODE, NIL), value_003);
                _return1 = value_003;
                return (((Cons*)(value_002)));
              }
            }
          }
          else {
          }
        }
        _return1 = SGT_WALK_STELLA_METHOD_CODE;
        return (tree);
      }
    }
    else {
      { Output_String_Stream* stream_000 = new_output_string_stream();

        *(stream_000->native_stream) << "`" << test_value_000 << "'" << " is not a valid case option";
        throw *new_stella_exception(stream_000->the_string_reader());
      }
    }
  }
}

Cons* de_uglify_arguments(Cons* uglyarguments) {
  { Cons* prettyarguments = NIL;

    { Object* arg = NULL;
      Cons* iter_000 = uglyarguments;
      Cons* collect_000 = NULL;

      while (!(iter_000 == NIL)) {
        arg = iter_000->value;
        iter_000 = iter_000->rest;
        if (!((boolean)(collect_000))) {
          {
            collect_000 = cons(de_uglify_parse_tree(arg), NIL);
            if (prettyarguments == NIL) {
              prettyarguments = collect_000;
            }
            else {
              add_cons_to_end_of_cons_list(prettyarguments, collect_000);
            }
          }
        }
        else {
          {
            collect_000->rest = cons(de_uglify_parse_tree(arg), NIL);
            collect_000 = collect_000->rest;
          }
        }
      }
    }
    return (prettyarguments);
  }
}

Object* de_uglify_parse_tree(Object* tree) {
  if (!((boolean)(tree))) {
    return (NULL);
  }
  { Surrogate* test_value_000 = safe_primary_type(tree);

    if (test_value_000 == SGT_WALK_STELLA_CONS) {
      { Object* tree_000 = tree;
        Cons* tree = ((Cons*)(tree_000));

        { Object* renamed_Operator = tree->value;

          if (!symbolP(renamed_Operator)) {
            return (de_uglify_arguments(tree));
          }
          { Generalized_Symbol* test_value_001 = ((Generalized_Symbol*)(renamed_Operator));

            if ((test_value_001 == SYM_WALK_STELLA_SYS_SLOT_VALUE) ||
                ((test_value_001 == SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER) ||
                 (test_value_001 == SYM_WALK_STELLA_SYS_CALL_METHOD))) {
              return (de_uglify_parse_tree(tree->rest->rest));
            }
            else if (test_value_001 == SYM_WALK_STELLA_SYS_CALL_FUNCTION) {
              return (de_uglify_parse_tree(tree->rest));
            }
            else if ((test_value_001 == SYM_WALK_STELLA_TYPED_SYS) ||
                ((test_value_001 == SYM_WALK_STELLA_VOID_SYS) ||
                 (test_value_001 == SYM_WALK_STELLA_BAD_SYS))) {
              return (de_uglify_parse_tree(tree->rest->value));
            }
            else if (test_value_001 == SYM_WALK_STELLA_GET_SYM) {
              return (get_sym_from_offset(((Integer_Wrapper*)(tree->rest->value))->wrapper_value));
            }
            else {
              { Cons* prettyarguments = NIL;

                { Object* arg = NULL;
                  Cons* iter_000 = tree->rest;
                  Cons* collect_000 = NULL;

                  while (!(iter_000 == NIL)) {
                    arg = iter_000->value;
                    iter_000 = iter_000->rest;
                    if (!((boolean)(collect_000))) {
                      {
                        collect_000 = cons(de_uglify_parse_tree(arg), NIL);
                        if (prettyarguments == NIL) {
                          prettyarguments = collect_000;
                        }
                        else {
                          add_cons_to_end_of_cons_list(prettyarguments, collect_000);
                        }
                      }
                    }
                    else {
                      {
                        collect_000->rest = cons(de_uglify_parse_tree(arg), NIL);
                        collect_000 = collect_000->rest;
                      }
                    }
                  }
                }
                return (cons(renamed_Operator, prettyarguments->concatenate(NIL, 0)));
              }
            }
          }
        }
      }
    }
    else if (subtype_ofP(test_value_000, SGT_WALK_STELLA_COMPOUND_TYPE_SPECIFIER)) {
      { Object* tree_001 = tree;
        Compound_Type_Specifier* tree = ((Compound_Type_Specifier*)(tree_001));

        return (yield_type_spec_tree(tree));
      }
    }
    else {
      return (tree);
    }
  }
}

Hash_Table* oNATIVE_NAME_TABLEo = NULL;

void register_native_name(Symbol* name, Keyword* language, Keyword* category) {
  { Key_Value_List* tableentry = ((Key_Value_List*)(oNATIVE_NAME_TABLEo->lookup(name)));
    List* categories = NULL;

    if (!((boolean)(tableentry))) {
      tableentry = new_key_value_list();
    }
    categories = ((List*)(tableentry->lookup(language)));
    if (!((boolean)(categories))) {
      categories = new_list();
    }
    categories->insert_new(category);
    tableentry->insert_at(language, categories);
    oNATIVE_NAME_TABLEo->insert_at(name, tableentry);
  }
}

boolean native_nameP(Symbol* name, Keyword* language, Keyword* category) {
  { Key_Value_List* tableentry = ((Key_Value_List*)(oNATIVE_NAME_TABLEo->lookup(name)));
    List* categories = NULL;

    if (!((boolean)(tableentry))) {
      return (FALSE);
    }
    categories = ((List*)(tableentry->lookup(language)));
    if (!((boolean)(categories))) {
      return (FALSE);
    }
    else {
      return (categories->memberP(category));
    }
  }
}

Symbol* yield_renamed_name_if_native(Symbol* name, Keyword* language, Keyword* category) {
  if (native_nameP(name, language, category)) {
    { Symbol* newname = NULL;

      if (language == KWD_WALK_COMMON_LISP) {
        newname = intern_derived_symbol(name, string_concatenate("%", name->symbol_name, 0));
      }
      else {
        newname = intern_derived_symbol(name, string_concatenate(name->symbol_name, "_", 0));
      }
      return (yield_renamed_name_if_native(newname, language, category));
    }
  }
  else {
    return (name);
  }
}

boolean method_needs_lisp_macroP(Method_Slot* method) {
  if (translate_to_common_lispP() &&
      method->method_parameter_names_reader()->non_emptyP()) {
    { Boolean_Wrapper* lispmacroP = ((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_LISP_MACROp, NULL)));

      return ((((boolean)(lispmacroP)) &&
          coerce_wrapped_boolean_to_boolean(lispmacroP)) ||
          (commandP(method) &&
           ((!((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_NATIVEp, FALSE_WRAPPER)))->wrapper_value) &&
            ((!((boolean)(lispmacroP))) ||
             coerce_wrapped_boolean_to_boolean(lispmacroP)))));
    }
  }
  else {
    return (FALSE);
  }
}

Cons* yield_lisp_macro_trees(Symbol* name, Method_Slot* method, Cons*& _return1) {
  { Module* module = (((boolean)(((Module*)(name->home_context)))) ? ((Module*)(name->home_context)) : oMODULEo.get());
    Object* translatedname = yield_global_lisp_symbol(module, name->symbol_name);
    Object* translatedfullname = yield_global_lisp_symbol(module, compute_full_name(name->symbol_name, module));
    char* documentation = ((String_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_DOCUMENTATION, NULL_STRING_WRAPPER)))->wrapper_value;

    { Cons* _return_temp = listO(6, SYM_WALK_STELLA_VERBATIM, KWD_WALK_COMMON_LISP, listO(3, intern_common_lisp_symbol("DEFMACRO"), translatedname, cons(listO(5, intern_common_lisp_symbol("&WHOLE"), SYM_WALK_STELLA_EXPRESSION, intern_common_lisp_symbol("&REST"), SYM_WALK_STELLA_IGNORE, NIL), (((documentation != NULL) ? cons(wrap_string(documentation), NIL) : NIL))->concatenate(listO(3, listO(3, intern_common_lisp_symbol("DECLARE"), listO(3, intern_common_lisp_symbol("IGNORE"), SYM_WALK_STELLA_IGNORE, NIL), NIL), listO(4, intern_common_lisp_symbol("LET"), cons(listO(3, SYM_WALK_STELLA_oIGNORETRANSLATIONERRORSpo, (commandP(method) ? SYM_WALK_STELLA_FALSE : SYM_WALK_STELLA_TRUE), NIL), NIL), listO(3, SYM_WALK_STELLA_CL_INCREMENTALLY_TRANSLATE, SYM_WALK_STELLA_EXPRESSION, NIL), NIL), NIL), 0))), KWD_WALK_OTHERWISE, SYM_WALK_STELLA_NULL, NIL);

      _return1 = listO(6, SYM_WALK_STELLA_VERBATIM, KWD_WALK_COMMON_LISP, listO(4, intern_common_lisp_symbol("SETF"), listO(3, intern_common_lisp_symbol("MACRO-FUNCTION"), listO(3, intern_common_lisp_symbol("QUOTE"), translatedfullname, NIL), NIL), listO(3, intern_common_lisp_symbol("MACRO-FUNCTION"), listO(3, intern_common_lisp_symbol("QUOTE"), translatedname, NIL), NIL), NIL), KWD_WALK_OTHERWISE, SYM_WALK_STELLA_NULL, NIL);
      return (_return_temp);
    }
  }
}

void create_lisp_macro_units(Symbol* name, Method_Slot* method) {
  if (name == method->slot_name) {
    register_native_name(name, KWD_WALK_COMMON_LISP, KWD_WALK_FUNCTION);
  }
  { Cons* macrotree = NULL;
    Cons* aliasdefinitiontree = NULL;

    macrotree = yield_lisp_macro_trees(name, method, aliasdefinitiontree);
    walk_auxiliary_tree(macrotree);
    walk_auxiliary_tree(aliasdefinitiontree);
  }
  if (name == method->slot_name) {
    walk_auxiliary_tree(listO(3, SYM_WALK_STELLA_STARTUP_TIME_PROGN, listO(5, SYM_WALK_STELLA_REGISTER_NATIVE_NAME, listO(3, SYM_WALK_STELLA_QUOTE, name, NIL), KWD_WALK_COMMON_LISP, KWD_WALK_FUNCTION, NIL), NIL));
  }
}

boolean commandP(Method_Slot* method) {
  // Return `true' if `method' is an evaluable command.
  return (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_COMMANDp, FALSE_WRAPPER)))->wrapper_value &&
      method->method_functionP);
}

Method_Slot* lookup_command(Symbol* name) {
  // If `name' names an evaluable command return its associated
  // command object;  otherwise, return `null'.  Currently, commands are not
  // polymorphic, i.e., they can only be implemented by functions.
  { Method_Slot* function = lookup_function(name);

    if (((boolean)(function)) &&
        commandP(function)) {
      return (function);
    }
    else {
      return (NULL);
    }
  }
}

boolean method_must_be_evaluableP(Method_Slot* method) {
  return (commandP(method) &&
      (!((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_NATIVEp, FALSE_WRAPPER)))->wrapper_value));
}

boolean method_needs_evaluator_wrapperP(Method_Slot* method) {
  { boolean test_value_000 = FALSE;

    test_value_000 = method_must_be_evaluableP(method);
    if (test_value_000) {
      if (((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value) {
        test_value_000 = TRUE;
      }
      else {
        if (method->method_parameter_names_reader()->length() > 5) {
          test_value_000 = TRUE;
        }
        else {
          if (method->method_return_type_specifiers_reader()->length() > 1) {
            test_value_000 = TRUE;
          }
          else {
            if (subtype_ofP(method->type(), SGT_WALK_STELLA_LITERAL)) {
              test_value_000 = TRUE;
            }
            else {
              { boolean foundP_000 = FALSE;

                { Standard_Object* tspec = NULL;
                  Cons* iter_000 = method->method_parameter_type_specifiers_reader()->the_cons_list;

                  while (!(iter_000 == NIL)) {
                    tspec = ((Standard_Object*)(iter_000->value));
                    iter_000 = iter_000->rest;
                    if (subtype_ofP(type_spec_to_base_type(tspec), SGT_WALK_STELLA_LITERAL)) {
                      foundP_000 = TRUE;
                      break;
                    }
                  }
                }
                test_value_000 = foundP_000;
              }
            }
          }
        }
      }
    }
    { boolean value_000 = test_value_000;

      return (value_000);
    }
  }
}

Symbol* yield_evaluator_wrapper_name(Symbol* methodname) {
  return (intern_derived_symbol(methodname, string_concatenate(methodname->symbol_name, "-EVALUATOR-WRAPPER", 0)));
}

Object* yield_argument_access_tree(Symbol* argumentsvariable, int index, boolean restargumentP) {
  { int limit = 5;
    Object* accesstree = argumentsvariable;

    if (index <= limit) {
      { int i = NULL_INTEGER;
        int iter_000 = 2;
        int upper_bound_000 = index;
        boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;

        while (unboundedP_000 ||
            (iter_000 <= upper_bound_000)) {
          i = iter_000;
          iter_000 = iter_000 + 1;
          accesstree = listO(3, SYM_WALK_STELLA_REST, accesstree, NIL);
        }
      }
    }
    else {
      accesstree = listO(3, SYM_WALK_STELLA_NTH_REST, accesstree, cons(wrap_integer(index - 1), NIL));
    }
    if (!restargumentP) {
      if (index <= limit) {
        accesstree = listO(3, SYM_WALK_STELLA_VALUE, accesstree, NIL);
      }
      else {
        ((Cons*)(accesstree))->value = SYM_WALK_STELLA_NTH;
      }
    }
    return (accesstree);
  }
}

Cons* yield_evaluator_wrapper_tree(Method_Slot* method) {
  { Standard_Object* dummy1;

    { Object* nametree = yield_evaluator_wrapper_name(method->slot_name);
      Cons* calltree = cons(method->slot_name, NIL);
      List* parametertypespecs = method->method_parameter_type_specifiers_reader();
      Surrogate* resulttype = method->type();
      int nofparameters = parametertypespecs->length();
      boolean variableargumentsP = ((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value;

      { int index = NULL_INTEGER;
        int iter_000 = 1;
        int upper_bound_000 = nofparameters;
        boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;
        Cons* collect_000 = NULL;

        while (unboundedP_000 ||
            (iter_000 <= upper_bound_000)) {
          index = iter_000;
          iter_000 = iter_000 + 1;
          if (!((boolean)(collect_000))) {
            {
              collect_000 = cons(yield_argument_access_tree(SYM_WALK_STELLA_ARGUMENTS, index, variableargumentsP &&
                  (index == nofparameters)), NIL);
              if (calltree == NIL) {
                calltree = collect_000;
              }
              else {
                add_cons_to_end_of_cons_list(calltree, collect_000);
              }
            }
          }
          else {
            {
              collect_000->rest = cons(yield_argument_access_tree(SYM_WALK_STELLA_ARGUMENTS, index, variableargumentsP &&
                  (index == nofparameters)), NIL);
              collect_000 = collect_000->rest;
            }
          }
        }
      }
      { Standard_Object* ptype = NULL;
        Cons* iter_001 = parametertypespecs->the_cons_list;
        Cons_Iterator* it = calltree->rest->allocate_iterator();

        while ((!(iter_001 == NIL)) &&
            it->nextP()) {
          ptype = ((Standard_Object*)(iter_001->value));
          iter_001 = iter_001->rest;
          if (sub_type_spec_ofP(ptype, SGT_WALK_STELLA_LITERAL)) {
            it->value_setter(listO(3, SYM_WALK_STELLA_WRAPPER_VALUE, listO(3, SYM_WALK_STELLA_SAFE_CAST, it->value, cons(type_spec_to_base_type(ptype)->type_to_wrapped_type(), NIL)), NIL));
          }
          else if (sub_type_spec_ofP(ptype, SGT_WALK_STELLA_ARGUMENT_LIST)) {
            push_variable_binding(SYM_WALK_STELLA_ARGUMENTS, yield_listified_variable_arguments_type(method));
            it->value_setter(sys_tree(walk_without_type_tree(it->value), ptype, dummy1));
            pop_variable_binding();
          }
          else if (!(ptype == SGT_WALK_STELLA_OBJECT)) {
            it->value_setter(listO(3, SYM_WALK_STELLA_SAFE_CAST, it->value, cons(yield_type_spec_tree(ptype), NIL)));
          }
        }
      }
      if (voidP(resulttype)) {
      }
      else if (subtype_ofP(resulttype, SGT_WALK_STELLA_LITERAL)) {
        nametree = cons(nametree, cons(resulttype->type_to_wrapped_type(), NIL));
        if (method->type() == SGT_WALK_STELLA_BOOLEAN) {
          calltree = listO(3, SYM_WALK_STELLA_RETURN, listO(3, SYM_WALK_STELLA_WRAP_BOOLEAN, calltree, NIL), NIL);
        }
        else {
          calltree = listO(4, SYM_WALK_STELLA_LET, cons(listO(3, SYM_WALK_STELLA_RESULT, calltree, NIL), NIL), listO(5, SYM_WALK_STELLA_IF, listO(3, SYM_WALK_STELLA_DEFINEDp, SYM_WALK_STELLA_RESULT, NIL), listO(3, SYM_WALK_STELLA_RETURN, listO(3, SYM_WALK_STELLA_WRAP_LITERAL, SYM_WALK_STELLA_RESULT, NIL), NIL), listO(3, SYM_WALK_STELLA_RETURN, SYM_WALK_STELLA_NULL, NIL), NIL), NIL);
        }
      }
      else {
        nametree = cons(nametree, cons(resulttype, NIL));
        calltree = listO(3, SYM_WALK_STELLA_RETURN, calltree, NIL);
      }
      return (listO(3, SYM_WALK_STELLA_DEFUN, nametree, listO(5, cons(listO(3, SYM_WALK_STELLA_ARGUMENTS, SYM_WALK_STELLA_CONS, NIL), NIL), KWD_WALK_AUXILIARYp, SYM_WALK_STELLA_TRUE, calltree, NIL)));
    }
  }
}

Method_Slot* create_evaluator_wrapper_unit(Method_Slot* method) {
  { boolean variableargumentsP = ((Boolean_Wrapper*)(dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value;
    boolean evaluateargumentsP = method->method_evaluate_argumentsP_reader();
    Cons* wrappertree = yield_evaluator_wrapper_tree(method);
    Method_Slot* wrappermethod = NULL;

    if (method->method_return_type_specifiers_reader()->length() > 1) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        signal_translation_warning();
        if (!(suppress_warningsP())) {
          print_error_context(">> WARNING: ", STANDARD_WARNING);
          *(STANDARD_WARNING->native_stream) << endl << " " << "During interpreted command evaluation only the" << endl << " first return value will be usable" << "." << endl;
        }
      }
    }
    if (!evaluateargumentsP) {
      set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_EVALUATE_ARGUMENTSp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
    }
    if (variableargumentsP) {
      set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, (FALSE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
    }
    wrappermethod = ((Method_Slot*)(help_walk_auxiliary_tree(wrappertree, TRUE)->the_object));
    if (!evaluateargumentsP) {
      set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_EVALUATE_ARGUMENTSp, (FALSE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
    }
    if (variableargumentsP) {
      set_dynamic_slot_value(method->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
    }
    return (wrappermethod);
  }
}

DEFINE_STELLA_SPECIAL(oEVALUATIONTREEo, Object* , NULL);

DEFINE_STELLA_SPECIAL(oEVALUATIONPARENTTREEo, Object* , NULL);

Object* evaluate(Object* expression) {
  // Evaluate the expression `expression' and return the result.
  // Currently, only the evaluation of (possibly nested) commands and global
  // variables is supported.  The second return value indicates the actual type 
  // of the result (which might have been wrapped), and the third return value
  // indicates whether an error occurred during the evaluation.  Expressions
  // are simple to program in Common Lisp, since they are built into the language,
  // and relatively awkward in Java and C++.  Users of either of those 
  // languages are more likely to want to call `evaluate-string'.
  { Standard_Object* dummy1;
    Standard_Object* dummy2;

    { Object* otree = NULL;

      if (((boolean)(expression))) {
        if (safe_primary_type(expression) == SGT_WALK_STELLA_CONS) {
          { Object* expression_000 = expression;
            Cons* expression = ((Cons*)(expression_000));

            otree = evaluate_cons_tree(expression, dummy1);
          }
        }
        else {
          otree = evaluate_atomic_tree(expression, dummy2);
        }
      }
      return (otree);
    }
  }
}

Object* evaluate_string(char* expression) {
  // Evaluate the expression represented by `expression' and return the result.
  // This is equivalent to '(evaluate (unstringify expression))'.
  return (evaluate(read_s_expression_from_string(expression)));
}

Object* try_to_evaluate(Object* tree) {
  // Variant of `evaluate' that only evaluates `tree' if it
  // represents an evaluable expression.  If it does not, `tree' is returned
  // unmodified.  This can be used to implement commands with mixed argument
  // evaluation strategies.
  { Standard_Object* dummy1;

    { Object* otree = tree;

      if (((boolean)(tree))) {
        if (safe_primary_type(tree) == SGT_WALK_STELLA_CONS) {
          { Object* tree_000 = tree;
            Cons* tree = ((Cons*)(tree_000));

            if (symbolP(tree->value) &&
                ((boolean)(lookup_command(((Symbol*)(tree->value)))))) {
              try {
                otree = evaluate_cons_tree(tree, dummy1);
              }
              catch (Stella_Exception& _e) {
                Stella_Exception* e = &_e;

                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  signal_translation_error();
                  if (!(suppress_warningsP())) {
                    print_error_context(">> ERROR: ", STANDARD_ERROR);
                    *(STANDARD_ERROR->native_stream) << endl << " " << "`" << exception_message(e) << "'" << "." << endl;
                  }
                }
              }
            }
          }
        }
        else {
        }
      }
      return (otree);
    }
  }
}

Object* evaluate_cons_tree(Cons* tree, Standard_Object*& _return1) {
  { 
    BIND_STELLA_SPECIAL(oEVALUATIONPARENTTREEo, Object*, oEVALUATIONTREEo.get());
    BIND_STELLA_SPECIAL(oEVALUATIONTREEo, Object*, tree);
    { Object* operatorname = tree->value;
      Cons* arguments = tree->rest;
      Method_Slot* renamed_Operator = NULL;

      if (subtype_of_symbolP(safe_primary_type(operatorname))) {
        { Object* operatorname_000 = operatorname;
          Symbol* operatorname = ((Symbol*)(operatorname_000));

          if (operatorname == SYM_WALK_STELLA_QUOTE) {
            if (arguments->length() == 1) {
              { Object* _return_temp = arguments->value;

                _return1 = arguments->value->primary_type();
                return (_return_temp);
              }
            }
            { Output_String_Stream* stream_000 = new_output_string_stream();

              *(stream_000->native_stream) << "While evaluating '" << oEVALUATIONTREEo.get();
              if (((boolean)(oEVALUATIONPARENTTREEo.get()))) {
                *(stream_000->native_stream) << endl << "' inside '" << oEVALUATIONPARENTTREEo.get();
              }
              *(stream_000->native_stream) << "':" << endl;
              *(stream_000->native_stream) << "Illegal QUOTE expression";
              throw *new_evaluation_exception(stream_000->the_string_reader());
            }
          }
          else {
            { boolean test_value_000 = FALSE;

              { 
                renamed_Operator = lookup_command(operatorname);
                test_value_000 = ((boolean)(renamed_Operator));
              }
              test_value_000 = !test_value_000;
              if (test_value_000) {
                { Output_String_Stream* stream_001 = new_output_string_stream();

                  *(stream_001->native_stream) << "While evaluating '" << oEVALUATIONTREEo.get();
                  if (((boolean)(oEVALUATIONPARENTTREEo.get()))) {
                    *(stream_001->native_stream) << endl << "' inside '" << oEVALUATIONPARENTTREEo.get();
                  }
                  *(stream_001->native_stream) << "':" << endl;
                  *(stream_001->native_stream) << "Undefined operator: " << "`" << operatorname << "'";
                  throw *new_evaluation_exception(stream_001->the_string_reader());
                }
              }
            }
          }
        }
      }
      else {
        { Output_String_Stream* stream_002 = new_output_string_stream();

          *(stream_002->native_stream) << "While evaluating '" << oEVALUATIONTREEo.get();
          if (((boolean)(oEVALUATIONPARENTTREEo.get()))) {
            *(stream_002->native_stream) << endl << "' inside '" << oEVALUATIONPARENTTREEo.get();
          }
          *(stream_002->native_stream) << "':" << endl;
          *(stream_002->native_stream) << "Illegal operator: " << "`" << operatorname << "'";
          throw *new_evaluation_exception(stream_002->the_string_reader());
        }
      }
      { cpp_function_code evaluatorwrappercode = ((Function_Code_Wrapper*)(dynamic_slot_value(renamed_Operator->dynamic_slots, SYM_WALK_STELLA_EVALUATOR_WRAPPER_CODE, NULL_FUNCTION_CODE_WRAPPER)))->wrapper_value;
        boolean evaluateargsP = renamed_Operator->method_evaluate_argumentsP_reader();
        boolean variableargsP = ((Boolean_Wrapper*)(dynamic_slot_value(renamed_Operator->dynamic_slots, SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp, FALSE_WRAPPER)))->wrapper_value;
        List* parametertypes = renamed_Operator->method_parameter_type_specifiers_reader();
        int nofparameters = parametertypes->length();
        int nofargs = arguments->length();
        int minargs = (variableargsP ? (nofparameters - 1) : nofparameters);
        int maxargs = (variableargsP ? ((int)(NULL_INTEGER)) : nofparameters);
        Cons* unevaluatedargs = arguments;
        Object* evaluatedarg = NULL;
        Standard_Object* evaluatedargtype = NULL;
        int argindex = 0;
        Surrogate* returntype = renamed_Operator->type();
        Object* result = NULL;

        if ((nofargs < minargs) ||
            ((maxargs != NULL_INTEGER) &&
             (nofargs > maxargs))) {
          { Output_String_Stream* stream_003 = new_output_string_stream();

            *(stream_003->native_stream) << "While evaluating '" << oEVALUATIONTREEo.get();
            if (((boolean)(oEVALUATIONPARENTTREEo.get()))) {
              *(stream_003->native_stream) << endl << "' inside '" << oEVALUATIONPARENTTREEo.get();
            }
            *(stream_003->native_stream) << "':" << endl;
            *(stream_003->native_stream) << "Wrong number of arguments";
            throw *new_evaluation_exception(stream_003->the_string_reader());
          }
        }
        while (!(unevaluatedargs == NIL)) {
          evaluatedarg = evaluate_argument_tree(unevaluatedargs->value, evaluateargsP, evaluatedargtype);
          evaluatedarg = coerce_evaluated_tree(evaluatedarg, unevaluatedargs->value, evaluatedargtype, ((argindex >= minargs) ? variable_arguments_type(renamed_Operator) : ((Standard_Object*)(parametertypes->nth(argindex)))), evaluateargsP, evaluatedargtype);
          unevaluatedargs->value = evaluatedarg;
          unevaluatedargs = unevaluatedargs->rest;
          argindex = argindex + 1;
        }
        if (evaluatorwrappercode != NULL) {
          if (voidP(returntype)) {
            ((void  (*) (Cons*))evaluatorwrappercode)(arguments);
          }
          else {
            result = ((Object*  (*) (Cons*))evaluatorwrappercode)(arguments);
          }
        }
        else {
          if (voidP(returntype)) {
            apply(renamed_Operator->function_code, arguments);
          }
          else {
            result = apply(renamed_Operator->function_code, arguments);
          }
        }
        _return1 = returntype;
        return (result);
      }
    }
  }
}

Object* evaluate_argument_tree(Object* tree, boolean evaluateP, Standard_Object*& _return1) {
  { Surrogate* test_value_000 = safe_primary_type(tree);

    if (test_value_000 == SGT_WALK_STELLA_CONS) {
      { Object* tree_000 = tree;
        Cons* tree = ((Cons*)(tree_000));

        if (evaluateP) {
          return (evaluate_cons_tree(tree, _return1));
        }
        else {
          _return1 = SGT_WALK_STELLA_CONS;
          return (tree);
        }
      }
    }
    else if (subtype_of_wrapperP(test_value_000)) {
      { Object* tree_001 = tree;
        Wrapper* tree = ((Wrapper*)(tree_001));

        { Object* _return_temp = tree;

          _return1 = wrapped_type_to_type(tree->primary_type());
          return (_return_temp);
        }
      }
    }
    else {
      if (evaluateP) {
        return (evaluate_atomic_tree(tree, _return1));
      }
      else {
        { Object* _return_temp = tree;

          _return1 = tree->primary_type();
          return (_return_temp);
        }
      }
    }
  }
}

Object* evaluate_atomic_tree(Object* tree, Standard_Object*& _return1) {
  { 
    BIND_STELLA_SPECIAL(oEVALUATIONPARENTTREEo, Object*, oEVALUATIONTREEo.get());
    BIND_STELLA_SPECIAL(oEVALUATIONTREEo, Object*, tree);
    { Surrogate* test_value_000 = safe_primary_type(tree);

      if (subtype_of_symbolP(test_value_000)) {
        { Object* tree_000 = tree;
          Symbol* tree = ((Symbol*)(tree_000));

          if (tree == SYM_WALK_STELLA_NULL) {
            _return1 = SGT_WALK_STELLA_UNKNOWN;
            return (tree);
          }
          else if (tree == SYM_WALK_STELLA_TRUE) {
            _return1 = SGT_WALK_STELLA_BOOLEAN;
            return (TRUE_WRAPPER);
          }
          else if (tree == SYM_WALK_STELLA_FALSE) {
            _return1 = SGT_WALK_STELLA_BOOLEAN;
            return (FALSE_WRAPPER);
          }
          else {
            { Output_String_Stream* stream_000 = new_output_string_stream();

              *(stream_000->native_stream) << "While evaluating '" << oEVALUATIONTREEo.get();
              if (((boolean)(oEVALUATIONPARENTTREEo.get()))) {
                *(stream_000->native_stream) << endl << "' inside '" << oEVALUATIONPARENTTREEo.get();
              }
              *(stream_000->native_stream) << "':" << endl;
              *(stream_000->native_stream) << "Variable evaluation not yet implemented";
              throw *new_evaluation_exception(stream_000->the_string_reader());
            }
          }
        }
      }
      else if (subtype_of_surrogateP(test_value_000)) {
        { Object* tree_001 = tree;
          Surrogate* tree = ((Surrogate*)(tree_001));

          _return1 = SGT_WALK_STELLA_SURROGATE;
          return (tree);
        }
      }
      else if (subtype_of_keywordP(test_value_000)) {
        { Object* tree_002 = tree;
          Keyword* tree = ((Keyword*)(tree_002));

          _return1 = SGT_WALK_STELLA_KEYWORD;
          return (tree);
        }
      }
      else if (subtype_of_wrapperP(test_value_000)) {
        { Object* tree_003 = tree;
          Wrapper* tree = ((Wrapper*)(tree_003));

          { Object* _return_temp = tree;

            _return1 = wrapped_type_to_type(tree->primary_type());
            return (_return_temp);
          }
        }
      }
      else {
        { Output_String_Stream* stream_001 = new_output_string_stream();

          *(stream_001->native_stream) << "While evaluating '" << oEVALUATIONTREEo.get();
          if (((boolean)(oEVALUATIONPARENTTREEo.get()))) {
            *(stream_001->native_stream) << endl << "' inside '" << oEVALUATIONPARENTTREEo.get();
          }
          *(stream_001->native_stream) << "':" << endl;
          *(stream_001->native_stream) << "Unknown atomic expression type: " << "`" << tree << "'";
          throw *new_evaluation_exception(stream_001->the_string_reader());
        }
      }
    }
  }
}

Object* coerce_evaluated_tree(Object* tree, Object* sourcetree, Standard_Object* sourcetype, Standard_Object* targettype, boolean evaluateP, Standard_Object*& _return1) {
  { 
    BIND_STELLA_SPECIAL(oEVALUATIONPARENTTREEo, Object*, oEVALUATIONTREEo.get());
    BIND_STELLA_SPECIAL(oEVALUATIONTREEo, Object*, sourcetree);
    { Surrogate* sourcebasetype = type_spec_to_base_type(sourcetype);
      Surrogate* targetbasetype = type_spec_to_base_type(targettype);

      if (tree == SYM_WALK_STELLA_NULL) {
        { Generalized_Symbol* test_value_000 = ((Generalized_Symbol*)(targettype));

          if (test_value_000 == SGT_WALK_STELLA_INTEGER) {
            _return1 = targettype;
            return (NULL_INTEGER_WRAPPER);
          }
          else if (test_value_000 == SGT_WALK_STELLA_FLOAT) {
            _return1 = targettype;
            return (NULL_FLOAT_WRAPPER);
          }
          else if (test_value_000 == SGT_WALK_STELLA_STRING) {
            _return1 = targettype;
            return (NULL_STRING_WRAPPER);
          }
          else if (test_value_000 == SGT_WALK_STELLA_BOOLEAN) {
            _return1 = targettype;
            return (FALSE_WRAPPER);
          }
          else {
            _return1 = targettype;
            return (NULL);
          }
        }
      }
      if ((targettype == SGT_WALK_STELLA_BOOLEAN) &&
          (!evaluateP)) {
        if (tree == SYM_WALK_STELLA_TRUE) {
          _return1 = targettype;
          return (TRUE_WRAPPER);
        }
        if (tree == SYM_WALK_STELLA_FALSE) {
          _return1 = targettype;
          return (FALSE_WRAPPER);
        }
      }
      if (subtype_ofP(sourcebasetype, targetbasetype)) {
        _return1 = sourcetype;
        return (tree);
      }
      if ((!evaluateP) &&
          (subtype_ofP(sourcebasetype, SGT_WALK_STELLA_LITERAL) &&
           subtype_ofP(sourcebasetype->type_to_wrapped_type(), targetbasetype))) {
        { Object* _return_temp = tree;

          _return1 = sourcebasetype->type_to_wrapped_type();
          return (_return_temp);
        }
      }
      if (subtype_ofP(targetbasetype, SGT_WALK_STELLA_LITERAL) &&
          subtype_ofP(sourcebasetype, targetbasetype->type_to_wrapped_type())) {
        _return1 = targettype;
        return (tree);
      }
      if (voidP(sourcetype)) {
        { Output_String_Stream* stream_000 = new_output_string_stream();

          *(stream_000->native_stream) << "While evaluating '" << oEVALUATIONTREEo.get();
          if (((boolean)(oEVALUATIONPARENTTREEo.get()))) {
            *(stream_000->native_stream) << endl << "' inside '" << oEVALUATIONPARENTTREEo.get();
          }
          *(stream_000->native_stream) << "':" << endl;
          *(stream_000->native_stream) << "`" << sourcetree << "'" << " does not return a value";
          throw *new_evaluation_exception(stream_000->the_string_reader());
        }
      }
      else {
        { Output_String_Stream* stream_001 = new_output_string_stream();

          *(stream_001->native_stream) << "While evaluating '" << oEVALUATIONTREEo.get();
          if (((boolean)(oEVALUATIONPARENTTREEo.get()))) {
            *(stream_001->native_stream) << endl << "' inside '" << oEVALUATIONPARENTTREEo.get();
          }
          *(stream_001->native_stream) << "':" << endl;
          *(stream_001->native_stream) << "Type conflict: " << "`" << yield_type_spec_tree(sourcetype) << "'" << " found where " << "`" << yield_type_spec_tree(targettype) << "'" << " expected";
          throw *new_evaluation_exception(stream_001->the_string_reader());
        }
      }
    }
  }
}

void help_startup_walk1() {
  {
    KWD_WALK_WARN_ABOUT_UNDEFINED_METHODS = ((Keyword*)(intern_rigid_symbol_wrt_module("WARN-ABOUT-UNDEFINED-METHODS", NULL, 2)));
    KWD_WALK_SUPPRESS_WARNINGS = ((Keyword*)(intern_rigid_symbol_wrt_module("SUPPRESS-WARNINGS", NULL, 2)));
    KWD_WALK_WARN_ABOUT_MISSING_METHODS = ((Keyword*)(intern_rigid_symbol_wrt_module("WARN-ABOUT-MISSING-METHODS", NULL, 2)));
    KWD_WALK_USE_HARDCODED_SYMBOLS = ((Keyword*)(intern_rigid_symbol_wrt_module("USE-HARDCODED-SYMBOLS", NULL, 2)));
    KWD_WALK_USE_COMMON_LISP_STRUCTS = ((Keyword*)(intern_rigid_symbol_wrt_module("USE-COMMON-LISP-STRUCTS", NULL, 2)));
    KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS = ((Keyword*)(intern_rigid_symbol_wrt_module("USE-COMMON-LISP-VECTOR-STRUCTS", NULL, 2)));
    KWD_WALK_USE_COMMON_LISP_CONSES = ((Keyword*)(intern_rigid_symbol_wrt_module("USE-COMMON-LISP-CONSES", NULL, 2)));
    KWD_WALK_USE_CPP_GARBAGE_COLLECTOR = ((Keyword*)(intern_rigid_symbol_wrt_module("USE-CPP-GARBAGE-COLLECTOR", NULL, 2)));
    KWD_WALK_TRANSLATE_WITH_COPYRIGHT_HEADER = ((Keyword*)(intern_rigid_symbol_wrt_module("TRANSLATE-WITH-COPYRIGHT-HEADER", NULL, 2)));
    SYM_WALK_STELLA_WHEN = ((Symbol*)(intern_rigid_symbol_wrt_module("WHEN", NULL, 0)));
    SYM_WALK_STELLA_AND = ((Symbol*)(intern_rigid_symbol_wrt_module("AND", NULL, 0)));
    SYM_WALK_STELLA_DEFINEDp = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFINED?", NULL, 0)));
    SYM_WALK_STELLA_oTRACED_KEYWORDSo = ((Symbol*)(intern_rigid_symbol_wrt_module("*TRACED-KEYWORDS*", NULL, 0)));
    SYM_WALK_STELLA_PRINT = ((Symbol*)(intern_rigid_symbol_wrt_module("PRINT", NULL, 0)));
    SYM_WALK_STELLA_EXISTS = ((Symbol*)(intern_rigid_symbol_wrt_module("EXISTS", NULL, 0)));
    SYM_WALK_STELLA_KWD = ((Symbol*)(intern_rigid_symbol_wrt_module("KWD", NULL, 0)));
    SYM_WALK_STELLA_IN = ((Symbol*)(intern_rigid_symbol_wrt_module("IN", NULL, 0)));
    SYM_WALK_STELLA_QUOTE = ((Symbol*)(intern_rigid_symbol_wrt_module("QUOTE", NULL, 0)));
    SYM_WALK_STELLA_WHERE = ((Symbol*)(intern_rigid_symbol_wrt_module("WHERE", NULL, 0)));
    SYM_WALK_STELLA_MEMBp = ((Symbol*)(intern_rigid_symbol_wrt_module("MEMB?", NULL, 0)));
    SYM_WALK_STELLA_TRACE_IF = ((Symbol*)(intern_rigid_symbol_wrt_module("TRACE-IF", NULL, 0)));
    SYM_WALK_STELLA_TRACE_IF_EXPANDER = ((Symbol*)(intern_rigid_symbol_wrt_module("TRACE-IF-EXPANDER", NULL, 0)));
    KWD_WALK_JAVA = ((Keyword*)(intern_rigid_symbol_wrt_module("JAVA", NULL, 2)));
    SGT_WALK_STELLA_TRANSLATION_UNIT = ((Surrogate*)(intern_rigid_symbol_wrt_module("TRANSLATION-UNIT", NULL, 1)));
    SYM_WALK_STELLA_TU_HOME_MODULE = ((Symbol*)(intern_rigid_symbol_wrt_module("TU-HOME-MODULE", NULL, 0)));
    SYM_WALK_STELLA_THE_OBJECT = ((Symbol*)(intern_rigid_symbol_wrt_module("THE-OBJECT", NULL, 0)));
    SYM_WALK_STELLA_CATEGORY = ((Symbol*)(intern_rigid_symbol_wrt_module("CATEGORY", NULL, 0)));
    SYM_WALK_STELLA_ANNOTATION = ((Symbol*)(intern_rigid_symbol_wrt_module("ANNOTATION", NULL, 0)));
    SYM_WALK_STELLA_AUXILIARYp = ((Symbol*)(intern_rigid_symbol_wrt_module("AUXILIARY?", NULL, 0)));
    SYM_WALK_STELLA_CODE_REGISTER = ((Symbol*)(intern_rigid_symbol_wrt_module("CODE-REGISTER", NULL, 0)));
    SYM_WALK_STELLA_TRANSLATION = ((Symbol*)(intern_rigid_symbol_wrt_module("TRANSLATION", NULL, 0)));
    SYM_WALK_STELLA_REFERENCED_GLOBALS = ((Symbol*)(intern_rigid_symbol_wrt_module("REFERENCED-GLOBALS", NULL, 0)));
    SYM_WALK_STELLA_CLASS = ((Symbol*)(intern_rigid_symbol_wrt_module("CLASS", NULL, 0)));
    SYM_WALK_STELLA_METHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD", NULL, 0)));
    SYM_WALK_STELLA_MACRO = ((Symbol*)(intern_rigid_symbol_wrt_module("MACRO", NULL, 0)));
    SYM_WALK_STELLA_GLOBAL_VARIABLE = ((Symbol*)(intern_rigid_symbol_wrt_module("GLOBAL-VARIABLE", NULL, 0)));
    KWD_WALK_COMMON_LISP = ((Keyword*)(intern_rigid_symbol_wrt_module("COMMON-LISP", NULL, 2)));
    KWD_WALK_CPP = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP", NULL, 2)));
    KWD_WALK_CPP_STANDALONE = ((Keyword*)(intern_rigid_symbol_wrt_module("CPP-STANDALONE", NULL, 2)));
    KWD_WALK_IDL = ((Keyword*)(intern_rigid_symbol_wrt_module("IDL", NULL, 2)));
    SYM_WALK_STELLA_PROGN = ((Symbol*)(intern_rigid_symbol_wrt_module("PROGN", NULL, 0)));
    SGT_WALK_STELLA_CONS = ((Surrogate*)(intern_rigid_symbol_wrt_module("CONS", NULL, 1)));
    SYM_WALK_STELLA_DEFCLASS = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFCLASS", NULL, 0)));
    SYM_WALK_STELLA_DEFSLOT = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFSLOT", NULL, 0)));
    SYM_WALK_STELLA_DEFMETHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFMETHOD", NULL, 0)));
    SYM_WALK_STELLA_DEFUN = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFUN", NULL, 0)));
    SYM_WALK_STELLA_DEFOPERATOR = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFOPERATOR", NULL, 0)));
    SYM_WALK_STELLA_DEFGLOBAL = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFGLOBAL", NULL, 0)));
    SYM_WALK_STELLA_DEFSPECIAL = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFSPECIAL", NULL, 0)));
    SYM_WALK_STELLA_DEFCONSTANT = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFCONSTANT", NULL, 0)));
    SYM_WALK_STELLA_DEFTYPE = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFTYPE", NULL, 0)));
    SYM_WALK_STELLA_DEFMODULE = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFMODULE", NULL, 0)));
    SYM_WALK_STELLA_DEFMACRO = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFMACRO", NULL, 0)));
    SYM_WALK_STELLA_STARTUP_TIME_PROGN = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-TIME-PROGN", NULL, 0)));
    SYM_WALK_STELLA_VERBATIM = ((Symbol*)(intern_rigid_symbol_wrt_module("VERBATIM", NULL, 0)));
    SYM_WALK_STELLA_MODULE = ((Symbol*)(intern_rigid_symbol_wrt_module("MODULE", NULL, 0)));
    SYM_WALK_STELLA_TYPE = ((Symbol*)(intern_rigid_symbol_wrt_module("TYPE", NULL, 0)));
    SYM_WALK_STELLA_PRINT_METHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("PRINT-METHOD", NULL, 0)));
    SYM_WALK_STELLA_SLOT = ((Symbol*)(intern_rigid_symbol_wrt_module("SLOT", NULL, 0)));
    KWD_WALK_JAVA_STANDALONE = ((Keyword*)(intern_rigid_symbol_wrt_module("JAVA-STANDALONE", NULL, 2)));
  }
}

void help_startup_walk2() {
  {
    SYM_WALK_STELLA_METHOD_CONSTRUCTORp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-CONSTRUCTOR?", NULL, 0)));
    KWD_WALK_DEFINE = ((Keyword*)(intern_rigid_symbol_wrt_module("DEFINE", NULL, 2)));
    KWD_WALK_FINALIZE = ((Keyword*)(intern_rigid_symbol_wrt_module("FINALIZE", NULL, 2)));
    KWD_WALK_WALK = ((Keyword*)(intern_rigid_symbol_wrt_module("WALK", NULL, 2)));
    KWD_WALK_TRANSLATE = ((Keyword*)(intern_rigid_symbol_wrt_module("TRANSLATE", NULL, 2)));
    SGT_WALK_STELLA_VOID = ((Surrogate*)(intern_rigid_symbol_wrt_module("VOID", NULL, 1)));
    SGT_WALK_STELLA_GLOBAL_VARIABLE = ((Surrogate*)(intern_rigid_symbol_wrt_module("GLOBAL-VARIABLE", NULL, 1)));
    SGT_WALK_STELLA_TYPE_SPEC = ((Surrogate*)(intern_rigid_symbol_wrt_module("TYPE-SPEC", NULL, 1)));
    SGT_WALK_STELLA_MODULE = ((Surrogate*)(intern_rigid_symbol_wrt_module("MODULE", NULL, 1)));
    SYM_WALK_STELLA_SLOT_AUXILIARYp = ((Symbol*)(intern_rigid_symbol_wrt_module("SLOT-AUXILIARY?", NULL, 0)));
    SYM_WALK_STELLA_METHOD_GLOBALLY_INLINEp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-GLOBALLY-INLINE?", NULL, 0)));
    SYM_WALK_STELLA_LOG_FUNCTION_CALL = ((Symbol*)(intern_rigid_symbol_wrt_module("LOG-FUNCTION-CALL", NULL, 0)));
    SYM_WALK_STELLA_START_FUNCTION_CALL_LOGGING = ((Symbol*)(intern_rigid_symbol_wrt_module("START-FUNCTION-CALL-LOGGING", NULL, 0)));
    SYM_WALK_STELLA_STOP_FUNCTION_CALL_LOGGING = ((Symbol*)(intern_rigid_symbol_wrt_module("STOP-FUNCTION-CALL-LOGGING", NULL, 0)));
    SYM_WALK_STELLA_SET_CALL_LOG_BREAK_POINT = ((Symbol*)(intern_rigid_symbol_wrt_module("SET-CALL-LOG-BREAK-POINT", NULL, 0)));
    SYM_WALK_STELLA_BREAK_PROGRAM = ((Symbol*)(intern_rigid_symbol_wrt_module("BREAK-PROGRAM", NULL, 0)));
    SYM_WALK_STELLA_TERMINATE_PROGRAM = ((Symbol*)(intern_rigid_symbol_wrt_module("TERMINATE-PROGRAM", NULL, 0)));
    SYM_WALK_STELLA_PO = ((Symbol*)(intern_rigid_symbol_wrt_module("PO", NULL, 0)));
    SYM_WALK_STELLA_NULLp = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL?", NULL, 0)));
    SYM_WALK_STELLA_EQp = ((Symbol*)(intern_rigid_symbol_wrt_module("EQ?", NULL, 0)));
    SYM_WALK_STELLA_EQLp = ((Symbol*)(intern_rigid_symbol_wrt_module("EQL?", NULL, 0)));
    SYM_WALK_STELLA_l = ((Symbol*)(intern_rigid_symbol_wrt_module("<", NULL, 0)));
    SYM_WALK_STELLA_le = ((Symbol*)(intern_rigid_symbol_wrt_module("<=", NULL, 0)));
    SYM_WALK_STELLA_g = ((Symbol*)(intern_rigid_symbol_wrt_module(">", NULL, 0)));
    SYM_WALK_STELLA_ge = ((Symbol*)(intern_rigid_symbol_wrt_module(">=", NULL, 0)));
    SYM_WALK_STELLA_CAST = ((Symbol*)(intern_rigid_symbol_wrt_module("CAST", NULL, 0)));
    SGT_WALK_STELLA_UNKNOWN = ((Surrogate*)(intern_rigid_symbol_wrt_module("UNKNOWN", NULL, 1)));
    SGT_WALK_STELLA_LISP_CODE = ((Surrogate*)(intern_rigid_symbol_wrt_module("LISP-CODE", NULL, 1)));
    SGT_WALK_STELLA_OUTPUT_STREAM = ((Surrogate*)(intern_rigid_symbol_wrt_module("OUTPUT-STREAM", NULL, 1)));
    SGT_WALK_STELLA_NATIVE_OUTPUT_STREAM = ((Surrogate*)(intern_rigid_symbol_wrt_module("NATIVE-OUTPUT-STREAM", NULL, 1)));
    SYM_WALK_STELLA_INLINE = ((Symbol*)(intern_rigid_symbol_wrt_module("INLINE", NULL, 0)));
    SGT_WALK_STELLA_UNINITIALIZED = ((Surrogate*)(intern_rigid_symbol_wrt_module("UNINITIALIZED", NULL, 1)));
    KWD_WALK_SECOND_CLASS = ((Keyword*)(intern_rigid_symbol_wrt_module("SECOND-CLASS", NULL, 2)));
    SGT_WALK_STELLA_KEY_VALUE_LIST = ((Surrogate*)(intern_rigid_symbol_wrt_module("KEY-VALUE-LIST", NULL, 1)));
    SYM_WALK_STELLA_NULL = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL", NULL, 0)));
    KWD_WALK_TYPE = ((Keyword*)(intern_rigid_symbol_wrt_module("TYPE", NULL, 2)));
    KWD_WALK_ALIAS = ((Keyword*)(intern_rigid_symbol_wrt_module("ALIAS", NULL, 2)));
    SYM_WALK_STELLA_ERROR = ((Symbol*)(intern_rigid_symbol_wrt_module("ERROR", NULL, 0)));
    SYM_WALK_STELLA_EOL = ((Symbol*)(intern_rigid_symbol_wrt_module("EOL", NULL, 0)));
    SYM_WALK_STELLA_ILLEGAL_EXPRESSION_FLAGGED_BY_THE_TRANSLATOR = ((Symbol*)(intern_rigid_symbol_wrt_module("ILLEGAL-EXPRESSION-FLAGGED-BY-THE-TRANSLATOR", NULL, 0)));
    SYM_WALK_STELLA_BAD_SYS = ((Symbol*)(intern_rigid_symbol_wrt_module("BAD-SYS", NULL, 0)));
    SGT_WALK_STELLA_INTEGER_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("INTEGER-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_INTEGER = ((Surrogate*)(intern_rigid_symbol_wrt_module("INTEGER", NULL, 1)));
    SGT_WALK_STELLA_FLOAT_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("FLOAT-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_FLOAT = ((Surrogate*)(intern_rigid_symbol_wrt_module("FLOAT", NULL, 1)));
    SGT_WALK_STELLA_NUMBER_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("NUMBER-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_NUMBER = ((Surrogate*)(intern_rigid_symbol_wrt_module("NUMBER", NULL, 1)));
    SGT_WALK_STELLA_STRING_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("STRING-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_STRING = ((Surrogate*)(intern_rigid_symbol_wrt_module("STRING", NULL, 1)));
    SGT_WALK_STELLA_MUTABLE_STRING_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("MUTABLE-STRING-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_MUTABLE_STRING = ((Surrogate*)(intern_rigid_symbol_wrt_module("MUTABLE-STRING", NULL, 1)));
    SGT_WALK_STELLA_CHARACTER_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("CHARACTER-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_CHARACTER = ((Surrogate*)(intern_rigid_symbol_wrt_module("CHARACTER", NULL, 1)));
    SGT_WALK_STELLA_BOOLEAN_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("BOOLEAN-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_BOOLEAN = ((Surrogate*)(intern_rigid_symbol_wrt_module("BOOLEAN", NULL, 1)));
    SGT_WALK_STELLA_FUNCTION_CODE_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("FUNCTION-CODE-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_FUNCTION_CODE = ((Surrogate*)(intern_rigid_symbol_wrt_module("FUNCTION-CODE", NULL, 1)));
    SGT_WALK_STELLA_METHOD_CODE_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("METHOD-CODE-WRAPPER", NULL, 1)));
    SGT_WALK_STELLA_METHOD_CODE = ((Surrogate*)(intern_rigid_symbol_wrt_module("METHOD-CODE", NULL, 1)));
    SGT_WALK_STELLA_OBJECT = ((Surrogate*)(intern_rigid_symbol_wrt_module("OBJECT", NULL, 1)));
  }
}

void help_startup_walk3() {
  {
    SYM_WALK_STELLA_INLINE_WRAP_BOOLEAN = ((Symbol*)(intern_rigid_symbol_wrt_module("INLINE-WRAP-BOOLEAN", NULL, 0)));
    SYM_WALK_STELLA_WRAP_LITERAL = ((Symbol*)(intern_rigid_symbol_wrt_module("WRAP-LITERAL", NULL, 0)));
    SYM_WALK_STELLA_INTEGER_TO_BOOLEAN_WRAPPER = ((Symbol*)(intern_rigid_symbol_wrt_module("INTEGER-TO-BOOLEAN-WRAPPER", NULL, 0)));
    SYM_WALK_STELLA_INTEGER_TO_BOOLEAN = ((Symbol*)(intern_rigid_symbol_wrt_module("INTEGER-TO-BOOLEAN", NULL, 0)));
    SYM_WALK_STELLA_MUTABLE_STRING_TO_STRING = ((Symbol*)(intern_rigid_symbol_wrt_module("MUTABLE-STRING-TO-STRING", NULL, 0)));
    SYM_WALK_STELLA_STRING_TO_MUTABLE_STRING = ((Symbol*)(intern_rigid_symbol_wrt_module("STRING-TO-MUTABLE-STRING", NULL, 0)));
    SGT_WALK_STELLA_SYMBOL = ((Surrogate*)(intern_rigid_symbol_wrt_module("SYMBOL", NULL, 1)));
    SYM_WALK_STELLA_INTERN_SYMBOL = ((Symbol*)(intern_rigid_symbol_wrt_module("INTERN-SYMBOL", NULL, 0)));
    SYM_WALK_STELLA_CHARACTER_TO_STRING = ((Symbol*)(intern_rigid_symbol_wrt_module("CHARACTER-TO-STRING", NULL, 0)));
    SYM_WALK_STELLA_SYMBOL_NAME = ((Symbol*)(intern_rigid_symbol_wrt_module("SYMBOL-NAME", NULL, 0)));
    SYM_WALK_STELLA_INLINE_UNWRAP_BOOLEAN = ((Symbol*)(intern_rigid_symbol_wrt_module("INLINE-UNWRAP-BOOLEAN", NULL, 0)));
    SYM_WALK_STELLA_WRAPPER_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("WRAPPER-VALUE", NULL, 0)));
    SYM_WALK_STELLA_NUMBER_WRAPPER_TO_FLOAT = ((Symbol*)(intern_rigid_symbol_wrt_module("NUMBER-WRAPPER-TO-FLOAT", NULL, 0)));
    SGT_WALK_STELLA_SURROGATE = ((Surrogate*)(intern_rigid_symbol_wrt_module("SURROGATE", NULL, 1)));
    SGT_WALK_STELLA_CLASS = ((Surrogate*)(intern_rigid_symbol_wrt_module("CLASS", NULL, 1)));
    SYM_WALK_STELLA_SURROGATE_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("SURROGATE-VALUE", NULL, 0)));
    SGT_WALK_STELLA_INPUT_STREAM = ((Surrogate*)(intern_rigid_symbol_wrt_module("INPUT-STREAM", NULL, 1)));
    SGT_WALK_STELLA_NATIVE_INPUT_STREAM = ((Surrogate*)(intern_rigid_symbol_wrt_module("NATIVE-INPUT-STREAM", NULL, 1)));
    SYM_WALK_STELLA_NATIVE_STREAM = ((Symbol*)(intern_rigid_symbol_wrt_module("NATIVE-STREAM", NULL, 0)));
    SYM_WALK_STELLA_lXg = ((Symbol*)(intern_rigid_symbol_wrt_module("<X>", NULL, 0)));
    SGT_WALK_STELLA_SINGLE_FLOAT = ((Surrogate*)(intern_rigid_symbol_wrt_module("SINGLE-FLOAT", NULL, 1)));
    SYM_WALK_STELLA_IDENTITY = ((Symbol*)(intern_rigid_symbol_wrt_module("IDENTITY", NULL, 0)));
    SGT_WALK_STELLA_DOUBLE_FLOAT = ((Surrogate*)(intern_rigid_symbol_wrt_module("DOUBLE-FLOAT", NULL, 1)));
    SGT_WALK_STELLA_SHORT_INTEGER = ((Surrogate*)(intern_rigid_symbol_wrt_module("SHORT-INTEGER", NULL, 1)));
    SGT_WALK_STELLA_LONG_INTEGER = ((Surrogate*)(intern_rigid_symbol_wrt_module("LONG-INTEGER", NULL, 1)));
    SGT_WALK_STELLA_UNSIGNED_SHORT_INTEGER = ((Surrogate*)(intern_rigid_symbol_wrt_module("UNSIGNED-SHORT-INTEGER", NULL, 1)));
    SGT_WALK_STELLA_UNSIGNED_LONG_INTEGER = ((Surrogate*)(intern_rigid_symbol_wrt_module("UNSIGNED-LONG-INTEGER", NULL, 1)));
    SYM_WALK_STELLA_VRLET = ((Symbol*)(intern_rigid_symbol_wrt_module("VRLET", NULL, 0)));
    SYM_WALK_STELLA_SYS_CALL_METHOD = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-CALL-METHOD", NULL, 0)));
    SYM_WALK_STELLA_SYS_CALL_FUNCTION = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-CALL-FUNCTION", NULL, 0)));
    SYM_WALK_STELLA_TYPED_SYS = ((Symbol*)(intern_rigid_symbol_wrt_module("TYPED-SYS", NULL, 0)));
    SYM_WALK_STELLA_SYS_INLINE_CALL = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-INLINE-CALL", NULL, 0)));
    SYM_WALK_STELLA_SAFE_CAST = ((Symbol*)(intern_rigid_symbol_wrt_module("SAFE-CAST", NULL, 0)));
    SGT_WALK_STELLA_NON_OBJECT = ((Surrogate*)(intern_rigid_symbol_wrt_module("NON-OBJECT", NULL, 1)));
    SGT_WALK_STELLA_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("WRAPPER", NULL, 1)));
    SYM_WALK_STELLA_LET = ((Symbol*)(intern_rigid_symbol_wrt_module("LET", NULL, 0)));
    SYM_WALK_STELLA_VOID_SYS = ((Symbol*)(intern_rigid_symbol_wrt_module("VOID-SYS", NULL, 0)));
    SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER = ((Symbol*)(intern_rigid_symbol_wrt_module("SLOT-TYPE-SPECIFIER", NULL, 0)));
    SYM_WALK_STELLA_SYS_FOREACH = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-FOREACH", NULL, 0)));
    SYM_WALK_STELLA_SETQ = ((Symbol*)(intern_rigid_symbol_wrt_module("SETQ", NULL, 0)));
    SYM_WALK_STELLA_SETF = ((Symbol*)(intern_rigid_symbol_wrt_module("SETF", NULL, 0)));
    SYM_WALK_STELLA_SLOT_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("SLOT-VALUE", NULL, 0)));
    SYM_WALK_STELLA_SLOT_VALUE_SETTER = ((Symbol*)(intern_rigid_symbol_wrt_module("SLOT-VALUE-SETTER", NULL, 0)));
    SYM_WALK_STELLA_AREF = ((Symbol*)(intern_rigid_symbol_wrt_module("AREF", NULL, 0)));
    SYM_WALK_STELLA_AREF_SETTER = ((Symbol*)(intern_rigid_symbol_wrt_module("AREF-SETTER", NULL, 0)));
    SYM_WALK_STELLA_WITH_PROCESS_LOCK = ((Symbol*)(intern_rigid_symbol_wrt_module("WITH-PROCESS-LOCK", NULL, 0)));
    SYM_WALK_STELLA_LOOP = ((Symbol*)(intern_rigid_symbol_wrt_module("LOOP", NULL, 0)));
    SYM_WALK_STELLA_VALUES = ((Symbol*)(intern_rigid_symbol_wrt_module("VALUES", NULL, 0)));
    SYM_WALK_STELLA_SPECIAL = ((Symbol*)(intern_rigid_symbol_wrt_module("SPECIAL", NULL, 0)));
    SYM_WALK_STELLA_SYS_SPECIAL = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-SPECIAL", NULL, 0)));
    SYM_WALK_STELLA_RETURN = ((Symbol*)(intern_rigid_symbol_wrt_module("RETURN", NULL, 0)));
    SYM_WALK_STELLA_MV_SETQ = ((Symbol*)(intern_rigid_symbol_wrt_module("MV-SETQ", NULL, 0)));
    SYM_WALK_STELLA_MV_BIND = ((Symbol*)(intern_rigid_symbol_wrt_module("MV-BIND", NULL, 0)));
    SYM_WALK_STELLA_BREAK = ((Symbol*)(intern_rigid_symbol_wrt_module("BREAK", NULL, 0)));
    SYM_WALK_STELLA_CONTINUE = ((Symbol*)(intern_rigid_symbol_wrt_module("CONTINUE", NULL, 0)));
    SYM_WALK_STELLA_WHILE = ((Symbol*)(intern_rigid_symbol_wrt_module("WHILE", NULL, 0)));
    SYM_WALK_STELLA_FOREACH = ((Symbol*)(intern_rigid_symbol_wrt_module("FOREACH", NULL, 0)));
    SYM_WALK_STELLA_FORALL = ((Symbol*)(intern_rigid_symbol_wrt_module("FORALL", NULL, 0)));
    SYM_WALK_STELLA_SOME = ((Symbol*)(intern_rigid_symbol_wrt_module("SOME", NULL, 0)));
    SYM_WALK_STELLA_SETOF = ((Symbol*)(intern_rigid_symbol_wrt_module("SETOF", NULL, 0)));
  }
}

void help_startup_walk4() {
  {
    SYM_WALK_STELLA_SET_OF = ((Symbol*)(intern_rigid_symbol_wrt_module("SET-OF", NULL, 0)));
    SYM_WALK_STELLA_SELECT = ((Symbol*)(intern_rigid_symbol_wrt_module("SELECT", NULL, 0)));
    SYM_WALK_STELLA_IF = ((Symbol*)(intern_rigid_symbol_wrt_module("IF", NULL, 0)));
    SYM_WALK_STELLA_UNLESS = ((Symbol*)(intern_rigid_symbol_wrt_module("UNLESS", NULL, 0)));
    SYM_WALK_STELLA_CHOOSE = ((Symbol*)(intern_rigid_symbol_wrt_module("CHOOSE", NULL, 0)));
    SYM_WALK_STELLA_OR = ((Symbol*)(intern_rigid_symbol_wrt_module("OR", NULL, 0)));
    SYM_WALK_STELLA_NOT = ((Symbol*)(intern_rigid_symbol_wrt_module("NOT", NULL, 0)));
    SYM_WALK_STELLA_COND = ((Symbol*)(intern_rigid_symbol_wrt_module("COND", NULL, 0)));
    SYM_WALK_STELLA_CASE = ((Symbol*)(intern_rigid_symbol_wrt_module("CASE", NULL, 0)));
    SYM_WALK_STELLA_TYPECASE = ((Symbol*)(intern_rigid_symbol_wrt_module("TYPECASE", NULL, 0)));
    SYM_WALK_STELLA_e = ((Symbol*)(intern_rigid_symbol_wrt_module("=", NULL, 0)));
    SYM_WALK_STELLA_i = ((Symbol*)(intern_rigid_symbol_wrt_module("+", NULL, 0)));
    SYM_WALK_STELLA__ = ((Symbol*)(intern_rigid_symbol_wrt_module("-", NULL, 0)));
    SYM_WALK_STELLA_o = ((Symbol*)(intern_rigid_symbol_wrt_module("*", NULL, 0)));
    SYM_WALK_STELLA_s = ((Symbol*)(intern_rigid_symbol_wrt_module("/", NULL, 0)));
    SYM_WALK_STELLA_BQUOTE = ((Symbol*)(intern_rigid_symbol_wrt_module("BQUOTE", NULL, 0)));
    SYM_WALK_STELLA_PRINT_STREAM = ((Symbol*)(intern_rigid_symbol_wrt_module("PRINT-STREAM", NULL, 0)));
    SYM_WALK_STELLA_PRINT_NATIVE_STREAM = ((Symbol*)(intern_rigid_symbol_wrt_module("PRINT-NATIVE-STREAM", NULL, 0)));
    SYM_WALK_STELLA_NEW = ((Symbol*)(intern_rigid_symbol_wrt_module("NEW", NULL, 0)));
    SYM_WALK_STELLA_ALLOCATE = ((Symbol*)(intern_rigid_symbol_wrt_module("ALLOCATE", NULL, 0)));
    SYM_WALK_STELLA_LOCAL_NEW = ((Symbol*)(intern_rigid_symbol_wrt_module("LOCAL-NEW", NULL, 0)));
    SYM_WALK_STELLA_MAKE = ((Symbol*)(intern_rigid_symbol_wrt_module("MAKE", NULL, 0)));
    SYM_WALK_STELLA_FUNCALL = ((Symbol*)(intern_rigid_symbol_wrt_module("FUNCALL", NULL, 0)));
    SYM_WALK_STELLA_SYS_CALL_FUNCTION_CODE = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-CALL-FUNCTION-CODE", NULL, 0)));
    SYM_WALK_STELLA_SYS_CALL_METHOD_CODE = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-CALL-METHOD-CODE", NULL, 0)));
    SYM_WALK_STELLA_THE_CODE = ((Symbol*)(intern_rigid_symbol_wrt_module("THE-CODE", NULL, 0)));
    SYM_WALK_STELLA_SIGNAL_EXCEPTION = ((Symbol*)(intern_rigid_symbol_wrt_module("SIGNAL-EXCEPTION", NULL, 0)));
    SYM_WALK_STELLA_EXCEPTION_CASE = ((Symbol*)(intern_rigid_symbol_wrt_module("EXCEPTION-CASE", NULL, 0)));
    SYM_WALK_STELLA_UNWIND_PROTECT = ((Symbol*)(intern_rigid_symbol_wrt_module("UNWIND-PROTECT", NULL, 0)));
    SYM_WALK_STELLA_SYS_UNWIND_PROTECT = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-UNWIND-PROTECT", NULL, 0)));
    SYM_WALK_STELLA_SYS_SLOT_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-SLOT-VALUE", NULL, 0)));
    SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-SLOT-VALUE-SETTER", NULL, 0)));
    SYM_WALK_STELLA_SYS_CALL_METHOD_SETTER = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-CALL-METHOD-SETTER", NULL, 0)));
    SYM_WALK_STELLA_SYS_NEW = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-NEW", NULL, 0)));
    SGT_WALK_STELLA_PROCESS_LOCK_OBJECT = ((Surrogate*)(intern_rigid_symbol_wrt_module("PROCESS-LOCK-OBJECT", NULL, 1)));
    KWD_WALK_PUBLICp = ((Keyword*)(intern_rigid_symbol_wrt_module("PUBLIC?", NULL, 2)));
    SYM_WALK_STELLA_TRUE = ((Symbol*)(intern_rigid_symbol_wrt_module("TRUE", NULL, 0)));
    KWD_WALK_SYMBOLS = ((Keyword*)(intern_rigid_symbol_wrt_module("SYMBOLS", NULL, 2)));
    SYM_WALK_STELLA_INTERN_RIGID_SYMBOL_WRT_MODULE = ((Symbol*)(intern_rigid_symbol_wrt_module("INTERN-RIGID-SYMBOL-WRT-MODULE", NULL, 0)));
    SYM_WALK_STELLA_GET_STELLA_MODULE = ((Symbol*)(intern_rigid_symbol_wrt_module("GET-STELLA-MODULE", NULL, 0)));
    SYM_WALK_STELLA_INTERN_SYMBOL_AT = ((Symbol*)(intern_rigid_symbol_wrt_module("INTERN-SYMBOL-AT", NULL, 0)));
    SYM_WALK_STELLA_INTERN_SURROGATE_AT = ((Symbol*)(intern_rigid_symbol_wrt_module("INTERN-SURROGATE-AT", NULL, 0)));
    SYM_WALK_STELLA_INTERN_KEYWORD_AT = ((Symbol*)(intern_rigid_symbol_wrt_module("INTERN-KEYWORD-AT", NULL, 0)));
    SYM_WALK_STELLA_FALSE = ((Symbol*)(intern_rigid_symbol_wrt_module("FALSE", NULL, 0)));
    SYM_WALK_STELLA_NIL = ((Symbol*)(intern_rigid_symbol_wrt_module("NIL", NULL, 0)));
    SYM_WALK_STELLA_NULL_INTEGER = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL-INTEGER", NULL, 0)));
    SYM_WALK_STELLA_NULL_FLOAT = ((Symbol*)(intern_rigid_symbol_wrt_module("NULL-FLOAT", NULL, 0)));
    KWD_WALK_INLINE_ARGUMENT = ((Keyword*)(intern_rigid_symbol_wrt_module("INLINE-ARGUMENT", NULL, 2)));
    SGT_WALK_STELLA_KEYWORD = ((Surrogate*)(intern_rigid_symbol_wrt_module("KEYWORD", NULL, 1)));
    KWD_WALK_PLACE_WHERE_VRLET_WAS = ((Keyword*)(intern_rigid_symbol_wrt_module("PLACE-WHERE-VRLET-WAS", NULL, 2)));
    KWD_WALK_PLACE_WHERE_LAST_EXPRESSION_WAS = ((Keyword*)(intern_rigid_symbol_wrt_module("PLACE-WHERE-LAST-EXPRESSION-WAS", NULL, 2)));
    KWD_WALK_DYNAMIC = ((Keyword*)(intern_rigid_symbol_wrt_module("DYNAMIC", NULL, 2)));
    SYM_WALK_STELLA_IT = ((Symbol*)(intern_rigid_symbol_wrt_module("IT", NULL, 0)));
    SYM_WALK_STELLA_ON = ((Symbol*)(intern_rigid_symbol_wrt_module("ON", NULL, 0)));
    SYM_WALK_STELLA_DYNAMIC_SLOTS = ((Symbol*)(intern_rigid_symbol_wrt_module("DYNAMIC-SLOTS", NULL, 0)));
    SYM_WALK_STELLA_KEY = ((Symbol*)(intern_rigid_symbol_wrt_module("KEY", NULL, 0)));
    SYM_WALK_STELLA_DO = ((Symbol*)(intern_rigid_symbol_wrt_module("DO", NULL, 0)));
    SYM_WALK_STELLA_WRAPPEDVALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("WRAPPEDVALUE", NULL, 0)));
    SYM_WALK_STELLA_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("VALUE", NULL, 0)));
    KWD_WALK_BIT = ((Keyword*)(intern_rigid_symbol_wrt_module("BIT", NULL, 2)));
  }
}

void help_startup_walk5() {
  {
    SYM_WALK_STELLA_DYNAMICSLOTS = ((Symbol*)(intern_rigid_symbol_wrt_module("DYNAMICSLOTS", NULL, 0)));
    SYM_WALK_STELLA_NEWVALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("NEWVALUE", NULL, 0)));
    SYM_WALK_STELLA_FOUNDMATCHINGENTRYp = ((Symbol*)(intern_rigid_symbol_wrt_module("FOUNDMATCHINGENTRY?", NULL, 0)));
    SYM_WALK_STELLA_THE_KV_LIST = ((Symbol*)(intern_rigid_symbol_wrt_module("THE-KV-LIST", NULL, 0)));
    SYM_WALK_STELLA_KV_CONS = ((Symbol*)(intern_rigid_symbol_wrt_module("KV-CONS", NULL, 0)));
    SYM_WALK_STELLA_OLDVALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("OLDVALUE", NULL, 0)));
    KWD_WALK_CONTEXT_SENSITIVEp = ((Keyword*)(intern_rigid_symbol_wrt_module("CONTEXT-SENSITIVE?", NULL, 2)));
    KWD_WALK_DONT_INHERITp = ((Keyword*)(intern_rigid_symbol_wrt_module("DONT-INHERIT?", NULL, 2)));
    SYM_WALK_STELLA_ACCESS_IN_CONTEXT = ((Symbol*)(intern_rigid_symbol_wrt_module("ACCESS-IN-CONTEXT", NULL, 0)));
    SYM_WALK_STELLA_HOME_CONTEXT = ((Symbol*)(intern_rigid_symbol_wrt_module("HOME-CONTEXT", NULL, 0)));
    KWD_WALK_COPY_TO_CHILDRENp = ((Keyword*)(intern_rigid_symbol_wrt_module("COPY-TO-CHILDREN?", NULL, 2)));
    SYM_WALK_STELLA_UPDATE_IN_CONTEXT = ((Symbol*)(intern_rigid_symbol_wrt_module("UPDATE-IN-CONTEXT", NULL, 0)));
    SYM_WALK_STELLA_CS_VALUEp = ((Symbol*)(intern_rigid_symbol_wrt_module("CS-VALUE?", NULL, 0)));
    SGT_WALK_STELLA_ARRAY = ((Surrogate*)(intern_rigid_symbol_wrt_module("ARRAY", NULL, 1)));
    SYM_WALK_STELLA_ANY_VALUE = ((Symbol*)(intern_rigid_symbol_wrt_module("ANY-VALUE", NULL, 0)));
    SYM_WALK_STELLA_VARIABLE_DECLARATION = ((Symbol*)(intern_rigid_symbol_wrt_module("VARIABLE-DECLARATION", NULL, 0)));
    SYM_WALK_STELLA_ILLEGAL_VARIABLE = ((Symbol*)(intern_rigid_symbol_wrt_module("ILLEGAL-VARIABLE", NULL, 0)));
    KWD_WALK_UNBIND_WITH_DESTRUCTORS = ((Keyword*)(intern_rigid_symbol_wrt_module("UNBIND-WITH-DESTRUCTORS", NULL, 2)));
    KWD_WALK_UNWIND_PROTECT = ((Keyword*)(intern_rigid_symbol_wrt_module("UNWIND-PROTECT", NULL, 2)));
    SGT_WALK_STELLA_METHOD_SLOT = ((Surrogate*)(intern_rigid_symbol_wrt_module("METHOD-SLOT", NULL, 1)));
    SYM_WALK_STELLA_OTHERWISE = ((Symbol*)(intern_rigid_symbol_wrt_module("OTHERWISE", NULL, 0)));
    SYM_WALK_STELLA_SYMBOL_ID = ((Symbol*)(intern_rigid_symbol_wrt_module("SYMBOL-ID", NULL, 0)));
    SYM_WALK_STELLA_GENERALIZED_SYMBOL = ((Symbol*)(intern_rigid_symbol_wrt_module("GENERALIZED-SYMBOL", NULL, 0)));
    SYM_WALK_STELLA_BOOLEANp = ((Symbol*)(intern_rigid_symbol_wrt_module("BOOLEAN?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_BOOLEANp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-BOOLEAN?", NULL, 0)));
    SYM_WALK_STELLA_INTEGERp = ((Symbol*)(intern_rigid_symbol_wrt_module("INTEGER?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_INTEGERp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-INTEGER?", NULL, 0)));
    SYM_WALK_STELLA_FLOATp = ((Symbol*)(intern_rigid_symbol_wrt_module("FLOAT?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_FLOATp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-FLOAT?", NULL, 0)));
    SYM_WALK_STELLA_STRINGp = ((Symbol*)(intern_rigid_symbol_wrt_module("STRING?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_STRINGp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-STRING?", NULL, 0)));
    SYM_WALK_STELLA_CHARACTERp = ((Symbol*)(intern_rigid_symbol_wrt_module("CHARACTER?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_CHARACTERp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-CHARACTER?", NULL, 0)));
    SYM_WALK_STELLA_WRAPPERp = ((Symbol*)(intern_rigid_symbol_wrt_module("WRAPPER?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_WRAPPERp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-WRAPPER?", NULL, 0)));
    SGT_WALK_STELLA_VERBATIM_STRING_WRAPPER = ((Surrogate*)(intern_rigid_symbol_wrt_module("VERBATIM-STRING-WRAPPER", NULL, 1)));
    SYM_WALK_STELLA_VERBATIM_STRINGp = ((Symbol*)(intern_rigid_symbol_wrt_module("VERBATIM-STRING?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_VERBATIM_STRINGp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-VERBATIM-STRING?", NULL, 0)));
    SYM_WALK_STELLA_SURROGATEp = ((Symbol*)(intern_rigid_symbol_wrt_module("SURROGATE?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_SURROGATEp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-SURROGATE?", NULL, 0)));
    SGT_WALK_STELLA_TYPE = ((Surrogate*)(intern_rigid_symbol_wrt_module("TYPE", NULL, 1)));
    SYM_WALK_STELLA_TYPEp = ((Symbol*)(intern_rigid_symbol_wrt_module("TYPE?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_TYPEp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-TYPE?", NULL, 0)));
    SYM_WALK_STELLA_SYMBOLp = ((Symbol*)(intern_rigid_symbol_wrt_module("SYMBOL?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_SYMBOLp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-SYMBOL?", NULL, 0)));
    SGT_WALK_STELLA_TRANSIENT_SYMBOL = ((Surrogate*)(intern_rigid_symbol_wrt_module("TRANSIENT-SYMBOL", NULL, 1)));
    SYM_WALK_STELLA_TRANSIENT_SYMBOLp = ((Symbol*)(intern_rigid_symbol_wrt_module("TRANSIENT-SYMBOL?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_TRANSIENT_SYMBOLp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-TRANSIENT-SYMBOL?", NULL, 0)));
    SYM_WALK_STELLA_KEYWORDp = ((Symbol*)(intern_rigid_symbol_wrt_module("KEYWORD?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_KEYWORDp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-KEYWORD?", NULL, 0)));
    SYM_WALK_STELLA_CONSp = ((Symbol*)(intern_rigid_symbol_wrt_module("CONS?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_CONSp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-CONS?", NULL, 0)));
    SYM_WALK_STELLA_STELLA_CLASSp = ((Symbol*)(intern_rigid_symbol_wrt_module("STELLA-CLASS?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_CLASSp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-CLASS?", NULL, 0)));
    SGT_WALK_STELLA_STORAGE_SLOT = ((Surrogate*)(intern_rigid_symbol_wrt_module("STORAGE-SLOT", NULL, 1)));
    SYM_WALK_STELLA_STORAGE_SLOTp = ((Symbol*)(intern_rigid_symbol_wrt_module("STORAGE-SLOT?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_STORAGE_SLOTp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-STORAGE-SLOT?", NULL, 0)));
    SYM_WALK_STELLA_METHOD_SLOTp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-SLOT?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_METHOD_SLOTp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-METHOD-SLOT?", NULL, 0)));
    SGT_WALK_STELLA_ANCHORED_TYPE_SPECIFIER = ((Surrogate*)(intern_rigid_symbol_wrt_module("ANCHORED-TYPE-SPECIFIER", NULL, 1)));
  }
}

void help_startup_walk6() {
  {
    SYM_WALK_STELLA_ANCHORED_TYPE_SPECIFIERp = ((Symbol*)(intern_rigid_symbol_wrt_module("ANCHORED-TYPE-SPECIFIER?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_ANCHORED_TYPE_SPECIFIERp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-ANCHORED-TYPE-SPECIFIER?", NULL, 0)));
    SGT_WALK_STELLA_PARAMETRIC_TYPE_SPECIFIER = ((Surrogate*)(intern_rigid_symbol_wrt_module("PARAMETRIC-TYPE-SPECIFIER", NULL, 1)));
    SYM_WALK_STELLA_PARAMETRIC_TYPE_SPECIFIERp = ((Symbol*)(intern_rigid_symbol_wrt_module("PARAMETRIC-TYPE-SPECIFIER?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OF_PARAMETRIC_TYPE_SPECIFIERp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF-PARAMETRIC-TYPE-SPECIFIER?", NULL, 0)));
    SYM_WALK_STELLA_ISAp = ((Symbol*)(intern_rigid_symbol_wrt_module("ISA?", NULL, 0)));
    SYM_WALK_STELLA_SUBTYPE_OFp = ((Symbol*)(intern_rigid_symbol_wrt_module("SUBTYPE-OF?", NULL, 0)));
    SGT_WALK_STELLA_GENERALIZED_SYMBOL = ((Surrogate*)(intern_rigid_symbol_wrt_module("GENERALIZED-SYMBOL", NULL, 1)));
    SYM_WALK_STELLA_STRING_EQLp = ((Symbol*)(intern_rigid_symbol_wrt_module("STRING-EQL?", NULL, 0)));
    SYM_WALK_STELLA_SAFE_PRIMARY_TYPE = ((Symbol*)(intern_rigid_symbol_wrt_module("SAFE-PRIMARY-TYPE", NULL, 0)));
    SGT_WALK_STELLA_STANDARD_OBJECT = ((Surrogate*)(intern_rigid_symbol_wrt_module("STANDARD-OBJECT", NULL, 1)));
    SGT_WALK_STELLA_LITERAL = ((Surrogate*)(intern_rigid_symbol_wrt_module("LITERAL", NULL, 1)));
    SYM_WALK_STELLA_EQL_TO_BOOLEANp = ((Symbol*)(intern_rigid_symbol_wrt_module("EQL-TO-BOOLEAN?", NULL, 0)));
    SYM_WALK_STELLA_EQL_TO_INTEGERp = ((Symbol*)(intern_rigid_symbol_wrt_module("EQL-TO-INTEGER?", NULL, 0)));
    SYM_WALK_STELLA_EQL_TO_FLOATp = ((Symbol*)(intern_rigid_symbol_wrt_module("EQL-TO-FLOAT?", NULL, 0)));
    SYM_WALK_STELLA_EQL_TO_STRINGp = ((Symbol*)(intern_rigid_symbol_wrt_module("EQL-TO-STRING?", NULL, 0)));
    SYM_WALK_STELLA_GET_QUOTED_TREE = ((Symbol*)(intern_rigid_symbol_wrt_module("GET-QUOTED-TREE", NULL, 0)));
    SYM_WALK_STELLA_METHOD_BODY_ARGUMENTp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-BODY-ARGUMENT?", NULL, 0)));
    KWD_WALK_WARN = ((Keyword*)(intern_rigid_symbol_wrt_module("WARN", NULL, 2)));
    KWD_WALK_ERROR = ((Keyword*)(intern_rigid_symbol_wrt_module("ERROR", NULL, 2)));
    KWD_WALK_CONTINUABLE_ERROR = ((Keyword*)(intern_rigid_symbol_wrt_module("CONTINUABLE-ERROR", NULL, 2)));
    SYM_WALK_STELLA_CONCATENATE = ((Symbol*)(intern_rigid_symbol_wrt_module("CONCATENATE", NULL, 0)));
    SYM_WALK_STELLA_WRAP_BOOLEAN = ((Symbol*)(intern_rigid_symbol_wrt_module("WRAP-BOOLEAN", NULL, 0)));
    SYM_WALK_STELLA_CONS = ((Symbol*)(intern_rigid_symbol_wrt_module("CONS", NULL, 0)));
    SYM_WALK_STELLA_CONS_LIST = ((Symbol*)(intern_rigid_symbol_wrt_module("CONS-LIST", NULL, 0)));
    SYM_WALK_STELLA_ALLOCATE_ITERATOR = ((Symbol*)(intern_rigid_symbol_wrt_module("ALLOCATE-ITERATOR", NULL, 0)));
    SGT_WALK_STELLA_ARGUMENT_LIST_ITERATOR = ((Surrogate*)(intern_rigid_symbol_wrt_module("ARGUMENT-LIST-ITERATOR", NULL, 1)));
    SYM_WALK_STELLA_LENGTH = ((Symbol*)(intern_rigid_symbol_wrt_module("LENGTH", NULL, 0)));
    SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-VARIABLE-ARGUMENTS?", NULL, 0)));
    SYM_WALK_STELLA_GET_SYM = ((Symbol*)(intern_rigid_symbol_wrt_module("GET-SYM", NULL, 0)));
    SYM_WALK_STELLA_GET_SGT = ((Symbol*)(intern_rigid_symbol_wrt_module("GET-SGT", NULL, 0)));
    SYM_WALK_STELLA_GET_KWD = ((Symbol*)(intern_rigid_symbol_wrt_module("GET-KWD", NULL, 0)));
    SGT_WALK_STELLA_ARGUMENT_LIST = ((Surrogate*)(intern_rigid_symbol_wrt_module("ARGUMENT-LIST", NULL, 1)));
    SYM_WALK_STELLA_CACHED_INLINABLE_METHOD_BODY = ((Symbol*)(intern_rigid_symbol_wrt_module("CACHED-INLINABLE-METHOD-BODY", NULL, 0)));
    KWD_WALK_INLINE_REFERENCES = ((Keyword*)(intern_rigid_symbol_wrt_module("INLINE-REFERENCES", NULL, 2)));
    SYM_WALK_STELLA_INLINE_CALL = ((Symbol*)(intern_rigid_symbol_wrt_module("INLINE-CALL", NULL, 0)));
    SYM_WALK_STELLA_METHOD_INHERITS_THROUGH = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-INHERITS-THROUGH", NULL, 0)));
    KWD_WALK_FUNCTION = ((Keyword*)(intern_rigid_symbol_wrt_module("FUNCTION", NULL, 2)));
    SYM_WALK_STELLA_FORWARD_DECLARATIONp = ((Symbol*)(intern_rigid_symbol_wrt_module("FORWARD-DECLARATION?", NULL, 0)));
    SYM_WALK_STELLA_METHOD_NATIVEp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-NATIVE?", NULL, 0)));
    SYM_WALK_STELLA_METHOD_COMMANDp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-COMMAND?", NULL, 0)));
    SYM_WALK_STELLA_MAIN = ((Symbol*)(intern_rigid_symbol_wrt_module("MAIN", NULL, 0)));
    KWD_WALK_METHODS = ((Keyword*)(intern_rigid_symbol_wrt_module("METHODS", NULL, 2)));
    KWD_WALK_CLASS = ((Keyword*)(intern_rigid_symbol_wrt_module("CLASS", NULL, 2)));
    KWD_WALK_EMBEDDED = ((Keyword*)(intern_rigid_symbol_wrt_module("EMBEDDED", NULL, 2)));
    SYM_WALK_STELLA_CLASS_CONSTRUCTOR_CODE = ((Symbol*)(intern_rigid_symbol_wrt_module("CLASS-CONSTRUCTOR-CODE", NULL, 0)));
    SYM_WALK_STELLA_CLASS_AUXILIARY_METHODS = ((Symbol*)(intern_rigid_symbol_wrt_module("CLASS-AUXILIARY-METHODS", NULL, 0)));
    SYM_WALK_STELLA_AUXILIARY_METHOD_REASON = ((Symbol*)(intern_rigid_symbol_wrt_module("AUXILIARY-METHOD-REASON", NULL, 0)));
    KWD_WALK_FIRST_CLASS_WITH_TYPECASE = ((Keyword*)(intern_rigid_symbol_wrt_module("FIRST-CLASS-WITH-TYPECASE", NULL, 2)));
    KWD_WALK_AUXILIARYp = ((Keyword*)(intern_rigid_symbol_wrt_module("AUXILIARY?", NULL, 2)));
    SYM_WALK_STELLA_PRINT_FORM = ((Symbol*)(intern_rigid_symbol_wrt_module("PRINT-FORM", NULL, 0)));
    SYM_WALK_STELLA_PRINT_OBJECT = ((Symbol*)(intern_rigid_symbol_wrt_module("PRINT-OBJECT", NULL, 0)));
    SYM_WALK_STELLA_SELF = ((Symbol*)(intern_rigid_symbol_wrt_module("SELF", NULL, 0)));
    SYM_WALK_STELLA_STREAM = ((Symbol*)(intern_rigid_symbol_wrt_module("STREAM", NULL, 0)));
    SYM_WALK_STELLA_NATIVE_OUTPUT_STREAM = ((Symbol*)(intern_rigid_symbol_wrt_module("NATIVE-OUTPUT-STREAM", NULL, 0)));
    SYM_WALK_STELLA_STRINGIFY = ((Symbol*)(intern_rigid_symbol_wrt_module("STRINGIFY", NULL, 0)));
    KWD_WALK_CLASSES = ((Keyword*)(intern_rigid_symbol_wrt_module("CLASSES", NULL, 2)));
    KWD_WALK_FINALIZE_CLASSES = ((Keyword*)(intern_rigid_symbol_wrt_module("FINALIZE-CLASSES", NULL, 2)));
    SYM_WALK_STELLA_FINALIZE_CLASSES = ((Symbol*)(intern_rigid_symbol_wrt_module("FINALIZE-CLASSES", NULL, 0)));
    KWD_WALK_FINALIZE_METHODS = ((Keyword*)(intern_rigid_symbol_wrt_module("FINALIZE-METHODS", NULL, 2)));
  }
}

void help_startup_walk7() {
  {
    SYM_WALK_STELLA_FINALIZE_SLOTS = ((Symbol*)(intern_rigid_symbol_wrt_module("FINALIZE-SLOTS", NULL, 0)));
    SYM_WALK_STELLA_CLEANUP_UNFINALIZED_CLASSES = ((Symbol*)(intern_rigid_symbol_wrt_module("CLEANUP-UNFINALIZED-CLASSES", NULL, 0)));
    SYM_WALK_STELLA_DEFINE_EXTERNAL_SLOT_FROM_STRINGIFIED_SOURCE = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFINE-EXTERNAL-SLOT-FROM-STRINGIFIED-SOURCE", NULL, 0)));
    SYM_WALK_STELLA_OBJECT = ((Symbol*)(intern_rigid_symbol_wrt_module("OBJECT", NULL, 0)));
    SYM_WALK_STELLA_METHOD_LISP_MACROp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-LISP-MACRO?", NULL, 0)));
    SYM_WALK_STELLA_STORE_MACRO = ((Symbol*)(intern_rigid_symbol_wrt_module("STORE-MACRO", NULL, 0)));
    KWD_WALK_UNBOUND_SPECIAL_VARIABLE = ((Keyword*)(intern_rigid_symbol_wrt_module("UNBOUND-SPECIAL-VARIABLE", NULL, 2)));
    KWD_WALK_GLOBALS = ((Keyword*)(intern_rigid_symbol_wrt_module("GLOBALS", NULL, 2)));
    SYM_WALK_STELLA_DEFINE_STELLA_GLOBAL_VARIABLE_FROM_STRINGIFIED_SOURCE = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE", NULL, 0)));
    SYM_WALK_STELLA_DEFINE_STELLA_TYPE_FROM_STRINGIFIED_SOURCE = ((Symbol*)(intern_rigid_symbol_wrt_module("DEFINE-STELLA-TYPE-FROM-STRINGIFIED-SOURCE", NULL, 0)));
    KWD_WALK_FINAL = ((Keyword*)(intern_rigid_symbol_wrt_module("FINAL", NULL, 2)));
    SYM_WALK_STELLA_CURRENT_STARTUP_TIME_PHASEp = ((Symbol*)(intern_rigid_symbol_wrt_module("CURRENT-STARTUP-TIME-PHASE?", NULL, 0)));
    KWD_WALK_MODULES = ((Keyword*)(intern_rigid_symbol_wrt_module("MODULES", NULL, 2)));
    SYM_WALK_STELLA_WITHIN_MODULE = ((Symbol*)(intern_rigid_symbol_wrt_module("WITHIN-MODULE", NULL, 0)));
    SYM_WALK_STELLA_oSTELLA_MODULEo = ((Symbol*)(intern_rigid_symbol_wrt_module("*STELLA-MODULE*", NULL, 0)));
    SYM_WALK_STELLA_oSTARTUP_TIME_PHASEo = ((Symbol*)(intern_rigid_symbol_wrt_module("*STARTUP-TIME-PHASE*", NULL, 0)));
    SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-STARTUP-CLASSNAME", NULL, 0)));
    SYM_WALK_STELLA_FUNCTION = ((Symbol*)(intern_rigid_symbol_wrt_module("FUNCTION", NULL, 0)));
    SYM_WALK_STELLA_LOOKUP_FUNCTION = ((Symbol*)(intern_rigid_symbol_wrt_module("LOOKUP-FUNCTION", NULL, 0)));
    SYM_WALK_STELLA_SYS_SIGNAL = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-SIGNAL", NULL, 0)));
    SGT_WALK_STELLA_STELLA_ROOT_EXCEPTION = ((Surrogate*)(intern_rigid_symbol_wrt_module("STELLA-ROOT-EXCEPTION", NULL, 1)));
    SYM_WALK_STELLA_SYS_HANDLER_CASE = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-HANDLER-CASE", NULL, 0)));
    SGT_WALK_STELLA_NATIVE_EXCEPTION = ((Surrogate*)(intern_rigid_symbol_wrt_module("NATIVE-EXCEPTION", NULL, 1)));
    SYM_WALK_STELLA_SYS_HANDLE_EXCEPTION = ((Symbol*)(intern_rigid_symbol_wrt_module("SYS-HANDLE-EXCEPTION", NULL, 0)));
    KWD_WALK_OTHERWISE = ((Keyword*)(intern_rigid_symbol_wrt_module("OTHERWISE", NULL, 2)));
    SYM_WALK_STELLA_INITIAL_ELEMENT = ((Symbol*)(intern_rigid_symbol_wrt_module("INITIAL-ELEMENT", NULL, 0)));
    SYM_WALK_STELLA_SIZE = ((Symbol*)(intern_rigid_symbol_wrt_module("SIZE", NULL, 0)));
    SYM_WALK_STELLA_STORAGE_SLOT = ((Symbol*)(intern_rigid_symbol_wrt_module("STORAGE-SLOT", NULL, 0)));
    SGT_WALK_STELLA_CODE = ((Surrogate*)(intern_rigid_symbol_wrt_module("CODE", NULL, 1)));
    KWD_WALK_METHOD = ((Keyword*)(intern_rigid_symbol_wrt_module("METHOD", NULL, 2)));
    SYM_WALK_STELLA_METHOD_CODE = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-CODE", NULL, 0)));
    SGT_WALK_STELLA_COMPOUND_TYPE_SPECIFIER = ((Surrogate*)(intern_rigid_symbol_wrt_module("COMPOUND-TYPE-SPECIFIER", NULL, 1)));
    SYM_WALK_STELLA_DOCUMENTATION = ((Symbol*)(intern_rigid_symbol_wrt_module("DOCUMENTATION", NULL, 0)));
    SYM_WALK_STELLA_EXPRESSION = ((Symbol*)(intern_rigid_symbol_wrt_module("EXPRESSION", NULL, 0)));
    SYM_WALK_STELLA_IGNORE = ((Symbol*)(intern_rigid_symbol_wrt_module("IGNORE", NULL, 0)));
    SYM_WALK_STELLA_oIGNORETRANSLATIONERRORSpo = ((Symbol*)(intern_rigid_symbol_wrt_module("*IGNORETRANSLATIONERRORS?*", NULL, 0)));
    SYM_WALK_STELLA_CL_INCREMENTALLY_TRANSLATE = ((Symbol*)(intern_rigid_symbol_wrt_module("CL-INCREMENTALLY-TRANSLATE", NULL, 0)));
    SYM_WALK_STELLA_REGISTER_NATIVE_NAME = ((Symbol*)(intern_rigid_symbol_wrt_module("REGISTER-NATIVE-NAME", NULL, 0)));
    SYM_WALK_STELLA_REST = ((Symbol*)(intern_rigid_symbol_wrt_module("REST", NULL, 0)));
    SYM_WALK_STELLA_NTH_REST = ((Symbol*)(intern_rigid_symbol_wrt_module("NTH-REST", NULL, 0)));
    SYM_WALK_STELLA_NTH = ((Symbol*)(intern_rigid_symbol_wrt_module("NTH", NULL, 0)));
    SYM_WALK_STELLA_ARGUMENTS = ((Symbol*)(intern_rigid_symbol_wrt_module("ARGUMENTS", NULL, 0)));
    SYM_WALK_STELLA_RESULT = ((Symbol*)(intern_rigid_symbol_wrt_module("RESULT", NULL, 0)));
    SYM_WALK_STELLA_METHOD_EVALUATE_ARGUMENTSp = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-EVALUATE-ARGUMENTS?", NULL, 0)));
    SYM_WALK_STELLA_EVALUATOR_WRAPPER_CODE = ((Symbol*)(intern_rigid_symbol_wrt_module("EVALUATOR-WRAPPER-CODE", NULL, 0)));
    SYM_WALK_STELLA_STARTUP_WALK = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-WALK", NULL, 0)));
  }
}

void help_startup_walk8() {
  {
    oAVAILABLE_ENVIRONMENT_FEATURESo = ((List*)(list(9, KWD_WALK_WARN_ABOUT_UNDEFINED_METHODS, KWD_WALK_SUPPRESS_WARNINGS, KWD_WALK_WARN_ABOUT_MISSING_METHODS, KWD_WALK_USE_HARDCODED_SYMBOLS, KWD_WALK_USE_COMMON_LISP_STRUCTS, KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS, KWD_WALK_USE_COMMON_LISP_CONSES, KWD_WALK_USE_CPP_GARBAGE_COLLECTOR, KWD_WALK_TRANSLATE_WITH_COPYRIGHT_HEADER)));
    oCURRENT_ENVIRONMENT_FEATURESo = new_list();
    oDEFAULT_ENVIRONMENT_FEATURESo = ((List*)(list(4, KWD_WALK_WARN_ABOUT_UNDEFINED_METHODS, KWD_WALK_WARN_ABOUT_MISSING_METHODS, KWD_WALK_USE_CPP_GARBAGE_COLLECTOR, KWD_WALK_USE_COMMON_LISP_CONSES)));
    reset_environment_features();
    oTRANSLATOROUTPUTLANGUAGEo.set(KWD_WALK_COMMON_LISP);
    oTARGETTYPEo.set(SGT_WALK_STELLA_VOID);
    oWRAPPED_TYPE_TABLEo = listO(10, listO(3, SGT_WALK_STELLA_INTEGER_WRAPPER, SGT_WALK_STELLA_INTEGER, NIL), listO(3, SGT_WALK_STELLA_FLOAT_WRAPPER, SGT_WALK_STELLA_FLOAT, NIL), listO(3, SGT_WALK_STELLA_NUMBER_WRAPPER, SGT_WALK_STELLA_NUMBER, NIL), listO(3, SGT_WALK_STELLA_STRING_WRAPPER, SGT_WALK_STELLA_STRING, NIL), listO(3, SGT_WALK_STELLA_MUTABLE_STRING_WRAPPER, SGT_WALK_STELLA_MUTABLE_STRING, NIL), listO(3, SGT_WALK_STELLA_CHARACTER_WRAPPER, SGT_WALK_STELLA_CHARACTER, NIL), listO(3, SGT_WALK_STELLA_BOOLEAN_WRAPPER, SGT_WALK_STELLA_BOOLEAN, NIL), listO(3, SGT_WALK_STELLA_FUNCTION_CODE_WRAPPER, SGT_WALK_STELLA_FUNCTION_CODE, NIL), listO(3, SGT_WALK_STELLA_METHOD_CODE_WRAPPER, SGT_WALK_STELLA_METHOD_CODE, NIL), NIL);
    oCOERSION_TABLEo = listO(39, listO(4, SGT_WALK_STELLA_BOOLEAN, SGT_WALK_STELLA_BOOLEAN_WRAPPER, SYM_WALK_STELLA_INLINE_WRAP_BOOLEAN, NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_INTEGER_WRAPPER, SYM_WALK_STELLA_WRAP_LITERAL, NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_BOOLEAN_WRAPPER, SYM_WALK_STELLA_INTEGER_TO_BOOLEAN_WRAPPER, NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_BOOLEAN, SYM_WALK_STELLA_INTEGER_TO_BOOLEAN, NIL), listO(4, SGT_WALK_STELLA_FLOAT, SGT_WALK_STELLA_FLOAT_WRAPPER, SYM_WALK_STELLA_WRAP_LITERAL, NIL), listO(4, SGT_WALK_STELLA_MUTABLE_STRING, SGT_WALK_STELLA_STRING, SYM_WALK_STELLA_MUTABLE_STRING_TO_STRING, NIL), listO(4, SGT_WALK_STELLA_MUTABLE_STRING, SGT_WALK_STELLA_MUTABLE_STRING_WRAPPER, SYM_WALK_STELLA_WRAP_LITERAL, NIL), listO(4, SGT_WALK_STELLA_STRING, SGT_WALK_STELLA_STRING_WRAPPER, SYM_WALK_STELLA_WRAP_LITERAL, NIL), listO(4, SGT_WALK_STELLA_STRING, SGT_WALK_STELLA_MUTABLE_STRING, SYM_WALK_STELLA_STRING_TO_MUTABLE_STRING, NIL), listO(4, SGT_WALK_STELLA_STRING, SGT_WALK_STELLA_SYMBOL, SYM_WALK_STELLA_INTERN_SYMBOL, NIL), listO(4, SGT_WALK_STELLA_CHARACTER, SGT_WALK_STELLA_CHARACTER_WRAPPER, SYM_WALK_STELLA_WRAP_LITERAL, NIL), listO(4, SGT_WALK_STELLA_CHARACTER, SGT_WALK_STELLA_STRING, SYM_WALK_STELLA_CHARACTER_TO_STRING, NIL), listO(4, SGT_WALK_STELLA_FUNCTION_CODE, SGT_WALK_STELLA_FUNCTION_CODE_WRAPPER, SYM_WALK_STELLA_WRAP_LITERAL, NIL), listO(4, SGT_WALK_STELLA_METHOD_CODE, SGT_WALK_STELLA_METHOD_CODE_WRAPPER, SYM_WALK_STELLA_WRAP_LITERAL, NIL), listO(4, SGT_WALK_STELLA_SYMBOL, SGT_WALK_STELLA_STRING, SYM_WALK_STELLA_SYMBOL_NAME, NIL), listO(4, SGT_WALK_STELLA_BOOLEAN_WRAPPER, SGT_WALK_STELLA_BOOLEAN, SYM_WALK_STELLA_INLINE_UNWRAP_BOOLEAN, NIL), listO(4, SGT_WALK_STELLA_INTEGER_WRAPPER, SGT_WALK_STELLA_INTEGER, SYM_WALK_STELLA_WRAPPER_VALUE, NIL), listO(4, SGT_WALK_STELLA_FLOAT_WRAPPER, SGT_WALK_STELLA_FLOAT, SYM_WALK_STELLA_WRAPPER_VALUE, NIL), listO(4, SGT_WALK_STELLA_NUMBER_WRAPPER, SGT_WALK_STELLA_FLOAT, SYM_WALK_STELLA_NUMBER_WRAPPER_TO_FLOAT, NIL), listO(4, SGT_WALK_STELLA_STRING_WRAPPER, SGT_WALK_STELLA_STRING, SYM_WALK_STELLA_WRAPPER_VALUE, NIL), listO(4, SGT_WALK_STELLA_MUTABLE_STRING_WRAPPER, SGT_WALK_STELLA_MUTABLE_STRING, SYM_WALK_STELLA_WRAPPER_VALUE, NIL), listO(4, SGT_WALK_STELLA_CHARACTER_WRAPPER, SGT_WALK_STELLA_CHARACTER, SYM_WALK_STELLA_WRAPPER_VALUE, NIL), listO(4, SGT_WALK_STELLA_FUNCTION_CODE_WRAPPER, SGT_WALK_STELLA_FUNCTION_CODE, SYM_WALK_STELLA_WRAPPER_VALUE, NIL), listO(4, SGT_WALK_STELLA_METHOD_CODE_WRAPPER, SGT_WALK_STELLA_METHOD_CODE, SYM_WALK_STELLA_WRAPPER_VALUE, NIL), listO(4, SGT_WALK_STELLA_SURROGATE, SGT_WALK_STELLA_CLASS, SYM_WALK_STELLA_SURROGATE_VALUE, NIL), listO(4, SGT_WALK_STELLA_SURROGATE, SGT_WALK_STELLA_MODULE, SYM_WALK_STELLA_SURROGATE_VALUE, NIL), listO(4, SGT_WALK_STELLA_INPUT_STREAM, SGT_WALK_STELLA_NATIVE_INPUT_STREAM, SYM_WALK_STELLA_NATIVE_STREAM, NIL), listO(4, SGT_WALK_STELLA_OUTPUT_STREAM, SGT_WALK_STELLA_NATIVE_OUTPUT_STREAM, SYM_WALK_STELLA_NATIVE_STREAM, NIL), listO(4, SGT_WALK_STELLA_NUMBER, SGT_WALK_STELLA_INTEGER, listO(4, SYM_WALK_STELLA_CAST, SYM_WALK_STELLA_lXg, SGT_WALK_STELLA_INTEGER, NIL), NIL), listO(4, SGT_WALK_STELLA_NUMBER, SGT_WALK_STELLA_FLOAT, listO(4, SYM_WALK_STELLA_CAST, SYM_WALK_STELLA_lXg, SGT_WALK_STELLA_FLOAT, NIL), NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_FLOAT, listO(4, SYM_WALK_STELLA_CAST, SYM_WALK_STELLA_lXg, SGT_WALK_STELLA_FLOAT, NIL), NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_SINGLE_FLOAT, listO(4, SYM_WALK_STELLA_CAST, SYM_WALK_STELLA_lXg, SGT_WALK_STELLA_SINGLE_FLOAT, NIL), NIL), listO(4, SGT_WALK_STELLA_FLOAT, SGT_WALK_STELLA_SINGLE_FLOAT, SYM_WALK_STELLA_IDENTITY, NIL), listO(4, SGT_WALK_STELLA_FLOAT, SGT_WALK_STELLA_DOUBLE_FLOAT, SYM_WALK_STELLA_IDENTITY, NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_SHORT_INTEGER, SYM_WALK_STELLA_IDENTITY, NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_LONG_INTEGER, SYM_WALK_STELLA_IDENTITY, NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_UNSIGNED_SHORT_INTEGER, SYM_WALK_STELLA_IDENTITY, NIL), listO(4, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_UNSIGNED_LONG_INTEGER, SYM_WALK_STELLA_IDENTITY, NIL), NIL);
    oSYMBOL_REGISTRYo = new_hash_table();
    oSYMBOL_SETo = new_list();
    oSPECIALVARIABLESTACKo.set(new_key_value_list());
    oTYPE_PREDICATE_TABLEo = listO(24, listO(4, SGT_WALK_STELLA_BOOLEAN, SYM_WALK_STELLA_BOOLEANp, SYM_WALK_STELLA_SUBTYPE_OF_BOOLEANp, NIL), listO(4, SGT_WALK_STELLA_INTEGER, SYM_WALK_STELLA_INTEGERp, SYM_WALK_STELLA_SUBTYPE_OF_INTEGERp, NIL), listO(4, SGT_WALK_STELLA_FLOAT, SYM_WALK_STELLA_FLOATp, SYM_WALK_STELLA_SUBTYPE_OF_FLOATp, NIL), listO(4, SGT_WALK_STELLA_STRING, SYM_WALK_STELLA_STRINGp, SYM_WALK_STELLA_SUBTYPE_OF_STRINGp, NIL), listO(4, SGT_WALK_STELLA_CHARACTER, SYM_WALK_STELLA_CHARACTERp, SYM_WALK_STELLA_SUBTYPE_OF_CHARACTERp, NIL), listO(4, SGT_WALK_STELLA_WRAPPER, SYM_WALK_STELLA_WRAPPERp, SYM_WALK_STELLA_SUBTYPE_OF_WRAPPERp, NIL), listO(4, SGT_WALK_STELLA_BOOLEAN_WRAPPER, SYM_WALK_STELLA_BOOLEANp, SYM_WALK_STELLA_SUBTYPE_OF_BOOLEANp, NIL), listO(4, SGT_WALK_STELLA_INTEGER_WRAPPER, SYM_WALK_STELLA_INTEGERp, SYM_WALK_STELLA_SUBTYPE_OF_INTEGERp, NIL), listO(4, SGT_WALK_STELLA_FLOAT_WRAPPER, SYM_WALK_STELLA_FLOATp, SYM_WALK_STELLA_SUBTYPE_OF_FLOATp, NIL), listO(4, SGT_WALK_STELLA_STRING_WRAPPER, SYM_WALK_STELLA_STRINGp, SYM_WALK_STELLA_SUBTYPE_OF_STRINGp, NIL), listO(4, SGT_WALK_STELLA_CHARACTER_WRAPPER, SYM_WALK_STELLA_CHARACTERp, SYM_WALK_STELLA_SUBTYPE_OF_CHARACTERp, NIL), listO(4, SGT_WALK_STELLA_VERBATIM_STRING_WRAPPER, SYM_WALK_STELLA_VERBATIM_STRINGp, SYM_WALK_STELLA_SUBTYPE_OF_VERBATIM_STRINGp, NIL), listO(4, SGT_WALK_STELLA_SURROGATE, SYM_WALK_STELLA_SURROGATEp, SYM_WALK_STELLA_SUBTYPE_OF_SURROGATEp, NIL), listO(4, SGT_WALK_STELLA_TYPE, SYM_WALK_STELLA_TYPEp, SYM_WALK_STELLA_SUBTYPE_OF_TYPEp, NIL), listO(4, SGT_WALK_STELLA_SYMBOL, SYM_WALK_STELLA_SYMBOLp, SYM_WALK_STELLA_SUBTYPE_OF_SYMBOLp, NIL), listO(4, SGT_WALK_STELLA_TRANSIENT_SYMBOL, SYM_WALK_STELLA_TRANSIENT_SYMBOLp, SYM_WALK_STELLA_SUBTYPE_OF_TRANSIENT_SYMBOLp, NIL), listO(4, SGT_WALK_STELLA_KEYWORD, SYM_WALK_STELLA_KEYWORDp, SYM_WALK_STELLA_SUBTYPE_OF_KEYWORDp, NIL), listO(4, SGT_WALK_STELLA_CONS, SYM_WALK_STELLA_CONSp, SYM_WALK_STELLA_SUBTYPE_OF_CONSp, NIL), listO(4, SGT_WALK_STELLA_CLASS, SYM_WALK_STELLA_STELLA_CLASSp, SYM_WALK_STELLA_SUBTYPE_OF_CLASSp, NIL), listO(4, SGT_WALK_STELLA_STORAGE_SLOT, SYM_WALK_STELLA_STORAGE_SLOTp, SYM_WALK_STELLA_SUBTYPE_OF_STORAGE_SLOTp, NIL), listO(4, SGT_WALK_STELLA_METHOD_SLOT, SYM_WALK_STELLA_METHOD_SLOTp, SYM_WALK_STELLA_SUBTYPE_OF_METHOD_SLOTp, NIL), listO(4, SGT_WALK_STELLA_ANCHORED_TYPE_SPECIFIER, SYM_WALK_STELLA_ANCHORED_TYPE_SPECIFIERp, SYM_WALK_STELLA_SUBTYPE_OF_ANCHORED_TYPE_SPECIFIERp, NIL), listO(4, SGT_WALK_STELLA_PARAMETRIC_TYPE_SPECIFIER, SYM_WALK_STELLA_PARAMETRIC_TYPE_SPECIFIERp, SYM_WALK_STELLA_SUBTYPE_OF_PARAMETRIC_TYPE_SPECIFIERp, NIL), NIL);
    oNUMERIC_TYPE_HIERARCHYo = list(3, SGT_WALK_STELLA_INTEGER, SGT_WALK_STELLA_FLOAT, SGT_WALK_STELLA_NUMBER);
    oMIXIN_IMPLEMENTATION_STYLEo = KWD_WALK_SECOND_CLASS;
    oMACRO_TABLEo = new_hash_table();
    oNATIVE_NAME_TABLEo = new_hash_table();
  }
}

void help_startup_walk9() {
  {
    define_function_object("DISPLAY-ENVIRONMENT-FEATURES", "(DEFUN DISPLAY-ENVIRONMENT-FEATURES () :DOCUMENTATION \"Print a list of the currently enabled environment \nfeatures and print a list of all Stella environment features.\")", ((cpp_function_code)(&display_environment_features)), NULL);
    define_function_object("ENABLE-ENVIRONMENT-FEATURE", "(DEFUN ENABLE-ENVIRONMENT-FEATURE (|&REST| (FEATURES KEYWORD)) :DOCUMENTATION \"Enable the environment feature(s) indicated by the\nkeywords in `features'.\")", ((cpp_function_code)(&enable_environment_feature)), NULL);
    define_function_object("DISABLE-ENVIRONMENT-FEATURE", "(DEFUN DISABLE-ENVIRONMENT-FEATURE (|&REST| (FEATURES KEYWORD)) :DOCUMENTATION \"Disable the environment feature(s) indicated by the\nkeywords in `features'.\")", ((cpp_function_code)(&disable_environment_feature)), NULL);
    define_function_object("RESET-ENVIRONMENT-FEATURES", "(DEFUN RESET-ENVIRONMENT-FEATURES () :DOCUMENTATION \"Reset the current environment features to their default\nsettings.\")", ((cpp_function_code)(&reset_environment_features)), NULL);
    define_function_object("TRACE-IF-EXPANDER", "(DEFUN (TRACE-IF-EXPANDER OBJECT) ((KEYWORD OBJECT) |&BODY| (BODY CONS)))", ((cpp_function_code)(&trace_if_expander)), NULL);
    define_function_object("TRACE-KEYWORD?", "(DEFUN (TRACE-KEYWORD? BOOLEAN) ((KEYWORD OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&trace_keywordP)), NULL);
    define_function_object("ADD-TRACE", "(DEFUN (ADD-TRACE LIST) (|&REST| (KEYWORDS GENERALIZED-SYMBOL)) :COMMAND? TRUE :PUBLIC? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Enable trace messages identified by any of the listed `keywords'.  After\ncalling (`add-trace' <keyword>) code guarded by (`trace-if' <keyword> ...)\nwill be executed when it is encountered.\")", ((cpp_function_code)(&add_trace)), ((cpp_function_code)(&add_trace_evaluator_wrapper)));
    define_function_object("DROP-TRACE", "(DEFUN (DROP-TRACE LIST) (|&REST| (KEYWORDS GENERALIZED-SYMBOL)) :COMMAND? TRUE :PUBLIC? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Disable trace messages identified by any of the listed `keywords'.  After\ncalling (`drop-trace' <keyword>) code guarded by (`trace-if' <keyword> ...)\nwill not be executed when it is encountered.\")", ((cpp_function_code)(&drop_trace)), ((cpp_function_code)(&drop_trace_evaluator_wrapper)));
    define_function_object("CLEAR-TRACE", "(DEFUN CLEAR-TRACE () :COMMAND? TRUE :PUBLIC? TRUE :DOCUMENTATION \"Disable all tracing previously enabled with `add-trace'.\")", ((cpp_function_code)(&clear_trace)), NULL);
    define_function_object("SET-OPTIMIZATION-LEVELS", "(DEFUN SET-OPTIMIZATION-LEVELS ((SAFETY INTEGER) (DEBUG INTEGER) (SPEED INTEGER) (SPACE INTEGER)) :DOCUMENTATION \"Set optimization levels for the qualities `safety', `debug',\n`speed', and `space'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&set_optimization_levels)), NULL);
    define_function_object("PRESERVE-TAIL-MERGE-OPTIMIZABILITY?", "(DEFUN (PRESERVE-TAIL-MERGE-OPTIMIZABILITY? BOOLEAN) ())", ((cpp_function_code)(&preserve_tail_merge_optimizabilityP)), NULL);
    define_function_object("METHOD-CALL-INLINING-ENABLED?", "(DEFUN (METHOD-CALL-INLINING-ENABLED? BOOLEAN) ())", ((cpp_function_code)(&method_call_inlining_enabledP)), NULL);
    define_function_object("OPTIMIZE-BOOLEAN-TESTS?", "(DEFUN (OPTIMIZE-BOOLEAN-TESTS? BOOLEAN) ())", ((cpp_function_code)(&optimize_boolean_testsP)), NULL);
    define_function_object("CHECK-FOR-ILLEGAL-RETURN?", "(DEFUN (CHECK-FOR-ILLEGAL-RETURN? BOOLEAN) ())", ((cpp_function_code)(&check_for_illegal_returnP)), NULL);
    define_function_object("NAME-QUOTED-TREE", "(DEFUN (NAME-QUOTED-TREE STRING) ((TREE CONS)))", ((cpp_function_code)(&name_quoted_tree)), NULL);
    define_function_object("GET-QUOTED-TREE", "(DEFUN (GET-QUOTED-TREE CONS) ((TREE-NAME STRING) (MODULENAME STRING)) :DOCUMENTATION \"Return the quoted tree with name `tree-name'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&get_quoted_tree)), NULL);
    define_function_object("CLEAR-TRANSLATION-UNIT", "(DEFUN CLEAR-TRANSLATION-UNIT ((SELF TRANSLATION-UNIT)))", ((cpp_function_code)(&clear_translation_unit)), NULL);
    define_function_object("PRINT-TRANSLATION-UNIT", "(DEFUN PRINT-TRANSLATION-UNIT ((SELF TRANSLATION-UNIT) (STREAM NATIVE-OUTPUT-STREAM)))", ((cpp_function_code)(&print_translation_unit)), NULL);
    define_function_object("USE-HARDCODED-SYMBOLS?", "(DEFUN (USE-HARDCODED-SYMBOLS? BOOLEAN) ())", ((cpp_function_code)(&use_hardcoded_symbolsP)), NULL);
    define_function_object("TRANSLATOR-OUTPUT-LANGUAGE", "(DEFUN (TRANSLATOR-OUTPUT-LANGUAGE KEYWORD) () :GLOBALLY-INLINE? TRUE (RETURN *TRANSLATOROUTPUTLANGUAGE*))", ((cpp_function_code)(&translator_output_language)), NULL);
    define_function_object("TRANSLATOR-OUTPUT-LANGUAGE-NAME", "(DEFUN (TRANSLATOR-OUTPUT-LANGUAGE-NAME STRING) ())", ((cpp_function_code)(&translator_output_language_name)), NULL);
    define_function_object("SET-TRANSLATOR-OUTPUT-LANGUAGE", "(DEFUN (SET-TRANSLATOR-OUTPUT-LANGUAGE KEYWORD) ((NEW-LANGUAGE KEYWORD)) :DOCUMENTATION \"Set output language to `new-language'.  Return previous language.\" :COMMAND? TRUE :PUBLIC? TRUE)", ((cpp_function_code)(&set_translator_output_language)), NULL);
    define_function_object("TRANSLATE-TO-COMMON-LISP?", "(DEFUN (TRANSLATE-TO-COMMON-LISP? BOOLEAN) () :DOCUMENTATION \"Return `true' if current output language is Common-Lisp.\")", ((cpp_function_code)(&translate_to_common_lispP)), NULL);
    define_function_object("TRANSLATE-TO-CPP?", "(DEFUN (TRANSLATE-TO-CPP? BOOLEAN) () :DOCUMENTATION \"Return `true' if current output language is C++\")", ((cpp_function_code)(&translate_to_cppP)), NULL);
    define_function_object("TRANSLATE-TO-JAVA?", "(DEFUN (TRANSLATE-TO-JAVA? BOOLEAN) () :DOCUMENTATION \"Return `true' if current output language is Java\")", ((cpp_function_code)(&translate_to_javaP)), NULL);
    define_function_object("TRANSLATE-TO-SINGLE-INHERITANCE-LANGUAGE?", "(DEFUN (TRANSLATE-TO-SINGLE-INHERITANCE-LANGUAGE? BOOLEAN) ())", ((cpp_function_code)(&translate_to_single_inheritance_languageP)), NULL);
    define_function_object("TOGGLE-OUTPUT-LANGUAGE", "(DEFUN (TOGGLE-OUTPUT-LANGUAGE KEYWORD) () :DOCUMENTATION \"Switch between Common Lisp and C++ as output languages.\")", ((cpp_function_code)(&toggle_output_language)), NULL);
    define_function_object("USE-BOOTSTRAP-TRANSLATOR?", "(DEFUN (USE-BOOTSTRAP-TRANSLATOR? BOOLEAN) ())", ((cpp_function_code)(&use_bootstrap_translatorP)), NULL);
    define_function_object("TOGGLE-TRANSLATORS", "(DEFUN (TOGGLE-TRANSLATORS KEYWORD) ())", ((cpp_function_code)(&toggle_translators)), NULL);
    define_function_object("DECLARATION-OPERATOR?", "(DEFUN (DECLARATION-OPERATOR? BOOLEAN) ((OPERATOR SYMBOL)))", ((cpp_function_code)(&declaration_operatorP)), NULL);
    define_function_object("DECLARATION-TREE?", "(DEFUN (DECLARATION-TREE? BOOLEAN) ((TREE CONS)) :PUBLIC? TRUE)", ((cpp_function_code)(&declaration_treeP)), NULL);
    define_function_object("CREATE-ANNOTATION", "(DEFUN (CREATE-ANNOTATION STRING) ((TOP-LEVEL-TREE CONS)))", ((cpp_function_code)(&create_annotation)), NULL);
    define_function_object("WALK-TOP-LEVEL-TREE", "(DEFUN WALK-TOP-LEVEL-TREE ((TREE CONS) (CREATEANNOTATION? BOOLEAN)) :PUBLIC? TRUE)", ((cpp_function_code)(&walk_top_level_tree)), NULL);
    define_function_object("WALK-TOP-LEVEL-EXPRESSION", "(DEFUN (WALK-TOP-LEVEL-EXPRESSION OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&walk_top_level_expression)), NULL);
    define_function_object("WALK-PHASE-ONE-UNIT", "(DEFUN WALK-PHASE-ONE-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_phase_one_unit)), NULL);
    define_function_object("WALK-AUXILIARY-TREE", "(DEFUN WALK-AUXILIARY-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_auxiliary_tree)), NULL);
    define_function_object("HELP-WALK-AUXILIARY-TREE", "(DEFUN (HELP-WALK-AUXILIARY-TREE TRANSLATION-UNIT) ((TREE CONS) (FINALIZE? BOOLEAN)))", ((cpp_function_code)(&help_walk_auxiliary_tree)), NULL);
    define_function_object("WALK-ALL-PHASE-ONE-UNITS", "(DEFUN WALK-ALL-PHASE-ONE-UNITS ())", ((cpp_function_code)(&walk_all_phase_one_units)), NULL);
    define_function_object("TRANSLATE-ALL-UNITS", "(DEFUN TRANSLATE-ALL-UNITS ())", ((cpp_function_code)(&translate_all_units)), NULL);
    define_function_object("UNIT-INCLUDED-IN-OTHER-UNITS?", "(DEFUN (UNIT-INCLUDED-IN-OTHER-UNITS? BOOLEAN) ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&unit_included_in_other_unitsP)), NULL);
    define_function_object("COMBINE-TRANSLATED-TREES", "(DEFUN (COMBINE-TRANSLATED-TREES CONS) ())", ((cpp_function_code)(&combine_translated_trees)), NULL);
    define_function_object("TRANSLATE-WALKED-TREE", "(DEFUN (TRANSLATE-WALKED-TREE OBJECT) ((CODETREE OBJECT)))", ((cpp_function_code)(&translate_walked_tree)), NULL);
    define_function_object("INCREMENTALLY-TRANSLATE", "(DEFUN (INCREMENTALLY-TRANSLATE OBJECT) ((TREE OBJECT)) :DOCUMENTATION \"Translate a single Stella expression `tree' and return\nthe result.  For C++ and Java print the translation to standard output and\nreturn NIL instead.\" :PUBLIC? TRUE)", ((cpp_function_code)(&incrementally_translate)), NULL);
    define_function_object("RESET-TRANSLATION-ERRORS", "(DEFUN RESET-TRANSLATION-ERRORS ())", ((cpp_function_code)(&reset_translation_errors)), NULL);
    define_function_object("SIGNAL-TRANSLATION-ERROR", "(DEFUN SIGNAL-TRANSLATION-ERROR () :PUBLIC? TRUE)", ((cpp_function_code)(&signal_translation_error)), NULL);
    define_function_object("SIGNAL-TRANSLATION-WARNING", "(DEFUN SIGNAL-TRANSLATION-WARNING () :PUBLIC? TRUE)", ((cpp_function_code)(&signal_translation_warning)), NULL);
    define_function_object("IGNORE-TRANSLATION-ERRORS?", "(DEFUN (IGNORE-TRANSLATION-ERRORS? BOOLEAN) ())", ((cpp_function_code)(&ignore_translation_errorsP)), NULL);
    define_function_object("TRANSLATION-ERRORS?", "(DEFUN (TRANSLATION-ERRORS? BOOLEAN) () :PUBLIC? TRUE)", ((cpp_function_code)(&translation_errorsP)), NULL);
    define_function_object("SUMMARIZE-TRANSLATION-ERRORS", "(DEFUN SUMMARIZE-TRANSLATION-ERRORS ())", ((cpp_function_code)(&summarize_translation_errors)), NULL);
    define_function_object("PRINT-ERROR-CONTEXT", "(DEFUN PRINT-ERROR-CONTEXT ((PREFIX STRING) (STREAM OUTPUT-STREAM)) :PUBLIC? TRUE)", ((cpp_function_code)(&print_error_context)), NULL);
    define_function_object("LOG-FUNCTION-CALL?", "(DEFUN (LOG-FUNCTION-CALL? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&log_function_callP)), NULL);
    define_function_object("WRAP-BODY-WITH-LOG-FUNCTION-CALL-TREE", "(DEFUN (WRAP-BODY-WITH-LOG-FUNCTION-CALL-TREE CONS) ((METHOD METHOD-SLOT) (BODY CONS)))", ((cpp_function_code)(&wrap_body_with_log_function_call_tree)), NULL);
    define_function_object("LOG-FUNCTION-CALL", "(DEFUN LOG-FUNCTION-CALL ((NAME STRING)))", ((cpp_function_code)(&log_function_call)), NULL);
    define_function_object("START-FUNCTION-CALL-LOGGING", "(DEFUN START-FUNCTION-CALL-LOGGING ((FILENAME STRING)) :DOCUMENTATION \"Start function call logging to `fileName'.\" :COMMAND? TRUE :PUBLIC? TRUE)", ((cpp_function_code)(&start_function_call_logging)), ((cpp_function_code)(&start_function_call_logging_evaluator_wrapper)));
    define_function_object("STOP-FUNCTION-CALL-LOGGING", "(DEFUN STOP-FUNCTION-CALL-LOGGING () :DOCUMENTATION \"Stop function call logging and close the current log file.\" :COMMAND? TRUE :PUBLIC? TRUE)", ((cpp_function_code)(&stop_function_call_logging)), NULL);
    define_function_object("SET-CALL-LOG-BREAK-POINT", "(DEFUN SET-CALL-LOG-BREAK-POINT ((COUNT INTEGER)) :DOCUMENTATION \"Set a call log break point to `count'.  Execution will be\ninterrupted right at the entry of the `count'th logged function call.\" :COMMAND? TRUE :PUBLIC? TRUE)", ((cpp_function_code)(&set_call_log_break_point)), ((cpp_function_code)(&set_call_log_break_point_evaluator_wrapper)));
    define_function_object("BREAK-PROGRAM", "(DEFUN BREAK-PROGRAM ((MESSAGE STRING)) :DOCUMENTATION \"Interrupt the program and print `message'.  Continue after\nconfirmation with the user.\" :PUBLIC? TRUE)", ((cpp_function_code)(&break_program)), NULL);
    define_function_object("TERMINATE-PROGRAM", "(DEFUN TERMINATE-PROGRAM () :DOCUMENTATION \"Terminate and exit the program with normal exit code.\" :PUBLIC? TRUE)", ((cpp_function_code)(&terminate_program)), NULL);
    define_function_object("PO", "(DEFUN PO ((THING OBJECT)))", ((cpp_function_code)(&po)), NULL);
    define_function_object("VARIABLE-EQL?", "(DEFUN (VARIABLE-EQL? BOOLEAN) ((VAR1 SYMBOL) (VAR2 SYMBOL)))", ((cpp_function_code)(&variable_eqlP)), NULL);
  }
}

void help_startup_walk10() {
  {
    define_function_object("LOOKUP-VARIABLE-TABLE", "(DEFUN (LOOKUP-VARIABLE-TABLE (LIKE (ANY-VALUE SELF))) ((SELF (KEY-VALUE-LIST OF SYMBOL OBJECT)) (VARIABLE SYMBOL)))", ((cpp_function_code)(&lookup_variable_table)), NULL);
    define_function_object("INSERT-AT-VARIABLE-TABLE", "(DEFUN INSERT-AT-VARIABLE-TABLE ((SELF (KEY-VALUE-LIST OF SYMBOL OBJECT)) (VARIABLE SYMBOL) (VALUE OBJECT)))", ((cpp_function_code)(&insert_at_variable_table)), NULL);
    define_function_object("PUSH-VARIABLE-BINDING", "(DEFUN PUSH-VARIABLE-BINDING ((VARIABLE SYMBOL) (TYPE TYPE-SPEC)))", ((cpp_function_code)(&push_variable_binding)), NULL);
    define_function_object("POP-VARIABLE-BINDING", "(DEFUN POP-VARIABLE-BINDING ())", ((cpp_function_code)(&pop_variable_binding)), NULL);
    define_function_object("LOOKUP-VARIABLE-TYPE", "(DEFUN (LOOKUP-VARIABLE-TYPE TYPE-SPEC) ((VARIABLENAME SYMBOL)))", ((cpp_function_code)(&lookup_variable_type)), NULL);
    define_function_object("SET-LOCAL-VARIABLE-INFO", "(DEFUN SET-LOCAL-VARIABLE-INFO ((VARIABLE SYMBOL) (KEY KEYWORD) (INFO OBJECT)))", ((cpp_function_code)(&set_local_variable_info)), NULL);
    define_function_object("GET-LOCAL-VARIABLE-INFO", "(DEFUN (GET-LOCAL-VARIABLE-INFO OBJECT) ((VARIABLE SYMBOL) (KEY KEYWORD)))", ((cpp_function_code)(&get_local_variable_info)), NULL);
    define_function_object("REGISTER-REFERENCE-TO-GLOBAL-VARIABLE", "(DEFUN REGISTER-REFERENCE-TO-GLOBAL-VARIABLE ((VARIABLENAME SYMBOL)))", ((cpp_function_code)(&register_reference_to_global_variable)), NULL);
    define_function_object("LOCAL-GENSYM", "(DEFUN (LOCAL-GENSYM SYMBOL) ((PREFIX STRING)))", ((cpp_function_code)(&local_gensym)), NULL);
    define_function_object("YIELD-LOCAL-VARIABLE-ALIAS", "(DEFUN (YIELD-LOCAL-VARIABLE-ALIAS SYMBOL) ((VARIABLE SYMBOL) (LEVEL INTEGER)))", ((cpp_function_code)(&yield_local_variable_alias)), NULL);
    define_function_object("RENAME-SHADOWING-LOCAL-VARIABLES?", "(DEFUN (RENAME-SHADOWING-LOCAL-VARIABLES? BOOLEAN) () :GLOBALLY-INLINE? TRUE (RETURN (EQL? (TRANSLATOR-OUTPUT-LANGUAGE) :JAVA)))", ((cpp_function_code)(&rename_shadowing_local_variablesP)), NULL);
    define_function_object("MAYBE-RENAME-LOCAL-VARIABLE", "(DEFUN MAYBE-RENAME-LOCAL-VARIABLE ((VARIABLE SYMBOL)))", ((cpp_function_code)(&maybe_rename_local_variable)), NULL);
    define_function_object("TRUE-VARIABLE-NAME", "(DEFUN (TRUE-VARIABLE-NAME SYMBOL) ((VARIABLE SYMBOL)))", ((cpp_function_code)(&true_variable_name)), NULL);
    define_function_object("SUPPRESS-WARNINGS?", "(DEFUN (SUPPRESS-WARNINGS? BOOLEAN) () :PUBLIC? TRUE)", ((cpp_function_code)(&suppress_warningsP)), NULL);
    define_function_object("BAD-ARGUMENT-RANGE?", "(DEFUN (BAD-ARGUMENT-RANGE? BOOLEAN) ((TREE CONS) (MINARITY INTEGER) (MAXARITY INTEGER)))", ((cpp_function_code)(&bad_argument_rangeP)), NULL);
    define_function_object("BAD-ARGUMENT-COUNT?", "(DEFUN (BAD-ARGUMENT-COUNT? BOOLEAN) ((TREE CONS) (ARITY INTEGER)))", ((cpp_function_code)(&bad_argument_countP)), NULL);
    define_function_object("YIELD-ILLEGAL-STATEMENT-TREE", "(DEFUN (YIELD-ILLEGAL-STATEMENT-TREE CONS) ())", ((cpp_function_code)(&yield_illegal_statement_tree)), NULL);
    define_function_object("YIELD-ILLEGAL-EXPRESSION-TREE", "(DEFUN (YIELD-ILLEGAL-EXPRESSION-TREE OBJECT) ())", ((cpp_function_code)(&yield_illegal_expression_tree)), NULL);
    define_function_object("WALK-DONT-CALL-ME-TREE", "(DEFUN (WALK-DONT-CALL-ME-TREE CONS TYPE-SPEC) ((TREE OBJECT) (RETURNTYPE TYPE-SPEC)))", ((cpp_function_code)(&walk_dont_call_me_tree)), NULL);
    define_function_object("ILLEGAL-TREE?", "(DEFUN (ILLEGAL-TREE? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&illegal_treeP)), NULL);
    define_function_object("NATIVE-CLASS-METHOD-INLINING?", "(DEFUN (NATIVE-CLASS-METHOD-INLINING? BOOLEAN) ())", ((cpp_function_code)(&native_class_method_inliningP)), NULL);
    define_function_object("PROGNIFY", "(DEFUN (PROGNIFY OBJECT) ((TREES CONS)))", ((cpp_function_code)(&prognify)), NULL);
    define_function_object("OPTIMIZE-PROGN", "(DEFUN (OPTIMIZE-PROGN OBJECT) ((PROGN-TREE CONS)))", ((cpp_function_code)(&optimize_progn)), NULL);
    define_function_object("TREE-TO-TREES", "(DEFUN (TREE-TO-TREES CONS) ((TREE OBJECT)))", ((cpp_function_code)(&tree_to_trees)), NULL);
    define_function_object("WRAPPED-TYPE-TO-TYPE", "(DEFUN (WRAPPED-TYPE-TO-TYPE TYPE) ((SELF TYPE)) :DOCUMENTATION \"Return the unwrapped type for the wrapped type `self',\nor `self' if it is not a wrapped type.\" :PUBLIC? TRUE)", ((cpp_function_code)(&wrapped_type_to_type)), NULL);
    define_function_object("WRAPPER-VALUE-TYPE", "(DEFUN (WRAPPER-VALUE-TYPE TYPE) ((SELF WRAPPER)) :DOCUMENTATION \"Return the type of the value stored in the wrapper `self'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&wrapper_value_type)), NULL);
    define_method_object("(DEFMETHOD (TYPE-TO-WRAPPED-TYPE TYPE) ((SELF TYPE)) :DOCUMENTATION \"Return the wrapped type for the type `self',\nor 'self' if it is not a bare literal type.\" :PUBLIC? TRUE)", ((cpp_method_code)(&Surrogate::type_to_wrapped_type)), ((cpp_method_code)(NULL)));
    define_function_object("LOOKUP-COERSION-METHOD", "(DEFUN (LOOKUP-COERSION-METHOD OBJECT) ((SOURCETYPE TYPE) (TARGETTYPE TYPE) (EXPRESSION OBJECT)))", ((cpp_function_code)(&lookup_coersion_method)), NULL);
    define_function_object("APPLY-COERCION-METHOD", "(DEFUN (APPLY-COERCION-METHOD OBJECT TYPE-SPEC) ((EXPRESSION OBJECT) (SOURCETYPE TYPE) (TARGETTYPE TYPE) (COERCIONMETHOD OBJECT)))", ((cpp_function_code)(&apply_coercion_method)), NULL);
    define_function_object("TRANSLATING-CODE?", "(DEFUN (TRANSLATING-CODE? BOOLEAN) ())", ((cpp_function_code)(&translating_codeP)), NULL);
    define_function_object("VERIFY-TYPE", "(DEFUN (VERIFY-TYPE TYPE) ((SELF TYPE)))", ((cpp_function_code)(&verify_type)), NULL);
    define_function_object("SAFE-SUBTYPE-OF?", "(DEFUN (SAFE-SUBTYPE-OF? BOOLEAN) ((SUBTYPE TYPE) (SUPERTYPE TYPE)))", ((cpp_function_code)(&safe_subtype_ofP)), NULL);
    define_function_object("WARN-ABOUT-UNKNOWN-SOURCE-TYPE", "(DEFUN WARN-ABOUT-UNKNOWN-SOURCE-TYPE ((TREE OBJECT)))", ((cpp_function_code)(&warn_about_unknown_source_type)), NULL);
    define_function_object("COERCE-A-TREE", "(DEFUN (COERCE-A-TREE OBJECT TYPE-SPEC) ((TREE OBJECT) (SOURCETYPE TYPE-SPEC) (TARGETTYPE TYPE-SPEC)))", ((cpp_function_code)(&coerce_a_tree)), NULL);
    define_function_object("HELP-COERCE-A-TREE", "(DEFUN (HELP-COERCE-A-TREE OBJECT TYPE-SPEC BOOLEAN) ((TREE OBJECT) (SOURCETYPE TYPE-SPEC) (TARGETTYPE TYPE-SPEC) (TESTONLY? BOOLEAN)))", ((cpp_function_code)(&help_coerce_a_tree)), NULL);
    define_function_object("COERCIBLE?", "(DEFUN (COERCIBLE? BOOLEAN) ((TREE OBJECT) (SOURCETYPE TYPE-SPEC) (TARGETTYPE TYPE-SPEC)))", ((cpp_function_code)(&coercibleP)), NULL);
    define_function_object("WALKING-EXPRESSION?", "(DEFUN (WALKING-EXPRESSION? BOOLEAN) ())", ((cpp_function_code)(&walking_expressionP)), NULL);
    define_function_object("WALK-EXPRESSION-TREE", "(DEFUN (WALK-EXPRESSION-TREE OBJECT TYPE-SPEC) ((EXPTREE OBJECT) (TARGETTYPE TYPE-SPEC) (OPERATORNAME SYMBOL) (VRLETISOK? BOOLEAN)))", ((cpp_function_code)(&walk_expression_tree)), NULL);
    define_function_object("WALK-WITHOUT-TYPE-TREE", "(DEFUN (WALK-WITHOUT-TYPE-TREE OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&walk_without_type_tree)), NULL);
    define_function_object("WALK-LIST-OF-TREES", "(DEFUN (WALK-LIST-OF-TREES CONS) ((TREES CONS)))", ((cpp_function_code)(&walk_list_of_trees)), NULL);
    define_function_object("WALK-LIST-OF-STATEMENTS", "(DEFUN (WALK-LIST-OF-STATEMENTS CONS) ((TREES CONS)))", ((cpp_function_code)(&walk_list_of_statements)), NULL);
    define_function_object("WALK-STATEMENT", "(DEFUN (WALK-STATEMENT OBJECT) ((TREE OBJECT) (WARNONATOM? BOOLEAN)))", ((cpp_function_code)(&walk_statement)), NULL);
    define_function_object("ELIMINATE-VRLET-STATEMENT", "(DEFUN (ELIMINATE-VRLET-STATEMENT CONS) ((TREE CONS)))", ((cpp_function_code)(&eliminate_vrlet_statement)), NULL);
    define_function_object("SYS-TREE", "(DEFUN (SYS-TREE CONS TYPE-SPEC) ((TREE OBJECT) (TYPESPEC TYPE-SPEC)))", ((cpp_function_code)(&sys_tree)), NULL);
    define_function_object("WALKED-EXPRESSION-TYPE", "(DEFUN (WALKED-EXPRESSION-TYPE TYPE-SPEC) ((TREE OBJECT)))", ((cpp_function_code)(&walked_expression_type)), NULL);
    define_function_object("NEED-IDENTICAL-METHOD-SIGNATURES?", "(DEFUN (NEED-IDENTICAL-METHOD-SIGNATURES? BOOLEAN) ())", ((cpp_function_code)(&need_identical_method_signaturesP)), NULL);
    define_function_object("COMPUTE-MOST-GENERAL-RETURN-TYPE", "(DEFUN (COMPUTE-MOST-GENERAL-RETURN-TYPE TYPE-SPEC) ((METHOD METHOD-SLOT) (RETURNTYPE TYPE-SPEC)))", ((cpp_function_code)(&compute_most_general_return_type)), NULL);
    define_function_object("COMPUTE-REAL-SLOT-TYPE", "(DEFUN (COMPUTE-REAL-SLOT-TYPE TYPE-SPEC) ((SLOT STORAGE-SLOT) (FIRSTARGTYPE TYPE-SPEC) (RETURNTYPE TYPE-SPEC)))", ((cpp_function_code)(&compute_real_slot_type)), NULL);
    define_function_object("SYS-TREE-IF-NEEDED", "(DEFUN (SYS-TREE-IF-NEEDED OBJECT TYPE-SPEC) ((SLOT SLOT) (TREE OBJECT) (FIRSTARGTYPE TYPE-SPEC) (RETURNTYPE TYPE-SPEC)))", ((cpp_function_code)(&sys_tree_if_needed)), NULL);
    define_function_object("TARGET-LANGUAGE-TYPE", "(DEFUN (TARGET-LANGUAGE-TYPE TYPE) ((TREE OBJECT)))", ((cpp_function_code)(&target_language_type)), NULL);
    define_function_object("SET-TARGET-LANGUAGE-TYPE", "(DEFUN (SET-TARGET-LANGUAGE-TYPE CONS TYPE-SPEC) ((SYSTREE CONS) (TYPE TYPE)))", ((cpp_function_code)(&set_target_language_type)), NULL);
    define_function_object("WALK-A-TREE", "(DEFUN (WALK-A-TREE OBJECT TYPE-SPEC) ((TREE OBJECT)))", ((cpp_function_code)(&walk_a_tree)), NULL);
    define_function_object("HELP-WALK-A-TREE", "(DEFUN (HELP-WALK-A-TREE OBJECT TYPE-SPEC) ((TREE OBJECT)))", ((cpp_function_code)(&help_walk_a_tree)), NULL);
    define_function_object("WALK-A-CONS-TREE", "(DEFUN (WALK-A-CONS-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_a_cons_tree)), NULL);
    define_function_object("INCREMENTAL-TRANSLATION?", "(DEFUN (INCREMENTAL-TRANSLATION? BOOLEAN) ())", ((cpp_function_code)(&incremental_translationP)), NULL);
    define_function_object("CONSTRUCT-SYMBOL-CONSTANT-NAME", "(DEFUN (CONSTRUCT-SYMBOL-CONSTANT-NAME STRING) ((SYMBOL GENERALIZED-SYMBOL)))", ((cpp_function_code)(&construct_symbol_constant_name)), NULL);
    define_function_object("YIELD-SYMBOL-CONSTANT-NAME", "(DEFUN (YIELD-SYMBOL-CONSTANT-NAME SYMBOL) ((SYMBOL GENERALIZED-SYMBOL)))", ((cpp_function_code)(&yield_symbol_constant_name)), NULL);
    define_function_object("CREATE-STARTUP-SYMBOL", "(DEFUN (CREATE-STARTUP-SYMBOL SYMBOL) ((SYMBOL GENERALIZED-SYMBOL)))", ((cpp_function_code)(&create_startup_symbol)), NULL);
    define_function_object("REGISTER-SYMBOL", "(DEFUN (REGISTER-SYMBOL GENERALIZED-SYMBOL) ((SYMBOL GENERALIZED-SYMBOL)))", ((cpp_function_code)(&register_symbol)), NULL);
    define_function_object("CLEAR-SYMBOL-REGISTRY", "(DEFUN CLEAR-SYMBOL-REGISTRY ())", ((cpp_function_code)(&clear_symbol_registry)), NULL);
  }
}

void help_startup_walk11() {
  {
    define_function_object("YIELD-HARDCODED-INTERN-REGISTERED-SYMBOLS-TREE", "(DEFUN (YIELD-HARDCODED-INTERN-REGISTERED-SYMBOLS-TREE CONS) ())", ((cpp_function_code)(&yield_hardcoded_intern_registered_symbols_tree)), NULL);
    define_method_object("(DEFMETHOD (WALK-ATOMIC-TREE OBJECT TYPE-SPEC) ((SELF OBJECT)))", ((cpp_method_code)(&Object::walk_atomic_tree)), ((cpp_method_code)(NULL)));
    define_method_object("(DEFMETHOD (WALK-ATOMIC-TREE OBJECT TYPE-SPEC) ((SELF WRAPPER)))", ((cpp_method_code)(&Wrapper::walk_atomic_tree)), ((cpp_method_code)(NULL)));
    define_function_object("LOOKUP-CONSTANT-SYMBOL", "(DEFUN (LOOKUP-CONSTANT-SYMBOL TYPE) ((SELF SYMBOL)))", ((cpp_function_code)(&lookup_constant_symbol)), NULL);
    define_function_object("CONSTANT-SYMBOL?", "(DEFUN (CONSTANT-SYMBOL? BOOLEAN) ((SYMBOL SYMBOL)))", ((cpp_function_code)(&constant_symbolP)), NULL);
    define_function_object("WALK-CONSTANT-SYMBOL", "(DEFUN (WALK-CONSTANT-SYMBOL SYMBOL TYPE) ((SELF SYMBOL)))", ((cpp_function_code)(&walk_constant_symbol)), NULL);
    define_method_object("(DEFMETHOD (WALK-ATOMIC-TREE OBJECT TYPE-SPEC) ((SELF SYMBOL)))", ((cpp_method_code)(&Symbol::walk_atomic_tree)), ((cpp_method_code)(NULL)));
    define_method_object("(DEFMETHOD (WALK-ATOMIC-TREE OBJECT TYPE-SPEC) ((SELF KEYWORD)))", ((cpp_method_code)(&Keyword::walk_atomic_tree)), ((cpp_method_code)(NULL)));
    define_method_object("(DEFMETHOD (WALK-ATOMIC-TREE OBJECT TYPE-SPEC) ((SELF SURROGATE)))", ((cpp_method_code)(&Surrogate::walk_atomic_tree)), ((cpp_method_code)(NULL)));
    define_function_object("ATOMIC-EXPRESSION?", "(DEFUN (ATOMIC-EXPRESSION? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&atomic_expressionP)), NULL);
    define_function_object("VARIABLE-EXPRESSION?", "(DEFUN (VARIABLE-EXPRESSION? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&variable_expressionP)), NULL);
    define_function_object("VRLET-EXPRESSION?", "(DEFUN (VRLET-EXPRESSION? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&vrlet_expressionP)), NULL);
    define_function_object("PROCEDURAL-EXPRESSION?", "(DEFUN (PROCEDURAL-EXPRESSION? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&procedural_expressionP)), NULL);
    define_function_object("SUBSTITUTE-ONCE", "(DEFUN (SUBSTITUTE-ONCE BOOLEAN) ((EXPRESSION OBJECT) (NEWTREE OBJECT) (OLDTREE OBJECT)))", ((cpp_function_code)(&substitute_once)), NULL);
    define_function_object("TRANSFORM-VRLET-EXPRESSION", "(DEFUN (TRANSFORM-VRLET-EXPRESSION CONS TYPE-SPEC) ((STATEMENT CONS) (VREXPRESSION CONS)))", ((cpp_function_code)(&transform_vrlet_expression)), NULL);
    define_function_object("HELP-TRANSFORM-BOOLEAN-PROCEDURAL-EXPRESSION", "(DEFUN (HELP-TRANSFORM-BOOLEAN-PROCEDURAL-EXPRESSION CONS) ((EXPRESSION OBJECT) (TESTVARIABLE SYMBOL)))", ((cpp_function_code)(&help_transform_boolean_procedural_expression)), NULL);
    define_function_object("TRANSFORM-BOOLEAN-PROCEDURAL-EXPRESSION", "(DEFUN (TRANSFORM-BOOLEAN-PROCEDURAL-EXPRESSION CONS TYPE-SPEC) ((STATEMENT CONS) (VREXPRESSION CONS)))", ((cpp_function_code)(&transform_boolean_procedural_expression)), NULL);
    define_function_object("PERCOLATE-OUT-BOOLEAN-VRLET-EXPRESSIONS", "(DEFUN (PERCOLATE-OUT-BOOLEAN-VRLET-EXPRESSIONS CONS) ((BOOLEANEXPRESSION CONS)))", ((cpp_function_code)(&percolate_out_boolean_vrlet_expressions)), NULL);
    define_function_object("TRANSFORM-PROCEDURAL-EXPRESSION", "(DEFUN (TRANSFORM-PROCEDURAL-EXPRESSION CONS TYPE-SPEC) ((STATEMENT CONS) (VREXPRESSION CONS) (TYPE TYPE-SPEC)))", ((cpp_function_code)(&transform_procedural_expression)), NULL);
    define_function_object("PERCOLATE-OUT-VRLET-EXPRESSION", "(DEFUN (PERCOLATE-OUT-VRLET-EXPRESSION CONS TYPE-SPEC) ((STATEMENT CONS) (VREXPRESSION CONS) (TYPE TYPE-SPEC)))", ((cpp_function_code)(&percolate_out_vrlet_expression)), NULL);
    define_function_object("WALK-SETQ-TREE", "(DEFUN (WALK-SETQ-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_setq_tree)), NULL);
    define_function_object("YIELD-FIXED-SLOT-VALUE-SETTER-TREE", "(DEFUN (YIELD-FIXED-SLOT-VALUE-SETTER-TREE CONS) ((SLOT SLOT) (OBJECTREF OBJECT) (VALUETREE OBJECT)))", ((cpp_function_code)(&yield_fixed_slot_value_setter_tree)), NULL);
    define_function_object("WALK-SETF-TREE", "(DEFUN (WALK-SETF-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_setf_tree)), NULL);
    define_function_object("YIELD-DYNAMIC-SLOT-VALUE-TREE", "(DEFUN (YIELD-DYNAMIC-SLOT-VALUE-TREE CONS) ((SLOT STORAGE-SLOT) (OBJECTREF OBJECT) (RETURNTYPE TYPE-SPEC)))", ((cpp_function_code)(&yield_dynamic_slot_value_tree)), NULL);
    define_function_object("YIELD-DYNAMIC-SLOT-VALUE-SETTER-TREE", "(DEFUN (YIELD-DYNAMIC-SLOT-VALUE-SETTER-TREE CONS) ((SLOT STORAGE-SLOT) (OBJECTREF OBJECT) (VALUEREF OBJECT) (RETURNTYPE TYPE-SPEC)))", ((cpp_function_code)(&yield_dynamic_slot_value_setter_tree)), NULL);
    define_function_object("YIELD-NATIVE-SLOT-VALUE-TREE", "(DEFUN (YIELD-NATIVE-SLOT-VALUE-TREE CONS) ((SLOT STORAGE-SLOT) (OBJECTREF OBJECT) (OBJECTTYPE TYPE-SPEC)))", ((cpp_function_code)(&yield_native_slot_value_tree)), NULL);
    define_function_object("YIELD-CONTEXT-SENSITIVE-SLOT-VALUE-TREE", "(DEFUN (YIELD-CONTEXT-SENSITIVE-SLOT-VALUE-TREE CONS TYPE-SPEC) ((SLOT STORAGE-SLOT) (OBJECTREF OBJECT) (RETURNTYPE TYPE-SPEC) (CSOPTIONS CONS)))", ((cpp_function_code)(&yield_context_sensitive_slot_value_tree)), NULL);
    define_function_object("YIELD-SLOT-VALUE-TREE", "(DEFUN (YIELD-SLOT-VALUE-TREE CONS TYPE-SPEC) ((SLOT STORAGE-SLOT) (OBJECTREF OBJECT) (OBJECTTYPE TYPE-SPEC) (RETURNTYPE TYPE-SPEC) (CSOPTIONS CONS)))", ((cpp_function_code)(&yield_slot_value_tree)), NULL);
    define_function_object("WALK-SLOT-VALUE-TREE", "(DEFUN (WALK-SLOT-VALUE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_slot_value_tree)), NULL);
    define_function_object("YIELD-NATIVE-SLOT-VALUE-SETTER-TREE", "(DEFUN (YIELD-NATIVE-SLOT-VALUE-SETTER-TREE CONS) ((SLOT STORAGE-SLOT) (OBJECTREF OBJECT) (OBJECTTYPE TYPE-SPEC) (VALUEREF OBJECT)))", ((cpp_function_code)(&yield_native_slot_value_setter_tree)), NULL);
    define_function_object("YIELD-CONTEXT-SENSITIVE-SLOT-VALUE-SETTER-TREE", "(DEFUN (YIELD-CONTEXT-SENSITIVE-SLOT-VALUE-SETTER-TREE CONS TYPE-SPEC) ((SLOT STORAGE-SLOT) (OBJECTREF OBJECT) (VALUEREF OBJECT) (RETURNTYPE TYPE-SPEC) (CSOPTIONS CONS)))", ((cpp_function_code)(&yield_context_sensitive_slot_value_setter_tree)), NULL);
    define_function_object("YIELD-SLOT-VALUE-SETTER-TREE", "(DEFUN (YIELD-SLOT-VALUE-SETTER-TREE CONS TYPE-SPEC) ((SLOT STORAGE-SLOT) (OBJECTREF OBJECT) (OBJECTTYPE TYPE-SPEC) (VALUEREF OBJECT) (RETURNTYPE TYPE-SPEC) (CSOPTIONS CONS)))", ((cpp_function_code)(&yield_slot_value_setter_tree)), NULL);
    define_function_object("WALK-SLOT-VALUE-SETTER-TREE", "(DEFUN (WALK-SLOT-VALUE-SETTER-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_slot_value_setter_tree)), NULL);
    define_function_object("WALK-AREF-TREE", "(DEFUN (WALK-AREF-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_aref_tree)), NULL);
    define_function_object("SAFE-YIELD-TYPE-SPECIFIER", "(DEFUN (SAFE-YIELD-TYPE-SPECIFIER TYPE-SPEC) ((TYPETREE OBJECT)))", ((cpp_function_code)(&safe_yield_type_specifier)), NULL);
    define_function_object("WALK-A-DECLARATION", "(DEFUN (WALK-A-DECLARATION CONS) ((VARIABLE SYMBOL) (TYPETREE OBJECT) (VALUE OBJECT) (INPUTPARAMETER? BOOLEAN)))", ((cpp_function_code)(&walk_a_declaration)), NULL);
    define_function_object("WALK-VARIABLE-DECLARATIONS", "(DEFUN (WALK-VARIABLE-DECLARATIONS CONS) ((DECLARATIONS CONS)))", ((cpp_function_code)(&walk_variable_declarations)), NULL);
    define_function_object("TRANSFORM-LET-WITH-PROCEDURAL-EXPRESSION", "(DEFUN (TRANSFORM-LET-WITH-PROCEDURAL-EXPRESSION CONS) ((TREE CONS)))", ((cpp_function_code)(&transform_let_with_procedural_expression)), NULL);
    define_function_object("POP-LOCAL-VARIABLE-BINDINGS", "(DEFUN POP-LOCAL-VARIABLE-BINDINGS ((DECLARATIONS CONS)))", ((cpp_function_code)(&pop_local_variable_bindings)), NULL);
    define_function_object("WALK-LET-TREE", "(DEFUN (WALK-LET-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_let_tree)), NULL);
    define_function_object("SPECIAL-IMPLEMENTATION-STYLE", "(DEFUN (SPECIAL-IMPLEMENTATION-STYLE KEYWORD) ())", ((cpp_function_code)(&special_implementation_style)), NULL);
    define_function_object("PUSH-SPECIAL", "(DEFUN PUSH-SPECIAL ((VARIABLE SYMBOL) (OLDVALUEVARIABLE SYMBOL)))", ((cpp_function_code)(&push_special)), NULL);
    define_function_object("POP-SPECIAL", "(DEFUN POP-SPECIAL ())", ((cpp_function_code)(&pop_special)), NULL);
    define_function_object("LOOKUP-OLD-VALUE-VARIABLE", "(DEFUN (LOOKUP-OLD-VALUE-VARIABLE SYMBOL) ((VARIABLE SYMBOL)))", ((cpp_function_code)(&lookup_old_value_variable)), NULL);
    define_function_object("NEED-TO-UNBIND-SPECIALS?", "(DEFUN (NEED-TO-UNBIND-SPECIALS? BOOLEAN) ())", ((cpp_function_code)(&need_to_unbind_specialsP)), NULL);
    define_function_object("YIELD-SPECIAL-UNBIND-TREE", "(DEFUN (YIELD-SPECIAL-UNBIND-TREE CONS) ((NOFBINDINGS INTEGER)))", ((cpp_function_code)(&yield_special_unbind_tree)), NULL);
    define_function_object("YIELD-RETURN-SPECIAL-UNBIND-TREE", "(DEFUN (YIELD-RETURN-SPECIAL-UNBIND-TREE CONS) ())", ((cpp_function_code)(&yield_return_special_unbind_tree)), NULL);
    define_function_object("YIELD-LOOP-EXIT-SPECIAL-UNBIND-TREE", "(DEFUN (YIELD-LOOP-EXIT-SPECIAL-UNBIND-TREE CONS) ())", ((cpp_function_code)(&yield_loop_exit_special_unbind_tree)), NULL);
    define_function_object("WALK-SPECIAL-TREE", "(DEFUN (WALK-SPECIAL-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_special_tree)), NULL);
    define_function_object("WALK-CL-SPECIAL-TREE", "(DEFUN (WALK-CL-SPECIAL-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_cl_special_tree)), NULL);
    define_function_object("WALK-UNWIND-PROTECT-SPECIAL-TREE", "(DEFUN (WALK-UNWIND-PROTECT-SPECIAL-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_unwind_protect_special_tree)), NULL);
    define_function_object("WALK-CAST-TREE", "(DEFUN (WALK-CAST-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_cast_tree)), NULL);
    define_function_object("VALUES-TREE?", "(DEFUN (VALUES-TREE? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&values_treeP)), NULL);
    define_function_object("WALK-VALUES-TREE", "(DEFUN (WALK-VALUES-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_values_tree)), NULL);
    define_function_object("WALK-MV-EXPRESSION-TREE", "(DEFUN (WALK-MV-EXPRESSION-TREE OBJECT CONS (LIST OF TYPE-SPEC)) ((TREE OBJECT) (TARGETTYPES (LIST OF TYPE-SPEC)) (OPERATOR SYMBOL)))", ((cpp_function_code)(&walk_mv_expression_tree)), NULL);
    define_function_object("LISTIFY-TYPE-SPEC", "(DEFUN (LISTIFY-TYPE-SPEC (LIST OF TYPE-SPEC)) ((TYPESPEC TYPE-SPEC)))", ((cpp_function_code)(&listify_type_spec)), NULL);
    define_function_object("SLOT-FROM-EXPRESSION-TREE", "(DEFUN (SLOT-FROM-EXPRESSION-TREE SLOT) ((TREE CONS)))", ((cpp_function_code)(&slot_from_expression_tree)), NULL);
    define_function_object("WALK-MV-TREE", "(DEFUN (WALK-MV-TREE OBJECT CONS (LIST OF TYPE-SPEC)) ((TREE OBJECT)))", ((cpp_function_code)(&walk_mv_tree)), NULL);
    define_function_object("WALK-MV-VALUES-TREE", "(DEFUN (WALK-MV-VALUES-TREE CONS CONS (LIST OF TYPE-SPEC)) ((TREE CONS)))", ((cpp_function_code)(&walk_mv_values_tree)), NULL);
    define_function_object("COERCE-MV-TREE", "(DEFUN (COERCE-MV-TREE OBJECT CONS (LIST OF TYPE-SPEC)) ((TREE OBJECT) (OPERATOR SYMBOL) (SOURCETYPES (LIST OF TYPE-SPEC)) (TARGETTYPES (LIST OF TYPE-SPEC))))", ((cpp_function_code)(&coerce_mv_tree)), NULL);
  }
}

void help_startup_walk12() {
  {
    define_function_object("WALK-RETURN-AND-UNBIND-SPECIALS", "(DEFUN (WALK-RETURN-AND-UNBIND-SPECIALS CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_return_and_unbind_specials)), NULL);
    define_function_object("WALK-RETURN-TREE", "(DEFUN (WALK-RETURN-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_return_tree)), NULL);
    define_function_object("WALK-MV-SETQ-TREE", "(DEFUN (WALK-MV-SETQ-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_mv_setq_tree)), NULL);
    define_function_object("WALK-MV-BIND-TREE", "(DEFUN (WALK-MV-BIND-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_mv_bind_tree)), NULL);
    define_function_object("WALK-LOOP-TREE", "(DEFUN (WALK-LOOP-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_loop_tree)), NULL);
    define_function_object("WALK-LOOP-EXIT-TREE", "(DEFUN (WALK-LOOP-EXIT-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_loop_exit_tree)), NULL);
    define_function_object("WALK-WHILE-TREE", "(DEFUN (WALK-WHILE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_while_tree)), NULL);
    define_function_object("WALK-CONDITIONAL-TREE", "(DEFUN (WALK-CONDITIONAL-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_conditional_tree)), NULL);
    define_function_object("WALK-CHOOSE-TREE", "(DEFUN (WALK-CHOOSE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_choose_tree)), NULL);
    define_function_object("WALK-BOOLEAN-TREE", "(DEFUN (WALK-BOOLEAN-TREE OBJECT TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_boolean_tree)), NULL);
    define_function_object("AND-OR-NOT-TREE?", "(DEFUN (AND-OR-NOT-TREE? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&and_or_not_treeP)), NULL);
    define_function_object("HELP-WALK-BOOLEAN-TREE", "(DEFUN (HELP-WALK-BOOLEAN-TREE OBJECT) ((TREE CONS)))", ((cpp_function_code)(&help_walk_boolean_tree)), NULL);
    define_function_object("WALK-COND-TREE", "(DEFUN (WALK-COND-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_cond_tree)), NULL);
    define_function_object("COND-TREE-TO-IF-TREE", "(DEFUN (COND-TREE-TO-IF-TREE CONS) ((TREE CONS)))", ((cpp_function_code)(&cond_tree_to_if_tree)), NULL);
    define_function_object("YIELD-HARDCODED-CASE-SYMBOL-ID-OR-IDS", "(DEFUN (YIELD-HARDCODED-CASE-SYMBOL-ID-OR-IDS OBJECT) ((CASETEST OBJECT)))", ((cpp_function_code)(&yield_hardcoded_case_symbol_id_or_ids)), NULL);
    define_function_object("WALK-HARDCODED-SYMBOL-CASE-TREE", "(DEFUN (WALK-HARDCODED-SYMBOL-CASE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_hardcoded_symbol_case_tree)), NULL);
    define_function_object("YIELD-ISA?-CASE-TEST", "(DEFUN (YIELD-ISA?-CASE-TEST CONS) ((CASEKEY TYPE) (TESTEXPRESSION OBJECT)))", ((cpp_function_code)(&yield_isaP_case_test)), NULL);
    define_function_object("YIELD-SUBTYPE-OF?-CASE-TEST", "(DEFUN (YIELD-SUBTYPE-OF?-CASE-TEST CONS) ((CASEKEY TYPE) (TESTEXPRESSION OBJECT)))", ((cpp_function_code)(&yield_subtype_ofP_case_test)), NULL);
    define_function_object("YIELD-COND-TEST", "(DEFUN (YIELD-COND-TEST CONS) ((CASETEST OBJECT) (TESTVARIABLE OBJECT) (EQUALITYTEST SYMBOL)))", ((cpp_function_code)(&yield_cond_test)), NULL);
    define_function_object("YIELD-COND-TEST-OR-TESTS", "(DEFUN (YIELD-COND-TEST-OR-TESTS OBJECT) ((CASETEST OBJECT) (TESTVARIABLE OBJECT) (EQUALITYTEST SYMBOL)))", ((cpp_function_code)(&yield_cond_test_or_tests)), NULL);
    define_function_object("COUNT-CASE-TEST-CLAUSES", "(DEFUN (COUNT-CASE-TEST-CLAUSES INTEGER BOOLEAN) ((TREE CONS)))", ((cpp_function_code)(&count_case_test_clauses)), NULL);
    define_function_object("ATTACH-UNHANDLED-CASE-ERROR-CLAUSE", "(DEFUN (ATTACH-UNHANDLED-CASE-ERROR-CLAUSE CONS) ((CASETREE CONS) (TESTTREE OBJECT)))", ((cpp_function_code)(&attach_unhandled_case_error_clause)), NULL);
    define_function_object("WALK-NON-BUILT-IN-CASE-TREE", "(DEFUN (WALK-NON-BUILT-IN-CASE-TREE CONS TYPE-SPEC) ((TREE CONS) (EQUALITYTEST SYMBOL)))", ((cpp_function_code)(&walk_non_built_in_case_tree)), NULL);
    define_function_object("WALK-CASE-TREE", "(DEFUN (WALK-CASE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_case_tree)), NULL);
    define_function_object("WALK-TYPECASE-TREE", "(DEFUN (WALK-TYPECASE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_typecase_tree)), NULL);
    define_function_object("YIELD-CASTED-TYPECASE-CLAUSE-TREES", "(DEFUN (YIELD-CASTED-TYPECASE-CLAUSE-TREES CONS) ((TESTVARIABLE SYMBOL) (CLAUSETYPE TYPE-SPEC) (CLAUSETREES CONS)))", ((cpp_function_code)(&yield_casted_typecase_clause_trees)), NULL);
    define_function_object("VERIFY-TYPECASE-CLAUSE-TYPE", "(DEFUN VERIFY-TYPECASE-CLAUSE-TYPE ((TESTTYPE TYPE) (CLAUSETYPE TYPE) (PRECEDINGCLAUSETYPES (LIST OF TYPE)) (CLAUSE CONS)))", ((cpp_function_code)(&verify_typecase_clause_type)), NULL);
    define_function_object("FINISH-WALKING-EQL-TREE", "(DEFUN (FINISH-WALKING-EQL-TREE CONS TYPE-SPEC) ((TREE CONS) (TYPE1 TYPE) (TYPE2 TYPE)))", ((cpp_function_code)(&finish_walking_eql_tree)), NULL);
    define_function_object("FINISH-WALKING-EQUAL-TREE", "(DEFUN (FINISH-WALKING-EQUAL-TREE CONS TYPE-SPEC) ((TREE CONS) (TYPE1 TYPE) (TYPE2 TYPE)))", ((cpp_function_code)(&finish_walking_equal_tree)), NULL);
    define_function_object("WALK-EQUALITY-TREE", "(DEFUN (WALK-EQUALITY-TREE OBJECT TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_equality_tree)), NULL);
    define_function_object("MORE-GENERAL-NUMERIC-TYPE", "(DEFUN (MORE-GENERAL-NUMERIC-TYPE TYPE) ((TYPE1 TYPE) (TYPE2 TYPE)))", ((cpp_function_code)(&more_general_numeric_type)), NULL);
    define_function_object("YIELD-ARITHMETIC-OPERATOR-CALL-TREE", "(DEFUN (YIELD-ARITHMETIC-OPERATOR-CALL-TREE CONS) ((OPERATOR SYMBOL) (OPERANDS CONS)))", ((cpp_function_code)(&yield_arithmetic_operator_call_tree)), NULL);
    define_function_object("YIELD-NESTED-ARITHMETIC-OPERATOR-CALL-TREE", "(DEFUN (YIELD-NESTED-ARITHMETIC-OPERATOR-CALL-TREE CONS) ((OPERATOR SYMBOL) (OPERANDS CONS)))", ((cpp_function_code)(&yield_nested_arithmetic_operator_call_tree)), NULL);
    define_function_object("WALK-ARITHMETIC-TREE", "(DEFUN (WALK-ARITHMETIC-TREE OBJECT TYPE) ((TREE CONS)))", ((cpp_function_code)(&walk_arithmetic_tree)), NULL);
    define_function_object("WALK-QUOTED-TREE", "(DEFUN (WALK-QUOTED-TREE OBJECT TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_quoted_tree)), NULL);
    define_function_object("WALK-BQUOTE-TREE", "(DEFUN (WALK-BQUOTE-TREE OBJECT TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_bquote_tree)), NULL);
    define_function_object("EXPAND-MACRO", "(DEFUN (EXPAND-MACRO OBJECT) ((TREE CONS)))", ((cpp_function_code)(&expand_macro)), NULL);
    define_function_object("WALK-MACRO-TREE", "(DEFUN (WALK-MACRO-TREE OBJECT TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_macro_tree)), NULL);
    define_function_object("WALK-PRINT-STREAM-TREE", "(DEFUN (WALK-PRINT-STREAM-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_print_stream_tree)), NULL);
    define_function_object("WARN-ABOUT-UNDEFINED-METHODS?", "(DEFUN (WARN-ABOUT-UNDEFINED-METHODS? BOOLEAN) ())", ((cpp_function_code)(&warn_about_undefined_methodsP)), NULL);
    define_function_object("WARN-ABOUT-MISSING-METHODS?", "(DEFUN (WARN-ABOUT-MISSING-METHODS? BOOLEAN) ())", ((cpp_function_code)(&warn_about_missing_methodsP)), NULL);
    define_function_object("WALK-UNDEFINED-SLOT-TREE", "(DEFUN (WALK-UNDEFINED-SLOT-TREE CONS TYPE) ((TREE CONS) (FIRSTARGTYPE TYPE-SPEC) (WARN? BOOLEAN)))", ((cpp_function_code)(&walk_undefined_slot_tree)), NULL);
    define_function_object("WALK-SLOT-ON-ABSTRACT-TYPE-TREE", "(DEFUN (WALK-SLOT-ON-ABSTRACT-TYPE-TREE CONS TYPE-SPEC) ((TREE CONS) (ABSTRACTTYPE TYPE)))", ((cpp_function_code)(&walk_slot_on_abstract_type_tree)), NULL);
    define_function_object("FIND-PROTOTYPE-METHOD", "(DEFUN (FIND-PROTOTYPE-METHOD METHOD-SLOT) ((METHODNAME SYMBOL) (ABSTRACTTYPE TYPE)))", ((cpp_function_code)(&find_prototype_method)), NULL);
    define_function_object("COMPATIBLE-REAL-METHODS", "(DEFUN (COMPATIBLE-REAL-METHODS INTEGER INTEGER BOOLEAN) ((PROTOTYPEMETHOD METHOD-SLOT) (ABSTRACTTYPE TYPE)))", ((cpp_function_code)(&compatible_real_methods)), NULL);
    define_method_object("(DEFMETHOD (FINISH-WALKING-CALL-SLOT-TREE OBJECT TYPE-SPEC) ((SELF SLOT) (TREE CONS) (FIRSTARGTYPE TYPE-SPEC)))", ((cpp_method_code)(&Slot::finish_walking_call_slot_tree)), ((cpp_method_code)(NULL)));
    define_method_object("(DEFMETHOD (FINISH-WALKING-CALL-SLOT-TREE OBJECT TYPE-SPEC) ((SELF STORAGE-SLOT) (TREE CONS) (FIRSTARGTYPE TYPE-SPEC)))", ((cpp_method_code)(&Storage_Slot::finish_walking_call_slot_tree)), ((cpp_method_code)(NULL)));
    define_function_object("STRING-CONCATENATE-METHOD?", "(DEFUN (STRING-CONCATENATE-METHOD? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&string_concatenate_methodP)), NULL);
    define_function_object("PASS-VARIABLE-ARGUMENTS-AS-LIST?", "(DEFUN (PASS-VARIABLE-ARGUMENTS-AS-LIST? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&pass_variable_arguments_as_listP)), NULL);
    define_function_object("VARIABLE-ARGUMENTS-TYPE", "(DEFUN (VARIABLE-ARGUMENTS-TYPE TYPE-SPEC) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&variable_arguments_type)), NULL);
    define_function_object("VARIABLE-ARGUMENTS-NAME", "(DEFUN (VARIABLE-ARGUMENTS-NAME SYMBOL) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&variable_arguments_name)), NULL);
    define_function_object("YIELD-LISTIFIED-VARIABLE-ARGUMENTS-TYPE", "(DEFUN (YIELD-LISTIFIED-VARIABLE-ARGUMENTS-TYPE TYPE-SPEC) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&yield_listified_variable_arguments_type)), NULL);
    define_function_object("YIELD-LISTIFIED-VARIABLE-ARGUMENTS", "(DEFUN (YIELD-LISTIFIED-VARIABLE-ARGUMENTS CONS) ((WALKEDARGS CONS) (TARGETTYPE TYPE-SPEC) (WRAPARGS? BOOLEAN)))", ((cpp_function_code)(&yield_listified_variable_arguments)), NULL);
    define_function_object("FINISH-WALKING-ARGUMENT-LIST-TREE", "(DEFUN (FINISH-WALKING-ARGUMENT-LIST-TREE OBJECT TYPE-SPEC) ((SELF SLOT) (TREE CONS) (FIRSTARGTYPE TYPE-SPEC)))", ((cpp_function_code)(&finish_walking_argument_list_tree)), NULL);
    define_function_object("WALK-VARIABLE-ARGUMENTS", "(DEFUN (WALK-VARIABLE-ARGUMENTS CONS) ((ARGUMENTS CONS) (METHOD METHOD-SLOT)))", ((cpp_function_code)(&walk_variable_arguments)), NULL);
    define_method_object("(DEFMETHOD (FINISH-WALKING-CALL-SLOT-TREE OBJECT TYPE-SPEC) ((SELF METHOD-SLOT) (TREE CONS) (FIRSTARGTYPE TYPE-SPEC)))", ((cpp_method_code)(&Method_Slot::finish_walking_call_slot_tree)), ((cpp_method_code)(NULL)));
    define_function_object("QUOTE-ARGUMENTS", "(DEFUN (QUOTE-ARGUMENTS CONS) ((METHOD METHOD-SLOT) (ARGUMENTS CONS)))", ((cpp_function_code)(&quote_arguments)), NULL);
    define_function_object("WALK-FIRST-ARGUMENT-TO-FUNCTION", "(DEFUN (WALK-FIRST-ARGUMENT-TO-FUNCTION TYPE-SPEC) ((FNSLOT METHOD-SLOT) (TREE CONS)))", ((cpp_function_code)(&walk_first_argument_to_function)), NULL);
    define_function_object("WALK-CALL-SLOT-TREE", "(DEFUN (WALK-CALL-SLOT-TREE OBJECT TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_call_slot_tree)), NULL);
    define_function_object("INLINE-METHOD-CALL?", "(DEFUN (INLINE-METHOD-CALL? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&inline_method_callP)), NULL);
  }
}

void help_startup_walk13() {
  {
    define_function_object("METHOD-INLINABLE?", "(DEFUN (METHOD-INLINABLE? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&method_inlinableP)), NULL);
    define_function_object("MOST-SPECIFIC-METHOD?", "(DEFUN (MOST-SPECIFIC-METHOD? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&most_specific_methodP)), NULL);
    define_function_object("HELP-MOST-SPECIFIC-METHOD?", "(DEFUN (HELP-MOST-SPECIFIC-METHOD? BOOLEAN) ((CLASS CLASS) (METHOD METHOD-SLOT)))", ((cpp_function_code)(&help_most_specific_methodP)), NULL);
    define_external_slot_from_stringified_source("(DEFSLOT METHOD-SLOT CACHED-INLINABLE-METHOD-BODY :TYPE OBJECT :ALLOCATION :DYNAMIC)");
    define_function_object("INLINABLE-METHOD-BODY", "(DEFUN (INLINABLE-METHOD-BODY OBJECT) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&inlinable_method_body)), NULL);
    define_function_object("WALK-INLINE-METHOD-CALL", "(DEFUN (WALK-INLINE-METHOD-CALL OBJECT) ((METHOD METHOD-SLOT) (WALKEDARGS CONS)))", ((cpp_function_code)(&walk_inline_method_call)), NULL);
    define_function_object("YIELD-VERBATIM-INLINE-CALL-TREE", "(DEFUN (YIELD-VERBATIM-INLINE-CALL-TREE CONS) ((METHOD METHOD-SLOT) (WALKEDARGS CONS)))", ((cpp_function_code)(&yield_verbatim_inline_call_tree)), NULL);
    define_function_object("INLINE-VARIABLE-REFERENCE?", "(DEFUN (INLINE-VARIABLE-REFERENCE? BOOLEAN) ((SELF SYMBOL)) :GLOBALLY-INLINE? TRUE (RETURN (AND *INLININGMETHODCALL?* (DEFINED? (GET-LOCAL-VARIABLE-INFO SELF :INLINE-ARGUMENT)))))", ((cpp_function_code)(&inline_variable_referenceP)), NULL);
    define_function_object("WALK-INLINE-VARIABLE-REFERENCE", "(DEFUN (WALK-INLINE-VARIABLE-REFERENCE CONS TYPE-SPEC) ((SELF SYMBOL)))", ((cpp_function_code)(&walk_inline_variable_reference)), NULL);
    define_function_object("SIDE-EFFECT-FREE-EXPRESSION?", "(DEFUN (SIDE-EFFECT-FREE-EXPRESSION? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&side_effect_free_expressionP)), NULL);
    define_function_object("ESTIMATED-EVALUATION-COST", "(DEFUN (ESTIMATED-EVALUATION-COST INTEGER) ((TREE OBJECT)))", ((cpp_function_code)(&estimated_evaluation_cost)), NULL);
    define_function_object("YIELD-SYNTHESIZED-METHOD-BODY", "(DEFUN (YIELD-SYNTHESIZED-METHOD-BODY CONS) ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&yield_synthesized_method_body)), NULL);
    define_function_object("WALK-METHOD-OBJECT", "(DEFUN (WALK-METHOD-OBJECT CONS) ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_method_object)), NULL);
    define_function_object("WALK-DEFMETHOD-TREE", "(DEFUN WALK-DEFMETHOD-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_defmethod_tree)), NULL);
    define_external_slot_from_stringified_source("(DEFSLOT METHOD-SLOT FORWARD-DECLARATION? :TYPE BOOLEAN :DEFAULT FALSE :DOCUMENTATION \"Set to `true' for walked methods without a body.\nUsed to flag functions and methods that were declared but not defined.\" :ALLOCATION :DYNAMIC)");
    define_function_object("WALK-METHOD-UNIT", "(DEFUN WALK-METHOD-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_method_unit)), NULL);
    define_function_object("PRINT-UNDEFINED-METHODS", "(DEFUN PRINT-UNDEFINED-METHODS ((MODULE MODULE) (LOCAL? BOOLEAN)) :PUBLIC? TRUE :DOCUMENTATION \"Print all declared but not yet defined functions and methods\nin `module'.  If `local?' is true, do not consider any parent modules of\n`module'.  If `module' is NULL, look at all modules in the system.  This\nis handy to pinpoint forward declarations that haven't been followed up\nby actual definitions.\")", ((cpp_function_code)(&print_undefined_methods)), NULL);
    define_function_object("WALK-SYS-CALL-TREE", "(DEFUN (WALK-SYS-CALL-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_sys_call_tree)), NULL);
    define_function_object("WALK-DEFCLASS-TREE", "(DEFUN WALK-DEFCLASS-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_defclass_tree)), NULL);
    define_function_object("YIELD-INITIALIZE-HARDWIRED-SLOTS", "(DEFUN (YIELD-INITIALIZE-HARDWIRED-SLOTS CONS) ((CLASS CLASS)))", ((cpp_function_code)(&yield_initialize_hardwired_slots)), NULL);
    define_function_object("YIELD-INITIAL-VALUE-EXPRESSION", "(DEFUN (YIELD-INITIAL-VALUE-EXPRESSION OBJECT) ((SLOT STORAGE-SLOT)))", ((cpp_function_code)(&yield_initial_value_expression)), NULL);
    define_function_object("YIELD-CONSTRUCTOR-ATTACHMENT", "(DEFUN (YIELD-CONSTRUCTOR-ATTACHMENT CONS) ((CLASS CLASS) (CLASSREF SYMBOL)))", ((cpp_function_code)(&yield_constructor_attachment)), NULL);
    define_external_slot_from_stringified_source("(DEFSLOT CLASS CLASS-AUXILIARY-METHODS :TYPE (LIST OF METHOD-SLOT) :DEFAULT NIL-LIST :DOCUMENTATION \"Stores auxiliary method objects such as slot accessors and\ninherited mixin methods that are not attached to the class as regular methods\nbut need to be accessible by the C++ translator in order to generate the\nnecessary signatures.\" :ALLOCATION :DYNAMIC)");
    define_external_slot_from_stringified_source("(DEFSLOT METHOD-SLOT AUXILIARY-METHOD-REASON :TYPE SLOT :ALLOCATION :DYNAMIC)");
    define_function_object("CLEANUP-AUXILIARY-METHODS", "(DEFUN CLEANUP-AUXILIARY-METHODS ((CLASS CLASS)))", ((cpp_function_code)(&cleanup_auxiliary_methods)), NULL);
    define_function_object("REGISTER-AUXILIARY-METHOD", "(DEFUN REGISTER-AUXILIARY-METHOD ((METHOD METHOD-SLOT) (REASON SLOT)))", ((cpp_function_code)(&register_auxiliary_method)), NULL);
    define_function_object("MIXIN-TYPE-VARIABLES-ALLOWED?", "(DEFUN (MIXIN-TYPE-VARIABLES-ALLOWED? BOOLEAN) () :GLOBALLY-INLINE? TRUE (RETURN (OR (NOT (TRANSLATE-TO-SINGLE-INHERITANCE-LANGUAGE?)) (NOT (EQL? *MIXIN-IMPLEMENTATION-STYLE* :SECOND-CLASS)))))", ((cpp_function_code)(&mixin_type_variables_allowedP)), NULL);
    define_function_object("CREATE-NATIVE-MIXIN-SLOT-ACCESSORS?", "(DEFUN (CREATE-NATIVE-MIXIN-SLOT-ACCESSORS? BOOLEAN) () :GLOBALLY-INLINE? TRUE (RETURN (AND (TRANSLATE-TO-SINGLE-INHERITANCE-LANGUAGE?) (MIXIN-TYPE-VARIABLES-ALLOWED?))))", ((cpp_function_code)(&create_native_mixin_slot_accessorsP)), NULL);
    define_function_object("CREATE-ACCESSOR-UNITS-FOR-SLOT", "(DEFUN CREATE-ACCESSOR-UNITS-FOR-SLOT ((SLOT STORAGE-SLOT) (CLASS CLASS) (MIXINACCESSORS? BOOLEAN) (SIGNATURESONLY? BOOLEAN)))", ((cpp_function_code)(&create_accessor_units_for_slot)), NULL);
    define_function_object("CREATE-ACCESSOR-UNITS-FOR-MIXIN-SLOT", "(DEFUN CREATE-ACCESSOR-UNITS-FOR-MIXIN-SLOT ((SLOT STORAGE-SLOT) (CLASS CLASS)))", ((cpp_function_code)(&create_accessor_units_for_mixin_slot)), NULL);
    define_function_object("CREATE-SLOT-ACCESSOR-UNITS", "(DEFUN CREATE-SLOT-ACCESSOR-UNITS ((CLASS CLASS)))", ((cpp_function_code)(&create_slot_accessor_units)), NULL);
    define_function_object("CREATE-INLINE-METHOD-UNITS", "(DEFUN CREATE-INLINE-METHOD-UNITS ((CLASS CLASS)))", ((cpp_function_code)(&create_inline_method_units)), NULL);
    define_function_object("YIELD-MIXIN-METHOD-TREE", "(DEFUN (YIELD-MIXIN-METHOD-TREE CONS) ((METHOD METHOD-SLOT) (CLASS CLASS)))", ((cpp_function_code)(&yield_mixin_method_tree)), NULL);
    define_function_object("CREATE-MIXIN-METHOD-UNITS", "(DEFUN CREATE-MIXIN-METHOD-UNITS ((CLASS CLASS)))", ((cpp_function_code)(&create_mixin_method_units)), NULL);
    define_function_object("CREATE-DEFPRINT-UNIT", "(DEFUN CREATE-DEFPRINT-UNIT ((CLASS CLASS)))", ((cpp_function_code)(&create_defprint_unit)), NULL);
    define_function_object("WALK-CLASS-UNIT", "(DEFUN WALK-CLASS-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_class_unit)), NULL);
    define_function_object("CREATE-FINALIZATION-UNITS", "(DEFUN CREATE-FINALIZATION-UNITS ())", ((cpp_function_code)(&create_finalization_units)), NULL);
    define_function_object("WALK-DEFSLOT-TREE", "(DEFUN WALK-DEFSLOT-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_defslot_tree)), NULL);
    define_function_object("WALK-SLOT-UNIT", "(DEFUN WALK-SLOT-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_slot_unit)), NULL);
    define_function_object("LOOKUP-MACRO", "(DEFUN (LOOKUP-MACRO METHOD-SLOT) ((NAME SYMBOL)) :PUBLIC? TRUE)", ((cpp_function_code)(&lookup_macro)), NULL);
    define_function_object("STORE-MACRO", "(DEFUN STORE-MACRO ((NAME SYMBOL) (EXPANDERNAME SYMBOL) (EXPANDERCODE FUNCTION-CODE)) :PUBLIC? TRUE)", ((cpp_function_code)(&store_macro)), NULL);
    define_function_object("UNDEFINE-MACRO", "(DEFUN UNDEFINE-MACRO ((NAME SYMBOL)))", ((cpp_function_code)(&undefine_macro)), NULL);
    define_function_object("YIELD-MACRO-EXPANDER-FUNCTION-NAME", "(DEFUN (YIELD-MACRO-EXPANDER-FUNCTION-NAME SYMBOL) ((MACRONAME SYMBOL)))", ((cpp_function_code)(&yield_macro_expander_function_name)), NULL);
    define_function_object("WALK-DEFMACRO-TREE", "(DEFUN WALK-DEFMACRO-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_defmacro_tree)), NULL);
    define_function_object("WALK-MACRO-UNIT", "(DEFUN WALK-MACRO-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_macro_unit)), NULL);
    define_function_object("WALK-DEFGLOBAL-TREE", "(DEFUN WALK-DEFGLOBAL-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_defglobal_tree)), NULL);
    define_function_object("WALK-GLOBAL-UNIT", "(DEFUN WALK-GLOBAL-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_global_unit)), NULL);
    define_function_object("WALK-DEFTYPE-TREE", "(DEFUN WALK-DEFTYPE-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_deftype_tree)), NULL);
    define_function_object("WALK-TYPE-UNIT", "(DEFUN WALK-TYPE-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_type_unit)), NULL);
    define_function_object("WALK-STARTUP-TIME-PROGN-TREE", "(DEFUN WALK-STARTUP-TIME-PROGN-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_startup_time_progn_tree)), NULL);
    define_function_object("EXTRACT-STARTUP-TIME-PHASE", "(DEFUN (EXTRACT-STARTUP-TIME-PHASE KEYWORD) ((TREE CONS)))", ((cpp_function_code)(&extract_startup_time_phase)), NULL);
    define_function_object("WALK-STARTUP-TIME-PROGN-UNIT", "(DEFUN WALK-STARTUP-TIME-PROGN-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_startup_time_progn_unit)), NULL);
    define_function_object("EARLIER-STARTUP-UNIT?", "(DEFUN (EARLIER-STARTUP-UNIT? BOOLEAN) ((UNIT1 TRANSLATION-UNIT) (UNIT2 TRANSLATION-UNIT)))", ((cpp_function_code)(&earlier_startup_unitP)), NULL);
    define_function_object("COMBINE-STARTUP-FUNCTION-UNITS", "(DEFUN (COMBINE-STARTUP-FUNCTION-UNITS CONS) ((STARTUPFNNAME SYMBOL)))", ((cpp_function_code)(&combine_startup_function_units)), NULL);
    define_external_slot_from_stringified_source("(DEFSLOT METHOD-SLOT METHOD-STARTUP-CLASSNAME :TYPE STRING :ALLOCATION :DYNAMIC)");
    define_method_object("(DEFMETHOD (METHOD-STARTUP-FUNCTION? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_method_code)(&Method_Slot::method_startup_functionP)), ((cpp_method_code)(NULL)));
    define_function_object("EXTRACT-STARTUP-FUNCTION-UNITS", "(DEFUN EXTRACT-STARTUP-FUNCTION-UNITS ((STARTUPUNITS (LIST OF TRANSLATION-UNIT)) (STARTUPFNNAME SYMBOL)))", ((cpp_function_code)(&extract_startup_function_units)), NULL);
    define_function_object("YIELD-STARTUP-FUNCTION-NAME", "(DEFUN (YIELD-STARTUP-FUNCTION-NAME SYMBOL) ((FILE STRING)))", ((cpp_function_code)(&yield_startup_function_name)), NULL);
    define_function_object("CREATE-STARTUP-FUNCTION-UNITS", "(DEFUN CREATE-STARTUP-FUNCTION-UNITS ())", ((cpp_function_code)(&create_startup_function_units)), NULL);
    define_function_object("DONT-GENERATE-STARTUP-CODE?", "(DEFUN (DONT-GENERATE-STARTUP-CODE? BOOLEAN) ())", ((cpp_function_code)(&dont_generate_startup_codeP)), NULL);
  }
}

void help_startup_walk14() {
  {
    define_function_object("REMOVE-ALL-STARTUP-TIME-PROGN-UNITS", "(DEFUN REMOVE-ALL-STARTUP-TIME-PROGN-UNITS ())", ((cpp_function_code)(&remove_all_startup_time_progn_units)), NULL);
    define_function_object("WALK-SIGNAL-EXCEPTION-TREE", "(DEFUN (WALK-SIGNAL-EXCEPTION-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_signal_exception_tree)), NULL);
    define_function_object("WALK-EXCEPTION-CASE-TREE", "(DEFUN (WALK-EXCEPTION-CASE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_exception_case_tree)), NULL);
    define_function_object("WALK-EXCEPTION-HANDLER-TREES", "(DEFUN (WALK-EXCEPTION-HANDLER-TREES CONS) ((TREE CONS)))", ((cpp_function_code)(&walk_exception_handler_trees)), NULL);
    define_function_object("WALK-UNWIND-PROTECT-TREE", "(DEFUN (WALK-UNWIND-PROTECT-TREE CONS TYPE) ((TREE CONS)))", ((cpp_function_code)(&walk_unwind_protect_tree)), NULL);
    define_function_object("WALK-VERBATIM-DEFINITION-TREE", "(DEFUN WALK-VERBATIM-DEFINITION-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_verbatim_definition_tree)), NULL);
    define_function_object("WALK-VERBATIM-UNIT", "(DEFUN WALK-VERBATIM-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_verbatim_unit)), NULL);
    define_function_object("WALK-VERBATIM-TREE", "(DEFUN (WALK-VERBATIM-TREE OBJECT TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_verbatim_tree)), NULL);
    define_function_object("LOOKUP-VERBATIM-TREE", "(DEFUN (LOOKUP-VERBATIM-TREE OBJECT) ((TREE CONS) (LANGUAGE KEYWORD)))", ((cpp_function_code)(&lookup_verbatim_tree)), NULL);
    define_function_object("VERBATIM-TREE?", "(DEFUN (VERBATIM-TREE? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&verbatim_treeP)), NULL);
    define_function_object("WALK-DEFMODULE-TREE", "(DEFUN WALK-DEFMODULE-TREE ((TREE CONS)))", ((cpp_function_code)(&walk_defmodule_tree)), NULL);
    define_function_object("WALK-MODULE-UNIT", "(DEFUN WALK-MODULE-UNIT ((UNIT TRANSLATION-UNIT)))", ((cpp_function_code)(&walk_module_unit)), NULL);
    define_function_object("EXTRACT-REQUIRED-ARGUMENT-VALUES", "(DEFUN (EXTRACT-REQUIRED-ARGUMENT-VALUES CONS) ((CLASSTYPE TYPE-SPEC) (SLOTSANDVALUES (PROPERTY-LIST OF SLOT OBJECT))))", ((cpp_function_code)(&extract_required_argument_values)), NULL);
    define_function_object("PREPROCESS-ARRAY-ARGUMENTS", "(DEFUN PREPROCESS-ARRAY-ARGUMENTS ((ARRAYTYPE PARAMETRIC-TYPE-SPECIFIER) (SLOTSANDVALUES (PROPERTY-LIST OF SLOT OBJECT))))", ((cpp_function_code)(&preprocess_array_arguments)), NULL);
    define_function_object("EVALUATE-ARRAY-ARGUMENT-VALUE", "(DEFUN EVALUATE-ARRAY-ARGUMENT-VALUE ((ARRAYTYPE PARAMETRIC-TYPE-SPECIFIER) (REQUIREDVALUES LIST) (SLOT SLOT) (VALUEREF OBJECT)))", ((cpp_function_code)(&evaluate_array_argument_value)), NULL);
    define_function_object("YIELD-NEW-ARGUMENTS-TREE", "(DEFUN (YIELD-NEW-ARGUMENTS-TREE CONS CONS) ((KEYWORDSANDVALUES CONS) (CLASSTYPE TYPE-SPEC) (SELFVARIABLE SYMBOL)))", ((cpp_function_code)(&yield_new_arguments_tree)), NULL);
    define_function_object("WALK-NEW-TREE", "(DEFUN (WALK-NEW-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_new_tree)), NULL);
    define_function_object("WALK-MAKE-TREE", "(DEFUN (WALK-MAKE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_make_tree)), NULL);
    define_function_object("WALK-FUNCALL-TREE", "(DEFUN (WALK-FUNCALL-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_funcall_tree)), NULL);
    define_function_object("WALK-THE-CODE-TREE", "(DEFUN (WALK-THE-CODE-TREE CONS TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&walk_the_code_tree)), NULL);
    define_function_object("DE-UGLIFY-ARGUMENTS", "(DEFUN (DE-UGLIFY-ARGUMENTS CONS) ((UGLYARGUMENTS CONS)))", ((cpp_function_code)(&de_uglify_arguments)), NULL);
    define_function_object("DE-UGLIFY-PARSE-TREE", "(DEFUN (DE-UGLIFY-PARSE-TREE OBJECT) ((TREE OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&de_uglify_parse_tree)), NULL);
    define_function_object("REGISTER-NATIVE-NAME", "(DEFUN REGISTER-NATIVE-NAME ((NAME SYMBOL) (LANGUAGE KEYWORD) (CATEGORY KEYWORD)))", ((cpp_function_code)(&register_native_name)), NULL);
    define_function_object("NATIVE-NAME?", "(DEFUN (NATIVE-NAME? BOOLEAN) ((NAME SYMBOL) (LANGUAGE KEYWORD) (CATEGORY KEYWORD)))", ((cpp_function_code)(&native_nameP)), NULL);
    define_function_object("YIELD-RENAMED-NAME-IF-NATIVE", "(DEFUN (YIELD-RENAMED-NAME-IF-NATIVE SYMBOL) ((NAME SYMBOL) (LANGUAGE KEYWORD) (CATEGORY KEYWORD)))", ((cpp_function_code)(&yield_renamed_name_if_native)), NULL);
    define_function_object("METHOD-NEEDS-LISP-MACRO?", "(DEFUN (METHOD-NEEDS-LISP-MACRO? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&method_needs_lisp_macroP)), NULL);
    define_function_object("YIELD-LISP-MACRO-TREES", "(DEFUN (YIELD-LISP-MACRO-TREES CONS CONS) ((NAME SYMBOL) (METHOD METHOD-SLOT)))", ((cpp_function_code)(&yield_lisp_macro_trees)), NULL);
    define_function_object("CREATE-LISP-MACRO-UNITS", "(DEFUN CREATE-LISP-MACRO-UNITS ((NAME SYMBOL) (METHOD METHOD-SLOT)))", ((cpp_function_code)(&create_lisp_macro_units)), NULL);
    define_function_object("COMMAND?", "(DEFUN (COMMAND? BOOLEAN) ((METHOD METHOD-SLOT)) :DOCUMENTATION \"Return `true' if `method' is an evaluable command.\" :PUBLIC? TRUE)", ((cpp_function_code)(&commandP)), NULL);
    define_function_object("LOOKUP-COMMAND", "(DEFUN (LOOKUP-COMMAND METHOD-SLOT) ((NAME SYMBOL)) :DOCUMENTATION \"If `name' names an evaluable command return its associated\ncommand object;  otherwise, return `null'.  Currently, commands are not\npolymorphic, i.e., they can only be implemented by functions.\" :PUBLIC? TRUE)", ((cpp_function_code)(&lookup_command)), NULL);
    define_function_object("METHOD-MUST-BE-EVALUABLE?", "(DEFUN (METHOD-MUST-BE-EVALUABLE? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&method_must_be_evaluableP)), NULL);
    define_function_object("METHOD-NEEDS-EVALUATOR-WRAPPER?", "(DEFUN (METHOD-NEEDS-EVALUATOR-WRAPPER? BOOLEAN) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&method_needs_evaluator_wrapperP)), NULL);
    define_function_object("YIELD-EVALUATOR-WRAPPER-NAME", "(DEFUN (YIELD-EVALUATOR-WRAPPER-NAME SYMBOL) ((METHODNAME SYMBOL)))", ((cpp_function_code)(&yield_evaluator_wrapper_name)), NULL);
    define_function_object("YIELD-ARGUMENT-ACCESS-TREE", "(DEFUN (YIELD-ARGUMENT-ACCESS-TREE OBJECT) ((ARGUMENTSVARIABLE SYMBOL) (INDEX INTEGER) (RESTARGUMENT? BOOLEAN)))", ((cpp_function_code)(&yield_argument_access_tree)), NULL);
    define_function_object("YIELD-EVALUATOR-WRAPPER-TREE", "(DEFUN (YIELD-EVALUATOR-WRAPPER-TREE CONS) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&yield_evaluator_wrapper_tree)), NULL);
    define_function_object("CREATE-EVALUATOR-WRAPPER-UNIT", "(DEFUN (CREATE-EVALUATOR-WRAPPER-UNIT METHOD-SLOT) ((METHOD METHOD-SLOT)))", ((cpp_function_code)(&create_evaluator_wrapper_unit)), NULL);
    define_function_object("EVALUATE", "(DEFUN (EVALUATE OBJECT) ((EXPRESSION OBJECT)) :DOCUMENTATION \"Evaluate the expression `expression' and return the result.\nCurrently, only the evaluation of (possibly nested) commands and global\nvariables is supported.  The second return value indicates the actual type \nof the result (which might have been wrapped), and the third return value\nindicates whether an error occurred during the evaluation.  Expressions\nare simple to program in Common Lisp, since they are built into the language,\nand relatively awkward in Java and C++.  Users of either of those \nlanguages are more likely to want to call `evaluate-string'.\" :LISP-MACRO? TRUE :PUBLIC? TRUE)", ((cpp_function_code)(&evaluate)), NULL);
    define_function_object("EVALUATE-STRING", "(DEFUN (EVALUATE-STRING OBJECT) ((EXPRESSION STRING)) :DOCUMENTATION \"Evaluate the expression represented by `expression' and return the result.\nThis is equivalent to '(evaluate (unstringify expression))'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&evaluate_string)), NULL);
    define_function_object("TRY-TO-EVALUATE", "(DEFUN (TRY-TO-EVALUATE OBJECT) ((TREE OBJECT)) :DOCUMENTATION \"Variant of `evaluate' that only evaluates `tree' if it\nrepresents an evaluable expression.  If it does not, `tree' is returned\nunmodified.  This can be used to implement commands with mixed argument\nevaluation strategies.\" :PUBLIC? TRUE)", ((cpp_function_code)(&try_to_evaluate)), NULL);
    define_function_object("EVALUATE-CONS-TREE", "(DEFUN (EVALUATE-CONS-TREE OBJECT TYPE-SPEC) ((TREE CONS)))", ((cpp_function_code)(&evaluate_cons_tree)), NULL);
    define_function_object("EVALUATE-ARGUMENT-TREE", "(DEFUN (EVALUATE-ARGUMENT-TREE OBJECT TYPE-SPEC) ((TREE OBJECT) (EVALUATE? BOOLEAN)))", ((cpp_function_code)(&evaluate_argument_tree)), NULL);
    define_function_object("EVALUATE-ATOMIC-TREE", "(DEFUN (EVALUATE-ATOMIC-TREE OBJECT TYPE-SPEC) ((TREE OBJECT)))", ((cpp_function_code)(&evaluate_atomic_tree)), NULL);
    define_function_object("COERCE-EVALUATED-TREE", "(DEFUN (COERCE-EVALUATED-TREE OBJECT TYPE-SPEC) ((TREE OBJECT) (SOURCETREE OBJECT) (SOURCETYPE TYPE-SPEC) (TARGETTYPE TYPE-SPEC) (EVALUATE? BOOLEAN)))", ((cpp_function_code)(&coerce_evaluated_tree)), NULL);
    define_function_object("STARTUP-WALK", "(DEFUN STARTUP-WALK () :PUBLIC? TRUE)", ((cpp_function_code)(&startup_walk)), NULL);
    { Method_Slot* function = lookup_function(SYM_WALK_STELLA_STARTUP_WALK);

      set_dynamic_slot_value(function->dynamic_slots, SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string("Startup-Walk"), NULL_STRING_WRAPPER);
    }
  }
}

void startup_walk() {
  if (current_startup_time_phaseP(0)) {
    oQUOTATION_TABLEo = new_string_hash_table();
  }
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oSTELLA_MODULEo);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (current_startup_time_phaseP(2)) {
      help_startup_walk1();
      help_startup_walk2();
      help_startup_walk3();
      help_startup_walk4();
      help_startup_walk5();
      help_startup_walk6();
      help_startup_walk7();
    }
    if (current_startup_time_phaseP(4)) {
      help_startup_walk8();
    }
    if (current_startup_time_phaseP(5)) {
      { Class* renamed_Class = define_class_from_stringified_source("TRANSLATION-UNIT", "(DEFCLASS TRANSLATION-UNIT (STANDARD-OBJECT) :PUBLIC? TRUE :PUBLIC-SLOTS ((TU-HOME-MODULE :TYPE MODULE :OPTION-KEYWORD :TU-HOME-MODULE) (THE-OBJECT :TYPE OBJECT) (CATEGORY :TYPE SYMBOL) (ANNOTATION :TYPE STRING) (AUXILIARY? :TYPE BOOLEAN) (CODE-REGISTER :TYPE OBJECT) (TRANSLATION :TYPE OBJECT) (REFERENCED-GLOBALS :TYPE (LIST OF GLOBAL-VARIABLE) :ALLOCATION :EMBEDDED :COMPONENT? TRUE)) :PUBLIC-METHODS ((HOME-MODULE ((SELF TRANSLATION-UNIT)) :TYPE MODULE (RETURN (TU-HOME-MODULE SELF)))) :PRINT-FORM (PRINT-TRANSLATION-UNIT SELF STREAM))");

        renamed_Class->class_constructor_code = ((cpp_function_code)(&new_translation_unit));
        renamed_Class->class_slot_accessor_code = ((cpp_function_code)(&access_translation_unit_slot_value));
      }
    }
    if (current_startup_time_phaseP(6)) {
      finalize_classes();
    }
    if (current_startup_time_phaseP(7)) {
      help_startup_walk9();
      help_startup_walk10();
      help_startup_walk11();
      help_startup_walk12();
      help_startup_walk13();
      help_startup_walk14();
    }
    if (current_startup_time_phaseP(8)) {
      finalize_slots();
      cleanup_unfinalized_classes();
    }
    if (current_startup_time_phaseP(9)) {
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *AVAILABLE-ENVIRONMENT-FEATURES* (LIST OF KEYWORD) (CAST (LIST :WARN-ABOUT-UNDEFINED-METHODS :SUPPRESS-WARNINGS :WARN-ABOUT-MISSING-METHODS :USE-HARDCODED-SYMBOLS :USE-COMMON-LISP-STRUCTS :USE-COMMON-LISP-VECTOR-STRUCTS :USE-COMMON-LISP-CONSES :USE-CPP-GARBAGE-COLLECTOR :TRANSLATE-WITH-COPYRIGHT-HEADER) (LIST OF KEYWORD)) :DOCUMENTATION \"List of Stella environment features.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *CURRENT-ENVIRONMENT-FEATURES* (LIST OF KEYWORD) (NEW (LIST OF KEYWORD)) :DOCUMENTATION \"List of currently enabled environment features.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *DEFAULT-ENVIRONMENT-FEATURES* (LIST OF KEYWORD) (CAST (LIST :WARN-ABOUT-UNDEFINED-METHODS :WARN-ABOUT-MISSING-METHODS :USE-CPP-GARBAGE-COLLECTOR :USE-COMMON-LISP-CONSES) (LIST OF KEYWORD)) :DOCUMENTATION \"List of features enabled upon call to\n`reset-environment-features'.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *TRACED-KEYWORDS* (LIST OF KEYWORD) NULL :PUBLIC? TRUE)");
      store_macro(SYM_WALK_STELLA_TRACE_IF, SYM_WALK_STELLA_TRACE_IF_EXPANDER, ((cpp_function_code)(&trace_if_expander)));
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *SAFETY* INTEGER 3 :DOCUMENTATION \"Integer between 0 and 3.  Higher levels call more\nsafety checks.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *DEBUGLEVEL* INTEGER 3 :DOCUMENTATION \"Integer between 0 and 3.  Higher levels generate more\ncode to aid debugging.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *OPTIMIZESPEEDLEVEL* INTEGER 3 :DOCUMENTATION \"Integer between 0 and 3.  Higher levels optimize for\ngreater execution speed.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *OPTIMIZESPACELEVEL* INTEGER 3 :DOCUMENTATION \"Integer between 0 and 3.  Higher levels optimize for\nless code size and memory consumption.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *QUOTATION-TABLE* (STRING-HASH-TABLE OF STRING CONS) NULL :DOCUMENTATION \"Stores quoted cons trees.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *TRANSLATIONUNITS* (LIST OF TRANSLATION-UNIT) NULL :PUBLIC? TRUE :DOCUMENTATION \"List of objects representing partially walked\ntop-level definitions and auxiliary code.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *CURRENTTRANSLATIONUNIT* TRANSLATION-UNIT NULL :PUBLIC? TRUE :DOCUMENTATION \"The translation unit currently operated on.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *TRANSLATIONPHASE* KEYWORD NULL :PUBLIC? TRUE :DOCUMENTATION \"Indicates the current translation phase which is one of\n:DEFINE, :FINALIZE, :WALK, or :TRANSLATE.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *TRANSLATIONVERBOSITYLEVEL* INTEGER 1 :PUBLIC? TRUE :DOCUMENTATION \"The higher the level, the more progress annotations are\ngenerated during the translation of Stella declarations.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *USEHARDCODEDSYMBOLS?* BOOLEAN FALSE)");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *TRANSLATOROUTPUTLANGUAGE* KEYWORD :COMMON-LISP :DOCUMENTATION \"Specifies the current translator output language; either\n:common-lisp, :idl, :java, :cpp, or :cpp-standalone.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *LOCALVARIABLETYPETABLE* (KEY-VALUE-LIST OF SYMBOL STANDARD-OBJECT) NULL :DOCUMENTATION \"Table mapping local variable names their declared types\n(declared explicitly or implicitly).\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *METHODBEINGWALKED* METHOD-SLOT NULL :DOCUMENTATION \"Contains the method or function being walked, or else `null'.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *FOUNDRETURN?* BOOLEAN FALSE :DOCUMENTATION \"Indicates that one or more return statements have been found\nduring the walk of the current method.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *TARGETTYPE* TYPE-SPEC @VOID :DOCUMENTATION \"Bound to the target type for an expression currently walked.\nUsed instead of an extra argument to `walk-a-tree', since only a few types\nof expressions need to know about their expected type (e.g., FUNCALL).\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *LOCALGENSYMTABLE* KEY-VALUE-LIST NULL :DOCUMENTATION \"Table that maps each prefix of a function-local gensym\nto its own gensym counter and/or to related gensyms.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *TRANSLATIONERRORS* INTEGER 0 :PUBLIC? TRUE)");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *TRANSLATIONWARNINGS* INTEGER 0 :PUBLIC? TRUE)");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *IGNORETRANSLATIONERRORS?* BOOLEAN TRUE :DOCUMENTATION \"If `true' all passes of a translation will always be performed\nregardless of whether any errors were encountered.  Otherwise, translation\nends with the first pass that encountered an error.\" :PUBLIC? TRUE)");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *FUNCTION-CALL-LOG-STREAM* OUTPUT-STREAM NULL :DOCUMENTATION \"The current log file to which function calls should be logged.\nA non-NULL value indicates that function call logging is enabled.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *LOG-FUNCTION-CALLS?* BOOLEAN FALSE :DOCUMENTATION \"Translation switch which indicates that methods should\nbe instrumented to log their calls to a file.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *LOG-BREAK-POINT-COUNTER* INTEGER NULL)");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *WRAPPED-TYPE-TABLE* (CONS OF CONS) (BQUOTE ((@INTEGER-WRAPPER @INTEGER) (@FLOAT-WRAPPER @FLOAT) (@NUMBER-WRAPPER @NUMBER) (@STRING-WRAPPER @STRING) (@MUTABLE-STRING-WRAPPER @MUTABLE-STRING) (@CHARACTER-WRAPPER @CHARACTER) (@BOOLEAN-WRAPPER @BOOLEAN) (@FUNCTION-CODE-WRAPPER @FUNCTION-CODE) (@METHOD-CODE-WRAPPER @METHOD-CODE))) :DOCUMENTATION \"Table of pairs used by `wrapper-value-type' and\n`type-to-wrapped-type'.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *COERSION-TABLE* (CONS OF CONS) (BQUOTE ((@BOOLEAN @BOOLEAN-WRAPPER INLINE-WRAP-BOOLEAN) (@INTEGER @INTEGER-WRAPPER WRAP-LITERAL) (@INTEGER @BOOLEAN-WRAPPER INTEGER-TO-BOOLEAN-WRAPPER) (@INTEGER @BOOLEAN INTEGER-TO-BOOLEAN) (@FLOAT @FLOAT-WRAPPER WRAP-LITERAL) (@MUTABLE-STRING @STRING MUTABLE-STRING-TO-STRING) (@MUTABLE-STRING @MUTABLE-STRING-WRAPPER WRAP-LITERAL) (@STRING @STRING-WRAPPER WRAP-LITERAL) (@STRING @MUTABLE-STRING STRING-TO-MUTABLE-STRING) (@STRING @SYMBOL INTERN-SYMBOL) (@CHARACTER @CHARACTER-WRAPPER WRAP-LITERAL) (@CHARACTER @STRING CHARACTER-TO-STRING) (@FUNCTION-CODE @FUNCTION-CODE-WRAPPER WRAP-LITERAL) (@METHOD-CODE @METHOD-CODE-WRAPPER WRAP-LITERAL) (@SYMBOL @STRING SYMBOL-NAME) (@BOOLEAN-WRAPPER @BOOLEAN INLINE-UNWRAP-BOOLEAN) (@INTEGER-WRAPPER @INTEGER WRAPPER-VALUE) (@FLOAT-WRAPPER @FLOAT WRAPPER-VALUE) (@NUMBER-WRAPPER @FLOAT NUMBER-WRAPPER-TO-FLOAT) (@STRING-WRAPPER @STRING WRAPPER-VALUE) (@MUTABLE-STRING-WRAPPER @MUTABLE-STRING WRAPPER-VALUE) (@CHARACTER-WRAPPER @CHARACTER WRAPPER-VALUE) (@FUNCTION-CODE-WRAPPER @FUNCTION-CODE WRAPPER-VALUE) (@METHOD-CODE-WRAPPER @METHOD-CODE WRAPPER-VALUE) (@SURROGATE @CLASS SURROGATE-VALUE) (@SURROGATE @MODULE SURROGATE-VALUE) (@INPUT-STREAM @NATIVE-INPUT-STREAM NATIVE-STREAM) (@OUTPUT-STREAM @NATIVE-OUTPUT-STREAM NATIVE-STREAM) (@NUMBER @INTEGER (CAST <X> @INTEGER)) (@NUMBER @FLOAT (CAST <X> @FLOAT)) (@INTEGER @FLOAT (CAST <X> @FLOAT)) (@INTEGER @SINGLE-FLOAT (CAST <X> @SINGLE-FLOAT)) (@FLOAT @SINGLE-FLOAT IDENTITY) (@FLOAT @DOUBLE-FLOAT IDENTITY) (@INTEGER @SHORT-INTEGER IDENTITY) (@INTEGER @LONG-INTEGER IDENTITY) (@INTEGER @UNSIGNED-SHORT-INTEGER IDENTITY) (@INTEGER @UNSIGNED-LONG-INTEGER IDENTITY))) :DOCUMENTATION \"Table of triples used by `lookup-coersion-method' to\nlocate a coersion method.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *SYMBOL-REGISTRY* (HASH-TABLE OF GENERALIZED-SYMBOL GENERALIZED-SYMBOL) (NEW (HASH-TABLE OF GENERALIZED-SYMBOL GENERALIZED-SYMBOL)) :PUBLIC? TRUE)");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *SYMBOL-SET* (LIST OF GENERALIZED-SYMBOL) (NEW (LIST OF GENERALIZED-SYMBOL)))");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *CURRENTFILE* STRING NULL :PUBLIC? TRUE :DOCUMENTATION \"Name of file that is currently being translated.\nA NULL value indicates an incremental translation.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *SPECIALVARIABLESTACK* (KEY-VALUE-LIST OF SYMBOL SYMBOL) (NEW (KEY-VALUE-LIST OF SYMBOL SYMBOL)) :DOCUMENTATION \"Stack mirroring the current state of bound specials\nwith their associated old-value variables.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *SPECIALSENABLED?* BOOLEAN TRUE :DOCUMENTATION \"`true' if using specials is enabled and legal.\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *NOFSPECIALSATLOOPENTRY* INTEGER 0 :DOCUMENTATION \"Number of specials bound at the most recent entry\nto a LOOP/WHILE/FOREACH construct.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *TYPE-PREDICATE-TABLE* (CONS OF CONS) (BQUOTE ((@BOOLEAN BOOLEAN? SUBTYPE-OF-BOOLEAN?) (@INTEGER INTEGER? SUBTYPE-OF-INTEGER?) (@FLOAT FLOAT? SUBTYPE-OF-FLOAT?) (@STRING STRING? SUBTYPE-OF-STRING?) (@CHARACTER CHARACTER? SUBTYPE-OF-CHARACTER?) (@WRAPPER WRAPPER? SUBTYPE-OF-WRAPPER?) (@BOOLEAN-WRAPPER BOOLEAN? SUBTYPE-OF-BOOLEAN?) (@INTEGER-WRAPPER INTEGER? SUBTYPE-OF-INTEGER?) (@FLOAT-WRAPPER FLOAT? SUBTYPE-OF-FLOAT?) (@STRING-WRAPPER STRING? SUBTYPE-OF-STRING?) (@CHARACTER-WRAPPER CHARACTER? SUBTYPE-OF-CHARACTER?) (@VERBATIM-STRING-WRAPPER VERBATIM-STRING? SUBTYPE-OF-VERBATIM-STRING?) (@SURROGATE SURROGATE? SUBTYPE-OF-SURROGATE?) (@TYPE TYPE? SUBTYPE-OF-TYPE?) (@SYMBOL SYMBOL? SUBTYPE-OF-SYMBOL?) (@TRANSIENT-SYMBOL TRANSIENT-SYMBOL? SUBTYPE-OF-TRANSIENT-SYMBOL?) (@KEYWORD KEYWORD? SUBTYPE-OF-KEYWORD?) (@CONS CONS? SUBTYPE-OF-CONS?) (@CLASS STELLA-CLASS? SUBTYPE-OF-CLASS?) (@STORAGE-SLOT STORAGE-SLOT? SUBTYPE-OF-STORAGE-SLOT?) (@METHOD-SLOT METHOD-SLOT? SUBTYPE-OF-METHOD-SLOT?) (@ANCHORED-TYPE-SPECIFIER ANCHORED-TYPE-SPECIFIER? SUBTYPE-OF-ANCHORED-TYPE-SPECIFIER?) (@PARAMETRIC-TYPE-SPECIFIER PARAMETRIC-TYPE-SPECIFIER? SUBTYPE-OF-PARAMETRIC-TYPE-SPECIFIER?))) :DOCUMENTATION \"Table of specialized type predicates for various types.\nThese predicates have to be used instead of `isa?', since they also work\nduring bootstrap when only some class objects are defined.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *NUMERIC-TYPE-HIERARCHY* (LIST OF TYPE) (LIST @INTEGER @FLOAT @NUMBER))");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *INLININGMETHODCALL?* BOOLEAN FALSE)");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *MIXIN-IMPLEMENTATION-STYLE* KEYWORD :SECOND-CLASS :DOCUMENTATION \"A keyword describing how mixin classes are handled in\nsingle-inheritance target languages.  The legal values are\n:FIRST-CLASS-WITH-METHOD, which means that variables of a mixin type\nare legal and that slot access on a mixin type is facilitated by\ninherited-down accessor methods and a catch-all method on OBJECT,\n:FIRST-CLASS-WITH-TYPECASE which is similar but replaces the catch-all\nmethod with a function using a TYPECASE, and :SECOND-CLASS, which\nmeans that variables of a mixin type are illegal and no additional\naccessors and catch-all methods are needed.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *MACRO-TABLE* (HASH-TABLE OF SYMBOL CONS) (NEW (HASH-TABLE OF SYMBOL CONS)) :DOCUMENTATION \"Associates macro names with the name and code of\ntheir expander function.\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *MAX-NUMBER-OF-STARTUP-UNITS* INTEGER 60 :DOCUMENTATION \"The maximum number of startup units that can be combined\ninto a single startup function (this avoids the construction of huge startup\nfunctions that would cause too much stress for some wimpy compilers).\")");
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *NATIVE-NAME-TABLE* (HASH-TABLE OF SYMBOL (KEY-VALUE-LIST OF KEYWORD LIST)) (NEW HASH-TABLE))");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *EVALUATIONTREE* OBJECT NULL :PUBLIC? TRUE)");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *EVALUATIONPARENTTREE* OBJECT NULL :PUBLIC? TRUE)");
    }
  }
}

Keyword* KWD_WALK_WARN_ABOUT_UNDEFINED_METHODS = NULL;

Keyword* KWD_WALK_SUPPRESS_WARNINGS = NULL;

Keyword* KWD_WALK_WARN_ABOUT_MISSING_METHODS = NULL;

Keyword* KWD_WALK_USE_HARDCODED_SYMBOLS = NULL;

Keyword* KWD_WALK_USE_COMMON_LISP_STRUCTS = NULL;

Keyword* KWD_WALK_USE_COMMON_LISP_VECTOR_STRUCTS = NULL;

Keyword* KWD_WALK_USE_COMMON_LISP_CONSES = NULL;

Keyword* KWD_WALK_USE_CPP_GARBAGE_COLLECTOR = NULL;

Keyword* KWD_WALK_TRANSLATE_WITH_COPYRIGHT_HEADER = NULL;

Symbol* SYM_WALK_STELLA_WHEN = NULL;

Symbol* SYM_WALK_STELLA_AND = NULL;

Symbol* SYM_WALK_STELLA_DEFINEDp = NULL;

Symbol* SYM_WALK_STELLA_oTRACED_KEYWORDSo = NULL;

Symbol* SYM_WALK_STELLA_PRINT = NULL;

Symbol* SYM_WALK_STELLA_EXISTS = NULL;

Symbol* SYM_WALK_STELLA_KWD = NULL;

Symbol* SYM_WALK_STELLA_IN = NULL;

Symbol* SYM_WALK_STELLA_QUOTE = NULL;

Symbol* SYM_WALK_STELLA_WHERE = NULL;

Symbol* SYM_WALK_STELLA_MEMBp = NULL;

Symbol* SYM_WALK_STELLA_TRACE_IF = NULL;

Symbol* SYM_WALK_STELLA_TRACE_IF_EXPANDER = NULL;

Keyword* KWD_WALK_JAVA = NULL;

Surrogate* SGT_WALK_STELLA_TRANSLATION_UNIT = NULL;

Symbol* SYM_WALK_STELLA_TU_HOME_MODULE = NULL;

Symbol* SYM_WALK_STELLA_THE_OBJECT = NULL;

Symbol* SYM_WALK_STELLA_CATEGORY = NULL;

Symbol* SYM_WALK_STELLA_ANNOTATION = NULL;

Symbol* SYM_WALK_STELLA_AUXILIARYp = NULL;

Symbol* SYM_WALK_STELLA_CODE_REGISTER = NULL;

Symbol* SYM_WALK_STELLA_TRANSLATION = NULL;

Symbol* SYM_WALK_STELLA_REFERENCED_GLOBALS = NULL;

Symbol* SYM_WALK_STELLA_CLASS = NULL;

Symbol* SYM_WALK_STELLA_METHOD = NULL;

Symbol* SYM_WALK_STELLA_MACRO = NULL;

Symbol* SYM_WALK_STELLA_GLOBAL_VARIABLE = NULL;

Keyword* KWD_WALK_COMMON_LISP = NULL;

Keyword* KWD_WALK_CPP = NULL;

Keyword* KWD_WALK_CPP_STANDALONE = NULL;

Keyword* KWD_WALK_IDL = NULL;

Symbol* SYM_WALK_STELLA_PROGN = NULL;

Surrogate* SGT_WALK_STELLA_CONS = NULL;

Symbol* SYM_WALK_STELLA_DEFCLASS = NULL;

Symbol* SYM_WALK_STELLA_DEFSLOT = NULL;

Symbol* SYM_WALK_STELLA_DEFMETHOD = NULL;

Symbol* SYM_WALK_STELLA_DEFUN = NULL;

Symbol* SYM_WALK_STELLA_DEFOPERATOR = NULL;

Symbol* SYM_WALK_STELLA_DEFGLOBAL = NULL;

Symbol* SYM_WALK_STELLA_DEFSPECIAL = NULL;

Symbol* SYM_WALK_STELLA_DEFCONSTANT = NULL;

Symbol* SYM_WALK_STELLA_DEFTYPE = NULL;

Symbol* SYM_WALK_STELLA_DEFMODULE = NULL;

Symbol* SYM_WALK_STELLA_DEFMACRO = NULL;

Symbol* SYM_WALK_STELLA_STARTUP_TIME_PROGN = NULL;

Symbol* SYM_WALK_STELLA_VERBATIM = NULL;

Symbol* SYM_WALK_STELLA_MODULE = NULL;

Symbol* SYM_WALK_STELLA_TYPE = NULL;

Symbol* SYM_WALK_STELLA_PRINT_METHOD = NULL;

Symbol* SYM_WALK_STELLA_SLOT = NULL;

Keyword* KWD_WALK_JAVA_STANDALONE = NULL;

Symbol* SYM_WALK_STELLA_METHOD_CONSTRUCTORp = NULL;

Keyword* KWD_WALK_DEFINE = NULL;

Keyword* KWD_WALK_FINALIZE = NULL;

Keyword* KWD_WALK_WALK = NULL;

Keyword* KWD_WALK_TRANSLATE = NULL;

Surrogate* SGT_WALK_STELLA_VOID = NULL;

Surrogate* SGT_WALK_STELLA_GLOBAL_VARIABLE = NULL;

Surrogate* SGT_WALK_STELLA_TYPE_SPEC = NULL;

Surrogate* SGT_WALK_STELLA_MODULE = NULL;

Symbol* SYM_WALK_STELLA_SLOT_AUXILIARYp = NULL;

Symbol* SYM_WALK_STELLA_METHOD_GLOBALLY_INLINEp = NULL;

Symbol* SYM_WALK_STELLA_LOG_FUNCTION_CALL = NULL;

Symbol* SYM_WALK_STELLA_START_FUNCTION_CALL_LOGGING = NULL;

Symbol* SYM_WALK_STELLA_STOP_FUNCTION_CALL_LOGGING = NULL;

Symbol* SYM_WALK_STELLA_SET_CALL_LOG_BREAK_POINT = NULL;

Symbol* SYM_WALK_STELLA_BREAK_PROGRAM = NULL;

Symbol* SYM_WALK_STELLA_TERMINATE_PROGRAM = NULL;

Symbol* SYM_WALK_STELLA_PO = NULL;

Symbol* SYM_WALK_STELLA_NULLp = NULL;

Symbol* SYM_WALK_STELLA_EQp = NULL;

Symbol* SYM_WALK_STELLA_EQLp = NULL;

Symbol* SYM_WALK_STELLA_l = NULL;

Symbol* SYM_WALK_STELLA_le = NULL;

Symbol* SYM_WALK_STELLA_g = NULL;

Symbol* SYM_WALK_STELLA_ge = NULL;

Symbol* SYM_WALK_STELLA_CAST = NULL;

Surrogate* SGT_WALK_STELLA_UNKNOWN = NULL;

Surrogate* SGT_WALK_STELLA_LISP_CODE = NULL;

Surrogate* SGT_WALK_STELLA_OUTPUT_STREAM = NULL;

Surrogate* SGT_WALK_STELLA_NATIVE_OUTPUT_STREAM = NULL;

Symbol* SYM_WALK_STELLA_INLINE = NULL;

Surrogate* SGT_WALK_STELLA_UNINITIALIZED = NULL;

Keyword* KWD_WALK_SECOND_CLASS = NULL;

Surrogate* SGT_WALK_STELLA_KEY_VALUE_LIST = NULL;

Symbol* SYM_WALK_STELLA_NULL = NULL;

Keyword* KWD_WALK_TYPE = NULL;

Keyword* KWD_WALK_ALIAS = NULL;

Symbol* SYM_WALK_STELLA_ERROR = NULL;

Symbol* SYM_WALK_STELLA_EOL = NULL;

Symbol* SYM_WALK_STELLA_ILLEGAL_EXPRESSION_FLAGGED_BY_THE_TRANSLATOR = NULL;

Symbol* SYM_WALK_STELLA_BAD_SYS = NULL;

Surrogate* SGT_WALK_STELLA_INTEGER_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_INTEGER = NULL;

Surrogate* SGT_WALK_STELLA_FLOAT_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_FLOAT = NULL;

Surrogate* SGT_WALK_STELLA_NUMBER_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_NUMBER = NULL;

Surrogate* SGT_WALK_STELLA_STRING_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_STRING = NULL;

Surrogate* SGT_WALK_STELLA_MUTABLE_STRING_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_MUTABLE_STRING = NULL;

Surrogate* SGT_WALK_STELLA_CHARACTER_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_CHARACTER = NULL;

Surrogate* SGT_WALK_STELLA_BOOLEAN_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_BOOLEAN = NULL;

Surrogate* SGT_WALK_STELLA_FUNCTION_CODE_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_FUNCTION_CODE = NULL;

Surrogate* SGT_WALK_STELLA_METHOD_CODE_WRAPPER = NULL;

Surrogate* SGT_WALK_STELLA_METHOD_CODE = NULL;

Surrogate* SGT_WALK_STELLA_OBJECT = NULL;

Symbol* SYM_WALK_STELLA_INLINE_WRAP_BOOLEAN = NULL;

Symbol* SYM_WALK_STELLA_WRAP_LITERAL = NULL;

Symbol* SYM_WALK_STELLA_INTEGER_TO_BOOLEAN_WRAPPER = NULL;

Symbol* SYM_WALK_STELLA_INTEGER_TO_BOOLEAN = NULL;

Symbol* SYM_WALK_STELLA_MUTABLE_STRING_TO_STRING = NULL;

Symbol* SYM_WALK_STELLA_STRING_TO_MUTABLE_STRING = NULL;

Surrogate* SGT_WALK_STELLA_SYMBOL = NULL;

Symbol* SYM_WALK_STELLA_INTERN_SYMBOL = NULL;

Symbol* SYM_WALK_STELLA_CHARACTER_TO_STRING = NULL;

Symbol* SYM_WALK_STELLA_SYMBOL_NAME = NULL;

Symbol* SYM_WALK_STELLA_INLINE_UNWRAP_BOOLEAN = NULL;

Symbol* SYM_WALK_STELLA_WRAPPER_VALUE = NULL;

Symbol* SYM_WALK_STELLA_NUMBER_WRAPPER_TO_FLOAT = NULL;

Surrogate* SGT_WALK_STELLA_SURROGATE = NULL;

Surrogate* SGT_WALK_STELLA_CLASS = NULL;

Symbol* SYM_WALK_STELLA_SURROGATE_VALUE = NULL;

Surrogate* SGT_WALK_STELLA_INPUT_STREAM = NULL;

Surrogate* SGT_WALK_STELLA_NATIVE_INPUT_STREAM = NULL;

Symbol* SYM_WALK_STELLA_NATIVE_STREAM = NULL;

Symbol* SYM_WALK_STELLA_lXg = NULL;

Surrogate* SGT_WALK_STELLA_SINGLE_FLOAT = NULL;

Symbol* SYM_WALK_STELLA_IDENTITY = NULL;

Surrogate* SGT_WALK_STELLA_DOUBLE_FLOAT = NULL;

Surrogate* SGT_WALK_STELLA_SHORT_INTEGER = NULL;

Surrogate* SGT_WALK_STELLA_LONG_INTEGER = NULL;

Surrogate* SGT_WALK_STELLA_UNSIGNED_SHORT_INTEGER = NULL;

Surrogate* SGT_WALK_STELLA_UNSIGNED_LONG_INTEGER = NULL;

Symbol* SYM_WALK_STELLA_VRLET = NULL;

Symbol* SYM_WALK_STELLA_SYS_CALL_METHOD = NULL;

Symbol* SYM_WALK_STELLA_SYS_CALL_FUNCTION = NULL;

Symbol* SYM_WALK_STELLA_TYPED_SYS = NULL;

Symbol* SYM_WALK_STELLA_SYS_INLINE_CALL = NULL;

Symbol* SYM_WALK_STELLA_SAFE_CAST = NULL;

Surrogate* SGT_WALK_STELLA_NON_OBJECT = NULL;

Surrogate* SGT_WALK_STELLA_WRAPPER = NULL;

Symbol* SYM_WALK_STELLA_LET = NULL;

Symbol* SYM_WALK_STELLA_VOID_SYS = NULL;

Symbol* SYM_WALK_STELLA_SLOT_TYPE_SPECIFIER = NULL;

Symbol* SYM_WALK_STELLA_SYS_FOREACH = NULL;

Symbol* SYM_WALK_STELLA_SETQ = NULL;

Symbol* SYM_WALK_STELLA_SETF = NULL;

Symbol* SYM_WALK_STELLA_SLOT_VALUE = NULL;

Symbol* SYM_WALK_STELLA_SLOT_VALUE_SETTER = NULL;

Symbol* SYM_WALK_STELLA_AREF = NULL;

Symbol* SYM_WALK_STELLA_AREF_SETTER = NULL;

Symbol* SYM_WALK_STELLA_WITH_PROCESS_LOCK = NULL;

Symbol* SYM_WALK_STELLA_LOOP = NULL;

Symbol* SYM_WALK_STELLA_VALUES = NULL;

Symbol* SYM_WALK_STELLA_SPECIAL = NULL;

Symbol* SYM_WALK_STELLA_SYS_SPECIAL = NULL;

Symbol* SYM_WALK_STELLA_RETURN = NULL;

Symbol* SYM_WALK_STELLA_MV_SETQ = NULL;

Symbol* SYM_WALK_STELLA_MV_BIND = NULL;

Symbol* SYM_WALK_STELLA_BREAK = NULL;

Symbol* SYM_WALK_STELLA_CONTINUE = NULL;

Symbol* SYM_WALK_STELLA_WHILE = NULL;

Symbol* SYM_WALK_STELLA_FOREACH = NULL;

Symbol* SYM_WALK_STELLA_FORALL = NULL;

Symbol* SYM_WALK_STELLA_SOME = NULL;

Symbol* SYM_WALK_STELLA_SETOF = NULL;

Symbol* SYM_WALK_STELLA_SET_OF = NULL;

Symbol* SYM_WALK_STELLA_SELECT = NULL;

Symbol* SYM_WALK_STELLA_IF = NULL;

Symbol* SYM_WALK_STELLA_UNLESS = NULL;

Symbol* SYM_WALK_STELLA_CHOOSE = NULL;

Symbol* SYM_WALK_STELLA_OR = NULL;

Symbol* SYM_WALK_STELLA_NOT = NULL;

Symbol* SYM_WALK_STELLA_COND = NULL;

Symbol* SYM_WALK_STELLA_CASE = NULL;

Symbol* SYM_WALK_STELLA_TYPECASE = NULL;

Symbol* SYM_WALK_STELLA_e = NULL;

Symbol* SYM_WALK_STELLA_i = NULL;

Symbol* SYM_WALK_STELLA__ = NULL;

Symbol* SYM_WALK_STELLA_o = NULL;

Symbol* SYM_WALK_STELLA_s = NULL;

Symbol* SYM_WALK_STELLA_BQUOTE = NULL;

Symbol* SYM_WALK_STELLA_PRINT_STREAM = NULL;

Symbol* SYM_WALK_STELLA_PRINT_NATIVE_STREAM = NULL;

Symbol* SYM_WALK_STELLA_NEW = NULL;

Symbol* SYM_WALK_STELLA_ALLOCATE = NULL;

Symbol* SYM_WALK_STELLA_LOCAL_NEW = NULL;

Symbol* SYM_WALK_STELLA_MAKE = NULL;

Symbol* SYM_WALK_STELLA_FUNCALL = NULL;

Symbol* SYM_WALK_STELLA_SYS_CALL_FUNCTION_CODE = NULL;

Symbol* SYM_WALK_STELLA_SYS_CALL_METHOD_CODE = NULL;

Symbol* SYM_WALK_STELLA_THE_CODE = NULL;

Symbol* SYM_WALK_STELLA_SIGNAL_EXCEPTION = NULL;

Symbol* SYM_WALK_STELLA_EXCEPTION_CASE = NULL;

Symbol* SYM_WALK_STELLA_UNWIND_PROTECT = NULL;

Symbol* SYM_WALK_STELLA_SYS_UNWIND_PROTECT = NULL;

Symbol* SYM_WALK_STELLA_SYS_SLOT_VALUE = NULL;

Symbol* SYM_WALK_STELLA_SYS_SLOT_VALUE_SETTER = NULL;

Symbol* SYM_WALK_STELLA_SYS_CALL_METHOD_SETTER = NULL;

Symbol* SYM_WALK_STELLA_SYS_NEW = NULL;

Surrogate* SGT_WALK_STELLA_PROCESS_LOCK_OBJECT = NULL;

Keyword* KWD_WALK_PUBLICp = NULL;

Symbol* SYM_WALK_STELLA_TRUE = NULL;

Keyword* KWD_WALK_SYMBOLS = NULL;

Symbol* SYM_WALK_STELLA_INTERN_RIGID_SYMBOL_WRT_MODULE = NULL;

Symbol* SYM_WALK_STELLA_GET_STELLA_MODULE = NULL;

Symbol* SYM_WALK_STELLA_INTERN_SYMBOL_AT = NULL;

Symbol* SYM_WALK_STELLA_INTERN_SURROGATE_AT = NULL;

Symbol* SYM_WALK_STELLA_INTERN_KEYWORD_AT = NULL;

Symbol* SYM_WALK_STELLA_FALSE = NULL;

Symbol* SYM_WALK_STELLA_NIL = NULL;

Symbol* SYM_WALK_STELLA_NULL_INTEGER = NULL;

Symbol* SYM_WALK_STELLA_NULL_FLOAT = NULL;

Keyword* KWD_WALK_INLINE_ARGUMENT = NULL;

Surrogate* SGT_WALK_STELLA_KEYWORD = NULL;

Keyword* KWD_WALK_PLACE_WHERE_VRLET_WAS = NULL;

Keyword* KWD_WALK_PLACE_WHERE_LAST_EXPRESSION_WAS = NULL;

Keyword* KWD_WALK_DYNAMIC = NULL;

Symbol* SYM_WALK_STELLA_IT = NULL;

Symbol* SYM_WALK_STELLA_ON = NULL;

Symbol* SYM_WALK_STELLA_DYNAMIC_SLOTS = NULL;

Symbol* SYM_WALK_STELLA_KEY = NULL;

Symbol* SYM_WALK_STELLA_DO = NULL;

Symbol* SYM_WALK_STELLA_WRAPPEDVALUE = NULL;

Symbol* SYM_WALK_STELLA_VALUE = NULL;

Keyword* KWD_WALK_BIT = NULL;

Symbol* SYM_WALK_STELLA_DYNAMICSLOTS = NULL;

Symbol* SYM_WALK_STELLA_NEWVALUE = NULL;

Symbol* SYM_WALK_STELLA_FOUNDMATCHINGENTRYp = NULL;

Symbol* SYM_WALK_STELLA_THE_KV_LIST = NULL;

Symbol* SYM_WALK_STELLA_KV_CONS = NULL;

Symbol* SYM_WALK_STELLA_OLDVALUE = NULL;

Keyword* KWD_WALK_CONTEXT_SENSITIVEp = NULL;

Keyword* KWD_WALK_DONT_INHERITp = NULL;

Symbol* SYM_WALK_STELLA_ACCESS_IN_CONTEXT = NULL;

Symbol* SYM_WALK_STELLA_HOME_CONTEXT = NULL;

Keyword* KWD_WALK_COPY_TO_CHILDRENp = NULL;

Symbol* SYM_WALK_STELLA_UPDATE_IN_CONTEXT = NULL;

Symbol* SYM_WALK_STELLA_CS_VALUEp = NULL;

Surrogate* SGT_WALK_STELLA_ARRAY = NULL;

Symbol* SYM_WALK_STELLA_ANY_VALUE = NULL;

Symbol* SYM_WALK_STELLA_VARIABLE_DECLARATION = NULL;

Symbol* SYM_WALK_STELLA_ILLEGAL_VARIABLE = NULL;

Keyword* KWD_WALK_UNBIND_WITH_DESTRUCTORS = NULL;

Keyword* KWD_WALK_UNWIND_PROTECT = NULL;

Surrogate* SGT_WALK_STELLA_METHOD_SLOT = NULL;

Symbol* SYM_WALK_STELLA_OTHERWISE = NULL;

Symbol* SYM_WALK_STELLA_SYMBOL_ID = NULL;

Symbol* SYM_WALK_STELLA_GENERALIZED_SYMBOL = NULL;

Symbol* SYM_WALK_STELLA_BOOLEANp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_BOOLEANp = NULL;

Symbol* SYM_WALK_STELLA_INTEGERp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_INTEGERp = NULL;

Symbol* SYM_WALK_STELLA_FLOATp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_FLOATp = NULL;

Symbol* SYM_WALK_STELLA_STRINGp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_STRINGp = NULL;

Symbol* SYM_WALK_STELLA_CHARACTERp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_CHARACTERp = NULL;

Symbol* SYM_WALK_STELLA_WRAPPERp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_WRAPPERp = NULL;

Surrogate* SGT_WALK_STELLA_VERBATIM_STRING_WRAPPER = NULL;

Symbol* SYM_WALK_STELLA_VERBATIM_STRINGp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_VERBATIM_STRINGp = NULL;

Symbol* SYM_WALK_STELLA_SURROGATEp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_SURROGATEp = NULL;

Surrogate* SGT_WALK_STELLA_TYPE = NULL;

Symbol* SYM_WALK_STELLA_TYPEp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_TYPEp = NULL;

Symbol* SYM_WALK_STELLA_SYMBOLp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_SYMBOLp = NULL;

Surrogate* SGT_WALK_STELLA_TRANSIENT_SYMBOL = NULL;

Symbol* SYM_WALK_STELLA_TRANSIENT_SYMBOLp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_TRANSIENT_SYMBOLp = NULL;

Symbol* SYM_WALK_STELLA_KEYWORDp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_KEYWORDp = NULL;

Symbol* SYM_WALK_STELLA_CONSp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_CONSp = NULL;

Symbol* SYM_WALK_STELLA_STELLA_CLASSp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_CLASSp = NULL;

Surrogate* SGT_WALK_STELLA_STORAGE_SLOT = NULL;

Symbol* SYM_WALK_STELLA_STORAGE_SLOTp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_STORAGE_SLOTp = NULL;

Symbol* SYM_WALK_STELLA_METHOD_SLOTp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_METHOD_SLOTp = NULL;

Surrogate* SGT_WALK_STELLA_ANCHORED_TYPE_SPECIFIER = NULL;

Symbol* SYM_WALK_STELLA_ANCHORED_TYPE_SPECIFIERp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_ANCHORED_TYPE_SPECIFIERp = NULL;

Surrogate* SGT_WALK_STELLA_PARAMETRIC_TYPE_SPECIFIER = NULL;

Symbol* SYM_WALK_STELLA_PARAMETRIC_TYPE_SPECIFIERp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OF_PARAMETRIC_TYPE_SPECIFIERp = NULL;

Symbol* SYM_WALK_STELLA_ISAp = NULL;

Symbol* SYM_WALK_STELLA_SUBTYPE_OFp = NULL;

Surrogate* SGT_WALK_STELLA_GENERALIZED_SYMBOL = NULL;

Symbol* SYM_WALK_STELLA_STRING_EQLp = NULL;

Symbol* SYM_WALK_STELLA_SAFE_PRIMARY_TYPE = NULL;

Surrogate* SGT_WALK_STELLA_STANDARD_OBJECT = NULL;

Surrogate* SGT_WALK_STELLA_LITERAL = NULL;

Symbol* SYM_WALK_STELLA_EQL_TO_BOOLEANp = NULL;

Symbol* SYM_WALK_STELLA_EQL_TO_INTEGERp = NULL;

Symbol* SYM_WALK_STELLA_EQL_TO_FLOATp = NULL;

Symbol* SYM_WALK_STELLA_EQL_TO_STRINGp = NULL;

Symbol* SYM_WALK_STELLA_GET_QUOTED_TREE = NULL;

Symbol* SYM_WALK_STELLA_METHOD_BODY_ARGUMENTp = NULL;

Keyword* KWD_WALK_WARN = NULL;

Keyword* KWD_WALK_ERROR = NULL;

Keyword* KWD_WALK_CONTINUABLE_ERROR = NULL;

Symbol* SYM_WALK_STELLA_CONCATENATE = NULL;

Symbol* SYM_WALK_STELLA_WRAP_BOOLEAN = NULL;

Symbol* SYM_WALK_STELLA_CONS = NULL;

Symbol* SYM_WALK_STELLA_CONS_LIST = NULL;

Symbol* SYM_WALK_STELLA_ALLOCATE_ITERATOR = NULL;

Surrogate* SGT_WALK_STELLA_ARGUMENT_LIST_ITERATOR = NULL;

Symbol* SYM_WALK_STELLA_LENGTH = NULL;

Symbol* SYM_WALK_STELLA_METHOD_VARIABLE_ARGUMENTSp = NULL;

Symbol* SYM_WALK_STELLA_GET_SYM = NULL;

Symbol* SYM_WALK_STELLA_GET_SGT = NULL;

Symbol* SYM_WALK_STELLA_GET_KWD = NULL;

Surrogate* SGT_WALK_STELLA_ARGUMENT_LIST = NULL;

Symbol* SYM_WALK_STELLA_CACHED_INLINABLE_METHOD_BODY = NULL;

Keyword* KWD_WALK_INLINE_REFERENCES = NULL;

Symbol* SYM_WALK_STELLA_INLINE_CALL = NULL;

Symbol* SYM_WALK_STELLA_METHOD_INHERITS_THROUGH = NULL;

Keyword* KWD_WALK_FUNCTION = NULL;

Symbol* SYM_WALK_STELLA_FORWARD_DECLARATIONp = NULL;

Symbol* SYM_WALK_STELLA_METHOD_NATIVEp = NULL;

Symbol* SYM_WALK_STELLA_METHOD_COMMANDp = NULL;

Symbol* SYM_WALK_STELLA_MAIN = NULL;

Keyword* KWD_WALK_METHODS = NULL;

Keyword* KWD_WALK_CLASS = NULL;

Keyword* KWD_WALK_EMBEDDED = NULL;

Symbol* SYM_WALK_STELLA_CLASS_CONSTRUCTOR_CODE = NULL;

Symbol* SYM_WALK_STELLA_CLASS_AUXILIARY_METHODS = NULL;

Symbol* SYM_WALK_STELLA_AUXILIARY_METHOD_REASON = NULL;

Keyword* KWD_WALK_FIRST_CLASS_WITH_TYPECASE = NULL;

Keyword* KWD_WALK_AUXILIARYp = NULL;

Symbol* SYM_WALK_STELLA_PRINT_FORM = NULL;

Symbol* SYM_WALK_STELLA_PRINT_OBJECT = NULL;

Symbol* SYM_WALK_STELLA_SELF = NULL;

Symbol* SYM_WALK_STELLA_STREAM = NULL;

Symbol* SYM_WALK_STELLA_NATIVE_OUTPUT_STREAM = NULL;

Symbol* SYM_WALK_STELLA_STRINGIFY = NULL;

Keyword* KWD_WALK_CLASSES = NULL;

Keyword* KWD_WALK_FINALIZE_CLASSES = NULL;

Symbol* SYM_WALK_STELLA_FINALIZE_CLASSES = NULL;

Keyword* KWD_WALK_FINALIZE_METHODS = NULL;

Symbol* SYM_WALK_STELLA_FINALIZE_SLOTS = NULL;

Symbol* SYM_WALK_STELLA_CLEANUP_UNFINALIZED_CLASSES = NULL;

Symbol* SYM_WALK_STELLA_DEFINE_EXTERNAL_SLOT_FROM_STRINGIFIED_SOURCE = NULL;

Symbol* SYM_WALK_STELLA_OBJECT = NULL;

Symbol* SYM_WALK_STELLA_METHOD_LISP_MACROp = NULL;

Symbol* SYM_WALK_STELLA_STORE_MACRO = NULL;

Keyword* KWD_WALK_UNBOUND_SPECIAL_VARIABLE = NULL;

Keyword* KWD_WALK_GLOBALS = NULL;

Symbol* SYM_WALK_STELLA_DEFINE_STELLA_GLOBAL_VARIABLE_FROM_STRINGIFIED_SOURCE = NULL;

Symbol* SYM_WALK_STELLA_DEFINE_STELLA_TYPE_FROM_STRINGIFIED_SOURCE = NULL;

Keyword* KWD_WALK_FINAL = NULL;

Symbol* SYM_WALK_STELLA_CURRENT_STARTUP_TIME_PHASEp = NULL;

Keyword* KWD_WALK_MODULES = NULL;

Symbol* SYM_WALK_STELLA_WITHIN_MODULE = NULL;

Symbol* SYM_WALK_STELLA_oSTELLA_MODULEo = NULL;

Symbol* SYM_WALK_STELLA_oSTARTUP_TIME_PHASEo = NULL;

Symbol* SYM_WALK_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

Symbol* SYM_WALK_STELLA_FUNCTION = NULL;

Symbol* SYM_WALK_STELLA_LOOKUP_FUNCTION = NULL;

Symbol* SYM_WALK_STELLA_SYS_SIGNAL = NULL;

Surrogate* SGT_WALK_STELLA_STELLA_ROOT_EXCEPTION = NULL;

Symbol* SYM_WALK_STELLA_SYS_HANDLER_CASE = NULL;

Surrogate* SGT_WALK_STELLA_NATIVE_EXCEPTION = NULL;

Symbol* SYM_WALK_STELLA_SYS_HANDLE_EXCEPTION = NULL;

Keyword* KWD_WALK_OTHERWISE = NULL;

Symbol* SYM_WALK_STELLA_INITIAL_ELEMENT = NULL;

Symbol* SYM_WALK_STELLA_SIZE = NULL;

Symbol* SYM_WALK_STELLA_STORAGE_SLOT = NULL;

Surrogate* SGT_WALK_STELLA_CODE = NULL;

Keyword* KWD_WALK_METHOD = NULL;

Symbol* SYM_WALK_STELLA_METHOD_CODE = NULL;

Surrogate* SGT_WALK_STELLA_COMPOUND_TYPE_SPECIFIER = NULL;

Symbol* SYM_WALK_STELLA_DOCUMENTATION = NULL;

Symbol* SYM_WALK_STELLA_EXPRESSION = NULL;

Symbol* SYM_WALK_STELLA_IGNORE = NULL;

Symbol* SYM_WALK_STELLA_oIGNORETRANSLATIONERRORSpo = NULL;

Symbol* SYM_WALK_STELLA_CL_INCREMENTALLY_TRANSLATE = NULL;

Symbol* SYM_WALK_STELLA_REGISTER_NATIVE_NAME = NULL;

Symbol* SYM_WALK_STELLA_REST = NULL;

Symbol* SYM_WALK_STELLA_NTH_REST = NULL;

Symbol* SYM_WALK_STELLA_NTH = NULL;

Symbol* SYM_WALK_STELLA_ARGUMENTS = NULL;

Symbol* SYM_WALK_STELLA_RESULT = NULL;

Symbol* SYM_WALK_STELLA_METHOD_EVALUATE_ARGUMENTSp = NULL;

Symbol* SYM_WALK_STELLA_EVALUATOR_WRAPPER_CODE = NULL;

Symbol* SYM_WALK_STELLA_STARTUP_WALK = NULL;

} // end of namespace stella
